/*====================================================================================================================*
 * FILE:                                     C D E V _ G L B . H                    
 *====================================================================================================================*
 *
 *   COPYRIGHT 2003 - 2017 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED
 *   ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH THE INCLUSION
 *   OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY OTHER COPIES THEREOF
 *   MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE
 *   AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS, INC. 
 *
 *====================================================================================================================*
 *
 * FUNCTION:    Application Programmer's Interface for a CEI-x30 ARINC device.
 *              Header File for API Global Variables and Data Structures
 *
 *--------------------------------------------------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date    By   Comments
 * -------- ----- ----------------------------------------------------------------------------------------------------
 * 10/21/05  BDW  Added support for the new CEI-430-supported API channel types
 *                   DIFFERENTIAL_IN_CHANNEL and DIFFERENTIAL_OUT_CHANNEL
 *                   to the channelTypeCrossRef[][] table
 * 09/25/06  BDW  Added prototype for download routine a429_utl_program_xilinx_9056_32bit
 * 10/29/07  BDW  Added prototype for open error handling, cei_get_open_error
 * 02/18/08  BDW  Removed the data struct array channelIdent[]; structure members
 *                   relocated in the new data structure array "deviceGlobal"
 *                Removed the following global arrays, relocated as structure
 *                   members in the new data structure array "deviceGlobal":
 *                      alternateJconfiguration      
 *                      deviceTimeTagReference       
 *                      deviceX20CompatibleResolution
 *                      userProgrammableIrigOffset   
 *                      currentIrigTimeUsec
 *                      irigReferenceTimeTag
 *                Removed the variable multiThreadProtect, relocated it as a 
 *                   structure member in the data structure array "device".
 *                Added prototypes for multi-process support: 
 *                   cei_open_shared_lock and cei_map_shared_memory
 *                Moved the remaining API local data arrays to CDEV_API.C.
 * 04/15/09  BDW  Incorporated support for the CEI-x30 Enhanced Firmware Design.
 *                adding prototypes for low-level hardware interrupt routines.
 * 06/16/09  BDW  Added prototype for cei_open_preload_lock.
 * 06/23/09  BDW  Added prototype for a429_utl_enh.
 * 07/01/09  BDW  Added prototype for cei_delay_milliseconds.
 * 12/20/13  BDW  Modified the prototype for cei_get_open_error when the directive
 *                INTEGRITY_PCI_PPC is defined to match the function in CDEV_INT.C
 * 01/08/14  BDW  Modified the data type for the Linux build definition of ARINC_LOCAL_TYPE 
 *                member sharedMemoryHandle and the corresponding cei_map_shared_memory parameter
 *                blockHandle used to acquire the handle, so they are integer & pointer to and not void
 * 06/11/14  BDW  Applied the 'U' suffix to unsigned integer constant values
 * 01/07/15  BDW  Added prototype for the utility routine a429_utl_configure_programmable_channel
 * 07/01/15  BDW  Added prototype for the utility routine a429_utl_initialize_device
 * 02/04/20  BDW  Added support for Integrity x86 via the directive INTEGRITY_PCI_X86 with cei_get_open_error
 *====================================================================================================================*/ 
#ifndef CEIDEV_GLB_H
#define CEIDEV_GLB_H
#include "cdev_api.h"

#define AR_USE_PROGRAMMED_TIMETAG_MODE      AR_TIMETAG_LAST_OPTION
#define AR_BIT_TIMER_DEVIATION              1000U
#define AR_WRAP_TEST_MSG_COUNT              20U
#define AR_STATIC                           0U
#define AR_SHIFT                            1U
#define AR_ADD                              2U
#define AR_NUM_PATTERNS                     6
#define AR_NUM_UNUSED_SRAM_LOCATIONS        6
#define AR_LOWER                            0
#define AR_UPPER                            1
#define LOCK_IDENT                          (CDEV_BOARD_TYPE)board

/* IRIG Adjustment Logic, DAC Count, and Delay Constants */
#define APPLY_ADJUSTMENT_AND_TEST          -1
#define NO_PREV_SIGNAL_DET_FINE_ADJUST      1
#define PREV_SIGNAL_DETECT_FINE_ADJUST      0

/* the following constants are used with the function a429_utl_wrap_test() */
#define LOCAL_EXTERNAL_WRAP                 0U
#define LOCAL_INTERNAL_WRAP                 1U

/* channel access validation definitions used with the function a429_utl_validate_channel_request() */
#define ARU_GET_DATA                        34U
#define ARU_PUT_DATA                        35U
#define ARU_FILTER_LABEL                    36U
#define ARU_ENHANCED_FILTER_LABEL           37U
#define ARU_ADD_SCHEDULED_MSG               39U

#define IRIG_DAC_MAX_VALUE                  255U
#define IRIG_DAC_ADJUST_INITIAL_COUNT       128U
#define COURSE_ADJUST_INITIAL_ADJUST_COUNT  10U
#define INITIAL_SIGNAL_DETECT_ADJUST        10U
#define INITIAL_SIGNAL_DETECT_NO_SIG_ADJUST 20U
#define COURSE_ADJUST_MIN_DAC_COUNT         4U
#define FINE_ADJUST_DAC_COUNT               1U
#define COURSE_ADJUST_DAC_COUNT             2U
#define SIGNAL_DETECT_THRESHOLD_CUSHION     1U
#define SIGNAL_LOST_THRESHOLD_CUSHION       2U

#define ADJUST_DELAY_FOR_TIME_SAMPLE        2000U
#define ADJUST_DELAY_FOR_EDGE_DETECT        5U
#define ADJUST_DELAY_30_MILLISECONDS        30U

#define INTERRUPT_QUEUE_SIZE    2048
/*========================================================================* 
 * CEI_INT_QUEUE_TYPE is used to maintain the interrupt queue entries for:
 * 1) head/tail pointers for access from a separate processing thread.
 * 2) last queue location read by the API ISR cei_utl_interrupt_handler
 * 3) last queue location read by the host via ar_interrupt_queue_read
 *========================================================================*/
typedef struct CEI_INT_QUEUE {
   CEI_UINT32 intQueueHeadPtr;
   CEI_UINT32 intQueueTailPtr;
   CEI_UINT32 intQueueEntries[INTERRUPT_QUEUE_SIZE];
   CEI_UINT32 lastReadIntQueueLocation;
   CEI_UINT32 lastApplicationReadIntQueueLocation;
} CEI_INT_QUEUE_TYPE;

/* Initialization cases for the big switch in ar_initialize_device and ar_init_slave */
typedef enum {
   INIT_DISABLE = 0U,
   INIT_TEST_SRAM = 1U,
   INIT_CLEAR_SCHED_TABLE = 2U,
   INIT_CLEAR_SNAPSHOT = 3U,
   INIT_RESET_TIMER = 4U,
   INIT_CONFIG_RX = 5U,
   INIT_FLUSH_TX = 6U,
   INIT_CONFIG_TX = 7U,
   INIT_CLEAR_RX_FILTER = 8U,
   INIT_717 = 9U,
   INIT_ENABLE_FLUSH_RX = 10U,
   INIT_DISABLE_AGAIN = 11U,
   INIT_DAC_IRIG = 12U,
   INIT_RESTORE = 13U,
   INIT_TERMINATE = 14U
} initProgressType;

/* Define the basis for a time delay using the system clock resolution */
#if defined (LABVIEW_RT) || defined (_WIN32)    
   #define TIMER_CLOCK_TICKS_PER_SECOND CLOCKS_PER_SEC
#elif defined (VXW_PCI_PPC) || defined (VXW_PCI_X86)  || defined (_LINUX_X86_) || defined (_LINUX_PPC_) || defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
   #define TIMER_CLOCK_TICKS_PER_SECOND 1000
#endif

/*===================================================================================================================* 
 * Macros for general board index and board initialized verification conditionals, required for most API routines
 *===================================================================================================================*/
#define VALID_BOARD_AND_INIT (board < MAX_BOARDS) && (deviceGlobal[board] != NULL) && (deviceGlobal[board]->boardInitialized)
#define VALID_BOARD_AND_INIT_II (board >= MAX_BOARDS) || (deviceGlobal[board] == NULL) || !(deviceGlobal[board]->boardInitialized)

#ifndef IN_LNX_SOURCE
/*===================================================================================================================* 
 *===================================================================================================================* 
 *                                                                                                                   *
 *                                   API Local (Private) Function prototypes                                         *
 *                                                                                                                   *
 *===================================================================================================================*
 *===================================================================================================================*/

/*===================================================================================================================*
 * ENTRY POINT:            A 4 2 9 _ U T L _ A C Q U I R E _ F O R M A T T E D _ T I M E _ T A G
 *===================================================================================================================*
 *
 * FUNCTION:    Read and potentially convert the active receive buffer 
 *              or IRIG time-stamp to the specified format
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board       -- (input) Board number (0 to 127) 
 *
 *              CEI_UINT32 chanMemOffset    -- (input) Offset into channel memory area
 *                                                     to read the time-tag
 *
 *              CEI_UINT32 timeSource       -- (input) Source (timer or receiver FIFO)
 *                                                     AR_TIMER_REGISTER_SET
 *                                                     AR_RECEIVER_FIFO_TIMETAG
 *                                                     AR_BYPASS_TIMER_READ
 *                                              
 *              CEI_UINT32 timeTagFormat    -- (input) Selects the format of the time-tag value returned:
 *                                                     AR_TIMETAG_EXT_IRIG_64BIT
 *                                                     AR_TIMETAG_INT_USEC_64BIT
 *                                                     AR_TIMETAG_INT_20USEC_32BIT *
 *                                                     AR_TIMETAG_INT_MSEC_32BIT   *
 *                                                     AR_TIMER_X20_COMPAT_32BIT   *
 *                                                     AR_USE_PROGRAMMED_TIMETAG_MODE
 *                                                     AR_LOG_EXT_IRIG_SAMPLE
 *                                                     AR_TIMETAG_REF_IRIG_64BIT
 *                                                     AR_TIMETAG_SYNC_1553_CH1
 *                                                     AR_TIMETAG_SYNC_1553_CH2
 *                                                     AR_TIMETAG_SYNC_1553_CH3
 *                                                     AR_TIMETAG_SYNC_1553_CH4
 *                                              
 *              CEI_UINT32 mergedMode       -- (input) buffering mode (merged/dedicated)
 *                                                     When mergedMode < CEIDEV_MAX_CHANNELS, merged mode is active 
 *                                                     and then if a 32-bit time-stamp is requested, the channel 
 *                                                     number is inserted in the upper 5 bits of the time-stamp
 *                                              
 *              pAR_TIMETAG_TYPE timeTagRef -- (output) 64-bit time-tag structure
 *                                              
 * RETURN VAL:  CEI_INT32    -- ARS_NORMAL      Success
 *                              ARS_INVBOARD    Invalid or uninitialized board
 *                              ARS_INVARG      Invalid timeSource or timeTagFormat
 *
 * GLOBALS:     device
 *              deviceGlobal
 * 
 * DESCRIPTION: This routine assumes the desired receive channel has been selected by the calling routine.  It reads 
 *              the 64-bit microsecond resolution time-tag from the receive buffer, then reformats it based on either 
 *              the requested resolution and size or the actively selected resolution/size.
 * 
 *===================================================================================================================*/ 
CEI_INT32 a429_utl_acquire_formatted_time_tag
(
   CDEV_BOARD_TYPE board,                                                /* (input)  Board number (0 to 127)         */
   CEI_UINT32 chanMemOffset,                                             /* (input)  Offset into channel memory area */
   CEI_UINT32 timeSource,                                                /* (input)  Source (timer or receiver FIFO) */
   CEI_UINT32 timeTagFormat,                                             /* (input)  Format of the time-tag returned */
   CEI_UINT32 mergedMode,                                                /* (input)  Buffering mode merged/dedicated */
   pAR_TIMETAG_TYPE timeTagRef                                           /* (output) 64-bit time-tag structure       */
);

/*===================================================================================================================*
 * ENTRY POINT:         A 4 2 9 _ U T L _ C O N F I G U R E _ P R O G R A M M A B L E _ C H A N N E L
 *===================================================================================================================*
 *
 * FUNCTION:    Configure a programmable channel as either a transmitter or receiver
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board  -- (input) Board number (0 to 127) 
 *              CEI_UINT32 channel     -- (input) programmable channel index (0 to 127)
 *              CEI_UINT32 channelType -- (input) desired configuration for channel
 *                                                ARU_RECEIVER    0
 *                                                ARU_TRANSMITTER 1
 *                                                ARU_BOTH        2
 *                                                ARU_DISABLE     3
 * 
 * RETURN VAL:  CEI_INT32    -- ARS_NORMAL     Success updating or already done
 *                              ARS_INVHARVAL  The channel parameter was an invalid channel, or the channel was not 
 *                                             programmable, or the channelType parameter was neither ARU_RECEIVER 
 *                                             or ARU_TRANSMITTER.
 * GLOBALS:     device
 *              deviceGlobal
 *
 * DESCRIPTION: This function is invoked from ar_set_device_config when the item parameter 
 *              ARU_CONFIG_PROGRAMMABLE_CHAN is supplied.  It configures a programmable channel pair such that the 
 *              I/O pins and respective channel functions as a transmitter or receiver, both transmitter and receiver, 
 *              or both disabled.
 *
 *              Note the channel parameter is not indexed by receive channel... it is indexed by programmable 
 *              channel, typically the transmit channel index.
 *
 *===================================================================================================================*/ 
CEI_INT32 a429_utl_configure_programmable_channel 
(
   CDEV_BOARD_TYPE board,                                                  /* Board number (0 to 127)                */
   CEI_UINT32 channel,                                                     /* programmable channel index (0 to 127)  */
   CEI_UINT32 channelType                                                  /* desired configuration for channel      */
);

/*===================================================================================================================*
 * ENTRY POINT:          A 4 2 9 _ U T L _ D E T E R M I N E _ C H A N N E L _ S U P P O R T
 *===================================================================================================================*
 *
 * FUNCTION:    Determine the channel type for each channel available on the specified device
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input) Board number (0 to 127) 
 * 
 * RETURN VAL:  N/A
 * 
 * GLOBALS:     device
 *              deviceGlobal
 *
 * DESCRIPTION: This function determines the number of channels installed on the board for each channel/protocol 
 *              type detected.
 *
 *===================================================================================================================*/ 
CEI_VOID  a429_utl_determine_channel_support
(
   CDEV_BOARD_TYPE board                                                        /* board to reference (0 to 127)     */
);

/*===================================================================================================================*
 * ENTRY POINT:                 A 4 2 9 _ U T L _ I N I T I A L I Z E _ D E V I C E
 *===================================================================================================================*
 *
 * FUNCTION:    Initialize the device to a ready state
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input) board to reference (0 to 127)
 *              CEI_UINT32 powerOnState -- (input) state of calling function
 *                                                 TRUE  - power on initialization
 *                                                 FALSE - request to reset device
 * 
 * RETURN VAL:  CDEV_API_RET_TYPE -- ARS_NORMAL       Success
 *                                   ARS_BAD_STATIC   Register read/write/verify failure
 *                                   ARS_MEMWRERR     Memory test failure
 *                                   ARS_INVBOARD     Invalid board parameter
 *                                   ARS_BOARD_MUTEX  Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     device
 *              deviceGlobal
 *
 * DESCRIPTION: This routine initializes the device I/O configuration registers to a state ready for use by the host 
 *              application.
 *
 *===================================================================================================================*/ 
CDEV_API_RET_TYPE a429_utl_initialize_device
(
   CDEV_BOARD_TYPE  board,                                            /* board to reference (0 to 127)               */
   CEI_UINT32       powerOnState                                      /* state of calling function                   */
);

/*===================================================================================================================*
 * ENTRY POINT:                A 4 2 9 _ U T L _ I N I T I A L I Z E _ G L O B A L _ D A T A
 *===================================================================================================================*
 *
 * FUNCTION:    Function that initializes most global data
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input) board to reference (0 to 127)
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     device
 *
 * DESCRIPTION: This routine is used to initialize global data structures and arrays shared between the API and the 
 *              operating system specifically-sized arrays based on MAX_BOARDS.
 *
 *===================================================================================================================*/ 
CEI_VOID  a429_utl_initialize_global_data 
(
   CDEV_BOARD_TYPE board                                                        /* board to reference (0 to 127)     */
);

/*===================================================================================================================*
 * ENTRY POINT:                       A 4 2 9 _ U T L _ M E M O R Y _ T E S T
 *===================================================================================================================*
 *
 * FUNCTION:    Perform a memory test of the CEI-x30 SRAM device
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board -- (input) Board number (0 to 127) 
 * 
 *              CEI_UINT32 testType   -- (input) size of memory test to execute
 *                                               FULL_SRAM   0 test all SRAM
 *                                               UNUSED_SRAM 1 test only unused SRAM (Tx Chan 16-31)
 *                                               SELECT_SRAM 2 test a portion of SRAM
 * 
 *              CEI_UINT32 testOption -- (input) when the testType is SELECT_SRAM, the testOption parameter specifies 
 *                                               which block of SRAM will be verified. Block select is based on a  
 *                                               length of 512 locations starting at SRAM address = 512*testOption
 * 
 * RETURN VAL:  CDEV_API_RET_TYPE -- ARS_NORMAL     Normal succesful completion
 *                                   ARS_INVARG     Invalid memory test request
 *                                   ARS_MEMWRERR   Memory test failure
 * GLOBALS:     device
 *              deviceGlobal
 *
 * DESCRIPTION: This function performs SRAM memory tests, either on a small unused portion (active processing test) 
 *              a selected portion of, or on the entire SRAM (destructive test that should only be executed when the 
 *              board is disabled, duration is six seconds to test all 2Mb of SRAM).
 *
 *===================================================================================================================*/ 
CDEV_API_RET_TYPE a429_utl_memory_test 
(
   CDEV_BOARD_TYPE board,                                                     /* board to reference (0 to 127)       */
   CEI_UINT32 testType,                                                       /* stype of memory test to execute     */
   CEI_UINT32 testOption                                                      /* SELECT_SRAM block identifier        */
);

/*===================================================================================================================*
 * ENTRY POINT:                      A 4 2 9 _ U T L _ R E A D _ M E S S A G E
 *===================================================================================================================*
 *
 * FUNCTION:    Read the next data and 64-bit time-tag from a receive FIFO,
 *              returns the extended time-tag structure supporting IRIG.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input)  Board number (0 to 127) 
 *              CDEV_CHAN_TYPE* channel -- (input)  recv channel to reference (0 to 255)
 *                                         (output) in merged mode, the channel this data was actually received on
 *              pCEI_UINT32 data        -- (output) location to store received data
 *
 *              pAR_TIMETAG_TYPE timeTagRef
 *                                      -- (output) location for time-tag struct
 *
 * RETURN VAL   CDEV_API_RET_TYPE ARS_GOTDATA           Success
 *                                ARS_INVARG            NULL data parameter value
 *                                ARS_BOARD_MUTEX       Access to the Board Lock timed-out/failed
 *                                ARS_NODATA            No data availalble
 *                                ARS_BAD_MESSAGE       An invalid length msg was received
 *                                ARS_RX_BUFFER_OVERRUN A receive buffer overrun was detected
 *
 * GLOBALS:     device
 *              deviceGlobal
 * 
 * DESCRIPTION: This routine reads the most recently received entry (data and timetag) from the specified receive 
 *              channel FIFO.  The channel number passed to this routine corresponds to the hardware channel 
 *              assignment on the device.
 *
 *===================================================================================================================*/ 
CDEV_API_RET_TYPE a429_utl_read_message
(
   CDEV_BOARD_TYPE board,                                               /* (input)  board to reference (0 to 127)    */
   CDEV_CHAN_TYPE * channel,                                            /* (input)  recv channel to ref (0 to 255)   */
   pCEI_UINT32 data,                                                    /* (output) location to store received data  */
   pAR_TIMETAG_TYPE timeTagRef                                          /* (output) location to store data time-tags */
);

/*===================================================================================================================*
 * ENTRY POINT:           A 4 2 9 _ U T L _ V A L I D A T E _ C H A N N E L _ R E Q U E S T 
 *===================================================================================================================*
 *
 * FUNCTION:    Validate the channel / item combination specified by the
 *              host application.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input) board to reference (0 to 127)
 *              CEI_UINT32 channel      -- (input) h/w channel select value
 *              CDEV_PARM_SSI_TYPE item -- (input) config register attribute to check
 * 
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL     Normal succesful completion.
 *                                ARS_INVHARVAL  Item was invalid or channel not supported on this device
 *                                ARS_INVBOARD   Board parameter value invalid or board has not been initialized
 * 
 * GLOBALS:     deviceGlobal
 *
 * DESCRIPTION: This function assures the provided channel and configuration item are valid for the specified 
 *              board configuration.
 *
 *===================================================================================================================*/ 
CDEV_API_RET_TYPE a429_utl_validate_channel_request 
(
   CDEV_BOARD_TYPE board,                                                     /* board to reference (0 to 127)       */
   CEI_UINT32 channel,                                                        /* h/w channel select value            */
   CDEV_PARM_SSI_TYPE item                                                    /* config register attribute to check  */
);

/*===================================================================================================================*
 * ENTRY POINT:                           A 4 2 9 _ U T L _ W R A P _ T E S T
 *===================================================================================================================*
 *
 * FUNCTION:    Perform an internal wrap test using the specified number of ARINC 429 messages on each transmit-
 *              receive channel pairs installed on the device.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board  -- (input) board to reference (0 to 127)
 * 
 *              CEI_UINT32 wrap        -- (input) type of wrap to enable: 
 *                                                LOCAL_INTERNAL_WRAP  for internal wrap 
 *                                                LOCAL_EXTERNAL_WRAP for external wrap
 * 
 *              CEI_UINT32 msgCount    -- (input) number of messages to transmit 
 *                                                limited to AR_WRAP_TEST_MSG_COUNT
 * 
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL          Normal succesful completion.
 *                                ARS_INVHARVAL       Item was invalid or channel not supported on this device
 *                                ARS_INVBOARD        Board parameter value invalid or board has not been initialized
 *                                ARS_BOARD_MUTEX     Access to the Board Lock timed-out/failed
 *                                ARS_WRAP_DROP_FAIL  Wrap Data Missing
 *                                ARS_WRAP_DATA_FAIL  Wrap Data Pattern Mismatch
 *                                ARS_WRAP_FLUSH_FAIL Unknown External data received
 *                                ARS_XMITOVRFLO      Transmit buffer overrun has occurred
 * GLOBALS:     device
 *              deviceGlobal
 *
 * DESCRIPTION: This function performs an internal wrap test based on a specified number of ARINC 429 messages on  
 *              each transmit/receive channel pair installed on the device.  Configuration parameters are set for  
 *              no parity and 100Kbps.
 *
 *===================================================================================================================*/ 
CDEV_API_RET_TYPE a429_utl_wrap_test 
(
   CDEV_BOARD_TYPE board,                                             /* (input)  board to reference (0 to 127)      */
   CEI_UINT32 wrap,                                                   /* (input)  type of wrap INTERNAL/EXTERNAL     */
   CEI_UINT32 msgCount                                                /* (input)  number of messages to transmit     */
);

#ifdef INCLUDE_ARINC_717
/*===================================================================================================================*
 * ENTRY POINT:                A 7 1 7 _ U T L _ I N I T I A L I Z E _ G L O B A L _ D A T A
 *===================================================================================================================*
 *
 * FUNCTION:    Function that initializes the ARINC 717 API stored sync words
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board -- (input) board to reference (0 to 127)
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     a717syncWords[][]
 *
 * DESCRIPTION: This routine initializes the internal API sub-frame sync word set (4) for the specified device.
 *
 *===================================================================================================================*/ 
CEI_VOID  a717_utl_initialize_global_data 
(
   CDEV_BOARD_TYPE board                                                        /* board to reference (0 to 127)     */
);
#endif

#ifdef INCLUDE_IRIG
/*===================================================================================================================*
 * ENTRY POINT:                A 4 2 9 _ U T L _ A D J U S T _ I R I G _ T H R E S H O L D
 *===================================================================================================================*
 *
 * FUNCTION:    Performs a thorough adjustment of the IRIG DAC for an external input IRIG signal, using valid 
 *              IRIG-B 32-bit time reception for verification of signal presence.
 *
 * DESCRIPTION: This routine sets the DAC to Vmin + (0.825 * (Vmax-Vmin)).  
 *              Vmin is the lower peak value and Vmax is the upper peak value.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board        -- (input)  Board number (0 to 127) 
 *              CEI_INT32 displayDacSettings -- (input)  when set enables DAC adjustment print to console
 * 
 * GLOBALS:     deviceGlobal[]
 *
 * RETURN VAL:  status   --  Any value from 0 to x7FFFFFFF is the DAC value
 *                           ARS_FAILURE (-1) Any issue encountered with incoming IRIG signal level
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_adjust_irig_threshold 
(
   CDEV_BOARD_TYPE board,                                             /* (input)  board to reference (0 to 127)      */
   CEI_INT32 displayDacSettings                         /* (input)  when set enables DAC adjustment print to console */
);

/*===================================================================================================================*
 * ENTRY POINT:           A 4 2 9 _ U T L _ Q U I C K _ A D J U S T _ I R I G _ T H R E S H O L D
 *===================================================================================================================*
 *
 * FUNCTION:    Performs a quick adjustment of the IRIG DAC for an external input IRIG signal, using signal edge 
 *              detection for verification of signal presence.
 *
 * DESCRIPTION: This routine sets the DAC to Vmin + (0.825 * (Vmax-Vmin)).  
 *              Vmin is the lower peak value and Vmax is the upper peak value.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board        -- (input)  Board number (0 to 127) 
 *              CEI_INT32 displayDacSettings -- (input)  when set enables DAC adjustment print to console
 *
 * GLOBALS:     deviceGlobal[]
 *
 * RETURN VAL:  CEI_INT32    -- Any value from 0 to x7FFFFFFF is the DAC value
 *                              ARS_FAILURE (-1) Any issue encountered with incoming IRIG signal level
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_quick_adjust_irig_threshold
(
   CDEV_BOARD_TYPE board,                                             /* (input)  board to reference (0 to 127)      */
   CEI_INT32 displayDacSettings                         /* (input)  when set enables DAC adjustment print to console */
);

/*===================================================================================================================*
 * ENTRY POINT:                    A 4 2 9 _ U T L _ I S _ I R I G _ C A L I B R A T E D
 *===================================================================================================================*
 *
 * FUNCTION:    Determine if the given board is synched to an IRIG signal.
 *
 * DESCRIPTION: Check if the IRIG base time has been updated in the last one second.  If it has, return; otherwise, 
 *              wait a second for the reference IRIG time-tag to update.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board  -- (input)  Board number (0 to 127) 
 * 
 * GLOBALS:     deviceGlobal[]
 *
 * RETURN VAL:  TRUE or ARS_NORMAL indicates valid IRIG signal reception
 *              ARS_INVHARVAL or other non-zero value indicates bad API status
 *              FALSE or 0 indicates failure to detect valid IRIG input
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_is_irig_calibrated (CDEV_BOARD_TYPE board);

/*===================================================================================================================*
 * ENTRY POINT:                    A 4 2 9 _ U T L _ F I N D _ D A C _ T H R E S H O L D
 *===================================================================================================================*
 *
 * FUNCTION:    Given a threshold type (lower or upper) and the bounds of the threshold search, find the DAC setting 
 *              at which the IRIG signal is lost, then return a DAC value just under that value.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board               -- (input)  Board number (0 to 127) 
 *              CEI_INT32 useActualTimeSample,      -- (input)  IRIG edge det. vs time cmpare      
 *              CEI_INT32 displayDacSettings,       -- (input)  conditional flag printf invoc      
 *              CEI_INT32 thresholdType,            -- (input)  find AR_LOWER or AR_UPPER threshold
 *              CEI_UINT32 voltageLimit,            -- (input)  outer bounds for the search        
 *              CEI_UINT32 initialDacAdjustment,    -- (input)  initial DAC increment              
 *              pCEI_UINT32 finalDacValue,          -- (output) dynamic DAC value during test      
 *              pCEI_UINT32 dacBoudaryCushionValue  -- (output) minor adj to final DAC value       
 *
 * GLOBALS:     device[]
 *
 * RETURN VAL:  CEI_INT32 -- ARS_NORMAL       threshold was detected
 *                           ARS_FAILURE (-1) Any issue encountered with incoming IRIG signal level
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_find_dac_threshold 
(
   CDEV_BOARD_TYPE board,                                            /* (input)  board to reference (0 to 127)       */
   CEI_INT32 useActualTimeSample,                                    /* (input)  IRIG edge det. vs time cmpare       */
   CEI_INT32 displayDacSettings,                                     /* (input)  conditional flag printf invoc       */
   CEI_INT32 thresholdType,                                          /* (input)  find AR_LOWER or AR_UPPER threshold */ 
   CEI_UINT32 voltageLimit,                                          /* (input)  outer bounds for the search         */
   CEI_UINT32 initialDacAdjustment,                                  /* (input)  initial DAC increment               */
   pCEI_UINT32 finalDacValue,                                        /* (output) dynamic DAC value during test       */
   pCEI_UINT32 dacBoudaryCushionValue                                /* (output) minor adj to final DAC value        */
);
#endif

#ifdef INCLUDE_MSG_SCHEDULER
/*===================================================================================================================*
 * ENTRY POINT:            A 4 2 9 _ U T L _ S T O R E _ M E S S A G E _ S C H E D U L E R _ E N T R Y
 *===================================================================================================================*
 *
 * FUNCTION:    Define or modify a single message scheduler table entry
 *
 * PARAMETERS:  CEI_INT32 modifyEntryFlag -- (input) indicates whether to create a new scheduled message 
 *                                                   or modify an existing message: FALSE - new : TRUE - modify
 *              CEI_INT32 numberOfEntries -- (input) the number of entries to define or modify, used as a
 *                                                   reference for the number of structure elements defined
 *                                                   via the messageEntry pointer
 *              pAR_SCHEDULED_MSG_ENTRY_TYPE
 *                        messageEntry    -- (input) structure of message content defined as follows:
 *
 *                                              struct {
 *                                                        CEI_UINT32 messageIndex;                  
 *                                                        CEI_UINT32 board;                         
 *                                                        CEI_UINT32 channel;                       
 *                                                        CEI_UINT32 rate;                          
 *                                                        CEI_UINT32 start;                         
 *                                                        CEI_UINT32 txCount;               
 *                                                        CEI_UINT32 data; 
 *                                                     }
 * RETURN VAL:  CEI_INT32  -- ARS_NORMAL      Success
 *                            ARS_FAILURE     message scheduler table full
 *
 * GLOBALS:     device[]
 * 
 * DESCRIPTION: This routine defines the supplied message(s) in the next available entires within the device 
 *              message scheduler table.
 * 
 *===================================================================================================================*/ 
CEI_INT32 a429_utl_store_message_scheduler_entry 
(
   CEI_INT32 modifyEntryFlag,                                                /* (input) modify existing entries flag */
   CEI_INT32 numberOfEntries,                                                /* (input) number of entries to process */
   pAR_SCHEDULED_MSG_ENTRY_TYPE messageEntry                                 /* (input) ptr to struct array of msgs  */
);

/*===================================================================================================================*
 * ENTRY POINT:     A 4 2 9 _ U T L _V A L I D A T E _ S C H E D U L E D _ M E S S A G E _ P A R A M E T E R S
 *===================================================================================================================*
 *
 * FUNCTION:    Validate the parameters for a scheduled transmission request.
 *
 * PARAMETERS:  CEI_UINT32 board     -- (input)  Board number (0 to 127) 
 *              CEI_UINT32 channel   -- (input)  ARINC 429 Tx channel selection, valid range is 0 to N, where N 
 *                                               refers to the ARINC 561 channel, and N-1 refers to the last 429 
 *                                               Tx channel installed.
 *              pCEI_INT32 hwChannel -- (output) resp. h/w-based channel number
 *
 * RETURN VAL:  CEI_INT32    -- ARS_NORMAL      Success
 *                              ARS_INVBOARD    Invalid or uninitialized board
 *                              ARS_INVARG      Invalid channel
 *                              ARS_INVHARVAL   Message scheduling is not supported on this device
 *
 * GLOBALS:     device[]
 * 
 * DESCRIPTION: This routine validates the board and channel parameter values for a host-requested message 
 *              scheduling function.  It then returns the 429-based channel-index-referenced hardware-based channel 
 *              number for use by the calling routine.
 * 
 *===================================================================================================================*/ 
CEI_INT32 a429_utl_validate_scheduled_message_parameters 
(
   CEI_UINT32 board,                                                  /* (input)  Board number (0 to 127)            */
   CEI_UINT32 channel,                                                /* (input)  API-based 429 Tx channel selection */
   pCEI_UINT32 hwChannel                                              /* (output) resp. h/w-based channel number     */
);
#endif

#ifdef INCLUDE_PLX_SUPPORT
/*===================================================================================================================*
 * ENTRY POINT:                               A 4 2 9 _ U T L _ E N H
 *===================================================================================================================*
 *
 * FUNCTION:    Open a session to the board and update the EEPROM for CEI-x30 Enhanced firmware operation
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board  -- (input) board to reference (0 to 127)
 *
 *              CEI_INT32 action -- (input) operation to perform:
 *                                          PLX_ENHANCED_CHECK   - check to see if BAR2 is 512Kb
 *                                          PLX_ENHANCED_UPDATE  - check to see if BAR2 is 512Kb and
 *                                                                 reprogram it if it is only 4Kb
 *                                          PLX_ORIGINAL_RESTORE - if BAR2 is not 4kb, this option will result in 
 *                                                                 reprogramming BAR2 to the original x30 4Kb size
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL      Success updating or already done
 *                                ARS_INVBOARD    Invalid board parameter
 *                                ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                                ARS_FAILURE     Invalid device (not x30)
 *                                ARS_MEMWRERR    PROM update failed
 *                                ARS_DRIVERFAIL  could not access the CEI_Install library or
 *                                                the Driver Open/Close Session invocation failed
 * 
 * GLOBALS:     device[board]
 * 
 * DESCRIPTION: This routine is a special utility routine to upgrade a CEI-x30 board for use with the enhanced 
 *              firmware and API.  It modifies the PLX PROM BAR2 size from 4Kb to 512Kb.
 *
 *===================================================================================================================*/
CDEV_API_RET_TYPE a429_utl_enh 
(
   CDEV_BOARD_TYPE board,                                             /* (input)  board to reference (0 to 127)      */
   CEI_INT32 action                                                   /* (input)  operation (check/update/restore)   */
);

/*===================================================================================================================*
 * ENTRY POINT:                          A 4 2 9 _ U T L _ E N H _ C H E C K
 *===================================================================================================================*
 *
 * FUNCTION:    Check the PLX EEPROM for CEI-x30 Enhanced firmware support
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board  -- (input) board to reference (0 to 127)
 *              CEI_INT32  deviceIdent -- (input) PCI Device ID for device
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL     Success updating or already done
 *                                ARS_INVBOARD   Invalid board parameter
 *                                ARS_FAILURE    Invalid device (not x30)
 * 
 * GLOBALS:     device[board]
 * 
 * DESCRIPTION: This routine is a special utility to check CEI-x30 board support for the enhanced firmware and API.
 *
 *===================================================================================================================*/
CDEV_API_RET_TYPE a429_utl_enh_check 
(
   CDEV_BOARD_TYPE board,                                                  /* (input)  board to reference (0 to 127) */
   CEI_INT32 deviceIdent                                                   /* (input)  PCI Device ID for device      */
);

#ifndef FLASH_BASED_TARGET
/*===================================================================================================================*
 * ENTRY POINT:                    A 4 2 9 _ U T L _ P R O G R A M _ X I L I N X _ 9 0 3 0
 *===================================================================================================================*
 *
 * FUNCTION:    Download the FPGA program to an x30 device having a PLX9030 PCI interface and Xilinx FPGA, 
 *              supporting 8-bit programming.
 *
 * RETURN VAL:  CEI_INT32  -- ARS_NORMAL      Success
 *                            ARS_BADLOAD     FPGA load process failed
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_program_xilinx_9030
(
   CDEV_BOARD_TYPE board,                                                /* (input)  board to reference (0 to 127)   */
   CEI_INT32   deviceIdent,                                              /* (input)  PCI Configuration Device ID     */
   CEI_UINT32  programSize,                                              /* (input)  size of FPGA program load       */
   pCEI_UCHAR  fwloadPtr                                                 /* (input)  pointer to the FPGA program     */
);

/*===================================================================================================================*
 * ENTRY POINT:          A 4 2 9 _ U T L _ P R O G R A M _ X I L I N X _ 9 0 3 0 _ 3 2 B I T
 *===================================================================================================================*
 *
 * FUNCTION:    Download the FPGA program to an x30 device having a PLX9030 PCI interface and Xilinx FPGA, 
 *              supporting 32-bit programming.
 *
 * RETURN VAL:  CEI_INT32  -- ARS_NORMAL      Success
 *                            ARS_BADLOAD     FPGA load process failed
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_program_xilinx_9030_32bit
(
   CDEV_BOARD_TYPE board,                                                /* (input)  board to reference (0 to 127)   */
   CEI_INT32    deviceIdent,                                             /* (input)  PCI Configuration Device ID     */
   CEI_UINT32   programSize,                                             /* (input)  size of FPGA program load       */
   pCEI_UINT32  fwloadPtr                                                /* (input)  pointer to the FPGA program     */
);

/*===================================================================================================================*
 * ENTRY POINT:               A 4 2 9 _ U T L _ P R O G R A M    X I L I N X   9 0 5 6
 *===================================================================================================================*
 *
 * FUNCTION:    Download the FPGA program to an x30 device having a PLX9056 PCI interface and Xilinx FPGA, 
 *              supporting 8-bit programming.
 *
 * RETURN VAL:  CEI_INT32  -- ARS_NORMAL      Success
 *                            ARS_BADLOAD     FPGA load process failed
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_program_xilinx_9056 
(
   CDEV_BOARD_TYPE board,                                                /* (input)  device ID                       */
   CEI_INT32   deviceIdent,                                              /* (input)  PCI Configuration Device ID     */
   CEI_UINT32  programSize,                                              /* (input)  size of FPGA program load       */
   pCEI_UCHAR  fwloadPtr                                                 /* (input)  pointer to the FPGA program     */
);

/*===================================================================================================================*
 * ENTRY POINT:          A 4 2 9 _ U T L _ P R O G R A M _ X I L I N X _ 9 0 5 6 _ 3 2 B I T
 *===================================================================================================================*
 *
 * FUNCTION:    Download the FPGA program to an x30 device having a PLX9056 PCI interface and Xilinx FPGA, 
 *              supporting 32-bit programming.
 *
 * RETURN VAL:  CEI_INT32  -- ARS_NORMAL      Success
 *                            ARS_BADLOAD     FPGA load process failed
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_program_xilinx_9056_32bit
(
   CDEV_BOARD_TYPE board,                                                /* (input)  device ID                       */
   CEI_INT32   deviceIdent,                                              /* (input)  PCI Configuration Device ID     */
   CEI_UINT32  programSize,                                              /* (input)  size of FPGA program load       */
   pCEI_UINT32 fwloadPtr                                                 /* (input)  pointer to the FPGA program     */
);

/*===================================================================================================================*
 * ENTRY POINT:                       A 4 2 9 _ U T L _ P L X _ E E P R O M _ R E A D
 *===================================================================================================================*
 *
 * FUNCTION:    Read and return the contents of the PLX EEPROM
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board      -- (input) board to reference (0 to 127)
 * 
 *              CEI_INT32 pciInterfaceType -- (input) PLX Part (9030 or 9056)
 * 
 *              pCEI_UINT32 pciConfigPtr   -- (input) ptr to the BAR0 memory region
 * 
 *              pCEI_UINT32 promValues     -- (output) ptr to array to store values
 * 
 * RETURN VAL:  CEI_INT32 -- ARS_NORMAL     Normal succesful completion.
 *                           other values   Platform specific device access failures
 * 
 * GLOBALS:     none
 *
 * DESCRIPTION: This function reads the PLX PROM and returns the values.
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_plx_eeprom_read 
(
   CDEV_BOARD_TYPE board,                                             /* (input)  board to reference (0 to 127)      */
   CEI_INT32   pciInterfaceType,                                      /* (input)  PLX Part Type (9030 or 9056)       */
   pCEI_UINT32 pciConfigPtr,                                          /* (input)  ptr to the BAR0 memory region      */
   pCEI_UINT32 promValues                                             /* (output) ptr to array to store values       */
);

/*===================================================================================================================*
 * ENTRY POINT:               A 4 2 9 _ U T L _ P L X _ E E P R O M _ W R I T E _ E N A B L E
 *===================================================================================================================*
 *
 * FUNCTION:    Enable write access to the PLX PROM
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board      -- (input) board to reference (0 to 127)
 * 
 *              CEI_INT32 pciInterfaceType -- (input) PLX Part (9030 or 9056)
 * 
 *              pCEI_UINT32 pciConfigPtr   -- (input) ptr to the BAR0 memory region
 * 
 * RETURN VAL:  CEI_INT32 -- ARS_NORMAL     Normal succesful completion.
 *                           other values   Platform specific device access failures
 * 
 * GLOBALS:     none
 *
 * DESCRIPTION: This function enables write access to the PLX PROM.
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_plx_eeprom_write_enable 
(
   CDEV_BOARD_TYPE board,                                             /* (input)  board to reference (0 to 127)      */
   CEI_INT32   pciInterfaceType,                                      /* (input)  PLX Part Type (9030 or 9056)       */
   pCEI_UINT32 pciConfigPtr                                           /* (input)  ptr to the BAR0 memory region      */
);

/*===================================================================================================================*
 * ENTRY POINT:                A 4 2 9 _ U T L _ P L X _ E E P R O M _ W R I T E _ D I S A B L E
 *===================================================================================================================*
 *
 * FUNCTION:    Disable write access to the PLX PROM
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board    -- (input) board to reference (0 to 127)
 * 
 *              CEI_INT32 pciInterfaceType -- (input) PLX Part (9030 or 9056)
 * 
 *              pCEI_UINT32 pciConfigPtr -- (input) ptr to the BAR0 memory region
 * 
 * RETURN VAL:  CEI_INT32 -- ARS_NORMAL     Normal succesful completion.
 *                           other values   Platform specific device access failures
 * 
 * GLOBALS:     none
 *
 * DESCRIPTION: This function disables write access to the PLX PROM.
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_plx_eeprom_write_disable 
(
   CDEV_BOARD_TYPE board,                                             /* (input)  board to reference (0 to 127)      */
   CEI_INT32   pciInterfaceType,                                      /* (input)  PLX Part Type (9030 or 9056)       */
   pCEI_UINT32 pciConfigPtr                                           /* (input)  ptr to the BAR0 memory region      */
);

/*===================================================================================================================*
 * ENTRY POINT:                    A 4 2 9 _ U T L _ P L X _ E E P R O M _ W R I T E
 *===================================================================================================================*
 *
 * FUNCTION:    Write a value to the specified PLX PROM location
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board      -- (input) board to reference (0 to 127)
 * 
 *              CEI_INT32 pciInterfaceType -- (input) PLX Part (9030 or 9056)
 * 
 *              pCEI_UINT32 pciConfigPtr   -- (input) ptr to the BAR0 memory region
 * 
 *              CEI_UINT32 address         -- (input) offset into the PLX PROM to write
 * 
 *              CEI_UINT32 data            -- (input) value to write to the PLX PROM
 * 
 * RETURN VAL:  CEI_INT32 -- ARS_NORMAL     Normal succesful completion.
 *                           other values   Platform specific device access failures
 * 
 * GLOBALS:     none
 *
 * DESCRIPTION: This function writes a value to the specified PLX PROM location.
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_plx_eeprom_write 
(
   CDEV_BOARD_TYPE board,                                             /* (input)  board to reference (0 to 127)      */
   CEI_INT32   pciInterfaceType,                                      /* (input)  PLX Part Type (9030 or 9056)       */
   pCEI_UINT32 pciConfigPtr,                                          /* (input)  ptr to the BAR0 memory region      */
   CEI_UINT32 address,                                                /* (input)  offset into the PLX PROM to write  */
   CEI_UINT32 data                                                    /* (input)  value to write to the PLX PROM     */
);
#endif
#endif


/*===================================================================================================================* 
 *===================================================================================================================* 
 *                                                                                                                   *
 *  Operating System Dependent API Local (Private) Function prototypes implemented across multiple C source files:   *
 *                                                                                                                   *
 *  CDEV_INT.C  - Support for Integrity 5.10 and 11.0                                                                *
 *  CDEV_LNX.C  - Support for Linux                                                                                  *
 *  CDEV_LRT.C  - Support for LabVIEW Real-Time (Pharlap DOS)                                                        *
 *  CDEV_VISA.C - Support for NI-VISA                                                                                *
 *  CDEV_VXW.C  - Support for VxWorks 6.x and 7.x                                                                    *
 *  CDEV_WIN.C  - Support for Windows                                                                                *
 *                                                                                                                   *
 *===================================================================================================================*
 *===================================================================================================================*/

/*===================================================================================================================*
 * ENTRY POINT:                         C E I _ A C Q U I R E _ D E V I C E
 *===================================================================================================================*
 *
 * FUNCTION:    Open a session with a device
 *              
 * PARAMETERS:  CDEV_BOARD_TYPE deviceIndex  -- (input)  device to reference (0 to 15, 127 for Windows)
 *              ARINC_LOCAL_TYPE * device    -- (output) reference to ARINC_LOCAL_TYPE struct
 * 
 * RETURN VAL:  ARS_NORMAL     -- Session was opened successfully
 *              ARS_DRIVERFAIL -- Under Windows, Linux, Visa, and LabVIEW R/T, the respective driver 'Open Session' 
 *                                or 'Resource Acquisition' operations failed.
 *              ARS_FAILURE    -- Under VxWorks and Integrity, the driver 'Open Session' opreation failed.
 *                                Under Windows, the CEI_Install.DLL cannot be accessed
 *
 * DESCRIPTION: This routine searches for the instance of a CEI-x30 device matching the supplied device ID, with the 
 *              resource acquisition for each instance of a device found limited to the first execution of this 
 *              routine.  In most cases subsequent invocations will simply retrieve the resources allocated for the 
 *              respective device index from global/static variables, (with the exception of some PPC SBCs such as  
 *              Thales where the "sysPciMemToLocalAdrs" translation will be applied to previously mapped addresses).
 *
 *===================================================================================================================*/
CEI_INT32 cei_acquire_device (CDEV_BOARD_TYPE deviceIndex, ARINC_LOCAL_TYPE * device);

/*===================================================================================================================*
 * ENTRY POINT:                    C E I _ A C Q U I R E _ D E V I C E _ I D E N T I T Y
 *===================================================================================================================*
 *
 * FUNCTION:    Open a session with a device
 *              
 * PARAMETERS:  CEI_INT32  deviceIndex -- (input)  device to reference (0 to 15, 127 for Windows)
 *              pCEI_INT32 vendorId    -- (output) ptr to storage of vendor PCI ID
 *              pCEI_INT32 deviceId    -- (output) ptr to storage of device PCI ID
 * 
 * RETURN VAL:  ARS_NORMAL     -- Successful acquisition of the device ID's
 *              ARS_FAILURE    -- Under Visa, LabVIEW R/T, VxWorks and Linux, the driver session is not open
 *                                Under Windows, the CEI_Install DLL cannot be accessed
 *              ARS_INVBOARD   -- Under Windows only, device configuration space is inaccessible (session not opened)
 *
 * DESCRIPTION: This routine reports the specified device PCI Configuration Space Vendor ID and Device ID values 
 *              either read directly from PCI Configuration Space or acquired and saved during initialization.
 *
 *===================================================================================================================*/
CEI_INT32 cei_acquire_device_identity (CEI_INT32 deviceIndex, pCEI_INT32 vendorId, pCEI_INT32 deviceId);

/*===================================================================================================================*
 * ENTRY POINT:                          C E I _ R E L E A S E _ D E V I C E
 *===================================================================================================================*
 *
 * FUNCTION:    Close a session with a device
 *              
 * PARAMETERS:  CDEV_BOARD_TYPE deviceIndex  -- (input)  device to reference (0 to 15, 127 for Windows)
 *              ARINC_LOCAL_TYPE * device    -- (output) reference to ARINC_LOCAL_TYPE struct
 * 
 * RETURN VAL:  ARS_NORMAL         - session was closed successfully
 *
 *              Status Values Applicable to limited operating systems...
 *
 *              ARS_BOARD_MUTEX    - failed to release the host interface thread lock mechanism (Windows, Integrity)
 *              ARS_ERR_SH_MEM_OBJ - failed to release shared memory resource (Windows only)
 *              ARS_DRIVERFAIL     - no access to the CEI_Install library or Driver Close Session failed (Windows only)
 *
 * DESCRIPTION: This routine releases a driver handle (closes the session), and releases all resources acquired
 *              for this device during exectuion of cei_acquire_device.
 *
 *===================================================================================================================*/
CEI_INT32 cei_release_device (CDEV_BOARD_TYPE deviceIndex, ARINC_LOCAL_TYPE * device);

/*===================================================================================================================*
 * ENTRY POINT:                           C E I _ R E A D _ 3 2 _ B I T
 *===================================================================================================================*
 *
 * FUNCTION:    Perform a 32-bit read from the specified device
 *
 * PARAMETERS:  CDEV_BOARD_TYPE deviceIndex  -- (input) device to reference (0 to 15, 127 for Windows)
 *              pCEI_UINT32 address          -- (input) location in host virtual memory to read
 *              pCEI_UINT32 data             -- (input) pointer in which to store the data value read
 * 
 * RETURN VAL:  For Visa and LabVIEW R/T: -- Visa 'viIn32' invocation return status, if not VI_SUCCESS
 *              ARS_NORMAL                -- Success
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine performs a 32-bit read access from a device PCI memory region
 *
 *===================================================================================================================*/
CEI_INT32 cei_read_32bit (CDEV_BOARD_TYPE board,  pCEI_UINT32 address, pCEI_UINT32 data);

/*===================================================================================================================*
 * ENTRY POINT:                           C E I _ W R I T E _ 3 2 _ B I T
 *===================================================================================================================*
 *
 * FUNCTION:    Perform a 32-bit write to the specified device
 *
 * PARAMETERS:  CDEV_BOARD_TYPE deviceIndex  -- (input) device to reference (0 to 15, 127 for Windows)
 *              pCEI_UINT32 address          -- (input) location in host virtual memory to write
 *              CEI_UINT32 data              -- (input) the data value to write
 * 
 * RETURN VAL:  For Visa and LabVIEW R/T: -- Visa 'viOut32' invocation return status, if not VI_SUCCESS
 *              ARS_NORMAL                -- Success
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine performs a 32-bit write access to a device PCI memory region
 *
 *===================================================================================================================*/
CEI_INT32 cei_write_32bit (CDEV_BOARD_TYPE board, pCEI_UINT32 address, CEI_UINT32 data);

/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ D E L A Y _ S E C O N D S
 *===================================================================================================================*
 *
 * FUNCTION:    Delay for a specified (fractional) number of seconds
 *
 * PARAMETERS:  CEI_FLOAT numberOfSeconds -- (input) a floating point representation of the number (or fraction) of
 *                                                   seconds to delay execution of the calling thread
 * 
 * RETURN VAL:  ARS_FAILURE -- For Integrity only, indicates a nanosleep invocation failure
 *              ARS_NORMAL  -- Success
 *
 * DESCRIPTION: This routine utilizes the following o/s-specific mechanism for delaying thread execution:
 *
 *              WINDOWS:    A delay-polling method is used based on repeated invocations of timeGetTime, having a 
 *                          typical resolution of 1 millisecond.
 *              INTEGRITY:  Delay method is optional - for Posix, nanosleep is used; otherwise, usleep is used.
 *              VXWORKS:    Delay method is BSP dependent - either sysUsDelay, sysMsDelay, or taskDelay is used.
 *              LINUX:      A delay-polling method is used based on repeated invocations of clock()
 *              VISA:       The C standard library Sleep() function is used.
 *              LABVIEW RT: The o/s function Delay() is used
 *
 *===================================================================================================================*/
CEI_INT32 cei_delay_seconds (CEI_FLOAT numberOfSeconds);

/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ D E L A Y _ M I L L I S E C O N D S
 *===================================================================================================================*
 *
 * FUNCTION:    Delay for a specified number of milliseconds
 *
 * PARAMETERS:  CEI_UINT32 numberOfMillieconds -- (input) the number of milliseconds to delay execution of the 
 *                                                        calling thread
 * 
 * RETURN VAL:  ARS_FAILURE -- For Integrity only, indicates a nanosleep invocation failure
 *              ARS_NORMAL  -- Success
 *
 * DESCRIPTION: This routine utilizes the following o/s-specific mechanism for delaying thread execution:
 *
 *              WINDOWS:    The C standard library Sleep() function is used.
 *              INTEGRITY:  Delay method is optional - for Posix, nanosleep is used; otherwise, usleep is used.
 *              VXWORKS:    Delay method is BSP dependent - either sysUsDelay, sysMsDelay, or taskDelay is used.
 *              LINUX:      A delay-polling method is used based on repeated invocations of clock()
 *              VISA:       The C standard library Sleep() function is used.
 *              LABVIEW RT: The o/s function Delay() is used
 *
 *===================================================================================================================*/
CEI_INT32 cei_delay_milliseconds (CEI_UINT32 numberOfMilliseconds);

/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ G E T _ O P E N _ E R R O R
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the error description for a device open error
 *
 * PARAMETERS:  openErrorString -- (output) reference to assign to error string
 * 
 * RETURN VAL:  ARS_NORMAL  -- Mutex was acquired successfully
 *
 * GLOBALS:     deviceOpenErrorDesc
 *
 * DESCRIPTION: This routine retrieves the error description previously saved when a device open error is encountered.
 *
 *===================================================================================================================*/
#if defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
CEI_INT32 cei_get_open_error (pCEI_CHAR openErrorString);
#else
pCEI_CHAR cei_get_open_error (CEI_VOID);
#endif

/*===================================================================================================================*
 * ENTRY POINT:                      C E I _ M A P _ S H A R E D _ M E M O R Y
 *===================================================================================================================*
 *
 *              NOTE: IMPLEMENTED FOR WINDOWS AND LINUX TARGET OPERATING SYSTEMS ONLY!
 *
 * FUNCTION:    Map the desired shared global memory block into the calling process's address space, allocating 
 *              the memory block if it doesn't already exist.  
 *
 * PARAMETERS:  CEI_UCHAR  blockId        -- (input)  unique global block id (0-15)
 *              CEI_UINT32 blockSize      -- (input)  memory block size (in bytes)
 *              CEI_VOID   **blockPtr     -- (output) receives pointer to memory block
 *              CEI_VOID   **blockHandle  -- (output) receives platform-specific handle to the shared memory block
 *              CEI_UINT32 *alreadyExists -- (output) receives true if the block already exists, false otherwise
 *              CEI_UINT32 *isExclusive   -- (output) receives true if no other processes currently have the block 
 *                                                    mapped, false otherwise
 *
 * RETURN VAL:  ARS_NORMAL          - operation completed successfully
 *              ARS_ERR_SH_MEM_MAP  - failed to map shared memory block
 *              ARS_ERR_SH_MEM_OBJ  - failed to allocate shared memory block
 *              ARS_INVARG          - board index is invalid
 *              ARS_FAILURE         - Windows only - the CEI_Install library is not loaded in memory
 *
 * DESCRIPTION: If a shared global memory block associated with the given application-specific id has already been 
 *              allocated on the target system, map the memory block into the calling process's address space and 
 *              set 'already_exists' to true.  Otherwise, allocate and map the specified global memory block
 *              and set 'already_exists' to false.
 *
 *              If the memory block is being allocated for the first time, the caller should zero-fill the block.
 *
 *===================================================================================================================*/
CEI_INT32 cei_map_shared_memory 
( 
   CEI_UCHAR    blockId,                                      /* (input)  unique global block id (0-15)              */
   CEI_UINT32   blockSize,                                    /* (input)  memory block size (in bytes)               */
   CEI_VOID  ** blockPtr,                                     /* (output) receives pointer to memory block           */
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)                                                                                         
   CEI_INT32  * blockHandle,                                  /* (output) handle to the shared memory block          */
#else                                                                                                                                      
   CEI_VOID  ** blockHandle,                                  /* (output) handle to the shared memory block          */
#endif                                                                             
   CEI_UINT32 * alreadyExists,                                /* (output) true - block already exists, false - new   */  
   CEI_UINT32 * isExclusive                                   /* (output) true = no other owners, false - in use     */                                                         
);                                                              

/*===================================================================================================================*
 * ENTRY POINT:                     C E I _ G E T _ H O S T _ S Y S T E M _ T I M E
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the host 32-bit 1ms reference time value, supplied via operating system "System Time"
 *
 * PARAMETERS:  pCEI_UINT32 hostTime -- (output) 32-bit 1 millisecond resolution time
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current System Time in milliseconds
 *
 *              WINDOWS:    System time is acquired via the o/s function timeGetTime().
 *              INTEGRITY:  System time is acquired via the o/s function clock_gettime (Posix), GetClockTime otherwise.
 *              VXWORKS:    System time is acquired via combination of localtime() and mktime() invocations.
 *              LINUX:      System time is acquired via the o/s function clock().
 *              VISA:       Invokes cei_get_host_system_time_x to acquire system time.
 *              LABVIEW RT: Invokes cei_get_host_system_time_x to acquire system time.
 *
 *===================================================================================================================*/
CEI_INT32 cei_get_host_system_time (pCEI_UINT32 hostTime);

/*===================================================================================================================*
 * ENTRY POINT:                     C E I _ G E T _ H O S T _ S Y S T E M _ T I M E _ X
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the host 64-bit 1us calendar time value, supplied via operating system as time since Jan 1.
 *
 * PARAMETERS:  TIME_TAG_TYPE hostTime -- (output) 64-bit 1us millisecond resolution time
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current 64-bit 1us resolution calendar time value
 *
 *              WINDOWS:    Calendar time is acquired via combination of localtime() and mktime() invocations.
 *              INTEGRITY:  Calendar time is acquired via combination of localtime() and mktime() invocations (Posix only).
 *              VXWORKS:    Calendar time is acquired via combination of localtime() and mktime() invocations.
 *              LINUX:      Calendar time is acquired via combination of localtime() and mktime() invocations.
 *              VISA:       Calendar time is acquired via combination of time() and localtime() inocations.
 *              LABVIEW RT: Calendar time is acquired via combination of time() and localtime() inocations.
 *
 *===================================================================================================================*/
CEI_VOID  cei_get_host_system_time_x (TIME_TAG_TYPE * hostTime);

/*===================================================================================================================*
 * ENTRY POINT:                  C E I _ G E T _ I R I G _ I N I T I A L _ T I M E
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the host reference calendar time/date value for CEI-x30 IRIG generator initialization
 *
 * PARAMETERS:  time -- (output) 32-bit packed IRIG time value
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current System Time in calendar time/date and formats it into an 
 *              IRIG bit-encoded format value.
 *
 *===================================================================================================================*/
CEI_VOID  cei_get_irig_initial_time (pCEI_UINT32 time);

#if defined (MULTI_PROCESS_SUPPORT) || !defined (TARGET_EMBEDDED)
/*===================================================================================================================*
 * ENTRY POINT:                         C E I _ O P E N _ S H A R E D _ L O C K
 *===================================================================================================================*
 *
 *              NOTE: IMPLEMENTED FOR WINDOWS AND LINUX TARGET OPERATING SYSTEMS ONLY.
 *                    FOR THE VISA DRIVER IMPLEMENTED AS AN BLOCK ON LOCK ACCESS
 *
 * FUNCTION:    Opens a shared (global) mutex/semaphore lock
 *
 * PARAMETERS:  CDEV_BOARD_TYPE sharedLockId -- (input) device to reference (0 to 15, 127 for Windows)
 *              pCEI_INT32 sharedLockState   -- (input) location to store the mutex/semaphore handle
 *
 * RETURN VAL:  ARS_NORMAL      -- global mutex/semaphore was opened successfully
 *              ARS_INVARG      -- global mutex/semaphore id too large
 *              ARS_BOARD_MUTEX -- failed to create/open the global mutex/semaphore
 *              ARS_FAILURE     -- Windows only, CEI_Install library is not loaded in memory
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: Opens the device-specific shared (global) lock for use in multiple process applications
 *
 *===================================================================================================================*/
CEI_INT32 cei_open_shared_lock (CDEV_BOARD_TYPE sharedLockId, pCEI_INT32 sharedLockState);

/*===================================================================================================================*
 * ENTRY POINT:                       C E I _ O P E N _ P R E L O A D _ L O C K
 *===================================================================================================================*
 *
 * FUNCTION:    Requests the initiation of a shared (global) lock for multiprocess application usage
 *
 * PARAMETERS:  CDEV_BOARD_TYPE device -- (input) index of the device for which to create the lock 
 *                                                (0 to 15, 127 for Windows)
 *
 * RETURN VAL:  ARS_NORMAL      - global mutex was opened successfully
 *              ARS_INVARG      - global mutex id too large
 *              ARS_BOARD_MUTEX - failed to create/open the global mutex
 *              ARS_FAILURE     - CEI_Install library is not loaded in memory
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: Create the device-specific shared (global) lock for use in multiple process applications for select
 *              operating systems.
 *
 *              WINDOWS:    Invokes cei_open_shared_lock.
 *              INTEGRITY:  Not implemented.
 *              VXWORKS:    Not implemented.
 *              LINUX:      Invokes cei_open_shared_lock.
 *              VISA:       Invokes viLock to create a shared (global) lock
 *              LABVIEW RT: Invokes viLock to create a shared (global) lock
 *
 *===================================================================================================================*/
CEI_INT32 cei_open_preload_lock (CDEV_BOARD_TYPE device);

/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ A C Q U I R E _ T H R E A D _ L O C K    
 *===================================================================================================================*
 *
 * FUNCTION:    Block on a mutex or semaphore previously reserved for the specific device.  
 *
 * PARAMETERS:  CDEV_BOARD_TYPE deviceIndex -- (input) device to reference (0 to 15, 127 for Windows)
 * 
 * RETURN VAL:  ARS_NORMAL      -- The mutex/semaphore was acquired successfully
 *              ARS_FAILURE     -- Acquisition of the lock mechanism timed-out/failed (Linux, VxWorks and Integrity)          
 *              ARS_BOARD_MUTEX -- Acquisition of the lock mechanism timed-out/failed (Windows, VISA, LabVIEW R/T)
 *
 * GLOBALS:     ceiAccessLock
 *
 * DESCRIPTION: This routine blocks execution, waiting for access to the locking mechanism previously allocated for  
 *              the specific device.  Some operating systems support a time-out error, others do not; for those that
 *              do, the ARS_BOARD_MUTEX status code is used in place of the generic ARS_FAILURE status code.
 *
 *              WINDOWS:    A mutex is used for single process multi-threading and multi-process applications.
 *              INTEGRITY:  A Posix mutex is used for single process multi-threading applications.
 *              VXWORKS:    A global semaphore is used for multi-process applications.
 *              LINUX:      A Posix mutex is used for single process multi-threading applications.
 *                          A global semaphore is used for multi-process applications.
 *              VISA:       A Visa Lock mechanism is used for multi-threading single and multi-process applications.
 *              LABVIEW RT: A Visa Lock mechanism is used for multi-threading single and multi-process applications.
 *
 *===================================================================================================================*/
CEI_INT32 cei_acquire_thread_lock (CDEV_BOARD_TYPE deviceIndex);

/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ R E L E A S E _ T H R E A D _ L O C K    
 *===================================================================================================================*
 *
 * FUNCTION:    Release a semaphore previously reserved/blocked by the application/API for the specific device.
 *
 *              WINDOWS:    A mutex is used for single process multi-threading and multi-process applications.
 *              INTEGRITY:  A Posix mutex is used for single process multi-threading applications.
 *              VXWORKS:    A global semaphore is used for multi-process applications.
 *              LINUX:      A Posix mutex is used for single process multi-threading applications.
 *                          A global semaphore is used for multi-process applications.
 *              VISA:       A Visa Lock mechanism is used for multi-threading single and multi-process applications.
 *              LABVIEW RT: A Visa Lock mechanism is used for multi-threading single and multi-process applications.
 *
 * PARAMETERS:  deviceIndex -- (input) device to reference (0 to 15, 127 for Windows)
 * 
 * RETURN VAL:  ARS_NORMAL      -- The mutex/semaphore was acquired successfully
 *              ARS_FAILURE     -- Acquisition of the lock mechanism timed-out/failed (Linux, VxWorks and Integrity)          
 *              ARS_BOARD_MUTEX -- Acquisition of the lock mechanism timed-out/failed (Windows, VISA, LabVIEW R/T)
 *
 * DESCRIPTION: This routine releases the resource access previously allocated as a locking mechanism for
 *              the specific device.  Some operating systems support a time-out error, others do not; for those that
 *              do, the ARS_BOARD_MUTEX status code is used in place of the generic ARS_FAILURE status code.
 *
 *===================================================================================================================*/
CEI_INT32 cei_release_thread_lock (CDEV_BOARD_TYPE deviceIndex);
#endif

#ifdef INCLUDE_INTERRUPTS
/*===================================================================================================================*
 * ENTRY POINT:                         C E I _ I N I T _ I N T E R R U P T
 *===================================================================================================================*
 *
 * FUNCTION:    Configures the device PCI interrupt enable and associates the supplied ISR function with it.
 *              
 * PARAMETERS:  CEI_INT32   deviceIndex       -- (input)  device to reference (0 to 127)
 *              CEI_INT32   _stdcall isr      -- (input)  ptr to the interrupt service routine
 *              CEI_INT32   priority          -- (input)  desired priority of the ISR thread
 *              pCEI_UINT32 deviceBaseAddress -- (input)  virtual address of BAR0/BAR2 region
 *              CEI_INT32   memRegion         -- (input)  PCI BAR region/index for interrupt clear bit (0 - BAR0 or 1 - BAR2)
 *              CEI_UINT32  disOffset         -- (input)  PCI Interrupt Disable register (PLX reg offset, 0 for no PLX interface)
 *              CEI_UINT32  interruptBit      -- (input)  Local Interrupt Input Enable bit for PLX interface control register
 *              CEI_UINT32  comboChannelCount -- (input)  1553/ARINC Combo Board configuration
 *              pCEI_UINT32 isrEnabled        -- (output) Thread successfully defined state (TRUE, FALSE)
 * 
 * RETURN VAL:  ARS_NORMAL         - Interrupt was initialized successfully
 *              ARS_DRIVERFAIL     - Device driver session never opened
 *              ARS_INT_ISR        - Interrupt was not initialized successfully
 *                                 - Driver interrupt event creation failure
 *                                 - Windows thread create failure
 *              ARS_NO_INT_SUPPORT - Driver interrupt handling not supported for this board
 *              ARS_NO_INT_ENABLED - Driver Interrupt handling not enabled for this board
 *
 * GLOBALS:     None
 *
 * DESCRIPTION: This routine enables the PCI interrupt, then connects the user-specified interrupt function   
 *              to the interrupt assigned to the respective device by the host o/s.
 *
 *===================================================================================================================*/
CEI_INT32 cei_init_interrupt 
( 
   CEI_INT32   deviceIndex,                /* (input)  device to reference (0 to 15, for Windows 127)                */
#if defined (_WIN32)    
   CEI_INT32   (_stdcall *isr)(CEI_INT32), /* (input)  ptr to the interrupt service routine                          */
#else
   pCEI_VOID   isr,                        /* (input)  ptr to the interrupt service routine                          */
#endif
   CEI_INT32   priority,                   /* (input)  desired priority of the ISR thread                            */         
   pCEI_UINT32 deviceBaseAddress,          /* (input)  virtual address of BAR0/BAR2 region                           */         
   CEI_INT32   memRegion,                  /* (input)  PCI BAR region/index for interrupt clear bit                  */
   CEI_UINT32  disOffset,                  /* (input)  PCI Interrupt Disable register                                */
   CEI_UINT32  interruptBit,               /* (input)  Local Interrupt Input Enable bit for PLX control register     */
   CEI_UINT32  comboChanCount,             /* (input)  1553/ARINC Combo Board configuration                          */         
   pCEI_UINT32 isrEnabled                  /* (output) Thread successfully defined state (TRUE, FALSE)               */         
);

/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ T E R M I N A T E _ I N T E R R U P T
 *===================================================================================================================*
 *
 * FUNCTION:    Provides the method to terminate an interrupt association to an ISR and clear the interrupt enable.
 * 
 * PARAMETERS:  CEI_INT32   deviceIndex       -- (input)  device to reference (0 to 127)
 *              CEI_UINT32  interruptBit      -- (input)  Local Interrupt Input Enable bit for PLX interface control register
 *              pCEI_UINT32 deviceBaseAddress -- (input)  virtual address of BAR2 region
 * 
 * RETURN VAL:  ARS_NORMAL   -- Interrupt was terminated successfully
 *              ARS_INT_ISR  -- Windows only, Interrupt was not terminated successfully
 *              ARS_FAILURE  -- Other than Windows, Interrupt was not terminated successfully
 *
 * GLOBALS:     None
 * 
 * DESCRIPTION: This routine terminates the specified device interrupt function and clears the PCI interrupt enable.
 * 
 *===================================================================================================================*/
CEI_INT32 cei_terminate_interrupt
(
   CEI_INT32   deviceIndex,                /* (input)  device to reference (0 to 15, for Windows 127)                */
   CEI_UINT32  interruptBit,               /* (input)  Local Interrupt Input Enable bit for PLX control register     */
   pCEI_UINT32 deviceBaseAddress           /* (input)  virtual address of BAR0/BAR2 region                           */         
);

/*===================================================================================================================*
 * ENTRY POINT:                 A 4 2 9 _ U T L _ I N I T I A L I Z E _ I N T _ Q U E U E
 *===================================================================================================================*
 *
 * FUNCTION:    Function that initializes the internal API interrupt queue
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board -- (input) board to reference (0 to 127)
 *              CEI_UINT32 initSource -- (input) source/calling function
 *                                               TRUE  - a429_utl_initialize_global_data (session init processing)
 *                                               FALSE - a429_utl_initialize_device      (API reset processing)
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     interruptQueue
 *              device
 *
 * DESCRIPTION: This routine initializes the internal API interrupt queue for the specified device.
 *
 *===================================================================================================================*/
CEI_VOID  a429_utl_initialize_int_queue 
(
   CDEV_BOARD_TYPE board,                                                       /* board to reference (0 to 127)     */
   CEI_UINT32      initSource                                                   /* source/calling function           */
);

/*===================================================================================================================*
 * ENTRY POINT:                       A 4 2 9 _ U T L _ P R O C E S S _ H W _ I N T
 *===================================================================================================================*
 *
 * FUNCTION:    Assign PCI Interrupt Control to the Board 
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board    -- (input) Board number (0 to 127) 
 *
 *              CDEV_PARM_SSI_TYPE value -- (input) interrupt enable/disable selection                                
 *                                                  AR_ON  :  enable PCI interrupts   
 *                                                  AR_OFF : disable PCI interrupts   
 *                                              
 * RETURN VAL:  CDEV_API_RET_TYPE EXIT_ROUTINE    Success
 *                                ARS_INVHARVAL   Invalid value parameter option
 *
 * GLOBALS:     device[]
 * 
 * DESCRIPTION: This routine assigns the state of PCI Interrupt Control to the Board, by set/clear of the PLX or 
 *              PCIe Core interrupt enable, and setup or removal of the interrupt task/thread from the 
 *              host-specific o/s (common named) interrupt routines.
 * 
 *===================================================================================================================*/
CDEV_API_RET_TYPE a429_utl_process_hw_int 
( 
   CDEV_BOARD_TYPE board,                                            /* (input) board to reference (0 to 127)        */
   CDEV_PARM_SSI_TYPE value                                          /* (input) interrupt enable/disable selection   */
);
#endif


/*====================================================================================================================* 
 *  For LabVIEW Real-Time with full VISA support, writes to BAR0 require a different argument to the viOut32
 *  invocation, so we use a different write access routine, see cdev_lrt.h for details; select via compiler directive.
 *  For all other operating systems BAR0 is accessed like BAR2.
 *====================================================================================================================*/
#if defined (NIVISA_PXI) 
   CEI_INT32 cei_read_plx_32bit (CDEV_BOARD_TYPE board,  pCEI_UINT32 address, pCEI_UINT32 data);
   CEI_INT32 cei_write_plx_32bit (CDEV_BOARD_TYPE board, pCEI_UINT32 address, CEI_UINT32 data);
   #define WRITE_32_BIT cei_write_plx_32bit
   #define READ_32_BIT  cei_read_plx_32bit
#else
   #define WRITE_32_BIT cei_write_32bit
   #define READ_32_BIT  cei_read_32bit
#endif

#endif

#endif /* CEIDEV_GLB_H */

