/*====================================================================================================================*
 * FILE:                     C D E V _ I N T . C
 *====================================================================================================================*
 *
 *   COPYRIGHT 2005 - 2019 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED
 *   ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH THE INCLUSION
 *   OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY OTHER COPIES THEREOF
 *   MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE
 *   AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS, INC. 
 *
 *====================================================================================================================*
 *
 * FUNCTION:    Application Programmer's Interface for a CEI-x30 ARINC device.
 *              Integrity Operating System Specific Functionality
 *
 *    Internal CEI-x30 routines implemented herein:
 *
 *    cei_acquire_device          - opens a session with the specified x30 device
 *    cei_acquire_device_identity - supplies the PCI configuration space ID values
 *    cei_acquire_thread_lock     - acquires sempahore/mutex lock
 *    cei_delay_milliseconds      - o/s specific delay utility in milliseconds
 *    cei_delay_seconds           - o/s specific delay utility in fractional seconds
 *    cei_get_host_system_time_x  - acquires host system time-of-year 64-bit uSec     
 *    cei_get_host_system_time    - acquires host system time-of-year 32-bit mSec     
 *    cei_get_irig_initial_time   - acquires host system time-of-year IRIG format
 *    cei_map_shared_memory       - maps memory region for sharing globals across processes
 *    cei_open_shared_lock        - creates a shared semaphore across processes
 *    cei_release_device          - closes a session with the specified x30 device
 *    cei_release_thread_lock     - releases sempahore/mutex lock     
 *    cei_read_32bit              - read a 32-bit value from an x30 device
 *    cei_write_32bit             - write a 32-bit value to an x30 device
 *    cei_init_interrupt          - configures an interrupt and ISR
 *    cei_terminate_interrupt     - terminate an interrupt
 *          
 *--------------------------------------------------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date    By        Comments
 * -------- -----      ------------------------------------------------------------
 * 11/28/05  BH        Created.
 * 04/11/06  BDW       Removed utility routines cei_read_plx_32bit & cei_write_plx_32bit
 * 05/04/06  BDW       Added basicConfiguration[] declaration
 * 08/24/06  BDW       Incorporated CEI_TYPES.H data type definitions in 
 *                       support of 64-bit Linux operating systems - this 
 *                       modified every data declaration and prototype.
 *                     Swapped the BAR reference order by the device structure 
 *                        members baseAddr and devicePtr
 * 02/18/08  BDW       Removed the boardInitialized array, relocated as a structure
 *                        member in the data structure array "device"
 * 03/14/08  BDW       Removed the declaration of the following global arrays, 
 *                     relocated to CDEV_GLB.H:
 *                        bypassWrapTest
 *                        irigDacThreshold
 *                        a717syncWords
 *                     Removed declaration of basicConfiguration, it was unused.
 * 04/15/08  BDW       Added support for initiation and termination of h/w interrupts
 *                     via routines cei_init_interrupt and cei_terminate_interrupt.
 * 06/16/09  BDW       Added the function cei_open_preload_lock for multiprocess support.
 * 07/01/09  BDW       Added new function cei_delay_milliseconds (o/s thread suspend).
 * 09/14/10  BDW       Replaced the fixed user-mode isr invocation in cei_init_interrupt
 *                     (API-defined func) with the function defined in x30InterruptFunctionPtr
 * 09/16/10  BDW       Renamed interruptFunctionPtr to x30InterruptFunctionPtr
 *                     Renamed interruptServiceRoutine to x30InterruptFunctionPtr
 *                     to eliminate duplicate names with the PIO product line API
 * 10/28/10  BDW       Modified the parameter list for cei_init_interrupt to account for      
 *                        interrupt setup requirements in the Windows-only CEI_Install library
 *
 *   Date    By   Ver. Comments
 * -------- ----- ---- ------------------------------------------------------------
 * 08/03/12  BDW  4.01 Beta release for RAR15-XMC-XT Integrity Support
 * 06/13/13  BDW  4.30 Release full distribution for the (R)AR15-XMC-XT and
 *                     RCEI-830X820 boards.  Added support for Integrity 178B
 *                     build without POSIX support.  Any standard Integrity
 *                     5.x build with POSIX support should define the directive
 *                     INTEGRITY_POSIX in the project compilation.
 * 08/07/15  BDW  4.71 Increased MAX_BOARDS from 8 to 16 for VxWorks and Integrity targets
 * 10/18/16  BDW  4.90 Introduced use of CEIx30 API-specific type defines via the file CEIX30_TYPES.H allowing
 *                     for easy build to common use of only signed integer return types and unsigned integer
 *                     parameter types (with the exception of 64-bit types); specifically replacing use of CEI_INT16
 *                     with CDEV_BOARD_TYPE for functions having a device ID parameter.
 * 02/04/20  BDW  5.1x Force a return status of ARS_NORMAL with cei_acquire_device_identity
 *====================================================================================================================*/ 
#include <string.h>
#ifdef INTEGRITY_POSIX
  #include <pthread.h>
#endif
#include <errno.h>
#include <util/error_string.h>
#include "cei_types.h"
#ifndef INTEGRITY_POSIX
  #include "target_defines.h"
#else
  #ifdef INTEGRITY_PCI_PPC
    #ifndef NON_INTEL_WORD_ORDER
      #define NON_INTEL_WORD_ORDER
    #endif
  #endif
#endif
#include "lowlevel.h"
#include "cdev_api.h"

#define MAX_DEV_OPEN_ERROR_STR_LEN 512

/* this array contains a reference to the active user-mode ISR */
extern void (*x30InterruptFunctionPtr[MAX_BOARDS])(CEI_INT32, pCEI_UINT32);

/*========================================================================* 
 * Board-based globals (Board-based items only supporting 2 devices)
 *========================================================================*/
static DEVMAP_T devObj[MAX_BOARDS];
static Task iTask[MAX_BOARDS];

/* hardware interrupts are only supported on a single CEI-830 device */
static CEI_INT32 hwInterruptDeviceAssignment = 99;

/*========================================================================* 
 * Operating System specific functionality
 *========================================================================*/
/*===========================================================================*
 * INTERNAL MACRO:        L I T T L E _ E N D I A N _ 3 2
 *===========================================================================*
 *
 * FUNCTION:    Return the little-endian equivalent of the 4-byte integral
 *              argument (endian swap if host is big-endian).
 *
 * RETURN VAL:  The little-endian representation of the argument.
 *
 *===========================================================================*/
#define little_endian_32(longword) ( ((longword & 0x000000ff) << 24) | \
                                     ((longword & 0x0000ff00) << 8)  | \
                                     ((longword & 0xff0000)   >> 8)  | \
                                     ((longword & 0xff000000) >> 24) )

/* pthread_mutex_t ceiAccessMutex[MAX_BOARDS] = {  Device Access mutex  */
int ceiAccessMutex[MAX_BOARDS] = { /* Device Access mutex  */
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};
#ifdef INTEGRITY_POSIX
pthread_mutexattr_t m_attr;
#endif

/* storage for Windows driver low-level board mapping error messages      */
static CEI_CHAR deviceOpenErrorDesc[MAX_DEV_OPEN_ERROR_STR_LEN] = "Specified PCI Device Not Found";


/*========================================================================* 
 * Non-Integrity Operating System specific functionality - here to resolve
 *========================================================================*/
/*===================================================================================================================*
 * ENTRY POINT:                         C E I _ O P E N _ S H A R E D _ L O C K
 *===================================================================================================================*
 *
 *              NOTE: IMPLEMENTED FOR WINDOWS AND LINUX TARGET OPERATING SYSTEMS ONLY.
 *
 *===================================================================================================================*/
CEI_INT32 cei_open_shared_lock (CDEV_BOARD_TYPE sharedLockId, pCEI_INT32 sharedLockState)
{
   return ARS_NORMAL;
}
/*===================================================================================================================*
 * ENTRY POINT:                       C E I _ O P E N _ P R E L O A D _ L O C K
 *===================================================================================================================*
 *
 *              NOTE: IMPLEMENTED FOR WINDOWS AND LINUX TARGET OPERATING SYSTEMS ONLY.
 *
 *===================================================================================================================*/
CEI_INT32 cei_open_preload_lock (CDEV_BOARD_TYPE device)
{
   return ARS_NORMAL;
}

/*===================================================================================================================*
 * ENTRY POINT:                      C E I _ M A P _ S H A R E D _ M E M O R Y
 *===================================================================================================================*
 *
 *              NOTE: IMPLEMENTED FOR WINDOWS AND LINUX TARGET OPERATING SYSTEMS ONLY!
 *
 *===================================================================================================================*/
CEI_INT32 cei_map_shared_memory 
( 
   CEI_UCHAR    blockId,
   CEI_UINT32   blockSize,
   CEI_VOID  ** blockPtr,
   CEI_VOID  ** blockHandle,
   CEI_UINT32 * alreadyExists,
   CEI_UINT32 * isExclusive 
)
{
   return 1;
}

/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ G E T _ O P E N _ E R R O R
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the error description for a device open error
 *
 * PARAMETERS:  openErrorString -- (output) reference to assign to error string
 * 
 * RETURN VAL:  ARS_NORMAL  -- Mutex was acquired successfully
 *
 * GLOBALS:     deviceOpenErrorDesc
 *
 * DESCRIPTION: This routine retrieves the error description previously saved when a device open error is encountered.
 *
 *===================================================================================================================*/
CEI_INT32 cei_get_open_error (pCEI_CHAR openErrorString)
{
  openErrorString = (pCEI_CHAR)&deviceOpenErrorDesc[0];
  return ARS_NORMAL;
}

/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ A C Q U I R E _ T H R E A D _ L O C K    
 *===================================================================================================================*
 *
 * FUNCTION:    Block on a mutex or semaphore previously reserved for the specific device.  
 *              WINDOWS:    A mutex is used for single process multi-threading and multi-process applications.
 *              INTEGRITY:  A Posix mutex is used for single process multi-threading applications.
 *              VXWORKS:    A global semaphore is used for multi-process applications.
 *              LINUX:      A Posix mutex is used for single process multi-threading applications.
 *                          A global semaphore is used for multi-process applications.
 *              VISA:       A Visa Lock mechanism is used for multi-threading single and multi-process applications.
 *              LABVIEW RT: A Visa Lock mechanism is used for multi-threading single and multi-process applications.
 *
 * PARAMETERS:  deviceIndex -- (input) device to reference (0 to 15, 127 for Windows)
 * 
 * RETURN VAL:  ARS_NORMAL      -- The mutex/semaphore was acquired successfully
 *              ARS_FAILURE     -- Acquisition of the lock mechanism timed-out/failed (Linux, VxWorks and Integrity)          
 *              ARS_BOARD_MUTEX -- Acquisition of the lock mechanism timed-out/failed (Windows, VISA, LabVIEW R/T)
 *
 * DESCRIPTION: This routine blocks execution, waiting for access to the locking mechanism previously allocated for  
 *              the specific device.  Some operating systems support a time-out error, others do not; for those that
 *              do, the ARS_BOARD_MUTEX status code is used in place of the generic ARS_FAILURE status code.
 *
 *===================================================================================================================*/
CEI_INT32 cei_acquire_thread_lock (CDEV_BOARD_TYPE deviceIndex)
{
#ifdef INTEGRITY_POSIX
   CEI_INT32 status=0;
   if((status = pthread_mutex_lock(&ceiAccessMutex[deviceIndex])) != Success) {
     printf("Fail: cei_acquire_thread_lock. pthread_mutex_lock - %d\n", status);
     return ARS_FAILURE; 
   }
#else
   CDEV_BOARD_TYPE status=deviceIndex;
#endif
   return ARS_NORMAL;
}

/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ R E L E A S E _ T H R E A D _ L O C K    
 *===================================================================================================================*
 *
 * FUNCTION:    Release a semaphore previously reserved/blocked by the application/API for the specific device.
 *
 *              WINDOWS:    A mutex is used for single process multi-threading and multi-process applications.
 *              INTEGRITY:  A Posix mutex is used for single process multi-threading applications.
 *              VXWORKS:    A global semaphore is used for multi-process applications.
 *              LINUX:      A Posix mutex is used for single process multi-threading applications.
 *                          A global semaphore is used for multi-process applications.
 *              VISA:       A Visa Lock mechanism is used for multi-threading single and multi-process applications.
 *              LABVIEW RT: A Visa Lock mechanism is used for multi-threading single and multi-process applications.
 *
 * PARAMETERS:  deviceIndex -- (input) device to reference (0 to 15, 127 for Windows)
 * 
 * RETURN VAL:  ARS_NORMAL      -- The mutex/semaphore was acquired successfully
 *              ARS_FAILURE     -- Acquisition of the lock mechanism timed-out/failed (Linux, VxWorks and Integrity)          
 *              ARS_BOARD_MUTEX -- Acquisition of the lock mechanism timed-out/failed (Windows, VISA, LabVIEW R/T)
 *
 * DESCRIPTION: This routine releases the resource access previously allocated as a locking mechanism for
 *              the specific device.  Some operating systems support a time-out error, others do not; for those that
 *              do, the ARS_BOARD_MUTEX status code is used in place of the generic ARS_FAILURE status code.
 *
 *===================================================================================================================*/
CEI_INT32 cei_release_thread_lock (CDEV_BOARD_TYPE deviceIndex)
{
#ifdef INTEGRITY_POSIX
   CEI_INT32 status=0;
   if((status = pthread_mutex_unlock(&ceiAccessMutex[deviceIndex])) != Success) {
     printf("Fail: cei_acquire_thread_lock. pthread_mutex_lock - %d\n", status);
     return ARS_FAILURE; 
   }
#else
   CDEV_BOARD_TYPE status=deviceIndex;
#endif
   return ARS_NORMAL;
}

/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ D E L A Y _ M I L L I S E C O N D S
 *===================================================================================================================*
 *
 * FUNCTION:    Delay for a specified number of milliseconds
 *
 * PARAMETERS:  CEI_UINT32 numberOfMillieconds -- (input) the number of milliseconds to delay execution of the 
 *                                                        calling thread
 * 
 * RETURN VAL:  ARS_FAILURE -- For Integrity only, indicates a nanosleep invocation failure
 *              ARS_NORMAL  -- Success
 *
 * DESCRIPTION: This routine utilizes the following o/s-specific mechanism for delaying thread execution:
 *
 *              WINDOWS:    The C standard library Sleep() function is used.
 *              INTEGRITY:  Delay method is optional - for Posix, nanosleep is used; otherwise, usleep is used.
 *              VXWORKS:    Delay method is BSP dependent - either sysUsDelay, sysMsDelay, or taskDelay is used.
 *              LINUX:      A delay-polling method is used based on repeated invocations of clock()
 *              VISA:       The C standard library Sleep() function is used.
 *              LABVIEW RT: The o/s function Delay() is used
 *
 *===================================================================================================================*/
CEI_INT32 cei_delay_milliseconds (CEI_UINT32 numberOfMilliseconds)
{
#ifdef INTEGRITY_POSIX
   struct timespec ts_d;
   struct timespec ts_rm;
   CEI_UINT32 numberOfSeconds;
   CEI_UINT32 remainingMilliSeconds;

   /* first handle parameters for a delay greater than one second */
   if (numberOfMilliseconds > 1000) {
      numberOfSeconds = numberOfMilliseconds/1000;
      remainingMilliSeconds = numberOfMilliseconds - (numberOfSeconds * 1000);
   }
   else {
      numberOfSeconds = 0;
      remainingMilliSeconds = numberOfMilliseconds;
   }
   ts_d.tv_sec = numberOfSeconds;
   ts_d.tv_nsec = remainingMilliSeconds * 1000000; 

   /* use nanosleep to suspend this thread/task for the specified delay duration */
   if(nanosleep(&ts_d, &ts_rm) != 0) {
     printf("Fail: cei_delay_seconds. nanosleep - %d\n", errno);
     return ARS_FAILURE; 
   }
#else   
   usleep(numberOfMilliseconds*1000);
#endif

   return ARS_NORMAL;
}


/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ D E L A Y _ S E C O N D S
 *===================================================================================================================*
 *
 * FUNCTION:    Delay for a specified (fractional) number of seconds
 *
 * PARAMETERS:  CEI_FLOAT numberOfSeconds -- (input) a floating point representation of the number (or fraction) of
 *                                                   seconds to delay execution of the calling thread
 * 
 * RETURN VAL:  ARS_FAILURE -- For Integrity only, indicates a nanosleep invocation failure
 *              ARS_NORMAL  -- Success
 *
 * DESCRIPTION: This routine utilizes the following o/s-specific mechanism for delaying thread execution:
 *
 *              WINDOWS:    A delay-polling method is used based on repeated invocations of timeGetTime, having a 
 *                          typical resolution of 1 millisecond.
 *              INTEGRITY:  Delay method is optional - for Posix, nanosleep is used; otherwise, usleep is used.
 *              VXWORKS:    Delay method is BSP dependent - either sysUsDelay, sysMsDelay, or taskDelay is used.
 *              LINUX:      A delay-polling method is used based on repeated invocations of clock()
 *              VISA:       The C standard library Sleep() function is used.
 *              LABVIEW RT: The o/s function Delay() is used
 *
 *===================================================================================================================*/
CEI_INT32 cei_delay_seconds (CEI_FLOAT numberOfSeconds)
{
#ifdef INTEGRITY_POSIX
   struct timespec ts_d;
   struct timespec ts_rm;

   /* first handle parameters for a delay greater than one second */
   ts_d.tv_sec = (CEI_UINT32)numberOfSeconds;
   ts_d.tv_nsec = (CEI_UINT32)(numberOfSeconds - (CEI_FLOAT)ts_d.tv_sec) * 1000000000; 
   
   /* use nanosleep to suspend this thread/task for the specified delay duration */
   if(nanosleep(&ts_d, &ts_rm) != 0) {
     printf("Fail: cei_delay_seconds. nanosleep - %d\n", errno);
     return ARS_FAILURE; 
   }
#else
   CEI_FLOAT fDelayTime;
   CEI_UINT32 uDelayTime;
   fDelayTime = numberOfSeconds * (CEI_FLOAT)1000000;
   uDelayTime = (CEI_UINT32)fDelayTime;
   usleep(uDelayTime);
#endif
   return ARS_NORMAL;
}

/*===================================================================================================================*
 * ENTRY POINT:                     C E I _ G E T _ H O S T _ S Y S T E M _ T I M E _ X
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the host 64-bit 1us calendar time value, supplied via operating system as time since Jan 1.
 *
 * PARAMETERS:  TIME_TAG_TYPE hostTime -- (output) 64-bit 1us millisecond resolution time
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current 64-bit 1us resolution calendar time value
 *
 *              WINDOWS:    Calendar time is acquired via combination of localtime() and mktime() invocations.
 *              INTEGRITY:  Calendar time is acquired via combination of localtime() and mktime() invocations (Posix only).
 *              VXWORKS:    Calendar time is acquired via combination of localtime() and mktime() invocations.
 *              LINUX:      Calendar time is acquired via combination of localtime() and mktime() invocations.
 *              VISA:       Calendar time is acquired via combination of time() and localtime() inocations.
 *              LABVIEW RT: Calendar time is acquired via combination of time() and localtime() inocations.
 *
 *===================================================================================================================*/
CEI_INT32 cei_get_host_system_time_x (TIME_TAG_TYPE * hostTime)
{
#ifdef INTEGRITY_POSIX
   time_t now;               /* Current time, modify to get Time Reference Point. */
   struct tm start;        /* Formatted "now" current time, modified for TRP.   */
   time_t then;            /* Time Reference Point, Midnight or Jan 1.          */
   struct timespec ts;
   TIME_TAG_TYPE diff;       /* This is the difference time we return in microseconds, */

   time(&now);               /* Time in seconds since midnight Jan 1, 1970 GMT */
   start = *localtime(&now); /* Local                                          */
   start.tm_sec  = 0;        /*   time                                         */
   start.tm_min  = 0;        /*     at                                         */
   start.tm_hour = 0;        /*       Jan 1                                    */
   start.tm_mday = 1;        /*         midnight (This is day 0)               */
   start.tm_mon  = 0;        /*           this                                 */
   start.tm_yday = 0;        /*             year                               */
   then = mktime(&start);    /* local Jan 1 to time in seconds since Jan 1     */
                             /* We do IRIG time, which makes Jan 1 "day 0"     */ 

   if(clock_gettime(CLOCK_REALTIME, &ts) != 0) { 
     printf("Fail: cei_get_host_system_time_x. clock_gettime - %d\n", errno);
     return ARS_FAILURE;
   }
   
   diff = ts.tv_sec - then;
   diff *= 1000000;           /* Number of Seconds converted to microseconds    */
   diff += (ts.tv_nsec / 1000);
   *hostTime = diff;          /* microseconds.  Passed back to the caller.      */
#else
   *hostTime = (TIME_TAG_TYPE)0;
#endif
   return ARS_NORMAL;
}

/*===================================================================================================================*
 * ENTRY POINT:                     C E I _ G E T _ H O S T _ S Y S T E M _ T I M E
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the host 32-bit 1ms reference time value, supplied via operating system "System Time"
 *
 * PARAMETERS:  pCEI_UINT32 hostTime -- (output) 32-bit 1 millisecond resolution time
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current System Time in milliseconds
 *
 *              WINDOWS:    System time is acquired via the o/s function timeGetTime().
 *              INTEGRITY:  System time is acquired via the o/s function clock_gettime (Posix), GetClockTime otherwise.
 *              VXWORKS:    System time is acquired via combination of localtime() and mktime() invocations.
 *              LINUX:      System time is acquired via the o/s function clock().
 *              VISA:       Invokes cei_get_host_system_time_x to acquire system time.
 *              LABVIEW RT: Invokes cei_get_host_system_time_x to acquire system time.
 *
 *===================================================================================================================*/
CEI_INT32 cei_get_host_system_time (pCEI_UINT32 hostTime)
{
#ifdef INTEGRITY_POSIX
   struct timespec ts;
	   
   if(clock_gettime(CLOCK_REALTIME, &ts) != 0) {
     printf("Fail: cei_get_host_system_time. clock_gettime - %d\n", errno);
     return ARS_FAILURE;
   }
   
   *hostTime = (ts.tv_sec * 1000) + (ts.tv_nsec / 1000000); 
#else
   Time T;
   GetClockTime(PrimaryClock, &T);
   *hostTime = T.Fraction/500;
#endif
   
   return ARS_NORMAL;
}

/*===================================================================================================================*
 * ENTRY POINT:                  C E I _ G E T _ I R I G _ I N I T I A L _ T I M E
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the host reference calendar time/date value for CEI-x30 IRIG generator initialization
 *
 * PARAMETERS:  time -- (output) 32-bit packed IRIG time value
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current System Time in calendar time/date and formats it into an 
 *              IRIG bit-encoded format value.
 *
 *===================================================================================================================*/
CEI_INT32 cei_get_irig_initial_time (pCEI_UINT32 time)
{
   TIME_TAG_TYPE system_time;
   CEI_NATIVE_UINT seconds;
   CEI_NATIVE_UINT tens_of_seconds;
   CEI_NATIVE_UINT minutes;
   CEI_NATIVE_UINT tens_of_minutes;
   CEI_NATIVE_UINT hours;
   CEI_NATIVE_UINT tens_of_hours;
   CEI_NATIVE_UINT days;
   CEI_NATIVE_UINT tens_of_days;
   CEI_NATIVE_UINT hundreds_of_days;

   union
   {
      struct irigTimeFormat
           {
              CEI_NATIVE_UINT seconds          :4;
              CEI_NATIVE_UINT tens_of_seconds  :3;
              CEI_NATIVE_UINT minutes          :4;
              CEI_NATIVE_UINT tens_of_minutes  :3;
              CEI_NATIVE_UINT hours            :4;
              CEI_NATIVE_UINT tens_of_hours    :2;
              CEI_NATIVE_UINT days             :4;
              CEI_NATIVE_UINT tens_of_days     :4;
              CEI_NATIVE_UINT hundreds_of_days :2;
           } IRIG_TIME_FORMAT;
      CEI_UINT32 timeValue;
   } timeConvert;

   if(cei_get_host_system_time_x (&system_time) != ARS_NORMAL)
     return ARS_FAILURE;

   seconds = (CEI_UINT32)(system_time / (TIME_TAG_TYPE)1000000);
   if ( (system_time % (TIME_TAG_TYPE)1000000) > (TIME_TAG_TYPE)500000) {
      seconds++;
   }

   /* Extract Minutes from Seconds */
   minutes = seconds/60;
   seconds = seconds%60;
   tens_of_seconds = seconds/10;
   seconds = seconds%10;

   /* Extract Hours from Minutes */
   hours = minutes/60;  
   minutes = minutes%60;
   tens_of_minutes = minutes/10;
   minutes = minutes%10;

   /* Extract Days from Hours */
   days = hours/24;     
   hours = hours%24;     
   tens_of_hours = hours/10;    
   hours = hours%10;    
   days++;                       /* account for no such thing as day 0 */
   hundreds_of_days = days/100;
   days = days%100;
   tens_of_days = days/10;
   days = days%10;

   timeConvert.IRIG_TIME_FORMAT.seconds          = seconds;
   timeConvert.IRIG_TIME_FORMAT.tens_of_seconds  = tens_of_seconds;
   timeConvert.IRIG_TIME_FORMAT.minutes          = minutes;
   timeConvert.IRIG_TIME_FORMAT.tens_of_minutes  = tens_of_minutes;
   timeConvert.IRIG_TIME_FORMAT.hours            = hours;
   timeConvert.IRIG_TIME_FORMAT.tens_of_hours    = tens_of_hours;
   timeConvert.IRIG_TIME_FORMAT.days             = days;
   timeConvert.IRIG_TIME_FORMAT.tens_of_days     = tens_of_days;
   timeConvert.IRIG_TIME_FORMAT.hundreds_of_days = hundreds_of_days;
   
   *time = timeConvert.timeValue;

   return ARS_NORMAL;
}

/*========================================================================* 
 * Device Access Routines for the Integrity Operating System
 *========================================================================*/

/*===================================================================================================================*
 * ENTRY POINT:                    C E I _ A C Q U I R E _ D E V I C E _ I D E N T I T Y
 *===================================================================================================================*
 *
 * FUNCTION:    Open a session with a device
 *              
 * PARAMETERS:  CEI_INT32  deviceIndex -- (input)  device to reference (0 to 15, 127 for Windows)
 *              pCEI_INT32 vendorId    -- (output) ptr to storage of vendor PCI ID
 *              pCEI_INT32 deviceId    -- (output) ptr to storage of device PCI ID
 * 
 * RETURN VAL:  ARS_NORMAL     -- Successful acquisition of the device ID's
 *              ARS_FAILURE    -- Under Visa, LabVIEW R/T, VxWorks and Linux, the driver session is not open
 *                                Under Windows, the CEI_Install DLL cannot be accessed
 *              ARS_INVBOARD   -- Under Windows only, device configuration space is inaccessible (session not opened)
 *
 * DESCRIPTION: This routine reports the specified device PCI Configuration Space Vendor ID and Device ID values 
 *              either read directly from PCI Configuration Space or acquired and saved during initialization.
 *
 *===================================================================================================================*/
CEI_INT32 cei_acquire_device_identity (
   CEI_INT32  deviceIndex,                                                        /* device number to test (0-7)     */
   pCEI_INT32 vendorId,                                                           /* ptr to storage of vendor PCI ID */
   pCEI_INT32 deviceId                                                            /* ptr to storage of device PCI ID */
)
{
   CEI_INT32 status = ARS_NORMAL;                                                 /* local function return status    */
   CEI_INT32 pciDeviceId;

   /* read the PCI manufacturer and device ID fields from PCI Configuration Space */
   status = vbtGetPCIConfigRegister ((unsigned int)deviceIndex, 0, 2, (unsigned int *)&pciDeviceId);
   *vendorId = (CEI_INT32)pciDeviceId;
   status = vbtGetPCIConfigRegister ((unsigned int)deviceIndex, 1, 2, (unsigned int *)&pciDeviceId);
   *deviceId = (CEI_INT32)pciDeviceId;

   return ARS_NORMAL;
}


/*===================================================================================================================*
 * ENTRY POINT:                         C E I _ A C Q U I R E _ D E V I C E
 *===================================================================================================================*
 *
 * FUNCTION:    Open a session with a device
 *              
 * PARAMETERS:  CDEV_BOARD_TYPE deviceIndex  -- (input)  device to reference (0 to 15, 127 for Windows)
 *              ARINC_LOCAL_TYPE * device    -- (output) reference to ARINC_LOCAL_TYPE struct
 * 
 * RETURN VAL:  ARS_NORMAL     -- Session was opened successfully
 *              ARS_DRIVERFAIL -- Under Windows, Linux, Visa, and LabVIEW R/T, the respective driver 'Open Session' 
 *                                or 'Resource Acquisition' operations failed.
 *              ARS_FAILURE    -- Under VxWorks and Integrity, the driver 'Open Session' opreation failed.
 *                                Under Windows, the CEI_Install.DLL cannot be accessed
 *
 * DESCRIPTION: This routine searches for the instance of a CEI-x30 device matching the supplied device ID, with the 
 *              resource acquisition for each instance of a device found limited to the first execution of this 
 *              routine.  In most cases subsequent invocations will simply retrieve the resources allocated for the 
 *              respective device index from global/static variables, (with the exception of some PPC SBCs such as  
 *              Thales where the "sysPciMemToLocalAdrs" translation will be applied to previously mapped addresses).
 *
 *===================================================================================================================*/
CEI_INT32 cei_acquire_device (CDEV_BOARD_TYPE deviceIndex, ARINC_LOCAL_TYPE * device)
{
   CEI_INT32 status = ARS_NORMAL;
	 
   memset (&devObj[deviceIndex], 0, sizeof(DEVMAP_T));

   /* request the addresses mapped to both BAR0 and BAR1 for the CEI-830 */
   devObj[deviceIndex].flagMapToHost[0] = TRUE;
   devObj[deviceIndex].flagMapToHost[1] = TRUE;

   if (vbtMapBoardAddresses(deviceIndex, &devObj[deviceIndex]) != BTD_OK)
     return ARS_FAILURE;
   
   device->baseAddr = (pCEI_UINT32)devObj[deviceIndex].memHostBase[0];
   device->devicePtr = (pCEI_UINT32)devObj[deviceIndex].memHostBase[1];

#ifdef INTEGRITY_POSIX
   /* initialize multi-threading mutex (using PTRHEADS) */
   if (pthread_mutexattr_init(&m_attr) == Success) {
      /* PTHREAD_PRIO_NONE, PTHREAD_PRIO_INHERIT, PTHREAD_PRIO_PROTECT */
      pthread_mutexattr_setprotocol(&m_attr, PTHREAD_PRIO_NONE);
      /* PTHREAD_MUTEX_DEFAULT, PTHREAD_MUTEX_NORMAL, PTHREAD_MUTEX_ERRORCHECK, PTHREAD_MUTEX_RECURSIVE */
      pthread_mutexattr_settype(&m_attr, PTHREAD_MUTEX_ERRORCHECK);
      status = pthread_mutex_init(&ceiAccessMutex[deviceIndex], &m_attr);
   }
   else {
      /* use the defaults (PTHREAD_PRIO_INHERIT and PTHREAD_MUTEX_DEFAULT) for the mutex */
      status = pthread_mutex_init(&ceiAccessMutex[deviceIndex], NULL);
   }

   /* convert return status to API type */
   status = (status == Success) ? ARS_NORMAL : ARS_FAILURE; 
#endif 
   return status;
}

/*===================================================================================================================*
 * ENTRY POINT:                          C E I _ R E L E A S E _ D E V I C E
 *===================================================================================================================*
 *
 * FUNCTION:    Close a session with a device
 *              
 * PARAMETERS:  CDEV_BOARD_TYPE deviceIndex  -- (input)  device to reference (0 to 15, 127 for Windows)
 *              ARINC_LOCAL_TYPE * device    -- (output) reference to ARINC_LOCAL_TYPE struct
 * 
 * RETURN VAL:  ARS_NORMAL         - session was closed successfully
 *
 *              Status Values Applicable to limited operating systems...
 *
 *              ARS_BOARD_MUTEX    - failed to release the host interface thread lock mechanism (Windows, Integrity)
 *              ARS_ERR_SH_MEM_OBJ - failed to release shared memory resource (Windows only)
 *              ARS_DRIVERFAIL     - no access to the CEI_Install library or Driver Close Session failed (Windows only)
 *
 * DESCRIPTION: This routine releases a driver handle (closes the session), and releases all resources acquired
 *              for this device during exectuion of cei_acquire_device.
 *
 *===================================================================================================================*/
CEI_INT32 cei_release_device (CDEV_BOARD_TYPE deviceIndex, ARINC_LOCAL_TYPE * device)
{
   CEI_INT32 status=0;

   vbtFreeBoardAddresses(&devObj[deviceIndex]);

#ifdef INTEGRITY_POSIX
   /* release multi-thread protection mutex */
   if ((status = pthread_mutex_destroy(&ceiAccessMutex[deviceIndex])) != Success) {
     printf("Fail: cei_release_device. pthread_mutex_destroy - %d\n", status);
     return ARS_BOARD_MUTEX; 
   }

   pthread_mutexattr_destroy(&m_attr);
#endif 

   return ARS_NORMAL;
}

/*===================================================================================================================*
 * ENTRY POINT:                           C E I _ R E A D _ 3 2 _ B I T
 *===================================================================================================================*
 *
 * FUNCTION:    Perform a 32-bit read from the specified device
 *
 * PARAMETERS:  CDEV_BOARD_TYPE deviceIndex  -- (input) device to reference (0 to 15, 127 for Windows)
 *              pCEI_UINT32 address          -- (input) location in host virtual memory to read
 *              pCEI_UINT32 data             -- (input) pointer in which to store the data value read
 * 
 * RETURN VAL:  For Visa and LabVIEW R/T: -- Visa 'viIn32' invocation return status, if not VI_SUCCESS
 *              ARS_NORMAL                -- Success
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine performs a 32-bit read access from a device PCI memory region
 *
 *===================================================================================================================*/
CEI_INT32 cei_read_32bit (CDEV_BOARD_TYPE board, pCEI_UINT32 address, pCEI_UINT32 data)
{
#ifdef NON_INTEL_WORD_ORDER
   CEI_UINT32 localData, modifiedData;
   localData = *address;
   modifiedData = little_endian_32(localData);
   *data = modifiedData;
#else
   *data = *address;
#endif          
   return ARS_NORMAL;
}

/*===================================================================================================================*
 * ENTRY POINT:                           C E I _ W R I T E _ 3 2 _ B I T
 *===================================================================================================================*
 *
 * FUNCTION:    Perform a 32-bit write to the specified device
 *
 * PARAMETERS:  CDEV_BOARD_TYPE deviceIndex  -- (input) device to reference (0 to 15, 127 for Windows)
 *              pCEI_UINT32 address          -- (input) location in host virtual memory to write
 *              CEI_UINT32 data              -- (input) the data value to write
 * 
 * RETURN VAL:  For Visa and LabVIEW R/T: -- Visa 'viOut32' invocation return status, if not VI_SUCCESS
 *              ARS_NORMAL                -- Success
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine performs a 32-bit write access to a device PCI memory region
 *
 *===================================================================================================================*/
CEI_INT32 cei_write_32bit (CDEV_BOARD_TYPE board, pCEI_UINT32 address, CEI_UINT32 data)
{
#ifdef NON_INTEL_WORD_ORDER
   *address = little_endian_32(data);
#else
   *address = data;
#endif          
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:   i n t e r r u p t S e r v i c e R o u t i n e
 *===========================================================================*
 *
 * DESCRIPTION: Invoked whenever an interrupt occurs, this routine clears 
 *              the hardware Interrupt Pending bit and invokes the ISR via 
 *              semaphore to launch "x30InterruptFunctionPtr".
 *
 *===========================================================================*/
void x30InterruptServiceRoutine (void)
{
   Error err;
   char dbuf[40];
   IODevice cboard;
   int device;
   short int headTail;

   /* retrieve the device index the ISR is assigned to */
   device = hwInterruptDeviceAssignment;

   /* redundant check we're dealing with a valid device index */
   if (hwInterruptDeviceAssignment < MAX_BOARDS) {
      
      /* connect to this device's driver resource id */
      sprintf (dbuf, "CEIboard%d", device);  
      err = RequestResource ((Object *)&cboard, dbuf, "!systempassword");
      if (err != Success) {
         printf("Error from resource acquisition\n");
         return;
      }
      
      /* this "isr" is really a task that blocks on any interrupt event for the resource id */
      while (TRUE) {
         err = SynchronousReceive ((Connection)cboard, NULL);
         if (err != Success) {
            printf("SynchronousReceive error:  %d - %s\n",err, ErrorString(err));
         }
   
         err = ReadIODeviceStatus (cboard, 3, (void *)&headTail, sizeof(headTail));
         if (err != Success) {
            printf("Error Calling 3 ReadIODeviceStatus %d - %s\n", err, ErrorString(err));
            continue;
         }

         /* if an event was encountered, invoke the specified ISR */
         (x30InterruptFunctionPtr[device])(device, 0);
      }	
   }
   return;
}


/*===================================================================================================================*
 * ENTRY POINT:                         C E I _ I N I T _ I N T E R R U P T
 *===================================================================================================================*
 *
 * FUNCTION:    Configures the device PCI interrupt enable and associates the supplied ISR function with it.
 *              
 * PARAMETERS:  CEI_INT32   deviceIndex       -- (input)  device to reference (0 to 127)
 *              CEI_INT32   _stdcall isr      -- (input)  ptr to the interrupt service routine
 *              CEI_INT32   priority          -- (input)  desired priority of the ISR thread
 *              pCEI_UINT32 deviceBaseAddress -- (input)  virtual address of BAR0/BAR2 region
 *              CEI_INT32   memRegion         -- (input)  PCI BAR region/index for interrupt clear bit (0 - BAR0 or 1 - BAR2)
 *              CEI_UINT32  disOffset         -- (input)  PCI Interrupt Disable register (PLX reg offset, 0 for no PLX interface)
 *              CEI_UINT32  interruptBit      -- (input)  Local Interrupt Input Enable bit for PLX interface control register
 *              CEI_UINT32  comboChannelCount -- (input)  1553/ARINC Combo Board configuration
 *              pCEI_UINT32 isrEnabled        -- (output) Thread successfully defined state (TRUE, FALSE)
 * 
 * RETURN VAL:  ARS_NORMAL         - Interrupt was initialized successfully
 *              ARS_DRIVERFAIL     - Device driver session never opened
 *              ARS_INT_ISR        - Interrupt was not initialized successfully
 *                                 - Driver interrupt event creation failure
 *                                 - Windows thread create failure
 *              ARS_NO_INT_SUPPORT - Driver interrupt handling not supported for this board
 *              ARS_NO_INT_ENABLED - Driver Interrupt handling not enabled for this board
 *
 * GLOBALS:     None
 *
 * DESCRIPTION: This routine enables the PCI interrupt, then connects the user-specified interrupt function   
 *              to the interrupt assigned to the respective device by the host o/s.
 *
 *===================================================================================================================*/
CEI_INT32 cei_init_interrupt 
( 
   CEI_INT32   deviceIndex,                /* (input)  device to reference (0 to 15, for Windows 127)                */
   pCEI_VOID   isr,                        /* (input)  ptr to the interrupt service routine                          */
   CEI_INT32   priority,                   /* (input)  desired priority of the ISR thread                            */         
   pCEI_UINT32 deviceBaseAddress,          /* (input)  virtual address of BAR0/BAR2 region                           */         
   CEI_INT32   memRegion,                  /* (input)  PCI BAR region/index for interrupt clear bit                  */
   CEI_UINT32  disOffset,                  /* (input)  PCI Interrupt Disable register                                */
   CEI_UINT32  interruptBit,               /* (input)  Local Interrupt Input Enable bit for PLX control register     */
   CEI_UINT32  comboChanCount,             /* (input)  1553/ARINC Combo Board configuration                          */         
   pCEI_UINT32 isrEnabled                  /* (output) thread handle if successfully defined                         */         
)
{
   CEI_UINT32 intCtrlStatReg;             /* PLX Int Control/Status Reg data */
   CEI_INT32 status = ARS_NORMAL;         /* local function return status    */
   /* local reference to the low-level global virtual base address array     */
   pCEI_UINT32 virtualBaseAddressPtr = deviceBaseAddress;
   char sbuf[40];
   Error err;

   /* allow only one device ISR assignment per API instance */
   if (deviceIndex < MAX_BOARDS) {
      if (hwInterruptDeviceAssignment == 99) {

         /* create a task that monitors for events for the specified device/resource */
         sprintf (sbuf,"intProcess%d",deviceIndex);
         err = CommonCreateTask (1, (Address)x30InterruptServiceRoutine, 0x1000, sbuf, &iTask[deviceIndex]);
         if (err != Success) {
            printf("Bad Status from CommonCreateTask: %d\n", (int)err);
            return ARS_FAILURE;
         }
      
         /* save the device index for the ISR assignment */
         hwInterruptDeviceAssignment = deviceIndex;
   
         /* launch the task we just created */
         err = RunTask (iTask[deviceIndex]);
         if (err != Success) {
            printf("Bad Status from RunTask: %d\n", (int)err);
            return ARS_FAILURE;
         }
         *isrEnabled = (CEI_UINT32)iTask[deviceIndex];
      }
   
      /* Set either the PLX PCI9056 or PCI9030 Interrupt Control/Status Register
       * Local Interrupt Input Enable, High to pass local bus interrupts
       */
      if (interruptBit != 0) {
         status = cei_read_32bit (deviceIndex, virtualBaseAddressPtr, &intCtrlStatReg);
         intCtrlStatReg |= interruptBit;
         status = cei_write_32bit (deviceIndex, virtualBaseAddressPtr, intCtrlStatReg);
      }
   }
   else {
      status = ARS_FAILURE;
   }
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ T E R M I N A T E _ I N T E R R U P T
 *===================================================================================================================*
 *
 * FUNCTION:    Provides the method to terminate an interrupt association to an ISR and clear the interrupt enable.
 * 
 * PARAMETERS:  CEI_INT32   deviceIndex       -- (input)  device to reference (0 to 127)
 *              CEI_UINT32  interruptBit      -- (input)  Local Interrupt Input Enable bit for PLX interface control register
 *              pCEI_UINT32 deviceBaseAddress -- (input)  virtual address of BAR2 region
 * 
 * RETURN VAL:  ARS_NORMAL   -- Interrupt was terminated successfully
 *
 * GLOBALS:     None
 * 
 * DESCRIPTION: This routine terminates the specified device interrupt function and clears the PCI interrupt enable.
 * 
 *===================================================================================================================*/
CEI_INT32 cei_terminate_interrupt
(
   CEI_INT32   deviceIndex,                /* (input)  device to reference (0 to 15, for Windows 127)                */
   CEI_UINT32  interruptBit,               /* (input)  Local Interrupt Input Enable bit for PLX control register     */
   pCEI_UINT32 deviceBaseAddress           /* (input)  virtual address of BAR0/BAR2 region                           */         
)
{
   IODevice cboard;
   char dbuf[40]; 
   Error err;
   CEI_UINT32 intCtrlStatReg;
   pCEI_UINT32 virtualBaseAddressPtr = deviceBaseAddress;
   CEI_INT32 status = ARS_NORMAL;
   
   /* Reset either the PLX PCI9056 or PCI9030 Interrupt Control/Status Register
    * Local Interrupt Input Enable, Low to disable local bus interrupts
    */
   if (interruptBit != 0) {
      status = cei_read_32bit (deviceIndex, virtualBaseAddressPtr, &intCtrlStatReg);
      intCtrlStatReg &= ~interruptBit;
      status = cei_write_32bit (deviceIndex, virtualBaseAddressPtr, intCtrlStatReg);
   }
   sprintf (dbuf, "CEIboard%d", deviceIndex);  
   err = RequestResource ((Object *)&cboard, dbuf, "!systempassword");       
   CommonCloseTask (iTask[deviceIndex]);
   hwInterruptDeviceAssignment = 99;
   if (err) ;
   return status;
}

