/*============================================================================*
 * FILE:                      C E I _ I N T _ P C I _ D R V . C
 *============================================================================*
 *
 *      COPYRIGHT (C) 2006 - 2016 BY ABACO SYSTEMS, INC. 
 *      ALL RIGHTS RESERVED.
 *
 *      THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND
 *      COPIED ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH
 *      THE INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY
 *      OTHER COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE
 *      AVAILABLE TO ANY OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF THE
 *      SOFTWARE IS HEREBY TRANSFERRED.
 *
 *      THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT
 *      NOTICE AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS.
 *
 *===========================================================================*
 *
 * FUNCTION:   Integrity PCIe/PCI device driver
 *
 *             This module provides the interface between the Abaco Systems
 *             Avionics BusTools/1553, CEI-x30, P-708 and P-SER APIs and the
 *             Integrity operating system.  It performs the actual I/O and
 *             memory-mapping functions needed to support operation in an
 *             Integrity environment.  Since this file is used by multiple
 *             products, use care when performing modifications.
 * 
 *
 * EXTERNAL ENTRY POINTS:
 * 
 *    CEIDevCreateIODevice
 *    
 *    CEIReadStatus
 *    
 *    CEIWriteStatus
 * 
 * 
 * EXTERNAL NON-USER ENTRY POINTS:
 * 
 *    cei_int_pci_drv_init    probes the PCI bus for supported Abaco Systems
 *                            Avionics PCI devices, then intitializes the device.
 *
 *    cei_int_pci_drv_match
 *    
 *    CEIDevHandlerUCA        the interrupt service routine for hardware
 *                            interrupts on boards with UCA32 or UCA firmware. 
 *
 *    CEIDevHandlerCOM        the interrupt service routine for hardware
 
 *                            interrupts on CEI-x30 or P-SER boards. 
 *  
 * INTERNAL ROUTINES: 
 * 
 *    CEIDeviceAllocate
 *
 *===========================================================================*
 *
 * Revision:   8.18
 *
 *    Date     Modifications
 * ----------  ---------------------------------------------------------------
 * 11/23/2005  Added support for CEI-x30 devices.  Added support to read a
 *             device's PCI config space registers.
 * 11/04/2008  BDW - Added CEI-430, RCEI-530, RAR-CPCI, R830RX, P-708 & P-SER 
 * 12/15/2008  BDW - Modified driver as follows:
 *                A) added a compiler directive NON_INTEL_WORD_ORDER around
 *                   Endian byte-swapping in CEIDevHandler
 *                B) corrected the type casting on destination for device PCI
 *                   configuration space read invocations in CEIReadStatus
 *                C) corrected the 1553 device recognition for h/w interrupt
 *                   initialization in cei_int_pci_drv_init().
 * 01/15/2010  Add support for the RXMC-1553.
 * 06/08/2012  Add RAR15-XMC and RAR15-XMC-XT
 * 11/12/2014  Change to use semaphore instead of connection
 * 07/11/2016  Added RPCIe-1553, R15-EC, R15-LPCIe, R15-MPCIe, CEI-R830X820,
 *             RCEI-830A, RP-708, RAR-XMC, and RAR-MPCIe. Removed RAR15-XMC,
 *             EPMC-1553 and PMC-D-1. Added CEIDevHandlerCOM. Merged
 *             CEIDevHandler and CEIDevHandlerV6 into CEIDevHandlerUCA. Added
 *             error checking and debugging. bch  
 *===========================================================================*/

#include <string.h>
#include "pci_config.h"
#include "bus.h"
#include "interrupt.h"


#define MAX_BTA  16

#define DEVICE_ID_1553_MASK   0x1500
#define DEVICE_ID_QPM_1553    0x1553
#define DEVICE_ID_QPCI_1553   0x1554
#define DEVICE_ID_QCP_1553    0x1555
#define DEVICE_ID_QPCX_1553   0x1556
#define DEVICE_ID_R15_EC      0x1557
#define DEVICE_ID_RXMC_1553   0x1559
#define DEVICE_ID_RPCIE_1553  0x155A
#define DEVICE_ID_R15_LPCIE   0x155B
#define DEVICE_ID_RXMC2_1553  0x155C
#define DEVICE_ID_RAR15_XT    0x1542
#define DEVICE_ID_R15_MPCIE   0x1545
#define DEVICE_ID_P_708       0x0708
#define DEVICE_ID_CEI_430     0x0430
#define DEVICE_ID_CEI_530     0x0530
#define DEVICE_ID_RAR_CPCI    0x0630
#define DEVICE_ID_CEI_830     0x0830
#define DEVICE_ID_R830RX      0x0831
#define DEVICE_ID_P_SER       0x1004
#define DEVICE_ID_R830X820    0x0832
#define DEVICE_ID_RCEI_830A   0x830A
#define DEVICE_ID_RP_708      0x708A
#define DEVICE_ID_RAR_XMC     0x100C
#define DEVICE_ID_RAR_MPCIE   0x100D

#define VENDOR_ID_CEI_AVIONICS 0x13C6

#define INTERRUPT_QUEUE_SIZE  64
#define INTERRUPT_QUEUE_MASK (INTERRUPT_QUEUE_SIZE - 1)

/* for CEIReadStatus */
#define STATUS_DEVICE_ID         1
#define STATUS_HOST_INTERFACE    2
#define STATUS_INTERRUPT_DATA    3
#define STATUS_PCI_CONFIG_SPACE  4

/* structures */
typedef struct CEIPCIDeviceStruct {
   struct IODeviceVectorStruct    device;
   InterruptHandler               handler;
   BusSpace                       config_space;
   BusHandle                      config_handle;
   BusSpace                       memory_space;
   BusHandle                      memory_handle[6];
   UINT4                          address;
   UINT4                          mem_size[6];
   UINT4                          mem_region_base[6];
   UINT4                          mem_base;
   UINT4                          mem_sections;
   UINT2                          devid;
   UINT4                          host_interface;
   UINT1                          interrupt_queue[INTERRUPT_QUEUE_SIZE];
   unsigned char                  head;
   unsigned char                  tail;
   UINT2                          devnum;
   DynamicMemoryRegionReservation bar_reservation[6];
   char                           name[40];
   char                           mem_res_name[6][40];
} CEIPCIDevice;

/* prototypes */
static Error CEIDeviceAllocate(UINT2 devId);
static Error CEIDevCreateIODevice(IODeviceVector iodev);
static Error CEIReadStatus(IODeviceVector iodev, Value StatusNumber, void* Buffer, Address Length);
static Error CEIWriteStatus(IODeviceVector iodev, Value StatusNumber, void* Buffer, Address Length);
static EVENT CEIDevHandlerUCA(XFRAME *xframe, InterruptHandler *handler);
static EVENT CEIDevHandlerCOM(XFRAME *xframe, InterruptHandler *handler);

/* module */
static Error cei_int_pci_drv_match(PciData* data, PciFunction* function, Value* fitness);
static Error cei_int_pci_drv_init(PciData* data, PciFunction* function);
PciDeviceDriver __ghsentry_pci_device_driver_cei_int_pci_drv = {
   "CEI Integrity PPC PCI Device Driver",  
   cei_int_pci_drv_match,  /* match */
   NULL,                   /* pre-init */
   NULL,                   /* requirements */
   cei_int_pci_drv_init,   /* init */
   NULL,                   /* device_name */
   NULL,                   /* device_bar */
};

/* globals */
static CEIPCIDevice cei_device[MAX_BTA];
static int num_devices=0;
/* set to "1" to enable output for debugging purposes */
static int debug=0;

/*===========================================================================*
 * ENTRY POINT:    cei_int_pci_drv_match
 *===========================================================================*
 *
 * FUNCTION:    Will be called for each PCI device detected.
 *
 * PARAMETERS:  PciData *data         -- (input) PCI header info on this device
 *              PciFunction *function -- (input) o/w function info on this device
 *              Value *fitness        -- (output)
 *  
 * RETURN VAL:  Success
 *              Failure
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: 
 *
 *===========================================================================*/
static Error cei_int_pci_drv_match(PciData* data, PciFunction* function, Value* fitness) {
  /* check for a valid manufacturer ID */  
  if((function->id & 0xFFFF) != VENDOR_ID_CEI_AVIONICS)
    return Failure;
  
  *fitness = 128;
  
  return Success;
}


/*===========================================================================*
 * ENTRY POINT:    C E I _ I N T _ P C I _ D R V _ I N I T
 *===========================================================================*
 *
 * FUNCTION:    Perform initialization on supported boards
 *
 * PARAMETERS:  PciData * data         -- (input) PCI header info on this device
 *              PciFunction * function -- (input) o/w function info on this device
 * 
 * RETURN VAL:  Success
 *              Failure
 *              Error - Integrity-specific driver error codes
 *
 * DESCRIPTION: This routine is invoked by the kernel for each instance of a 
 *              detected PCI device.
 *
 *===========================================================================*/
static Error cei_int_pci_drv_init(PciData *data, PciFunction *function) {
  Error status;
  UINT4 mask=0, bar_offset=0, bar_addr=0, pciConfigCommand=0;
  int indx=0, pciInterruptLine=0;
  VECT interrupt=0;
  UINT2 devId=0;
  CEIPCIDevice *device=NULL;

  /* extract the PCI Device ID for this device */
  devId = (UINT2)((function->id & 0xFFFF0000)>>16);
  
  if(debug)
    db_printf("CEI PCI Driver: Abaco Systems avionics device detected (device ID 0x%X)\n", devId);

  if((status = CEIDeviceAllocate(devId)) != Success)
    return status;
  
  device = &cei_device[num_devices];
  device->devnum = num_devices;
  device->devid = devId;
  device->config_space = data->config_space;
  device->config_handle = function->handle;
  device->memory_space = data->memory_space;
  device->head = 0;  /*initialize interrupt FIFO head pointer */
  device->tail = 0;  /*initialize interrupt FIFO tail pointer */
  /* determine the PCI BAR configuration and map each defined & desired region */
  for(indx=0; indx<6; indx++) {
    /* if BAR1 (I/O Region) then skip since memory mapping can cause the driver to fail */
    if(indx == 1)
      continue;
    
    bar_offset = (indx * 4) + 0x10;  /* BAR region offset	*/
    /* get the BAR address, if it doesn't exist, go to the next one  */
    if((bar_addr = BusRead4B(device->config_space, device->config_handle, bar_offset)) != 0) {
      /* determine BAR region size, limit to 4Kb minimum */
      BusWrite4B(device->config_space, device->config_handle, bar_offset, ~0);
      mask = BusRead4B(device->config_space, device->config_handle, bar_offset);
      BusWrite4B(device->config_space, device->config_handle, bar_offset, bar_addr);
      device->mem_region_base[indx] = bar_addr & ~0xf;  /* Strip non-address bits */
      device->mem_size[indx] = ~mask+1;  /* Find the memory region size */
      /* Limit BAR size smaller than 4Kb page */
      if(device->mem_size[indx] < 0x1000)
        device->mem_size[indx] = 0x1000;
      if(debug)
        db_printf("CEI PCI Driver (dev%d): BAR%d = phys addr 0x%X, size %d\n", device->devnum, indx, device->mem_region_base[indx], device->mem_size[indx]);
   
      /* map the BAR region  */
      if(BusMap(device->memory_space, device->mem_region_base[indx], device->mem_size[indx], 0, &device->memory_handle[indx])) {
        db_printf("CEI PCI Driver (dev%d): failed to create Bus Handle for BAR%d\n", device->devnum, indx);
        return Failure;
      }

      /* check for a Q104-1553P */
      if((indx == 0) && (devId == DEVICE_ID_QPM_1553)) {
        if((BusRead2B(device->memory_space, device->memory_handle[0], 0) & 0x3E) == 0x0A)
          strcpy(cei_device[device->devnum].mem_res_name[0], "Q1041553P_BASE0");
      }  
        
      /* record the information required to access these memory regions later */
      device->bar_reservation[indx].First = (Address)device->memory_handle[indx];
      device->bar_reservation[indx].Last = (Address)device->memory_handle[indx] + device->mem_size[indx] - 1;
      device->bar_reservation[indx].PhysicalSpaceNumber = 0;
      device->bar_reservation[indx].RegistrationName = device->mem_res_name[indx];
      if(debug)
        db_printf("CEI PCI Driver (dev%d): BAR%d RegistrationName = %s\n", device->devnum, indx, device->bar_reservation[indx].RegistrationName);
      if(RegisterDynamicMemoryRegion(&device->bar_reservation[indx]) != Success) {
        db_printf("CEI PCI Driver (dev%d): failed to register BAR%d\n", device->devnum, indx);
        return Failure;
      }

      device->mem_sections++;
    }
  }
   
  /* The devices supported by this driver will have 1 (BAR0) or 2 (BAR0 and BAR2) PCI memory regions */
  if((device->mem_sections == 0) || (device->mem_sections > 2)) {
    db_printf("CEI PCI Driver (dev%d): invalid memory sections detected (%d)\n", device->devnum, device->mem_sections);
    return Failure;
  }
  indx = device->mem_sections & 0x2;
  device->mem_base = device->mem_region_base[indx];

  /* set the device PCI Command Word for Memory space access but no I/O space access */
  pciConfigCommand = (BusRead4B(device->config_space, device->config_handle, 0x04) & 0xFFFF) | 0x02;
  BusWrite4B(device->config_space, device->config_handle, 0x04, pciConfigCommand);

  /* read the CSC regiser on the device */
  device->host_interface = BusRead4B(device->memory_space, device->memory_handle[indx], 0) & 0xFFFF;
  if(debug)
    db_printf("CEI PCI Driver (dev%d): host interface = 0x%X\n", device->devnum, device->host_interface);
    
  /* setup for hardware interrupts, supported on all boards except P-708 and RP-708 */
  if((device->devid != DEVICE_ID_P_708) && (device->devid != DEVICE_ID_RP_708)) {
    /* read the Interrupt Line assignment from the PCI Config Space */
    pciInterruptLine = BusRead4B(device->config_space, device->config_handle, 0x3C) & 0xFF;
    if(BSP_PciIrqLineToVector(pciInterruptLine, &interrupt) != Success) {
      db_printf("CEI PCI Driver (dev%d): failed to get vector for IRQ %d\n", device->devnum, pciInterruptLine);
      return Failure;
    }
    if(debug)
      db_printf("CEI PCI Driver (dev%d): using vector %d for IRQ %d\n", device->devnum, interrupt, pciInterruptLine);

    /* set the ISR */
    if((device->devid & DEVICE_ID_1553_MASK) == DEVICE_ID_1553_MASK)
      device->handler.handler = CEIDevHandlerUCA;
    else 
      device->handler.handler = CEIDevHandlerCOM;

    device->handler.vect = interrupt;
    device->handler.id = (Address) device;
    SetChainedExceptionHandler(&device->handler, device->name);
    SetChainedExceptionPriority(&device->handler, 1);
  }

  device->address = function->location;

  /* Register this new device with the system. */
  device->device.Create                = CEIDevCreateIODevice;
  device->device.ReadBlock             = NULL;
  device->device.WriteBlock            = NULL;
  device->device.ReadBuffers           = NULL;
  device->device.WriteBuffers          = NULL;
  device->device.ReadRegister          = NULL;
  device->device.WriteRegister         = NULL;
  device->device.ReadStatus            = CEIReadStatus;
  device->device.WriteStatus           = CEIWriteStatus;
  device->device.IOCoherentNotRequired = 1;
  device->device.IOSynchronizationNotRequired = 0;
  device->device.DeviceSpecificInfo    = (void *)device;
  if((status = RegisterIODeviceVector(&device->device, device->name)) != Success) {
    db_printf("CEI PCI Driver (dev%d): failed to register %s, status: %d\n", device->devnum, device->name, status);
    return Failure;
  }
  
  num_devices++;
  if(debug)
    db_printf("CEI PCI Driver (dev%d): registered %s\n", device->devnum, device->name);
    
  return Success;
}


/*===========================================================================*
 * ENTRY POINT:    CEIDeviceAllocate
 *===========================================================================*
 *
 * FUNCTION:   Will be called by "cei_int_pci_drv_init"  
 *
 * PARAMETERS:  UINT2 devId -- (input)
 * 
 * RETURN VAL:
 *              Success
 *              Failure
 *              MaximumNumberOfResourcesExceeded 
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: assign a device name and memory resource names to this device
 *
 *===========================================================================*/
static Error CEIDeviceAllocate(UINT2 devId) {
  char res_name[2][40];
    
  if(debug)
    db_printf("CEI PCI Driver (dev%d): CEIDeviceAllocate\n", num_devices);

  /* check if maximum supported boards have been reached */ 
  if(num_devices >= MAX_BTA)
    return MaximumNumberOfResourcesExceeded;

  sprintf(cei_device[num_devices].name, "CEIboard%d (0x%X)", num_devices, devId);
    
  memset(res_name, 0, sizeof(res_name));
    
  switch(devId) {
   case DEVICE_ID_QPM_1553:  /* QPM-1553, QPMC-1553 and Q104-1553P */ 
     sprintf(res_name[0], "QPM1553_BASE%d", num_devices);
     break;
   case DEVICE_ID_QPCI_1553:
     sprintf(res_name[0], "QPCI1553_PLX_CONFIG_BASE%d", num_devices);
     sprintf(res_name[1], "QPCI1553_BASE%d", num_devices);
     break;
   case DEVICE_ID_QCP_1553:
     sprintf(res_name[0], "QCP1553_PLX_CONFIG_BASE%d", num_devices);
     sprintf(res_name[1], "QCP1553_BASE%d", num_devices);
     break;
   case DEVICE_ID_QPCX_1553:
     sprintf(res_name[0], "QPCX1553_PLX_CONFIG_BASE%d", num_devices);
     sprintf(res_name[1], "QPCX1553_BASE%d", num_devices);
     break;
   case DEVICE_ID_R15_EC:
     sprintf(res_name[0], "R15_EC_BASE%d", num_devices);
     break;
   case DEVICE_ID_RXMC_1553:
     sprintf(res_name[0], "RXMC1553_BASE%d", num_devices);
     break;
   case DEVICE_ID_RPCIE_1553:
     sprintf(res_name[0], "RPCIE1553_BASE%d", num_devices);
     break;
   case DEVICE_ID_R15_LPCIE:
     sprintf(res_name[0], "R15_LPCIE_BASE%d", num_devices);
     break;
   case DEVICE_ID_RXMC2_1553:
     sprintf(res_name[0], "RXMC21553_BASE%d", num_devices);
     break;
   case DEVICE_ID_RAR15_XT:
     sprintf(res_name[0], "RAR15_XMC_XT_BASE%d", num_devices);
     break;
   case DEVICE_ID_R15_MPCIE:
     sprintf(res_name[0], "R15_MPCIE_BASE%d", num_devices);
     break;   
   case DEVICE_ID_CEI_430:
     sprintf(res_name[0], "CEI430_PLX_CONFIG_BASE%d", num_devices);
     sprintf(res_name[1], "CEI430_BASE%d", num_devices);
     break;
   case DEVICE_ID_CEI_530:
     sprintf(res_name[0], "CEI530_PLX_CONFIG_BASE%d", num_devices);
     sprintf(res_name[1], "CEI530_BASE%d", num_devices);
     break;
   case DEVICE_ID_RAR_CPCI:
     sprintf(res_name[0], "RAR_CPCI_PLX_CONFIG_BASE%d", num_devices);
     sprintf(res_name[1], "RAR_CPCI_BASE%d", num_devices);
     break;
   case DEVICE_ID_CEI_830:
     sprintf(res_name[0], "CEI830_PLX_CONFIG_BASE%d", num_devices);
     sprintf(res_name[1], "CEI830_BASE%d", num_devices);
     break;
   case DEVICE_ID_R830RX:
     sprintf(res_name[0], "R830RX_PLX_CONFIG_BASE%d", num_devices);
     sprintf(res_name[1], "R830RX_BASE%d", num_devices);
     break;
   case DEVICE_ID_R830X820:
     sprintf(res_name[0], "R830X820_CONFIG_BASE%d", num_devices);
     sprintf(res_name[1], "R830X820_BASE%d", num_devices);
     break;     
   case DEVICE_ID_RCEI_830A:
     sprintf(res_name[0], "RCEI_830A_CONFIG_BASE%d", num_devices);
     sprintf(res_name[1], "RCEI_830A_BASE%d", num_devices);
     break;
   case DEVICE_ID_RAR_XMC:
     sprintf(res_name[0], "RAR_XMC_BASE%d", num_devices);
     break;     
   case DEVICE_ID_RAR_MPCIE:
     sprintf(res_name[0], "RAR_MPCIE_BASE%d", num_devices);
     break;     
   case DEVICE_ID_P_708:
     sprintf(res_name[0], "P708_PLX_CONFIG_BASE%d", num_devices);
     sprintf(res_name[1], "P708_BASE%d", num_devices);
     break;
   case DEVICE_ID_RP_708:
     sprintf(res_name[0], "RP_708_CONFIG_BASE%d", num_devices);
     sprintf(res_name[1], "RP_708_BASE%d", num_devices);
     break;     
   case DEVICE_ID_P_SER:
     sprintf(res_name[0], "PSER_PLX_CONFIG_BASE%d", num_devices);
     sprintf(res_name[1], "PSER_BASE%d", num_devices);
     break;     
   default:
     if(debug)
       db_printf("CEI PCI Driver (dev%d): detected unsupported board, devID 0x%X\n", num_devices, devId);
     return Failure;
  };

  strcpy(cei_device[num_devices].mem_res_name[0], res_name[0]);
  strcpy(cei_device[num_devices].mem_res_name[2], res_name[1]);
  
  return Success;
}


/*===========================================================================*
 * ENTRY POINT:    CEIDevCreateIODevice
 *===========================================================================*
 *
 * FUNCTION:     
 *
 * PARAMETERS:  IODeviceVector iodev -- (input)
 * 
 * RETURN VAL:  Success
 *              Failure
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: initializes the device 
 *
 *===========================================================================*/
static Error CEIDevCreateIODevice(IODeviceVector iodev) {
  CEIPCIDevice *device=(CEIPCIDevice*)iodev->DeviceSpecificInfo;

  if(debug)
    db_printf("CEI PCI Driver (dev%d): CEIDevCreateIODevice\n", device->devnum);

  device->tail = 0;
  device->head = 0;

  return Success;
}


/*===========================================================================*
 * ENTRY POINT:    CEIReadStatus
 *===========================================================================*
 *
 * FUNCTION:    Read board information    
 *
 * PARAMETERS:  IODeviceVector iodev  -- (input) 
 *              Value StatusNumber    -- (input) the status to read 
 *              void* Buffer          -- (output)
 *              Address Length        -- (input) length for reading PCI config space
 * 
 * RETURN VAL:  Success
 *              Failure
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: retrieves the device ID, the host interface, PCI config space or
 *              hardware interrupt status for a board. 
 *
 *===========================================================================*/
static Error CEIReadStatus(IODeviceVector iodev, Value StatusNumber, void* Buffer, Address Length) {
  CEIPCIDevice *device=(CEIPCIDevice*)iodev->DeviceSpecificInfo;
  int offset=0;

/* db_printf("CEI PCI Driver (dev%d): CEIReadStatus, status number %d\n", device->devnum, StatusNumber); */

  switch(StatusNumber) {
   case STATUS_DEVICE_ID:  /* return the device ID */
     *((UINT2*)Buffer) = device->devid;
     break;
   case STATUS_HOST_INTERFACE:  /* returns the host interface register */
     *((UINT2*)Buffer) = device->host_interface;
     break;
   case STATUS_INTERRUPT_DATA:
     if(device->tail != device->head) {
       *((UINT1*)Buffer) = device->interrupt_queue[device->tail];
       device->tail++;
       device->tail &= INTERRUPT_QUEUE_MASK;
     }
     break;
   case STATUS_PCI_CONFIG_SPACE:  /* returns the PCI config space register with the user supplied offset   */
     offset = (*((UINT4*)Buffer)*Length);    
     switch(Length) {
      case 1:
        *((UINT4*)Buffer) = BusRead1(device->config_space,device->config_handle,(Value)offset);
        break;
      case 2:
        *((UINT4*)Buffer) = BusRead2(device->config_space,device->config_handle,(Value)offset);
        break;
      case 4:       
        *((UINT4*)Buffer) = BusRead4(device->config_space,device->config_handle,(Value)offset);
        break;
      default:
        *((UINT4*)Buffer) = ~0;
     };
     break;
   default:
     db_printf("CEI PCI Driver (dev%d): invalid status number %d\n", device->devnum, StatusNumber);
     return Failure;
   };

   return Success;
}


/*===========================================================================*
 * ENTRY POINT:    CEIWriteStatus
 *===========================================================================*
 *
 * FUNCTION:    Write board information 
 *
 * PARAMETERS:  IODeviceVector iodev  -- (input) 
 *              Value StatusNumber    -- (input) the status to read 
 *              void* Buffer          -- (input)
 *              Address Length        -- (input) length for reading PCI config space
 * 
 * RETURN VAL:  Success
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: not used 
 *
 *===========================================================================*/
static Error CEIWriteStatus(IODeviceVector iodev, Value StatusNumber, void* Buffer, Address Length) {
  CEIPCIDevice *device=(CEIPCIDevice*)iodev->DeviceSpecificInfo;
  
  if(debug)
    db_printf("CEI PCI Driver (dev%d): CEIWriteStatus, status number %d\n", device->devnum, StatusNumber);

  return Success;
}


/*===========================================================================*
 * ENTRY POINT:    CEIDevHandlerUCA
 *===========================================================================*
 *
 * FUNCTION:    ISR for UCA and UCA32 firmware boards   
 *
 * PARAMETERS:  XFRAME *xframe            -- (input) 
 *              InterruptHandler *handler -- (input)
 * 
 * RETURN VAL:  EVENT_HANDLED
 *              EVENT_UNKNOWN
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: 
 *
 *===========================================================================*/
static EVENT CEIDevHandlerUCA(XFRAME *xframe, InterruptHandler *handler) {
  CEIPCIDevice *device=(CEIPCIDevice*)handler->id;
  UINT4 nchan=0, handled=0, chan_addr=0, chan_offset=0, reg_offset=0, temp=0;
  UINT1 indx=0, bar_region=0;

  
  if(device->host_interface & 0x4000) {
    chan_offset = 0x4000;
    reg_offset = 0x30058;
  }
  else {
    /* determine the BAR region */
    bar_region = device->mem_sections & 0x2;
    chan_offset = 0x200000;
    reg_offset = 0x800; 
  }
  
  /* maximum number of channels is 4 */
  if((nchan = (device->host_interface >> 6) & 0x1F) > 4)
    nchan = 4;
          
  /* need to check each channel on the board for interrupts */
  for(indx = 0;indx < nchan; indx++) {
    /* byte address for "1553_control_reg" register */
    chan_addr = (indx * chan_offset) + reg_offset;
    /* check for interrupt: "CPU interrupt" bit */
    if(BusRead2B(device->memory_space, device->memory_handle[bar_region], chan_addr) & 0x200) {
      /* clear the interrupt: "write_interrupt_bit" register */
      if(device->host_interface & 0x4000)
        BusWrite4B(device->memory_space, device->memory_handle[bar_region], chan_addr, 0x40000200); /* UCA32 */
      else
        BusWrite2B(device->memory_space, device->memory_handle[bar_region], chan_addr + 0x16, 0x0); /* UCA */
      /* signal the user app that there was an interrupt */
      device->interrupt_queue[device->head] = indx;
      device->head++;
      device->head &= INTERRUPT_QUEUE_MASK;
      handled++;
    }
  }

  /* if a multi-protocol board then check for an ARINC interrupt */
  if(device->devid == DEVICE_ID_RAR15_XT) {
    /* read the "Global Enable Register" anc check that both the "Interrupt Pending" and
     * "Interrupt Enable" bits are set */
    if((temp = BusRead4B(device->memory_space,device->memory_handle[bar_region], 0x100000)) & 0x30) {
      BusWrite4B(device->memory_space,device->memory_handle[bar_region], 0x100000, temp);
      /* signal the user app that there was an interrupt */
      device->interrupt_queue[device->head] = 0xff;
      device->head++;
      device->head &= INTERRUPT_QUEUE_MASK;
      handled++;
    }
  }

  
/*
  if(debug)
    db_printf("CEI PCI Driver (dev%d): CEIDevHandlerUCA, head: %d\n", device->devnum, device->head);
*/
    
  if(handled) {
    INTERRUPT_IODeviceNotifyTask(&device->device);
    return EVENT_HANDLED;
  }

  return EVENT_UNKNOWN;
}


/*===========================================================================*
 * ENTRY POINT:   CEIDevHandlerCOM 
 *===========================================================================*
 *
 * FUNCTION:    ISR for CEI-x30 and P-SER boards
 *
 * PARAMETERS:  XFRAME *xframe            -- (input) 
 *              InterruptHandler *handler -- (input)
 * 
 * RETURN VAL:  EVENT_HANDLED
 *              EVENT_UNKNOWN
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: 
 *
 *===========================================================================*/
static EVENT CEIDevHandlerCOM(XFRAME *xframe, InterruptHandler *handler) {
  CEIPCIDevice *device=(CEIPCIDevice*)handler->id;
  UINT4 temp=0;
  int bar_region=0;

  /* determine the BAR region */
  bar_region = device->mem_sections & 0x2;
  
  /* read the "Global Enable Register" */
  if((temp = BusRead4B (device->memory_space, device->memory_handle[bar_region], 0)) & 0x30 ) {
    /* need to write back the values along with clearing the interrupt */
    BusWrite4B (device->memory_space, device->memory_handle[bar_region], 0, temp);
      
    /* signal the user app that there was an interrupt */
    device->head++;
    device->head &= INTERRUPT_QUEUE_MASK;

    INTERRUPT_IODeviceNotifyTask(&device->device);
    
    return EVENT_HANDLED;
  }

  return EVENT_UNKNOWN;
}

