/*============================================================================*
 * FILE:                     M E M _ I N T E G R I T Y . C
 *============================================================================*
 *
 *      COPYRIGHT (C) 2001 - 2016 BY ABACO SYSTEMS, INC.
 *      ALL RIGHTS RESERVED.
 *
 *      THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND
 *      COPIED ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH
 *      THE INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY
 *      OTHER COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE
 *      AVAILABLE TO ANY OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF THE
 *      SOFTWARE IS HEREBY TRANSFERRED.
 *
 *      THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT
 *      NOTICE AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS.
 *
 *===========================================================================*
 *
 * FUNCTION:   Integrity PCI device driver interface functions.
 *
 *             This module provides the interface between the Abaco Systems
 *             Avionics BusTools/1553, CEI-x30, P-708 and P-SER APIs and the
 *             Integrity operating system.  It performs the actual I/O and
 *             memory-mapping functions needed to support operation in an
 *             Integrity environment.  Since this file is used by multiple
 *             products, use care when performing modifications.
 *
 *
 * EXTERNAL ENTRY POINTS:
 *
 *    vbtMapBoardAddresses    This routine acquires the location for the board
 *                            PCI memory region(s), then requests the kernel 
 *                            to map the region(s) and returns pointer(s)
 *                            that may be used to directly access the memory-
 *                            mapped board PCI BAR(s).  
 *
 *    vbtFreeBoardAddresses   Free any system resources which were obtained in
 *                            the mapping of the phyical board address.
 *
 *    vbtGetPCIConfigRegister Reads the specified device's PCI configuration
 *                            register. 
 *                            
 *    vbtGetDevInfo           Retrieves specific information concerning the 
 *                            device.
 *
 *    vbtMapBoardAddress      Maps the device A32 VME physical base address
 *                            for a fixed size applicable to the (R)QVME-1553
 *                            board (VME only).
 *
 *    vbtMapIoAddress         Maps the device A16 VME physical base address
 *                            for a fixed size applicable to the (R)QVME-1553
 *                            board (VME only).
 *
 *    vbtFreeBoardAddress     Releases the A16 and A32 VME base addresses.
 *
 * EXTERNAL NON-USER ENTRY POINTS:
 *
 * INTERNAL ROUTINES:
 *
 *===========================================================================*
 *
 * $Revision:  1.07 Release $
 *
 *    Date     Modifications
 * ----------  ---------------------------------------------------------------
 * 06/04/2001  Configure current API release
 * 11/25/2003  Add support for VME SUP and USR mapping
 * 11/15/2005  Add support for CEI-x30 API.  Added vbtGetPCIConfigRegister,
 *             vbtFreeBoardAddresses and vbtGetDevInfo. Added error checking for
 *             Integrity system calls.
 * 04/07/2009  Completed integration of CEI-830, P-SER, and P-708 devices.
 * 09/17/2012  Added support for the new XMC 1553 and Combo 1553/429 devices.
 * 09/25/2013  Added support for the new x30 devices.
 * 08/16/2016  Rebranded and added Mini-PCIe boards
 * 09/28/2016  revised. bch
 *===========================================================================*/

/*---------------------------------------------------------------------------*
 *                    INCLUDES, DEFINES and GLOBALS
 *---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <integrity.h>
#include <memory_region.h>
#include <util/error_string.h>
#ifdef INTEGRITY_VME_PPC
#include <vme/vme_api.h>
#endif
#include "lowlevel.h"


/* This is the password used when requesting resources, modify if needed.  The
 * default system password is "!systempassword".
 */
#define RESOURCE_PASSWORD  "!systempassword"

/*---------------------------------------------------------------------------*
 * Constants for identifying supported Abaco Systems Avionics devices
 *---------------------------------------------------------------------------*/
#define DEVICE_ID_1553_MASK   0x1500
#define DEVICE_ID_QPM_1553    0x1553
#define DEVICE_ID_QPCI_1553   0x1554
#define DEVICE_ID_QCP_1553    0x1555
#define DEVICE_ID_QPCX_1553   0x1556
#define DEVICE_ID_R15_EC      0x1557
#define DEVICE_ID_RXMC_1553   0x1559
#define DEVICE_ID_RPCIE_1553  0x155A
#define DEVICE_ID_R15_LPCIE   0x155B
#define DEVICE_ID_RXMC2_1553  0x155C
#define DEVICE_ID_RAR15_XT    0x1542
#define DEVICE_ID_R15_MPCIE   0x1545
#define DEVICE_ID_P_708       0x0708
#define DEVICE_ID_CEI_430     0x0430
#define DEVICE_ID_CEI_530     0x0530
#define DEVICE_ID_RAR_CPCI    0x0630
#define DEVICE_ID_CEI_830     0x0830
#define DEVICE_ID_R830RX      0x0831
#define DEVICE_ID_P_SER       0x1004
#define DEVICE_ID_R830X820    0x0832
#define DEVICE_ID_RCEI_830A   0x830A
#define DEVICE_ID_RP_708      0x708A
#define DEVICE_ID_RAR_XMC     0x100C
#define DEVICE_ID_RAR_MPCIE   0x100D

#define VENDOR_ID_CEI_AVIONICS  0x13C6

/*---------------------------------------------------------------------------*
 * Constants for the driver interface CEIReadStatus
 * A matching set of constants exists in the device driver file
 *---------------------------------------------------------------------------*/
#define STATUS_DEVICE_ID        1
#define STATUS_HOST_INTERFACE   2
#define STATUS_INTERRUPT_DATA   3
#define STATUS_PCI_CONFIG_SPACE 4

#ifdef _CEI_DEBUG
#define ceiprint(a,b,c) printf(a,b,c)
#define cei_err(_1_,_2_,_3_) printf("Fail (dev%d):  %s error %d - %s. Line %d in %s\n", _1_, _2_, _3_, ErrorString(_2_), __line__, __func__)
#else
#define ceiprint(a,b,c)
#define cei_err(_1_,_2_,_3_)
#endif


#ifdef INTEGRITY_VME_PPC
//extern Error CreateWindow(Value num, Value am, Address vme_start, Address length, MemoryRegion *mr, Address *pci_start_p);
//extern Error InitializeVMEBridge();
static int One_Time_Init=0;
static int vme_device_use[MAX_DEVICES];
static int vme_device_cnt=0;
static MemoryRegion vme_mem[MAX_DEVICES][2];
#endif

#if defined(INTEGRITY_PCI_PPC) || defined(INTEGRITY_PCI_X86)
typedef struct DEVICE_MAP_INFO {
  int deviceBarCount;  /* Number of BARs on device */
  unsigned long *deviceBarLocation[6];  /* Plug'n'Play Address to which each BAR is mapped */
}DEVICE_MAP_INFO_TYPE;

extern MemoryRegionPool* __ghs_VirtualMemoryRegionPool;
static MemoryRegion v_mem_regions[MAX_DEVICES][MAX_MEMORY];  /* holds the virtual memory addresses for each mapped BAR region */
static DEVICE_MAP_INFO_TYPE devInfo[MAX_DEVICES];

/****************************************************************************
*
*  PROCEDURE - vbtMapBoardAddresses
*
*  FUNCTION
*     This routine acquires the location for the board PCI memory region(s), 
*     then requests the kernel to map the region(s) and returns pointer(s)
*     that may be used to directly access the memory-mapped board PCI BAR(S).  
*
*  PARAMETERS
*     device          -  device to open (0 - 15)
*     PDEVMAP_T pMap  - pointer to region mapping structure
*
*  RETURNS
*     BTD_OK          - Mapped address with no error
*     SYSTEM_ERROR    - any errors
*     
****************************************************************************/
int vbtMapBoardAddresses(UINT device, PDEVMAP_T pMap) {
  Error err;
  int i, status=BTD_OK;
  char dbuf[20]="";
  char res_name[2][40];
  UINT2 boardType=0, wVal=0;
  IODevice cboard;
  MemoryRegion p_mem_reg;
  unsigned long lpbase=0;
  Value mem_size=0;

 
  /* if device has already mapped, then get the host address */
  if(pMap->use_count != 0) {
    if((devInfo[device].deviceBarCount == 1) || (devInfo[device].deviceBarCount == 2)) {
      pMap->memHostBase[0] = (char*)devInfo[device].deviceBarLocation[0]; /* hack for multiprotocol access */
      pMap->memHostBase[1] = (char*)devInfo[device].deviceBarLocation[1];
    }
  }
  else {   
    memset(res_name, 0, sizeof(res_name));
    devInfo[device].deviceBarCount = 0;
       
    /* obtain the resource handle from the kernel */
    sprintf(dbuf, "CEIboard%d", device);  
    if((err = RequestResource((Object*)&cboard, dbuf, RESOURCE_PASSWORD)) != Success) {
      cei_err(device, "RequestResource", err);
      status = BTD_RESOURCE_ERR;
    }
    
    if(status == BTD_OK) {
      /* read PCI configuration space word 1 - the PCI Device ID */
      if((err = ReadIODeviceStatus(cboard, STATUS_DEVICE_ID, (void*)&boardType, sizeof(boardType))) != Success) {
        cei_err(device, "ReadIODeviceStatus", err);
        status = BTD_READ_IODEV_ERR;
      }
    }

    if(status == BTD_OK) {
      /* Reference the device memory resource by name based on the board type.
       * Note the name strings must match those used in the driver CEI_INT_PCI_DRV.C
       */
      switch(boardType) {
        case DEVICE_ID_QPM_1553:
          /* read the CSC */
          if((err = ReadIODeviceStatus(cboard, STATUS_HOST_INTERFACE, (void*)&wVal, sizeof(wVal))) != Success) {
            cei_err(device, "ReadIODeviceStatus", err);
            status = BTD_READ_IODEV_ERR;
          }
          else {
            if((wVal & 0x3E) == 0xA)  /* Q104-1553P */
              sprintf(res_name[0], "Q1041553P_BASE%d", device);
            else  /* QPM-1553 and QPMC-1553 */
              sprintf(res_name[0], "QPM1553_BASE%d", device);
          }
          break;
        case DEVICE_ID_QPCI_1553:
          sprintf(res_name[0], "QPCI1553_PLX_CONFIG_BASE%d", device);
          sprintf(res_name[1], "QPCI1553_BASE%d", device);
          break;
        case DEVICE_ID_QCP_1553:
          sprintf(res_name[0], "QCP1553_PLX_CONFIG_BASE%d", device);
          sprintf(res_name[1], "QCP1553_BASE%d", device);
          break;
        case DEVICE_ID_QPCX_1553:
          sprintf(res_name[0], "QPCX1553_PLX_CONFIG_BASE%d", device);
          sprintf(res_name[1], "QPCX1553_BASE%d", device);
          break;
        case DEVICE_ID_R15_EC:
          sprintf(res_name[0], "R15_EC_BASE%d", device);
          break;
        case DEVICE_ID_RXMC_1553:
          sprintf(res_name[0], "RXMC1553_BASE%d", device);
          break;
        case DEVICE_ID_RPCIE_1553:
          sprintf(res_name[0], "RPCIE1553_BASE%d", device);
          break;
        case DEVICE_ID_R15_LPCIE:
          sprintf(res_name[0], "R15_LPCIE_BASE%d", device);
          break;
        case DEVICE_ID_RXMC2_1553:
          sprintf(res_name[0], "RXMC21553_BASE%d", device);
          break;
        case DEVICE_ID_RAR15_XT:
          sprintf(res_name[0], "RAR15_XMC_XT_BASE%d", device);
          break;
        case DEVICE_ID_R15_MPCIE:
          sprintf(res_name[0], "R15_MPCIE_BASE%d", device);
          break;   
        case DEVICE_ID_CEI_430:
          sprintf(res_name[0], "CEI430_PLX_CONFIG_BASE%d", device);
          sprintf(res_name[1], "CEI430_BASE%d", device);
          break;
        case DEVICE_ID_CEI_530:
          sprintf(res_name[0], "CEI530_PLX_CONFIG_BASE%d", device);
          sprintf(res_name[1], "CEI530_BASE%d", device);
          break;
        case DEVICE_ID_RAR_CPCI:
          sprintf(res_name[0], "RAR_CPCI_PLX_CONFIG_BASE%d", device);
          sprintf(res_name[1], "RAR_CPCI_BASE%d", device);
          break;
        case DEVICE_ID_CEI_830:
          sprintf(res_name[0], "CEI830_PLX_CONFIG_BASE%d", device);
          sprintf(res_name[1], "CEI830_BASE%d", device);
          break;
        case DEVICE_ID_R830RX:
          sprintf(res_name[0], "R830RX_PLX_CONFIG_BASE%d", device);
          sprintf(res_name[1], "R830RX_BASE%d", device);
          break;
        case DEVICE_ID_R830X820:
          sprintf(res_name[0], "R830X820_CONFIG_BASE%d", device);
          sprintf(res_name[1], "R830X820_BASE%d", device);
          break;     
        case DEVICE_ID_RCEI_830A:
          sprintf(res_name[0], "RCEI_830A_CONFIG_BASE%d", device);
          sprintf(res_name[1], "RCEI_830A_BASE%d", device);
          break;
        case DEVICE_ID_RAR_XMC:
          sprintf(res_name[0], "RAR_XMC_BASE%d", device);
          break;     
        case DEVICE_ID_RAR_MPCIE:
          sprintf(res_name[0], "RAR_MPCIE_BASE%d", device);
          break;     
        case DEVICE_ID_P_708:
          sprintf(res_name[0], "P708_PLX_CONFIG_BASE%d", device);
          sprintf(res_name[1], "P708_BASE%d", device);
          break;
        case DEVICE_ID_RP_708:
          sprintf(res_name[0], "RP_708_CONFIG_BASE%d", device);
          sprintf(res_name[1], "RP_708_BASE%d", device);
          break;     
        case DEVICE_ID_P_SER:
          sprintf(res_name[0], "PSER_PLX_CONFIG_BASE%d", device);
          sprintf(res_name[1], "PSER_BASE%d", device);
          break;     
        default:
          ceiprint("Fail (dev%d): unsupported board, devID 0x%X\n", device, boardType);
          status = BTD_ERR_BADBOARDTYPE;
      };
    }

    if(status == BTD_OK) {
      if(strcmp(res_name[1], "") == 0)
        pMap->memSections = 1;
      else
        pMap->memSections = 2;      
      /* for any valid board, acquire the memory regions and map them in the virtual memory pool */
      for(i=0; (i < pMap->memSections) && (status == BTD_OK); i++) {
        if(v_mem_regions[device][i] != 0) {
          ceiprint("Error (dev%d):  memory region %d already mapped\n", device, i);
          continue;
        }
        if((err = RequestResource((Object*)&p_mem_reg, res_name[i], RESOURCE_PASSWORD)) != Success) {
          cei_err(device, "RequestResource", err);
          status = BTD_RESOURCE_ERR;
        }
        else { 
#ifdef GHS_KERNEL
          Address start, end;
          if((err = GetMemoryRegionAddresses(p_mem_reg, &start, &end)) == Success) {
            lpbase = (unsigned long)start;
            mem_size = (end - start) + 1;
          }
          else {
            cei_err(device, "GetMemoryRegionAddresses", err);
            status = BTD_MEMREG_ERR;
          }
#else
          if((err = MapMemory(__ghs_VirtualMemoryRegionPool, p_mem_reg, &v_mem_regions[device][i], &lpbase, &mem_size)) != Success) {
            cei_err(device, "MapMemory", err);
            status = BTD_MEM_MAP_ERR;
          }
#endif    
          if(err == Success) {
            pMap->memHostBase[i] = (char*)lpbase;
            pMap->memLengthBytes[i] = (unsigned)mem_size;
            pMap->flagMapToHost[i] = 1;
            devInfo[device].deviceBarCount++;
            devInfo[device].deviceBarLocation[i] = (unsigned long*)lpbase;
          }
        }
      }
    }
      
    if(status == BTD_OK) {
      pMap->VendorID = VENDOR_ID_CEI_AVIONICS;
      pMap->DeviceID = boardType;
      pMap->device = device;
      pMap->busType = BUS_PCI;
      pMap->use_count++;
    }
  }
  
  return status;
}


/****************************************************************
*
*  PROCEDURE - vbtFreeBoardAddresses
*
*  FUNCTION
*     This routine frees any system resources which were obtained
*     in the mapping of the phyical board address.
*
*  PARAMETERS
*     See below.
*
*  RETURNS
*     BTD_OK
*     BTD_MEM_MAP_ERR
****************************************************************/
int vbtFreeBoardAddresses(PDEVMAP_T pMap) {
  Error err;
  int i, status=BTD_OK;
  
  if(pMap->use_count > 0)
    pMap->use_count--;   
#ifndef GHS_KERNEL
  if(pMap->use_count == 0) {
    for(i = 0;i < MAX_MEMORY;i++) {
      if(pMap->flagMapToHost[i] == 1) {
        if((err = UnmapMemory(__ghs_VirtualMemoryRegionPool, v_mem_regions[pMap->device][i])) == Success)
          v_mem_regions[pMap->device][i] = 0;
        else {
          cei_err(pMap->device, "UnmapMemory", err);
          status = BTD_MEM_MAP_ERR;
        }
      }
    }
  }
#endif
  
  return status;
}

/****************************************************************
*
*  PROCEDURE - vbtGetPCIConfigRegister
*
*  FUNCTION
*     This routine takes the device number, an offset in PCI
*     configuration space, a byte length, and a pointer for the 
*     value of the specified PCI configuration space register.
*
*  PARAMETERS
*     See below.
*
*  RETURNS
*     BTD_OK             
*     BTD_RESOURCE_ERR 
*     BTD_READ_IODEV_ERR
****************************************************************/
int vbtGetPCIConfigRegister(UINT device, UINT offset, UINT length, UINT* value) {
  int status=BTD_OK;
  Error err;
  IODevice ioDev;
  char dbuf[20]="";
  UINT4 val=offset;  
    
  sprintf(dbuf,"CEIboard%d", device);  
  if((err = RequestResource((Object*)&ioDev,dbuf, RESOURCE_PASSWORD)) != Success) {
    cei_err(device, "RequestResource", err);
    status = BTD_RESOURCE_ERR;
  }
  else {
    if((err = ReadIODeviceStatus(ioDev, STATUS_PCI_CONFIG_SPACE, &val, length)) != Success) {
      cei_err(device, "ReadIODeviceStatus", err);
      status = BTD_READ_IODEV_ERR;
    }
    else
      *value = (UINT)val;
  }

  return status;
}


/****************************************************************
*
*  PROCEDURE - vbtGetDevInfo
*
*  FUNCTION
*     This rountine retrieves specific information concerning the 
*     device.
*
*  PARAMETERS
*     See below.
*
*  RETURNS
*     BTD_OK
*     BTD_ERR_PARAM
*     BTD_RESOURCE_ERR
*     BTD_READ_IODEV_ERR
****************************************************************/
int vbtGetDevInfo(int device, char *devInfo, void *pData) {
  Error err;
  int status=BTD_OK;
  IODevice ioDev;
  char dbuf[20]="";
  UINT2 val=0;
  int offset=0;
     
  if(strcmp(devInfo, "id") == 0)  /* the device id */
    offset = STATUS_DEVICE_ID;
  else if(strcmp(devInfo, "hir") == 0)  /* host interface register */
    offset = STATUS_HOST_INTERFACE;
  else if(strcmp(devInfo, "iq") == 0)  /* interrupt queue */
    offset = STATUS_INTERRUPT_DATA;
  else
    status = BTD_ERR_PARAM;    

  if(status == BTD_OK) {
    sprintf(dbuf,"CEIboard%d", device);  
    if((err = RequestResource ((Object*)&ioDev, dbuf, RESOURCE_PASSWORD)) != Success) {
      cei_err(device, "RequestResource", err);
      status = BTD_RESOURCE_ERR;
    }
    else {
      if((err = ReadIODeviceStatus (ioDev, offset, (void*)&val, sizeof(val))) != Success) {
        cei_err(device, "ReadIODeviceStatus", err);
        status = BTD_READ_IODEV_ERR;
      }
      else
        *((unsigned long*)pData) = (unsigned long)val;
    }
  }
  
  return status;
}
#endif /* INTEGRITY_PCI_PPC, INTEGRITY_PCI_X86 */



#ifdef INTEGRITY_VME_PPC
/****************************************************************
*
*  PROCEDURE - vbtMapIoAddress
*
*  FUNCTION
*     Maps the device A16 VME physical base address
*
*  PARAMETERS
*     BT_U32BIT  base_address      - Card I/O base address.
*     char      ** addr            - Returned address of board.
*
*  RETURNS
*     BTD_OK
*     BTD_MEM_MAP_ERR
****************************************************************/
int vbtMapIoAddress(unsigned base_address, char **addr) {
  int status=BTD_OK, dev=0;
  Error err;

  if((One_Time_Init == 1) && (vme_mem[dev][0] == NULL)) {
    if((err = CreateWindow(0, A16, base_address, 0x40, &vme_mem[dev][0], (Address*)addr[0])) != Success) {
      cei_err(device, "CreateWindow", err);
      status = BTD_MEM_MAP_ERR;
    }
  }
  
  return status;
}


/****************************************************************
*
*  PROCEDURE - vbtMapBoardAddress
*
*  FUNCTION
*     Maps the device A32 VME physical base address                             
*
*  PARAMETERS
*     BT_U32BIT  base_address      - Card physical base address.
*     BT_U32BIT  map_length        - Size of region to map in bytes.
*     int        addressing_mode   - VME Addressing Mode.
*     char     **addr              - Returned address of board.
*     void      *junk              - not used
*
*  RETURNS
*     BTD_OK
*     BTD_RESOURCE_ERR
*     BTD_MEM_MAP_ERR
****************************************************************/
int vbtMapBoardAddress(unsigned base_addr, unsigned map_length, char **addr, void *junk, int addressing_mode) {
  int status=BTD_OK, dev=0;
  Error err;

  if(One_Time_Init == 0) {
    vme_mem[dev][0] = NULL;
    vme_mem[dev][1] = NULL;
    /* Only initialize the VME bridge once */
    if((err = InitializeVMEBridge()) != Success) {
      cei_err(device, "InitializeVMEBridge", err);
      status = BTD_RESOURCE_ERR;
    }
    else {
      One_Time_Init = 1;
  }
  
  if((One_Time_Init == 1) && (vme_mem[dev][1] == NULL)) {
    if((err = CreateWindow(1, A32, base_addr, map_length /*0x800000*/, &vme_mem[dev][1], (Address*)addr[0])) != Success) {
      cei_err(device, "CreateWindow", err);
      status = BTD_MEM_MAP_ERR;
    }
  }
  
  return status;
}


/****************************************************************
*
*  PROCEDURE - vbtFreeBoardAddress
*
*  FUNCTION
*     This routine is a placeholder that does nothing.
*
*  PARAMETERS
*     See below.
*
*  RETURNS
*     Nothing
****************************************************************/
void vbtFreeBoardAddress(UINT cardnum) {}
  int dev=0;
  
  if(vme_device_use[dev] == 0) {
    if(vme_mem[dev][0] != NULL) {
      /* close the A16 window */
      if((err = CloseWindow(0, &vme_mem[dev][0])) != Success)
        cei_err(device, "CloseWindow", err);
      else {
        vme_mem[dev][0] = NULL);
        vme_device_use[dev]--;
      }
    }
    
    if(vme_mem[dev][1] != NULL) {
      /* close the A32 window */
      if((err = CloseWindow(1, &vme_mem[dev][1])) != Success)
        cei_err(device, "CloseWindow", err);
      else {
        vme_mem[dev][1] = NULL);
        vme_device_use[dev]--;
      }
    }
  }
}

#endif /*INTEGRITY_VME_PPC */
