/*===================================================================================================================*
 * FILE:                                   A V I O   V X W   D R V . C
 *===================================================================================================================*
 *
 *   COPYRIGHT 2011 - 2021 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED ONLY IN ACCORDANCE WITH THE
 *   TERMS OF SUCH LICENSE AND WITH THE INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY 
 *   OTHER COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE AND SHOULD NOT BE CONSTRUED 
 *   AS A COMMITMENT BY ABACO SYSTEMS, INC.
 *
 *===================================================================================================================*
 *
 * FUNCTION:   Avionics I/O Boards VxBus Driver Source
 *
 *===================================================================================================================*
 *
 *   Date       By   Modifications
 * ----------  ----  ---------------------------------------------------------------
 * 09/20/2011   BC   Initial Release of VxBus Driver for QPM-1553 and RXMC-1553 
 * 08/06/2012   BC   Adds support for RAR15-XMC-XT, and RXMC2-1553 includes ISR support for ARINC
 * 03/19/2013   BC   Update for 64-bit driver.
 * 11/13/2013   BW   Add support for new ARINC boards
 * 11/13/2013   BC   Add support for QCP-1553
 * 11/05/2014   BC   Fix GEIP_XMC2_DEVID
 * 07/23/2015   BW   Corrected improperly declared address offset constants, general rewrite and cleanup:
 *                   Properly disconnect the assigned ISR using vxbIntDisconnect invocation in geipAvDevRemove
 * 12/10/2015   BW   Added commercial message queue, impacting the following routines:
 *                      geipAvDriverCtrl() - added msg_q_com_id[] device info structure member setup
 *                      geipAvDevRemove() - delete the commercial message queue during driver removal cleanup
 *                      avioInstInit2() - when required, create the commercial message queue
 *                      intProcess32() - if ARINC channel is present (combo card), if the ARINC f/w generates an
 *                                       interrupt, send a message to the commercial queue instead of the 1553 queue
 *                      intProcessCommercial() - if an interrupt is processed, send a message to the commercial queue
 *                      intProcessCommercialPcie() - if an interrupt is processed, send a message to the commercial queue
 * 05/12/2016   BW   Renaming all things to generic corp references, in addition to the changes described in avioVxwDrv.h
 *                      geipInst -> avioInst
 *                      geipAvInstInit2 -> avioInstInit2
 *                      geipAvDevProbe -> avioDevProbe
 *                      geipAvDriverCtrl -> avioDriverCtrl
 *                      geipAvDevRemove -> avioDevRemove
 *                      geipAvDeviceShow -> avioDeviceShow
 *                      geipAvDrvMethods -> avioDrvMethods
 *                      geipAvDrvFuncs -> avioDrvFuncs
 *                      geipAvDrvIDList -> avioDrvIDList
 *                      geipAvCfg1553Devices -> avioCfg1553Devices
 *                      geipAvConfigDevices -> avioConfigDevices
 *                      geipAvDrvRegister -> avioDrvRegister
 *                      geipAvDevRegistration -> avioDevRegistration
 *                      geipAvUsDelay -> avioUsDelay
 * 08/31/2016   BW   Relocated all include files not required for API build from avioVxwDrv.h to this file
 * 12/26/2019   BH   Modifed avioInstInit2 for change to device and message queue define.
 * 10/05/2021   BDW  Renamed local variables geipAvRefCnt and geipPCIDrvNum to avioAvRefCnt and avioPciDrvNum.
 *
 *===================================================================================================================*/

#include <errno.h>
#include <semLib.h>
#include <drv/pci/pciConfigLib.h>
#include <drv/pci/pciIntLib.h>
#include <pci.h>
#include <sysLib.h>
#include <vxWorks.h>
#include <version.h>
#include <stdlib.h>
#include <hwif/vxbus/vxBus.h>
#include <hwif/vxbus/vxbPciLib.h>
#include <hwif/vxbus/hwConf.h>
#include <hwif/vxbus/vxbMethodDecl.h> /* for DEVMETHOD_CALL(driverControl) */
#include <vxAtomicLib.h>
#include <selectLib.h>
#include <errnoLib.h> 
#include <vxBusLib.h>
#include "avioVxwDrv.h"

/*-------------------------------------------------------------------------------------------------------------------*
 * Global Data
 *-------------------------------------------------------------------------------------------------------------------*/
VXB_DEVICE_ID avioInst[DRV_MAX_DEVICES] = {NULL};

/*-------------------------------------------------------------------------------------------------------------------*
 * Avionics I/O VxBus Driver Routine Prototypes
 *-------------------------------------------------------------------------------------------------------------------*/
LOCAL VOID avioInstInit2 (VXB_DEVICE_ID pInst);
LOCAL BOOL avioDevProbe (VXB_DEVICE_ID pInst);
LOCAL STATUS avioDriverCtrl (VXB_DEVICE_ID pInst, struct vxbDriverControl *pControl);
LOCAL STATUS avioDevRemove (VXB_DEVICE_ID pInst, struct vxbDriverControl *pControl);

/*-------------------------------------------------------------------------------------------------------------------*
 * External Avionics I/O VxBus Utility Routine Prototypes
 *-------------------------------------------------------------------------------------------------------------------*/
void avioDeviceShow(void);

/*-------------------------------------------------------------------------------------------------------------------*
 * Avionics I/O Supported VxBus "Other-Class" Driver Methods and Routines
 *-------------------------------------------------------------------------------------------------------------------*/
LOCAL DRIVER_METHOD avioDrvMethods[] =
{
   DEVMETHOD(driverControl, avioDriverCtrl),
   DEVMETHOD(vxbDrvUnlink,  avioDevRemove),
   DEVMETHOD_END
};

LOCAL struct drvBusFuncs avioDrvFuncs =
{
   NULL,              /* devInstanceInit */
   avioInstInit2,     /* devInstanceInit2 */
   NULL               /* devConnect */
};

/*-------------------------------------------------------------------------------------------------------------------*
 * Avionics I/O VxBus Driver Private Data
 *-------------------------------------------------------------------------------------------------------------------*/
/* Device Search List for all supported Avionics I/O Devices */
LOCAL PCI_DEVVEND avioDrvIDList[] =
{    /* PCI Device ID         PCI Manuf. ID  */
   { AVIO_PCI_QPM_DEVID,       AVIO_PCI_VENDID },
   { AVIO_PCI_QCP_DEVID,       AVIO_PCI_VENDID },
   { AVIO_PCI_XMC_DEVID,       AVIO_PCI_VENDID },
   { AVIO_PCI_XMC2_DEVID,      AVIO_PCI_VENDID },
   { AVIO_PCI_RAR15_DEVID,     AVIO_PCI_VENDID },
   { AVIO_PCI_RAR15XT_DEVID,   AVIO_PCI_VENDID },
   { AVIO_PCI_R15PMC_DEVID,    AVIO_PCI_VENDID },
   { AVIO_PCI_R15_MPCIE_DEVID, AVIO_PCI_VENDID },
   { AVIO_PCI_RCEI830A_DEVID,  AVIO_PCI_VENDID },
   { AVIO_PCI_P708_DEVID,      AVIO_PCI_VENDID },
   { AVIO_PCI_PSER_DEVID,      AVIO_PCI_VENDID },
   { AVIO_PCI_PDIS_DEVID,      AVIO_PCI_VENDID },
   { AVIO_PCI_RAR_CPCI_DEVID,  AVIO_PCI_VENDID },
   { AVIO_PCI_R830RX_DEVID,    AVIO_PCI_VENDID },
   { AVIO_PCI_R830X820_DEVID,  AVIO_PCI_VENDID },
   { AVIO_PCI_CEI830_DEVID,    AVIO_PCI_VENDID },
   { AVIO_PCI_RARXMC_DEVID,    AVIO_PCI_VENDID },
   { AVIO_PCI_RP708_DEVID,     AVIO_PCI_VENDID },
   { AVIO_PCI_RAR_MPCIE_DEVID, AVIO_PCI_VENDID }
};

/*-------------------------------------------------------------------------------------------------------------------*
 * Avionics I/O VxBus Driver Registration Structure
 *-------------------------------------------------------------------------------------------------------------------*/
LOCAL PCI_DRIVER_REGISTRATION avioDevRegistration =
{
   {
      NULL,                  /* pNext */
      VXB_DEVID_DEVICE,      /* devID */
      VXB_BUSID_PCI,         /* busID = PCI */
#ifdef VXB_LEGACY_ACCESS
      VXBUS_VERSION_3,       /* vxbVersion */
#else
  #ifdef _WRS_CONFIG_LP64  
      VXB_VER_5_0_0,         /* 64 bit Version */
  #else
      VXB_VER_4_0_0,         /* 32 bit Version */
  #endif
#endif
      AVIO_DRV_NAME,        /* drvName */
      &avioDrvFuncs,        /* pDrvBusFuncs */
      avioDrvMethods,       /* pMethods */
      avioDevProbe,         /* devProbe */
      NULL                  /* pParamDefaults */
   },
   NELEMENTS(avioDrvIDList),
   avioDrvIDList
};

/* Total number of Avionics I/O Devices Detected - use to determine the Device Number assigned to each board */
LOCAL atomic_t avioAvRefCnt = 0; 

/* Driver Instance Counter - every instantiation of the Avionics I/O driver is tracked */
LOCAL atomic_t avioPciDrvNum = ERROR;  

/* ISR Message Queue values passed back to the API to indicate which "channel" generated the interrupt */
LOCAL unsigned int mq_data[MAX_CHANNELS] = {0,1,2,3};

/*********************************************************************************************************************
**
** NAME:
** avioDriverCtrl()
**
** DESCRIPTION:
** Implements the VxBus "other-class driver" driverControl function for the Avionics IO driver.
** This "method" is activated from the device API via execution of the VxBus utility function as follows:
** 
**   vxbDevMethodRun (DEVMETHOD_CALL(driverControl), &<controlStructure>) 
**
** This function checks the invocation to be the type "get Avionics IO device information", and fills the
** structure if true.
**
** ARGUMENTS:
** pInst    - VxBus device instance
** pControl - method argument
**
** RETURN:
** OK or ERROR on error.
**
**********************************************************************************************************************/
LOCAL STATUS avioDriverCtrl
(
   VXB_DEVICE_ID pInst,
   struct vxbDriverControl *pControl
)
{
   AVIO_DEVICE_CTRL_TYPE *pDrvCtrl = NULL;   /* reference to the Device Control/Info Structure for this device */
   STATUS driverControlStatus = OK;             /* parameter validation status */
   
   if ((pInst == NULL) || (pControl == NULL) || (pControl->driverName == NULL)) {
      /* Any invocation of this function by the API prior to proper session initialization is an error */
      driverControlStatus = ERROR;
   }
   else {
      pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)pInst->pDrvCtrl;
      
      dprint("avioDriverCtrl: invoked for pControl->drvCtrlCmd = %x\n",pControl->drvCtrlCmd);
      dprint("avioDriverCtrl: pInst->pName = %s\n",pInst->pName);
      dprint("avioDriverCtrl: pControl->driverName = %s\n",pControl->driverName);
      dprint("avioDriverCtrl: pDrvCtrl->dev_id = %x\n",pDrvCtrl->dev_id);
   
      switch ( pControl->drvCtrlCmd ) {
         /* the only valid application query of the Avionics VxBus Driver is to map an existing board */
         case VXB_GET_DEV_MAP: 
            dprint("avioDriverCtrl: VXB_GET_DEV_MAP request for device %x\n",pDrvCtrl->dev_id);
            ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->dev_id[pInst->unitNumber] = pDrvCtrl->dev_id;
            ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->irq[pInst->unitNumber] = pDrvCtrl->irq;
            ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->dev_chan[pInst->unitNumber] = pDrvCtrl->dev_chan;
            ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->membase[pInst->unitNumber] = pDrvCtrl->membase;
            ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->membaseBar2[pInst->unitNumber] = pDrvCtrl->membaseBar2;
            ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->baseAddressBar0[pInst->unitNumber] = pDrvCtrl->baseAddressBar0;
            ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->baseAddressBar2[pInst->unitNumber] = pDrvCtrl->baseAddressBar2;
            ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->device[pInst->unitNumber] = pDrvCtrl->device;
            ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->msg_q_id[pInst->unitNumber] = pDrvCtrl->msg_q_id;
            ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->msg_q_com_id[pInst->unitNumber] = pDrvCtrl->msg_q_com_id;
            ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->uca32[pInst->unitNumber] = pDrvCtrl->uca32;
            ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->arinc[pInst->unitNumber] = pDrvCtrl->arinc;
            ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->size[pInst->unitNumber] = pDrvCtrl->size;
            ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->hif[pInst->unitNumber] = pDrvCtrl->hif;
            ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->CondorDeviceID[pInst->unitNumber] = pDrvCtrl->CondorDeviceID;
            ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->device_present[pInst->unitNumber] = 0xcafe0000+pInst->unitNumber;
            ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->number_of_devices++;    
            break; 
   
         default:
            /* Any invocation of this function by the API other than for VXB_GET_DEV_MAP functionality is an error */
            driverControlStatus = ERROR;
       }
   }
   return driverControlStatus;
} /* avioDriverCtrl() */


/*********************************************************************************************************************
**
** NAME:
** avioDevRemove()
**
** DESCRIPTION:
** Implements the VxBus device removal functionality for the Avionics IO driver.
** This "method" is activated from the device API via execution of the VxBus utility function as follows:
** 
**   vxbDevMethodRun (DEVMETHOD_CALL(vxbDrvUnlink), &<controlStructure>) 
**
** ARGUMENTS:
** pInst    - VxBus device instance.
** pControl - method argument
**
** RETURN:
** OK or ERROR on error.
**
**********************************************************************************************************************/
LOCAL STATUS avioDevRemove
(
   VXB_DEVICE_ID pInst,
   struct vxbDriverControl *pControl
)
{
   AVIO_DEVICE_CTRL_TYPE *pDrvCtrl = NULL;   /* reference to the Device Control/Info Structure for this device */
   STATUS driverRemoveStatus = OK;           /* parameter validation and resource disconnect status */
   int deviceIndex;
   
   if (pInst != NULL) {
      dprint("avioDevRemove: invoked for pControl->drvCtrlCmd = %x\n",pControl->drvCtrlCmd);
      dprint("avioDevRemove: pInst->pName = %s\n",pInst->pName);
      dprint("avioDevRemove: pControl->driverName = %s\n",pControl->driverName);
      
      /* if this instance is a defined session, check if it is active */
      pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)pInst->pDrvCtrl;
      if (pDrvCtrl != NULL) {
         dprint("avioDevRemove: Device %d", (int)pDrvCtrl->device);
         deviceIndex = pInst->unitNumber;

         /* Disconnect the respective ISR from this session, note the handler routine is specific to the board type. */  
         switch (pDrvCtrl->CondorDeviceID) {
            /* 1553 PCI/PMC/XMC products */
            case AVIO_PCI_QPM_DEVID:
            case AVIO_PCI_QCP_DEVID:
            case AVIO_PCI_XMC_DEVID:
            case AVIO_PCI_XMC2_DEVID:
            case AVIO_PCI_R15PMC_DEVID:
            case AVIO_PCI_R15_MPCIE_DEVID:
               /* v6 firmware and later (UCA32) */
               if (pDrvCtrl->uca32) {   
                  vxbIntDisable (pInst, 0, (VOIDFUNCPTR)intProcess32, pInst);
                  vxbIntDisconnect (pInst, 0, (VOIDFUNCPTR)intProcess32, pInst);
               }
               else {  /* v5 firmware (UCA) */
                  vxbIntDisable (pInst, 0, (VOIDFUNCPTR)intProcess, pInst);
                  vxbIntDisconnect (pInst, 0, (VOIDFUNCPTR)intProcess, pInst);
               }
               driverRemoveStatus = msgQDelete(pDrvCtrl->msg_q_id);
               break;
      
            /* 1553/ARINC combo cards are UCA32-based, with two message queues to remove */
            case AVIO_PCI_RAR15_DEVID:
            case AVIO_PCI_RAR15XT_DEVID:
               vxbIntDisable (pInst, 0, (VOIDFUNCPTR)intProcess32, pInst);
               vxbIntDisconnect (pInst, 0, (VOIDFUNCPTR)intProcess32, pInst);
               driverRemoveStatus = msgQDelete(pDrvCtrl->msg_q_id);
               driverRemoveStatus = msgQDelete(pDrvCtrl->msg_q_com_id);

            /* commercial PCI/PMC products with a PLX PCI9056 interface */
            case AVIO_PCI_CEI830_DEVID:
            case AVIO_PCI_R830RX_DEVID:
            case AVIO_PCI_R830X820_DEVID:
            case AVIO_PCI_RCEI830A_DEVID:
            case AVIO_PCI_RAR_CPCI_DEVID:
            case AVIO_PCI_PSER_DEVID:
               vxbIntDisable (pInst, 0, (VOIDFUNCPTR)intProcessCommercial, pInst);
               vxbIntDisconnect (pInst, 0,(VOIDFUNCPTR)intProcessCommercial, pInst);
               driverRemoveStatus = msgQDelete(pDrvCtrl->msg_q_com_id);
               break;
      
            case AVIO_PCI_RARXMC_DEVID:        /* commercial PCI Express products */
            case AVIO_PCI_RAR_MPCIE_DEVID:
               vxbIntDisable (pInst, 0, (VOIDFUNCPTR)intProcessCommercialPcie, pInst);
               vxbIntDisconnect (pInst, 0,(VOIDFUNCPTR)intProcessCommercialPcie, pInst);
               driverRemoveStatus = msgQDelete(pDrvCtrl->msg_q_com_id);
               break;

            default:
               driverRemoveStatus = ERROR;
               break;
         }
         
         /* proceed only if the remove request was for a valid board */
         if (driverRemoveStatus == OK) {
            /* Check to see if this is the last instance of this driver, if so de-register it from the o/s */
            if ( vxAtomicDec(&avioPciDrvNum) == ERROR ) {
               /* last instance of this driver */
               if (vxbDriverUnregister((DRIVER_REGISTRATION*)&avioDevRegistration) == ERROR) {
                  printf ("Error de-registering Avionics I/O Driver.\n");  
                  driverRemoveStatus = ERROR;
/*                  errnoSet(EIO);*/
               }
             
               /* free dynamically allocated device control structure */
               if( pDrvCtrl->name ) {
                  free(pDrvCtrl->name);
               }
               free(pDrvCtrl);      
               pInst->pDrvCtrl = NULL;
               avioInst[deviceIndex] = NULL;
            }
         }
      }
      else {  
         /* invalid session referenced - pInst->pDrvCtrl is NULL */
/*         errnoSet(EACCES);*/
         driverRemoveStatus = ERROR;
      }
   }
   else {  
      /* bad argument - pInst is NULL */
/*      errnoSet(EINVAL);*/
      driverRemoveStatus = ERROR;
   }
   return driverRemoveStatus;
}

/*********************************************************************************************************************
**
** NAME:
** avioDevProbe()
**
** DESCRIPTION:
** Queries each VxBus device instance for presence of a Avionics PCI/PCIe IO device.
**
** ARGUMENTS:
** pInst - VxBus device instance.
**
** RETURN:
** BOOL: TRUE when a supported Avionics IO device is found
**       FLASE otherwise.
**********************************************************************************************************************/
LOCAL BOOL avioDevProbe(VXB_DEVICE_ID pInst)
{
   BOOL deviceProbeStatus = TRUE; /* parameter and device reference validation status */

   /* validate the parameter and pertinant information therein */
   if ( (pInst == NULL) || (pInst->pParentBus == NULL) || (pInst->pParentBus->pBusType == NULL) ) {  
      errnoSet(EINVAL);
      deviceProbeStatus = FALSE;
   }
   else {
      dprint ("avioDevProbe: Device %d\n", (int)pInst->unitNumber);
      
      /* Only PCI devices are valid Avionics VxBus devices, for today */
      if ( pInst->pParentBus->pBusType->busID == VXB_BUSID_PCI ) {
         UINT32 subId;
   
         /* read this device's subsystem device and vendor ID from PCI configuration space */
         VXB_PCI_BUS_CFG_READ (pInst, PCI_CFG_VENDOR_ID, 4, subId);
         dprint ("avioDevProbe: PCI Device/Vendor ID = %x\n", subId);
         switch (subId) {
            case AVIO_PMC_ID:
            case AVIO_QCP_ID:
            case AVIO_XMC_ID:
            case AVIO_XMC2_ID:
            case AVIO_RAR15_ID:
            case AVIO_RAR15XT_ID:
            case AVIO_R15PMC_ID:
            case AVIO_PCI_R15_MPCIE_ID:
            case AVIO_RAR_CPCI_ID:
            case AVIO_RP708_ID:
            case AVIO_P708_ID:
            case AVIO_PSER_ID:
            case AVIO_PDIS_ID:
            case AVIO_CEI830_ID:
            case AVIO_R830RX_ID:
            case AVIO_R830X820_ID:
            case AVIO_RCEI830A_ID:
            case AVIO_RARXMC_ID:
            case AVIO_PCI_RAR_MPCIE_ID:
               break;

            default:
               deviceProbeStatus = FALSE;
               break;
         }
      }
      else
      {
         /* device not found */
         deviceProbeStatus = FALSE;
      }
   }

   return deviceProbeStatus;
}

/*********************************************************************************************************************
**
** NAME:
** avioDeviceShow()
**
** DESCRIPTION:
** Displays information for each Avionics PCI/PCIe IO device discovered and supported by this VxBus driver.
**
** ARGUMENTS:
** N/A
**
** RETURN:
** N/A
**
**********************************************************************************************************************/
void avioDeviceShow(void)
{
   int devcnt = 0;                           /* device index for displaying information for all discovered devices */
   AVIO_DEVICE_CTRL_TYPE *pDrvCtrl = NULL;   /* reference to the Device Control/Info Structure for this device */
   
   if (avioInst == NULL) {  
      printf("\nNo Avionics I/O Devices Detected\n");
      errnoSet(EINVAL);
      return;
   }

   printf("Avionics I/O Device(s) Detected:\n\n");
#ifdef _WRS_CONFIG_LP64 
   printf("Unit  Device ID   Base Address        IRQ   Channels  UCA32  ARINC\n");
#else
   printf("Unit  Device ID   Base Address  IRQ   Channels  UCA32    ARINC\n");
#endif

   printf("-----------------------------------------------------------------------------\n");  
   while (avioInst[devcnt]) {
      pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)avioInst[devcnt]->pDrvCtrl;   
#ifdef _WRS_CONFIG_LP64 
      printf("%2d    %4x        %p  %d        %d        %d      %d  \n",
#else
      printf("%2d    %4x        0x%08X    0x%X     %2d       %d        %d  \n",
#endif
      pDrvCtrl->device,
      pDrvCtrl->CondorDeviceID,
      pDrvCtrl->membase,
      pDrvCtrl->irq,
      pDrvCtrl->dev_chan,
      pDrvCtrl->uca32,
      pDrvCtrl->arinc);
      devcnt++;
   }
}

/*********************************************************************************************************************
** NAME:
** avioUsDelay()
**
** DESCRIPTION:
** A delay function ?
**
** ARGUMENTS:
** unsigned int delay - duration to delay, in microseconds?
**
** RETURNS:
** N/A
**********************************************************************************************************************/
void avioUsDelay(unsigned int delay)
{
   vxbUsDelay(delay);
}

/*********************************************************************************************************************
** NAME:
** avioDrvRegister2()
**
** DESCRIPTION:
** Called from avioDrvRegister, see the description below.  When the value for AVIO_DEBUG as defined in the
** configuration file 40avioVxwDrv.cdf is TRUE, this routine is not executed at boot time; however, late driver
** registration can be performed by invoking this routine from the console or command line.
**
** ARGUMENTS:
** N/A
**
** RETURNS:
** N/A
**********************************************************************************************************************/
void avioDrvRegister2(void)
{
   STATUS status; /* driver registration status */
   dprint("avioDrvRegister%d\n",2);

   status = vxbDevRegister ((DRIVER_REGISTRATION*)&avioDevRegistration);
   
   dprint("avioDrvRegister2: vxbDevRegister status = %d\n",status);
   if( status == ERROR ) {
      printf( "avioDrvRegister2: Error registering Avionics I/O Driver.\n");
      errnoSet(EIO);
   }
} /* avioDrvRegister2() */


/*********************************************************************************************************************
** NAME:
** avioDrvRegister()
**
** DESCRIPTION:
** This routine is referenced in the Avionics IO VxBus configuration file 40avioAvDrv.cdf.  It is the  
** Avionics IO main VxBus Driver registration routine, notifying VxWorks the driver is available and registering
** the structure "avioDevRegistration" as the source of information required for Avionics IO devices.
**
** Registration is actually performed via invocation of the routine avioDrvRegister2 when not executing this 
** driver source in "debug" mode (Configuration File assignment via the parameter AVIO_DEBUG).
**
** ARGUMENTS:
** N/A
**
** RETURNS:
** N/A
**********************************************************************************************************************/
void avioDrvRegister(void)
{
#ifndef AVIO_DEBUG
   avioDrvRegister2();
#endif /* AVIO_DEBUG */
}

/*********************************************************************************************************************
** NAME:
** avioCfg1553Devices()
**
** DESCRIPTION:
** This routine configures Avionics 1553 device information in the structure available to the application.
**
** ARGUMENTS:                                         
** AVIO_DEVICE_CTRL_TYPE * pDevCtrl  device control & info structure
**
** RETURNS:
** N/A
**********************************************************************************************************************/
int avioCfg1553Devices(AVIO_DEVICE_CTRL_TYPE * pDevCtrl)
{
   UINT btype;                     /* 1553 board attributes */
   UINT nchan;
   unsigned short host_interface;
   
   host_interface = *(unsigned short *)pDevCtrl->baseAddressBar0;
   flipw(&host_interface);
   pDevCtrl->hif = host_interface;
   btype = (host_interface & 0x3e) >> 1;
   nchan = (host_interface & 0x7c0) >> 6;
   pDevCtrl->dev_chan = nchan;
   pDevCtrl->uca32 = (host_interface & 0x4000) ? 1 : 0;
   pDevCtrl->arinc = (btype == AVIO_MUL_PROT_TYP2 || btype == AVIO_MUL_PROT_TYP1) ? 1 : 0;
   
   dprint("avioCfg1553Devices: base addr = %x\n",pDevCtrl->membase);
   dprint("avioCfg1553Devices: Host_interface = %x\n",pDevCtrl->hif);                   
   dprint("avioCfg1553Devices: PCI Device ID = %x\n",pDevCtrl->CondorDeviceID);
   dprint("avioCfg1553Devices: 1553 Qboard host interface = %x\n",host_interface);
   dprint("avioCfg1553Devices: uca32 = %x\n",pDevCtrl->uca32);
   dprint("avioCfg1553Devices: ARINC = %d\n",pDevCtrl->arinc);
   dprint("avioCfg1553Devices: btype = %x\n",btype);
   dprint("avioCfg1553Devices: nchan = %d\n",nchan);
   dprint("avioCfg1553Devices: Complete for device %d\n",pDevCtrl->devcnt);
   return 0;   
}


/*********************************************************************************************************************
** NAME:
** avioConfigDevices()
**
** DESCRIPTION:
** This routine configures Avionics generic device information in the structure available to the application,
** as well as handling registration of the device BAR memory region(s) to VxWorks.
**
** ARGUMENTS:                                         
** VXB_DEVICE_ID pInst -  Avionics PCI/PCIe IO device driver instance 
**
** RETURNS:
** N/A
** 
**********************************************************************************************************************/
void avioConfigDevices(VXB_DEVICE_ID pInst)
{
   AVIO_DEVICE_CTRL_TYPE *pDrvCtrl = NULL;   /* reference to the Device Control/Info Structure for this device */
   UINT16 devID;                             /* PCI Device Identification value */
   unsigned int commercialHostInterface;     /* commercial (ARINC/P-SER) device Global Enable register value */
   STATUS driverConfigStatus = OK;           /* validity of the device referenced by the open session */

   pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)pInst->pDrvCtrl;
   
   /* the only valid Avionics device configuration is one where the first mapped region is a memory region (not I/O) */
   if (pInst->regBaseFlags[0] == VXB_REG_MEM) 
   {
      /* Initialize some generic information regarding this device */
      pDrvCtrl->bustype = 2;
      pDrvCtrl->mem_base_reg = 0;
      pDrvCtrl->region_count = 1;
      pDrvCtrl->size = BAR_SIZE_8MB;
      pDrvCtrl->dev_chan = 0;
      pDrvCtrl->uca32 = 0; 
      VXB_PCI_BUS_CFG_READ(pInst, PCI_CFG_DEVICE_ID, 2, devID);
      devID &= 0xFFFF;
      pDrvCtrl->CondorDeviceID = (UINT16)devID;
      dprint("avioConfigDevices: Device ID = %x\n",devID);
      
      /* for 1553 devices, the dev_id stucture member is translated into some sort of board type, not so for commercial */
      switch (devID) {
         case AVIO_PCI_QPM_DEVID:
            pDrvCtrl->dev_id = DRV_QPM1553;
            break;
         case AVIO_PCI_QCP_DEVID:
            pDrvCtrl->dev_id = DRV_QCP1553;
            break;
         case AVIO_PCI_XMC_DEVID:
            pDrvCtrl->dev_id = DRV_RXMC1553;
            break;
         case AVIO_PCI_R15PMC_DEVID:
            pDrvCtrl->dev_id = DRV_R15PMC;
            break;
         case AVIO_PCI_XMC2_DEVID:
            pDrvCtrl->dev_id = DRV_R15XMC2;
            break;
         case AVIO_PCI_RAR15_DEVID:
            pDrvCtrl->dev_id = DRV_RAR15XMC;
            break;
         case AVIO_PCI_RAR15XT_DEVID:
            pDrvCtrl->dev_id = DRV_RAR15XMCXT;
            pDrvCtrl->size = BAR_SIZE_16MB;
            break;
         case AVIO_PCI_R15_MPCIE_DEVID:
            pDrvCtrl->dev_id = DRV_R15_MPCIE;
            break;
      
         /* PCIbus Commercial Avionics Devices (PLX based with host ifc in BAR2) */
         case AVIO_PCI_CEI830_DEVID:
         case AVIO_PCI_R830RX_DEVID:
         case AVIO_PCI_R830X820_DEVID:
         case AVIO_PCI_RCEI830A_DEVID:
         case AVIO_PCI_RAR_CPCI_DEVID:
         case AVIO_PCI_RP708_DEVID:
         case AVIO_PCI_P708_DEVID:
         case AVIO_PCI_PSER_DEVID:
         case AVIO_PCI_PDIS_DEVID:
            pDrvCtrl->mem_base_reg = 2;
            pDrvCtrl->region_count = 2;
            pDrvCtrl->dev_id = devID;
            if ((devID == AVIO_PCI_RP708_DEVID) || (devID == AVIO_PCI_P708_DEVID) || (devID == AVIO_PCI_PSER_DEVID) || (devID == AVIO_PCI_PDIS_DEVID)) {
               pDrvCtrl->size = BAR_SIZE_4MB;
            }
            else {
               pDrvCtrl->size = BAR_SIZE_512KB;
            }
            /* Map the physical memory region for BAR2 */
            vxbRegMap (pInst, 2, &pDrvCtrl->baseHandle[2]);
            dprint("avioConfigDevices: Device BAR2 = %x\n", pDrvCtrl->baseHandle[2]);
            
            /* used for Legacy API expectation of integer structure member storage of BAR pointers */
            pDrvCtrl->membaseBar2 = (unsigned long int)pInst->pRegBase[2];
            dprint("avioConfigDevices: Device Virtual BAR2 = %x\n", pDrvCtrl->membaseBar2);
            
            pDrvCtrl->baseAddressBar2 = pInst->pRegBase[2];
#ifdef VXB_VER_5_0_0
            pInst->pRegBasePhys[2] = pInst->pRegBase[2];
#endif
            commercialHostInterface = *(unsigned int *)pDrvCtrl->baseAddressBar2;
            pDrvCtrl->hif = (unsigned short)commercialHostInterface;
            pDrvCtrl->arinc = 1;
            break;
      
         /* PCIe Commercial Avionics Devices */
         case AVIO_PCI_RARXMC_DEVID:
         case AVIO_PCI_RAR_MPCIE_DEVID:
            pDrvCtrl->dev_id = devID;
            pDrvCtrl->size = BAR_SIZE_512KB;
            pDrvCtrl->arinc = 1;
            break;
      
         default:
            dprint("avioConfigDevices:  Unknown device ID %x\n", devID); 
            driverConfigStatus = ERROR;
            break;
      }
      
      if (driverConfigStatus == OK) {
         pDrvCtrl->device = pInst->unitNumber;
         
         /* Map the physical memory region for BAR0 common to all Avionics generic devices. */
         vxbRegMap (pInst, 0, &pDrvCtrl->baseHandle[0]);
         
         pDrvCtrl->baseAddressBar0 = pInst->pRegBase[0];
         pDrvCtrl->membase = (unsigned long)pInst->pRegBase[0];
   
#ifdef VXB_VER_5_0_0 
         pInst->pRegBasePhys[0] = pInst->pRegBase[0];
#endif
      
         dprint("avioConfigDevices:  Base address = %x\n", pDrvCtrl->baseAddressBar0);
         dprint("avioConfigDevices:  Complete with unit %d\n", pInst->unitNumber);
      }
      else
      {  /* error */ 
         errnoSet(EFAULT);
      }
   }
   else
   {  /* error */ 
      vxAtomicDec(&avioAvRefCnt);
      errnoSet(EFAULT);
   }
}

/*********************************************************************************************************************
** NAME:
** avioInstInit2()
**
** DESCRIPTION:
** This is the main Avionics IO devices VxBus driver post-kernel initialization routine.
**
** ARGUMENTS:                                         
** VXB_DEVICE_ID pInst
**
** RETURNS
**    BTD_OK          - Mapped address with no error
**    ERROR           - Error Mapping PCI addresses
***********************************************************************************************************************/
LOCAL void avioInstInit2(VXB_DEVICE_ID pInst)
{
   int deviceIndex;                          /* device number referenced in the call */
   char mq_name[40];                         /* message queue name */
   char mq2_name[40];                        /* message queue name */
   AVIO_DEVICE_CTRL_TYPE *pDrvCtrl = NULL;   /* reference to the Device Control/Info Structure for this device */
   
   dprint("avioInstInit2: Begin execution%d\n", 2);
   if( !pInst )
   {  /* bad argument - pInst is NULL */
      errnoSet(EINVAL);
      dprint("avioInstInit2: Instance pInst is NULL%d\n", 2);
      return;
   }
   
   /* Increment the reference counter indicating the number of instances this driver has been initialized. */
   deviceIndex = vxAtomicInc (&avioAvRefCnt);
   pInst->unitNumber = deviceIndex;
   
   /* create a local copy of this device driver instance */
   avioInst[deviceIndex] = pInst;
   
   pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)malloc(sizeof(AVIO_DEVICE_CTRL_TYPE));

   while (pDrvCtrl) {
      int i;

      /* create device name */
      if ( (i = snprintf(NULL, 0, "/%s/%d", AVIO_DEV_NAME, deviceIndex)) > 0 ) {
         if ((pDrvCtrl->name = malloc(i + 1)) > 0) {          
            if ( (i = snprintf(pDrvCtrl->name, i + 1, "/%s/%d", AVIO_DEV_NAME, deviceIndex)) > 0 ) {

               /* increment the number of driver instances / open sessions */
               vxAtomicInc(&avioPciDrvNum);
               
               /* connect to VxBus */
               pInst->pDrvCtrl = pDrvCtrl;
               
               dprint("avioInstInit2: avioPciDrvNum = %d\n", avioPciDrvNum);
               dprint("avioInstInit2: calling avioConfigDevices for device %s\n", pDrvCtrl->name);
         
               /* define the contents of the device control & information structure */
               avioConfigDevices(pInst);
               
               /* connect and enable the interrupt. */
               switch (pDrvCtrl->CondorDeviceID) {
                  case AVIO_PCI_QPM_DEVID:
                  case AVIO_PCI_QCP_DEVID:
                  case AVIO_PCI_XMC_DEVID:
                  case AVIO_PCI_XMC2_DEVID:
                  case AVIO_PCI_R15PMC_DEVID:
                  case AVIO_PCI_R15_MPCIE_DEVID:

                     /* create a 1553 Message Queue for this device for interrupt notification purposes */
                     sprintf (mq_name,"/%s%d", AVIO_MSG_Q_NAME, deviceIndex);
                     pDrvCtrl->msg_q_id = msgQOpen (mq_name, 100, 4, MSG_Q_FIFO, OM_CREATE, NULL);
                     
                     dprint ("avioInstInit2: calling avioCfg1553Devices for device %d\n",avioAvRefCnt);
                     /* further initialize information for all 1553 devices */
                     avioCfg1553Devices (pInst->pDrvCtrl);

                     if (pDrvCtrl->uca32) {
                        dprint("avioInstInit2: setting up UCA32 interrupts %d",pDrvCtrl->uca32);
                        vxbIntConnect(pInst, 0, (VOIDFUNCPTR)intProcess32, pInst);
#ifdef SPIN_LOCK_PROTECT
                        spinLockIsrInit(&spinlock[deviceIndex],0); 
#endif    
                        vxbIntEnable(pInst, 0, (VOIDFUNCPTR)intProcess32, pInst);
                     }
                     else {
                        dprint("avioInstInit2: setting up UCA interrupts %d",pDrvCtrl->uca32);
                        vxbIntConnect(pInst, 0, (VOIDFUNCPTR)intProcess, pInst);
#ifdef SPIN_LOCK_PROTECT
                        spinLockIsrInit(&spinlock[deviceIndex],0); 
#endif    
                        vxbIntEnable(pInst, 0, (VOIDFUNCPTR)intProcess, pInst); 
                     }
                     break;
         
                  case AVIO_PCI_RAR15_DEVID:
                  case AVIO_PCI_RAR15XT_DEVID:
                     /* create a 1553 Message Queue for this device for interrupt notification purposes */
                     sprintf (mq_name,"/%s%d", AVIO_MSG_Q_NAME, deviceIndex);
                     pDrvCtrl->msg_q_id = msgQOpen (mq_name, 100, 4, MSG_Q_FIFO, OM_CREATE, NULL);

                     /* create a Commercial Message Queue for this device for interrupt notification purposes */
                     sprintf (mq2_name,"/%s%d", AVIO_COM_MQ_NAME, deviceIndex);
                     pDrvCtrl->msg_q_com_id = msgQOpen (mq2_name, 100, 4, MSG_Q_FIFO, OM_CREATE, NULL);
                     
                     dprint ("avioInstInit2: calling avioCfg1553Devices for device %d\n",avioAvRefCnt);
                     /* further initialize information for all 1553 devices */
                     avioCfg1553Devices (pInst->pDrvCtrl);

                     dprint("avioInstInit2: setting up UCA32 interrupts %d",pDrvCtrl->uca32);
                     vxbIntConnect (pInst, 0, (VOIDFUNCPTR)intProcess32, pInst);
#ifdef SPIN_LOCK_PROTECT
                     spinLockIsrInit (&spinlock[deviceIndex],0); 
#endif    
                     vxbIntEnable (pInst, 0, (VOIDFUNCPTR)intProcess32, pInst);
                     break;
         
                  /* Setup ISR for PLX-based PMC devices (x30 & P-SER) */
                  case AVIO_PCI_CEI830_DEVID:
                  case AVIO_PCI_R830RX_DEVID:
                  case AVIO_PCI_R830X820_DEVID:
                  case AVIO_PCI_RCEI830A_DEVID:
                  case AVIO_PCI_RAR_CPCI_DEVID:
                  case AVIO_PCI_PSER_DEVID:
                     /* create a Commercial Product Message Queue for this device for interrupt notification purposes */
                     sprintf (mq_name,"/%s%d", AVIO_COM_MQ_NAME, deviceIndex);
                     pDrvCtrl->msg_q_com_id = msgQOpen(mq_name, 100, 4, MSG_Q_FIFO, OM_CREATE, NULL);
                     
                     vxbIntConnect(pInst, 0, (VOIDFUNCPTR)intProcessCommercial, pInst);
#ifdef SPIN_LOCK_PROTECT
                     spinLockIsrInit(&spinlock[deviceIndex],0); 
#endif    
                     vxbIntEnable(pInst, 0, (VOIDFUNCPTR)intProcessCommercial, pInst);
                     break;
            
                  /* Setup ISR for PCI Express devices (RAR-XMC) */
                  case AVIO_PCI_RARXMC_DEVID:
                  case AVIO_PCI_RAR_MPCIE_DEVID:
                     /* create a Commercial Product Message Queue for this device for interrupt notification purposes */
                     sprintf (mq_name,"/%s%d", AVIO_COM_MQ_NAME, deviceIndex);
                     pDrvCtrl->msg_q_com_id = msgQOpen (mq_name, 100, 4, MSG_Q_FIFO, OM_CREATE,NULL);
                     
                     vxbIntConnect (pInst, 0, (VOIDFUNCPTR)intProcessCommercialPcie, pInst);
#ifdef SPIN_LOCK_PROTECT
                     spinLockIsrInit (&spinlock[deviceIndex],0); 
#endif    
                     vxbIntEnable (pInst, 0, (VOIDFUNCPTR)intProcessCommercialPcie, pInst);
                     break;
            
                  /* Interrupts are not supported with the P708/RP-708 or P-DIS */
                  case AVIO_PCI_RP708_DEVID:
                  case AVIO_PCI_P708_DEVID:
                  case AVIO_PCI_PDIS_DEVID:
                     break;
               }
               pDrvCtrl->devcnt++;
               return;
            }
            else {
               break;
            }
         }
         else {
            break;
         }
      }
      else {
         break;
      }
   }
   
   /* disconnect from VxBus */
   dprint("avioInstInit2: This should not happen %d\n",0);
   pInst->pDrvCtrl = NULL;
   vxAtomicDec(&avioAvRefCnt);
   free(pDrvCtrl);
   errnoSet(EFAULT);
}

/*********************************************************************************************************************
** NAME:
** intProcess()
**
** DESCRIPTION:
** ISR for legacy 1553 boards running UCA f/w v4/5.
**
** ARGUMENTS:                                         
** VXB_DEVICE_ID pInst
**
** RETURNS
**    N/A
**********************************************************************************************************************/
void intProcess(VXB_DEVICE_ID pInst)
{
   unsigned short  control_data;  /* control register contents */
   unsigned short * chanHiPtr[4]; /* pointers to each 1553 channel host interface */
   UINT int_bit;
   int chan;
   MSG_Q_ID mq_id;
   AVIO_DEVICE_CTRL_TYPE *pDrvCtrl = NULL;

#ifdef SPIN_LOCK_PROTECT
   int deviceIndex = pInst->unitNumber;
#endif 
   /* setup the interrupt bit mask */ 
   int_bit = UCA_INT_ASSERTED;
   flipw(&int_bit); 
   pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)pInst->pDrvCtrl;

   /* define the pointers to each 1553 channel's control register offset into the device host interface */
   chanHiPtr[0] = (unsigned short *)(pDrvCtrl->baseAddressBar0 + UCA16_1553_CTL_REG);
   chanHiPtr[1] = (unsigned short *)(pDrvCtrl->baseAddressBar0 + UCA16_1553_CH1_OFF + UCA16_1553_CTL_REG);
   chanHiPtr[2] = (unsigned short *)(pDrvCtrl->baseAddressBar0 + UCA16_1553_CH2_OFF + UCA16_1553_CTL_REG);
   chanHiPtr[3] = (unsigned short *)(pDrvCtrl->baseAddressBar0 + UCA16_1553_CH3_OFF + UCA16_1553_CTL_REG);
   
   mq_id = pDrvCtrl->msg_q_id;
   
   /* Clear the interrupt */
   /* read the CSC control reg then acknowledge via the int ack reg */

#ifdef SPIN_LOCK_PROTECT
   spinLockIsrTake(&spinlock[deviceIndex]);
#endif                            
   /* check each potential channel location */
   for (chan = 0; (chan < pDrvCtrl->dev_chan) && (chan < MAX_CHANNELS); chan++) {
      control_data = *chanHiPtr[chan];
      /* if the "interrupt pending" bit is high for this channel, acknowledge the interrupt */
      if (control_data & int_bit) {
         /* write 0 to the write_interrupt_bit */
         (chanHiPtr[chan])[UCA16_1553_ACK_REG] = 0; 

         /* Send a message to the host thread indicating the respective 1553 channel interrupt event occurred. */
         msgQSend (mq_id, (char *)&mq_data[chan], sizeof(unsigned), NO_WAIT, MSG_PRI_NORMAL);
      }
   }


#ifdef SPIN_LOCK_PROTECT
   spinLockIsrGive(&spinlock[deviceIndex]);
#endif
}

/*********************************************************************************************************************
** NAME:
** intProcess32()
**
** DESCRIPTION:
** ISR for Avionics IO 1553 and MultiProtocol boards (UCA32 f/w v6).
**
** ARGUMENTS:                                         
** VXB_DEVICE_ID pInst
**
** RETURNS
**    N/A
**********************************************************************************************************************/
void intProcess32(VXB_DEVICE_ID pInst)
{
   UINT control_data;                           /* control register contents */
   UINT temp;                                   /* interrupt acknowledge data word */
   unsigned int * interruptAckRegPtr;           /* pointer to the device host interface */
   unsigned int ctrlRegOffset;                  /* offset into the host interface for the next channel */
   int chan;                                    /* 1553 channel index */
   MSG_Q_ID mq_id;                              /* local 1553 interrupt event message queue handle, shared with API */
   AVIO_DEVICE_CTRL_TYPE *pDrvCtrl = NULL;      /* local reference to this device's Driver Control structure */
#ifdef SPIN_LOCK_PROTECT
   int deviceIndex = pInst->unitNumber;         /* device index */
#endif 
   
   /* access this device's Driver Control structure and 1553 interrupt event message queue handle */
   pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)pInst->pDrvCtrl;
   mq_id = pDrvCtrl->msg_q_id;

   /* assign the local reference to this device's base host interface */
   interruptAckRegPtr = (unsigned int *)pDrvCtrl->baseAddressBar0;
                                                      
#ifdef SPIN_LOCK_PROTECT
   spinLockIsrTake(&spinlock[deviceIndex]);
#endif                            
   /* check each installed 1553 channel to determine if one or more of them asserted the interrupt */
   ctrlRegOffset = UCA32_1553_CTL_REG;
   for (chan = 0; (chan < pDrvCtrl->dev_chan) && (chan < MAX_CHANNELS); chan++) {
      control_data = *(&interruptAckRegPtr[ctrlRegOffset]);
      fliplong(&control_data);
      if (control_data & UCA_INT_ASSERTED) {
         /* Clear the interrupt in the respective channel's 1553 Control Register */
         temp = UCA32_CLEAR_INT;
         fliplong(&temp);
         *(&interruptAckRegPtr[ctrlRegOffset]) = temp; 
         
         /* Send a message to the 1553-specific message queue, indicating to the host ISR thread an interrupt event occurred. */
         msgQSend (mq_id, (char *)&mq_data[chan], sizeof(unsigned), NO_WAIT, MSG_PRI_NORMAL);
      }

      /* increment for the next channel */
      ctrlRegOffset += UCA32_1553_CH_OFST;
   }

    /* if this device contains an ARINC component (I.E. a combo card), check for an ARINC firmware interrupt assertion */
   if (pDrvCtrl->arinc) {
      control_data = *(&interruptAckRegPtr[MULT_PT_ARINC_OFST]);
      fliplong(&control_data);
      /* if an ARINC interrupt is asserted, handle it */
      if (control_data & ARINC_INT_ASSERTED) {
         /* Clear the interrupt in the ARINC Global Enable Register */
         temp = ARINC_CLEAR_INT;
         fliplong(&temp);
         *(&interruptAckRegPtr[MULT_PT_ARINC_OFST]) = temp; 
         
         /* Send a message to the "commercial product" message queue, indicating to the host ISR thread an interrupt event occurred. */
         msgQSend (pDrvCtrl->msg_q_com_id, (char *)&mq_data[0], sizeof(unsigned), NO_WAIT, MSG_PRI_NORMAL);
      }
   }
   
#ifdef SPIN_LOCK_PROTECT
   spinLockIsrGive(&spinlock[deviceIndex]);
#endif
}

/*********************************************************************************************************************
** NAME:
** intProcessCommercial()
**
** DESCRIPTION:
** ISR for PCI based Commercial Avionics IO cards (x30 and P-SER), all having a PLX PCI9056 interface
**
** ARGUMENTS:                                         
** VXB_DEVICE_ID pInst
**
** RETURNS
**    N/A
**********************************************************************************************************************/
void intProcessCommercial(VXB_DEVICE_ID pInst)
{
   unsigned int control_data;                       /* control register contents */      
   unsigned int temp;                               /* interrupt acknowledge data word */
   AVIO_DEVICE_CTRL_TYPE *pDrvCtrl = NULL;          /* local reference to this device's Driver Control structure */
#ifdef SPIN_LOCK_PROTECT
   int deviceIndex = pInst->unitNumber;             /* device index */
#endif 

   /* access this device's Driver Control structure */
   pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)pInst->pDrvCtrl;

#ifdef SPIN_LOCK_PROTECT
   spinLockIsrTake(&spinlock[deviceIndex]);
#endif                            
   /* assign the local reference to this device's base host interface */
   control_data = *(unsigned int *)pDrvCtrl->baseAddressBar2;
   temp = control_data;
   fliplong(&control_data);
   /* if a "commercial product" interrupt is asserted, handle it */
   if (control_data & ARINC_INT_ASSERTED) {
      /* Clear the interrupt */
      *(unsigned int *)pDrvCtrl->baseAddressBar2 = temp; /* Clear the interrupt assertion by writing control register value back */
         
      /* Send a message to the "commercial product" message queue, indicating to the host ISR thread an interrupt event occurred. */
      msgQSend (pDrvCtrl->msg_q_com_id, (char *)&mq_data[0], sizeof(unsigned), NO_WAIT, MSG_PRI_NORMAL);
   }
#ifdef SPIN_LOCK_PROTECT
   spinLockIsrGive(&spinlock[deviceIndex]);
#endif
}



/*********************************************************************************************************************
** NAME:
** intProcessCommercialPcie()
**
** DESCRIPTION:
** ISR for PCI Express Commercial Avionics IO cards (RAR-XMC)
**
** ARGUMENTS:                                         
** VXB_DEVICE_ID pInst
**
** RETURNS
**    N/A
**********************************************************************************************************************/
void intProcessCommercialPcie(VXB_DEVICE_ID pInst)
{
   unsigned int control_data;                       /* control register contents */      
   unsigned int temp;                               /* interrupt acknowledge data word */
   AVIO_DEVICE_CTRL_TYPE *pDrvCtrl = NULL;          /* local reference to this device's Driver Control structure */
#ifdef SPIN_LOCK_PROTECT
   int deviceIndex = pInst->unitNumber;             /* device index */
#endif 
   
   /* access this device's Driver Control structure */
   pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)pInst->pDrvCtrl;

#ifdef SPIN_LOCK_PROTECT
   spinLockIsrTake(&spinlock[deviceIndex]);
#endif                            
   /* assign the local reference to this device's base host interface */
   control_data = *(unsigned int *)pDrvCtrl->baseAddressBar0;
   fliplong(&control_data);
   /* if a "commercial product" interrupt is asserted, handle it */
   if (control_data & ARINC_INT_ASSERTED) {
      /* Clear the interrupt */
      temp = ARINC_CLEAR_INT;
      fliplong (&temp);
      *(unsigned int *)pDrvCtrl->baseAddressBar0 = temp; /* Atomic write to "clear the interrupt asserted bit" to clear the interrupt */
         
      /* Send a message to the "commercial product" message queue, indicating to the host ISR thread an interrupt event occurred. */
      msgQSend (pDrvCtrl->msg_q_com_id, (char *)&mq_data[0], sizeof(unsigned), NO_WAIT, MSG_PRI_NORMAL);
   }
#ifdef SPIN_LOCK_PROTECT
   spinLockIsrGive(&spinlock[deviceIndex]);
#endif
}

