/*===================================================================================================================*
 * FILE:                                 A V I O   V X W   IO   D R V . C
 *===================================================================================================================*
 *
 *   COPYRIGHT 2019 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED ONLY IN ACCORDANCE WITH THE
 *   TERMS OF SUCH LICENSE AND WITH THE INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY
 *   OTHER COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE AND SHOULD NOT BE CONSTRUED
 *   AS A COMMITMENT BY ABACO SYSTEMS, INC.
 *
 *===================================================================================================================*
 *
 * Function: VxBus Gen2 Device Driver IO Support File
 *
 *===================================================================================================================*
 *
 * DESCRIPTION
 *
 * This library contains the support routines for the vxBus gen2 subsystem.
 *
 * INCLUDE FILES: 
 *
 *===================================================================================================================*
 *
 * Modification History
 *
 *   Date    By   Modifications
 * --------  ---  ---------------------------------------------------------------------------------------------------
 * 12/04/18  DMH  Initial
 * 09/12/19  BCH  Added tracking for open/close file descriptors to unmap the board with last instance.  Modified 
 *                 functions to check/handle status returns.  Removed avioIoRead and avioDrvInit.
 *
 *===================================================================================================================*/

/* for VxWorks 32-bit need to define before includes */
/* #define _WRS_STOP_USING_FUNCPTR */

#include <vxWorks.h>
#include <errnoLib.h>
#include <stdio.h>
#include <iosLib.h>
#include <hwif/buslib/vxbPciLib.h>
#include <devMemLib.h>
#include <sys/mman.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <devMemLib.h>
#include <sysLib.h>
#include <sllLib.h>
#include <avioVxwDrv.h>


typedef struct avio_pci_dev
{
  DEV_HDR devHdr;
  VXB_DEV_ID vxbDevId;
  void *mappedRegion;
  int cnt;
} AVIO_PCI_DEV;
typedef AVIO_PCI_DEV *AVIO_DESCRIPTOR;

/* prototypes */
LOCAL STATUS avioIoDevCreate(VXB_DEV_ID vxbDevId);
LOCAL void * avioIoOpen(AVIO_DESCRIPTOR pDesc, char *name, int flags);
LOCAL STATUS avioIoIoctl(AVIO_DESCRIPTOR pDesc, int request, int *argptr);
LOCAL int avioIoClose(AVIO_DESCRIPTOR pDesc);

/* globals */
LOCAL int avioIoDrvNum=ERROR; /* driver loaded/unloaded */
LOCAL SL_LIST avioDevList;  /* list of devices */


/*******************************************************************************
*
* avioIoDrvInit - Initialize the avio IO device driver
*
* RETURNS: OK or ERROR
*
*/
STATUS avioIoDrvInit()
{
  SL_NODE *pNode=NULL;

  
  dprint("avioIoDrvInit\n", 0);

  /* check if IO driver is not installed */
  if(avioIoDrvNum == ERROR) {
    /* load the avio IO driver with its services routines. */
    avioIoDrvNum = iosDrvInstall((DRV_CREATE_PTR)NULL, /* UNUSED */
                                 (DRV_REMOVE_PTR)NULL, /* UNUSED */
                                 (DRV_OPEN_PTR)avioIoOpen,
                                 (DRV_CLOSE_PTR)avioIoClose,
                                 (DRV_READ_PTR)NULL,
                                 (DRV_WRITE_PTR)NULL,
                                 (DRV_IOCTL_PTR)avioIoIoctl);
    if(avioIoDrvNum == ERROR) {
   	  dprint("avioIoDrvInit:  iosDrvInstall failed (errno: %d)\n", errnoGet());
      errnoSet(S_ioLib_NO_DRIVER);   	
      return ERROR;
    }
  }
  
  /* get the first node in the device list */
  pNode = SLL_FIRST(&avioDevList);
  /* search the device list for the node */
  while(pNode != NULL) {
    /* create an avio IO device at the empty node in the device list */
    avioIoDevCreate(((AVIO_DEVICE_CTRL_TYPE*)pNode)->vxbDevId);
    pNode = SLL_NEXT(pNode);
  }

  dprint("avioIoDrvInit exit\n", 0);
  
  return OK;
}


/*******************************************************************************
*
* avioIoDevRegister - adds an avio IO device for the VxBus IO system
*
* RETURNS: OK, EINVAL or ENXIO
*
*/
STATUS avioIoDevRegister(VXB_DEV_ID vxbDevId)
{
  STATUS status=OK;
  AVIO_DEVICE_CTRL_TYPE *pDrvCtrl=NULL;
  static BOOL avio_list_initialized=FALSE;

  
  if(vxbDevId == NULL)
    return EINVAL;

  dprint("avioIoDevRegister: 0x%X\n", vxbDevId);

  /* create device list if needed */
  if(!avio_list_initialized) {
    SLL_INIT(&avioDevList);
    avio_list_initialized = TRUE;
  }
  
  /* get the VxBus 'device context' also known as the 'driver control' */
  pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)vxbDevSoftcGet(vxbDevId);
  if(pDrvCtrl == NULL) {
    dprint("avioIoDevCreate:  vxbDevSoftcGet fail\n", 0);
    return ENXIO;
  }
  
  /* add the device to the list */
  SLL_PUT_AT_TAIL(&avioDevList, &pDrvCtrl->avioNode);

  /* if the avio IO driver is loaded then create an avio IO device in the device list */
  if(avioIoDrvNum != ERROR)
    status = avioIoDevCreate(vxbDevId);

  dprint("avioIoDevRegister exit\n", 0);
  
  return status;
}


/*******************************************************************************
*
* avioIoDevCreate - create an "avioDev" device
*
* This routine creates a avioDev device.
* It allocates memory for the necessary structures and initializes the device.
*
* RETURNS: OK, EINVAL, ENXIO or ERROR
*
*/
LOCAL STATUS avioIoDevCreate(VXB_DEV_ID vxbDevId)
{
  STATUS status=OK;
  char device_name[MAX_FILENAME_LENGTH];
  AVIO_PCI_DEV *pDev=NULL;
  AVIO_DEVICE_CTRL_TYPE *pDrvCtrl=NULL;
  PHYS_ADDR address=0;
  UINT devnum=0;

  
  if(vxbDevId == NULL)
    return EINVAL;

  dprint("avioIoDevCreate: 0x%X\n", vxbDevId);

  /* check if the IO driver is loaded */
  if(avioIoDrvNum == ERROR) {
    dprint("avioIoDevCreate: driver not loaded\n", 0);
    errnoSet(S_ioLib_NO_DRIVER); 
    return ERROR;
  }

  pDev = (AVIO_PCI_DEV*)malloc(sizeof(AVIO_PCI_DEV));
  if(pDev == NULL) {
    dprint("avioIoDevCreate:  malloc failed (errno: %d)\n", errnoGet());
    return ERROR;
  }
  memset(pDev, 0, sizeof(AVIO_PCI_DEV));
  
  /* get the device unit number */
  devnum = vxbDevUnitGet(vxbDevId);
  sprintf(device_name, "/%s/%d", AVIO_DEV_IO_NAME, devnum);
  pDev->vxbDevId = vxbDevId;
  
  /* add instance of I/O device to system */
  if(iosDevAdd(&pDev->devHdr, device_name, avioIoDrvNum) != OK) {
    dprint("avioIoDevCreate:  iosDevAdd failed (errno: %d)\n", errnoGet());
    free(pDev);
    return ERROR;
  }
  
  /* get the context for the VxBus device */
  pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)vxbDevSoftcGet(vxbDevId);
  if(pDrvCtrl == NULL) {
    dprint("avioIoDevCreate:  vxbDevSoftcGet fail\n", 0);
    free(pDev);
    return ENXIO;
  }
  
  address = (PHYS_ADDR)((long)pDrvCtrl->baseHandle[0]);
  sprintf(device_name, "/%s%d", AVIO_MEM_NAME, devnum);
  /* create a device mem object for the avio IO device */
  status = devMemCreate(device_name, MAP_SHARED, address, pDrvCtrl->size, MMU_ATTR_SUP_RW | MMU_ATTR_USR_RW | MMU_ATTR_VALID | MMU_ATTR_CACHE_OFF);
  if(status == ERROR) {
    iosDevDelete(&pDev->devHdr);
    dprint("avioIoDevCreate:  devMemCreate failed (errno: %d)\n", errnoGet());
  }
  
  dprint("avioIoDevCreate exit\n", 0);
  
  return status;
}

/*******************************************************************************
*
* avioIoOpen - open an avio IO device file
*
* This routine is called to open the avio IO driver.  If the "flag" is O_RDWR then
* map the board, this will allow the ioctls to return information from kernel
* to process space.
*
* RETURNS: pointer to AVIO_DESCRIPTOR or NULL (check errno) if it fails
*
*/
LOCAL void * avioIoOpen(AVIO_DESCRIPTOR pDesc, char *name, int flags)
{
  int fd=0;
  char device_name[MAX_FILENAME_LENGTH];
  AVIO_DEVICE_CTRL_TYPE *pDrvCtrl=NULL;

  
  if(pDesc == NULL) {
    errnoSet(EINVAL);
    return NULL;
  }

  dprint("avioIoOpen: %d\n", pDesc->cnt);

  /* if board is not mapped */
  if(pDesc->mappedRegion == NULL) {
     /* open the avio MEM device */
    sprintf(device_name, "/%s%d", AVIO_MEM_NAME, vxbDevUnitGet(pDesc->vxbDevId));
    fd = devMemOpen(device_name, O_RDWR, S_IRUSR | S_IWUSR);
    if(fd == ERROR) {
      dprint("avioIoOpen:  devMemOpen failed (errno: %d)\n", errnoGet()); 
      return NULL;
    }
    
    pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)vxbDevSoftcGet(pDesc->vxbDevId);
    if(pDrvCtrl == NULL) {
      dprint("avioIoOpen:  vxbDevSoftcGet fail\n", 0);
      errnoSet(ENXIO);
      close(fd);
      return NULL;
    }
    
    /* map the board to RTP space and set to share for RTP */
    pDesc->mappedRegion = mmap(NULL, pDrvCtrl->size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    if(pDesc->mappedRegion == (void *)MAP_FAILED) {
      dprint("avioIoOpen:  mmap failed (errno: %d)\n", errnoGet()); 
      pDesc->mappedRegion = NULL;
    }
    
    dprint("avioIoOpen:  mmap address 0x%lX\n", (unsigned long)((char*)pDesc->mappedRegion));
    
    close(fd);
  }
  
  pDesc->cnt++;
  
  dprint("avioIoOpen exit\n", 0);

  return ((void*)pDesc);
}

/*******************************************************************************
*
* avioIoIoctl - do device specific control function
*
* RETURNS: OK or ERROR (check errno)
*
*/
LOCAL STATUS avioIoIoctl(AVIO_DESCRIPTOR pDesc, int request, int *argptr)
{
  STATUS status=OK;
  AVIO_DEVICE_CTRL_TYPE *pDrvCtrl=NULL;
  PDEVMAP_T devPTR=NULL;
  UINT i;
  
  if((pDesc == NULL) || (argptr == NULL)) {
    errnoSet(EINVAL);
    return ERROR;
  }
  
  dprint("avioIoIoctl: %d\n", pDesc->cnt);

  pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)vxbDevSoftcGet(pDesc->vxbDevId);
  if(pDrvCtrl == NULL) {
    dprint("avioIoIoctl:  vxbDevSoftcGet fail\n", 0);
    errnoSet(ENXIO);
    return ERROR;
  }
  
  switch(request) {
    case PCI_GET_BOARD_TYPE:
      *argptr = (int)pDrvCtrl->dev_id;
      break;
    case PCI_GET_DEV_MAP:
      devPTR = (PDEVMAP_T)argptr;
      for(i = 0; i < MAX_MEMORY; i++) {
        devPTR->memStartPhy[i]    = 0;    /*Physical address is null.*/
        devPTR->memLengthBytes[i] = 0;    /*Length is null.*/
        devPTR->memHostBase[i]    = NULL; /*Host address is null.*/
        devPTR->flagMapToHost[i]  = 0;
      }
      devPTR->device = pDrvCtrl->device;
      devPTR->busType = pDrvCtrl->bustype;
      devPTR->interruptNumber = pDrvCtrl->irq;
      devPTR->VendorID = AVIO_PCI_VENDID;
      devPTR->DeviceID = pDrvCtrl->CondorDeviceID;
      devPTR->mapping = PHYSICAL_MAP;
      devPTR->memSections = pDrvCtrl->region_count;
      devPTR->memStartPhy[0] = (unsigned long)pDrvCtrl->baseHandle[0];
      devPTR->memHostBase[0] = (char*)pDesc->mappedRegion;
      devPTR->memLengthBytes[0] = (unsigned int)pDrvCtrl->size;
      devPTR->flagMapToHost[0] = 1;
      break;
    default:
      errnoSet(EINVAL);
      status = ERROR;
      break;
  }
  
  dprint("avioIoIoctl exit\n", 0);

  return status;
}


/*******************************************************************************
*
* avioIoClose - called when the device is closed
*
* For a simple device there are no resources to reclaim and you may specify
* a NULL pointer instead of creating an empty function. For more complex
* devices the resources allocated in your open function should be freed here.
*
* RETURNS: OK, ENXIO or ERROR
*
*/
LOCAL STATUS avioIoClose(AVIO_DESCRIPTOR pDesc)
{
  STATUS status=OK;
  AVIO_DEVICE_CTRL_TYPE *pDrvCtrl=NULL;

  if(pDesc == NULL) {
    errnoSet(EINVAL);
    return ERROR;
  }

  dprint("avioIoClose: %d\n", pDesc->cnt);
  
  if(pDesc->cnt > 0) {
    /* check if device is mapped */
    if((pDesc->cnt == 1) && (pDesc->mappedRegion != NULL)) {
      dprint("avioIoClose:  mmap address 0x%lX\n", (unsigned long)((char*)pDesc->mappedRegion));
        
      pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)vxbDevSoftcGet(pDesc->vxbDevId);
      if(pDrvCtrl == NULL) {
        dprint("avioIoClose:  vxbDevSoftcGet fail\n", 0);
        status = ENXIO;
      }
      else {
        if((status = munmap(pDesc->mappedRegion, pDrvCtrl->size)) != OK) {
          dprint("avioIoClose:  munmap fail (errno: %d)\n", errnoGet());
          status = ERROR;
        }
      }
      /* if munmap should fail it will not be an issue since when the RTP application exits the OS will release all resources */
      pDesc->mappedRegion = NULL;
    }
    pDesc->cnt--;
  }
    
  dprint("avioIoClose exit\n", 0);
 
  return status;
}
