/*===================================================================================================================*
 * FILE:                              C o n d o r V X W R T P D r v . c
 *===================================================================================================================*
 *
 *   COPYRIGHT (C) 1998-2021 BY ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND
 *   COPIED ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH
 *   THE INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY
 *   OTHER COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE
 *   AVAILABLE TO ANY OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF THE
 *   SOFTWARE IS HEREBY TRANSFERRED.
 *
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT
 *   NOTICE AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS.
 *
 *===================================================================================================================*
 *
 * FUNCTION:   CondorVXWRTPDrv.c
 *
 *===================================================================================================================*/

/* Revision:   1.23
   Date        Modification Notes
  ----------   ---------------------------------------------------------------
  02/29/2008   Unified PowerPC and x86 installation for Kernel Module and RTP
  10/22/2008   Added VxW_GetCommercialInfo in support of all Commercial APIs
  03/12/2009   Added support for the new IRQ_OFFSET and SPIN_LOCK_PROTECT 
               defines with values assigned in the CDF file.
  08/13/2009   Merged Commerical and 1553 product VxWorks driver - lets be common
  07/19/2010   Trying to keep up with unannounced 1553 VxWorks driver changes
                  Changed gefDrvVer routine name to be geipDrvVer
                  Changed gefHelp routine name to be geipHelp
                  Changed gefBoardShow routine name to be geipBoardShow
  03/07/2013   Add commercial compatibility
  11/13/2013   Add support for new ARINC boards
  11/05/2014   Change intProcess32 from intProcess
  12/13/2016   Renamed geipBoardShow to avioDeviceShow
               Renamed geipDrvVer to avioDrvVer
               Renamed geipHelp to avioHelp
               Added support for the Mini-PCIe boards CONDOR_R15_MPCIE & CONDOR_RAR_MPCIE
  12/06/2019   Added interrupt support for non-1553 boards (intProcessCommercial), assigned to the kernel via the
               routine rtpInterruptConnect (modified to add a parameter designating board type as 1553 or not).
  10/05/2021   Removing any instance of the character sequences "geip", "gefes", "gef" or related,
               replaced with "avio" unless otherwise noted; included consistency modifications for hump-style
               function names and underscore-delimiter variable names.
 *===================================================================================================================*/

#include "CondorVXWRTPDrv.h"
#include <sysLib.h>

#define COMMERCIAL_DEVICE_INTERRUPT_BIT 0x00000010;

UINT VxW_GetCommercialInfo(UINT device, 
                           UINT *bar0addr,
                           UINT *bar2addr,
                           UINT *manufId,
                           UINT *deviceId,
                           UINT *irq
                           )
{
   *bar0addr = pci_state[device].membase[0];
   *bar2addr = pci_state[device].membase[2];    
   *manufId  = pci_state[device].manufacturerID;    
   *deviceId = pci_state[device].CondorDeviceID;
   *irq      = pci_state[device].irq;
   return 1;
}

void avioDrvVer()
{
   printf("Abaco Systems Avionics Common VxWorks Drvier version %3.2f\n",CONDOR_VXW_DRV_VER);
   printf("Abaco Systems Avionics Common VxWorks Driver date (%s)\n\n",CONDOR_VXW_DRV_DATE);
}

void avioHelp()
{
   printf("       Abaco Systems Avionics Debug functions\n\n");
   printf("VxW_GetCommercialInfo(...) -- Returns commercial avionics board info\n");
   printf("VxW_Get1553CardID(device)  -- Returns the 1553 board type\n");
   printf("VxW_GetMQID(device)        -- Returns interrupt message queue ID\n");
   printf("VxW_Get1553HIF(device)     -- Returns the 1553 Host Interface value\n");
   printf("VxW_GetDevChan(device)     -- Returns number channels on the device\n");
   printf("VxW_GetDevIRQ(device)      -- Returns the devices IRQ (int line)\n");
   printf("avioBoardShow              -- Lists all Abaco Avionics boards\n");
   printf("memDescshow                -- Lists contents of sysPhysMemDesc table\n");
   printf("mapCondorPCIAddress        -- Shows mapping parameter (debug mode only)\n");
   printf("avioCfg1553Devices(device) -- Shows 1553 device configuration(debug mode only)\n");
   printf("avioInitPCI                -- Show Device and interrupt configuration \n");
   printf("avioDrvVer                 -- Display driver version and release date\n\n");
}

float VxW_GetDrvVer()
{
   return CONDOR_VXW_DRV_VER;
}

char * VxW_GetDrvDate()
{
   return CONDOR_VXW_DRV_DATE;
}

UINT VxW_Get1553CardID(UINT device )
{
   return pci_state[device].dev_id;
}

MSG_Q_ID VxW_GetMQID(UINT device)
{
   return pci_state[device].msg_q_id;
}

UINT VxW_Get1553HIF(int dev)
{
   return pci_state[dev].hif;
}

UINT VxW_GetDevChan(UINT device)
{
   return pci_state[device].dev_chan;
}

UINT VxW_GetDevIRQ(UINT device)
{
   return pci_state[device].irq;
}

UINT VxW_GetDevID(UINT device)
{
   return dev_id[device];
}

UINT get_device(UINT cardnum)
{
   return cardNumberIndex[cardnum];
}

void get_dev_id(UINT *devid)
{
   int i;
   for (i=0; i<MAX_BTA; i++)
     devid[i] = dev_id[i];
}

/****************************************************************
*
*  PROCEDURE - rtpInterruptConnect()
*
*  PARAMETERS - device - Board Number, valid range 0 - 15
*               commericalDevice - any non-zero value indicates 
*                                  invoked for a non-1553 board
*
*  RETURNS
*     BTD_OK          - Mapped address with no error
*     ERROR           - Error Mapping PCI addresses
****************************************************************/
int rtpInterruptConnect(int device, int commericalDevice)
{
   char mq_name[40];
   irq_value[device] = VxW_GetDevIRQ(device);
   dprint("rtpInterruptConnect for device",device);
  
#if _WRS_VXWORKS_MAJOR == 6  
   dprint("VxWorks 6.x interrupt setup o device %d\n", device);
   sprintf(mq_name,"/msg_q_%d",device);
   pci_state[device].msg_q_id = msgQOpen(mq_name,100,4,MSG_Q_FIFO,OM_CREATE,NULL);
#else
   dprint("VxWorks 5.5 interrupt setup o device %d\n", device);
   pci_state[device].msg_q_id = msgQCreate(100,4,MSG_Q_FIFO);
#endif  /* VXW_RTP_MODE */
   if(pci_state[device].msg_q_id == NULL)
   {
      dprint("Message Queue Create Error on %s\n",mq_name);
      dprint("msgQOpen fails with %x\n",errno);
      return 0x444;
   }  
   dprint("Message Queue ID = %x",pci_state[device].msg_q_id);
   int_level[device] = irq_value[device];

#ifdef VXW_PCI_X86
   sysIntDisablePIC(irq_value[device]);
   if (commericalDevice)
   {
      if(pciIntConnect (INUM_TO_IVEC (INT_NUM_GET (irq_value[device])),intProcessCommercial,device) == ERROR) 
         return ERROR;
   }
   else if(pci_state[device].hif & UCA32)
   {
      if(pciIntConnect (INUM_TO_IVEC (INT_NUM_GET (irq_value[device])),intProcess32,device) == ERROR) 
         return ERROR;
   }
   else
   {
      if(pciIntConnect (INUM_TO_IVEC (INT_NUM_GET (irq_value[device])),intProcess,device) == ERROR) 
         return ERROR;
   }
  
   sysIntEnablePIC(irq_value[device]);
   dprint("pciIntConnect for IVEC %x\n",INUM_TO_IVEC(irq_value[device] + INT_NUM_IRQ0 + IRQ_OFFSET));
   dprint("INT_NUM_GET(irq) = %x\n",INT_NUM_GET(irq_value[device]));
#else
#if defined (VXW_PCI_INT_CON)
   dprint("VXW_PCI_INT_CON pciIntConnect with int line %x\n", int_level[device]);
   if (commericalDevice)
   {
      if(pciIntConnect(INUM_TO_IVEC(int_level[device] + IRQ_OFFSET),intProcessCommercial,device))
         return ERROR;
   }
   else if(pci_state[device].hif & UCA32)
   {
      if(pciIntConnect(INUM_TO_IVEC(int_level[device] + IRQ_OFFSET),intProcess32,device))
         return ERROR;
   }
   else
   {
      if(pciIntConnect(INUM_TO_IVEC(int_level[device] + IRQ_OFFSET),intProcess,device))
         return ERROR;
   }   
   if(intEnable(int_level[device]) == ERROR)
      return ERROR;
#elif defined (VXW_INT_LINE_CON)
   dprint("VXW_INT_LINE_CON int line %x\n", int_level[device]);
   int_level[device] = int_level[device] + INT_VEC_IRQ0;
   dprint("intConnect using int line %x\n", int_level[device]);
   if (commericalDevice)
   {
      if(intConnect((void *)(int_level[device] + IRQ_OFFSET),intProcessCommercial,device))
         return ERROR;
      dprint("Using intProcessCommercial on device",device);
   }
   else if(pci_state[device].hif & UCA32)
   {
      if(intConnect((void *)(int_level[device] + IRQ_OFFSET),intProcess32,device))
         return ERROR;
       dprint("Using intProcess32 on device",device);
   }
   else
   {
      if(intConnect((void *)(int_level[device] + IRQ_OFFSET),intProcess,device))
         return ERROR;
      dprint("Using intProcess on device",device);
   }   
   if(intEnable(int_level[device]) == ERROR)
      return ERROR;
#else /* VXW_PCI_INT_CON */
   intDisable(int_level[device]);

   if (commericalDevice)
   {
      dprint("Using intConnnect on intProcessCommercial for IRQ %x\n",int_level[device]);
      if(intConnect(INUM_TO_IVEC(INT_VEC_IRQ0 + int_level[device] + IRQ_OFFSET),intProcessCommercial,device))
         return ERROR;
   }
   else if(pci_state[device].hif & UCA32)
   {
      dprint("Using intConnnect on intProcess32 for IRQ %x\n",int_level[device]);
      if(intConnect(INUM_TO_IVEC(INT_VEC_IRQ0 + int_level[device] + IRQ_OFFSET),intProcess32,device))
         return ERROR;
   }
   else
   {
      dprint("Using intConnnect on intProcess for IRQ %x\n",int_level[device]);
      if(intConnect(INUM_TO_IVEC(INT_VEC_IRQ0 + int_level[device] + IRQ_OFFSET),intProcess,device))
         return ERROR;	   
   }
        
 if(intEnable(int_level[device]) == ERROR)
      return ERROR;
#endif /* VXW_PCI_INT_CON */
#endif
   return 0;
}

int cnicInterruptConnect(int* params)
{
   /* params is pointer to three 32-bit ints:  [0] = pPdx, [1] = pPdx->sys_board_num, [2] = pointer to ISR */
   int pPdx = params[0];
   int device = params[1];
   int isr = params[2];
   int irqLevel = VxW_GetDevIRQ(device);

   irq_value[device] = irqLevel;
/*   pci_state[device].irq = irqLevel; */

   dprint("Entering cnicInterruptConnect :\n",0);
   dprint("  device = %d\n", device);
   dprint("  isr = 0x%X\n", isr);
   dprint("  irqLevel = %d\n", irqLevel);

#if defined(VXW_PCI_X86)

   sysIntDisablePIC(irqLevel);
   dprint("cnicInterruptConnect: calling pciIntConnect\n", 0);

   if (pciIntConnect(INUM_TO_IVEC(irq_value[device] + INT_NUM_IRQ0 + IRQ_OFFSET), (VOIDFUNCPTR)isr, pPdx) == ERROR)
   {
      dprint("cnicInterruptConnect: error calling pciIntConnect - errno = %d\n", errno);
      return ERROR;
   }
   dprint("cnicInterruptConnect: success calling pciIntConnect\n",0);
   sysIntEnablePIC(irqLevel);

#else
   #if defined (VXW_PCI_INT_CON)

      dprint("cnicInterruptConnect: calling pciIntConnect\n", 0);
      if (pciIntConnect(INUM_TO_IVEC(irqLevel + IRQ_OFFSET), (VOIDFUNCPTR)isr, pPdx))
      {
         dprint("cnicInterruptConnect: error calling pciIntConnect\n", 0);
         return ERROR;
      }
           
      if (intEnable(irqLevel) == ERROR)
      {
         dprint("cnicInterruptConnect: error calling intEnable\n", 0);
         return ERROR;
      }

   #else /* VXW_PCI_INT_CON */

      dprint("cnicInterruptConnect: calling intConnect\n", 0);
      if (intConnect(INUM_TO_IVEC(INT_VEC_IRQ0 + irqLevel + IRQ_OFFSET), (VOIDFUNCPTR)isr, pPdx))
      {
         dprint("cnicInterruptConnect: error calling intConnect\n", 0);
         return ERROR;
      }
      dprint("cnicInterruptConnect: success calling intConnect\n", 0);
           
      if (intEnable(irqLevel) == ERROR)
      {
         dprint("cnicInterruptConnect: error calling intEnable\n", 0);
         return ERROR;
      }
      dprint("cnicInterruptConnect: success calling intEnable\n", 0);

   #endif /* VXW_PCI_INT_CON */
#endif

   return 0;
}


/****************************************************************
*
*  PROCEDURE - mapCondorPCIAddress()
*
*  PARAMETERS - None
*
*  RETURNS
*     BTD_OK          - Mapped address with no error
*     ERROR           - Error Mapping PCI addresses
****************************************************************/
int mapCondorPCIAddress(void)
{
   int i,j, status,unit, memRegion, instance;
   int busNumber;
   int deviceNumber;
   int functionNumber;

   UINT32 mem, mask, page_addr, page_offset;
   pciregion_state * pPCIState;
   unsigned short *addr;

   status = SYSTEM_OK;

   dprint("mapCondorPCIAddress in debug mode %d\n",0);
   unit = 0;
   for (i = 0; i < num_boards; i++) {
      for (instance = 0; instance < MAX_BTA; instance++) {
         if (pciFindDevice(CONDOR_ENGINEERING,board_to_map[i],instance,&busNumber,&deviceNumber,&functionNumber) == ERROR)
            break;

         /* read address */
         pPCIState = &pci_state[0];

         for (j = 0; j < MAX_BTA; j++, pPCIState++) {
            if ((j != unit) && (pPCIState->busNumber == busNumber) &&
                (pPCIState->deviceNumber == deviceNumber) && (pPCIState->functionNumber == functionNumber))
               continue;
         }
      
         pci_state[unit].busNumber = busNumber;
         pci_state[unit].CondorDeviceID = board_to_map[i];
         pci_state[unit].functionNumber = functionNumber;
         pci_state[unit].deviceNumber = deviceNumber;
         pci_state[unit].bustype = 2;
         pci_state[unit].mem_base_reg = 0;
         pci_state[unit].region_count = 0;
         pci_state[unit].mapping = PHYSICAL_MAP; /* default to physical mapping */
      
         dprint("PCI device found on bus %d\n",busNumber);
         dprint("PCI device number = %x\n",board_to_map[i]);

         for (memRegion = 0; memRegion < MAX_MEMORY_REGIONS; memRegion++) {
            /* determine region size */
            pciConfigInLong(busNumber, deviceNumber,functionNumber,  addresses[memRegion], &mem);
            pciConfigOutLong(busNumber, deviceNumber,functionNumber,addresses[memRegion], (unsigned)~0);
            pciConfigInLong(busNumber, deviceNumber,functionNumber,addresses[memRegion], &mask);
            pciConfigOutLong(busNumber, deviceNumber,functionNumber,addresses[memRegion], mem);
         
            if(mem & PCI_BASE_ADDRESS_SPACE)
               continue;
            
            pci_state[unit].membase[memRegion] = mem & PCI_MEMBASE_MASK;    
            pci_state[unit].size[memRegion] = ~mask+1;

            if(!mask)
               continue;

            pci_state[unit].region_count++;
            
            page_addr = pci_state[unit].membase[memRegion] & 0xfffff000;
            dprint("page_addr = %x\n", page_addr);
            page_offset = pci_state[unit].membase[memRegion] - page_addr;
            dprint("page_offset = %x\n", page_offset);

            if(pci_state[unit].size[memRegion] < PAGE_SIZE) {
               dprint("PCI unadjusted region size = %lx\n",pci_state[unit].size[memRegion]);
               pci_state[unit].size[memRegion] = PAGE_SIZE;
            }
            
            if(pci_state[unit].size[memRegion] == 0x800000 || pci_state[unit].size[memRegion] == 0x400000 || pci_state[unit].size[memRegion] == 0x7ffffc) {
               pci_state[unit].mem_base_reg = memRegion;
               dprint("mem_base_region = %lx\n",pci_state[unit].mem_base_reg);
            }   

#if defined (VXW_X86_MAP_ADD)
            if((mem & PCI_BASE_ADDRESS_SPACE) == 0) /* map only memory regions not I/O space */
            { 
               dprint("Calling sysMmuMapAdd for x86 unit = %d\n",unit);
               if(sysMmuMapAdd((void *)mem,pci_state[unit].size[memRegion],VM_STATE_MASK_FOR_ALL,VM_STATE_FOR_PCI) == ERROR) 
                  return ERROR;
            }
#endif  /* VXW_X86_MAP_ADD */
         
            dprint("Found PCI memory region %d\n", memRegion);
            dprint("PCI base address = %lx\n",pci_state[unit].membase[memRegion]);
            dprint("PCI region size = %lx\n",pci_state[unit].size[memRegion]);      
         }     
      
         pciConfigInWord (busNumber, deviceNumber, functionNumber, PCI_CFG_VENDOR_ID, (UINT16 *)(&pci_state[unit].manufacturerID));  
         pciConfigInWord (busNumber, deviceNumber, functionNumber, PCI_CFG_DEVICE_ID, (UINT16 *)(&pci_state[unit].CondorDeviceID));  
         pciConfigInByte (busNumber, deviceNumber, functionNumber, PCI_CFG_DEV_INT_LINE, &pci_state[unit].irq);

         addr = (unsigned short *)pci_state[unit].membase[pci_state[unit].mem_base_reg];
         dprint("Mem base addr = %x\n",addr);

         dprint("Complete with unit %d\n",unit);
         unit++;
      }  
   }
   bd_mapped = 1;
   dprint("mapCondorPCIAddress Complete %d\n\n",unit);
   number_of_avio_devices = unit;
   return unit;
}

int avioCfg1553Devices(int bd_indx)
{
   UINT btype,nchan;
   int ci;
   unsigned short host_interface;
   UINT chndx;
   UINT card_table[] = {0,4,8,12};
   
   dprint("avioCfg1553Devices %d\n",bd_indx);
   chndx = card_table[bd_indx];
   
   dprint("base addr = %x\n",pci_state[bd_indx].membase[pci_state[bd_indx].mem_base_reg]);
   host_interface = *(unsigned short *)(pci_state[bd_indx].membase[pci_state[bd_indx].mem_base_reg]);
   flipw(&host_interface);
   dprint("Host_interface = %x\n",host_interface);                   

   dprint("Condor Device ID = %x\n",pci_state[bd_indx].CondorDeviceID);
   pci_state[bd_indx].hif=host_interface;
      
   dprint("1553 Qboard host interface = %x\n",host_interface);
   btype = (host_interface & 0x3e)>>1;
   nchan = (host_interface & 0x7c0)>>6;
   dev_chan[bd_indx]=nchan;
   pci_state[bd_indx].dev_chan = nchan;
   pci_state[bd_indx].dev_id = QdevID[btype-1];
   dev_id[bd_indx] = QdevID[btype-1];
   for(ci=0;ci<nchan;ci++)
   {
      cardNumberIndex[chndx] = bd_indx;
      chndx++;
   }
   dprint("btype = %x\n",btype);
   dprint("nchan = %d\n",nchan);
   dprint("dev_id = %x\n", dev_id[bd_indx]);

   dprint("avioCfg1553Devices Complete for device %d\n",bd_indx);
   fflush(stdout);
   return 0;   
}


/*******************************************************************************
*
* avioPciDrvInit - initialize the device driver
*
* This routine initializes and installs the driver.  It must be called
* before any avionics PCI devices are created.  
*
* RETURNS: OK, or ERROR if the driver installation fails.
*/

STATUS avioPciDrvInit (void)
{
    /* check if driver already installed */
   
   if(avio_pci_drv_num == -1)
      avio_pci_drv_num = 0;
   
   if (avio_pci_drv_num == 0)
   {
      avio_pci_drv_num = iosDrvInstall ( avioPciDrvCreate, 
                                         avioPciDrvDelete, 
                                         avioPciDrvOpen,
                                         avioPciDrvClose, 
                                         avioPciDrvRead, 
                                         avioPciDrvWrite, 
                                         avioPciDrvIoctl);
   }
   
   if(avio_pci_drv_num == ERROR)
   {
      return -1; /* ERROR;*/
   }
   else
      return 0;
}
/*******************************************************************************
*
* avioPciDrvCreate - create an avionics PCI device instance
*
* This routine creates an avionics PCI device instance.
* It allocates memory for the necessary structures and initializes the device.
*   
* RETURNS: OK, or ERROR if the call fails.
*/

static STATUS avioPciDrvCreate (DEV_HDR *hdr, char *name, void * mode)      /* name of device to be created      */
{

   CEI_INT devnum;          
   char device_name[80];
   
   if (avio_pci_drv_num < 1)
   {
      errno = S_ioLib_NO_DRIVER;
      return (ERROR);
   }
 
   devnum = *(int *)name;      
   dprint("avioPciDrvCreate for device %d\n",devnum);
   
   sprintf(device_name,"/%s/%d",AVIO_DEV_IO_NAME,devnum);
   dprint("device_name = %s\n",device_name);
   
   p_avio_pci_device[devnum] = (AVIO_PCI_DEV *) malloc (sizeof (AVIO_PCI_DEV));

   if (p_avio_pci_device[devnum] == NULL)
      return (ERROR);

   /* add instance of I/O device to system */
   if (iosDevAdd (&p_avio_pci_device[devnum]->devHdr, device_name, avio_pci_drv_num) != OK)
   {
      free ((char *) p_avio_pci_device);
      return (ERROR);
   }
   p_avio_pci_device[devnum]->thisDevice = devnum;
   p_avio_pci_device[devnum]->pci_state = &pci_state[devnum];
   p_avio_pci_device[devnum]->read_target = 0;

   return(OK);  
}

/*******************************************************************************
*
* avioPciDrvDelete- Destroy an avionics PCI device instance
*
*/

static STATUS avioPciDrvDelete(char * name) /* name of an avionics PCI device instance to be deleted */
{
   AVIO_PCI_DEV *p_avio_pci_device;
   char NameTail;
   char *pNameTail; 

   pNameTail = &NameTail;

    /* driver must be initialized */

   if (avio_pci_drv_num == ERROR)
   {
      errno = S_ioLib_NO_DRIVER;
      return (ERROR);
   }

   /* get pointer to an avionics PCI device instance descriptor */

   if ((p_avio_pci_device = (AVIO_PCI_DEV *) iosDevFind (name, &pNameTail)) == NULL)
   {
      return (ERROR);
   }

   /* remove the instance of the I/O device from the system */
   iosDevDelete (&p_avio_pci_device->devHdr);

   /* free memory */
   free ((char *) p_avio_pci_device);

   return (OK);
}

/*******************************************************************************
*
* avioPciDrvOpen - open an avionics PCI device file
*
* This routine is called to open an avionics PCI driver instance.
*    
* Whatever you return, the system will associate with the file
* descriptor and hand to your read and write function when they are called.
* For a simple device with no files simply return the device descriptor.  
* More complex devices will have an array of file specific structures 
* that includes  a pointer back to device descriptor.
*
* If no resources are avalible you can also return ERROR, which will cause 
* the open() call in the code to return error also.    
*
* NOMANUAL
*/

LOCAL int avioPciDrvOpen
(
   AVIO_PCI_DEV *p_avio_pci_device,     /* avionics PCI device instance descriptor */
   char *name,                          /* name of the file with device stipped off */ 
   int flags
)
{ 
   return ((int)p_avio_pci_device);
} 

/*******************************************************************************
*
* avioPciDrvWrite - write bytes to an avionics PCI device
*
* This routine writes a message of `nbytes' to an avionics PCI device.
*
* RETURNS: number of bytes written or ERROR
*
* NOMANUAL
*/

LOCAL int avioPciDrvWrite
(
   FAST AVIO_PCI_DEV *p_avio_pci_device,    /* pointer to an avionics PCI device descriptor */
   char *buffer,               /* buffer from which to copy bytes */
   int nbytes                  /* number of bytes to copy from buffer */
)
{
   /* TODO: do somthing with the data */
   return (nbytes);
}

/*******************************************************************************
*
* avioPciDrvIoctl - do device specific control function
*
* The ioctl requests recognized are entirely up to the designer, though
* select support requires implimentation of FIOSELECT and FIOUNSELECT
*
* The ioctl command also provides an excellent way to pass out of band
* status mode, and configuration information to or from the driver using
* non-standard ioctl codes.
*
* RETURNS:
*  OK 
*  ERROR if request is not supported
*
* NOMANUAL
*/

LOCAL STATUS avioPciDrvIoctl
(
   FAST AVIO_PCI_DEV *p_avio_pci_device, /* pointer to an avionics PCI device descriptor */
   int request,                          /* ioctl code */
   int *argptr                           /* where to send answer */
)
{
   STATUS   status = OK;
   int offset;

   switch (request)
   {
      case PCI_GET_REGION_SIZE:
         *argptr = (int)p_avio_pci_device->pci_state->size[p_avio_pci_device->pci_state->region];
         break;
      case PCI_GET_REGION_MEM:
         *argptr = (int)p_avio_pci_device->pci_state->membase[p_avio_pci_device->pci_state->region];
         break;
      case PCI_GET_INTERRUPT:
         *argptr = (int)p_avio_pci_device->pci_state->irq;
         break;
      case PCI_SET_REGION:
         p_avio_pci_device->pci_state->region = *argptr; 
         break;
      case PCI_SET_CNIC_ISR:
         /* params is pointer to three 32-bit ints:  [0] = pPdx, [1] = pPdx->sys_board_num, [2] = pointer to ISR */
         cnicInterruptConnect(argptr); 
         break;
      case PCI_GET_NUM_REGS:
         *argptr = p_avio_pci_device->pci_state->num_regions;
      case PCI_GET_HOST_INTRF:
         break;
      case PCI_GET_MSG_Q_ID:
         *argptr = (unsigned)p_avio_pci_device->pci_state->msg_q_id;
         break;
      case PCI_GET_PCIREGION_BYTE:
         offset = *argptr;
         pciConfigInByte(p_avio_pci_device->pci_state->busNumber,
                         p_avio_pci_device->pci_state->deviceNumber,
                         p_avio_pci_device->pci_state->functionNumber,
                         offset,
                        (UINT8 *)argptr);
         break;
      case PCI_GET_PCIREGION_WORD:
         offset = *argptr;
         pciConfigInWord(p_avio_pci_device->pci_state->busNumber,
                         p_avio_pci_device->pci_state->deviceNumber,
                         p_avio_pci_device->pci_state->functionNumber,
                         offset,
                        (UINT16 *)argptr);
         break;
      case PCI_GET_PCIREGION_DWORD:
         offset = *argptr;
         pciConfigInLong(p_avio_pci_device->pci_state->busNumber,
                         p_avio_pci_device->pci_state->deviceNumber,
                         p_avio_pci_device->pci_state->functionNumber,
                         offset,
                        (UINT32 *)argptr);
         break;
      case PCI_GET_BOARD_TYPE:
         *argptr = (int)p_avio_pci_device->pci_state->dev_id;
         break;
      case PCI_GET_DEVICE_ID:
         *argptr = (int)p_avio_pci_device->pci_state->CondorDeviceID;
         break;    
      case PCI_GET_NUM_CHAN:
         *argptr = (int)p_avio_pci_device->pci_state->dev_chan;
         break;
      case PCI_SYS_CLK_SET:
         {
            int crate;                    
            crate = (int)*argptr;
            if(sysClkRateSet(crate)== ERROR) /* This gives a accurate millisecond timer */
               *argptr = BTD_CLK_RATE_NOT_SET;  
            else
               *argptr = BTD_OK;
            break;  
         }
      case PCI_GET_DEV_MAP:
         {
            PDEVMAP_T devPTR;
            CEI_INT mindx,device,i;
            CEI_UINT membase,region_count,size;
          
            devPTR = (PDEVMAP_T)argptr;
            device = devPTR->device;

            /*Clear the region_count and struct*/
            region_count=0;
            for ( i = 0; i < MAX_MEMORY; i++ )
            {
               devPTR->memStartPhy[i]    = 0;    /*Physical address is null.*/
               devPTR->memLengthBytes[i] = 0;    /*Length is null.*/
               devPTR->memHostBase[i]    = NULL; /*Host address is null.*/
               devPTR->flagMapToHost[i]  = 0;
               
            }
          
            devPTR->busType = pci_state[device].bustype;
            devPTR->interruptNumber = pci_state[device].irq;
            devPTR->memSections = pci_state[device].region_count;
            devPTR->VendorID = 0x13c6;
            devPTR->DeviceID = pci_state[device].CondorDeviceID;

            for(mindx=0;mindx<MAX_MEMORY; mindx++)
            {
               membase = pci_state[device].membase[mindx];
               size = pci_state[device].size[mindx];

               if(size>0)
               {
                  devPTR->memStartPhy[region_count] = membase;
                  devPTR->memHostBase[region_count] = (pCEI_CHAR)membase;
                  devPTR->memLengthBytes[region_count] = size;
                  devPTR->flagMapToHost[region_count] = 1;
                  region_count++;
               }  
            }
            break;
         }
     default:
       return ERROR;
   }
    return (status); 
}


/*******************************************************************************
*
* avioPciDrvRead - does the reading for each file descritor
*
* RETURNS: number of bytes read or ERROR
*
* NOMANUAL
*/

static int avioPciDrvRead 
(
   AVIO_PCI_DEV *p_avio_pci_device, 
   char *  pBuf,    /* addr of destination buffer */
   int     maxBytes /* maximum bytes to read into buffer */
)
{
   unsigned request, offset;
   
   request = *(unsigned *)pBuf;
   
   switch (request)
   {
      case PCI_GET_REGION_SIZE:
         *(int *)pBuf = (int)p_avio_pci_device->pci_state->size[p_avio_pci_device->pci_state->region];
         break;
      case PCI_GET_REGION_MEM:
         *(unsigned *)pBuf = (int)p_avio_pci_device->pci_state->membase[p_avio_pci_device->pci_state->region];
         break;
      case PCI_GET_INTERRUPT:
         *(int *)pBuf = (int)p_avio_pci_device->pci_state->irq;
         break;
      case PCI_GET_NUM_REGS:
         *(int *)pBuf = p_avio_pci_device->pci_state->num_regions;
         break;
      case PCI_GET_HOST_INTRF:
         break;
      case PCI_GET_MSG_Q_ID:
         *(int *)pBuf = (unsigned)p_avio_pci_device->pci_state->msg_q_id;
         break;
      case PCI_GET_PCIREGION_BYTE:
         pBuf++;
         offset = *(unsigned *)pBuf;
         pciConfigInByte(p_avio_pci_device->pci_state->busNumber,
                         p_avio_pci_device->pci_state->deviceNumber,
                         p_avio_pci_device->pci_state->functionNumber,
                         offset,
                         (UINT8 *)pBuf);
         break;
      case PCI_GET_PCIREGION_WORD:
         pBuf++;
         offset = *(unsigned *)pBuf;
         pciConfigInWord(p_avio_pci_device->pci_state->busNumber,
                         p_avio_pci_device->pci_state->deviceNumber,
                         p_avio_pci_device->pci_state->functionNumber,
                         offset,
                         (UINT16 *)pBuf);
         break;
      case PCI_GET_PCIREGION_DWORD:
         pBuf++;
         offset = *(unsigned *)pBuf;
         pciConfigInLong(p_avio_pci_device->pci_state->busNumber,
                         p_avio_pci_device->pci_state->deviceNumber,
                         p_avio_pci_device->pci_state->functionNumber,
                         offset,
                         (UINT32 *)pBuf);
         break;
      case PCI_GET_BOARD_TYPE:
         *(int *)pBuf = (int)p_avio_pci_device->pci_state->dev_id;
         break;
      case PCI_GET_DEVICE_ID:
         *(int *)pBuf = (int)p_avio_pci_device->pci_state->CondorDeviceID;
         break;    
      case PCI_GET_NUM_CHAN:
         *(int *)pBuf = (int)p_avio_pci_device->pci_state->dev_chan;
         break;
      case PCI_GET_DEV_MAP:
      {
         PDEVMAP_T devPTR;
         CEI_INT mindx,device,i;
         CEI_UINT membase,region_count,size;
        
         devPTR = (PDEVMAP_T)pBuf;
         device = p_avio_pci_device->thisDevice;
        
         /*Clear the region_count and struct*/
         region_count=0;
         for (i = 0; i < MAX_MEMORY; i++)
         {
            devPTR->memStartPhy[i]    = 0;    /*Physical address is null.*/
            devPTR->memLengthBytes[i] = 0;    /*Length is null.*/
            devPTR->memHostBase[i]    = NULL; /*Host address is null.*/
            devPTR->flagMapToHost[i]  = 0;              
         }
        
         devPTR->busType = pci_state[device].bustype;
         devPTR->interruptNumber = pci_state[device].irq;
         devPTR->memSections = pci_state[device].region_count;
         devPTR->VendorID = pci_state[device].manufacturerID;       /* don't assume this is an Abaco Goleta board */
         devPTR->DeviceID = pci_state[device].CondorDeviceID;
         devPTR->mapping = pci_state[device].mapping;
        
         for (mindx = 0; mindx < MAX_MEMORY; mindx++)
         {
            membase = pci_state[device].membase[mindx];
            size = pci_state[device].size[mindx];

            if (size > 0)
            {
               devPTR->memStartPhy[region_count] = membase;
               devPTR->memHostBase[region_count] = (pCEI_CHAR)membase;
               devPTR->memLengthBytes[region_count] = size;
               devPTR->flagMapToHost[region_count] = 1;
               region_count++;
            }   
         }
      }
      break;
   }
   return maxBytes;
}    
     
/*******************************************************************************
*   
* avioPciDrvClose - called when the device is closed
*   
* For a simple device there are no resources to reclaim and you may specify 
* a NULL pointer instead of createing an empty function. For more complex
* devices the resources allocated in your open function should be freed here. 
*   
* RETURNS: OK or ERROR
*   
* NOMANUAL
*/  
    
static int avioPciDrvClose (AVIO_PCI_DEV *p_avio_pci_device)
{   
  return(OK);
}   
    
int avioPciDriverControl(int flag,int devnum)
{   
   STATUS status;
   int i;
   char device_name[80];
   
   if (flag)
   {
      dprint("avioPciDriverControl add driver %d\n",number_of_avio_devices);
      if(avioPciDrvInit() == ERROR)
      {
         return ERROR;
      }
   
      for(i = 0; i < number_of_avio_devices; i++)
      {
         dprint("calling avioPciDrvCreate for device %d\n",i);
         status = avioPciDrvCreate(NULL,(char *)&i,0);
         if (status == ERROR)
         {
            dprint("avioPciDrvCreate error %d\n",status);
            continue;
         }
      }
      return number_of_avio_devices;
   }
   else
   {
      sprintf(device_name,"/%s/%d",AVIO_DEV_IO_NAME,devnum);
      status = avioPciDrvDelete(device_name);
      return status;
   }
}

void avioInitPCI()
{
   int num_devices,dev_indx, status;   
   
   if (bd_mapped == 0 || vxwdebug == 1)
   {
      status = mapCondorPCIAddress();   /* detect and map all Abaco Systems Avionics Avionic boards */ 
   } 
  
   dprint("avioInitPCI %d\n",0);
   num_devices = avioPciDriverControl(DRV_INSTALL,0); 
   
   for(dev_indx =0; dev_indx<num_devices; dev_indx++)
   {
      switch (pci_state[dev_indx].CondorDeviceID)
      {
         case CONDOR_PCI_1553:   
         case CONDOR_QPCI_1553:
         case CONDOR_QCP_1553: 
         case CONDOR_QPCX_1553:
         case CONDOR_RXMC_1553:
         case CONDOR_RPCIE_1553:
         case CONDOR_LPCIE_1553:
         case CONDOR_RXMC2_1553:
         case CONDOR_RAR15_XMC_XT:
         case CONDOR_R15_PMC:
         case CONDOR_R15_MPCIE:

            dprint("calling avioCfg1553Devices for device %d\n",dev_indx);
            avioCfg1553Devices(dev_indx);
            dprint("calling rtpInterruptConnect for device %d\n",dev_indx);
            if(rtpInterruptConnect(dev_indx, 0) == ERROR)
               logMsg("Error setting the ISR for device %d\n",dev_indx,0,0,0,0,0);
            break;

         case CONDOR_CEI_520:
         case CONDOR_CEI_620:
         case CONDOR_CEI_820:
         case CONDOR_CEI_820TX:
            /* do x20 init stuff here */
            break;
         case CONDOR_CEI_830:
         case CONDOR_R830RX:
         case CONDOR_CEI_430:
         case CONDOR_CEI_530:
         case CONDOR_RAR_CPCI:
         case CONDOR_CEI_A30:
         case CONDOR_RAR_PCIE:
         case CONDOR_R830X820:
         case CONDOR_RCEI830A:
         case CONDOR_RAR_XMC:
         case CONDOR_RAR_MPCIE:
            /* do x30 init stuff here */
            if(rtpInterruptConnect(dev_indx, 1) == ERROR)
               logMsg("Error setting the ISR for device %d\n",dev_indx,0,0,0,0,0);
            break;
         case CONDOR_P708:
         case CONDOR_RP708:
         case CONDOR_PSER:
         case CONDOR_PDIS:  /* P-DIS   */
               /* do init stuff here */
            if(rtpInterruptConnect(dev_indx, 1) == ERROR)
               logMsg("Error setting the ISR for device %d\n",dev_indx,0,0,0,0,0);
            break;
         case CONDOR_AFDX:
            /* do CNIC init stuff here */
            break;
         case CONDOR_PMC_D1:
            /* do PMC-D1 init stuff here */
            break;
         default:
            logMsg("Unknown device type %x for device %d\n",pci_state[dev_indx].CondorDeviceID,dev_indx,0,0,0,0);
      }
   }
}

void avioDeviceShow(void)
{
   CEI_INT unit;
   char mapstr[80];    

   if (number_of_avio_devices > 0)
   {
      printf("%d Abaco Systems Avionics Device(s) Detected on the PCI Buses\n\n", number_of_avio_devices);
      printf("Unit  Bus  Dev  Func  Device ID        addr0        addr1     irq     Mapping\n");
      printf("-----------------------------------------------------------------------------\n");
   
      for (unit = 0; unit < number_of_avio_devices; unit++)
      {
         if (pci_state[unit].mapping)
            strcpy(mapstr,"  Shared");
         else
            strcpy(mapstr,"physical");
         
         printf("%2d   %4p %4p   %p     %4p   %10p   %10p    %4p    %8s\n",
                unit,
                pci_state[unit].busNumber,
                pci_state[unit].deviceNumber,
                pci_state[unit].functionNumber,
                pci_state[unit].CondorDeviceID,
                pci_state[unit].membase[0],
                pci_state[unit].membase[2],
                pci_state[unit].irq,
                mapstr);
      }
      printf("\n");
   }
   else
      printf("\nNo Abaco Systems Avionics Boards installed\n");
}

void memDescShow ()
{
   int i;

   for (i = 0; i < sysPhysMemDescNumEnt; ++i)
   {
      printf ("\n\n entry %d", i);
      printf ("\n virtual addr %p", sysPhysMemDesc[i].virtualAddr);
      printf ("\nphysical addr %p", sysPhysMemDesc[i].physicalAddr);
      printf ("\nsize in bytes %u", sysPhysMemDesc[i].len);
      printf ("\ninitialStateMask %p", sysPhysMemDesc[i].initialStateMask);
      printf ("\ninitialState %p", sysPhysMemDesc[i].initialState);
   }
   putchar ('\n');
}

static void intProcessCommercial(unsigned device)
{
   UINT control_data, temp, base_addr;
   UINT *HWR_Addr[5]; 
   int chan;
   MSG_Q_ID mq_id;
   
   base_addr = pci_state[device].membase[pci_state[device].mem_base_reg];
   mq_id = pci_state[device].msg_q_id;
                                                      
   HWR_Addr[0] = (UINT *)(base_addr); /* Ch 1 Interrupt  */

#ifdef SPIN_LOCK_PROTECT
   spinLockIsrTake(&spinlock[device]);
#endif                            
   
   control_data = *HWR_Addr[0];
   temp = control_data;                             /* BDW - save control register read value */
   fliplong(&control_data);
   if(control_data & COMMERCIAL_DEVICE_INTERRUPT_BIT)
   {
      (HWR_Addr[0])[0x0] = temp;    /* Clear the interrupt on arinc */
      msgQSend(mq_id,(char *)&mq_data[4],sizeof(unsigned),NO_WAIT, MSG_PRI_NORMAL);
   }

#ifdef SPIN_LOCK_PROTECT
   spinLockIsrGive(&spinlock[device]);
#endif
}

static void intProcess(unsigned device)
{
   unsigned short  control_data;
   short *HIR_Addr, *HWR_Addr[4]; 
   unsigned base_addr;
   int chan, status;
   MSG_Q_ID mq_id;
   
   unsigned channel_1=0x0;       
   unsigned channel_2=0x00200000;      
   unsigned channel_3=0x00400000;      
   unsigned channel_4=0x00600000;
   unsigned DATA_RAM =0x00100000;    
   unsigned HW_REG   =0x00000800;      
 
#ifdef VXW_PCI_PPC
#define int_bit 0x02
#else
#define int_bit 0x0200
#endif

   base_addr = pci_state[device].membase[pci_state[device].mem_base_reg];
   mq_id = pci_state[device].msg_q_id;
 
   HIR_Addr    = (short *)(base_addr);          /* Host Interface Registers */
   HWR_Addr[0] = (short *)(base_addr + channel_1 + HW_REG); /* HW Registers  */
   HWR_Addr[1] = (short *)(base_addr + channel_2 + HW_REG); /* HW Registers  */
   HWR_Addr[2] = (short *)(base_addr + channel_3 + HW_REG); /* HW Registers  */
   HWR_Addr[3] = (short *)(base_addr + channel_4 + HW_REG); /* HW Registers  */

   /* Clear the interrupt */
   /* read the CSC control reg */
   /* read the 1553_control_register on channel 1 */

#ifdef SPIN_LOCK_PROTECT
   spinLockIsrTake(&spinlock[device]);
#endif                            
   for(chan=0;chan<pci_state[device].dev_chan;chan++)
   {
      control_data = *HWR_Addr[chan]; 
      if(control_data & int_bit)
      {
                                             /* write 0 to the write_interrupt_bit */ 
         (HWR_Addr[chan])[0xb] = 0x0;        /* Clear the interrupt on the Channel */  
         status = msgQSend(mq_id,(char *)&mq_data[chan],sizeof(unsigned),NO_WAIT, MSG_PRI_NORMAL);
      }
   }
#ifdef SPIN_LOCK_PROTECT
   spinLockIsrGive(&spinlock[device]);
#endif
}

void intProcess32(UINT device)
{
   UINT control_data, temp, base_addr, int_bit32, arinc_int_bit;
   UINT *HWR_Addr[5]; 
   int chan;
   MSG_Q_ID mq_id;

   int_bit32 = 0x0200;
   arinc_int_bit = 0x10;

   base_addr = pci_state[device].membase[pci_state[device].mem_base_reg];
   mq_id = pci_state[device].msg_q_id;
                                                      
   HWR_Addr[0] = (UINT *)(base_addr + 0x30000l + 0x58); /* Ch 1 Interrupt  */
   HWR_Addr[1] = (UINT *)(base_addr + 0x34000l + 0x58); /* Ch 2 Interrupt  */
   HWR_Addr[2] = (UINT *)(base_addr + 0x38000l + 0x58); /* Ch 3 Interrupt  */
   HWR_Addr[3] = (UINT *)(base_addr + 0x3c000l + 0x58); /* Ch 4 Interrupt  */
   HWR_Addr[4] = (UINT *)(base_addr + 0x100000l);       /* ARINC on multiprotocol board */

   /* Clear the interrupt */
   /* read the CSC control reg */
   /* read the 1553_control_register on channel 1 */
                     
   for(chan=0;chan<pci_state[device].dev_chan;chan++)
   {
      control_data = *HWR_Addr[chan];
      fliplong(&control_data);
      if(control_data & int_bit32)
      {
         /* write 0 to the write_interrupt_bit */ 
         temp = 0x40000200;
         fliplong(&temp);
         (HWR_Addr[chan])[0x0] = temp;    /* Clear the interrupt on Channel 1 */ 
         msgQSend(mq_id,(char *)&mq_data[chan],sizeof(unsigned),NO_WAIT, MSG_PRI_NORMAL);
      }
   }
   if(pci_state[device].CondorDeviceID == 0x1542)
   {
      control_data = *HWR_Addr[4];
      fliplong(&control_data);
      if(control_data & arinc_int_bit)
      {
         temp = 0x40000010;
         fliplong(&temp);
         (HWR_Addr[4])[0x0] = temp;    /* Clear the interrupt on arinc */
         msgQSend(mq_id,(char *)&mq_data[4],sizeof(unsigned),NO_WAIT, MSG_PRI_NORMAL);
      }
   }
}

