/*====================================================================================================================*
 * FILE:                     C D E V _ V X W . C
 *====================================================================================================================*
 *
 *   COPYRIGHT 2004 - 2021 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED
 *   ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH THE INCLUSION
 *   OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY OTHER COPIES THEREOF
 *   MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE
 *   AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS, INC. 
 *
 *====================================================================================================================*
 *
 * FUNCTION:    Application Programmer's Interface for a CEI-x30 ARINC device.
 *              VxWorks 32-bit Operating System Specific Functionality
 *
 *    Internal CEI-x30 routines implemented herein:
 *
 *    timeGetTime                 - replacement for Windows O/S Get Timer call
 *    cei_acquire_device          - opens a session with the specified x30 device
 *    cei_acquire_device_identity - supplies the PCI configuration space ID values
 *    cei_acquire_thread_lock     - acquires sempahore/mutex lock
 *    cei_delay_milliseconds      - o/s specific delay utility in milliseconds
 *    cei_delay_seconds           - o/s specific delay utility in fractional seconds
 *    cei_get_host_system_time_x  - acquires host system time-of-year 64-bit uSec     
 *    cei_get_host_system_time    - acquires host system time-of-year 32-bit mSec     
 *    cei_get_irig_initial_time   - acquires host system time-of-year IRIG format
 *    cei_map_shared_memory       - maps memory region for sharing globals across processes
 *    cei_open_shared_lock        - creates a shared semaphore across processes
 *    cei_release_device          - closes a session with the specified x30 device
 *    cei_release_thread_lock     - releases sempahore/mutex lock     
 *    cei_read_32bit              - read a 32-bit value from an x30 device
 *    cei_write_32bit             - write a 32-bit value to an x30 device
 *    cei_init_interrupt          - configures an interrupt and ISR
 *    cei_terminate_interrupt     - terminate an interrupt
 *          
 *--------------------------------------------------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date    By   Comments
 * -------- ----- ------------------------------------------------------------
 * 10/06/04  BDW  Created.
 * 04/06/06  BDW  Added support for the Motorola MCP750 cPCI controller BSP.
 *                Corrected a bug in the cei_delay_seconds routine for MVME2300.
 * 04/11/06  BDW  Removed utility routines cei_read_plx_32bit & cei_write_plx_32bit
 * 08/24/06  BDW  Incorporated CEI_TYPES.H data type definitions in 
 *                  support of 64-bit Linux operating systems. 
 *                Swapped the BAR reference order by the device structure 
 *                   members baseAddr and devicePtr
 * 02/27/07  BDW  Added cei_acquire_device_identity() to supply the PCI 
 *                   Manufacturer and Device IDs for x30 products 
 * 08/15/07  BDW  Added support for the CEI-530 and R830RX boards
 * 02/18/08  BDW  Removed the boardInitialized array, relocated as a structure
 *                   member in the data structure array "device"
 * 03/10/08  BDW  Removed the declaration of the following global arrays, 
 *                relocated to CDEV_GLB.H:
 *                   bypassWrapTest
 *                   irigDacThreshold
 *                   a717syncWords
 *                Removed declaration of basicConfiguration, it was unused.
 *                Added reference to device[] structure as a parameter to
 *                   cei_acquire_device and cei_release_device, so they no 
 *                   longer access the global data structure directly.
 * 04/15/08  BDW  Added support for initiation and termination of h/w interrupts
 *                via routines cei_init_interrupt and cei_terminate_interrupt.
 * 06/16/09  BDW  Added the function cei_open_preload_lock for multiprocess support.
 * 07/01/09  BDW  Added new function cei_delay_milliseconds (o/s thread suspend).
 * 10/28/10  BDW  Modified the parameter list for cei_init_interrupt to account for      
 *                   interrupt setup requirements in the Windows-only CEI_Install library
 * 10/08/12  BDW  Added support for the VxBus driver interface.
 * 10/09/13  BDW  Moved the routine cei_vxb_interrupt_setup so it is encapsulated in
 *                the compilation when VXW_VXB_DRIVER is defined.
 *                
 *   Date    By   Ver. Comments
 * -------- ----- ---- -----------------------------------------------------------------------------------------------
 * 02/20/14  BDW  4.70 Eliminated some compiler warnings generated in cei_acquire_device
 * 08/07/15  BDW  4.71 Increased MAX_BOARDS from 8 to 16 for VxWorks and Integrity targets
 * 12/01/15  BDW  4.81 Modified cei_terminate_interrupt to include invocation of taskDelete
 *                     for VxBus based target systems, conditionally executed upon notification
 *                     to interruptProcessTaskVxb to release and close the message queue.
 *                     Modified interruptProcessTaskVxb to use the new commercial-only 
 *                     (CEI-x30 and P-SER boards) message queue for notification of an interrupt,
 *                     and to close this queue if notified to terminate by cei_terminate_interrupt.
 * 08/10/16  BDW  4.90 Modified to migrate to the Abaco VxBus Driver naming convention
 * 08/31/16  BDW  4.90 Removed unused legacy header include statements
 * 10/18/16  BDW  4.90 Introduced use of CEIx30 API-specific type defines via the file CEIX30_TYPES.H allowing
 *                     for easy build to common use of only signed integer return types and unsigned integer
 *                     parameter types (with the exception of 64-bit types).
 * 12/02/19  BDW  5.10 Switched the legacy PCI driver interrupt handling out of the API, as it was
 *                     moved to the driver by BC at some point, replaced with message queue usage.
 * 03/02/20  BDW  5.13 Revised the implementation for the taskDelay option in cei_delay_milliseconds.
 * 10/08/21  BDW  5.13 Reinstated support for the VxWorks 6.x Legacy PCI Driver.
 *====================================================================================================================*/ 
#include <stdio.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sysLib.h>
#include <semLib.h>
#include <taskLib.h>

#ifdef _WRS_KERNEL  
  #include <intLib.h>
#endif

#include "target_defines.h"

#if defined(VXW_VXB_DRIVER) || defined(VXW_DRIVER_OPTION)
  #include "avioVxwDrv.h"
#else
  #define AVIO_DEV_IO_NAME "avioDev"
#endif

#ifdef VXW_DRIVER_OPTION
  #include <ioLib.h> 
  #include "lowlevel.h"
#endif

#include "cdev_api.h"
#include "cdev_glb.h"

extern void (*x30InterruptFunctionPtr[MAX_BOARDS])(CEI_INT32, pCEI_UINT32);
#define MAX_DEV_OPEN_ERROR_STR_LEN  512
#define MAX_MEMORY_REGIONS          6        /* max # of BAR regions on a device */
#define TERMINATE_INTERRUPT_HANDLER 256

/* interrupt processing message queue, one per board, sourced by the driver */
static MSG_Q_ID mq_id[MAX_BOARDS] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

#ifdef VXW_VXB_DRIVER

/* The following section pertains to VxBus specific Device Driver types and structures */
static AVIO_DEVICE_INFO_TYPE *AvVxBusInfo = NULL;

#else

/* The following section pertains to non-VxBus Device Driver types and structures */
extern UINT VxW_GetCommercialInfo( UINT device, UINT *bar0addr, UINT *bar2addr, UINT *manufId, UINT *deviceId, UINT *irq );

#endif 

typedef struct device_attributes_type {
   CEI_UINT32 irq;                              /* Interrupt Request Level */
   CEI_UINT32 deviceID;
   CEI_UINT32 manufID;
   CEI_NATIVE_INT fd;
} DEVICE_ATTRIBUTES_TYPE;
  
/* PCI Device information structure - all things PCI for VxWorks */
typedef struct pciregion_state_type {
   CEI_UINT32       busNumber;                  /* PCI bus num of found device   */
   CEI_UINT32       deviceNumber;               /* PCI dev num of found device   */
   CEI_UINT32       functionNumber;             /* PCI func # of found device    */
   CEI_UINT32       irq;                        /* Device Interrupt Request Level*/
   CEI_UINT32       size[MAX_MEMORY_REGIONS];   /* size of each BAR for each dev */
   CEI_NATIVE_ULONG membase[MAX_MEMORY_REGIONS];/* base addr of each BAR for each*/
   CEI_UINT32       initialStateMask;           /* mask parameter to vmStateSet  */
   CEI_UINT32       initialState;               /* state parameter to vmStateSet */
   CEI_UINT32       deviceID;                   /* device ID for found device    */
   CEI_UINT32       manufID;                    /* manuf ID for found device     */
} CEI_PCI_REGION_STATE;
  
static DEVICE_ATTRIBUTES_TYPE deviceAttributes[MAX_BOARDS];

/* PCI Device information array */
CEI_PCI_REGION_STATE devicePciAttributes[MAX_BOARDS];

/* storage for Windows driver low-level board mapping error messages      */
static pCEI_CHAR deviceOpenErrorDesc = "Specified PCI Device Not Found";

/* Number of times the ISR has executed for each device - debug use mainly */
static CEI_INT32 interruptCounter[MAX_BOARDS] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

/* Task ID for the interrupt processing task */
#if defined(VXW_VXB_DRIVER) || defined(VXW_DRIVER_OPTION)
static TASK_ID interruptProcessTaskId[MAX_BOARDS] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
#else
static CEI_INT32 interruptProcessTaskId[MAX_BOARDS] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
#endif

/*===========================================================================*
 * INTERNAL MACRO:        L I T T L E _ E N D I A N _ 3 2
 *===========================================================================*
 *
 * FUNCTION:    Return the little-endian equivalent of the 4-byte integral
 *              argument (endian swap if host is big-endian).
 *
 * RETURN VAL:  The little-endian representation of the argument.
 *
 *===========================================================================*/
#define little_endian_32(longword) ( ((longword & 0x000000ff) << 24) | \
                                     ((longword & 0x0000ff00) << 8)  | \
                                     ((longword & 0xff0000)   >> 8)  | \
                                     ((longword & 0xff000000) >> 24) )

static SEM_ID ceiAccessMutex[MAX_BOARDS] = {         /* Device Access mutex  */
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL 
};


/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ G E T _ O P E N _ E R R O R
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the error description for a device open error
 *
 * PARAMETERS:  openErrorString -- (output) reference to assign to error string
 * 
 * RETURN VAL:  ARS_NORMAL  -- Mutex was acquired successfully
 *
 * GLOBALS:     deviceOpenErrorDesc
 *
 * DESCRIPTION: This routine retrieves the error description previously saved when a device open error is encountered.
 *
 *===================================================================================================================*/
pCEI_CHAR cei_get_open_error (CEI_VOID)
{
  return deviceOpenErrorDesc;
}


/*========================================================================* 
 * Non-VxWorks Operating System specific functionality - here to resolve
 *========================================================================*/
/*===================================================================================================================*
 * ENTRY POINT:                         C E I _ O P E N _ S H A R E D _ L O C K
 *===================================================================================================================*
 *
 *              NOTE: IMPLEMENTED FOR WINDOWS AND LINUX TARGET OPERATING SYSTEMS ONLY.
 *
 *===================================================================================================================*/
CEI_INT32 cei_open_shared_lock (CDEV_BOARD_TYPE sharedLockId, pCEI_INT32 sharedLockState)
{
   return ARS_NORMAL;
}
/*===================================================================================================================*
 * ENTRY POINT:                       C E I _ O P E N _ P R E L O A D _ L O C K
 *===================================================================================================================*
 *
 *              NOTE: IMPLEMENTED FOR WINDOWS AND LINUX TARGET OPERATING SYSTEMS ONLY.
 *
 *===================================================================================================================*/
CEI_INT32 cei_open_preload_lock (CDEV_BOARD_TYPE device)
{
   return ARS_NORMAL;
}

/*===================================================================================================================*
 * ENTRY POINT:                      C E I _ M A P _ S H A R E D _ M E M O R Y
 *===================================================================================================================*
 *
 *              NOTE: IMPLEMENTED FOR WINDOWS AND LINUX TARGET OPERATING SYSTEMS ONLY!
 *
 *===================================================================================================================*/
CEI_INT32 cei_map_shared_memory 
( 
   CEI_UCHAR    blockId,
   CEI_UINT32   blockSize,
   CEI_VOID  ** blockPtr,
   CEI_VOID  ** blockHandle,
   CEI_UINT32 * alreadyExists,
   CEI_UINT32 * isExclusive 
)
{
   return 1;
}

/*========================================================================* 
 * Operating System specific functionality
 *========================================================================*/
/*===========================================================================*
 * ENTRY POINT:            T I M E   G E T   T I M E
 *===========================================================================*
 *
 * FUNCTION:    Replicate the Windows O/S System Time Acquisition Method
 *
 * RETURN VAL:  32-bit 1 millisecond resolution system time
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine utilizes the VxWorks clock_gettime invocation
 *              to obtain the current system timer value with a
 *              resolution of 1 millisecond, replicating the Windows O/S 
 *              system time acquisition method of the same name.
 *
 *===========================================================================*/
DWORD timeGetTime()
{
   struct timespec ts;        /* Current time, use to get milliseconds */
   clock_gettime(CLOCK_REALTIME, &ts);
   return(DWORD)(ts.tv_nsec/1000000);
}

/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ A C Q U I R E _ T H R E A D _ L O C K    
 *===================================================================================================================*
 *
 * FUNCTION:    Block on a mutex or semaphore previously reserved for the specific device.  
 *              WINDOWS:    A mutex is used for single process multi-threading and multi-process applications.
 *              INTEGRITY:  A Posix mutex is used for single process multi-threading applications.
 *              VXWORKS:    A global semaphore is used for multi-process applications.
 *              LINUX:      A Posix mutex is used for single process multi-threading applications.
 *                          A global semaphore is used for multi-process applications.
 *              VISA:       A Visa Lock mechanism is used for multi-threading single and multi-process applications.
 *              LABVIEW RT: A Visa Lock mechanism is used for multi-threading single and multi-process applications.
 *
 * PARAMETERS:  deviceIndex -- (input) device to reference (0 to 15, 127 for Windows)
 * 
 * RETURN VAL:  ARS_NORMAL      -- The mutex/semaphore was acquired successfully
 *              ARS_FAILURE     -- Acquisition of the lock mechanism timed-out/failed (Linux, VxWorks and Integrity)          
 *              ARS_BOARD_MUTEX -- Acquisition of the lock mechanism timed-out/failed (Windows, VISA, LabVIEW R/T)
 *
 * GLOBALS:     ceiAccessLock
 *
 * DESCRIPTION: This routine blocks execution, waiting for access to the locking mechanism previously allocated for  
 *              the specific device.  Some operating systems support a time-out error, others do not; for those that
 *              do, the ARS_BOARD_MUTEX status code is used in place of the generic ARS_FAILURE status code.
 *
 *===================================================================================================================*/
CEI_INT32 cei_acquire_thread_lock (CDEV_BOARD_TYPE deviceIndex)
{
   CEI_INT32 status = ARS_NORMAL;
   if (ceiAccessMutex[deviceIndex] != NULL) {
      if (semTake(ceiAccessMutex[deviceIndex], WAIT_FOREVER) == OK) {
         status = ARS_FAILURE;
      }
   }
   else {
      status = ARS_FAILURE;
   }
   return status;
}

/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ R E L E A S E _ T H R E A D _ L O C K    
 *===================================================================================================================*
 *
 * FUNCTION:    Release a semaphore previously reserved/blocked by the application/API for the specific device.
 *
 *              WINDOWS:    A mutex is used for single process multi-threading and multi-process applications.
 *              INTEGRITY:  A Posix mutex is used for single process multi-threading applications.
 *              VXWORKS:    A global semaphore is used for multi-process applications.
 *              LINUX:      A Posix mutex is used for single process multi-threading applications.
 *                          A global semaphore is used for multi-process applications.
 *              VISA:       A Visa Lock mechanism is used for multi-threading single and multi-process applications.
 *              LABVIEW RT: A Visa Lock mechanism is used for multi-threading single and multi-process applications.
 *
 * PARAMETERS:  deviceIndex -- (input) device to reference (0 to 15, 127 for Windows)
 * 
 * RETURN VAL:  ARS_NORMAL      -- The mutex/semaphore was acquired successfully
 *              ARS_FAILURE     -- Acquisition of the lock mechanism timed-out/failed (Linux, VxWorks and Integrity)          
 *              ARS_BOARD_MUTEX -- Acquisition of the lock mechanism timed-out/failed (Windows, VISA, LabVIEW R/T)
 *
 * DESCRIPTION: This routine releases the resource access previously allocated as a locking mechanism for
 *              the specific device.  Some operating systems support a time-out error, others do not; for those that
 *              do, the ARS_BOARD_MUTEX status code is used in place of the generic ARS_FAILURE status code.
 *
 *===================================================================================================================*/
CEI_INT32 cei_release_thread_lock (CDEV_BOARD_TYPE deviceIndex)
{
   CEI_INT32 status = ARS_NORMAL;
   if (ceiAccessMutex[deviceIndex] != NULL) {
      if (semGive(ceiAccessMutex[deviceIndex]) == OK) {
         status = ARS_FAILURE;
      }
   }
   else {
      status = ARS_FAILURE;
   }
   return status;
}

/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ D E L A Y _ S E C O N D S
 *===================================================================================================================*
 *
 * FUNCTION:    Delay for a specified (fractional) number of seconds
 *
 * PARAMETERS:  CEI_FLOAT numberOfSeconds -- (input) a floating point representation of the number (or fraction) of
 *                                                   seconds to delay execution of the calling thread
 * 
 * RETURN VAL:  ARS_FAILURE -- For Integrity only, indicates a nanosleep invocation failure
 *              ARS_NORMAL  -- Success
 *
 * DESCRIPTION: This routine utilizes the following o/s-specific mechanism for delaying thread execution:
 *
 *              WINDOWS:    A delay-polling method is used based on repeated invocations of timeGetTime, having a 
 *                          typical resolution of 1 millisecond.
 *              INTEGRITY:  Delay method is optional - for Posix, nanosleep is used; otherwise, usleep is used.
 *              VXWORKS:    Delay method is BSP dependent - either sysUsDelay, sysMsDelay, or taskDelay is used.
 *              LINUX:      A delay-polling method is used based on repeated invocations of clock()
 *              VISA:       The C standard library Sleep() function is used.
 *              LABVIEW RT: The o/s function Delay() is used
 *
 *===================================================================================================================*/
CEI_INT32 cei_delay_seconds (CEI_FLOAT numberOfSeconds)
{
   CEI_UINT32 delay;

#if defined (DELAY_USE_SYS_US_DELAY)
   delay = (CEI_UINT32)((float)TIMER_CLOCK_TICKS_PER_SECOND * numberOfSeconds);
   sysUsDelay(delay);
#elif defined (DELAY_USE_SYS_MS_DELAY)
   delay = (CEI_UINT32)((float)TIMER_CLOCK_TICKS_PER_SECOND * numberOfSeconds);
   sysMsDelay(delay);
#else
   delay = (CEI_UINT32)(((float)sysClkRateGet() * numberOfSeconds) + 1U);
   taskDelay((CEI_INT32)delay);
#endif

   return ARS_NORMAL;
}

/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ D E L A Y _ M I L L I S E C O N D S
 *===================================================================================================================*
 *
 * FUNCTION:    Delay for a specified number of milliseconds
 *
 * PARAMETERS:  CEI_UINT32 numberOfMillieconds -- (input) the number of milliseconds to delay execution of the 
 *                                                        calling thread
 * 
 * RETURN VAL:  ARS_FAILURE -- For Integrity only, indicates a nanosleep invocation failure
 *              ARS_NORMAL  -- Success
 *
 * DESCRIPTION: This routine utilizes the following o/s-specific mechanism for delaying thread execution:
 *
 *              WINDOWS:    The C standard library Sleep() function is used.
 *              INTEGRITY:  Delay method is optional - for Posix, nanosleep is used; otherwise, usleep is used.
 *              VXWORKS:    Delay method is BSP dependent - either sysUsDelay, sysMsDelay, or taskDelay is used.
 *              LINUX:      A delay-polling method is used based on repeated invocations of clock()
 *              VISA:       The C standard library Sleep() function is used.
 *              LABVIEW RT: The o/s function Delay() is used
 *
 *===================================================================================================================*/
CEI_INT32 cei_delay_milliseconds (CEI_UINT32 numberOfMilliseconds)
{
   CEI_UINT32 delay;
   CEI_UINT32 delayResolution;

#if defined (DELAY_USE_SYS_US_DELAY)
   delayResolution = (CEI_UINT32)(TIMER_CLOCK_TICKS_PER_SECOND/1000);
   if (delayResolution == 0U) {
      delayResolution = 1U;
   }
   delay = (CEI_UINT32)(delayResolution * numberOfMilliseconds);
   sysUsDelay(delay);
#elif defined (DELAY_USE_SYS_MS_DELAY)
   delayResolution = (CEI_UINT32)(TIMER_CLOCK_TICKS_PER_SECOND/1000);
   if (delayResolution == 0U) {
      delayResolution = 1U;
   }
   delay = (CEI_UINT32)(delayResolution * numberOfMilliseconds);
   sysMsDelay(delay);
#else
   delayResolution = (CEI_UINT32)(1000000/((CEI_UINT32)sysClkRateGet()));
   if (delayResolution > (numberOfMilliseconds*1000)) {
      delay = 1U;
   }
   else {
      delay = (CEI_UINT32)((numberOfMilliseconds*1000)/delayResolution);
   }
   taskDelay((CEI_INT32)delay);
#endif

   return ARS_NORMAL;
}


/*===================================================================================================================*
 * ENTRY POINT:                     C E I _ G E T _ H O S T _ S Y S T E M _ T I M E
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the host 32-bit 1ms reference time value, supplied via operating system "System Time"
 *
 * PARAMETERS:  pCEI_UINT32 hostTime -- (output) 32-bit 1 millisecond resolution time
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current System Time in milliseconds
 *
 *              WINDOWS:    System time is acquired via the o/s function timeGetTime().
 *              INTEGRITY:  System time is acquired via the o/s function clock_gettime (Posix), GetClockTime otherwise.
 *              VXWORKS:    System time is acquired via combination of localtime() and mktime() invocations.
 *              LINUX:      System time is acquired via the o/s function clock().
 *              VISA:       Invokes cei_get_host_system_time_x to acquire system time.
 *              LABVIEW RT: Invokes cei_get_host_system_time_x to acquire system time.
 *
 *===================================================================================================================*/
CEI_INT32 cei_get_host_system_time (pCEI_UINT32 hostTime)
{
   time_t   now;             /* Current time, modify to get Time Reference Point. */
   struct   tm start;        /* Formatted "now" current time, modified for TRP.   */
   time_t   then;            /* Time Reference Point, Midnight or Jan 1.          */

   CEI_INT32 status;
   struct timespec ts;                     /* Current time, use to get milliseconds */
   
   CEI_UINT32 last_milli; /* This is the time we use to wait for time to change.    */
   TIME_TAG_TYPE diff;    /* This is the difference time we return in microseconds, */
                          /*  either time since Midnight or time since Jan 1.       */
   time(&now);                    /* Time in seconds since midnight Jan 1, 1970 GMT */
   start = *localtime(&now);      /* Local                                          */
   start.tm_sec  = 0;             /*   time                                         */
   start.tm_min  = 0;             /*     at                                         */
   start.tm_hour = 0;             /*       Jan 1                                    */
   start.tm_mday = 1;             /*         midnight (This is day 0)               */
   start.tm_mon  = 0;             /*           this                                 */
   start.tm_yday = 0;             /*             year                               */
   then = mktime(&start);         /* local Jan 1 to time in seconds since Jan 1     */
                                  /* We do IRIG time, which makes Jan 1 "day 0"     */

   status = clock_gettime(CLOCK_REALTIME, &ts);
   last_milli = (CEI_UINT32)(ts.tv_nsec/1000000);   /*  save millisecond */
   while (1) {                  /* Wait for milliseconds counter to change */
      clock_gettime(CLOCK_REALTIME, &ts);
      if ((CEI_UINT32)(ts.tv_nsec/1000000) != last_milli) {
         break;                 /* Break at exact time tick change */
      }
   }
   diff = ts.tv_sec - then;     /* We only want the time since Midnight or Jan 1  */
   diff *= (TIME_TAG_TYPE)1000000; /* Number of Seconds converted to microseconds */
   diff += (ts.tv_nsec/1000);   /* plus number of Milliseconds converted to       */
   *hostTime = (CEI_UINT32)diff;
   status = ARS_NORMAL;
   return status;
}

/*===================================================================================================================*
 * ENTRY POINT:                     C E I _ G E T _ H O S T _ S Y S T E M _ T I M E _ X
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the host 64-bit 1us calendar time value, supplied via operating system as time since Jan 1.
 *
 * PARAMETERS:  TIME_TAG_TYPE hostTime -- (output) 64-bit 1us millisecond resolution time
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current 64-bit 1us resolution calendar time value
 *
 *              WINDOWS:    Calendar time is acquired via combination of localtime() and mktime() invocations.
 *              INTEGRITY:  Calendar time is acquired via combination of localtime() and mktime() invocations (Posix only).
 *              VXWORKS:    Calendar time is acquired via combination of localtime() and mktime() invocations.
 *              LINUX:      Calendar time is acquired via combination of localtime() and mktime() invocations.
 *              VISA:       Calendar time is acquired via combination of time() and localtime() inocations.
 *              LABVIEW RT: Calendar time is acquired via combination of time() and localtime() inocations.
 *
 *===================================================================================================================*/
CEI_VOID cei_get_host_system_time_x (TIME_TAG_TYPE * hostTime)
{
   CEI_INT32 status;
   CEI_UINT32 time;
   status = cei_get_host_system_time(&time);
   *hostTime = (TIME_TAG_TYPE)time;
   if (status) ;
}


/*===================================================================================================================*
 * ENTRY POINT:                  C E I _ G E T _ I R I G _ I N I T I A L _ T I M E
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the host reference calendar time/date value for CEI-x30 IRIG generator initialization
 *
 * PARAMETERS:  time -- (output) 32-bit packed IRIG time value
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current System Time in calendar time/date and formats it into an 
 *              IRIG bit-encoded format value.
 *
 *===================================================================================================================*/
CEI_VOID cei_get_irig_initial_time (pCEI_UINT32 time)
{
   TIME_TAG_TYPE system_time;
   CEI_NATIVE_UINT seconds;
   CEI_NATIVE_UINT tens_of_seconds;
   CEI_NATIVE_UINT minutes;
   CEI_NATIVE_UINT tens_of_minutes;
   CEI_NATIVE_UINT hours;
   CEI_NATIVE_UINT tens_of_hours;
   CEI_NATIVE_UINT days;
   CEI_NATIVE_UINT tens_of_days;
   CEI_NATIVE_UINT hundreds_of_days;

   union
   {
      struct irigTimeFormat
           {
              CEI_NATIVE_UINT seconds          :4;
              CEI_NATIVE_UINT tens_of_seconds  :3;
              CEI_NATIVE_UINT minutes          :4;
              CEI_NATIVE_UINT tens_of_minutes  :3;
              CEI_NATIVE_UINT hours            :4;
              CEI_NATIVE_UINT tens_of_hours    :2;
              CEI_NATIVE_UINT days             :4;
              CEI_NATIVE_UINT tens_of_days     :4;
              CEI_NATIVE_UINT hundreds_of_days :2;
           } IRIG_TIME_FORMAT;
      CEI_UINT32 timeValue;
   } timeConvert;

   cei_get_host_system_time_x (&system_time);

   seconds = (CEI_UINT32)(system_time / (TIME_TAG_TYPE)1000000);
   if ( (system_time % (TIME_TAG_TYPE)1000000) > (TIME_TAG_TYPE)500000) {
      seconds++;
   }

   /* Extract Minutes from Seconds */
   minutes = seconds/60;
   seconds = seconds%60;
   tens_of_seconds = seconds/10;
   seconds = seconds%10;

   /* Extract Hours from Minutes */
   hours = minutes/60;  
   minutes = minutes%60;
   tens_of_minutes = minutes/10;
   minutes = minutes%10;

   /* Extract Days from Hours */
   days = hours/24;     
   hours = hours%24;     
   tens_of_hours = hours/10;    
   hours = hours%10;    
   days++;                       /* account for no such thing as day 0 */
   hundreds_of_days = days/100;
   days = days%100;
   tens_of_days = days/10;
   days = days%10;

   timeConvert.IRIG_TIME_FORMAT.seconds          = seconds;
   timeConvert.IRIG_TIME_FORMAT.tens_of_seconds  = tens_of_seconds;
   timeConvert.IRIG_TIME_FORMAT.minutes          = minutes;
   timeConvert.IRIG_TIME_FORMAT.tens_of_minutes  = tens_of_minutes;
   timeConvert.IRIG_TIME_FORMAT.hours            = hours;
   timeConvert.IRIG_TIME_FORMAT.tens_of_hours    = tens_of_hours;
   timeConvert.IRIG_TIME_FORMAT.days             = days;
   timeConvert.IRIG_TIME_FORMAT.tens_of_days     = tens_of_days;
   timeConvert.IRIG_TIME_FORMAT.hundreds_of_days = hundreds_of_days;
   
   *time = timeConvert.timeValue;
}

#define PCI_MEMBASE_MASK        ~0xf     /* mask for memory base address */
#define PCI_BASE_ADDRESS_SPACE  0x01     /* 0 = memory, 1 = I/O */

/*===================================================================================================================*
 * ENTRY POINT:                         C E I _ A C Q U I R E _ D E V I C E
 *===================================================================================================================*
 *
 * FUNCTION:    Open a session with a device
 *              
 * PARAMETERS:  CDEV_BOARD_TYPE deviceIndex  -- (input)  device to reference (0 to 15, 127 for Windows)
 *              ARINC_LOCAL_TYPE * device    -- (output) reference to ARINC_LOCAL_TYPE struct
 * 
 * RETURN VAL:  ARS_NORMAL     -- Session was opened successfully
 *              ARS_DRIVERFAIL -- Under Windows, Linux, Visa, and LabVIEW R/T, the respective driver 'Open Session' 
 *                                or 'Resource Acquisition' operations failed.
 *              ARS_FAILURE    -- Under VxWorks and Integrity, the driver 'Open Session' opreation failed.
 *                                Under Windows, the CEI_Install.DLL cannot be accessed
 *
 * DESCRIPTION: This routine searches for the instance of a CEI-x30 device matching the supplied device ID, with the 
 *              resource acquisition for each instance of a device found limited to the first execution of this 
 *              routine.  In most cases subsequent invocations will simply retrieve the resources allocated for the 
 *              respective device index from global/static variables, (with the exception of some PPC SBCs such as  
 *              Thales where the "sysPciMemToLocalAdrs" translation will be applied to previously mapped addresses).
 *
 *===================================================================================================================*/
CEI_INT32 cei_acquire_device (CDEV_BOARD_TYPE deviceIndex, ARINC_LOCAL_TYPE * device)
{
   CEI_INT32  status = ARS_NORMAL;
#ifdef VXW_VXB_DRIVER
   struct vxbDriverControl ctrl;
  
   if(AvVxBusInfo == NULL) {
      AvVxBusInfo = (AVIO_DEVICE_INFO_TYPE *)malloc(sizeof(AVIO_DEVICE_INFO_TYPE));
      if(AvVxBusInfo == NULL) {
         return ARS_FAILURE;
      }
      AvVxBusInfo->dev_use_count[deviceIndex] = 0;
   }

   ctrl.driverName = AVIO_DRV_NAME;
   ctrl.drvCtrlCmd = VXB_GET_DEV_MAP;
   ctrl.drvCtrlInfo = AvVxBusInfo;
   if (vxbDevMethodRun (DEVMETHOD_CALL(driverControl), &ctrl) != OK) {
      free (AvVxBusInfo);
      AvVxBusInfo = NULL;
      return API_VXB_METHOD_ERR;
   }

#if 0
   printf("AvVxBusInfo.device_present[%d] = %x\n", deviceIndex, AvVxBusInfo->device_present[deviceIndex]);
   printf("AvVxBusInfo.CondorDeviceID[%d] = %x\n", deviceIndex, AvVxBusInfo->CondorDeviceID[deviceIndex]);
   printf("AvVxBusInfo.membase[%d] = %x\n",deviceIndex, (int)AvVxBusInfo->membase[deviceIndex]);
   printf("AvVxBusInfo.msg_q_id[%d] = %x\n",deviceIndex, (int)AvVxBusInfo->msg_q_id[deviceIndex]);
#endif
   
   /* increment the local API x30 device-in-use counter */
   AvVxBusInfo->dev_use_count[deviceIndex]++;
   
   /* log the device attributes actually used */
   deviceAttributes[deviceIndex].manufID  = (CEI_UINT32)AVIO_PCI_VENDID;  
   deviceAttributes[deviceIndex].deviceID = (CEI_UINT32)AvVxBusInfo->CondorDeviceID[deviceIndex];  
   deviceAttributes[deviceIndex].irq   = (CEI_UINT32)AvVxBusInfo->irq[deviceIndex];
   
   device->baseAddr  = (pCEI_UINT32)AvVxBusInfo->baseAddressBar0[deviceIndex];
   device->devicePtr = (pCEI_UINT32)AvVxBusInfo->baseAddressBar2[deviceIndex];
   
   devicePciAttributes[deviceIndex].manufID =  (CEI_UINT32)AVIO_PCI_VENDID;         
   devicePciAttributes[deviceIndex].deviceID = (CEI_UINT32)AvVxBusInfo->CondorDeviceID[deviceIndex];
   devicePciAttributes[deviceIndex].irq = (CEI_UINT32)AvVxBusInfo->irq[deviceIndex];
   devicePciAttributes[deviceIndex].membase[0] = (CEI_NATIVE_ULONG)AvVxBusInfo->membase[deviceIndex];
   devicePciAttributes[deviceIndex].membase[2] = (CEI_NATIVE_ULONG)AvVxBusInfo->membaseBar2[deviceIndex];

#elif defined(VXW_DRIVER_OPTION)
   CEI_NATIVE_INT fd=0;
   CEI_CHAR device_name[40];   

   sprintf (device_name, "/%s/%d", AVIO_DEV_IO_NAME, deviceIndex);
   fd = open(device_name, O_RDWR, 0);
   if(fd == -1) {
     status = ARS_FAILURE;
   } else { /* acquire the necessary device attributes for the API */
     DEVMAP_T devmap;
     ioctl(fd, PCI_GET_DEV_MAP, &devmap);
     deviceAttributes[deviceIndex].manufID  = (CEI_UINT32)devmap.VendorID;
     deviceAttributes[deviceIndex].deviceID = (CEI_UINT32)devmap.DeviceID;
     deviceAttributes[deviceIndex].irq = 0;
     device->baseAddr  = (pCEI_UINT32)devmap.memHostBase[0];
     device->devicePtr = (pCEI_UINT32)devmap.memHostBase[2];
     devicePciAttributes[deviceIndex].manufID = (CEI_UINT32)devmap.VendorID;
     devicePciAttributes[deviceIndex].deviceID = (CEI_UINT32)devmap.DeviceID;
     devicePciAttributes[deviceIndex].irq = 0;
     devicePciAttributes[deviceIndex].membase[0] = (CEI_NATIVE_ULONG)devmap.memStartPhy[0];
     devicePciAttributes[deviceIndex].membase[2] = (CEI_NATIVE_ULONG)devmap.memStartPhy[2];
     deviceAttributes[deviceIndex].fd = fd;
     status = ARS_NORMAL;
   }

#else   
   CEI_CHAR   device_name[40];
   UINT manufId;
   UINT deviceId;
   UINT irq;
   UINT bar2addr, localDeviceIndex, bar0addr;
   CEI_NATIVE_INT fd;

   localDeviceIndex = (CEI_UINT32)deviceIndex; 
   sprintf (device_name, "/%s/%d", AVIO_DEV_IO_NAME, (CEI_NATIVE_INT)localDeviceIndex);
   fd = open (device_name, O_RDWR, 0);
   if (fd == -1) {
      status = ARS_FAILURE;
   }
   else { /* acquire the necessary device attributes for the API */
      status = (CEI_INT32)VxW_GetCommercialInfo ((UINT)deviceIndex, &bar0addr, &bar2addr, &manufId, &deviceId, &irq);

      deviceAttributes[deviceIndex].manufID  = (CEI_UINT32)manufId;  
      deviceAttributes[deviceIndex].deviceID = (CEI_UINT32)deviceId;  
      deviceAttributes[deviceIndex].irq   = (CEI_UINT32)irq;

      device->baseAddr  = (pCEI_INT32)bar0addr;
      device->devicePtr = (pCEI_INT32)bar2addr;
      
      devicePciAttributes[deviceIndex].manufID = (CEI_UINT32)manufId;         
      devicePciAttributes[deviceIndex].deviceID = (CEI_UINT32)deviceId;
      devicePciAttributes[deviceIndex].irq = (CEI_UINT32)irq;
      devicePciAttributes[deviceIndex].membase[0] = (CEI_NATIVE_ULONG)bar0addr;
      devicePciAttributes[deviceIndex].membase[2] = (CEI_NATIVE_ULONG)bar2addr;
      status = ARS_NORMAL;
   }
   close (fd);
#endif
   /* create thread protection on a first-in/first-out basis */
   ceiAccessMutex[deviceIndex] = semMCreate(SEM_Q_FIFO);

   return status;
}


/*===========================================================================*
 * ENTRY POINT:             R E L E A S E   D E V I C E
 *===========================================================================*
 *
 * FUNCTION:    Close a session with a device
 *              
 * PARAMETERS:  CDEV_BOARD_TYPE deviceIndex  -- (input)  device to reference (0 to 15, 127 for Windows)
 *              ARINC_LOCAL_TYPE * device    -- (output) reference to ARINC_LOCAL_TYPE struct
 * 
 * RETURN VAL:  fixed - ARS_NORMAL
 *
 * GLOBALS:     ceiAccessMutex[8][3]
 *
 * DESCRIPTION: This routine releases a driver handle and unmaps the two PCI 
 *              memory regions previously mapped for the specific device.
 *
 *===========================================================================*/
CEI_INT32 cei_release_device (CDEV_BOARD_TYPE deviceIndex, ARINC_LOCAL_TYPE * device)
{
#ifdef VXW_DRIVER_OPTION
   close(deviceAttributes[deviceIndex].fd);
#elif defined(VXW_VXB_DRIVER)
   struct vxbDriverControl ctrl;
   ctrl.driverName = AVIO_DRV_NAME;
   ctrl.drvCtrlCmd = VXB_GET_DEV_MAP;
   ctrl.drvCtrlInfo = AvVxBusInfo;
   CEI_INT32 status = ARS_NORMAL;
   
   AvVxBusInfo->dev_use_count[deviceIndex] -= 1;
   if (AvVxBusInfo->dev_use_count[deviceIndex] == 0) {
      free (AvVxBusInfo);
      AvVxBusInfo = NULL;
   }
#endif
   /* release multi-thread protection */
   if (ceiAccessMutex[deviceIndex] != NULL) {
      semDelete( ceiAccessMutex[deviceIndex] );
   }
   deviceAttributes[deviceIndex].manufID = 0;
   deviceAttributes[deviceIndex].deviceID = 0;
   return ARS_NORMAL;
}


/*===================================================================================================================*
 * ENTRY POINT:                    C E I _ A C Q U I R E _ D E V I C E _ I D E N T I T Y
 *===================================================================================================================*
 *
 * FUNCTION:    Open a session with a device
 *              
 * PARAMETERS:  CEI_INT32  deviceIndex -- (input)  device to reference (0 to 15, 127 for Windows)
 *              pCEI_INT32 vendorId    -- (output) ptr to storage of vendor PCI ID
 *              pCEI_INT32 deviceId    -- (output) ptr to storage of device PCI ID
 * 
 * RETURN VAL:  ARS_NORMAL     -- Successful acquisition of the device ID's
 *              ARS_FAILURE    -- Under Visa, LabVIEW R/T, VxWorks and Linux, the driver session is not open
 *                                Under Windows, the CEI_Install DLL cannot be accessed
 *              ARS_INVBOARD   -- Under Windows only, device configuration space is inaccessible (session not opened)
 *
 * DESCRIPTION: This routine reports the specified device PCI Configuration Space Vendor ID and Device ID values 
 *              either read directly from PCI Configuration Space or acquired and saved during initialization.
 *
 *===================================================================================================================*/
CEI_INT32 cei_acquire_device_identity (
   CEI_INT32  deviceIndex,                                                        /* device number to test (0-7)     */
   pCEI_INT32 vendorId,                                                           /* ptr to storage of vendor PCI ID */
   pCEI_INT32 deviceId                                                            /* ptr to storage of device PCI ID */
)
{
   CEI_INT32 status = ARS_NORMAL;                                                 /* local function return status    */
   
   if (deviceAttributes[deviceIndex].manufID != 0) {
      *vendorId = (CEI_INT32)deviceAttributes[deviceIndex].manufID;
      *deviceId = (CEI_INT32)deviceAttributes[deviceIndex].deviceID;
   }
   else {
      status = ARS_FAILURE;
   }
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                           C E I _ R E A D _ 3 2 _ B I T
 *===================================================================================================================*
 *
 * FUNCTION:    Perform a 32-bit read from the specified device
 *
 * PARAMETERS:  CDEV_BOARD_TYPE deviceIndex  -- (input) device to reference (0 to 15, 127 for Windows)
 *              pCEI_UINT32 address          -- (input) location in host virtual memory to read
 *              pCEI_UINT32 data             -- (input) pointer in which to store the data value read
 * 
 * RETURN VAL:  For Visa and LabVIEW R/T: -- Visa 'viIn32' invocation return status, if not VI_SUCCESS
 *              ARS_NORMAL                -- Success
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine performs a 32-bit read access from a device PCI memory region
 *
 *===================================================================================================================*/
CEI_INT32 cei_read_32bit (CDEV_BOARD_TYPE board, pCEI_UINT32 address, pCEI_UINT32 data)
{
#ifdef NON_INTEL_WORD_ORDER
   CEI_UINT32 localData, modifiedData;
   localData = *address;
   modifiedData = little_endian_32(localData);
   *data = modifiedData;
#else
   *data = *address;
#endif          
   return ARS_NORMAL;
}

/*===================================================================================================================*
 * ENTRY POINT:                           C E I _ W R I T E _ 3 2 _ B I T
 *===================================================================================================================*
 *
 * FUNCTION:    Perform a 32-bit write to the specified device
 *
 * PARAMETERS:  CDEV_BOARD_TYPE deviceIndex  -- (input) device to reference (0 to 15, 127 for Windows)
 *              pCEI_UINT32 address          -- (input) location in host virtual memory to write
 *              CEI_UINT32 data              -- (input) the data value to write
 * 
 * RETURN VAL:  For Visa and LabVIEW R/T: -- Visa 'viOut32' invocation return status, if not VI_SUCCESS
 *              ARS_NORMAL                -- Success
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine performs a 32-bit write access to a device PCI memory region
 *
 *===================================================================================================================*/
CEI_INT32 cei_write_32bit (CDEV_BOARD_TYPE board, pCEI_UINT32 address, CEI_UINT32 data)
{
#ifdef NON_INTEL_WORD_ORDER
   *address = little_endian_32(data);
#else
   *address = data;
#endif          
   return ARS_NORMAL;
}


/*****************************************************************************
 *  L O C A L   I N T E R R U P T   P R O C E S S I N G   R O U T I N E S
 *****************************************************************************/

#if defined(VXW_VXB_DRIVER) || defined(VXW_DRIVER_OPTION)
/*===========================================================================*
 * ENTRY POINT:   i n t e r r u p t P r o c e s s T a s k V x b
 *===========================================================================*
 *
 * DESCRIPTION: Spawned as a task from within "cei_interrupt_setup", this
 *              routine/task will block on the respective semaphore assigned 
 *              to this interrupt/device, then invoke the application/API
 *              assigned interrupt processing function when flagged to do so.
 *
 *===========================================================================*/
CEI_VOID interruptProcessTaskVxb ( CEI_INT32 deviceIndex,
                                   CEI_INT32 b,
                                   CEI_INT32 c,
                                   CEI_INT32 d,
                                   CEI_INT32 e,
                                   CEI_INT32 f,
                                   CEI_INT32 g,
                                   CEI_INT32 h,
                                   CEI_INT32 j,
                                   CEI_INT32 k )
{
   unsigned int cardNumber = 1;
   CEI_CHAR     sname[16];
   ssize_t      numbytes;
   STATUS       closeStatus;

   while (cardNumber < TERMINATE_INTERRUPT_HANDLER) {
      sprintf (sname, "/%s%d", AVIO_COM_MQ_NAME, deviceIndex);
      mq_id[deviceIndex] = msgQOpen (sname, 100, 4, MSG_Q_FIFO, 0, NULL);
      if (mq_id[deviceIndex] == NULL) {
         return;
      }
      while (cardNumber < TERMINATE_INTERRUPT_HANDLER) {
         numbytes = msgQReceive (mq_id[deviceIndex], (char *)&cardNumber, sizeof(cardNumber), WAIT_FOREVER); /* wait for event to occur */
         if ((cardNumber < TERMINATE_INTERRUPT_HANDLER) && (numbytes == sizeof(cardNumber))) {
            (x30InterruptFunctionPtr[deviceIndex])(deviceIndex, 0);
            interruptCounter[deviceIndex]++;
         }
      }
   }
   closeStatus = msgQClose (mq_id[deviceIndex]); /* close the message queue */
   mq_id[deviceIndex] = 0;
   if (closeStatus) ;
}


/*===========================================================================*
 * ENTRY POINT:     C E I _ V X B _ I N T E R R U P T _ S E T U P
 *===========================================================================*
 *
 * FUNCTION:    Setup for VxBus interrupt processing.
 *
 * PARAMETERS:  deviceIndex -- (input) device to reference (0 to 15)
 * 
 * RETURN VAL:  ARS_NORMAL   -- successful interrupt setup
 *              ARS_INVBOARD -- invalid device index
 *              ARS_FAILURE  -- VxWorks interrupt connect failure
 *                              VxWorks semaphore create failure
 *                              VxWorks task create failure
 *
 * GLOBALS:     
 *              interruptCounter
 *              interruptProcessTaskId
 *
 * DESCRIPTION: Setup the hardware interrupt service routine thread - task
 *              activated via semaphore.
 *
 *===========================================================================*/
CEI_INT32 cei_vxb_interrupt_setup (CEI_UINT32 deviceIndex)
{
   CEI_INT32 status = ARS_NORMAL;                  /* Function return status */
   CEI_CHAR  intProcessName[16];
   
   if (deviceIndex > MAX_BOARDS) {
      status = ARS_INVBOARD;
   }
   else {
      interruptCounter[deviceIndex] = 0;
   
      sprintf (intProcessName,"t_cei_isr_%d",(int)deviceIndex);
      /* with the semaphore created, now create the interrupt processing task */      
      interruptProcessTaskId[deviceIndex] = taskSpawn ( intProcessName,
                                                        16 + deviceIndex,
                                                        0, 
                                                        20000,
                                                        (FUNCPTR)interruptProcessTaskVxb, 
                                                        deviceIndex, 
                                                        0,0,0,0,0,0,0,0,0 );
   
      if (interruptProcessTaskId[deviceIndex] == NULL) {
         /* if an error occurred, remove anything created */
         taskDelete (interruptProcessTaskId[deviceIndex]);
         interruptProcessTaskId[deviceIndex] = 0;
         status = ARS_FAILURE;
      }
   }
   return status;
}

#else

/*===========================================================================*
 * ENTRY POINT:   i n t e r r u p t P r o c e s s T a s k
 *===========================================================================*
 *
 * DESCRIPTION: Spawned as a task from within "cei_interrupt_setup", this
 *              routine/task will block on the respective semaphore assigned 
 *              to this interrupt/device, then invoke the application/API
 *              assigned interrupt processing function when flagged to do so.
 *
 * GLOBALS:     
 *              interruptCounter
 *              mq_id
 *              x30InterruptFunctionPtr
 *
 *===========================================================================*/
CEI_VOID interruptProcessTask ( CEI_INT32 deviceIndex,
                                CEI_INT32 b,
                                CEI_INT32 c,
                                CEI_INT32 d,
                                CEI_INT32 e,
                                CEI_INT32 f,
                                CEI_INT32 g,
                                CEI_INT32 h,
                                CEI_INT32 j,
                                CEI_INT32 k )
{
   unsigned int cardNumber = 1;
   char         sname[16];
   int          numbytes;
   STATUS       closeStatus;

   while (cardNumber < TERMINATE_INTERRUPT_HANDLER) {
      sprintf (sname, "/msg_q_%d", deviceIndex);
      mq_id[deviceIndex] = msgQOpen (sname, 100, 4, MSG_Q_FIFO, 0, NULL);
      if (mq_id[deviceIndex] == NULL) {
         return;
      }
      while (cardNumber < TERMINATE_INTERRUPT_HANDLER) {
         numbytes = msgQReceive (mq_id[deviceIndex], (char *)&cardNumber, sizeof(cardNumber), WAIT_FOREVER); /* wait for event to occur */
         if ((cardNumber < TERMINATE_INTERRUPT_HANDLER) && (numbytes == sizeof(cardNumber))) {
            (x30InterruptFunctionPtr[deviceIndex])(deviceIndex, 0);
            interruptCounter[deviceIndex]++;
         }
      }
   }
   closeStatus = msgQClose (mq_id[deviceIndex]); /* close the message queue */
   mq_id[deviceIndex] = NULL;
}


/*===========================================================================*
 * ENTRY POINT:     C E I _ I N T E R R U P T _ S E T U P
 *===========================================================================*
 *
 * FUNCTION:    Setup for interrupt processing.
 *
 * PARAMETERS:  deviceIndex -- (input) device to reference (0 to 15)
 * 
 * RETURN VAL:  ARS_NORMAL   -- successful interrupt setup
 *              ARS_INVBOARD -- invalid device index
 *              ARS_FAILURE  -- VxWorks interrupt connect failure
 *                              VxWorks semaphore create failure
 *                              VxWorks task create failure
 *
 * GLOBALS:     
 *              interruptCounter
 *              interruptProcessTaskId
 *
 * DESCRIPTION: Setup the hardware interrupt service routine thread - task
 *              activated via semaphore.
 *
 *===========================================================================*/
CEI_INT32 cei_interrupt_setup (CEI_UINT32 deviceIndex)
{
   CEI_INT32 status = ARS_NORMAL;                  /* Function return status */
   CEI_CHAR  intProcessName[16];

   if (deviceIndex > MAX_BOARDS) {
      status = ARS_INVBOARD;
   }
   else {
      interruptCounter[deviceIndex] = 0;
   
      sprintf (intProcessName,"t_cei_isr_%d",(int)deviceIndex);
      /* with the semaphore created, now create the interrupt processing task */      
      interruptProcessTaskId[deviceIndex] = taskSpawn ( intProcessName,
                                                        16 + deviceIndex,
                                                        0, 
                                                        20000,
                                                        (FUNCPTR)interruptProcessTask, 
                                                        deviceIndex, 
                                                        0,0,0,0,0,0,0,0,0 );
   
      if (interruptProcessTaskId[deviceIndex] == ERROR) {
         taskDelete (interruptProcessTaskId[deviceIndex]);
         interruptProcessTaskId[deviceIndex] = 0;
         status = ARS_FAILURE;
      }
   }
   
   return status;
}

#endif /* ifdef VXW_VXB_DRIVER */

/*===================================================================================================================*
 * ENTRY POINT:                         C E I _ I N I T _ I N T E R R U P T
 *===================================================================================================================*
 *
 * FUNCTION:    Configures the device PCI interrupt enable and associates the supplied ISR function with it.
 *              
 * PARAMETERS:  CEI_INT32   deviceIndex       -- (input)  device to reference (0 to 127)
 *              CEI_INT32   _stdcall isr      -- (input)  ptr to the interrupt service routine
 *              CEI_INT32   priority          -- (input)  desired priority of the ISR thread
 *              pCEI_UINT32 deviceBaseAddress -- (input)  virtual address of BAR0/BAR2 region
 *              CEI_INT32   memRegion         -- (input)  PCI BAR region/index for interrupt clear bit (0 - BAR0 or 1 - BAR2)
 *              CEI_UINT32  disOffset         -- (input)  PCI Interrupt Disable register (PLX reg offset, 0 for no PLX interface)
 *              CEI_UINT32  interruptBit      -- (input)  Local Interrupt Input Enable bit for PLX interface control register
 *              CEI_UINT32  comboChannelCount -- (input)  1553/ARINC Combo Board configuration
 *              pCEI_UINT32 isrEnabled        -- (output) Thread successfully defined state (TRUE, FALSE)
 * 
 * RETURN VAL:  ARS_NORMAL         - Interrupt was initialized successfully
 *              ARS_DRIVERFAIL     - Device driver session never opened
 *              ARS_INT_ISR        - Interrupt was not initialized successfully
 *                                 - Driver interrupt event creation failure
 *                                 - Windows thread create failure
 *              ARS_NO_INT_SUPPORT - Driver interrupt handling not supported for this board
 *              ARS_NO_INT_ENABLED - Driver Interrupt handling not enabled for this board
 *
 * GLOBALS:     None
 *
 * DESCRIPTION: This routine enables the PCI interrupt, then connects the user-specified interrupt function   
 *              to the interrupt assigned to the respective device by the host o/s.
 *
 *===================================================================================================================*/
CEI_INT32 cei_init_interrupt 
( 
   CEI_INT32   deviceIndex,                /* (input)  device to reference (0 to 15, for Windows 127)                */
#if defined (_WIN32)    
   CEI_INT32   (_stdcall *isr)(CEI_INT32), /* (input)  ptr to the interrupt service routine                          */
#else
   pCEI_VOID   isr,                        /* (input)  ptr to the interrupt service routine                          */
#endif
   CEI_INT32   priority,                   /* (input)  desired priority of the ISR thread                            */         
   pCEI_UINT32 deviceBaseAddress,          /* (input)  virtual address of BAR0/BAR2 region                           */         
   CEI_INT32   memRegion,                  /* (input)  PCI BAR region/index for interrupt clear bit                  */
   CEI_UINT32  disOffset,                  /* (input)  PCI Interrupt Disable register                                */
   CEI_UINT32  interruptBit,               /* (input)  Local Interrupt Input Enable bit for PLX control register     */
   CEI_UINT32  comboChanCount,             /* (input)  1553/ARINC Combo Board configuration                          */         
   pCEI_UINT32 isrEnabled                  /* (output) thread handle if successfully defined                         */         
)
{
   CEI_UINT32 intCtrlStatReg;             /* PLX Int Control/Status Reg data */
   CEI_INT32 status = ARS_NORMAL;         /* local function return status    */
   /* local reference to the low-level global virtual base address array     */
   pCEI_UINT32 virtualBaseAddressPtr = deviceBaseAddress;

#if defined(VXW_VXB_DRIVER) || defined(VXW_DRIVER_OPTION)
   status = cei_vxb_interrupt_setup ((CEI_UINT32)deviceIndex);
#else
   /* setup the non-vxbus interrupt task and semaphore, then connect to the h/w int */
   status = cei_interrupt_setup ((CEI_UINT32)deviceIndex);
   *isrEnabled = 1;
#endif

   /* Set either the PLX PCI9056 or PCI9030 Interrupt Control/Status Register
    * Local Interrupt Input Enable, High to pass local bus interrupts
    */
   if ((status == ARS_NORMAL) && (interruptBit != 0)) {
      status = cei_read_32bit ((CDEV_BOARD_TYPE)deviceIndex, virtualBaseAddressPtr, &intCtrlStatReg);
      intCtrlStatReg |= interruptBit;
      status = cei_write_32bit ((CDEV_BOARD_TYPE)deviceIndex, virtualBaseAddressPtr, intCtrlStatReg);
   }

   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ T E R M I N A T E _ I N T E R R U P T
 *===================================================================================================================*
 *
 * FUNCTION:    Provides the method to terminate an interrupt association to an ISR and clear the interrupt enable.
 * 
 * PARAMETERS:  CEI_INT32   deviceIndex       -- (input)  device to reference (0 to 127)
 *              CEI_UINT32  interruptBit      -- (input)  Local Interrupt Input Enable bit for PLX interface control register
 *              pCEI_UINT32 deviceBaseAddress -- (input)  virtual address of BAR2 region
 * 
 * RETURN VAL:  ARS_NORMAL   -- Interrupt was terminated successfully
 *              ARS_FAILURE  -- Interrupt was not terminated successfully
 *
 * GLOBALS:     
 *              interruptProcessTaskId
 *              mq_id
 * 
 * DESCRIPTION: This routine terminates the specified device interrupt function and clears the PCI interrupt enable.
 * 
 *===================================================================================================================*/
CEI_INT32 cei_terminate_interrupt
(
   CEI_INT32   deviceIndex,                /* (input)  device to reference (0 to 15, for Windows 127)                */
   CEI_UINT32  interruptBit,               /* (input)  Local Interrupt Input Enable bit for PLX control register     */
   pCEI_UINT32 deviceBaseAddress           /* (input)  virtual address of BAR0/BAR2 region                           */         
)
{
   CEI_INT32 status = ARS_NORMAL;         /* local function return status    */
   CEI_UINT32 intCtrlStatReg;             /* PLX Int Control/Status Reg data */
   
   /* For PCI boards, reset either the PLX PCI9056 or PCI9030 Interrupt Control/Status Register
    * Local Interrupt Input Enable, Low to disable local bus interrupts
    */
   if (interruptBit != 0) {
      status = cei_read_32bit ((CDEV_BOARD_TYPE)deviceIndex, deviceBaseAddress, &intCtrlStatReg);
      intCtrlStatReg &= ~interruptBit;
      status = cei_write_32bit ((CDEV_BOARD_TYPE)deviceIndex, deviceBaseAddress, intCtrlStatReg);
   }
   
#if defined(VXW_VXB_DRIVER) || defined(VXW_DRIVER_OPTION)
   /* if the interrupt task was created and the driver message queue utilized, send a message
    * that indicates to the task to terminate and reset the mq_id as an indication on its way out
    */   
   if (mq_id[deviceIndex] != 0) {
      status = TERMINATE_INTERRUPT_HANDLER;
      msgQSend (mq_id[deviceIndex], (char *)&status, sizeof(unsigned), NO_WAIT, MSG_PRI_URGENT);
      while (mq_id[deviceIndex] > 0) {
         cei_delay_milliseconds(1);
      }
   }
   /* now that the task is not blocking on the message queue, delete the interrupt processing task */  
   if (interruptProcessTaskId[deviceIndex] != 0) {
      taskDelete ((TASK_ID)interruptProcessTaskId[deviceIndex]);
      interruptProcessTaskId[deviceIndex] = 0;
   }
#else
   /* Delete the interrupt processing task */  
   if (interruptProcessTaskId[deviceIndex] != 0) {
      taskDelete (interruptProcessTaskId[deviceIndex]);
      interruptProcessTaskId[deviceIndex] = 0;
	}
#endif

   return status;
}
