/*====================================================================================================================*
 * FILE:                                     C D E V _ A P I _ E X P _ R X. C                    
 *====================================================================================================================*
 *
 *   COPYRIGHT 2016 - 2019 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED
 *   ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH THE INCLUSION
 *   OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY OTHER COPIES THEREOF
 *   MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE
 *   AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS, INC. 
 *
 *====================================================================================================================*
 *
 * FUNCTION:    The Expanded Receive Message portion of the Application Programmer's Interface for a CEI-x30 device.
 *
 * DESCRIPTION: These routines are provided to demonstrate optional message reception-specific programming primitives 
 *              for the CEI-x30 API using the C programming language.  Each routine implements a particular API  
 *              function; however, a user application program may need to combine functions in different ways to 
 *              perform the desired operation.
 *
 *              Most routines utilize a board number parameter, zero-based on the Device ID/Board Number assigned 
 *              during the installation of CEI-x30-SW distribution.  Channel parameter values are all zero-based.
 *
 * EXTERNALLY DEFINED ROUTINE ENTRY POINTS:         
 *
 *    CEI-x30 ARINC 429 Message Reception API routines:
 *
 *    ar_getblock           - read a block of messages from a receiver FIFO
 *    ar_get_data           - retrieves data and time-tag from a receive FIFO
 *    ar_get_data_xt        - retrieves data and time-tag from a receive FIFO
 *    ar_get_error          - returns error string for a status code
 *    ar_get_429_message    - retrieves an ARINC 429 message
 *    ar_get_latest         - retrieves latest data from the snapshot buffer
 *    ar_getnext            - wait for next data word from a receiver
 *    ar_getnextt           - wait for next data, return with 32-bit timetag
 *    ar_getnext_xt         - wait for next data, return with 64-bit timetag
 *    ar_getwordt           - read data and 32-bit timetag from a receiver FIFO
 *
 * INTERNAL UTILITY ROUTINES:
 *
 *    a429_utl_get_time_msec
 *
 *--------------------------------------------------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date    By   Ver. Comments
 * -------- ----- ---- -----------------------------------------------------------------------------------------------
 * 10/18/16  BDW  4.90 Initial version.
 *
 *====================================================================================================================*/ 
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
   #include <unistd.h>
   #include <fcntl.h>
   #include <string.h>
   #include <pthread.h>
   #include <errno.h>
   #include <sys/stat.h> 
   #include <sys/shm.h>
   #include <sys/sem.h>
   #include <sys/syscall.h>
   #define VOLATILE volatile
#elif defined (VXW_PCI_X86) || defined (VXW_PCI_PPC)
   #include <stdio.h>
   #include <stdlib.h>
   #include <unistd.h>
   #define VOLATILE 
#elif defined (NIVISA_PXI)
   #ifdef LABVIEW_RT
      #include <ansi_c.h> 
   #endif
   #include <visa.h>  
   #define VOLATILE 
#elif defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
   #include <string.h>
   #define VOLATILE 
#else
   #define VOLATILE volatile
#endif

#include "cdev_glb.h"

extern VOLATILE ARINC_LOCAL_TYPE device[MAX_BOARDS];
extern ARINC_GLOBAL_TYPE * deviceGlobal[MAX_BOARDS];


/*===================================================================================================================*
 * ENTRY POINT:                      A 4 2 9 _ U T L _ G E T _ T I M E _ M S E C
 *===================================================================================================================*
 *
 * FUNCTION:    This routine returns the value of the device timer in a 32-bit format with a resolution 
 *              of 1 millisecond.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board       -- (input) Board number (0 to 127) 
 *
 * RETURN VAL:  CEI_INT32    -- ARS_NORMAL      Success
 *                              ARS_INVBOARD    Invalid or uninitialized board
 *                              ARS_INVARG      Invalid timeSource or timeTagFormat
 *
 *===================================================================================================================*/ 
CEI_UINT32 a429_utl_get_time_msec
(
   CDEV_BOARD_TYPE board                                                 /* (input)  Board number (0 to 127)         */
)
{
   AR_TIMETAG_TYPE timeTag;                                              /* temp struct storage for current time     */
   CEI_INT32 status = ARS_NORMAL;                                        /* operation status value                   */

   status = a429_utl_acquire_formatted_time_tag ( board,
                                                  0U,
                                                  AR_TIMER_REGISTER_SET, 
                                                  AR_TIMETAG_INT_MSEC_32BIT, 
                                                  AR_OFF, 
                                                  &timeTag );
      
   return ( (CEI_UINT32)timeTag.timeTag );
}


/*===================================================================================================================*
 * ENTRY POINT:                                     A R _ G E T N E X T
 *===================================================================================================================*
 *
 * FUNCTION:    Read a message from the specified receive buffer.  This routine will wait up to a maximum of 1/2 
 *              second for a message to arrive before timing out.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board               -- (input)  board to reference
 *              CDEV_CHAN_TYPE channel              -- (input)  recvr channel to use (0 to 31)
 *              CDEV_PARM_VOID_PTR_TYPE destination -- (output) location to receive data
 *
 * RETURN VAL   CDEV_API_RET_TYPE ARS_GOTDATA           Success
 *                                ARS_INVHARVAL         Unsupported channel #
 *                                ARS_INVARG            A NULL data parameter value was detected
 *                                ARS_BOARD_MUTEX       Access to the Board Lock timed-out/failed
 *                                ARS_CHAN_TIMEOUT      No data available
 *                                ARS_BAD_MESSAGE       An invalid length msg was received
 *                                ARS_RX_BUFFER_OVERRUN A receive buffer overrun was detected
 *                                ARS_INVBOARD          Invalid board parameter or board not initialized
 * GLOBALS:     none
 * 
 * DESCRIPTION: This routine reads the most recently received ARINC 429 message.  The channel number passed to this 
 *              routine corresponds to the ARINC 429 channel number index, starting with zero.
 * 
 *              This routine polls the receiver continuously, incorporating a 1 millisecond sleep/task delay until:
 *                 1) An ARINC 429 message is received or
 *                 2) The timeout threshold of 1/2 second is reached.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_getnext 
( 
   CDEV_BOARD_TYPE board,                                                /* (input)  board to reference (0 to 127)   */
   CDEV_CHAN_TYPE channel,                                               /* (input)  recv channel to ref (0 to 31)   */
   CDEV_PARM_VOID_PTR_TYPE destination                                   /* (output) location to store received data */
)
{
   CDEV_API_RET_TYPE status;                                                 /* function return status               */
   CEI_UINT32 goal;                                                          /* local time-out target value          */
   CEI_INT32 passes = 50;                                                    /* 50 10ms retries to receive data      */
   CDEV_CHAN_TYPE localChannel;                                       /* channel reference for a429_utl_read_message */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      /* channel is 0-referenced ARINC 429 receive channel index */
      if (channel < (CDEV_CHAN_TYPE)deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) {
#endif
         localChannel = channel + (CDEV_CHAN_TYPE)(deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]);
         goal = a429_utl_get_time_msec(board) + ((CEI_INT32)TIMER_CLOCK_TICKS_PER_SECOND / 100);

         /* Wait for a message to arrive in the buffer, or exit if timeout. */
         while ((status = a429_utl_read_message (board, &localChannel, (pCEI_UINT32)destination, NULL)) == ARS_NODATA) {
            /* Initialize variable for time to wait until if this is first pass. */
            if ( (passes > 0) && (a429_utl_get_time_msec(board) > goal) ) {
               passes--;
               goal = a429_utl_get_time_msec(board) + ((CEI_INT32)TIMER_CLOCK_TICKS_PER_SECOND / 100);
            }
      
            /* If more than 1/2 second of processing time has passed, exit. */
            else if ( (passes == 0) && (a429_utl_get_time_msec(board) > goal) ) {
               status = ARS_CHAN_TIMEOUT;
               break;
            }
            else { /* release this thread from the o/s for 1 millisecond */
               ar_sleep(1);
            }
         }
         
         /* if we timed-out, make one last attempt to get some data */
         if (status == ARS_CHAN_TIMEOUT) {
            status = a429_utl_read_message ( board, &localChannel, (pCEI_UINT32)destination, NULL );
            if (status == ARS_NODATA) {
               status = ARS_CHAN_TIMEOUT;
            }
         }
#ifndef TARGET_EMBEDDED
      }
      else {
         status = ARS_INVHARVAL;
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif
   return status;
}
                                                 

/*===================================================================================================================*
 * ENTRY POINT:                                  A R _ G E T N E X T T
 *===================================================================================================================*
 *
 * FUNCTION:    Read a message and time stamp from the specified receive buffer.  This routine will wait up to a  
 *              maximum of 1/2 second for a message to arrive before timing out.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board               -- (input)  board to reference
 *              CDEV_CHAN_TYPE channel              -- (input)  recvr channel to use (0 to 31)
 *              CDEV_PARM_VOID_PTR_TYPE destination -- (output) location to store the received message
 *              CDEV_PARM_VOID_PTR_TYPE timeTag     -- (output) location to store the received message time stamp
 *
 * RETURN VAL   CDEV_API_RET_TYPE ARS_GOTDATA           Success
 *                                ARS_CHAN_TIMEOUT      No data available
 *                                ARS_INVHARVAL         Unsupported channel parameter
 *                                ARS_INVARG            A NULL data parameter value was detected
 *                                ARS_INVBOARD          Invalid board parameter
 *                                ARS_BOARD_MUTEX       Access to the Board Lock timed-out/failed
 *                                ARS_BAD_MESSAGE       An invalid length msg was received
 *                                ARS_RX_BUFFER_OVERRUN A receive buffer overrun was detected
 *
 * GLOBALS:     deviceGlobal[]
 * 
 * DESCRIPTION: This routine reads the most recently received ARINC 429 message and time stamp.  The channel number  
 *              passed to this routine corresponds to the ARINC 429 channel number index, starting with zero.
 * 
 *              This routine polls the receiver continuously, incorporating a 1 millisecond sleep/task delay until:
 *                 1) An ARINC 429 message is received or
 *                 2) The timeout threshold of 1/2 second is reached.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_getnextt 
( 
   CDEV_BOARD_TYPE board,                                                /* (input)  board to reference (0 to 127)   */
   CDEV_CHAN_TYPE channel,                                               /* (input)  recv channel to ref (0 to 31)   */
   CDEV_PARM_VOID_PTR_TYPE destination,                                  /* (output) location to store received data */
   CDEV_PARM_VOID_PTR_TYPE timeTag                                       /* (output) location to store data time-tag */
)
{
   CDEV_API_RET_TYPE status;                                                   /* function return status             */
   CEI_UINT32 goal;                                                            /* local time-out target value        */
   CEI_INT32 passes = 50;                                                      /* 50 10ms retries to receive data    */
   AR_TIMETAG_TYPE tempTimeTag;                                                /* local time-tag reference structure */
   CEI_INT32 timeTagReference;                                                 /* local Time-Tag Mode reference      */
   CDEV_CHAN_TYPE localChannel;                                                /* merged/individual channel select   */
   pCEI_UINT32 timeTagPtr = (pCEI_UINT32)timeTag;                              /* local time stamp storage reference */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      /* legacy ar_getword channel is 0-referenced ARINC 429 receive channel index;
       * a429_utl_read_message uses the board-referenced channel number 0-255, where
       * ARINC 429 receive channels can be located anywhere in that range.
       */
      if (channel < (CDEV_CHAN_TYPE)(deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX])) {
#endif
         /* if a 32-bit time-tag ref. isn't already selected, force 32-bit as a default */
         timeTagReference = deviceGlobal[board]->deviceTimeTagReference;
         if ( (timeTagReference != AR_TIMER_X20_COMPAT_32BIT) && (timeTagReference != AR_TIMETAG_INT_20USEC_32BIT) ) {
            timeTagReference = AR_DEFAULT_RECEIVE_DATA_TIMETAG_FORMAT;
         }
         localChannel = channel + (CDEV_CHAN_TYPE)(deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]);
         goal = a429_utl_get_time_msec(board) + ((CEI_INT32)TIMER_CLOCK_TICKS_PER_SECOND / 100);
      
         /* Wait for something, or exit if timeout. */
         tempTimeTag.timeTagFormat = timeTagReference;
      
         while ( (status = a429_utl_read_message (board, &localChannel, (pCEI_UINT32)destination, &tempTimeTag)) == ARS_NODATA ) {
            /* Initialize variable for time to wait until if this is first pass. */
            if ( (passes > 0) && (a429_utl_get_time_msec(board) > goal) ) {
               passes--;
               goal = a429_utl_get_time_msec(board) + ((CEI_INT32)TIMER_CLOCK_TICKS_PER_SECOND / 100);
            }
      
            /* If more than 1/2 second of processing time has passed, exit. */
            else if ( (passes == 0) && (a429_utl_get_time_msec(board) > goal) ) {
               status = ARS_CHAN_TIMEOUT;
               break;
            }
            else { /* release this thread from the o/s for 1 millisecond */
               ar_sleep(1);
            }
         }
         
         /* if we timed-out, make one last attempt to get some data */
         if (status == ARS_CHAN_TIMEOUT) {
            status = a429_utl_read_message (board, &localChannel, (pCEI_UINT32)destination, &tempTimeTag);
            if (status == ARS_NODATA) {
               status = ARS_CHAN_TIMEOUT;
            }
         }
         if (((status == ARS_GOTDATA) || (status == ARS_BAD_MESSAGE)) & (timeTag != NULL)) {
            *timeTagPtr = (CEI_UINT32)(tempTimeTag.timeTag & 0xFFFFFFFF);
         }

#ifndef TARGET_EMBEDDED
      }
      else {
         status = ARS_INVHARVAL;
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif
   return (status);
}


/*===================================================================================================================*
 * ENTRY POINT:                                   A R _ G E T N E X T _ X T
 *===================================================================================================================*
 *
 * FUNCTION:    Read a message and time stamp from the specified receive buffer.  This routine will wait up to a  
 *              maximum of 1/2 second for a message to arrive before timing out.  This routine supports returning an 
 *              extended time-tag structure supporting IRIG.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board    -- (input)  board to reference (0 to 127)
 *
 *              CDEV_CHAN_TYPE channel   -- (input)  recvr channel to use (0 to 31)
 *
 *              pCEI_UINT32 data         -- (output) location to store the received message
 *
 *              pAR_TIMETAG_TYPE timeTagRef.timeTagFormat
 *
 *                                       -- (input) desired source, size and/or resolution for the message time 
 *                                                  stamp, valid selections are:
 *
 *                                                       AR_TIMETAG_EXT_IRIG_64BIT      0
 *                                                       AR_TIMETAG_INT_USEC_64BIT      1
 *                                                       AR_TIMETAG_HOST_USEC_64BIT     2
 *                                                       AR_TIMETAG_INT_20USEC_32BIT    3
 *                                                       AR_TIMETAG_INT_MSEC_32BIT      4
 *                                                       AR_TIMETAG_SYNC_1553_CH1       11
 *                                                       AR_TIMETAG_SYNC_1553_CH2       12
 *                                                       AR_TIMETAG_SYNC_1553_CH3       13
 *                                                       AR_TIMETAG_SYNC_1553_CH4       14
 *                                                       AR_TIMETAG_USE_PROGRAMMED_MODE 99
 *
 *              pAR_TIMETAG_TYPE timeTagRef
 *                                      -- (output) location to store the time-tag structure
 *
 * RETURN VAL   CDEV_API_RET_TYPE ARS_GOTDATA           Success
 *                                ARS_CHAN_TIMEOUT      No data available
 *                                ARS_INVHARVAL         Unsupported channel parameter
 *                                ARS_INVARG            A NULL data parameter value was detected
 *                                ARS_INVBOARD          Invalid board parameter
 *                                ARS_BOARD_MUTEX       Access to the Board Lock timed-out/failed
 *                                ARS_BAD_MESSAGE       An invalid length msg was received
 *                                ARS_RX_BUFFER_OVERRUN A receive buffer overrun was detected
 *
 * GLOBALS:     deviceGlobal[]
 * 
 * DESCRIPTION: This routine reads the most recently received ARINC 429 data.
 *              The channel number passed to this routine corresponds to the
 *              ARINC 429 channel number index, starting with zero.
 * 
 *              This routine polls the receiveer continuously until:
 *                 1) An ARINC word is received or
 *                 2) The timeout threshold (1/2 second) is reached.
 * 
 *              The channel number passed to this routine corresponds to the
 *              ARINC 429 channel number index, starting with zero.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_getnext_xt 
( 
   CDEV_BOARD_TYPE board,                                               /* (input)  board to reference (0 to 127)    */
   CDEV_CHAN_TYPE channel,                                              /* (input)  recv channel to ref (0 to 31)    */
   pCEI_UINT32 data,                                                    /* (output) location to store received data  */
   pAR_TIMETAG_TYPE timeTagRef                                          /* (output) location to store data time-tags */
)
{
   CDEV_CHAN_TYPE localChannel;                                                 /* merged/individual channel select  */
   CDEV_API_RET_TYPE status;                                                    /* function return status            */
   CEI_UINT32 goal;                                                             /* local time-out target value       */
   CEI_INT32 passes = 50;                                                       /* 50 10ms retries to receive data   */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      /* legacy ar_getword channel is 0-referenced ARINC 429 receive channel index;
       * a429_utl_read_message uses the board-referenced channel number 0-255, where
       * ARINC 429 receive channels can be located anywhere in that range.
       */
      if (channel < (CDEV_CHAN_TYPE)(deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX])) {
#endif
         localChannel = channel + (CDEV_CHAN_TYPE)(deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]);
         goal = a429_utl_get_time_msec(board) + ((CEI_INT32)TIMER_CLOCK_TICKS_PER_SECOND / 100);
      
         /*  Wait for something, or exit if timeout. */
         while ( (status = a429_utl_read_message (board, &localChannel, data, timeTagRef)) == ARS_NODATA ) {
      
            /* Initialize variable for time to wait until if this is first pass. */
            if ( (passes > 0) && (a429_utl_get_time_msec(board) > goal) ) {
               passes--;
               goal = a429_utl_get_time_msec(board) + ((CEI_INT32)TIMER_CLOCK_TICKS_PER_SECOND / 100);
            }
            /* If more than 1/2 second of processing time has passed, exit. */
            else if ( (passes == 0) && (a429_utl_get_time_msec(board) > goal) ) {
               status = ARS_CHAN_TIMEOUT;
               break;
            }
         }
         
         /* if we timed-out, make one last attempt to get some data */
         if (status == ARS_CHAN_TIMEOUT) {
            status = a429_utl_read_message (board, &localChannel, data, timeTagRef);
            if (status == ARS_NODATA) {
               status = ARS_CHAN_TIMEOUT;
            }
         }

#ifndef TARGET_EMBEDDED
      }
      else {
         status = ARS_INVHARVAL;
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif
   return (status);
}


/*===================================================================================================================*
 * ENTRY POINT:                                    A R _ G E T W O R D T
 *===================================================================================================================*
 *
 * FUNCTION:    Read an ARINC message and time stamp from a receive buffer.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board               -- (input)  board to reference (0 to 127)
 *              CDEV_CHAN_TYPE channel              -- (input)  recvr channel to use (0 to 31)
 *              CDEV_PARM_VOID_PTR_TYPE destination -- (output) location to receive data
 *              CDEV_PARM_VOID_PTR_TYPE timeTag     -- (output) location to store the time stamp
 *
 * RETURN VAL   CDEV_API_RET_TYPE ARS_GOTDATA           Success
 *                                ARS_NODATA            No data availalble
 *                                ARS_INVHARVAL         Protocol not available on device
 *                                ARS_INVARG            A NULL data parameter value was detected
 *                                ARS_INVBOARD          Invalid board parameter or board not initialized
 *                                ARS_BOARD_MUTEX       Access to the Board Lock timed-out/failed
 *                                ARS_BAD_MESSAGE       An invalid length msg was received
 *                                ARS_RX_BUFFER_OVERRUN A receive buffer overrun was detected
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine reads the most recently received ARINC message and time stamp from the specified 
 *              receive channel buffer.  This routine only supports ARINC 429 channel types. 
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_getwordt 
( 
   CDEV_BOARD_TYPE board,                                               /* (input)  board to reference (0 to 127)    */
   CDEV_CHAN_TYPE channel,                                              /* (input)  recv channel to ref (0 to 31)    */
   CDEV_PARM_VOID_PTR_TYPE destination,                                 /* (output) location to store received msg   */
   CDEV_PARM_VOID_PTR_TYPE timeTag                                      /* (output) location to store msg time stamp */
)
{
   CDEV_CHAN_TYPE localChannel;                                                /* merged/individual channel select   */
   CDEV_API_RET_TYPE status;                                                   /* function return status             */
   AR_TIMETAG_TYPE tempTimeTag;                                                /* local time-tag reference structure */
   CEI_INT32 timeTagReference;                                                 /* local Time-Tag Mode reference      */
   pCEI_UINT32 timeTagPtr = (pCEI_UINT32)timeTag;                              /* local time stamp storage reference */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      /* legacy ar_getwordt channel is 0-referenced ARINC 429 receive channel index;
       * a429_utl_read_message uses the board-referenced channel number 0-255, where
       * ARINC 429 receive channels can be located anywhere in that range.
       */
      if (channel < (CDEV_CHAN_TYPE)(deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX])) {
#endif
         /* if a 32-bit time-tag ref. isn't already selected, force 32-bit as a default */
         timeTagReference = deviceGlobal[board]->deviceTimeTagReference;
         if ( (timeTagReference != AR_TIMER_X20_COMPAT_32BIT) && (timeTagReference != AR_TIMETAG_INT_20USEC_32BIT) ) {
            timeTagReference = AR_DEFAULT_RECEIVE_DATA_TIMETAG_FORMAT;
         }
         localChannel = channel + (CDEV_CHAN_TYPE)(deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]);
         if (timeTag == NULL) {
            status = a429_utl_read_message ( board, &localChannel, (pCEI_UINT32)destination, NULL );
         }
         else {
            tempTimeTag.timeTagFormat = timeTagReference;
            status = a429_utl_read_message ( board, &localChannel, (pCEI_UINT32)destination, &tempTimeTag );
            if ((status == ARS_GOTDATA) || (status == ARS_BAD_MESSAGE)) {
               *timeTagPtr = (CEI_UINT32)(tempTimeTag.timeTag & 0xFFFFFFFF);
            }
         }
         
#ifndef TARGET_EMBEDDED
      }
      else {
         status = ARS_INVHARVAL;
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif
   return (status);
}


/*===================================================================================================================*
 * ENTRY POINT:                             A R _ G E T _ D A T A  
 *===================================================================================================================*
 *
 * FUNCTION:    Read the next message and 64-bit 1 microsecond resolution time stamp from a receive buffer.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input)  board to reference (0 to 127)
 *              pCDEV_CHAN_TYPE channel -- (input)  receive channel to reference (0 to 63)
 *                                         (output) assigned if merged mode is enabled, this value indicates the  
 *                                                  channel on which the message was actually received.
 *              pCEI_UINT32 data        -- (output) location to store the received message
 *              pCEI_UINT32 timeTagLo   -- (output) location to store the time stamp least significant 32-bits
 *              pCEI_UINT32 timeTagHi   -- (output) location to store the time stamp most significant 32-bits
 *
 * RETURN VAL   CDEV_API_RET_TYPE ARS_GOTDATA           Success
 *                                ARS_NODATA            No data availalble
 *                                ARS_INVARG            A NULL data parameter value was detected
 *                                ARS_INVHARVAL         Unsupported channel parameter
 *                                ARS_INVBOARD          Invalid board parameter or board not initialized
 *                                ARS_BOARD_MUTEX       Access to the Board Lock timed-out/failed
 *                                ARS_BAD_MESSAGE       An invalid length msg was received
 *                                ARS_RX_BUFFER_OVERRUN A receive buffer overrun was detected
 *
 * GLOBALS:     none
 * 
 * DESCRIPTION: This routine reads the most recently received entry (message and time stamp) from the specified 
 *              receive channel buffer.  The channel number passed to this routine corresponds to the hardware 
 *              channel assignment on the device:
 *                                                   0-31 ARINC 429 Receive Channel 0 to 31
 *                                                    32  ARINC 717 Receive Channel
 *                                                    63  Merged Receive Buffer
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_get_data 
( 
   CDEV_BOARD_TYPE board,                                         /* (input)  board to reference (0 to 127)          */
#if defined (LABVIEW_RT)                                                                                             
   CDEV_CHAN_TYPE channel,                                        /* (input)  receive channel number to reference    */
#else                                                                                                                
   CDEV_CHAN_TYPE * channel,                                      /* (output) in merged mode, msg rx channel source  */
#endif                                                                                                               
   pCEI_UINT32 data,                                              /* (output) location to store received message     */
   pCEI_UINT32 timeTagLo,                                         /* (output) location to store time-tag low 32 bits */
   pCEI_UINT32 timeTagHi                                          /* (output) location to store time-tag hi 32 bits  */
)
{
#if defined (LABVIEW_RT)
   CDEV_CHAN_TYPE localChannel = channel;
#else   
   CDEV_CHAN_TYPE localChannel = *channel;
#endif
   CDEV_API_RET_TYPE status;
   AR_TIMETAG_TYPE timeTagStruct;
   timeTagStruct.timeTagFormat = deviceGlobal[board]->deviceTimeTagReference;
   status = a429_utl_read_message (board, &localChannel, data, &timeTagStruct);
   /* for any case where a message is returned, parse the 64-bit time-stamp into two 32-bit elements */
   if ((status == ARS_GOTDATA) || (status == ARS_RX_BUFFER_OVERRUN) || (status == ARS_BAD_MESSAGE)) {
      *timeTagLo = (CEI_UINT32)(timeTagStruct.timeTag & 0xFFFFFFFFL);
      *timeTagHi = (CEI_UINT32)(timeTagStruct.timeTag >> 32);
#ifndef LABVIEW_RT
      *channel = localChannel;
#endif
   }
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                            A R _ G E T _ D A T A _ X T
 *===================================================================================================================*
 *
 * FUNCTION:    Read the next message and 64-bit time stamp from a receive buffer, returning an extended time-tag 
 *              structure supporting IRIG.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input)  board to reference (0 to 127)
 *
 *              pCDEV_CHAN_TYPE channel -- (input)  receive channel to reference (0 to 63)
 *                                         (output) assigned if merged mode is enabled, this value indicates the  
 *                                                  channel on which the message was actually received.
 *
 *              pCEI_UINT32 data         -- (output) location to store the received message
 *
 *              pAR_TIMETAG_TYPE timeTagRef.timeTagFormat
 *
 *                                       -- (input) desired source, size and/or resolution for the message time 
 *                                                  stamp, valid selections are:
 *
 *                                                       AR_TIMETAG_EXT_IRIG_64BIT      0
 *                                                       AR_TIMETAG_INT_USEC_64BIT      1
 *                                                       AR_TIMETAG_HOST_USEC_64BIT     2
 *                                                       AR_TIMETAG_INT_20USEC_32BIT    3
 *                                                       AR_TIMETAG_INT_MSEC_32BIT      4
 *                                                       AR_TIMETAG_SYNC_1553_CH1       11
 *                                                       AR_TIMETAG_SYNC_1553_CH2       12
 *                                                       AR_TIMETAG_SYNC_1553_CH3       13
 *                                                       AR_TIMETAG_SYNC_1553_CH4       14
 *                                                       AR_TIMETAG_USE_PROGRAMMED_MODE 99
 *
 *              pAR_TIMETAG_TYPE timeTagRef
 *                                      -- (output) location to store the time-tag structure
 *
 * RETURN VAL   CDEV_API_RET_TYPE ARS_GOTDATA           Success
 *                                ARS_INVHARVAL         Unsupported channel parameter
 *                                ARS_INVARG            A NULL data parameter value was detected
 *                                ARS_BOARD_MUTEX       Access to the Board Lock timed-out/failed
 *                                ARS_NODATA            No data availalble
 *                                ARS_INVBOARD          Invalid board parameter or board not initialized
 *                                ARS_BAD_MESSAGE       An invalid length msg was received
 *                                ARS_RX_BUFFER_OVERRUN A receive buffer overrun was detected
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: This routine reads the most recently received entry (message and time stamp) from the specified 
 *              receive channel buffer.  If the requested time-tag format is External IRIG 64-bit, the time stamp
 *              will be calculated as an extrapolated IRIG time-of-year value based on the original message time
 *              stamp and the IRIG sample time stamp. The channel number passed to this routine corresponds to the  
 *              hardware channel assignment on the device:
 *                                                         0-31 ARINC 429 Receive Channel 0 to 31
 *                                                          32  ARINC 717 Receive Channel
 *                                                          63  Merged Receive Buffer
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_get_data_xt
(
   CDEV_BOARD_TYPE board,                                               /* (input)  board to reference (0 to 127)    */
   CDEV_CHAN_TYPE * channel,                                            /* (input)  recv channel to ref (0 to 63 )   */
   pCEI_UINT32 data,                                                    /* (output) location to store received data  */
   pAR_TIMETAG_TYPE timeTagRef                                          /* (output) location to store data time-tags */
)
{
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                                /* function return status   */
   
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      if (data == NULL) {
         status = ARS_INVARG;
      }
      else {
#endif
         status = a429_utl_read_message (board, channel, data, timeTagRef);

#ifndef TARGET_EMBEDDED
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   return (status);
}


/*===================================================================================================================*
 * ENTRY POINT:                           A R _ G E T B L O C K
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve multiple messages from a specified receive buffer.
 *
 * PARAMETERS:  CEI_UINT32 board       -- (input)  board to reference (0 to 127)
 *              CEI_UINT32 channel     -- (input)  recvr channel to use (0 to 31)
 *              CEI_INT32  maxMessages -- (input)  number of messages to retrieve
 *              CEI_INT32  offset      -- (input)  unused parameter (legacy support)
 *              pCEI_INT32 actualCount -- (output) number of messages retrieved
 *              pCEI_INT32 data        -- (output) location to store data
 *              pCEI_INT32 timeTags    -- (output) location to store timetags
 *
 * RETURN VAL:  ARS_GOTDATA           one or more ARINC messages were retrieved
 *              ARS_NODATA            no ARINC messages were available in the buffer
 *              ARS_INVBOARD          invalid board number
 *              ARS_INVHARVAL         Item was invalid or channel
 *              ARS_INVARG            invalid parameter value
 *              ARS_BOARD_MUTEX       Access to the Board Lock timed-out/failed
 *              ARS_BAD_MESSAGE       An invalid length msg was received
 *              ARS_RX_BUFFER_OVERRUN A receive buffer overrun was detected
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: This routine retrieves all of the available ARINC messages (up to the number specified via 
 *              maxMesages), from the specified receive channel buffer, and copies them to the desired destination.
 *
 *              If the TimeTag pointer is not NULL, the time tags associated with each retrieved message are also 
 *              copied; however, only 32-bit time-tagging options are supported.
 *
 *              If the buffer is empty this routine returns ARS_NODATA.
 *
 *              If the return value for actualCount is equal to maxMessages, additional unread messages may be  
 *              available in the respective receive buffer. 
 *
 *===================================================================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_getblock
(
   CEI_UINT32 board,                                                   /* (input)  board to reference (0 to 127)     */
   CEI_UINT32 channel,                                                 /* (input)  recvr channel to use (0 to 31)    */
   CEI_INT32  maxMessages,                                             /* (input)  number of messages to retrieve    */
   CEI_INT32  offset,                                                  /* (input)  unused parameter (legacy support) */
   pCEI_INT32 actualCount,                                             /* (output) number of messages retrieved      */
   pCEI_INT32 data,                                                    /* (output) location to store data            */
   pCEI_INT32 timeTags                                                 /* (output) location to store timetags        */
)
{
   CEI_UINT32 channelStatus = 0U;                                             /* Rx channel FIFO data avail status   */
   CEI_UINT32 localChannel = 0U;                                              /* merged/individual channel select    */
   pCEI_INT32 dataPtr;                                                        /* local pointer to store Rx data      */
   pCEI_INT32 timeTagPtr;                                                     /* local pointer to store Rx timetag   */
   CEI_INT32 messageTotal;                                                    /* receiver buffer message counter     */
   CEI_INT32 messageCount;                                                    /* receiver buffer message counter     */
   CEI_INT32 mergedModeState =  AR_OFF;                                       /* 429 Rx ch merged/individual state   */
   CEI_UINT32 mergedModeChan = 0;                                             /* 429 Rx ch merged/individual state   */
   CEI_INT32 status = ARS_NORMAL;                                             /* function return status              */
   CEI_INT32 recvStatus = ARS_NORMAL;                                         /* FIFO receive status                 */
#ifndef TARGET_EMBEDDED
   CEI_INT32 lockStatus = ARS_NORMAL;                                         /* status of board access lock acquis. */
#endif
   CEI_INT32 retrieveTimeTag = FALSE;                                         /* indicates to pass time-tags back    */
   CEI_INT32 accessStatus = ARS_NORMAL;                                       /* status of device read/write access  */
   CEI_INT32 timeTagReference = offset;                                       /* currently selected time-tag format  */
   AR_TIMETAG_TYPE timeTagRef;                                                /* local time-tag reference structure  */
   CEI_UINT32 chanOffset;                                                     /* local offset to channel memory area */
   CEI_UINT32 dataOffset;                                                     /* local offset to channel memory area */
   CEI_UINT32 memOffset;                                                      /* local offset to channel memory area */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr = NULL;                   /* pointer to board local structure    */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr = NULL;                          /* pointer to board global structure   */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   else if ((data == NULL) || (actualCount == NULL) || (maxMessages < 0)) {
      status = ARS_INVARG;
   }
   else {
#endif
      /* initialize local return message count and board pointers */
      *actualCount = 0;
      deviceGlobalStructPtr = deviceGlobal[board];
      deviceLocalStructPtr = &device[board];

      /* reference supplied pointers locally */
      timeTagPtr = timeTags;
      timeTagReference = (CEI_INT32)deviceGlobalStructPtr->deviceTimeTagReference;
      
      /* if the channel number passed by the host application exceeds the ARINC 429 channel count, */
      /* assume if available, the ARINC 717 channel is desired; otherwise return failure. */
      if (channel >= deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) {
         if (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX] != CEIDEV_MAX_CHANNELS) {
            localChannel = deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX];
         }
         else {
            status = ARS_INVHARVAL;
         }
      }
      else {
         localChannel = channel + deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
      }
      
      /* Check to see which Receive Mode is active for this channel, Individual or Merged */
      if (status == ARS_NORMAL) {
         if (deviceGlobalStructPtr->receiveChannelBufferMode[localChannel] == ARU_MERGED) {
            localChannel = (CEI_UINT32)deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_MERGED_RX];
            mergedModeState = AR_ON;
         }
         else {
            mergedModeChan = CEIDEV_MAX_CHANNELS;
         
         }
      }

#ifndef TARGET_EMBEDDED
      if (status == ARS_NORMAL) {
         status = a429_utl_validate_channel_request ((CDEV_BOARD_TYPE)board, localChannel, ARU_GET_DATA);
      }
      if ((accessStatus == ARS_NORMAL) && (lockStatus == ARS_NORMAL) && (status == ARS_NORMAL) && (deviceGlobalStructPtr != NULL) && (deviceLocalStructPtr != NULL)) {
         /* if multi-thread protection requested, acquire semaphore/mutex-locking */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            lockStatus = cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (lockStatus == ARS_NORMAL) {
#endif
            /* check the receive channel status and assign return status in the following priority order...
             * If it indicates "BUFFER OVERRUN", save and return a status indicating a buffer overrun was encountered (don't read the buffer)
             * If it indicates "BAD MESSAGE RECEIVED", save and return a status indicating an invalid 429 message was previously detected (read the buffer)
             * If it indicates "NOT EMPTY", then retrieve message(s) and optionally the time-stamp(s) from the buffer. 
             */
            memOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS + (localChannel * CEIDEV_CHANNEL_REGION_SIZE);
            accessStatus = cei_read_32bit ((CDEV_BOARD_TYPE)board, &deviceLocalStructPtr->devicePtr[memOffset], &channelStatus);
            if (accessStatus == ARS_NORMAL) {
               if (channelStatus & CEIDEV_CSR_RECV_FIFO_OVERFLOW) {
                  recvStatus = ARS_RX_BUFFER_OVERRUN;
                  accessStatus = cei_write_32bit ((CDEV_BOARD_TYPE)board, &deviceLocalStructPtr->devicePtr[memOffset], CEIDEV_CSR_RECV_FIFO_OVERFLOW);
               }
               else if (channelStatus & CEIDEV_CSR_RECV_FIFO_INVALID_MSG) {
                  recvStatus = ARS_BAD_MESSAGE;
                  accessStatus = cei_write_32bit ((CDEV_BOARD_TYPE)board, &deviceLocalStructPtr->devicePtr[memOffset], CEIDEV_CSR_RECV_FIFO_INVALID_MSG);
               }
            }
         
            if ((accessStatus == ARS_NORMAL) && (channelStatus & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY)) {
               
               /* limit requested message count by the FIFO count in the status register */
               channelStatus = (channelStatus & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT;
               messageTotal = maxMessages < (CEI_INT32)channelStatus ? maxMessages : (CEI_INT32)channelStatus;

               /* reference supplied pointers locally */
               dataPtr = data;
               if (timeTags != NULL) {
                  retrieveTimeTag = TRUE;
               
                  /* if a 32-bit time-tag ref. isn't already selected, force 32-bit as a default */
                  if ( (timeTagReference != AR_TIMER_X20_COMPAT_32BIT) &&
                       (timeTagReference != AR_TIMETAG_INT_20USEC_32BIT) && 
                       (timeTagReference != AR_TIMETAG_INT_MSEC_32BIT) ) {
                     timeTagReference = AR_DEFAULT_RECEIVE_DATA_TIMETAG_FORMAT;
                  }
               }
   
               /* setup all offsets that might be used in the message/time-stamp retreival loop */
               memOffset =  CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_BUFFER_0 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE);
               chanOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_BUFFER_2 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE);
               dataOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_BUFFER_3 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE);
   
               /* attempt to retrieve the requested block of messages */
               messageCount = 0;
               while ((messageCount < messageTotal) && (accessStatus == ARS_NORMAL)) {
                  if (retrieveTimeTag) {
                     /* if merged mode is enabled for this channel read the actual 
                      * channel on which the message was received from the Rx buffer */
                     if (mergedModeState == AR_ON) {
                        accessStatus = cei_read_32bit ((CDEV_BOARD_TYPE)board, &deviceLocalStructPtr->devicePtr[chanOffset], (pCEI_UINT32)&mergedModeChan);
                     }
   
                     accessStatus = a429_utl_acquire_formatted_time_tag ((CDEV_BOARD_TYPE)board, 
                                                                         memOffset,
                                                                         AR_RECEIVER_FIFO_TIMETAG, 
                                                                         timeTagReference, 
                                                                         mergedModeChan, 
                                                                         &timeTagRef );
                     if (accessStatus == ARS_NORMAL) {
                        /* return the modified 32-bit time-tag */
                        *timeTagPtr = (CEI_UINT32)(timeTagRef.timeTag & 0xFFFFFFFF);
                        timeTagPtr++;
                     }
                  }
         
                  /* return the 32-bit data word */
                  accessStatus = cei_read_32bit ((CDEV_BOARD_TYPE)board, &deviceLocalStructPtr->devicePtr[dataOffset], (pCEI_UINT32)dataPtr);
                  dataPtr++;
                  messageCount++;
               }
               
               if (messageCount) {
                  status = ARS_GOTDATA;
                  *actualCount = messageCount;
               }
            }
            else { /* no data available, exit with status = ARS_NODATA */
               status = ARS_NODATA;
            }
#ifndef TARGET_EMBEDDED
            /* if multi-thread protection is active, free any acquired lock */
            if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
               lockStatus = cei_release_thread_lock(LOCK_IDENT);
               if (lockStatus != ARS_NORMAL) {
                  status = lockStatus;
               }
            }
         }
         else {
            status = lockStatus;
         }
      }
      if ( ((status == ARS_NORMAL) || (status == ARS_NODATA)) && (accessStatus != ARS_NORMAL)) {
         status = accessStatus;
      }
      if ((recvStatus != ARS_NORMAL) && (accessStatus == ARS_NORMAL) && (lockStatus == ARS_NORMAL)) {
#else
      if ((recvStatus != ARS_NORMAL) && (accessStatus == ARS_NORMAL)) {
#endif
         status = recvStatus;
      }
#ifndef TARGET_EMBEDDED
   }
#endif
   
   return (status);
}


/*===================================================================================================================*
 * ENTRY POINT:                                A R _ G E T _ L A T E S T
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the latest message received for the specified channel/label combination from the snapshot 
 *              buffer, (for legacy API compatibility with the CEI-x20 API).
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board        -- (input)  Board number (0 to 127)  
 *              CDEV_CHAN_TYPE channel       -- (input)  ARINC 429 receive channel to reference (0 to 31)
 *              CDEV_PARM_USI_TYPE label     -- (input)  Label of interest
 *              CDEV_PARM_VOID_PTR_TYPE data -- (output) Pointer to the location to store the message
 *              pCEI_CHAR seq_num            -- (output) Legacy parameter, implemented by inserting the least
 *                                                       significant byte of the message time stamp.
 * 
 * RETURN VAL:  NONE
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 *
 * DESCRIPTION: This routine copies the latest ARINC 492 message received for the specified channel/label combination 
 *              from the snapshot buffer.  When using this routine, the host application should set the snapshot 
 *              storage mode to "label field only".  This will setup the device to store snapshot data based on the 
 *              label field value only, ignoring the SDI bit field value.
 *
 *              The snapshot buffer access by this routine for a single label/channel combination is defined with 
 *              the bits of the ARINC message as follows:
 *
 *                     field    --------label---------
 *                      bits    7  6  5  4  3  2  1  0
 *
 *              If the label requested is label 256 (an ARINC label that does ot exist) or the value ARU_ALL_LABELS, 
 *              this function returns all 255 ARINC labels for the specified channel (000-255), in sequence, in 
 *              successive 'data' array elements.  This function assumes the caller has allocated at least 256*4 
 *              bytes for 'data' when used in this mode.
 *                    
 *===================================================================================================================*/ 
EXPORT32 CEI_VOID  DLL_EXPORTED ar_get_latest 
(
   CDEV_BOARD_TYPE board,                                               /* (input)  Board number (0 to 127)          */
   CDEV_CHAN_TYPE  channel,                                             /* (input)  429 Rx channel selection         */
   CDEV_PARM_USI_TYPE label,                                            /* (input)  Label of interest                */
   CDEV_PARM_VOID_PTR_TYPE data,                                        /* (output) pointer to the host data storage */
   pCEI_CHAR seq_num                                                    /* (n/a)    legacy parameter, not supported  */
)
{
   CEI_UINT32 snapshotBufferIndex = 0;                                        /* snapshot buffer offset label/sdi    */
   CEI_UINT32 snapStorageMode = 0;                                            /* local snapshot storage selection    */
   CEI_UINT32 labelIndex;                                                     /* local temp enh filter table index   */
   CEI_UINT32 localChannel;                                                   /* merged/individual channel select    */
   pCEI_UINT32 localDataPtr;                                                  /* temp pointer to application data    */
   CEI_UINT32 memOffset;                                                      /* local offset to channel memory area */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                     /* function return status              */
   CEI_INT32 accessStatus = ARS_NORMAL;                                       /* status of device read/write access  */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr = NULL;                   /* pointer to board local structure    */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   /* Snapshot Buffer access requests are limited to the number of ARINC 429 receive channels installed on the device */
   else if ( ((CEI_UINT32)channel > deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX] ) || (data == NULL) ) {
      status = ARS_INVARG;
   }
   else {
#endif
      deviceLocalStructPtr = &device[board];
      localChannel = (CEI_UINT32)channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
      
#ifndef TARGET_EMBEDDED
      status = a429_utl_validate_channel_request (board, localChannel, ARU_ACCESS_SNAPSHOT_BUFFER);
      if (status == ARS_NORMAL) {
#endif
         if (seq_num != NULL) {
            *seq_num = (CEI_CHAR)0;
         }
#ifndef TARGET_EMBEDDED
         /* if multi-thread protection requested, acquire semaphore/mutex-locking */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
         }
#endif
         if (accessStatus == ARS_NORMAL) {
            /* read the current snapshot storage mode selection: label only or label/SDI combined */
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], &snapStorageMode);
         }
         /* this routine must only be invoked when the snapshot storage mode selection is label only */
         if ((accessStatus == ARS_NORMAL) && (snapStorageMode & CEIDEV_SNAPSHOT_STORAGE_ENHANCED)) {
            if ((accessStatus == ARS_NORMAL) && (label < 256U)) {
               /* access the specific location in the snapshot buffer for the specified channel and label */
               memOffset = CEIDEV_SNAPSHOT_BUFFER_START + (CEI_UINT32)(localChannel << CEIDEV_SNAPSHOT_BUFFER_CH_SHIFT) + 
                                                          (CEI_UINT32)(label << CEIDEV_SNAPSHOT_BUFFER_LB_SHIFT);
               if (seq_num != NULL) {
                  accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &snapshotBufferIndex);
                  *seq_num = (CEI_CHAR)(snapshotBufferIndex & 255U);
               }

               memOffset += CEIDEV_SNAPSHOT_BUFFER_DATA_OFF;
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], (pCEI_UINT32)data);
            }
            else if ((accessStatus == ARS_NORMAL) && ((label == 256U) || (label == ARU_ALL_LABELS)) ) {
               localDataPtr = (pCEI_UINT32)data;
               snapshotBufferIndex = (CEI_UINT32)(localChannel << CEIDEV_SNAPSHOT_BUFFER_CH_SHIFT) + CEIDEV_SNAPSHOT_BUFFER_DATA_OFF;
               memOffset = CEIDEV_SNAPSHOT_BUFFER_START + snapshotBufferIndex;
               for (labelIndex = 0U; labelIndex < 256U; labelIndex++) {
                  accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], localDataPtr);
                  memOffset += 4U;
                  localDataPtr++;
               }
            }
         }
#ifndef TARGET_EMBEDDED
         /* if multi-thread protection is active, free any acquired lock */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
         }
#endif
         if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL)) {
            status = (CDEV_API_RET_TYPE)accessStatus;
         }
#ifndef TARGET_EMBEDDED
      }
   }
#endif
}



/*===================================================================================================================*
 * ENTRY POINT:                          A R _ G E T _ 4 2 9 _ M E S S A G E
 *===================================================================================================================*
 *
 * FUNCTION:    Read an ARINC 429 message from the specified 429 receive channel buffer.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board           -- (input)  board to reference (0 to 127)
 *              CDEV_CHAN_TYPE channel          -- (input)  429 receive channel to reference (0 to 31)
 *              CDEV_PARM_SSI_TYPE waitState    -- (input)  specifies whether or not to wait up to 1/2 second for a, 
 *                                                          message to appear in the receive buffer, designated as 
 *                                                          AR_ON (wait) or AR_OFF (don't wait)
 *              CDEV_PARM_VOID_PTR_TYPE data    -- (output) location to store the received message
 *              CDEV_PARM_VOID_PTR_TYPE timeTag -- (output) location to store the message time stamp
 *
 * RETURN VAL   CDEV_API_RET_TYPE ARS_GOTDATA           Success
 *                                ARS_NODATA            No data available (waitState = AR_OFF)
 *                                ARS_CHAN_TIMEOUT      No data available (waitState = AR_ON)
 *                                ARS_BOARD_MUTEX       Access to the Board Lock timed-out/failed
 *                                ARS_INVBOARD          Invalid board parameter or board not initialized
 *                                ARS_INVHARVAL         Unsupported channel parameter
 *                                ARS_INVARG            A NULL data parameter value was detected
 *                                ARS_BAD_MESSAGE       An invalid length msg was received
 *                                ARS_RX_BUFFER_OVERRUN A receive buffer overrun was detected
 * GLOBALS:     none
 * 
 * DESCRIPTION: This routine invokes either ar_getnextt or ar_getwordt, based on whether or not the waitState 
 *              parameter indicates AR_ON or AR_OFF, respectively.
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_get_429_message  
( 
   CDEV_BOARD_TYPE board,                                                /* (input)  board to reference (0 to 127)   */
   CDEV_CHAN_TYPE  channel,                                              /* (input)  channel number to ref (0 to 31) */
   CDEV_PARM_SSI_TYPE waitState,                                         /* (input)  to wait for 1/2 second or not   */
   CDEV_PARM_VOID_PTR_TYPE data,                                         /* (output) location to store received data */
   CDEV_PARM_VOID_PTR_TYPE timeTag                                       /* (output) location to store data timeTag  */
)
{
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                               /* function return status    */
      
   if (waitState == AR_ON) {
      status = ar_getnextt ( board, channel, data, timeTag );
   }
   else {
      status = ar_getwordt ( board, channel, data, timeTag );
   }

   return (status);
}
           


