/*====================================================================================================================*
 * FILE:                                     C D E V _ A P I _ L E G A C Y _ A P I. C                    
 *====================================================================================================================*
 *
 *   COPYRIGHT 2002 - 2019 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED
 *   ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH THE INCLUSION
 *   OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY OTHER COPIES THEREOF
 *   MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE
 *   AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS, INC. 
 *
 *====================================================================================================================*
 *
 * FUNCTION:    The Legacy Routines part of the Application Programmer's Interface for a CEI-x30 ARINC device.
 *
 * DESCRIPTION: These routines are provided to demonstrate the backward compatible API programming primitives for the 
 *              CEI-x30 API using the C programming language.  Each routine implements a particular API function; 
 *              however, a user application program may need to combine functions in different ways to perform the 
 *              desired operation.
 *
 *              Most routines utilize a board number parameter, zero-based on the Device ID/Board Number assigned 
 *              during the installation of CEI-x30-SW distribution.  Channel parameter values are all zero-based.
 *
 * EXTERNALLY DEFINED ROUTINE ENTRY POINTS:         
 *
 *    CEI-x30 Legacy ARINC API routines:
 *
 *    ar_loadslv                  - initializes the API and device
 *    ar_flush_receiver       
 *    ar_get_raw_mode
 *    ar_get_status               - returns FIFO Data Avail status for all rx
 *    ar_get_timercntl            - returns a 32-bit rep. of current device timer value
 *    ar_get_timercnt             - returns a 16-bit rep. of current device timer value
 *    ar_init_dual_port
 *    ar_init_slave
 *    ar_recreate_parity
 *    ar_reset_int
 *    ar_reset_timercnt
 *    ar_set_config               - configure using x20-generic 429 def's
 *    ar_set_raw_mode             - enables/disables parity for a channel
 *    ar_set_timerrate            - sets an x20-compatible device timer resolution
 *    ar_setchparms
 *    ar_setinterrupts
 *    ar_wait                     - delay a specified number of seconds
 *    ar_xmit_sync
 *    
 *    The following routines are present for compilation purposes only, they do not perform any operations on the board.
 *    
 *    ar_dump_dp 
 *    ar_force_version 
 *    ar_formatarinclabel
 *    ar_formatbinarylabel
 *    ar_get_errorLV
 *    ar_get_harris
 *    ar_int_set
 *    ar_int_slave
 *    ar_loadslv_init_discrete
 *    ar_msg_control
 *    ar_putword2x1
 *    ar_reformat
 *    ar_set_arinc_config
 *    ar_set_harris 
 *    ar_timetag_control
 *    ar_unformat
 *                            
 *    
 *--------------------------------------------------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date    By   Ver. Comments
 * -------- ----- ---- -----------------------------------------------------------------------------------------------
 * 10/18/16  BDW  4.90 Initial version.
 *
 *====================================================================================================================*/ 
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
   #include <unistd.h>
   #include <fcntl.h>
   #include <string.h>
   #include <pthread.h>
   #include <errno.h>
   #include <sys/stat.h> 
   #include <sys/shm.h>
   #include <sys/sem.h>
   #include <sys/syscall.h>
   #define VOLATILE volatile
#elif defined (VXW_PCI_X86) || defined (VXW_PCI_PPC)
   #include <stdio.h>
   #include <stdlib.h>
   #include <unistd.h>
   #define VOLATILE 
#elif defined (NIVISA_PXI)
   #ifdef LABVIEW_RT
      #include <ansi_c.h> 
   #endif
   #include <visa.h>  
   #define VOLATILE 
#elif defined (INTEGRITY_PCI_PPC)
   #include <string.h>
   #define VOLATILE 
#else
   #define VOLATILE volatile
#endif

#include "cdev_glb.h"

extern VOLATILE ARINC_LOCAL_TYPE device[MAX_BOARDS];
extern ARINC_GLOBAL_TYPE * deviceGlobal[MAX_BOARDS];


#ifdef INCLUDE_ARINC_717
extern EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_set_573_config (CDEV_BOARD_TYPE board, CDEV_PARM_SSI_TYPE item, CDEV_PARM_SI_TYPE value);  
#endif


/*===================================================================================================================*
 * ENTRY POINT:                                    A R _ L O A D S L V
 *===================================================================================================================*
 *
 * FUNCTION:    This routine initializes the API and device.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board     -- (input) device number (0-127)
 *              CEI_UINT32 base_seg       -- (input) not used       (present for
 *              CEI_INT32 base_port       -- (input) not used  API compatibility
 *              CEI_UINT16 ram_size       -- (input) not used              only)
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL           Success
 *                                ARS_INVBOARD         board parm value not supported
 *                                ARS_BADLOAD          FPGA load process failed
 *                                ARS_BOARD_MUTEX      Creation of the Board Lock timed-out/failed
 *                                ARS_DRIVERFAIL       Board session open failed
 *                                ARS_HW_DETECT        Supported device not detected
 *                                ARS_ERR_SH_MEM_MAP   failed to map shared memory block (multi-process
 *                                ARS_ERR_SH_MEM_OBJ   failed to allocate shared memory block      only)
 *                                ARS_FW_NOT_SUPPORTED Device firmware revision not supported
 *                                ARS_BADLOAD     FPGA load process failed
 *                                ARS_MEMWRERR         Device memory test failure
 *                                ARS_FAILURE          <Windows only> CEI_Install library failed to load in memory
 *                                                     <Linux multi-process> failed to create a shared memory block
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine fulfills the standard ARINC API initialization functionality by invoking ar_open.
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_loadslv 
(
   CDEV_BOARD_TYPE board,                                                    /* (input) device number (0-127)        */
   CEI_UINT32 base_seg,                                                      /* (input) legacy parameter, not used   */
   CEI_INT32  base_port,                                                     /* (input) legacy parameter, not used   */
   CEI_UINT16 ram_size                                                       /* (input) legacy parameter, not used   */
)
{
   base_seg = (CEI_UINT32)ram_size;
   base_seg = (CEI_UINT32)base_port;
   return ar_open (board);
}


/*===================================================================================================================*
 * ENTRY POINT:                                 A R _ G E T _ S T A T U S
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve FIFO Data Available bit status for all receivers
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board        -- (input)  board to reference (0 to 127)
 *              CDEV_PARM_SSI_PTR_TYPE state -- (output) bitwise value or'ing the Rx Status Reg msg avialable status 
 * 
 * RETURN VAL   CDEV_API_RET_TYPE ARS_NORMAL    Success
 *                                ARS_INVBOARD  Invalid board parameter
 * 
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: Supplies the combined state of each Rx FIFO status register FIFO Data Available bit.  Supports up 
 *              to 16 multi-protocol receivers.
 *
 *              For the CEI-x30 API, this routine returns the combined state of the Rx FIFO Status Register Data 
 *              Available Bit for each of the individual receiver FIFOs. 
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_get_status 
( 
   CDEV_BOARD_TYPE board,                                           /* (input)  board to reference (0 to 127)        */
   CDEV_PARM_SSI_PTR_TYPE state                                     /* (output) bit value from each FIFO status reg. */
)
{
   CEI_UINT32 channelIndex;                                                    /* local channel index                */
   CEI_UINT32 channelStatus;                                                   /* temp FIFO status by channel        */
   CEI_UINT32 memOffset;                                                     /* local offset to channel memory area  */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                      /* function return status             */
   CEI_INT32 accessStatus = ARS_NORMAL;                                        /* status of device read/write access */
   CEI_INT32 combinedStatus = 0;                                               /* accum. FIFO status for all chan's  */
   
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II)
      status = ARS_INVBOARD;
   else {
#endif
      memOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS + (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] * CEIDEV_CHANNEL_REGION_SIZE);
      for ( channelIndex = 0; 
            (channelIndex < deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) && (channelIndex < 16) && (accessStatus == ARS_NORMAL); 
            channelIndex++ ) {
         accessStatus = cei_read_32bit (board, &device[board].devicePtr[memOffset], &channelStatus);
         combinedStatus |= ( (channelStatus & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) << channelIndex);
         memOffset += CEIDEV_CHANNEL_REGION_SIZE;                        /* setup for next channel */
      }
      
      *state = (CDEV_PARM_SSI_TYPE)(combinedStatus & 0xFFFF);
      
#ifndef TARGET_EMBEDDED
      if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL)) {
         status = (CDEV_API_RET_TYPE)accessStatus;
      }
   }
#endif
   
   return status;
}
 

/*===================================================================================================================*
 * ENTRY POINT:                                A R _ G E T _ T I M E R C N T L
 *===================================================================================================================*
 *
 * FUNCTION:    This routine returns the value of the device timer in a 32-bit format.  If the current selected 
 *              system time format is 64-bit or IRIG, the 1 millisecond 32-bit format is used.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input) Board number (0 to 127)  
 * 
 * RETURN VAL:  CEI_UINT32 -- the value of the device timer
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 *===================================================================================================================*/
EXPORT32 CEI_UINT32 DLL_EXPORTED ar_get_timercntl
(
   CDEV_BOARD_TYPE board                                                   /* (input)  board to reference (0 to 127) */
) 
{
   CEI_UINT32 currentTime;                                                   /* temp storage for current timer value */
   AR_TIMETAG_TYPE timeTag;                                                  /* temp struct storage for current time */
   CEI_INT32 timeTagReference;
   CEI_INT32 status;

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      /* if a 32-bit time-tag ref. isn't already selected, force 32-bit default */
      timeTagReference = (CEI_INT32)deviceGlobal[board]->deviceTimeTagReference;
      if ( (timeTagReference != AR_TIMER_X20_COMPAT_32BIT) &&
           (timeTagReference != AR_TIMETAG_INT_20USEC_32BIT) ) {
         timeTagReference = AR_TIMETAG_INT_MSEC_32BIT;
      }
      
      status = a429_utl_acquire_formatted_time_tag ( board,
                                                     0U,
                                                     AR_TIMER_REGISTER_SET, 
                                                     timeTagReference, 
                                                     AR_OFF, 
                                                     &timeTag );
      
      currentTime = (CEI_UINT32)timeTag.timeTag;
      
      if (status) ;
#ifndef TARGET_EMBEDDED
   }
   else {
      currentTime = ARS_INVBOARD;
   }
#endif
   return ( currentTime );
}
EXPORT32 CDEV_PARM_USI_TYPE DLL_EXPORTED ar_get_timercnt (CDEV_BOARD_TYPE board) 
{
   return (CDEV_PARM_USI_TYPE)ar_get_timercntl(board);
}


/*===========================================================================*  
 * CEI-x20 API Functions - Providing Limited Compatibility from x20 to x30
 *===========================================================================*/             
/*===================================================================================================================*
 * ENTRY POINT:                         A R _ G E T _ R A W _ M O D E
 *===================================================================================================================*
 *
 * FUNCTION:    Report the active ARINC 429 Parity Mode
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board        -- (input) board to reference (0 to 127)
 *              CDEV_PARM_SSI_TYPE direction -- (input) TRANSMIT_CHANNEL / RECEIVE_CHANNEL
 *                                                      alternates: ARU_XMIT / ARU_RECV
 *              CDEV_CHAN_TYPE channel       -- (input) channel number (0 to 31)
 *
 * GLOBALS:     none
 * 
 * RETURN VAL:  CDEV_API_RET_TYPE AR_ON      Transmit Parity assignment is enabled/Receive Parity checking is enabled
 *                                AR_OFF     Transmit Parity generation or Receive Parity checking is disabled
 * 
 * DESCRIPTION: This routine reports the state of the transmitter and receiver parity control bits based on the 
 *              whether or not the Parity Enable bit is set in the respective channel configuration register.
 * 
 *              This routine is provided for x20 compatibility.  The proper method of checking parity is via 
 *              ar_get_device_config().
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_get_raw_mode (CDEV_BOARD_TYPE board, CDEV_PARM_SSI_TYPE direction, CDEV_CHAN_TYPE channel) 
{
   CDEV_API_RET_TYPE status;
   CDEV_PARM_SSI_TYPE parity;
   if (direction == ARU_RECV) {
      status = ar_get_device_config (board, channel, ARU_RX_PARITY, &parity);
      parity = (CDEV_PARM_SSI_TYPE)((parity == (CEI_INT16)AR_ON) ? AR_OFF : AR_ON);
   }
   else {
      status = ar_get_device_config (board, channel, ARU_TX_PARITY, &parity);
      parity = (CDEV_PARM_SSI_TYPE)((parity == (CEI_INT16)ARU_PARITY_NONE) ? AR_ON : AR_OFF);
   }
   if (status) ;
   return (CDEV_API_RET_TYPE)parity;
}

/*===================================================================================================================*
 * ENTRY POINT:            A R _ I N I T _ D U A L _ P O R T
 *===================================================================================================================*
 *
 * FUNCTION:    Initialize the device to a ready state
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board  -- (input) device number to reference (0-15)
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL       Success
 *                                ARS_MEMWRERR     Memory test failure
 *                                ARS_INVBOARD     Invalid board parameter
 *                                ARS_BOARD_MUTEX  Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     none
 * 
 * DESCRIPTION: This routine initializes the device I/O configuration registers to a state ready for use by the 
 *              host application.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_init_dual_port (CDEV_BOARD_TYPE board) 
{
   return ar_initialize_device(board);
}
/*===================================================================================================================*
 * ENTRY POINT:                                 A R _ I N I T _ S L A V E
 *===================================================================================================================*
 *
 * FUNCTION:    Initialize the device transmit and receive buffers
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board  -- (input) device number to reference (0-127)
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL       Success
 *                                ARS_INVBOARD     Invalid board parameter
 *                                ARS_BOARD_MUTEX  Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     deviceGlobal
 *              device
 *
 * DESCRIPTION: This routine initializes/flushes the device's transmit and receive buffers.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_init_slave (CDEV_BOARD_TYPE board)
{
   CEI_UINT32 labelIndex;                                                       /* local label filter table index    */
   CEI_UINT32 globalEnableRegValue = 0;                                         /* temp var for device register data */
   CEI_UINT32 bitFieldValue;                                                    /* temp var for device register data */
   CEI_UINT32 configRegValue[64];                                               /* temp var for device register data */
   CEI_UINT32 deviceChannel;                                                    /* temp var for device register data */
   CEI_UINT32 memOffset;                                                    /* local offset to channel memory area   */
   CEI_UINT32 channelOffset;                                          /* chan reg region offset in host interface    */
   CEI_UINT32 bufferOffset;                                           /* chan buffer region offset in host interface */
   CEI_UINT32 initProgress = INIT_FLUSH_TX;                                          /* init state of progression    */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                         /* local invocation return status  */
   CEI_INT32  accessStatus = ARS_NORMAL;                                       /* status of device read/write access */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                                /* ptr to board local structure  */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                                        /* ptr to brd global structure  */

#ifndef TARGET_EMBEDDED
   if ( (board < MAX_BOARDS) && (deviceGlobal[board] != NULL) && (deviceGlobal[board]->attachedProcessCount > 0) ) {
#endif
      /* assign references to this board's device and deviceGlobal structure-array elements */
      deviceLocalStructPtr = &device[board];
      deviceGlobalStructPtr = deviceGlobal[board];

#ifndef TARGET_EMBEDDED
      /* if multi-thread protection requested, acquire semaphore/mutex-locking */
      if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
         accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
      }
#endif

      /* Execute the initialization sequence for the board */
      while ((accessStatus == ARS_NORMAL) && (status == ARS_NORMAL) && (initProgress != INIT_TERMINATE)) {
         switch (initProgress) {
            case INIT_FLUSH_TX:
               /*------------------------------------------------------------------------------------------*
                *  Save the current Global Enable Register state and enable data processing 
                *------------------------------------------------------------------------------------------*/
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], &globalEnableRegValue);
               if (accessStatus == ARS_NORMAL) {
                  accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], CEIDEV_GLOBAL_ENABLE_AUTON_SET + CEIDEV_GLOBAL_ENABLE_ON);
               }
            
               /*--------------------------------------------------------------------*
                * Flush all ARINC 429 transmit buffers using the following process:
                *   Save the original transmit channel configuration
                *   Disable external transmission
                *   Set the receive rate to 8Mb/s
                *   Delay long enough to flush the entire buffer
                *--------------------------------------------------------------------*/
               if ((accessStatus == ARS_NORMAL) && (deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX] > 0)) {
                  channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX] * CEIDEV_CHANNEL_REGION_SIZE);
                  labelIndex = 0;  
                  for (deviceChannel = deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
                       (deviceChannel <= deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_TX]) && (accessStatus == ARS_NORMAL); 
                       deviceChannel++) {
                     accessStatus = cei_read_32bit  (board, &deviceLocalStructPtr->devicePtr[channelOffset], &configRegValue[labelIndex]);
                     if (accessStatus == ARS_NORMAL) {
                        bitFieldValue = (CEI_UINT32)0x00018041U;
                        accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], bitFieldValue);
#if defined (PPC_SYNC)
                     IO_SYNC;
#endif
                        channelOffset += CEIDEV_CHANNEL_REGION_SIZE;
                        labelIndex++;
                     }
                  }
                  
                  ar_sleep (9);

                  /* restore the original transmit channel configuration */
                  channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX] * CEIDEV_CHANNEL_REGION_SIZE);
                  labelIndex = 0;
                  for (deviceChannel = deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];  
                       deviceChannel <= deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_TX] && (accessStatus == ARS_NORMAL); 
                       deviceChannel++) {
                     accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], configRegValue[labelIndex]);
                     channelOffset += CEIDEV_CHANNEL_REGION_SIZE;
                     labelIndex++;
                  }
               }
               if (deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC573_RX] > 0) {
                  initProgress = INIT_717;
               }
               else {
                  initProgress = INIT_ENABLE_FLUSH_RX;
               }
               break;

            case INIT_717:
               /*--------------------------------------------------------------------*
                * if present, flush the 717 receiver
                *--------------------------------------------------------------------*/
               /* setup the receiver for raw bit reception (disable autosync) */
               bitFieldValue = CEIDEV_573_FRAME_768_64 + CEIDEV_573_RX_BPRZ_SELECT + CEIDEV_573_RX_MODE_RAW;
               channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX] * CEIDEV_CHANNEL_REGION_SIZE);
               accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], bitFieldValue);
               
               if (accessStatus == ARS_NORMAL) {
                  channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX] * CEIDEV_CHANNEL_REGION_SIZE);
                  bufferOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_BUFFER_3 + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX] * CEIDEV_CHANNEL_REGION_SIZE);
                  /* clear any outstanding error status bits */
                  bitFieldValue = CEIDEV_CSR_RECV_FIFO_NOT_EMPTY + CEIDEV_CSR_RECV_FIFO_INVALID_MSG;
                  accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], bitFieldValue);
#if defined (PPC_SYNC)
                  IO_SYNC;
#endif
                  while ((bitFieldValue & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) && (accessStatus == ARS_NORMAL)) {
                     accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[bufferOffset], &labelIndex);
                     if (accessStatus == ARS_NORMAL) {
                        accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &bitFieldValue);
                     }
                  }
                  /* now enable auto-sync */
                  if (accessStatus == ARS_NORMAL) {
                     bitFieldValue = CEIDEV_573_FRAME_768_64 + CEIDEV_573_RX_BPRZ_SELECT;
                     channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX] * CEIDEV_CHANNEL_REGION_SIZE);
                     accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], bitFieldValue);
                  }
               }
               initProgress = INIT_ENABLE_FLUSH_RX;
               break;
            
            case INIT_ENABLE_FLUSH_RX:
               /*--------------------------------------------------------------------*
                *  Enable and Flush the ARINC 429 receivers
                *--------------------------------------------------------------------*/
               memOffset = CEIDEV_CHANNEL_REGION_START + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] * CEIDEV_CHANNEL_REGION_SIZE);
               channelOffset = memOffset + CEIDEV_CHANNEL_CONFIG_1;
               bufferOffset = memOffset + CEIDEV_CHANNEL_BUFFER_3;
               if (deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX] > 0) {
                  for (deviceChannel = deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]; 
                       (deviceChannel <= deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]) && (accessStatus == ARS_NORMAL); 
                       deviceChannel++) {
                     /* enable the receiver, using current configuration settings */
                     accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &bitFieldValue);
                     if (accessStatus == ARS_NORMAL) {
                        accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], (bitFieldValue+CEIDEV_CHANNEL_ENABLE));
                        if (accessStatus == ARS_NORMAL) {
                           /* clear any outstanding error status bits */
                           bitFieldValue = CEIDEV_CSR_RECV_FIFO_NOT_EMPTY + CEIDEV_CSR_RECV_FIFO_INVALID_MSG;
                           accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], bitFieldValue);
                           if (accessStatus == ARS_NORMAL) {
#if defined (PPC_SYNC)
                              IO_SYNC;
#endif
                              accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &bitFieldValue);    /* read the status register - check for presense of data */
                              while ((bitFieldValue & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) && (accessStatus == ARS_NORMAL)) {
                                 accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[bufferOffset], &labelIndex); /* read the data register to reset status for this message */
                                 if (accessStatus == ARS_NORMAL) {
                                    accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &bitFieldValue); /* read the status register - check for presense of more data */
                                 }
                              }
                              memOffset  += CEIDEV_CHANNEL_REGION_SIZE;                            /* setup for next channel */
                              channelOffset += CEIDEV_CHANNEL_REGION_SIZE;
                              bufferOffset += CEIDEV_CHANNEL_REGION_SIZE;
                           }
                        }
                     }
                  }
               
                  /*--------------------------------------------------------------------*
                   * Flush the merged mode receiver FIFO
                   *--------------------------------------------------------------------*/
                  if (accessStatus == ARS_NORMAL) {
                     bufferOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_BUFFER_3 + (deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_MERGED_RX] * CEIDEV_CHANNEL_REGION_SIZE);
                     channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS + (deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_MERGED_RX] * CEIDEV_CHANNEL_REGION_SIZE);
                     /* clear any outstanding error status bits */
                     bitFieldValue = CEIDEV_CSR_RECV_FIFO_NOT_EMPTY + CEIDEV_CSR_RECV_FIFO_INVALID_MSG;
                     accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], bitFieldValue);
                     if (accessStatus == ARS_NORMAL) {
#if defined (PPC_SYNC)
                        IO_SYNC;
#endif
                        accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &bitFieldValue);    /* read the status register - check for presense of data */
                        while ((bitFieldValue == CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) && (accessStatus == ARS_NORMAL)) {
                           accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[bufferOffset], &labelIndex);     /* read the data register to reset status for this message */
                           if (accessStatus == ARS_NORMAL) {
                              accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &bitFieldValue); /* read the status register - check for presense of more data */
                           }
                        }
                     }
                  }
               }
               initProgress = INIT_RESTORE;
               break;
         
            case INIT_RESTORE:
               /* restore the global enable register */
               accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], globalEnableRegValue);
               initProgress = INIT_TERMINATE;
               break;
         }
#ifndef TARGET_EMBEDDED
         /* if multi-thread protection is active, free any acquired lock */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
         }
#endif
      }
#ifndef TARGET_EMBEDDED
   }
   else {
      status = ARS_INVBOARD;
   }
#endif
   if (accessStatus) ;
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                               A R _ R E S E T _ T I M E R C N T
 *===================================================================================================================*
 *
 * FUNCTION:    This routine resets the 64-bit device timer to zero.
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board        -- (input) Board number (0 to 127)  
 *
 * GLOBALS:     device[]
 *
 *===================================================================================================================*/
EXPORT32 CEI_VOID DLL_EXPORTED ar_reset_timercnt
(
   CDEV_BOARD_TYPE board                                                    /* (input)  board to reference (0 to 127)*/
)
{
#ifndef TARGET_EMBEDDED
   CDEV_API_RET_TYPE status = ARS_NORMAL;       /* local return status       */
#endif
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;  /* ptr to board local structure */
   
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      deviceLocalStructPtr = &device[board];
      /* if multi-thread protection is active, acquire a lock on the host interface */
#ifndef TARGET_EMBEDDED
      if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
         status = (CDEV_API_RET_TYPE)cei_acquire_thread_lock(LOCK_IDENT);
      }
      
      if (status == ARS_NORMAL) {
#endif
         accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_TIMER_LOW], (CEI_UINT32)0);
         if (accessStatus == ARS_NORMAL) {
            accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_TIMER_HIGH], (CEI_UINT32)0);
         }
#ifndef TARGET_EMBEDDED
         /* if multi-thread protection is active, free any acquired lock */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            status = (CDEV_API_RET_TYPE)cei_release_thread_lock(LOCK_IDENT);
         }
      }
   }
#endif
}

/*===================================================================================================================*
 * ENTRY POINT:                                  A R _ S E T _ C O N F I G
 *===================================================================================================================*
 *
 * FUNCTION:    Initialize channel configuration, BusTools product specific.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE    board  -- (input) board to reference (0 to 127)
 *              CDEV_PARM_SSI_TYPE item   -- (input) item for which to set the 
 *                                                   current state:
 * 
 *                    ARU_RX_CH**_BIT_RATE          ARU_DISCRETE_OUT*
 *                    ARU_TX_CH**_BIT_RATE          ARU_DISCRETE_OUT
 *                    ARU_RX_CH**_PARITY            ARU_DISCRETE_VALUES/ARU_DISCRETE_OUTPUTS
 *                    ARU_TX_CH**_PARITY            ARU_IRIG_INPUT_THRESHOLD  
 *                    ARU_TX_CH**_SHUT_OFF          ARU_INTERNAL_WRAP         
 *                    ARU_TX_CH**_HB_INJ            ARU_IRIG_QUICK_ADJUSTMENT 
 *                    ARU_TX_CH**_LB_INJ            ARU_IRIG_SET_BIAS         
 *                    ARU_TX_CH**_GAP_INJ           ARU_IRIG_WRAP_ENABLE
 *                    ARU_ACCESS_SNAPSHOT_BUFFER    ARU_IRIG_ADJUST_THRESHOLD
 *                                                  
 *                    ARU_RX_TIMETAG_MODE w/values of:                          
 *                                                     
 *                              AR_TIMETAG_EXT_IRIG_64BIT                         
 *                              AR_TIMETAG_INT_USEC_64BIT                         
 *                              AR_TIMETAG_INT_20USEC_32BIT                       
 *                              AR_TIMETAG_INT_MSEC_32BIT                         
 *                              AR_TIMETAG_HOST_SEC_32BIT                       
 *                              AR_TIMETAG_SYNC_1553_CH1 (for combo cards only)               
 *                              AR_TIMETAG_SYNC_1553_CH2                
 *                              AR_TIMETAG_SYNC_1553_CH3                
 *                              AR_TIMETAG_SYNC_1553_CH4                
 *                                                     
 *              CEI_UINT32 value -- (input) this is the value to which
 *                                      the specified item will be set
 * 
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL      Normal succesful completion.
 *                                ARS_INVARG      "item" parameter value is not
 *                                                supported by this function
 *                                ARS_INVHARVAL   Item was invalid or channel
 *                                                not supported on this device
 *                                ARS_INVBOARD    Board parameter value invalid or
 *                                                board has not been initialized
 *                                ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 * 
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: Calls the function ar_set_device_config () with the channel(s) of interest based on the expected 
 *              item definition.
 *      
 *        NOTE: The value passed into ar_set_device_config is UNSIGNED SHORT, the value passed here is CEI_UINT32, 
 *              so only the least signifciant 16 bits are utilized.
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_set_config 
( 
   CDEV_BOARD_TYPE board,                                                  /* (input)  board to reference (0 to 127) */
   CDEV_PARM_SSI_TYPE item,                                                /* (input)  item to process               */
   CEI_UINT32 value                                                        /* (input)  state of the item             */
)
{
   CDEV_API_RET_TYPE status;                                                    /* function return status            */
   CDEV_CHAN_TYPE channel;                                                      /* local item-derived channel index  */
   CDEV_CHAN_TYPE channelIndex;                                                 /* local item-derived channel index  */
   CDEV_CHAN_TYPE total429TxChannels;                                           /* local Tx total channel count      */
   CDEV_CHAN_TYPE total429RxChannels;                                           /* local Rx total channel count      */
   CEI_UINT16 locValue;                                                         /* local 16bit version of value parm */
   CEI_INT32 locLongValue;                                                      /* temp 32bit variable               */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                                   /* pointer to board global structure */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   else {
#endif
      deviceGlobalStructPtr = deviceGlobal[board];
      locValue = (CEI_UINT16)value;
      locLongValue = (CEI_INT32)value;
      status = ARS_NORMAL;
   
      /*-----------------------------------------------------------------*
       *  CEI-x20 ARINC 429 Set/Get Config Support
       *-----------------------------------------------------------------*/
      /* Configure receive channel bit rate in Hertz */
      if ( ((item >= ARU_RX_CH01_BIT_RATE) && (item <= ARU_RX_CH32_BIT_RATE)) || 
           ((item >= ARU_OLD_RX_CH01_BIT_RATE) && (item <= ARU_OLD_RX_CH16_BIT_RATE)) ) {
   
         if (item >= ARU_RX_CH01_BIT_RATE) {
            channel = (CEI_INT16)(item - ARU_RX_CH01_BIT_RATE);
         }
         else {
            channel = (CEI_INT16)(item - ARU_OLD_RX_CH01_BIT_RATE);
         }
   
         if (value > ARU_SPEED_LOW) {
            if ((deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15) ||
                (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15XT)) {
               locLongValue = (CEIDEV_20MHZ_REF_CLOCK / (CEI_INT32)value) - 2;
            }
            else {
               locLongValue = (CEIDEV_16MHZ_REF_CLOCK / (CEI_INT32)value) - 2;
            }
         }
         
         status = ar_set_device_config (board, channel, ARU_RX_BITRATE, (CEI_INT16)locLongValue);
      }
   
      /* Configure transmit channel bit rate in Hertz */
      else if ( ((item >= ARU_TX_CH01_BIT_RATE) && (item <= ARU_TX_CH32_BIT_RATE)) || 
                ((item >= ARU_OLD_TX_CH01_BIT_RATE) && (item <= ARU_OLD_TX_CH16_BIT_RATE)) ) {
   
         if (item >= ARU_TX_CH01_BIT_RATE) {
            channel = (CEI_INT16)(item - ARU_TX_CH01_BIT_RATE);
         }
         else {
            channel = (CEI_INT16)(item - ARU_OLD_TX_CH01_BIT_RATE);
         }
   
         if (value > ARU_SPEED_LOW) {
            if ((deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15) ||
                (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15XT)) {
               locLongValue = (CEIDEV_20MHZ_REF_CLOCK / (CEI_INT32)value) - 2;
            }
            else {
               locLongValue = (CEIDEV_16MHZ_REF_CLOCK / (CEI_INT32)value) - 2;
            }
         }
         
         status = ar_set_device_config (board, channel, ARU_TX_BITRATE, (CEI_INT16)locLongValue);
      }
   
      /* Configure receive channel parity */
      else if ( ((item >= ARU_RX_CH01_PARITY) && (item <= ARU_RX_CH32_PARITY)) ||
                ((item >= ARU_OLD_RX_CH01_PARITY) && (item <= ARU_OLD_RX_CH16_PARITY)) ) {    
         
         if (item >= ARU_RX_CH01_PARITY) {
            channel = (CEI_INT16)(item - ARU_RX_CH01_PARITY);
         }
         else {
            channel = (CEI_INT16)(item - ARU_OLD_RX_CH01_PARITY);
         }
               
         status = ar_set_device_config (board, channel, ARU_RX_PARITY, locValue);
      }
   
      /* Configure transmit channel parity */
      else if ( ((item >= ARU_TX_CH01_PARITY) && (item <= ARU_TX_CH32_PARITY)) ||
                ((item >= ARU_OLD_TX_CH01_PARITY) && (item <= ARU_OLD_TX_CH16_PARITY)) ) {    
         
         if (item >= ARU_TX_CH01_PARITY) {
            channel = (CEI_INT16)(item - ARU_TX_CH01_PARITY);
         }
         else {
            channel = (CEI_INT16)(item - ARU_OLD_TX_CH01_PARITY);
         }
   
         status = ar_set_device_config (board, channel, ARU_TX_PARITY, locValue);
      }
   
      /* Configure transmit enable */
      else if ( ((item >= ARU_TX_CH01_SHUT_OFF) && (item <= ARU_TX_CH32_SHUT_OFF)) ||
                ((item >= ARU_OLD_TX_CH01_SHUT_OFF) && (item <= ARU_OLD_TX_CH16_SHUT_OFF)) ) {    
         
         if (item >= ARU_TX_CH01_SHUT_OFF) {
            channel = (CEI_INT16)(item - ARU_TX_CH01_SHUT_OFF);
         }
         else {
            channel = (CEI_INT16)(item - ARU_OLD_TX_CH01_SHUT_OFF);
         }
         
         status = ar_set_device_config (board, channel, ARU_TX_DISABLE, locValue);
      }
   
      /* Assign transmit channel bit count error injection */
      else if ( ((item >= ARU_TX_CH01_HB_INJ) && (item <= ARU_TX_CH32_HB_INJ)) ||
                ((item >= ARU_OLD_TX_CH01_HB_INJ) && (item <= ARU_OLD_TX_CH16_HB_INJ)) ) {    
         
         if (item >= ARU_TX_CH01_HB_INJ) {
            channel = (CEI_INT16)(item - ARU_TX_CH01_HB_INJ);
         }
         else {
            channel = (CEI_INT16)(item - ARU_OLD_TX_CH01_HB_INJ);
         }
   
         if (value == AR_ON) {
            locValue = AR_HI;
         }
         else {
            locValue = AR_OFF;
         }
   
         status = ar_set_device_config (board, channel, ARU_TX_BIT_ERROR, locValue);
      }
   
      /* Assign transmit channel bit count error injection */
      else if ( ((item >= ARU_TX_CH01_LB_INJ) && (item <= ARU_TX_CH32_LB_INJ)) ||
                ((item >= ARU_OLD_TX_CH01_LB_INJ) && (item <= ARU_OLD_TX_CH16_LB_INJ)) ) {    
         
         if (item >= ARU_TX_CH01_LB_INJ) {
            channel = (CEI_INT16)(item - ARU_TX_CH01_LB_INJ);
         }
         else {
            channel = (CEI_INT16)(item - ARU_OLD_TX_CH01_LB_INJ);
         }
   
         if (value == AR_ON) {
            locValue = AR_LO;
         }
         else {
            locValue = AR_OFF;
         }
   
         status = ar_set_device_config (board, channel, ARU_TX_BIT_ERROR, locValue);
      }
   
      /* Assign transmit channel message gap error injection */
      else if ( ((item >= ARU_TX_CH01_GAP_INJ) && (item <= ARU_TX_CH32_GAP_INJ)) ||
                ((item >= ARU_OLD_TX_CH01_GAP_INJ) && (item <= ARU_OLD_TX_CH16_GAP_INJ)) ) {    
         
         if (item >= ARU_TX_CH01_GAP_INJ) {
            channel = (CEI_INT16)(item - ARU_TX_CH01_GAP_INJ);
         }
         else {
            channel = (CEI_INT16)(item - ARU_OLD_TX_CH01_GAP_INJ);
         }
   
         status = ar_set_device_config (board, channel, ARU_TX_GAP_ERROR, locValue);
      }
   
      else {
         /* for the backward compatibility with old x20 options, only support
          * the total number of 429 channels here, omit 717 channels */
         total429TxChannels = ar_num_xchans(board);
         total429RxChannels = ar_num_rchans(board);
         if (deviceGlobalStructPtr->alternateJconfiguration == CEIDEV_ALT_CONFIGURATION_717) {
            total429TxChannels--;
            total429RxChannels--;
         }
   
         switch (item)
         {
            case ARU_XMIT_RATE:
               if (value > ARU_SPEED_LOW) {
                  if ((deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15) ||
                      (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15XT)) {
                     locLongValue = (CEIDEV_20MHZ_REF_CLOCK / (CEI_INT32)value) - 2;
                  }
                  else {
                     locLongValue = (CEIDEV_16MHZ_REF_CLOCK / (CEI_INT32)value) - 2;
                  }
               }
   
               for (channel = 0; channel < total429TxChannels; channel++) {
                  status = ar_set_device_config ( board, channel, ARU_TX_BITRATE, (CEI_INT16)locLongValue );
               }
               break;
      
            case ARU_RECV_RATE:
               if (value > ARU_SPEED_LOW) {
                  if ((deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15) ||
                      (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15XT)) {
                     locLongValue = (CEIDEV_20MHZ_REF_CLOCK / (CEI_INT32)value) - 2;
                  }
                  else {
                     locLongValue = (CEIDEV_16MHZ_REF_CLOCK / (CEI_INT32)value) - 2;
                  }
               }
   
               for (channel = 0; channel < total429RxChannels; channel++) {
                  status = ar_set_device_config ( board, channel, ARU_RX_BITRATE, (CEI_INT16)locLongValue );
               }
               break;
      
            case ARU_PARITY:
               for (channel = 0; channel < total429RxChannels; channel++) {
                  status = ar_set_device_config ( board, channel, ARU_RX_PARITY, locValue );
               }
               for (channel = 0; channel < total429TxChannels; channel++) {
                  status = ar_set_device_config ( board, channel, ARU_TX_PARITY, locValue );
               }
               break;
      
            case ARU_INTERNAL_WRAP:
#ifdef INCLUDE_ARINC_717
               /* include 717 channel with internal wrap */
               if (deviceGlobalStructPtr->alternateJconfiguration == CEIDEV_ALT_CONFIGURATION_717) {
                  status = ar_set_573_config (board, ARU_RECV_MODE, locValue);
               }
#endif
               for (channel = 0; channel < total429RxChannels; channel++) {
                  status = ar_set_device_config ( board, channel, ARU_RECV_MODE, locValue );
               }
               break;
      
            /* assign output discrete register "output discrete" values */
            case ARU_DISCRETE_OUT1:
            case ARU_DISCRETE_OUT2:
            case ARU_DISCRETE_OUT3:
            case ARU_DISCRETE_OUT4:
               channel = (CEI_INT16)(item - ARU_DISCRETE_OUT1);  /* determine channel number */
               status = ar_set_device_config (board, channel, ARU_DISCRETE_OUT, locValue);
               break;
      
            case ARU_IRIG_OUTPUT_ENABLE:
            case ARU_ACCESS_SNAPSHOT_BUFFER:
            case ARU_RX_TIMETAG_MODE:
            case ARU_IRIG_WRAP_ENABLE:
            case ARU_IRIG_INPUT_THRESHOLD:
            case ARU_IRIG_ADJUST_THRESHOLD:
            case ARU_IRIG_QUICK_ADJUSTMENT:
            case ARU_IRIG_SET_BIAS:
               status = ar_set_device_config (board, 0, item, locValue);
               break;
      
            case ARU_DISCRETE_VALUES:
            case ARU_DISCRETE_OUTPUTS:
               for (channelIndex = 0; channelIndex < (CDEV_CHAN_TYPE)deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_DISC_OUT]; channelIndex++) {
                  status = ar_set_device_config (board, channelIndex, ARU_DISCRETE_OUT, (CEI_INT16)(locValue & 1));
                  locValue >>= 1;
               }
               break;
   
            default : status = ARS_INVARG; 
               break;
         }
      }
#ifndef TARGET_EMBEDDED
   }
#endif
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                           A R _ S E T _ T I M E R R A T E
 *===================================================================================================================*
 *
 * FUNCTION:    This routine assigns the CEI-x20 compatible timer resolution
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input) board index (0 to 127)
 *              CDEV_PARM_SSI_TYPE rate -- (input) resolution of the CEI-x20-type timer in 250 nanoseconds.  
 *                                                 Note the CEI-x30 message scheduler rate resolution is
 *                                                 1ms, so the resolution of the rate value should be 4000 to match.
 * RETURN VAL:  none
 *
 * GLOBALS:     deviceGlobal[]
 * 
 *===================================================================================================================*/
EXPORT32 CEI_VOID  DLL_EXPORTED ar_set_timerrate 
(
   CDEV_BOARD_TYPE board,                                    /* (input) board index (0 to 127)                       */
   CDEV_PARM_SSI_TYPE rate                                   /* (input) resolution of the CEI-x20-type timer (usec)  */
)
{
   if ((board < MAX_BOARDS) && (deviceGlobal[board] != NULL)) {
      deviceGlobal[board]->deviceTimeTagReference = AR_TIMER_X20_COMPAT_32BIT;
      if (rate > 0) {
         deviceGlobal[board]->deviceX20CompatibleResolution = (CEI_UINT32)rate;
      }
   }
}


/*===================================================================================================================*
 * ENTRY POINT:                                          A R _ W A I T 
 *===================================================================================================================*
 *
 * FUNCTION:    Wait for the specified number of seconds
 * 
 * PARAMETERS:  CEI_FLOAT nsecs -- (input) number of seconds to wait for
 * 
 * RETURN VAL:  CEI_VOID
 *
 * GLOBALS:     none
 * 
 * DESCRIPTION: Delay the specified number of seconds based on the OS timer function clock() and the value of 
 *              TIMER_CLOCK_TICKS_PER_SECOND (see <time.h>).
 *                    
 *===================================================================================================================*/ 
EXPORT32 CEI_VOID  DLL_EXPORTED ar_wait (CEI_FLOAT nsecs) 
{
   cei_delay_seconds(nsecs);
}


#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC)

/* Legacy Even/odd parity conversion table for the routine ar_recreate_parity */
#define ODD     0x80                  /* indicates odd parity                 */
#define EVN     0x00                  /* indicates even parity                */

static CEI_UCHAR byte_parity[] = {
   EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD, ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN,
   ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN, EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD,
   ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN, EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD,
   EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD, ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN,
   ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN, EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD,
   EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD, ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN,
   EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD, ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN,
   ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN, EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD,
   ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN, EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD,
   EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD, ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN,
   EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD, ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN,
   ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN, EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD,
   EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD, ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN,
   ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN, EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD,
   ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN, EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD,
   EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD, ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN
};
EXPORT32 CEI_VOID  DLL_EXPORTED ar_recreate_parity (pCEI_UCHAR word) 
{
   CEI_NATIVE_INT i;         /* loop index                    */
   CEI_UCHAR parity;         /* stores the word parity        */

   /* calculate parity for the word  */
   for (i = 0,parity = (CEI_UCHAR)0x80; i < 4; i++)
      parity += (CEI_UCHAR)byte_parity[word[i]];

   /* set/clear the parity bit  */
   if (parity != (CEI_UCHAR)0) {
      word[3] |= 0x80;
   }
   else {
      word[3] &= 0x7F;
   }
}

EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_reset_int (CDEV_BOARD_TYPE board)  {board=0; return ARS_NORMAL;}
EXPORT32 CEI_INT32 DLL_EXPORTED ar_setinterrupts (CEI_UINT32 cardnum, CEI_UINT32 channel, CEI_INT32 TransRcv, CEI_UINT32 Count, CEI_UINT32 Mask) {Count=cardnum; Mask=channel+TransRcv; return ARS_NORMAL;}
#endif

/*===================================================================================================================*
 * ENTRY POINT:                                A R _ S E T C H P A R M S
 *===================================================================================================================*
 *
 * FUNCTION:    Define the operating parameters and modes for an ARINC 429 receiver and/or transmitter.  This 
 *              routine provides compatibility with the CEI-x20 API method for assigning channel configurations.
 *
 * PARAMETERS:  CEI_UINT32 cardnum    (input) The board number of interest
 *              CEI_UINT32 channel    (input) The channel to which to set the data (0-based)
 *              CEI_INT32 TransRcv    (input) Set parameters for ARU_TRANSMITTER or ARU_RECEIVER
 *              pAR_CHANNEL_PARMS def (input) Pointer to channel parameter structure
 *
 * RETURN VAL:  CEI_INT32 status ARS_NORMAL      Normal succesful completion.
 *                               ARS_INVARG      Invalid channel parameter value was supplied for the TransRcv value
 *                               ARS_INVHARVAL   Structure member value is invalid
 *                               ARS_INVBOARD    Board parameter value invalid or board has not been initialized
 *                               ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *
 * DESCRIPTION: This function is used to configure an ARINC 429 transmitter or receiver.
 *
 * GLOBALS:     deviceGlobal[board]
 *
 *===================================================================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_setchparms 
(
   CEI_UINT32 cardnum, 
   CEI_UINT32 channel, 
   CEI_INT32 TransRcv, 
   pAR_CHANNEL_PARMS def
)
{
   CDEV_API_RET_TYPE status = ARS_NORMAL; /* function return status          */
   CEI_INT32 locLongValue;                /* temp 32bit variable             */
   CDEV_BOARD_TYPE board = (CEI_INT16)cardnum;

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      return ARS_INVBOARD;
   }
#endif

   if ( (TransRcv == ARU_TRANSMITTER) && (channel < deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX]) ) {
      def->ChannelType = CHAN_ARINC429;
   }
   else if ( (TransRcv == ARU_RECEIVER) && (channel < deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) ) {
      def->ChannelType = CHAN_ARINC429;
   }
   else if (deviceGlobal[board]->alternateJconfiguration == CEIDEV_ALT_CONFIGURATION_717) {
      def->ChannelType = CHAN_ARINC717;
   }
   else {
      def->ChannelType = CHAN_NOT_IMPLEMENTED;
   }

   if (def->ChannelType == CHAN_ARINC429) {
      if (TransRcv == ARU_TRANSMITTER) {
         /* first assign the specified bus speed */
         if (def->BaudRate > ARU_SPEED_LOW) {
            if ((deviceGlobal[board]->configuration == CEIDEV_CONFIG_RAR15) ||
                (deviceGlobal[board]->configuration == CEIDEV_CONFIG_RAR15XT)) {
               locLongValue = (CEIDEV_20MHZ_REF_CLOCK / (CEI_INT32)def->BaudRate) - 2;
            }
            else {
               locLongValue = (CEIDEV_16MHZ_REF_CLOCK / (CEI_INT32)def->BaudRate) - 2;
            }
         }
         else {
            locLongValue = def->BaudRate;
         }
   
         status = ar_set_device_config (board, (CEI_INT16)channel, ARU_TX_BITRATE, (CEI_INT16)locLongValue);
         
         /* next assign the channel parity */
         if (status == ARS_NORMAL) {
            status = ar_set_device_config (board, (CEI_INT16)channel, ARU_TX_PARITY, (CEI_INT16)def->Parity);
         }
      }
      else if (TransRcv == ARU_RECEIVER) {
         /* first assign the specified bus speed */
         if (def->BaudRate > ARU_SPEED_LOW) {
            if ((deviceGlobal[board]->configuration == CEIDEV_CONFIG_RAR15) ||
                (deviceGlobal[board]->configuration == CEIDEV_CONFIG_RAR15XT)) {
               locLongValue = (CEIDEV_20MHZ_REF_CLOCK / (CEI_INT32)def->BaudRate) - 2;
            }
            else {
               locLongValue = (CEIDEV_16MHZ_REF_CLOCK / (CEI_INT32)def->BaudRate) - 2;
            }
         }
         else {
            locLongValue = def->BaudRate;
         }
      
         status = ar_set_device_config (board, (CEI_INT16)channel, ARU_RX_BITRATE, (CEI_INT16)locLongValue);
         
         /* next assign the channel parity */
         if (status == ARS_NORMAL) {
            status = ar_set_device_config (board, (CEI_INT16)channel, ARU_RX_PARITY, (CEI_INT16)def->Parity);
         }
         /* next assign the receiver wrap selection */
         if (status == ARS_NORMAL) {
            status = ar_set_device_config (board, (CEI_INT16)channel, ARU_RECV_MODE, (CEI_INT16)def->InternalWrap);
         }
         /* next assign the receiver storage mode selection */
         if (status == ARS_NORMAL) {
            status = ar_set_storage_mode (board, (CEI_INT16)def->OperateMode);
         }
      }
      else {
         status = ARS_INVARG;
      }
   }
#ifdef INCLUDE_ARINC_717
   else if (def->ChannelType == CHAN_ARINC717) {
      if (TransRcv == ARU_TRANSMITTER) {
         /* assign the specified bus speed */
         locLongValue = (def->BaudRate >> 8) & 0x0000000F;
         status = ar_set_573_config (board, ARU_TX_BITRATE, locLongValue);
         
         /* next assign the channel encoding */
         if (status == ARS_NORMAL) {
            if (def->HWEncoding == ARU_717_H_BI_PHASE) {
               status = ar_set_573_config (board, ARU_573_TX_BPRZ_SELECT, AR_OFF);
               status = ar_set_573_config (board, ARU_573_TX_HBP_SELECT, AR_ON);
            }
            else {
               status = ar_set_573_config (board, ARU_573_TX_BPRZ_SELECT, AR_ON);
               status = ar_set_573_config (board, ARU_573_TX_HBP_SELECT, AR_OFF);
            }
            /* last enable the channel buffer */
            status = ar_set_573_config ( board, ARU_TX_FIFO_ENABLE, AR_ON );
         }
      }
      else if (TransRcv == ARU_RECEIVER) {
         /* assign the specified bus speed */
         locLongValue = (def->BaudRate >> 8) & 0x0000000F;
         status = ar_set_573_config (board, ARU_RX_BITRATE, locLongValue);
         
         /* next assign autosync selection */
         if (status == ARS_NORMAL) {
            locLongValue = (def->AutoSync == ARU_717_AUTOSYNC) ? AR_ON : AR_OFF;
            status = ar_set_573_config (board, ARU_573_RX_AUTO_DETECT, locLongValue);
         }
         /* next assign the receiver wrap selection */
         if (status == ARS_NORMAL) {
            status = ar_set_573_config (board, ARU_RECV_MODE, (CEI_INT16)def->InternalWrap);
         }
         /* next assign the receiver storage mode selection */
         if (status == ARS_NORMAL) {
            status = ar_set_storage_mode (board, (CEI_INT16)def->OperateMode);
         }

         if (status == ARS_NORMAL) {
            locLongValue = (def->HWEncoding == ARU_717_BIPOLAR) ? AR_ON : AR_OFF;
            status = ar_set_573_config (board, ARU_573_RX_BPRZ_SELECT, locLongValue);
         }

         if (status == ARS_NORMAL) {
            status = ar_set_573_config (board, ARU_573_SYNC_WORD1, (CEI_INT16)def->Sync1);
            status = ar_set_573_config (board, ARU_573_SYNC_WORD2, (CEI_INT16)def->Sync2);
            status = ar_set_573_config (board, ARU_573_SYNC_WORD3, (CEI_INT16)def->Sync3);
            status = ar_set_573_config (board, ARU_573_SYNC_WORD4, (CEI_INT16)def->Sync4);
         }
         /* last enable the channel buffer */
         status = ar_set_573_config ( board, ARU_RX_FIFO_ENABLE, AR_ON );
      }
      else {
         status = ARS_INVARG;
      }
   }
#endif

   return (CEI_INT32)status;
}

/*===========================================================================*
 * ENTRY POINT:             A R _ S E T _ R A W _ M O D E
 *===========================================================================*
 *
 * FUNCTION:    Enable/Disable ARINC 429 Parity Mode
 * 
 *===========================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_set_raw_mode 
( 
   CDEV_BOARD_TYPE board,      /* (input) board to reference (0 to 127)      */
   CDEV_PARM_SSI_TYPE direction, /* (input) TRANSMIT_CHANNEL/RECEIVE_CHANNEL */
   CDEV_CHAN_TYPE channel,     /* (input) channel number (0 to 7)            */
   CDEV_PARM_SSI_TYPE control  /* (input) AR_ON / AR_OFF                     */
)
{
   CEI_INT16 item = ARU_TX_PARITY;      /* Rx/Tx parity select               */
   CEI_INT16 value = ARU_PARITY_NONE;   /* Odd/None parity enable            */
   CDEV_API_RET_TYPE status = ARS_NORMAL;       /* function return status    */
   
   if ( control == AR_ON ) {
      value = ARU_PARITY_NONE;
   }
   else if ( control == AR_OFF ) {
      value = ARU_PARITY_ODD;
   }
   else {
      status = ARS_INVARG;
   }

   if ( (direction == TRANSMIT_CHANNEL) || (direction == ARU_XMIT) ) {
      item = ARU_TX_PARITY;
   }
   else if ( (direction == RECEIVE_CHANNEL) || (direction == ARU_RECV) ) {
      item = ARU_RX_PARITY;
   }
   else {
      status = ARS_INVARG;
   }

   /* Command the respective channel to the specified parity. */
   if (status == ARS_NORMAL) {
      status = ar_set_device_config ( board, channel, item, value );
   }
   
   return status;
}


/*===========================================================================*
 * ENTRY POINT:         A R _ X M I T _ S Y N C
 *===========================================================================*
 *
 * FUNCTION:    This routine is used to wait for all of the data in the
 *              transmit queue to be sent.  It is useful for a program
 *              that wants to reset the slave, but needs to wait for it
 *              to complete transmission before halting it.  It will
 *              self-terminate if the transmit FIFO hasn't completed
 *              flushing all messages after 12 seconds.
 *
 *===========================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_xmit_sync 
(
   CDEV_BOARD_TYPE board,            /* (input)  board to reference (0 to 127)     */
   CDEV_CHAN_TYPE channel           /* (input)  transmit channel to use (0 to N)  */
) 
{
   CEI_INT32 localChannel;     /* channel offset into the label filter table */
   CEI_UINT32 channelStatus;            /* TX channel status register value  */
   CEI_UINT32 globalEnableRegData;      /* contents of the Global Enable reg */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   CDEV_API_RET_TYPE status = ARS_NORMAL;  /* function return status         */
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */
   AR_TIMETAG_TYPE boardTime;
   AR_TIMETAG_TYPE exitTime;

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II)  {
      status = ARS_INVBOARD;
   }
   else if ((CEI_UINT32)channel >= deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX]) {
      status = ARS_INVARG;
   }
   else {
#endif
      status = ar_get_time (board, AR_TIMETAG_INT_USEC_64BIT, &exitTime); 
      exitTime.timeTag += 12000000;

      localChannel = (CEI_INT32)channel + (CEI_INT32)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
      boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS +
                                  (localChannel * CEIDEV_CHANNEL_REGION_SIZE));

#ifndef TARGET_EMBEDDED
      /* if multi-thread protection is active, acquire a lock on the host interface */
      if (device[board].multiThreadProtect != AR_CONC_NONE) {
         accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
      }
      if (accessStatus == ARS_NORMAL) {
#endif
         accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], &globalEnableRegData);
         if ((globalEnableRegData & CEIDEV_GLOBAL_ENABLE_MASK) == 0) {
            status = ARS_NODATA;
         }
         else {
            accessStatus = cei_read_32bit (board, boardPointer, &channelStatus);
            channelStatus = (channelStatus & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT;
            
            /* read the channel status continuously until it indicates the FIFO is empty */
            while (channelStatus > 0) {
               accessStatus = cei_read_32bit (board, boardPointer, &channelStatus);
               channelStatus = (channelStatus & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT;
               status = ar_get_time (board, AR_TIMETAG_INT_USEC_64BIT, &boardTime);
               if (boardTime.timeTag > exitTime.timeTag) {
                  accessStatus = ARS_NOSYNC;
                  break;
               }
            }
         }
#ifndef TARGET_EMBEDDED
         if (device[board].multiThreadProtect != AR_CONC_NONE) {
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
         }
      }
      status = (CDEV_API_RET_TYPE)accessStatus;
   }
#endif
   
   return status;
}

/*===========================================================================*  
 * CEI-x20 API Functions (NO FUNCTIONALITY) Providing Link-Compatibility ONLY
 *===========================================================================*/             

#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (INTEGRITY_PCI_PPC)
EXPORT32 CEI_INT32 DLL_EXPORTED ar_dump_dp (CDEV_BOARD_TYPE board, pCEI_CHAR file_name, pCEI_CHAR msg) {board=0; msg[0]=file_name[0]; return ARS_NO_HW_SUPRT;}
EXPORT32 CEI_INT16 DLL_EXPORTED ar_force_version (CDEV_BOARD_TYPE board, CEI_INT16 number) {number=(CEI_INT16)board; return ARS_NO_HW_SUPRT;}
EXPORT32 CEI_VOID  DLL_EXPORTED ar_formatarinclabel (CEI_UINT32 label_data, pCEI_CHAR labelstr) {label_data=0; labelstr[0]='0';}
EXPORT32 CEI_VOID  DLL_EXPORTED ar_formatbinarylabel (CEI_UINT32 label_data, CEI_INT32 trunc_label, CEI_CHAR *labelstr) {trunc_label=label_data; labelstr[0]='0';}
EXPORT32 CEI_VOID  DLL_EXPORTED ar_get_errorLV (pCEI_CHAR string, CEI_INT16 status) {string[0]='0'; status=0;}
EXPORT32 CEI_INT32 DLL_EXPORTED ar_get_harris (CDEV_BOARD_TYPE board, CEI_INT16 item) {item=(CEI_INT16)board; return ARS_NO_HW_SUPRT;}
EXPORT32 CEI_INT16 DLL_EXPORTED ar_int_set (CDEV_BOARD_TYPE board, CEI_INT32 int_num) {int_num=(CEI_INT32)board; return ARS_NORMAL;}
EXPORT32 CEI_INT16 DLL_EXPORTED ar_int_slave (CDEV_BOARD_TYPE board) {board=0; return ARS_NO_HW_SUPRT;}
EXPORT32 CEI_INT16 DLL_EXPORTED ar_loadslv_init_discrete (CDEV_BOARD_TYPE board, CEI_UINT32 base_seg, CEI_INT32 base_port, CEI_UINT16 ram_size, CEI_UINT32 disc_init)  {base_seg=(CEI_UINT32)board+(CEI_UINT32)base_port+(CEI_UINT32)ram_size+(CEI_UINT32)disc_init; return ARS_NO_HW_SUPRT;}
EXPORT32 CEI_INT16 DLL_EXPORTED ar_msg_control (CDEV_BOARD_TYPE board, CEI_INT16 control) {control=(CEI_INT16)board; return ARS_NORMAL;}
EXPORT32 CEI_INT16 DLL_EXPORTED ar_putword2x16 (CDEV_BOARD_TYPE board, CDEV_CHAN_TYPE channel, CEI_INT16 lsw, CEI_INT16 msw) {channel=board; lsw=msw; return ARS_NO_HW_SUPRT;}
EXPORT32 CEI_VOID  DLL_EXPORTED ar_reformat (pCEI_VOID lsw, pCEI_VOID msw) {lsw = msw;}
EXPORT32 CEI_INT16 DLL_EXPORTED ar_set_arinc_config (CDEV_BOARD_TYPE board, CDEV_CHAN_TYPE channel, CEI_INT16 item, CEI_UINT32 value) {value=item+board+channel; return ARS_NO_HW_SUPRT;}
EXPORT32 CEI_INT16 DLL_EXPORTED ar_set_harris (CDEV_BOARD_TYPE board, CEI_INT16 item, CEI_UINT32 value) {value=item+board; return ARS_NO_HW_SUPRT;}
EXPORT32 CEI_INT16 DLL_EXPORTED ar_timetag_control (CDEV_BOARD_TYPE board, CEI_INT16 flag) {flag=(CEI_INT16)board; return ARS_NORMAL;}
EXPORT32 CEI_VOID  DLL_EXPORTED ar_unformat (pCEI_INT16 lsword, pCEI_INT16 msword) {*lsword = *msword;}

EXPORT32 CEI_INT32 DLL_EXPORTED ar_flush_receiver
(
   CEI_UINT32 board,           /* (input)  board to reference (0 to 127)     */
   CEI_UINT32 channel          /* (input)  recvr channel to use (0 to 31)    */
)
{
   CEI_UINT32 bufferCount = 0;       /* Receive channel buffer message count */
   CEI_UINT32 data;                    /* local storage for buffered message */
   CEI_UINT32 dataOffset;             /* local offset to channel memory area */
   CEI_UINT32 memOffset;              /* local offset to channel memory area */
   CEI_UINT32 messageCount = 0;                /* local message read counter */
#ifndef TARGET_EMBEDDED
   CEI_INT32 lockStatus = ARS_NORMAL; /* status of board access lock acquis. */
#endif
   CEI_INT32 accessStatus = ARS_NORMAL;/* status of device read/write access */
   CEI_INT32 status = ARS_NORMAL;                  /* function return status */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr = NULL;  /* board local structure */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr = NULL; /* board global structure */

   deviceGlobalStructPtr = deviceGlobal[board];
   deviceLocalStructPtr = &device[board];

#ifndef TARGET_EMBEDDED
   /* if multi-thread protection requested, acquire semaphore/mutex-locking */
   if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
      lockStatus = cei_acquire_thread_lock(LOCK_IDENT);
   }
   if (lockStatus == ARS_NORMAL) {
#endif
      memOffset = CEIDEV_CHANNEL_REGION_START + ((channel + deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]) * CEIDEV_CHANNEL_REGION_SIZE);
      dataOffset = memOffset + CEIDEV_CHANNEL_BUFFER_3;
      
      /* read the FIFO count in the status register */
      accessStatus = cei_read_32bit ((CEI_INT16)board, &deviceLocalStructPtr->devicePtr[memOffset], &bufferCount);
      
      if (accessStatus == ARS_NORMAL) {
         bufferCount = (bufferCount & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT;
         
         /* flush by reading the 32-bit data word buffer location for each message in the buffer */
         while (messageCount++ < bufferCount) {
            accessStatus = cei_read_32bit ((CEI_INT16)board, &deviceLocalStructPtr->devicePtr[dataOffset], &data);
         }
      }
       
#ifndef TARGET_EMBEDDED
      /* if multi-thread protection is active, free any acquired lock */
      if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
         lockStatus = cei_release_thread_lock(LOCK_IDENT);
      }
   }    
#endif
      
   /* return any bad status indication */
   if (accessStatus != ARS_NORMAL) {
      status = accessStatus;
   }
#ifndef TARGET_EMBEDDED
   else if (lockStatus != ARS_NORMAL) {
      status = lockStatus;
   }
#endif
   
   return (status);
}

#endif

