/*====================================================================================================================*
 * FILE:                                     C D E V _ A P I _ P L X _ P G M. C                    
 *====================================================================================================================*
 *
 *   COPYRIGHT 2016 - 2019 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED
 *   ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH THE INCLUSION
 *   OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY OTHER COPIES THEREOF
 *   MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE
 *   AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS, INC. 
 *
 *====================================================================================================================*
 *
 * FUNCTION:    Local API Routines used for programming a PLX-based CEI-x30 ARINC device.
 *
 * DESCRIPTION: These routines are provided specificically for programming PLX-based CEI-x30 ARINC devices using
 *              using the C programming language.  
 *
 *              Most routines utilize a board number parameter, zero-based on the Device ID/Board Number assigned 
 *              during the installation of CEI-x30-SW distribution.  Channel parameter values are all zero-based.
 *
 * INTERNAL UTILITY ROUTINES:
 *
 *    a429_utl_enh                      
 *    a429_utl_enh_check                
 *    a429_utl_program_xilinx_9030      
 *    a429_utl_program_xilinx_9030_32bit
 *    a429_utl_program_xilinx_9056      
 *    a429_utl_program_xilinx_9056_32bit
 *    a429_utl_plx_eeprom_read          
 *    a429_utl_plx_eeprom_write_enable  
 *    a429_utl_plx_eeprom_write_disable 
 *    a429_utl_plx_eeprom_write         
 *
 *--------------------------------------------------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date    By   Ver. Comments
 * -------- ----- ---- -----------------------------------------------------------------------------------------------
 * 10/18/16  BDW  4.90 Initial version.
 *
 *====================================================================================================================*/ 
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
   #include <unistd.h>
   #include <fcntl.h>
   #include <string.h>
   #include <pthread.h>
   #include <errno.h>
   #include <sys/stat.h> 
   #include <sys/shm.h>
   #include <sys/sem.h>
   #include <sys/syscall.h>
   #define VOLATILE volatile
#elif defined (VXW_PCI_X86) || defined (VXW_PCI_PPC)
   #include <stdio.h>
   #include <stdlib.h>
   #include <unistd.h>
   #define VOLATILE 
#elif defined (NIVISA_PXI)
   #ifdef LABVIEW_RT
      #include <ansi_c.h> 
   #endif
   #include <visa.h>  
   #define VOLATILE 
#elif defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
   #include <string.h>
   #define VOLATILE 
#else
   #define VOLATILE volatile
#endif

#include "cdev_glb.h"
#ifdef DEBUG_PROGRAMMING
extern CEI_UINT32 programCsum;              /* device programming checksum       */
#endif   

extern VOLATILE ARINC_LOCAL_TYPE device[MAX_BOARDS];
extern ARINC_GLOBAL_TYPE * deviceGlobal[MAX_BOARDS];


/*===================================================================================================================*
 * ENTRY POINT:                       A 4 2 9 _ U T L _ P L X _ E E P R O M _ R E A D
 *===================================================================================================================*
 *
 * FUNCTION:    Read and return the contents of the PLX EEPROM
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board      -- (input) board to reference (0 to 127)
 * 
 *              CEI_INT32 pciInterfaceType -- (input) PLX Part (9030 or 9056)
 * 
 *              pCEI_UINT32 pciConfigPtr   -- (input) ptr to the BAR0 memory region
 * 
 *              pCEI_UINT32 promValues     -- (output) ptr to array to store values
 * 
 * RETURN VAL:  CEI_INT32 -- ARS_NORMAL     Normal succesful completion.
 *                           other values   Platform specific device access failures
 * 
 * GLOBALS:     none
 *
 * DESCRIPTION: This function reads the PLX PROM and returns the values.
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_plx_eeprom_read 
(
   CDEV_BOARD_TYPE board,                                             /* (input)  board to reference (0 to 127)      */
   CEI_INT32   pciInterfaceType,                                      /* (input)  PLX Part Type (9030 or 9056)       */
   pCEI_UINT32 pciConfigPtr,                                          /* (input)  ptr to the BAR0 memory region      */
   pCEI_UINT32 promValues                                             /* (output) ptr to array to store values       */
)
{
   CEI_UINT32 copy;
   CEI_UINT32 readValue;
   CEI_UINT32 x;
   CEI_UINT32 promData;
   CEI_UINT32 addr;
   pCEI_UINT32 promValuesPtr = promValues;
   CEI_INT32  status = ARS_NORMAL;                                                        /* function return status  */

   promData = 0;
   status = READ_32_BIT (board, pciConfigPtr, &copy);

   /* (1) Clear the Input Enable Bits */
   copy &= ~EECS;
   copy &= ~EESK;
   copy &= ~EEDI;
   if (pciInterfaceType == PLX_9056)  {
      copy &= ~EEDS;
   }
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (2) Set EECS high */
   copy |=  EECS;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (3) Toggle EESK high, then low */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (4) Set the EEDI bit high (Start bit) */
   copy |=  EEDI;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (5) Toggle EESK high, then low */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (6) Toggle EESK high, then low */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (7) Clear EEDI */
   copy &= ~EEDI;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (8) Toggle EESK high, then low */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (9) Toggle EESK high, then low, 8 times (serial EEPROM address 0) */
   for (x = 0; x < 8; x++) {
      copy |=  EESK;
      status = WRITE_32_BIT (board, pciConfigPtr, copy);
      copy &= ~EESK;
      status = WRITE_32_BIT (board, pciConfigPtr, copy);
   }

   /* (10) Set the EEDO Input Enable to 1 to float the EEDO input for reading */
   copy |= (pciInterfaceType == PLX_9056) ? EEDS : EEDO;

   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (11) Toggle EESK high, then low, 16 times, (12) reading in a bit after each clock */
   for (addr = 0; addr < PLX_EEPROM_SIZE; addr++) {
      promData = 0;
      for (x = 0; x <= 15; x++) {
         copy |=  EESK;
         status = WRITE_32_BIT (board, pciConfigPtr, copy);
         copy &= ~EESK;
         status = WRITE_32_BIT (board, pciConfigPtr, copy);
         status = READ_32_BIT (board, pciConfigPtr, &readValue);
         if (readValue & EEDO) {
            promData |= 0x0001;
         }
         else {
            promData &= 0xFFFE;
         }
         if (x!=15) {
            promData = promData << 1;
         }
         else {
            *promValuesPtr = promData;
            promValuesPtr++;
         }
      }
   }

   /* (13) Clear the EECS bit */
   copy &= ~EECS;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (14) Toggle EESK high, then low.  The read is complete */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:               A 4 2 9 _ U T L _ P L X _ E E P R O M _ W R I T E _ E N A B L E
 *===================================================================================================================*
 *
 * FUNCTION:    Enable write access to the PLX PROM
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board      -- (input) board to reference (0 to 127)
 * 
 *              CEI_INT32 pciInterfaceType -- (input) PLX Part (9030 or 9056)
 * 
 *              pCEI_UINT32 pciConfigPtr   -- (input) ptr to the BAR0 memory region
 * 
 * RETURN VAL:  CEI_INT32 -- ARS_NORMAL     Normal succesful completion.
 *                           other values   Platform specific device access failures
 * 
 * GLOBALS:     none
 *
 * DESCRIPTION: This function enables write access to the PLX PROM.
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_plx_eeprom_write_enable 
(
   CDEV_BOARD_TYPE board,                                             /* (input)  board to reference (0 to 127)      */
   CEI_INT32   pciInterfaceType,                                      /* (input)  PLX Part Type (9030 or 9056)       */
   pCEI_UINT32 pciConfigPtr                                           /* (input)  ptr to the BAR0 memory region      */
)
{
   CEI_UINT32 copy;
   CEI_INT32  x;
   CEI_INT32  y[11] = {1,0,0,1,1,0,0,0,0,0,0};
   CEI_INT32  status = ARS_NORMAL;                                                        /* function return status  */

   status = READ_32_BIT (board, pciConfigPtr, &copy);

   /* (1) Clear the Input Enable Bits */
   copy &= ~EECS;
   copy &= ~EESK;
   copy &= ~EEDI;
   if (pciInterfaceType == PLX_9056)  {
      copy &= ~EEDS;
   }
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (2) Set EECS high */
   copy |=  EECS;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (3) Toggle EESK high, then low */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (4) Clock in the 11 bits (Start bit, cmd bits, etc.) */

   for (x = 0; x <= 10; x++) {
      if (y[x] == 0) {
         copy &= ~EEDI;
      }
      else {
         copy |=  EEDI;
      }
      status = WRITE_32_BIT (board, pciConfigPtr, copy);
      copy |=  EESK;
      status = WRITE_32_BIT (board, pciConfigPtr, copy);
      copy &= ~EESK;
      status = WRITE_32_BIT (board, pciConfigPtr, copy);
   }


   /* (5) Clear the EECS bit */
   copy &= ~EECS;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (6) Toggle EESK high, then low.  The command is complete */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                A 4 2 9 _ U T L _ P L X _ E E P R O M _ W R I T E _ D I S A B L E
 *===================================================================================================================*
 *
 * FUNCTION:    Disable write access to the PLX PROM
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board    -- (input) board to reference (0 to 127)
 * 
 *              CEI_INT32 pciInterfaceType -- (input) PLX Part (9030 or 9056)
 * 
 *              pCEI_UINT32 pciConfigPtr -- (input) ptr to the BAR0 memory region
 * 
 * RETURN VAL:  CEI_INT32 -- ARS_NORMAL     Normal succesful completion.
 *                           other values   Platform specific device access failures
 * 
 * GLOBALS:     none
 *
 * DESCRIPTION: This function disables write access to the PLX PROM.
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_plx_eeprom_write_disable 
(
   CDEV_BOARD_TYPE board,                                             /* (input)  board to reference (0 to 127)      */
   CEI_INT32   pciInterfaceType,                                      /* (input)  PLX Part Type (9030 or 9056)       */
   pCEI_UINT32 pciConfigPtr                                           /* (input)  ptr to the BAR0 memory region      */
)
{
   CEI_UINT32 copy;
   CEI_INT32  x;
   CEI_INT32  y[11] = {1,0,0,0,0,0,0,0,0,0,0};
   CEI_INT32  status = ARS_NORMAL;                                                         /* function return status */

   status = READ_32_BIT (board, pciConfigPtr, &copy);

   /* (1) Clear the Input Enable Bits */
   copy &= ~EECS;
   copy &= ~EESK;
   copy &= ~EEDI;
   if (pciInterfaceType == PLX_9056)  {
      copy &= ~EEDS;
   }
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (2) Set EECS high */
   copy |=  EECS;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (3) Toggle EESK high, then low */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (4) Clock in the 11 bits (Start bit, cmd bits, etc.) */

   for (x = 0; x <= 10; x++) {
      if (y[x] == 0) { 
         copy &= ~EEDI;
      }
      else {
         copy |=  EEDI;
      }
      status = WRITE_32_BIT (board, pciConfigPtr, copy);
      copy |=  EESK;
      status = WRITE_32_BIT (board, pciConfigPtr, copy);
      copy &= ~EESK;
      status = WRITE_32_BIT (board, pciConfigPtr, copy);
   }


   /* (5) Clear the EECS bit */
   copy &= ~EECS;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (6) Toggle EESK high, then low.  The command is complete */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   return status;
}

/*===================================================================================================================*
 * ENTRY POINT:                    A 4 2 9 _ U T L _ P L X _ E E P R O M _ W R I T E
 *===================================================================================================================*
 *
 * FUNCTION:    Write a value to the specified PLX PROM location
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board      -- (input) board to reference (0 to 127)
 * 
 *              CEI_INT32 pciInterfaceType -- (input) PLX Part (9030 or 9056)
 * 
 *              pCEI_UINT32 pciConfigPtr   -- (input) ptr to the BAR0 memory region
 * 
 *              CEI_UINT32 address         -- (input) offset into the PLX PROM to write
 * 
 *              CEI_UINT32 data            -- (input) value to write to the PLX PROM
 * 
 * RETURN VAL:  CEI_INT32 -- ARS_NORMAL     Normal succesful completion.
 *                           other values   Platform specific device access failures
 * 
 * GLOBALS:     none
 *
 * DESCRIPTION: This function writes a value to the specified PLX PROM location.
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_plx_eeprom_write 
(
   CDEV_BOARD_TYPE board,                                             /* (input)  board to reference (0 to 127)      */
   CEI_INT32   pciInterfaceType,                                      /* (input)  PLX Part Type (9030 or 9056)       */
   pCEI_UINT32 pciConfigPtr,                                          /* (input)  ptr to the BAR0 memory region      */
   CEI_UINT32 address,                                                /* (input)  offset into the PLX PROM to write  */
   CEI_UINT32 data                                                    /* (input)  value to write to the PLX PROM     */
)
{
   CEI_UINT32 copy;
   CEI_INT32  x;
   CEI_INT32  y[27] = {1, 0,1, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0};
   CEI_INT32  status = ARS_NORMAL;                                                         /* function return status */

   if (address & 0x0080)  y[3]  = 1;
   if (address & 0x0040)  y[4]  = 1;
   if (address & 0x0020)  y[5]  = 1;
   if (address & 0x0010)  y[6]  = 1;
   if (address & 0x0008)  y[7]  = 1;
   if (address & 0x0004)  y[8]  = 1;
   if (address & 0x0002)  y[9]  = 1;
   if (address & 0x0001)  y[10] = 1;
   if (data    & 0x8000)  y[11] = 1;
   if (data    & 0x4000)  y[12] = 1;
   if (data    & 0x2000)  y[13] = 1;
   if (data    & 0x1000)  y[14] = 1;
   if (data    & 0x0800)  y[15] = 1;
   if (data    & 0x0400)  y[16] = 1;
   if (data    & 0x0200)  y[17] = 1;
   if (data    & 0x0100)  y[18] = 1;
   if (data    & 0x0080)  y[19] = 1;
   if (data    & 0x0040)  y[20] = 1;
   if (data    & 0x0020)  y[21] = 1;
   if (data    & 0x0010)  y[22] = 1;
   if (data    & 0x0008)  y[23] = 1;
   if (data    & 0x0004)  y[24] = 1;
   if (data    & 0x0002)  y[25] = 1;
   if (data    & 0x0001)  y[26] = 1;

   status = READ_32_BIT (board, pciConfigPtr, &copy);

   /* (1) Clear the Input Enable Bits */
   copy &= ~EECS;
   copy &= ~EESK;
   copy &= ~EEDI;
   if (pciInterfaceType == PLX_9056)  {
      copy &= ~EEDS;
   }
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (2) Set EECS high */
   copy |=  EECS;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (3) Toggle EESK high, then low */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (4) Clock in the 27 bits (Start bit, cmd bits, addr, data) */

   for (x = 0; x <= 26; x++) {
      if (y[x] == 0) { 
         copy &= ~EEDI;
      }
      else {
         copy |=  EEDI;
      }
      status = WRITE_32_BIT (board, pciConfigPtr, copy);
      copy |=  EESK;
      status = WRITE_32_BIT (board, pciConfigPtr, copy);
      copy &= ~EESK;
      status = WRITE_32_BIT (board, pciConfigPtr, copy);
   }


   /* (5) Clear the EECS bit */
   copy &= ~EECS;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (6) Toggle EESK high, then low.  The command is complete */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   ar_sleep (15);  /* Write time */
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                               A 4 2 9 _ U T L _ E N H
 *===================================================================================================================*
 *
 * FUNCTION:    Open a session to the board and update the EEPROM for CEI-x30 Enhanced firmware operation
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board  -- (input) board to reference (0 to 127)
 *
 *              CEI_INT32 action -- (input) operation to perform:
 *                                          PLX_ENHANCED_CHECK   - check to see if BAR2 is 512Kb
 *                                          PLX_ENHANCED_UPDATE  - check to see if BAR2 is 512Kb and
 *                                                                 reprogram it if it is only 4Kb
 *                                          PLX_ORIGINAL_RESTORE - if BAR2 is not 4kb, this option will result in 
 *                                                                 reprogramming BAR2 to the original x30 4Kb size
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL      Success updating or already done
 *                                ARS_INVBOARD    Invalid board parameter
 *                                ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                                ARS_FAILURE     Invalid device (not x30)
 *                                ARS_MEMWRERR    PROM update failed
 *                                ARS_DRIVERFAIL  could not access the CEI_Install library or
 *                                                the Driver Open/Close Session invocation failed
 * 
 * GLOBALS:     device[board]
 * 
 * DESCRIPTION: This routine is a special utility routine to upgrade a CEI-x30 board for use with the enhanced 
 *              firmware and API.  It modifies the PLX PROM BAR2 size from 4Kb to 512Kb.
 *
 *===================================================================================================================*/
CDEV_API_RET_TYPE a429_utl_enh 
(
   CDEV_BOARD_TYPE board,                                             /* (input)  board to reference (0 to 127)      */
   CEI_INT32 action                                                   /* (input)  operation (check/update/restore)   */
)
{
   CEI_INT32   status = ARS_NORMAL;                                              /* function return status           */
   CEI_INT32   tempStatus = ARS_NORMAL;                                          /* function return status           */
   CEI_INT32   plxType = 0;                                                      /* type of PLX PCI intfc 9056/9030  */
   CEI_INT32   bar2SizeWord1 = 0;                                                /* PCI Cfg offset for PCI BAR2 Size */
   CEI_INT32   bar2SizeWord2 = 0;                                                /* PCI Cfg offset for PCI BAR2 Size */
   CEI_INT32   vendorIdent;                                                      /* PCI Vendor ID for opened device  */
   CEI_INT32   deviceIdent;                                                      /* PCI Device ID for opened device  */
   CEI_UINT32  promContent[PLX_EEPROM_SIZE];                                     /* contents of PLX EEPROM           */
   pCEI_UINT32 pciConfigMemPtr = 0;                                              /* BAR1 offset for PLX EEPROM       */

   /* open a session with the specified device */
   status = cei_acquire_device (board, (ARINC_LOCAL_TYPE *)&device[board]);
   if (status == ARS_NORMAL) {
      /* force exclusive access to the board for the duration of this routine */
#if !defined (LABVIEW_RT) && !defined (TARGET_EMBEDDED)
      status = cei_acquire_thread_lock(LOCK_IDENT);
#endif      
      if (status == ARS_NORMAL) {
         /* assure this is a CEI-x30 board before continuing */
         status = cei_acquire_device_identity ((CEI_INT32)board, &vendorIdent, &deviceIdent);
         if ( (status != ARS_FAILURE) && (vendorIdent == PCI_X30_MANUF_ID) ) {
            switch (deviceIdent) {
               case PCI_DEVICE_ID_530:
               case PCI_DEVICE_ID_630:
               case PCI_DEVICE_ID_830:
               case PCI_DEVICE_ID_830RX:
               case PCI_DEVICE_ID_A30:
               case PCI_DEVICE_ID_830X820:
               case PCI_DEVICE_ID_RCEI830A:
                  plxType       = PLX_9056;
                  bar2SizeWord1 = PLX_9056_BAR2_SIZE_WORD_1;
                  bar2SizeWord2 = PLX_9056_BAR2_SIZE_WORD_2;
                  pciConfigMemPtr = device[board].baseAddr + PLX_CONFIG_MEM_CONTROL_REG;
                  break;
               case PCI_DEVICE_ID_430:
               case PCI_DEVICE_ID_430A:
               case PCI_DEVICE_ID_RAREC:
                  plxType       = PLX_9030;
                  bar2SizeWord1 = PLX_9030_BAR2_SIZE_WORD_1;
                  bar2SizeWord2 = PLX_9030_BAR2_SIZE_WORD_2;
                  pciConfigMemPtr = device[board].baseAddr + PLX9030_PCI_TARGET_RESPONSE;
                  break;
               case PCI_DEVICE_ID_RAR_PCIE:
               case PCI_DEVICE_ID_RAR_XMC:
               case PCI_DEVICE_ID_RAR15:
               case PCI_DEVICE_ID_RAR15XT:
               case PCI_DEVICE_ID_RAR_MPCIE:
                  status = ARS_INVBOARD;
                  break;
               default:
                  status = ARS_FAILURE;
            }

            if (status == ARS_NORMAL) {
               /* read the current content of the PROM */
               status = a429_utl_plx_eeprom_read (board, plxType, pciConfigMemPtr, promContent);
               /* if the PROM is currently programmed for the old CEI-x30 BAR2 size, update it to 512Kb */
               if ((promContent[bar2SizeWord1] != PLX_BAR2_ENHANCED_SIZE_1) || (promContent[bar2SizeWord2] != PLX_BAR2_ENHANCED_SIZE_2)) {
                  if (action == PLX_ENHANCED_CHECK) {
                     status = ARS_HW_CONSISTENCY;
                  }
                  else if (action == PLX_ENHANCED_UPDATE) {
                     status = a429_utl_plx_eeprom_write_enable(board, plxType, pciConfigMemPtr);
                     ar_sleep(52);
                     status = a429_utl_plx_eeprom_write(board, plxType, pciConfigMemPtr, bar2SizeWord1, PLX_BAR2_ENHANCED_SIZE_1);
                     ar_sleep(52);
                     status = a429_utl_plx_eeprom_write(board, plxType, pciConfigMemPtr, bar2SizeWord2, PLX_BAR2_ENHANCED_SIZE_2);
                     ar_sleep(52);
                     status = a429_utl_plx_eeprom_write_disable(board, plxType, pciConfigMemPtr);
                     ar_sleep(52);  
      
                     /* read/verify the updated content of the PROM */
                     status = a429_utl_plx_eeprom_read (board, plxType, pciConfigMemPtr, promContent);
                     if ((promContent[bar2SizeWord1] != PLX_BAR2_ENHANCED_SIZE_1) || (promContent[bar2SizeWord2] != PLX_BAR2_ENHANCED_SIZE_2)) {
                        status = ARS_MEMWRERR;
                     }
                     else {
                        status = ARS_NORMAL;
                     }
                  }
                  else {
                     status = ARS_INVARG;
                  }
               }
               else if (action == PLX_ORIGINAL_RESTORE) {
                  status = a429_utl_plx_eeprom_write_enable(board, plxType, pciConfigMemPtr);
                  ar_sleep(52);
                  status = a429_utl_plx_eeprom_write(board, plxType, pciConfigMemPtr, bar2SizeWord1, PLX_BAR2_ORIGINAL_SIZE_1);
                  ar_sleep(52);
                  status = a429_utl_plx_eeprom_write(board, plxType, pciConfigMemPtr, bar2SizeWord2, PLX_BAR2_ORIGINAL_SIZE_2);
                  ar_sleep(52);
                  status = a429_utl_plx_eeprom_write_disable(board, plxType, pciConfigMemPtr);
                  ar_sleep(52);  
   
                  /* read/verify the updated content of the PROM */
                  status = a429_utl_plx_eeprom_read (board, plxType, pciConfigMemPtr, promContent);
                  if ((promContent[bar2SizeWord1] != PLX_BAR2_ORIGINAL_SIZE_1) || (promContent[bar2SizeWord2] != PLX_BAR2_ORIGINAL_SIZE_2)) {
                     status = ARS_MEMWRERR;
                  }
                  else {
                     status = ARS_NORMAL;
                  }
               }
               else {
                  status = ARS_NORMAL;
               }
            }
         }
         else {
            status = ARS_FAILURE;
         }
#if !defined (LABVIEW_RT) && !defined (TARGET_EMBEDDED)
         tempStatus = cei_release_thread_lock(LOCK_IDENT);
#endif      
      }
      tempStatus = cei_release_device(board, (ARINC_LOCAL_TYPE *)&device[board]);
   }
   if (tempStatus) ;
   return (CDEV_API_RET_TYPE)status;
}


/*===================================================================================================================*
 * ENTRY POINT:                          A 4 2 9 _ U T L _ E N H _ C H E C K
 *===================================================================================================================*
 *
 * FUNCTION:    Check the PLX EEPROM for CEI-x30 Enhanced firmware support
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board  -- (input) board to reference (0 to 127)
 *              CEI_INT32  deviceIdent -- (input) PCI Device ID for device
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL     Success updating or already done
 *                                ARS_INVBOARD   Invalid board parameter
 *                                ARS_FAILURE    Invalid device (not x30)
 * 
 * GLOBALS:     device[board]
 * 
 * DESCRIPTION: This routine is a special utility to check CEI-x30 board support for the enhanced firmware and API.
 *
 *===================================================================================================================*/
CDEV_API_RET_TYPE a429_utl_enh_check 
(
   CDEV_BOARD_TYPE board,                                                  /* (input)  board to reference (0 to 127) */
   CEI_INT32 deviceIdent                                                   /* (input)  PCI Device ID for device      */
)
{
   CEI_INT32   status = ARS_NORMAL;                                              /* function return status           */
   CEI_INT32   plxType = 0;                                                      /* type of PLX PCI intfc 9056/9030  */
   CEI_INT32   bar2SizeWord1 = 0;                                                /* PCI Cfg offset for PCI BAR2 Size */
   CEI_INT32   bar2SizeWord2 = 0;                                                /* PCI Cfg offset for PCI BAR2 Size */
   CEI_UINT32  promContent[PLX_EEPROM_SIZE];                                     /* contents of PLX EEPROM           */
   pCEI_UINT32 pciConfigMemPtr = 0;                                              /* BAR1 offset for PLX EEPROM       */

   switch (deviceIdent) {
      case PCI_DEVICE_ID_530:
      case PCI_DEVICE_ID_630:
      case PCI_DEVICE_ID_830:
      case PCI_DEVICE_ID_830RX:
      case PCI_DEVICE_ID_A30:
      case PCI_DEVICE_ID_830X820:
      case PCI_DEVICE_ID_RCEI830A:
         plxType       = PLX_9056;
         bar2SizeWord1 = PLX_9056_BAR2_SIZE_WORD_1;
         bar2SizeWord2 = PLX_9056_BAR2_SIZE_WORD_2;
         pciConfigMemPtr = device[board].baseAddr + PLX_CONFIG_MEM_CONTROL_REG;
         break;
      case PCI_DEVICE_ID_430:
      case PCI_DEVICE_ID_430A:
      case PCI_DEVICE_ID_RAREC:
         plxType       = PLX_9030;
         bar2SizeWord1 = PLX_9030_BAR2_SIZE_WORD_1;
         bar2SizeWord2 = PLX_9030_BAR2_SIZE_WORD_2;
         pciConfigMemPtr = device[board].baseAddr + PLX9030_PCI_TARGET_RESPONSE;
         break;
      case PCI_DEVICE_ID_RAR_PCIE:
      case PCI_DEVICE_ID_RAR_XMC:
      case PCI_DEVICE_ID_RAR_MPCIE:
      case PCI_DEVICE_ID_RAR15:
      case PCI_DEVICE_ID_RAR15XT:
         status = ARS_INVBOARD;
         break;
      default:
         status = ARS_FAILURE;
   }

   if (status == ARS_NORMAL) {
      /* read the current content of the PROM */
      status = a429_utl_plx_eeprom_read (board, plxType, pciConfigMemPtr, promContent);
      /* if the PROM is currently programmed for the old CEI-x30 BAR2 size, update it to 512Kb */
      if ((promContent[bar2SizeWord1] != PLX_BAR2_ENHANCED_SIZE_1) || (promContent[bar2SizeWord2] != PLX_BAR2_ENHANCED_SIZE_2)) {
         status = ARS_HW_CONSISTENCY;
      }
   }
   
   
   return (CDEV_API_RET_TYPE)status;
}


/*===================================================================================================================*
 * ENTRY POINT:                    A 4 2 9 _ U T L _ P R O G R A M _ X I L I N X _ 9 0 3 0
 *===================================================================================================================*
 *
 * FUNCTION:    Download the FPGA program to an x30 device having a PLX9030 PCI interface and Xilinx FPGA, 
 *              supporting 8-bit programming.
 *
 * RETURN VAL:  CEI_INT32  -- ARS_NORMAL      Success
 *                            ARS_BADLOAD     FPGA load process failed
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_program_xilinx_9030
(
   CDEV_BOARD_TYPE board,                                                /* (input)  board to reference (0 to 127)   */
   CEI_INT32   deviceIdent,                                              /* (input)  PCI Configuration Device ID     */
   CEI_UINT32  programSize,                                              /* (input)  size of FPGA program load       */
   pCEI_UCHAR  fwloadPtr                                                 /* (input)  pointer to the FPGA program     */
)
{
   CEI_UINT32 tempData;                                                         /* temp variable for read/write      */
   CEI_UINT32 programIndex;                                                     /* temporary FPGA program load index */
   CEI_UCHAR  programData;                                                      /* temporary FPGA program data       */
   CEI_INT32  status = ARS_NORMAL;                                              /* function return status            */
   pCEI_UINT32 pciConfigPtr;                                             /* dev PCI confg space virtual base address */
   pCEI_UINT32 localMemPtr;                                              /* device local memory virtual base address */
   
#ifdef FW_FILE_LOAD
   CEI_UINT32 ucProgramIndex;                                                   /* index into the fpga load array    */
   CEI_UCHAR  fwDataCount;                                                      /* the count for the next sequence   */
   unsigned char data, temp;
   int state = 0;
   FILE *bit_file_ptr;                                                          /* input file pointer                */
   CEI_UINT32 fdata[] = {0xFF,0xFF,0xFF,0xFF,0x55,0x99};
#endif

   pciConfigPtr = device[board].baseAddr;
   localMemPtr  = device[board].devicePtr;

   /**************************************************************
    * Initialize PLX PCI9030 Local Configuration Register Set
    **************************************************************/
   /* Perform a software reset w/delay */
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_PCI_TARGET_RESPONSE], PLX9030_CTRL_REG_SET_SW_RESET); 
   status = cei_delay_milliseconds(20U);
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_PCI_TARGET_RESPONSE], PLX9030_CTRL_REG_DEFAULT); 
   
   /* Delay for 100 milliseconds to allow all configuration memory to clear */
   status = cei_delay_milliseconds(100U);

   status = cei_read_32bit  (board, &pciConfigPtr[PLX9030_LOC_AD_SP_BUS0_DESC], &tempData);
   
   /* set the local addr space 0 bus region descriptor */
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_LOC_AD_SP_BUS0_DESC], PLX9030_LOC_ADDR_SPACE_0_DESC); 

   /* Setup the General Purpose I/O Control register to program the Xilinx  */
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_1);
   status = cei_delay_milliseconds(10U);
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_2);
   status = cei_delay_milliseconds(10U);
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_3);
   status = cei_delay_milliseconds(10U);
   
   /* download the FPGA program file */
#ifndef FW_FILE_LOAD
   for (programIndex=0; programIndex < programSize; programIndex++) {
      programData = *fwloadPtr++;
      status = cei_write_32bit (board, &localMemPtr[0], (CEI_UINT32)programData);
   }
#else
   if ((bit_file_ptr = fopen("cei430.bit", "rb")) != NULL) {
      status = cei_write_32bit (board, &localMemPtr[0], fdata[0]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[1]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[2]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[3]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[4]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[5]);
      while (fread(&data, 1, 1, bit_file_ptr) == 1) {
         switch (state) {
            case 0: if      (data==0xFF) state=1;
                    else                 state=0;
                    break;
            case 1: if      (data==0xFF) state=2;
                    else                 state=0;
                    break;
            case 2: if      (data==0xFF) state=3;
                    else                 state=0;
                    break;
            case 3: if      (data==0xFF) state=4;
                    else                 state=0;
                    break;
            case 4: if      (data==0xAA) state=5;
                    else if (data==0xFF) state=1;
                    else                 state=0;
                    break;
            case 5: if      (data==0x99) state=6;
                    else if (data==0xFF) state=1;
                    else                 state=0;
                    break;
            case 6: /* we need to reverse bit ordering for parallel loading */
                    temp = 0;
                    if (data & 0x80)
                       temp |= 0x01;
                    if (data & 0x40)
                       temp |= 0x02;
                    if (data & 0x20)
                       temp |= 0x04;
                    if (data & 0x10)
                       temp |= 0x08;
                    if (data & 0x08)
                       temp |= 0x10;
                    if (data & 0x04)
                       temp |= 0x20;
                    if (data & 0x02)
                       temp |= 0x40;
                    if (data & 0x01)
                       temp |= 0x80;
 
                    status = cei_write_32bit (board, &localMemPtr[0], temp);
                    break;
         }
      }
      
      fclose(bit_file_ptr);
   }
#endif
   
   /* flush the PLX write fifo */
   for (programIndex=0; programIndex < 3; programIndex++) {  
      status = cei_write_32bit (board, &localMemPtr[0], (CEI_UINT32)0);
   }

   status = cei_delay_milliseconds(3U);
   status = cei_write_32bit  (board, &pciConfigPtr[PLX9030_LOC_AD_SP_BUS0_DESC], tempData);
   
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_3);
   status = cei_delay_milliseconds(10U);
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_2);
   status = cei_delay_milliseconds(10U);
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_1);

   /* Delay for 10 milliseconds to allow the FPGA to complete load/execution */
   status = cei_delay_milliseconds(10U);
   status = cei_read_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], &tempData);
   if ((tempData & PLX9030_FPGA_LOAD_COMPLETE_MASK) == 0) { 
      status = ARS_BADLOAD;
   }
   else {
      /* read the FPGA program load checksum, write it back for load confirmation */
      status = cei_read_32bit (board, &pciConfigPtr[PLX9030_CHIP_SEL_0_BASE_ADD], &tempData);
#if defined DEBUG_PROGRAMMING
      #include "dbg_30defs.h"
#else
      deviceIdent = 0;
#endif
      tempData &= PLX9030_CONFIG_MEM_CSM_MASK;
      status = cei_write_32bit (board, &localMemPtr[CEIDEV_CONFIG_MEM_CSUM], tempData);
      status = cei_delay_milliseconds(3U);
   }
#if defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
   if (deviceIdent) ;
#endif
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:          A 4 2 9 _ U T L _ P R O G R A M _ X I L I N X _ 9 0 3 0 _ 3 2 B I T
 *===================================================================================================================*
 *
 * FUNCTION:    Download the FPGA program to an x30 device having a PLX9030 PCI interface and Xilinx FPGA, 
 *              supporting 32-bit programming.
 *
 * RETURN VAL:  CEI_INT32  -- ARS_NORMAL      Success
 *                            ARS_BADLOAD     FPGA load process failed
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_program_xilinx_9030_32bit
(
   CDEV_BOARD_TYPE board,                                                /* (input)  board to reference (0 to 127)   */
   CEI_INT32    deviceIdent,                                             /* (input)  PCI Configuration Device ID     */
   CEI_UINT32   programSize,                                             /* (input)  size of FPGA program load       */
   pCEI_UINT32  fwloadPtr                                                /* (input)  pointer to the FPGA program     */
)
{
#if defined(FW_FILE_LOAD)
   CEI_UINT32 fwDataWord;                                                       /* the first data word in sequence   */
   CEI_UINT32 fwDataCount;                                                      /* the count for the next sequence   */
   CEI_UINT32 ucProgramIndex;                                                   /* index into the fpga load array    */
   CEI_NATIVE_INT lineCtr = 5;
#endif   
   CEI_UINT32 tempData;                                                         /* temp variable for read/write      */
   CEI_UINT32 programIndex;                                                     /* temporary FPGA program load index */
   CEI_INT32  status = ARS_NORMAL;                                              /* function return status            */
   pCEI_UINT32 pciConfigPtr;                                             /* device PCI config virtual base address   */
   pCEI_UINT32 localMemPtr;                                              /* device local memory virtual base address */
   
#ifdef FW_FILE_LOAD
   unsigned char data, byte3, byte2, byte1, byte0;
   int state = 0;
   FILE *bit_file_ptr;                                                                         /* input file pointer */
   CEI_UINT32 fdata[] = {0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,0xAA995566};
#endif

   pciConfigPtr = device[board].baseAddr;
   localMemPtr  = device[board].devicePtr;

   /**************************************************************
    * Initialize PLX PCI9030 Local Configuration Register Set
    **************************************************************/
   /* Perform a software reset w/delay */
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_PCI_TARGET_RESPONSE], PLX9030_CTRL_REG_SET_SW_RESET); 
   status = cei_delay_milliseconds(20U);
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_PCI_TARGET_RESPONSE], PLX9030_CTRL_REG_DEFAULT); 
   
   /* Delay for 100 milliseconds to allow all configuration memory to clear */
   status = cei_delay_milliseconds(100U);

   status = cei_read_32bit  (board, &pciConfigPtr[PLX9030_LOC_AD_SP_BUS0_DESC], &tempData);
   
   /* set the local addr space 0 bus region descriptor */
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_LOC_AD_SP_BUS0_DESC], PLX9030_LOC_ADDR_SPACE_0_DESC); 

   /* Setup the General Purpose I/O Control register to program the Xilinx  */
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_1);
   status = cei_delay_milliseconds(10U);
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_2);
   status = cei_delay_milliseconds(10U);
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_3);
   status = cei_delay_milliseconds(10U);
   
#ifndef FW_FILE_LOAD
   /* load the firmware by writing to any location in the local bus memory region */
   for (programIndex=0; programIndex < programSize; programIndex++) {
      status = cei_write_32bit (board, &localMemPtr[0], *fwloadPtr++);
   }
#else
   if ((bit_file_ptr = fopen("fpga_ec.bit", "rb")) == NULL) {
      /* load the firmware by writing to any location in the local bus memory region */
      for (programIndex=0; programIndex < programSize; programIndex+=2) {
         fwDataCount = *fwloadPtr++;
         fwDataWord = *fwloadPtr++;
         lineCtr++;
         if (fwDataWord == 0) {
            for (ucProgramIndex=0; ucProgramIndex < fwDataCount; ucProgramIndex++) {
               status = cei_write_32bit (board, &localMemPtr[0], 0);
            }
         }
         else {
            status = cei_write_32bit (board, &localMemPtr[0], fwDataWord);
            if (--fwDataCount > 0) {
               for (ucProgramIndex=0; ucProgramIndex < fwDataCount; ucProgramIndex++) {
                  status = cei_write_32bit (board, &localMemPtr[0], *fwloadPtr++);
                  programIndex++;
               }
            }
         }
      }
   }
   else {
      status = cei_write_32bit (board, &localMemPtr[0], fdata[0]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[1]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[2]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[3]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[4]);
      while (fread(&data, 1, 1, bit_file_ptr) == 1) {
         switch (state) {
             case 0: if      (data==0xFF) state=1;
                     else                 state=0;
                     break;
             case 1: if      (data==0xFF) state=2;
                     else                 state=0;
                     break;
             case 2: if      (data==0xFF) state=3;
                     else                 state=0;
                     break;
             case 3: if      (data==0xFF) state=4;
                     else                 state=0;
                     break;
             case 4: if      (data==0xAA) state=5;
                     else if (data==0xFF) state=1;
                     else                 state=0;
                     break;
             case 5: if      (data==0x99) state=8;
                     else if (data==0xFF) state=1;
                     else                 state=0;
                     break;
             case 6: state = 7;
                     byte3 = data;
                     break;
             case 7: state = 8;
                     byte2 = data;
                     break;
             case 8: state = 9;
                     byte1 = data;
                     break;
             case 9: state = 6;
                     byte0 = data;
                     fwDataWord = (byte3 << 24) + (byte2 << 16) + (byte1 << 8) + byte0;
      
                     status = cei_write_32bit (board, &localMemPtr[0], fwDataWord);
                     byte3 = 0;
                     byte2 = 0;
                     byte1 = 0;
                     byte0 = 0;
                     break;
         }
      }
      
      fclose(bit_file_ptr);
   }
#endif
   
   /* flush the PLX write fifo */
   for (programIndex=0; programIndex < 3; programIndex++) {  
      status = cei_write_32bit (board, &localMemPtr[0], (CEI_UINT32)0);
   }

   status = cei_delay_milliseconds(3U);
   status = cei_write_32bit  (board, &pciConfigPtr[PLX9030_LOC_AD_SP_BUS0_DESC], tempData);
   
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_3);
   status = cei_delay_milliseconds(10U);
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_2);
   status = cei_delay_milliseconds(10U);
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_1);

   /* Delay for 10 milliseconds to allow the FPGA to complete load/execution */
   status = cei_delay_milliseconds(10U);
   status = cei_read_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], &tempData);
   if ((tempData & PLX9030_FPGA_LOAD_COMPLETE_MASK) == 0) { 
      status = ARS_BADLOAD;
   }
   else {
   
      /* read the FPGA program load checksum, write it back for load confirmation */
      status = cei_read_32bit (board, &pciConfigPtr[PLX9030_CHIP_SEL_0_BASE_ADD], &tempData);
#if defined DEBUG_PROGRAMMING
      #include "dbg_30defs.h"
#else
      deviceIdent = 0;
#endif
      tempData &= PLX9030_CONFIG_MEM_CSM_MASK;
      status = cei_write_32bit (board, &localMemPtr[CEIDEV_CONFIG_MEM_CSUM], tempData);
      status = cei_delay_milliseconds(3U);
      status = cei_read_32bit (board, &pciConfigPtr[PLX9030_CHIP_SEL_1_BASE_ADD], &tempData);
      deviceGlobal[board]->serialNumber = (tempData >> 4) & 0x00FFFFFF;
   }
#if defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
   if (deviceIdent) ;
#endif
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:               A 4 2 9 _ U T L _ P R O G R A M    X I L I N X   9 0 5 6
 *===================================================================================================================*
 *
 * FUNCTION:    Download the FPGA program to an x30 device having a PLX9056 PCI interface and Xilinx FPGA, 
 *              supporting 8-bit programming.
 *
 * RETURN VAL:  CEI_INT32  -- ARS_NORMAL      Success
 *                            ARS_BADLOAD     FPGA load process failed
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_program_xilinx_9056 
(
   CDEV_BOARD_TYPE board,                                                /* (input)  device ID                       */
   CEI_INT32   deviceIdent,                                              /* (input)  PCI Configuration Device ID     */
   CEI_UINT32  programSize,                                              /* (input)  size of FPGA program load       */
   pCEI_UCHAR  fwloadPtr                                                 /* (input)  pointer to the FPGA program     */
)
{
   CEI_UINT32 configuration;                                                    /* device channel configuration      */
#ifndef DEBUG_PROGRAMMING
   CEI_UINT32 programCsum;                                                      /* device programming checksum       */
#endif   
   CEI_UINT32 programIndex;                                                     /* temporary FPGA program load index */
   CEI_UCHAR  programData;                                                      /* temporary FPGA program data       */
   CEI_INT32  status = ARS_NORMAL;                                              /* function return status            */
   pCEI_UINT32 pciConfigPtr;                                             /* device PCI config virtual base address   */
   pCEI_UINT32 localMemPtr;                                              /* device local memory virtual base address */

#ifdef FW_830_FILE_LOAD
   CEI_UINT32 ucProgramIndex;                                                   /* index into the fpga load array    */
   CEI_UCHAR  fwDataCount;                                                      /* the count for the next sequence   */
   unsigned char data, temp;
   int state = 0;
   FILE *bit_file_ptr;                                                          /* input file pointer                */
   CEI_UINT32 fdata[] = {0xFF,0xFF,0xFF,0xFF,0x55,0x99};
#endif   

   pciConfigPtr = device[board].baseAddr;
   localMemPtr  = device[board].devicePtr;
   
   /* Assign the startup value for the PLX CONTROL register */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_DEFAULT);
   
   /* Set the Software Reset b30 and delay for at least 1ms */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_SET_SW_RESET);
   status = cei_delay_milliseconds(3U);

   /* Clear Software Reset b30
    * Delay for 100 milliseconds to allow all configuration memory to clear */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_DEFAULT);
   status = cei_delay_milliseconds(100U);

   /* Restore the local configuration registers by reloading from the serial 
    * EEPROM (cleared by the software reset) */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_SET_EE_RELOAD);
   status = cei_delay_milliseconds(10U);
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_DEFAULT);
   status = cei_delay_milliseconds(10U);

   /* Assign the local address space 0 bus region descriptor */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_DESCRIPTOR_REG], PLX9056_MEM_DESC_REG_BUS0_SEL);

   /* set USERO pin low */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_SET_USER0_LOW);

   /* download the FPGA program file */
#ifndef FW_830_FILE_LOAD
   for (programIndex=0; programIndex < programSize; programIndex++) {
      programData = *fwloadPtr++;
      status = cei_write_32bit (board, &localMemPtr[5], (CEI_UINT32)programData);
   }
#else
   if ((bit_file_ptr = fopen("rcei_830A.bit", "rb")) != NULL) {
      status = cei_write_32bit (board, &localMemPtr[0], fdata[0]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[1]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[2]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[3]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[4]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[5]);
      while (fread(&data, 1, 1, bit_file_ptr) == 1) {
         switch (state) {
            case 0: if      (data==0xFF) state=1;
                    else                 state=0;
                    break;
            case 1: if      (data==0xFF) state=2;
                    else                 state=0;
                    break;
            case 2: if      (data==0xFF) state=3;
                    else                 state=0;
                    break;
            case 3: if      (data==0xFF) state=4;
                    else                 state=0;
                    break;
            case 4: if      (data==0xAA) state=5;
                    else if (data==0xFF) state=1;
                    else                 state=0;
                    break;
            case 5: if      (data==0x99) state=6;
                    else if (data==0xFF) state=1;
                    else                 state=0;
                    break;
            case 6: /* we need to reverse bit ordering for parallel loading */
                    temp = 0;
                    if (data & 0x80)
                       temp |= 0x01;
                    if (data & 0x40)
                       temp |= 0x02;
                    if (data & 0x20)
                       temp |= 0x04;
                    if (data & 0x10)
                       temp |= 0x08;
                    if (data & 0x08)
                       temp |= 0x10;
                    if (data & 0x04)
                       temp |= 0x20;
                    if (data & 0x02)
                       temp |= 0x40;
                    if (data & 0x01)
                       temp |= 0x80;
 
                    status = cei_write_32bit (board, &localMemPtr[0], temp);
                    break;
         }
      }
      
      fclose(bit_file_ptr);
   }
#endif
   
   /* flush the PLX write fifo */
   for (programIndex=0; programIndex < 3; programIndex++) {
      status = cei_write_32bit (board, &localMemPtr[5], (CEI_UINT32)0);
   }

   status = cei_delay_milliseconds(3U);
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_DEFAULT);
   status = cei_delay_milliseconds(3U);
   
   /* check if DONE signal is high */
   status = READ_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], &configuration);
   if ((configuration & PLX9056_FPGA_LOAD_COMPLETE_MASK) == 0) { 
      status = ARS_BADLOAD;
   }
   else {
      /* Restore the Bus 0 descriptor register to it's normal operational value */
      status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_DESCRIPTOR_REG], PLX9056_MEM_DESC_REG_BUS0_NORM);
   
      /* write/validate the checksum to local memory */
      status = READ_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CSUM_READ], &programCsum);
#if defined DEBUG_PROGRAMMING
      #include "dbg_defs.h"
#else
      programCsum &= PLX9056_CONFIG_MEM_CSUM_MASK;
      deviceIdent = 0;
#endif
      status = cei_write_32bit (board, &localMemPtr[CEIDEV_CONFIG_MEM_CSUM], programCsum);
      status = cei_delay_milliseconds(3U);
   }
#if defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
   if (deviceIdent) ;
#endif
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:          A 4 2 9 _ U T L _ P R O G R A M _ X I L I N X _ 9 0 5 6 _ 3 2 B I T
 *===================================================================================================================*
 *
 * FUNCTION:    Download the FPGA program to an x30 device having a PLX9056 PCI interface and Xilinx FPGA, 
 *              supporting 32-bit programming.
 *
 * RETURN VAL:  CEI_INT32  -- ARS_NORMAL      Success
 *                            ARS_BADLOAD     FPGA load process failed
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_program_xilinx_9056_32bit
(
   CDEV_BOARD_TYPE board,                                                /* (input)  device ID                       */
   CEI_INT32   deviceIdent,                                              /* (input)  PCI Configuration Device ID     */
   CEI_UINT32  programSize,                                              /* (input)  size of FPGA program load       */
   pCEI_UINT32 fwloadPtr                                                 /* (input)  pointer to the FPGA program     */
)
{
   CEI_UINT32 configuration;                                                    /* device channel configuration      */
#ifndef DEBUG_PROGRAMMING
   CEI_UINT32 programCsum;                                                      /* device programming checksum       */
#endif   
   CEI_UINT32 programIndex;                                                     /* temporary FPGA program load index */
   pCEI_UINT32 pciConfigPtr;                                             /* device PCI config virtual base address   */
   pCEI_UINT32 localMemPtr;                                              /* device local memory virtual base address */
   CEI_INT32 status = ARS_NORMAL;                                               /* function return status            */

   pciConfigPtr = device[board].baseAddr;
   localMemPtr  = device[board].devicePtr;

   /* Assign the startup value for the PLX CONTROL register */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_DEFAULT);
   
   /* Set the Software Reset b30 and delay for at least 1ms */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_SET_SW_RESET);
   status = cei_delay_milliseconds(3U);

   /* Clear Software Reset b30
    * Delay for 100 milliseconds to allow all configuration memory to clear */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_DEFAULT);
   status = cei_delay_milliseconds(100U);

   /* Restore the local configuration registers by reloading from the serial 
    * EEPROM (cleared by the software reset) */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_SET_EE_RELOAD);
   status = cei_delay_milliseconds(10U);
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_DEFAULT);
   status = cei_delay_milliseconds(10U);

   /* Assign the local address space 0 bus region descriptor */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_DESCRIPTOR_REG], PLX9056_MEM_DESC_REG_BUS0_SEL);

   /* set USERO pin low */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_SET_USER0_LOW);
   
   /* load the firmware by writing to any location in the local bus memory region */
   for (programIndex=0; programIndex < programSize; programIndex++) {
      status = cei_write_32bit (board, &localMemPtr[5], *fwloadPtr++);
   }

   status = cei_delay_milliseconds(3U);
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_DEFAULT);
   status = cei_delay_milliseconds(3U);

   /* check if DONE signal is high */
   status = READ_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], &configuration);
   if ((configuration & PLX9056_FPGA_LOAD_COMPLETE_MASK) == 0) { 
      status = ARS_BADLOAD;
   }
   else {
      /* Restore the Bus 0 descriptor register to it's normal operational value */
      status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_DESCRIPTOR_REG], PLX9056_MEM_DESC_REG_BUS0_NORM);
   
      /* write/validate the checksum to local memory */
      status = READ_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CSUM_READ], &programCsum);
   
#if defined DEBUG_PROGRAMMING
      #include "dbg_32defs.h"
#else
      deviceIdent = 0;
#endif
      status = cei_write_32bit (board, &localMemPtr[CEIDEV_CONFIG_MEM_CSUM], programCsum);
      status = cei_delay_milliseconds(3U);
   }
#if defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
   if (deviceIdent) ;
#endif

   return status;
}

