/*====================================================================================================================*
 * FILE:                                     C D E V _ A P I _ R X _ F I L T E R . C
 *====================================================================================================================*
 *
 *   COPYRIGHT 2016 - 2019 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED
 *   ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH THE INCLUSION
 *   OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY OTHER COPIES THEREOF
 *   MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE
 *   AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS, INC. 
 *
 *====================================================================================================================*
 *
 * FUNCTION:    The Label Filtering Routines of the Application Programmer's Interface for a CEI-x30 ARINC device.
 *
 * DESCRIPTION: These routines are provided to demonstrate receive label filtering-specific programming primitives for
 *              the CEI-x30 API using the C programming language.  Each routine implements a particular API function; 
 *              however, a user application program may need to combine functions in different ways to perform the 
 *              desired operation.
 *
 *              Most routines utilize a board number parameter, zero-based on the Device ID/Board Number assigned 
 *              during the installation of CEI-x30-SW distribution.  Channel parameter values are all zero-based.
 *
 * EXTERNALLY DEFINED ROUTINE ENTRY POINTS:         
 *
 *    CEI-x30 Receiver Label Filtering related API routines:
 *
 *    ar_enh_label_filter         - enable/disable enhanced label filtering on the device
 *    ar_getfilter                - retrieve the device filter table contents
 *    ar_get_label_filter         - returns label filter state on all rx chans
 *    ar_label_filter             - define the device label filtering by channel/label values
 *    ar_putfilter                - define the device filter table content
 *
 *--------------------------------------------------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date    By   Ver. Comments
 * -------- ----- ---- -----------------------------------------------------------------------------------------------
 * 10/18/16  BDW  4.90 Initial version.  For modification history related to the routines contained herein prior to
 *                     and including v4.90, see the history section in the header of the file CDEV_API.C.
 * 06/26/17  BDW  4.93 Added CERTIFICATION_BUILD compiler directive control over exclusion of all but the
 *                     ar_enh_label_filter routine in the certification build.
 * 01/22/18  BDW  5.10 Added verification of a valid board number and open session in ar_get_label_filter 
 *
 *====================================================================================================================*/ 
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
   #include <unistd.h>
   #include <fcntl.h>
   #include <string.h>
   #include <pthread.h>
   #include <errno.h>
   #include <sys/stat.h> 
   #include <sys/shm.h>
   #include <sys/sem.h>
   #include <sys/syscall.h>
   #define VOLATILE volatile
#elif defined (VXW_PCI_X86) || defined (VXW_PCI_PPC)
   #include <stdio.h>
   #include <stdlib.h>
   #include <unistd.h>
   #define VOLATILE 
#elif defined (NIVISA_PXI)
   #ifdef LABVIEW_RT
      #include <ansi_c.h> 
   #endif
   #include <visa.h>  
   #define VOLATILE 
#elif defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
   #include <string.h>
   #define VOLATILE 
#else
   #define VOLATILE volatile
#endif

#include "cdev_glb.h"

extern VOLATILE ARINC_LOCAL_TYPE device[MAX_BOARDS];
extern ARINC_GLOBAL_TYPE * deviceGlobal[MAX_BOARDS];


/*===================================================================================================================*
 * ENTRY POINT:                             A R _ E N H _ L A B E L _ F I L T E R
 *===================================================================================================================*
 *
 * FUNCTION:    Function to enable/disable enhanced label filtering on the device.
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board     -- (input) Board number (0 to 127)  
 *                                                   
 *              CDEV_CHAN_TYPE channel    -- (input) ARINC 429 receive channel selection, valid range is 0 to N,  
 *                                                   where N refers to the ARINC 561 channel, and N-1 refers to the  
 *                                                   last ARINC 429 channel installed
 *                                                   
 *              CDEV_PARM_USI_TYPE label  -- (input) Label of interest.  Valid range is 0-255.  Also valid is 
 *                                                   ARU_ALL_LABELS which takes the action for all labels. 
 *                                                   
 *              CDEV_PARM_USI_TYPE sdi    -- (input) SDI of interest.  Valid range is 0-3.  Also valid is 
 *                                                   ARU_ALL_SDI which takes the action for all SDI values. 
 *                                                   
 *              CDEV_PARM_USI_TYPE essm   -- (input) ESSM of interest.  Valid range is 0-7.  Also valid is 
 *                                                   ARU_ALL_ESSM which takes the action for all ESSM values. 
 *                                                   
 *              CDEV_PARM_SSI_TYPE action -- (input) Enable or disable filtering for this board/channel/label.  
 *                                                   Valid values can be a bitwise value including: 
 * 
 *                                                   FILTER_SEQUENTIAL  0x10  If b4 is CLEAR, messages having a
 *                                                                            matching label/SDI/ESSM bit field 
 *                                                                            value are logged to the individual  
 *                                                                            receive buffer.
 *                                                                            If b4 is SET, messages having a
 *                                                                            matching label/SDI/ESSM bit field 
 *                                                                            value are NOT logged to the individual  
 *                                                                            receive buffer.
 *
 *                                                   FILTER_SNAPSHOT    0x20  If b5 is CLEAR, messages having a
 *                                                                            matching label/SDI/ESSM bit field 
 *                                                                            value are logged to the snapshot 
 *                                                                            buffer.
 *                                                                            If b5 is SET, messages having a
 *                                                                            matching label/SDI/ESSM bit field 
 *                                                                            value are NOT logged to the snapshot 
 *                                                                            buffer.
 *                                                   
 *                                                   FILTER_INTERRUPT   0x40  If b6 is SET, reception of messages 
 *                                                                            having a matching label/SDI/ESSM bit 
 *                                                                            field value will result in the 
 *                                                                            insertion of the channel # (offset by
 *                                                                            64) as an entry in the interrupt queue, 
 *                                                                            and if enabled, generation of a PCI 
 *                                                                            interrupt.
 * 
 * RETURN VAL:  CEI_INT32 -- ARS_NORMAL      Success
 *                        -- ARS_INVBOARD    Invalid board argument
 *                        -- ARS_INVHARVAL   Invalid receive channel argument
 *                        -- ARS_INVARG      Invalid label, sdi, essm, or action argument
 *                        -- ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: The device supports hardware label filtering based on the label value for standard label filtering, 
 *              or the combination of the label, SDI, and ESSM field values for enhanced label filtering.  This 
 *              rroutine assigns the specified filter action to the respective enhanced label filter table location  
 *              for the device receive channel.
 *
 *===================================================================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_enh_label_filter 
(
   CDEV_BOARD_TYPE board,                                                       /* (input) Board number (0 to 127)   */
   CDEV_CHAN_TYPE  channel,                                                     /* (input) 429 Rx channel selection  */
   CDEV_PARM_USI_TYPE label,                                                    /* (input) Label of interest         */
   CDEV_PARM_USI_TYPE sdi,                                                      /* (input) SDI of interest           */
   CDEV_PARM_USI_TYPE essm,                                                     /* (input) ESSM of interest          */
   CDEV_PARM_SSI_TYPE action                                                    /* (input) Filtering Enable/Disable  */
)
{
   CEI_UINT32 labelFilterIndex;                                                /* label filter channel select index  */
   CEI_UINT32 localChannel = 0U;                                               /* receive channel select             */
   CEI_UINT32 filterActivation;                                                /* local translation of action parm   */
   CEI_UINT32 filterTableValue;                                                /* local translation of action parm   */
   CEI_UINT32 filterTableMask;                                                 /* mask of the current table entry    */
   CEI_UINT32 essmEnd;                                                         /* local temp enh filter table index  */
   CEI_UINT32 essmStart;                                                       /* local temp enh filter table index  */
   CEI_UINT32 essmIndex;                                                       /* local temp enh filter table index  */
   CEI_UINT32 sdiEnd;                                                          /* local temp enh filter table index  */
   CEI_UINT32 sdiStart;                                                        /* local temp enh filter table index  */
   CEI_UINT32 sdiIndex;                                                        /* local temp enh filter table index  */
   CEI_UINT32 labelEnd;                                                        /* local temp enh filter table index  */
   CEI_UINT32 labelStart;                                                      /* local temp enh filter table index  */
   CEI_UINT32 labelIndex;                                                      /* local temp enh filter table index  */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                      /* function return status             */
   CEI_INT32 accessStatus = ARS_NORMAL;                                        /* status of device read/write access */
#ifndef TARGET_EMBEDDED
   CEI_INT32 tempStatus = ARS_NORMAL;                                          /* local temp return status           */
#endif

   localChannel = (CEI_UINT32)channel;
   filterActivation = FILTER_SNAPSHOT;

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   /* For current configurations, any label filter channel request is limited
    * to the number of ARINC 429 channels installed on the device
    */
   else if (channel > (CDEV_CHAN_TYPE)(deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX])) {
      status = ARS_INVHARVAL;
   }
   else {
#endif
      localChannel = (CEI_UINT32)channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
#ifndef TARGET_EMBEDDED
   }
   
   /* check the values passed in for all remaining parameters */
   if (status == ARS_NORMAL) {
      if ( (label > ARU_ALL_LABELS) || (sdi > ARU_ALL_SDI) || (essm > ARU_ALL_ESSM) ) {
         status = ARS_INVARG;
      }
      else {
         status = a429_utl_validate_channel_request (board, localChannel, ARU_ENHANCED_FILTER_LABEL);
      }
      
      if (status == ARS_NORMAL) {
#endif
         if (action == ARU_FILTER_ON) {
            filterActivation = FILTER_SEQUENTIAL + FILTER_SNAPSHOT;
         }
         else if (action == ARU_FILTER_OFF) {
            filterActivation = 0;
         }
         else if ((action < FILTER_SEQUENTIAL) || (action > (FILTER_SEQUENTIAL + FILTER_SNAPSHOT + FILTER_INTERRUPT))) {
            status = ARS_INVARG;
         }
         else {
            filterActivation = (CEI_INT32)action;
         }
         
         filterActivation = filterActivation >> 4;    /* API definitions provide legacy support with x20 */
#ifndef TARGET_EMBEDDED
      }
   }
   
   if (status == ARS_NORMAL) {
      
      /* if multi-thread protection is active, acquire a lock on the host interface */
      if (device[board].multiThreadProtect != AR_CONC_NONE) {
         status = (CDEV_API_RET_TYPE)cei_acquire_thread_lock(LOCK_IDENT);
      }
      
      if (status == ARS_NORMAL) {
#endif
         filterTableValue = 0;
         filterTableMask = 0;
   
         /* if the same action applies to all essm/sdi/label combinations, loop through all
          * and assign the specified action
          */
         if ((label == ARU_ALL_LABELS) && (sdi == ARU_ALL_SDI) && (essm == ARU_ALL_ESSM)) {
   
            for (essmIndex = 0; essmIndex < ARU_ALL_ESSM; essmIndex++) {
               filterTableValue |= filterActivation << (essmIndex * CEIDEV_ENH_LABEL_ESSM_FILTER_SHIFT);
            }
   
            labelFilterIndex = CEIDEV_ENH_LABEL_FILTER_START + (localChannel * CEIDEV_ENH_LABEL_FILTER_CHAN_SIZE);
            labelEnd = labelFilterIndex + CEIDEV_ENH_LABEL_FILTER_CHAN_SIZE;
            while ((labelFilterIndex < labelEnd) && (accessStatus == ARS_NORMAL)) {
               /* load the External Memory Address Register with the Label Filter Table SRAM long-word address  */
               accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_ADDR], labelFilterIndex);
               /* write the filter table entry data pattern to this location */
               if (accessStatus == ARS_NORMAL) {
#if defined (INTEGRITY_PPC_SYNC)
                  #pragma asm
                  eieio
                  sync
                  #pragma endasm
#elif defined (PPC_SYNC)
                  IO_SYNC;
#endif
                  accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_DATA], filterTableValue);
                  labelFilterIndex++;
               }
            }
         }
         else {   /* otherwise, apply the specified action to the specified essm/sdi/label location(s) */
            if (label == ARU_ALL_LABELS) {
               labelStart = 0;
               labelEnd = 256;
            }
            else {
               labelStart = label;
               labelEnd = label+1;
            }
            if (sdi == ARU_ALL_SDI) {
               sdiStart = 0;
               sdiEnd = ARU_ALL_SDI;
            }
            else {
               sdiStart = sdi;
               sdiEnd = sdi+1;
            }
            if (essm == ARU_ALL_ESSM) {
               essmStart = 0;
               essmEnd = ARU_ALL_ESSM;
            }
            else {
               essmStart = essm;
               essmEnd = essm+1;
            }
            /* create the 32-bit filter activation definition for each 4-bit ESSM field 
             * along with a mask for the current content of the respective table entry.
             */
            for (essmIndex = essmStart; essmIndex < essmEnd; essmIndex++) {
               filterTableValue |= filterActivation << (essmIndex * CEIDEV_ENH_LABEL_ESSM_FILTER_SHIFT);
               filterTableMask  |= CEIDEV_ENH_FLTR_BIT_MASK  << (essmIndex * CEIDEV_ENH_LABEL_ESSM_FILTER_SHIFT);
            }
            filterTableMask = ~filterTableMask;
   
            /* now assign the action to the specified locations in the table */
            for (labelIndex = labelStart; labelIndex < labelEnd; labelIndex++) {
               for (sdiIndex = sdiStart; sdiIndex < sdiEnd; sdiIndex++) {
                  labelFilterIndex = CEIDEV_ENH_LABEL_FILTER_START + (localChannel * CEIDEV_ENH_LABEL_FILTER_CHAN_SIZE);
                  labelFilterIndex += (labelIndex << CEIDEV_ENH_LABEL_FILTER_LB_SHIFT) + sdiIndex;
                  
                  /* load the External Memory Address Register with the Label Filter Table SRAM long-word address  */
                  accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_ADDR], labelFilterIndex);
                  /* read the current contents of the label filter table location(s) to be updated */
                  if (accessStatus == ARS_NORMAL) {
#if defined (INTEGRITY_PPC_SYNC)
                     #pragma asm
                     eieio
                     sync
                     #pragma endasm
#elif defined (PPC_SYNC)
                     IO_SYNC;
#endif
                     accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_DATA], &filterActivation);
                     /* mask out only the selected ESSM field in this entry for the requested filter update, then add the new setting(s) */
                     if (accessStatus == ARS_NORMAL) {
                        filterActivation &= filterTableMask;
                        filterActivation |= filterTableValue;
#if defined (INTEGRITY_PPC_SYNC)
                        #pragma asm
                        eieio
                        sync
                        #pragma endasm
#elif defined (PPC_SYNC)
                        IO_SYNC;
#endif
                        /* write the updated filter table entry back to this location */
                        accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_DATA], filterActivation);
                     }
                  }
               }
            }
         }
#ifndef TARGET_EMBEDDED
      }
      /* if multi-thread protection is active, free any acquired lock */
      if (device[board].multiThreadProtect != AR_CONC_NONE) {
         if (status == ARS_NORMAL) {
            status = (CDEV_API_RET_TYPE)cei_release_thread_lock(LOCK_IDENT);
         }
         else {
            tempStatus = cei_release_thread_lock(LOCK_IDENT);
         }
      }
   }
   
   if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL)) {
      status = (CDEV_API_RET_TYPE)accessStatus;
   }
   if (tempStatus) ;
#endif
   return (CEI_INT32)status;
}


#ifndef CERTIFICATION_BUILD
/*===================================================================================================================*
 * ENTRY POINT:                                  A R _ G E T F I L T E R
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the label interrupt and filter buffer definition used by the device enhanced label 
 *              filter feature, (for legacy API compatibility with the CEI-x20 API).
 * 
 * PARAMETERS:  CEI_UINT32 board      -- (input) Board number (0 to 127)  
 *              CEI_UINT32 channel    -- (input) ARINC 429 Rx channel selection, valid range is 0 to one less than  
 *                                               the number of enhanced ARINC 429 channels installed.
 *              pCEI_CHAR filterTable -- (input) Pointer to the host filter table at least 8Kb in size.
 * 
 * RETURN VAL:  ARS_NORMAL        operation completed successfully
 *              ARS_INVBOARD      invalid board number or board not initialized
 *              ARS_INVARG        invalid argument
 *              ARS_BOARD_MUTEX   Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 *
 * DESCRIPTION: This routine copies the specified label filter buffer entry from the enhanced label filter table on 
 *              the device to the caller's memory space.  Each element of the label filter buffer returned consists 
 *              of a bit field defined as follows:
 *
 *                                                   FILTER_SEQUENTIAL  0x10  If b4 is CLEAR, messages having a
 *                                                                            matching label/SDI/ESSM bit field 
 *                                                                            value are logged to the individual  
 *                                                                            receive buffer.
 *                                                                            If b4 is SET, messages having a
 *                                                                            matching label/SDI/ESSM bit field 
 *                                                                            value are NOT logged to the individual  
 *                                                                            receive buffer.
 *
 *                                                   FILTER_SNAPSHOT    0x20  If b5 is CLEAR, messages having a
 *                                                                            matching label/SDI/ESSM bit field 
 *                                                                            value are logged to the snapshot 
 *                                                                            buffer.
 *                                                                            If b5 is SET, messages having a
 *                                                                            matching label/SDI/ESSM bit field 
 *                                                                            value are NOT logged to the snapshot 
 *                                                                            buffer.
 *                                                   
 *                                                   FILTER_INTERRUPT   0x40  If b6 is SET, reception of messages 
 *                                                                            having a matching label/SDI/ESSM bit 
 *                                                                            field value will result in the 
 *                                                                            insertion of the channel # (offset by
 *                                                                            64) as an entry in the interrupt queue, 
 *                                                                            and if enabled, generation of a PCI 
 *                                                                            interrupt.
 * 
 *              NOTE: the following CEI-x20 label filter table bit field assigned features are not supported:
 *
 *                                                   FILTER_SIGNAL       0x80 Transmit label signaling
 *                                                   FILTER_CHANNEL      0x0F Transmit channel number (0-15) 
 *                                                                            to signal
 *
 *              The filter buffer for a single channel on a CEI-x30 device is defined as follows:
 *
 *                 CEI_UCHAR filterTable[MAX_ESSM][MAX_SDI][MAX_LABEL]
 *
 *              where the bits of the ARINC word are split up as follows:
 *
 *              field   ---eSSM---  -SDI-  --------label---------
 *               bits   30  29  28   9 8   7  6  5  4  3  2  1  0
 *                    
 *              On a CEI-x30 device each 32-bit entry in the filter table is allocated once for each label/SDI 
 *              combination, divided into individual filter activation fields by eSSM:
 *
 *                 Label Filter Table [MAX_LABEL][MAX_SDI]
 *
 *               bits   31 30 29 28 27 ... 8  7  6  5  4  3  2  1  0
 *              field  | eSSM = 7  |   ...   | eSSM = 1  | eSSM = 0 |
 *
 *===================================================================================================================*/ 
EXPORT32 CEI_INT32 DLL_EXPORTED ar_getfilter 
(
   CEI_UINT32 board,                                                     /* (input) Board number (0 to 127)          */
   CEI_UINT32 channel,                                                   /* (input) 429 Rx channel selection         */
   pCEI_CHAR filterTable                                                 /* (input) pointer to the host filter table */
)
{
   CEI_UINT32 localChannel;                                            /* channel offset into the label filter table */
   CEI_INT32  labelFilterIndex;                                                /* label filter channel select index  */
   CEI_UINT32 filterActivation;                                                /* local translation of action parm   */
   CEI_UINT32 essmIndex;                                                       /* local temp enh filter table index  */
   CEI_UINT32 labelIndex;                                                      /* local temp enh filter table index  */
   CEI_UINT32 sdiIndex;                                                        /* local temp enh filter table index  */
   pCEI_CHAR  hostFilterTablePtr;                                              /* local pointer to host filterTable  */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                      /* function return status             */
   CEI_INT32  accessStatus = ARS_NORMAL;                                       /* status of device read/write access */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   /* Label Filter Table access requests are limited to the number of ARINC 429 receive channels installed on the device */
   else if ( ((CEI_UINT32)channel >= deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) || (filterTable == NULL) ) {
      status = ARS_INVARG;
   }
   else {
#endif
      localChannel = (CEI_UINT32)channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
   
#ifndef TARGET_EMBEDDED
      /* if multi-thread protection is active, acquire a lock on the host interface */
      if (device[board].multiThreadProtect != AR_CONC_NONE) {
         status = (CDEV_API_RET_TYPE)cei_acquire_thread_lock(LOCK_IDENT);
      }
      
      if (status == ARS_NORMAL) {
#endif
         /* copy the specified label filter table entry states from the device table to the host table */
         labelFilterIndex = CEIDEV_ENH_LABEL_FILTER_START + (localChannel * CEIDEV_ENH_LABEL_FILTER_CHAN_SIZE);
         for (labelIndex = 0; (labelIndex < MAX_LABEL) && (accessStatus == ARS_NORMAL); labelIndex++) {
            for (sdiIndex = 0; (sdiIndex < MAX_SDI) && (accessStatus == ARS_NORMAL); sdiIndex++) {
               /* load the External Memory Address Register with the Label Filter Table SRAM long-word address  */
               accessStatus = cei_write_32bit ((CDEV_BOARD_TYPE)board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_ADDR], labelFilterIndex);
               if (accessStatus == ARS_NORMAL) {
#if defined (INTEGRITY_PPC_SYNC)
                  #pragma asm
                  eieio
                  sync
                  #pragma endasm
#elif defined (PPC_SYNC)
                  IO_SYNC;
#endif
                  /* read the filter table entry data pattern from this location */
                  accessStatus = cei_read_32bit ((CDEV_BOARD_TYPE)board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_DATA], &filterActivation);
                  /* read eight ESSM entries for each combination of SDI/label entry */
                  for (essmIndex = 0; essmIndex < ARU_ALL_ESSM; essmIndex++) {
                     hostFilterTablePtr = filterTable  + (essmIndex << 10) + (sdiIndex << 8) + labelIndex;
                     *hostFilterTablePtr = (CEI_CHAR)(((filterActivation >> (essmIndex * CEIDEV_ENH_LABEL_ESSM_FILTER_SHIFT)) & 0x0000000F) << 4);
                  }
               }
               labelFilterIndex++;  
            }
         }

#ifndef TARGET_EMBEDDED
         /* if multi-thread protection is active, free any acquired lock */
         if (device[board].multiThreadProtect != AR_CONC_NONE) {
            status = (CDEV_API_RET_TYPE)cei_release_thread_lock(LOCK_IDENT);
         }
         if (accessStatus != ARS_NORMAL) {
            status = (CDEV_API_RET_TYPE)accessStatus;
         }
      }
   }
#endif

   
   return (CEI_INT32)status;
}


/*===================================================================================================================*
 * ENTRY POINT:                           A R _ G E T _ L A B E L _ F I L T E R
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the active state of filtering for all receive channels for the specified label, (for legacy
 *              API compatibility with the CEI-x20 API).
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board    -- (input) Board number (0 to 127)  
 *              CDEV_PARM_USI_TYPE label -- (input) label to query the filter state
 * 
 * RETURN VAL:  Label Filter State -- each bit indicates the label state on the respective receive channel 0-15.
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 *
 * DESCRIPTION: This routine indicates the active state of label filtering for the specified label on each receive 
 *              channel.  The state indicates active filtering with a "1" bit state for the bit respective to the 
 *              receive channel.  Active filtering is defined as any label filter table entry for any Enhanced SSM
 *              and SDI value combination where either sequential or snapshot buffer filterin is enabled.
 *
 *===================================================================================================================*/ 
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_get_label_filter 
(
   CDEV_BOARD_TYPE board,                                             /* (input) Board number (0 to 127)             */
   CDEV_PARM_USI_TYPE label                                           /* (input) the label to query the filter state */
) 
{
   CEI_UINT32 status = 0;                                                       /* function return status            */
   CEI_UINT32 sdiIndex;                                                         /* local temp enh filter table index */
   CEI_UINT32 localChannel = 0U;                                                /* receive channel select            */
   CEI_UINT32 labelFilterIndex;                                                 /* label filter channel select index */
   CEI_UINT32 filterActivation;                                                 /* local translation of action parm  */
   CEI_INT32 accessStatus = ARS_NORMAL;                                         /* status of device read/write access*/
   
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      /* read the label filter table for the specified filter on all channels, 
       * and if any eSSM enties are non-Zero, this label is considered "filtered" */
      for (localChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
           (localChannel <= deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]) && (accessStatus == ARS_NORMAL); 
           localChannel++) {
         for (sdiIndex = 0; (sdiIndex < ARU_ALL_SDI) && (accessStatus == ARS_NORMAL); sdiIndex++) {
            labelFilterIndex = CEIDEV_ENH_LABEL_FILTER_START + (localChannel * CEIDEV_ENH_LABEL_FILTER_CHAN_SIZE);
            labelFilterIndex += ((CEI_UINT32)label << CEIDEV_ENH_LABEL_FILTER_LB_SHIFT) + sdiIndex;
            
            /* load the External Memory Address Register with the Label Filter Table SRAM long-word address  */
            accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_ADDR], labelFilterIndex);
            if (accessStatus == ARS_NORMAL) {
#if defined (INTEGRITY_PPC_SYNC)
               #pragma asm
               eieio
               sync
               #pragma endasm
#elif defined (PPC_SYNC)
               IO_SYNC;
#endif
               /* read the current contents of the label filter table location(s) to be updated */
               accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_DATA], &filterActivation);
               /* mask out only the selected ESSM field in this entry for the requested filter update, then add the new setting(s) */
               if (filterActivation & 0x33333333) {
                  status |= 1 << (localChannel - deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]);
               }
            }
         }
      }
#ifndef TARGET_EMBEDDED
   }
   else {
      status = ARS_INVBOARD;
   }
#endif
   
   if (accessStatus) ;
   return (CDEV_API_RET_TYPE)status;
}


/*===========================================================================*
 * ENTRY POINT:         A R _ L A B E L _ F I L T E R
 *===========================================================================*
 *
 * FUNCTION:    Function to enable/disable standard label filtering on device.
 * 
 *===========================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_label_filter 
( 
   CDEV_BOARD_TYPE    board,         /* (input) Board number (0 to 127)      */
   CDEV_CHAN_TYPE     channel,       /* (input) 429 Rx channel selection     */
   CDEV_PARM_USI_TYPE label,         /* (input) Label of interest.           */
   CDEV_PARM_SSI_TYPE action         /* (input) Enable or disable filtering  */
)
{
   CDEV_API_RET_TYPE status;                              /* function return status  */
   /* the routine ar_enh_label_filter handles label filter table processing  */
   status = (CDEV_API_RET_TYPE)ar_enh_label_filter (board, channel, label, ARU_ALL_SDI, ARU_ALL_ESSM, action);
   return (status);
}


/*===================================================================================================================*
 * ENTRY POINT:                                  A R _ P U T F I L T E R
 *===================================================================================================================*
 *
 * FUNCTION:    Copy the raw label label filter table from the application to the device, (for legacy API
 *              compatibility with the CEI-x20 API).
 * 
 * PARAMETERS:  CEI_UINT32 board      -- (input) Board number (0 to 127)  
 *              CEI_UINT32 channel    -- (input) 429 Rx channel selection,
 *                                               valid range is 0 to one less
 *                                               than the number of enhanced
 *                                               ARINC 429 channels installed
 *              pCEI_CHAR filterTable -- (input) pointer to the host filter table
 * 
 * RETURN VAL:  ARS_NORMAL        operation completed successfully
 *              ARS_INVBOARD      invalid board number or board not initialized
 *              ARS_INVARG        invalid receive channel or null table ptr
 *              ARS_BOARD_MUTEX   Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 *
 * DESCRIPTION: This routine takes the content of the specified label filter table, defined in CEI-x20 label filter 
 *              terms, and assigns the equivalent filter table values to the CEI-x30 enhanced label filter table.
 *              Each element of the application filter buffer consists of a bit field defined as follows:
 *
 *                                                   FILTER_SEQUENTIAL  0x10  If b4 is CLEAR, messages having a
 *                                                                            matching label/SDI/ESSM bit field 
 *                                                                            value are logged to the individual  
 *                                                                            receive buffer.
 *                                                                            If b4 is SET, messages having a
 *                                                                            matching label/SDI/ESSM bit field 
 *                                                                            value are NOT logged to the individual  
 *                                                                            receive buffer.
 *
 *                                                   FILTER_SNAPSHOT    0x20  If b5 is CLEAR, messages having a
 *                                                                            matching label/SDI/ESSM bit field 
 *                                                                            value are logged to the snapshot 
 *                                                                            buffer.
 *                                                                            If b5 is SET, messages having a
 *                                                                            matching label/SDI/ESSM bit field 
 *                                                                            value are NOT logged to the snapshot 
 *                                                                            buffer.
 *                                                   
 *                                                   FILTER_INTERRUPT   0x40  If b6 is SET, reception of messages 
 *                                                                            having a matching label/SDI/ESSM bit 
 *                                                                            field value will result in the 
 *                                                                            insertion of the channel # (offset by
 *                                                                            64) as an entry in the interrupt queue, 
 *                                                                            and if enabled, generation of a PCI 
 *                                                                            interrupt.
 * 
 *              NOTE: the following CEI-x20 label filter table bit field assigned features are not supported:
 *
 *                                                   FILTER_SIGNAL       0x80 Transmit label signaling
 *                                                   FILTER_CHANNEL      0x0F Transmit channel number (0-15) 
 *                                                                            to signal
 *
 *              The filter buffer for a single channel on a CEI-x30 device is defined as follows:
 *
 *                 CEI_UCHAR filterTable[MAX_ESSM][MAX_SDI][MAX_LABEL]
 *
 *              where the bits of the ARINC word are split up as follows:
 *
 *              field   ---eSSM---  -SDI-  --------label---------
 *               bits   30  29  28   9 8   7  6  5  4  3  2  1  0
 *                    
 *              On a CEI-x30 device each 32-bit entry in the filter table is allocated once for each label/SDI 
 *              combination, divided into individual filter activation fields by eSSM:
 *
 *                 Label Filter Table [MAX_LABEL][MAX_SDI]
 *
 *               bits   31 30 29 28 27 ... 8  7  6  5  4  3  2  1  0
 *              field  | eSSM = 7  |   ...   | eSSM = 1  | eSSM = 0 |
 *                    
 *===================================================================================================================*/ 
EXPORT32 CEI_INT32 DLL_EXPORTED ar_putfilter 
(
   CEI_UINT32 board,                                                     /* (input) Board number (0 to 127)          */
   CEI_UINT32 channel,                                                   /* (input) 429 Rx channel selection         */
   pCEI_CHAR filterTable                                                 /* (input) pointer to the host filter table */
)
{
   CEI_INT32  labelFilterIndex;                                               /* label filter channel select index   */
   CEI_UINT32 filterActivation;                                               /* local translation of action parm    */
   CEI_UINT32 essmIndex;                                                      /* local temp enh filter table index   */
   CEI_UINT32 labelIndex;                                                     /* local temp enh filter table index   */
   CEI_UINT32 sdiIndex;                                                       /* local temp enh filter table index   */
   CEI_INT32  localChannel = 0;                                               /* merged/individual channel select    */
   CEI_CHAR   hostFilterTableData;                                            /* local copy of host filterTable ent  */
   pCEI_CHAR  hostFilterTablePtr;                                             /* local pointer to host filterTable   */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                     /* function return status              */
   CEI_INT32  accessStatus = ARS_NORMAL;                                      /* status of device read/write access  */
#ifndef TARGET_EMBEDDED
   CEI_INT32  tempStatus = ARS_NORMAL;                                        /* local temp return status            */
#endif
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr = NULL;                   /* pointer to board local structure    */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr = NULL;                          /* pointer to board global structure   */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   else {
      deviceGlobalStructPtr = deviceGlobal[board];

      /* Label Filter Table access requests are limited to the number of ARINC 429 receive channels installed on the device */
      if ( ((CEI_UINT32)channel >= deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) || (filterTable == NULL) ) {
         status = ARS_INVARG;
      }
      else {
#else
         deviceGlobalStructPtr = deviceGlobal[board];
#endif
         deviceLocalStructPtr = &device[board];
         localChannel = (CEI_INT32)channel + (CEI_INT32)deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
   
#ifndef TARGET_EMBEDDED
         /* if multi-thread protection is active, acquire a lock on the host interface */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            status = (CDEV_API_RET_TYPE)cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (status == ARS_NORMAL) {
#endif
            /* assign the specified action from the host table to the device table */
            hostFilterTablePtr = filterTable;
      
            labelFilterIndex = CEIDEV_ENH_LABEL_FILTER_START + (localChannel * CEIDEV_ENH_LABEL_FILTER_CHAN_SIZE);
            for (labelIndex = 0; labelIndex < MAX_LABEL; labelIndex++) {
               for (sdiIndex = 0; sdiIndex < MAX_SDI; sdiIndex++) {
                  filterActivation = 0;
                  for (essmIndex = 0; essmIndex < MAX_ESSM; essmIndex++) {
                     hostFilterTablePtr = filterTable  + (essmIndex << 10) + (sdiIndex << 8) + labelIndex;
                     hostFilterTableData = *hostFilterTablePtr;
                     filterActivation <<= 4;
                     filterActivation |= (CEI_UINT32)(hostFilterTableData >> 4) & CEIDEV_ENH_FLTR_BIT_MASK;
                  }
   
                  /* load the External Memory Address Register with the Label Filter Table SRAM long-word address  */
                  accessStatus = cei_write_32bit ((CDEV_BOARD_TYPE)board, &deviceLocalStructPtr->devicePtr[CEIDEV_ACCESS_SRAM_ADDR], labelFilterIndex);
                  /* write the filter table entry data pattern to this location */
                  accessStatus = cei_write_32bit ((CDEV_BOARD_TYPE)board, &deviceLocalStructPtr->devicePtr[CEIDEV_ACCESS_SRAM_DATA], filterActivation);
                  labelFilterIndex++;  
               }
            }
   
#ifndef TARGET_EMBEDDED
            /* if multi-thread protection is active, free any acquired lock */
            if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
               if (status == ARS_NORMAL) {
                  status = (CDEV_API_RET_TYPE)cei_release_thread_lock(LOCK_IDENT);
               }
               else {
                  tempStatus = cei_release_thread_lock(LOCK_IDENT);
               }
            }
         }
      }
   }
   
   if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL)) {
      status = (CDEV_API_RET_TYPE)accessStatus;
   }
   if (tempStatus) ;
#endif
      
   return (CEI_INT32)status;
}
#endif

