/*====================================================================================================================*
 * FILE:                                     C D E V _ A P I _ U T I L I T Y . C                    
 *====================================================================================================================*
 *
 *   COPYRIGHT 2016-2019 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED
 *   ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH THE INCLUSION
 *   OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY OTHER COPIES THEREOF
 *   MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE
 *   AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS, INC. 
 *
 *====================================================================================================================*
 *
 * FUNCTION:    The Utility Routines portion of the Application Programmer's Interface for a CEI-x30 ARINC device.
 *
 * DESCRIPTION: These routines are provided to demonstrate the API utility programming primitives for the 
 *              CEI-x30 API using the C programming language.  Each routine implements a particular API function; 
 *              however, a user application program may need to combine functions in different ways to perform the 
 *              desired operation.
 *
 *              Most routines utilize a board number parameter, zero-based on the Device ID/Board Number assigned 
 *              during the installation of CEI-x30-SW distribution.  Channel parameter values are all zero-based.
 *
 * EXTERNALLY DEFINED ROUTINE ENTRY POINTS:         
 *
 *    CEI-x30 Utility API routines:
 *    
 *    ar_convert_time_to_string     
 *    ar_convert_1553_time_to_string
 *    ar_get_base_addr            - returns the virtual address of the device
 *    ar_get_boardname            - returns a device description
 *    ar_get_boardnameLV          - returns the target board name string
 *    ar_get_boardtype            - returns the target board type
 *    ar_get_channel_index_info   - returns the channel index information
 *    ar_get_config               - returns 429 config register bit states
 *    ar_get_storage_mode         - returns current receiver storage mode
 *    ar_query_device             - queries the PCI IDs and returns the board type
 *    ar_version                  - get the API version string.
 *
 *--------------------------------------------------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date    By   Ver. Comments
 * -------- ----- ---- -----------------------------------------------------------------------------------------------
 * 10/18/16  BDW  4.90 Initial version.  For modification history related to the routines contained herein prior to
 *                     and including v4.90, see the history section in the header of the file CDEV_API.C.
 * 03/08/17  BDW  4.92 Add ATP capability to read the RCEI-530 and R830RX PLX Configuration Space Mode/DMA
 *                     Arbitration register value.
 * 11/17/17  BDW  5.00 Modified the ar_get_base_addr function call return type from pCEI_UINT32 to pCEI_VOID to make it
 *                     64-bit compliant.
 * 01/22/18  BDW  5.10 Added verification of a valid open session for the AR_TIMER_X20_COMPAT_32BIT resolution time 
 *                     conversion option in ar_convert_time_to_string
 *
 *====================================================================================================================*/ 
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
   #include <unistd.h>
   #include <fcntl.h>
   #include <string.h>
   #include <pthread.h>
   #include <errno.h>
   #include <sys/stat.h> 
   #include <sys/shm.h>
   #include <sys/sem.h>
   #include <sys/syscall.h>
   #define VOLATILE volatile
#elif defined (VXW_PCI_X86) || defined (VXW_PCI_PPC)
   #include <stdio.h>
   #include <stdlib.h>
   #include <unistd.h>
   #define VOLATILE 
#elif defined (NIVISA_PXI)
   #ifdef LABVIEW_RT
      #include <ansi_c.h> 
   #endif
   #include <visa.h>  
   #define VOLATILE 
#elif defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
   #include <string.h>
   #define VOLATILE 
#else
   #define VOLATILE volatile
#endif

#include "cdev_glb.h"

extern VOLATILE ARINC_LOCAL_TYPE device[MAX_BOARDS];
extern ARINC_GLOBAL_TYPE * deviceGlobal[MAX_BOARDS];

/**********************************************************************************************************************
 *                             I N I T I A L I Z E D   P R I V A T E   D A T A                                        *
 **********************************************************************************************************************/
#if defined (VXW_PCI_PPC) || defined (VXW_PCI_X86) || defined (_LINUX_X86_) || defined (_LINUX_PPC_) || defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
  #if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
    #ifdef MULTI_PROTOCOL_SUPPORT
       extern void BusTools_TimeGetFmtString(CEI_INT32 format, CEI_UINT64 * timeTag, CEI_CHAR * timeString);
    #endif
  #endif
#else
  void (*ptr_BusTools_TimeGetFmtString)(CEI_INT32 format, CEI_UINT64 * timeTag, CEI_CHAR * timeString);
  static HINSTANCE bt1553apiDllHandle = NULL;
#endif


/* Board names used by ar_get_boardname */
static pCEI_CHAR boardNames[CEIDEV_LAST_CEIX30_CONFIG - CEIDEV_FIRST_CEIX30_CONFIG + 1] = {  
   "CEI-830", 
   "CEI-430", 
   "AMC-A30", 
   "CEI-530", 
   "R830RX", 
   "RAR-CPCI", 
   "RAR-EC", 
   "RAR-PCIE", 
   "CEI-430A", 
   "RAR15-XMC", 
   "RAR15-XMC",
   "RCEI-830X820",
   "RAR-XMC",
   "RCEI-830A",
   "RAR-MPCIE"
};


/*===================================================================================================================*
 * ENTRY POINT:                       A R _ C O N V E R T _ T I M E _ T O _ S T R I N G   
 *===================================================================================================================*
 *
 * FUNCTION:    Convert standard x30 64-bit time value to character string
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board            -- (input)  Board number (0 to 127)  
 *              CDEV_PARM_SSI_TYPE displayFormat -- (input)  Format for the returned string:
 *                                                           AR_TD_REL_MIDNIGHT  Relative to Midnight Format and
 *                                                           AR_TD_IRIG          Full IRIG Format, both defined as:
 *                                                                                  "(DDD)hh:mm:ss.uuuuuu"           
 *                                                           AR_TD_DATE          Date Format defined as:
 *                                                                                  "(MM/DD)hh:mm:ss.uuuuuu"
 * 
 *              pAR_TIMETAG_TYPE timeIn          -- (input)  Source 64-bit time structure      
 *              pCEI_CHAR timeString             -- (output) Pointer to destination string    
 * 
 * RETURN VAL:  none
 *
 * DESCRIPTION: This routine converts the time value provided in the timeIn structure to a character string 
 *              representation of date/time, format based on what is specified via the displayFormat parameter.
 *              The supplied time format (LSB resolution) must be specified in the timeIn structure member 
 *              "timeTagFormat", representing the resolution of the respective "timeTag" member data.
 *
 *===================================================================================================================*/  
EXPORT32 CEI_VOID DLL_EXPORTED ar_convert_time_to_string 
( 
   CDEV_BOARD_TYPE board,                                                 /* (input)  Board number (0 to 127)        */
   CDEV_PARM_SSI_TYPE displayFormat,                                      /* (input)  Format of returned time string */
   pAR_TIMETAG_TYPE timeIn,                                               /* (input)  64-bit time-tag structure      */
   pCEI_CHAR timeString                                                   /* (output) Pointer to store time string   */ 
)
{
   /***********************************************************************
    *  Local variables
    ***********************************************************************/
   CEI_NATIVE_INT days;                                                   /* Number of days in time tag, 0 to 365(6) */
   CEI_NATIVE_INT daysSinceJan;                                           /* Number of days in time tag, 0 to 365(6) */
   CEI_NATIVE_INT leapday;                                                /* Leap year extra day tracker             */
   CEI_NATIVE_INT hours;                                                  /* Number of hours, 0 to 32                */
   CEI_NATIVE_INT minutes;                                                /* Number of minutes, 0 to 59              */
   CEI_NATIVE_UINT seconds;                                               /* Number of seconds, 0 to 59              */
   CEI_NATIVE_UINT microseconds;                                          /* Number of microseconds, 0 to 4294967296 */
   time_t   now;                                                          /* Current time                            */
   struct   tm start;                                                     /* Formatted "now" current time            */
   CEI_NATIVE_INT year;                                                   /* Current year (since 1900)               */
   TIME_TAG_TYPE timeTag64Value;                                          /* temp 64-bit time value placeholder      */
   CEI_NATIVE_INT month = 1;                                              /* The first month is January              */
   CEI_NATIVE_INT resulting_month = 1;                                    /* Just in case the day of the year should */
   CEI_NATIVE_INT resulting_day = 0;                                      /*   happen to be zero (which is illegal). */
   TIME_TAG_TYPE maxMicroSecondsPerYear;                                  /* Max number of usec in this year         */

   /* first convert the incoming time value to have a 1 microsecond resolution */
   timeTag64Value = timeIn->timeTag;
   switch ((CEI_NATIVE_INT)timeIn->timeTagFormat) {
      case AR_TIMETAG_INT_20USEC_32BIT: timeTag64Value *= (TIME_TAG_TYPE)20; break;
      case AR_TIMETAG_INT_MSEC_32BIT:   timeTag64Value *= (TIME_TAG_TYPE)1000; break;
      case AR_TIMETAG_HOST_SEC_32BIT:   timeTag64Value *= (TIME_TAG_TYPE)AR_MICROSECONDS_PER_SECOND; break;
      case AR_TIMER_X20_COMPAT_32BIT:
         if ( (board < MAX_BOARDS) && (deviceGlobal[board] != NULL) && (deviceGlobal[board]->attachedProcessCount > 0) ) {
            timeTag64Value *= (TIME_TAG_TYPE)(deviceGlobal[board]->deviceX20CompatibleResolution/AR_X20_SCALE_TIMETAG_VALUE); 
         }
         break;
      default: ;
   }
   
   /* Acquire the current year */
   time(&now);               /* Time in seconds since midnight Jan 1, 1970 GMT */
   start = *localtime(&now); /* Local time in seconds, min, hour, day, month, year */
   year = (CEI_NATIVE_INT)start.tm_year + AR_LOCAL_TIME_YEAR_BASIS;

   /* Determine if this is a leap year for total days in Feb and max microseconds/year */
   if ( ((year % AR_LEAP_YEAR_DIVISOR) == 0) && (year != AR_LEAP_YEAR_EXCEPTION) ) {
      leapday = 1;
      maxMicroSecondsPerYear = AR_IRIG_ROLLOVER_LEAP_YEAR;
   }
   else {
      leapday = 0;
      maxMicroSecondsPerYear = AR_IRIG_ROLLOVER_NON_LEAP_YEAR;
   }

   /* trim off years worth of microseconds if the supplied time value exceeds maximum microseconds per year */
   while (timeTag64Value > maxMicroSecondsPerYear) {
      timeTag64Value -= maxMicroSecondsPerYear;
   }

   /* convert the supplied time value into the appropriate Month/Day/Time components */
   if (timeTag64Value <= maxMicroSecondsPerYear) {
      microseconds = (CEI_NATIVE_UINT)(timeTag64Value % AR_MICROSECONDS_PER_SECOND);
      seconds = (CEI_NATIVE_UINT)(timeTag64Value / AR_MICROSECONDS_PER_SECOND);
   
      /* Extract Minutes from Seconds */
      minutes = seconds/60;
      seconds = seconds%60;
   
      /* Extract Hours from Minutes */
      hours = minutes/60;  
      minutes = minutes%60;
   
      /* Extract Days from Hours */
      days = hours/24;     
      hours = hours%24;    
   
      /* no such thing as day 0, days range from 1 to 365 (366 for leap yr) */
      days++;
      daysSinceJan = days;
      resulting_day = days;

      /* Compute the current month to extract current day-of-month */
      while ( days > 0 ) {
         resulting_day   = days;
         resulting_month = month;
   
         switch (month) {
            case 1:               /* Jan     */
               days -= 31; 
               break;             
            case 2:               /* Feb     */
               days -= (28 + leapday); 
               break;             
            case 3:               /* March   */
               days -= 31; 
               break;             
            case 4:               /* April   */
               days -= 30; 
               break;             
            case 5:               /* May     */
               days -= 31; 
               break;             
            case 6:               /* June    */
               days -= 30; 
               break;             
            case 7:               /* July    */
               days -= 31; 
               break;             
            case 8:               /* August  */
               days -= 31; 
               break;             
            case 9:               /* Sept    */
               days -= 30; 
               break;             
            case 10:              /* Oct     */
               days -= 31; 
               break;             
            case 11:              /* Nov     */
               days -= 30; 
               break;             
            case 12:              /* Dec     */
               days -= 31; 
               break;             
         }
         month++;  /* Go to next month */
      }
   }
   else {
      resulting_day   = 1;
      resulting_month = 1;
      daysSinceJan = 1;
      hours = 0;
      minutes = 0;
      seconds = 0;
      microseconds = 0;
   }

   /* Define the return string using only non-zero components (zero values omitted). */
   /* When all components are defined, format is based on the display format requested */
   switch ( displayFormat ) {
      case AR_TD_REL_MIDNIGHT:      /* Relative to Midnight Format "(ddd)hh:mm:ss.uuuuuu" */
      case AR_TD_IRIG:              /* Full IRIG Format "(ddd)hh:mm:ss.uuuuuu" */
         sprintf(timeString,"(%d)%2.2d:%2.2d:%2.2d.%6.6d", daysSinceJan, hours, minutes, seconds, microseconds);
         break;
   
      case AR_TD_DATE :             /* Date Format "(MM/dd)hh:mm:ss.uuuuuu" */
         sprintf(timeString,"(%d/%d)%2.2d:%2.2d:%2.2d.%6.6d", resulting_month, resulting_day, hours, minutes, seconds, microseconds);
         break;
   }
}


/*===================================================================================================================*
 * ENTRY POINT:                 A R _ C O N V E R T _ 1 5 5 3 _ T I M E _ T O _ S T R I N G   
 *===================================================================================================================*
 *
 * FUNCTION:    Convert a 1553 synchronized 64-bit time value to a character string
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board            -- (input)  Board number (0 to 127)  
 *              CDEV_PARM_SSI_TYPE displayFormat -- (input)  Format for the returned string:
 *                                                           AR_TD_REL_MIDNIGHT  Relative to Midnight Format and
 *                                                           AR_TD_IRIG          Full IRIG Format, both defined as:
 *                                                                                  "(DDD)hh:mm:ss.uuuuuu"           
 *                                                           AR_TD_DATE          Date Format defined as:
 *                                                                                  "(MM/DD)hh:mm:ss.uuuuuu"
 * 
 *              pAR_TIMETAG_TYPE timeIn          -- (input)  Source 64-bit time structure      
 *              pCEI_CHAR timeString             -- (output) Pointer to destination string    
 * 
 * RETURN VAL:  none
 *
 * DESCRIPTION: This routine converts the time value provided in the timeIn structure to a character string 
 *              representation of date/time, format based on what is specified via the displayFormat parameter.
 *              The supplied time format (LSB resolution) must be specified in the timeIn structure member 
 *              "timeTagFormat", representing the resolution of the respective "timeTag" member data.
 * 
 *              Since this routine is only provided for use with MIL-STD-1553/ARINC 429 Combo cards, the 
 *              timeIn parameter "timeTag" structure member will have a ONE NANOSECOND resolution.  The 
 *              timeIn parameter "referenceTimeTag" structure member will still have a one microsecond resolution, 
 *              as it retains the original CEI-x30 64-bit time-stamp associated with the MIL-STD-1553 timer/time-stamp
 *              value read from the host interface source common to the 1553/ARINC IP core.
 * 
 *===================================================================================================================*/  
EXPORT32 CEI_VOID  DLL_EXPORTED ar_convert_1553_time_to_string 
( 
   CDEV_BOARD_TYPE board,                                                 /* (input)  Board number (0 to 127)        */
   CDEV_PARM_SSI_TYPE displayFormat,                                      /* (input)  Format of returned time string */
   pAR_TIMETAG_TYPE timeIn,                                               /* (input)  64-bit time-tag structure      */
   pCEI_CHAR timeString                                                   /* (output) Pointer to store time string   */ 
)
{
#if defined (VXW_PCI_PPC) || defined (VXW_PCI_X86) || defined (_LINUX_X86_) || defined (_LINUX_PPC_) || defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
  #if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
    #ifdef MULTI_PROTOCOL_SUPPORT
      TIME_TAG_TYPE timeTag64Value = timeIn->timeTag; /* temp 64-bit time value placeholder */
      BusTools_TimeGetFmtString ( (CEI_INT32)(displayFormat) | 0x00000010, (CEI_UINT64 *)&timeTag64Value, timeString);
    #else
      timeIn->timeTag /= 1000;
      ar_convert_time_to_string (board, (CDEV_PARM_SSI_TYPE)(displayFormat & 3), timeIn, timeString);
    #endif
  #endif
#else
   TIME_TAG_TYPE timeTag64Value = timeIn->timeTag; /* temp 64-bit time value placeholder */

   /* map the BusTools-1553/API Dynamic Link Library... try 64-bit version if 32-bit isn't found */
   if (bt1553apiDllHandle == NULL) {
      bt1553apiDllHandle = LoadLibraryEx("busapi32.dll", NULL, 0);
      if (bt1553apiDllHandle == NULL) {
         bt1553apiDllHandle = LoadLibraryEx("busapi64.dll", NULL, 0);
      }
      /* if a DLL is found, map a pointer to the time-tag string format routine */
      if (bt1553apiDllHandle != NULL) {
         ptr_BusTools_TimeGetFmtString =
         (CEI_VOID (*)(CEI_INT32, CEI_UINT64 *, CEI_CHAR * ))GetProcAddress(bt1553apiDllHandle, "BusTools_TimeGetFmtString");
      }
   }

   /* if we have a reference to the 1553 API conversion routine, do the conversion */
   if (ptr_BusTools_TimeGetFmtString != NULL) {
      ptr_BusTools_TimeGetFmtString ( (CEI_INT32)(displayFormat) | 0x00000010, (CEI_UINT64 *)&timeTag64Value, timeString); 
   }
   else { /* BUSAPIxx.DLL is not available, at least convert a 1 usec value locally */
      timeIn->timeTag /= 1000;
      ar_convert_time_to_string (board, (CDEV_PARM_SSI_TYPE)(displayFormat & 3), timeIn, timeString);
   }
#endif
}


/*===================================================================================================================*
 * ENTRY POINT:                           A R _ G E T _ B A S E _ A D D R 
 *===================================================================================================================*
 *
 * FUNCTION:    This routine returns a pointer to the base address assigned to the device from the (plug and play) 
 *              host.  Note this routine does not support 64-bit operating systems and is not recommended for use.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board       -- (input) board to reference (0 to 127)
 *
 * RETURN VAL:  pCEI_VOID devicePtr -- pointer to the respective device host 
 *                                     interface location in virtual memory.
 * 
 * GLOBALS:     device[board].devicePtr
 * 
 * DESCRIPTION: Return the base address of the device.  Using the address returned, any register on the device can 
 *              be accessed directly.
 * 
 *===================================================================================================================*/
EXPORT32 pCEI_VOID DLL_EXPORTED ar_get_base_addr
(
   CDEV_BOARD_TYPE board                                                    /* (input) board to reference (0 to 127) */
)
{
   if (board < MAX_BOARDS) {
      return device[board].devicePtr;
   }
   else {
      return device[0].devicePtr;
   }
}


/*===================================================================================================================*
 * ENTRY POINT:                                A R _ G E T _ B O A R D N A M E
 *===================================================================================================================*
 *
 * FUNCTION:    Return a reference to a string describing the board type. 
 *                       
 * PARAMETERS:  CEI_UINT32 board     -- The board number of interest
 *              pCEI_CHAR  boardName -- Reference to store a board type string, consisting of one of the following:
 *
 *                                      "CEI-830", "CEI-430", "AMC-A30", "CEI-530", "RAR-CPCI", "RAR-EC", "RAR-PCIE",
 *                                      "R830RX",  "CEI-430A","RAR15-XMC", "RAR-XMC", "RCEI-830A", "RAR-MPCIE"
 *
 * RETURN VAL:  NULL                Board parameter value invalid or board has not been initialized
 *              pointer to char     Ptr to string for the board type detected, consisting of one of the following:
 *
 *                                      "CEI-830", "CEI-430", "AMC-A30", "CEI-530", "RAR-CPCI", "RAR-EC", "RAR-PCIE",
 *                                      "R830RX",  "CEI-430A","RAR15-XMC", "RAR-XMC", "RCEI-830A", "RAR-MPCIE"
 * GLOBALS:     deviceGlobal[]
 * 
 * DESCRIPTION: This routine returns this API-defined board name string for device referenced by the specified 
 *              board index.
 * 
 *===================================================================================================================*/ 
EXPORT32 pCEI_CHAR DLL_EXPORTED ar_get_boardname 
(
   CEI_UINT32 board,                                                       /* (input)  board to reference (0 to 127) */
   pCEI_CHAR  boardName                                                    /* (output) reference to store the name   */
) 
{
   CEI_UINT32 configuration;
   pCEI_CHAR localStrPtr = NULL;

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      configuration = deviceGlobal[board]->configuration;
      if ((configuration >= CEIDEV_FIRST_CEIX30_CONFIG) && (configuration <= CEIDEV_LAST_CEIX30_CONFIG)) {
         if (boardName != NULL) {
            strcpy (boardName, boardNames[configuration - CEIDEV_FIRST_CEIX30_CONFIG]); 
         }
         localStrPtr = boardNames[configuration - CEIDEV_FIRST_CEIX30_CONFIG];
      }
#ifndef TARGET_EMBEDDED
   }
#endif
   return localStrPtr;
}


/*===================================================================================================================*
 * ENTRY POINT:                             A R _ G E T _ B O A R D N A M E L V
 *===================================================================================================================*
 *
 * FUNCTION:    Return a reference to a string describing the board type. 
 *                       
 * PARAMETERS:  CEI_UINT32 board     -- The board number of interest
 *              pCEI_CHAR  boardName -- Reference to store a board type string, consisting of one of the following:
 *
 *                                      "CEI-830", "CEI-430", "AMC-A30", "CEI-530", "RAR-CPCI", "RAR-EC", "RAR-PCIE",
 *                                      "R830RX",  "CEI-430A","RAR15-XMC", "RAR-XMC", "RCEI-830A", "RAR-MPCIE"
 *
 * RETURN VAL:  CEI_INT32 -- ARS_NORMAL   Normal succesful completion.
 *                           ARS_INVBOARD Board parameter value invalid or
 *                                        board has not been initialized
 *                           ARS_INVARG   boardName parameter not valid memory
 * 
 * GLOBALS:     deviceGlobal[]
 * 
 * DESCRIPTION: This routine returns this API-defined board name string for device referenced by the specified 
 *              board index, defined specifically for compatibility with the LabVIEW API.
 * 
 *===================================================================================================================*/ 
EXPORT32 CEI_INT32 DLL_EXPORTED ar_get_boardnameLV 
(
   CEI_UINT32 board,                                                       /* (input)  board to reference (0 to 127) */
   pCEI_CHAR  boardName                                                    /* (output) reference to store the name   */
) 
{
   CEI_UINT32 configuration;
   CEI_INT32  status;

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      if (boardName != NULL) {
#endif
         configuration = deviceGlobal[board]->configuration;
         if ( (configuration < CEIDEV_FIRST_CEIX30_CONFIG) || (configuration > CEIDEV_LAST_CEIX30_CONFIG) ) {
            strcpy (boardName, "");
            status = ARS_INVBOARD;
         }
         else {
            strcpy (boardName, boardNames[configuration - CEIDEV_FIRST_CEIX30_CONFIG]); 
            status = ARS_NORMAL;
         }
#ifndef TARGET_EMBEDDED
      }
      else {
         status = ARS_INVARG;
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                                A R _ G E T _ B O A R D T Y P E
 *===================================================================================================================*
 *
 * FUNCTION:    Return the CEI-x30 board type for the specified board index.
 *                       
 * PARAMETERS:  CDEV_BOARD_TYPE board -- The board number of interest
 *
 *                                              CEI-830       19
 *                                              CEI-430       21
 *                                              AMC-A30       22
 *                                              RCEI-530      26
 *                                              R830RX        27
 *                                              RAR-CPCI      28
 *                                              RAR-EC        29
 *                                              RAR-PCIE      30
 *                                              CEI-430A      31
 *                                              CEI-830X820   34
 *                                              RAR-XMC       35
 *                                              RCEI830A      36
 *                                              RAR-MPCIE     38
 *                                              RAR-15       118
 *                                              RAR-15XT     119
 *
 * 
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL    Normal succesful completion.
 *                                ARS_INVBOARD  Board parameter value invalid or board has not been initialized
 * GLOBALS:     deviceGlobal[]
 * 
 * DESCRIPTION: This routine returns the CEI-x30 specific board type referenced by the specified board index.
 * 
 *===================================================================================================================*/ 
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_get_boardtype
(
   CDEV_BOARD_TYPE board                                                   /* (input)  board to reference (0 to 127) */
)
{
   CDEV_API_RET_TYPE configuration = ARS_INVBOARD;

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      switch (deviceGlobal[board]->configuration) {
         case CEIDEV_CONFIG_CEI830:
            configuration = CEI830;
            break;
         case CEIDEV_CONFIG_CEI430:
            configuration = CEI430;
            break;
         case CEIDEV_CONFIG_CEI430A:
            configuration = CEI430A;
            break;
         case CEIDEV_CONFIG_AMCA30:
            configuration = AMCA30;
            break;
         case CEIDEV_CONFIG_CEI530:
            configuration = CEI530;
            break;
         case CEIDEV_CONFIG_R830RX:
            configuration = R830RX;
            break;
         case CEIDEV_CONFIG_RAR_CPCI:
            configuration = RARCPCI;
            break;
         case CEIDEV_CONFIG_RAR_EC:
            configuration = RAREC;
            break;
         case CEIDEV_CONFIG_RAR_PCIE:
            configuration = RARPCIE;
            break;
         case CEIDEV_CONFIG_R830X820:
            configuration = CEI830X820;
            break;
         case CEIDEV_CONFIG_RAR_XMC:
            configuration = RARXMC;
            break;
         case CEIDEV_CONFIG_RCEI830A:
            configuration = RCEI830A;
            break;
         case CEIDEV_CONFIG_RAR_MPCIE:
            configuration = RARMPCIE;
            break;
         case CEIDEV_CONFIG_RAR15:
            configuration = RAR15;
            break;
         case CEIDEV_CONFIG_RAR15XT:
            configuration = RAR15XT;
            break;
         default:
            configuration = CEIDEV_CONFIGURATION_NULL;
      }
#ifndef TARGET_EMBEDDED
   }
#endif
   return configuration;
}             


/*===================================================================================================================*
 * ENTRY POINT:                   A R _ G E T _ C H A N N E L _ I N D E X _ I N F O
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve channel indexing information for the specified board.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE              board -- The board number of interest
 *              AR_CHANNEL_INDEX_INFO_TYPE * chan_index_info - structure containing the channel index information
 *                                                             based on the following definitions
 *              
 *              typedef enum 
 *              {
 *                 AR_CHAN_TYPE_429_RX = 0,                        a fixed receive channel
 *                 AR_CHAN_TYPE_429_TX = 1,                        a fixed transmit channel
 *                 AR_CHAN_TYPE_429_RX_OR_TX = 2,                  a s/w programmable channel
 *                 AR_CHAN_TYPE_UNDEFINED = 255                    non-existent channel
 *              } channelType;
 *
 *              typedef struct
 *              {
 *                 CEI_UINT32 num_channels;                        total number of channels               
 *                 CEI_UINT32 chan_type[CEIDEV_MAX_CHANNELS];      channelType enum value AR_CHAN_TYPE_429_<x> 
 *                 CEI_UINT32 chan_rx_index[CEIDEV_MAX_CHANNELS];  associated receive channel index       
 *                 CEI_UINT32 chan_tx_index[CEIDEV_MAX_CHANNELS];  associated transmit channel index      
 *              } AR_CHANNEL_INDEX_INFO_TYPE;
 *
 *              Each entry in "chan_type" (indexed 0 to "num_channels-1") is defined as a "channelType" value.
 *              For each "channelType" value, the following rule applies:
 *
 *              If chan_type[index] = AR_CHAN_TYPE_429_RX
 *                 the respective receive channel index = chan_rx_index[index]
 *                 the respective transmit channel index = N/A
 *
 *              Else If chan_type[index] = AR_CHAN_TYPE_429_TX
 *                 the respective transmit channel index = chan_tx_index[index]
 *                 the respective receive channel index = N/A
 *
 *              Else If chan_type[index] = AR_CHAN_TYPE_429_RX_OR_TX
 *                 the respective receive channel index = chan_rx_index[index]
 *                 the respective transmit channel index = chan_tx_index[index]
 *
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL    Normal succesful completion.
 *                                ARS_INVBOARD  Board parameter value invalid or
 *                                              board has not been initialized
 *
 * GLOBALS:     deviceGlobal[]
 * 
 * DESCRIPTION: This routine returns the CEI-x30 channel indexing configuration for the specified board.
 *
 *              For each chan_type entry, the respective chan_*x_index references the respective receive/transmit 
 *              channel index available.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_get_channel_index_info
(
   CDEV_BOARD_TYPE board,                                                    /* (input)  desired board id            */
   AR_CHANNEL_INDEX_INFO_TYPE * chan_index_info                              /* (output) receives channel index info */
)
{
   CEI_UINT32 index;                                                                   /* local Channel Type index   */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                              /* function return status     */
   AR_CHANNEL_INDEX_INFO_TYPE * channelInfoTablePtr = (AR_CHANNEL_INDEX_INFO_TYPE *)&deviceGlobal[board]->boardChannelIndexInfoStruct;
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      chan_index_info->num_channels = channelInfoTablePtr->num_channels;
      for (index = 0; index < CEIDEV_MAX_CHANNELS; index++) {
         chan_index_info->chan_type[index]     = channelInfoTablePtr->chan_type[index];
         chan_index_info->chan_rx_index[index] = channelInfoTablePtr->chan_rx_index[index];
         chan_index_info->chan_tx_index[index] = channelInfoTablePtr->chan_tx_index[index];
      }
#ifndef TARGET_EMBEDDED
   }
   else {
      status = ARS_INVBOARD;
   }
#endif
   return (status);
}


/*===================================================================================================================*
 * ENTRY POINT:                                    A R _ G E T _ C O N F I G
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve board and channel configuration attributes (for legacy compatibility with previous API's)
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input) board to reference (0 to 127)
 *              CDEV_PARM_SSI_TYPE item -- (input) item for which to return the 
 *                                                 current state:
 * 
 *                     ARU_RX_CH0*_BIT_RATE        ARU_CONFIGURATION
 *                     ARU_RX_CH1*_BIT_RATE        ARU_DISCRETE_IN1          
 *                     ARU_TX_CH0*_BIT_RATE        ARU_DISCRETE_IN2          
 *                     ARU_TX_CH1*_BIT_RATE        ARU_DISCRETE_IN3          
 *                     ARU_RX_CH0*_PARITY          ARU_DISCRETE_IN4          
 *                     ARU_RX_CH1*_PARITY          ARU_DISCRETE_OUT1         
 *                     ARU_TX_CH0*_PARITY          ARU_DISCRETE_OUT2         
 *                     ARU_TX_CH1*_PARITY          ARU_DISCRETE_OUT2         
 *                     ARU_TX_CH0*_SHUT_OFF        ARU_DISCRETE_OUT3         
 *                     ARU_TX_CH1*_SHUT_OFF        ARU_ACCESS_SNAPSHOT_BUFFER
 *                     ARU_TX_CH**_HB_INJ          ARU_IRIG_WRAP_ENABLE      
 *                     ARU_TX_CH**_LB_INJ          ARU_IRIG_AVAILABLE        
 *                     ARU_TX_CH**_GAP_INJ         ARU_IRIG_CALIBRATED       
 *                     ARU_FW_VERSION              ARU_IRIG_INPUT_TIME
 *                     ARU_SERIAL_NUMBER
 *                     ARU_RX_TIMETAG_MODE w/values of:                          
 *
 *                              AR_TIMETAG_EXT_IRIG_64BIT                         
 *                              AR_TIMETAG_INT_USEC_64BIT                         
 *                              AR_TIMETAG_INT_20USEC_32BIT                       
 *                              AR_TIMETAG_INT_MSEC_32BIT                         
 *                              AR_TIMER_X20_COMPAT_32BIT
 *                              AR_TIMETAG_SYNC_1553_CH1 (for combo cards only)               
 *                              AR_TIMETAG_SYNC_1553_CH2                
 *                              AR_TIMETAG_SYNC_1553_CH3                
 *                              AR_TIMETAG_SYNC_1553_CH4                
 *
 * RETURN VAL:  CEI_INT32 state -- the state of the item specified
 *                             ARS_INVARG      "item" parameter value is not supported by this API function
 *                             ARS_INVHARVAL   Channel value is not supported on device
 *                             ARS_INVBOARD    Board parameter value invalid or board has not been initialized
 *                             ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 * 
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: Supplies the current state of the "item" requested.
 *
 *===================================================================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_get_config 
(
   CDEV_BOARD_TYPE board,                                              /* board to reference (0 to 127)              */
   CDEV_PARM_SSI_TYPE item                                             /* item for which to return the current state */
)
{
   CEI_INT32 accessStatus = ARS_NORMAL;                                     /* status of device read/write access    */
   CDEV_PARM_SSI_TYPE value;                                                /* function return - item value          */
   CDEV_CHAN_TYPE channel = 0;                                              /* local item-derived channel index      */
   CEI_UINT32 state = 0;                                                    /* function return - item value          */
   CEI_UINT32 bitFieldValue;                                                /* temp for device register read content */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                   /* local invocation return status        */
   CEI_UINT32 deviceChannel;                                                /* current channel to select             */
   CEI_UINT32 memOffset;                                                    /* local offset to channel memory area   */
   CEI_FLOAT parametricValue;                                               /* FPGA parametric working variable      */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                        /* pointer to board local structure      */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                               /* pointer to board global structure     */
   CEI_UINT32 localItem = (CEI_UINT32)item;

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      /* assign references to this board's device and deviceGlobal structure-array elements */
      deviceLocalStructPtr = &device[board];
      deviceGlobalStructPtr = deviceGlobal[board];

      /*-----------------------------------------------------------------*
       *  CEI-x20 ARINC 429 Set/Get Config Support
       *-----------------------------------------------------------------*/
      /* Report receive channel bit rate */
      if ( ((localItem >= ARU_RX_CH01_BIT_RATE) && (localItem <= ARU_RX_CH32_BIT_RATE)) || 
           ((localItem >= ARU_OLD_RX_CH01_BIT_RATE) && (localItem <= ARU_OLD_RX_CH16_BIT_RATE)) ) {

         if (localItem >= ARU_RX_CH01_BIT_RATE) {
            channel = (CDEV_CHAN_TYPE)(localItem - ARU_RX_CH01_BIT_RATE);
         }
         else {
            channel = (CDEV_CHAN_TYPE)(localItem - ARU_OLD_RX_CH01_BIT_RATE);
         }
            
         status = ar_get_device_config (board, channel, (CDEV_PARM_SSI_TYPE)ARU_RX_BITRATE, &value);
         
         if (status == ARS_NORMAL) {
            if (value > ARU_SPEED_LOW) {
               if ((deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15) ||
                   (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15XT)) {
                  state = CEIDEV_20MHZ_REF_CLOCK / ((CEI_UINT32)value + 2U);
               }
               else {
                  state = CEIDEV_16MHZ_REF_CLOCK / ((CEI_UINT32)value + 2U);
               }
            }
            else
               state = (CEI_UINT32)value;
         }
         else
            state = (CEI_UINT32)status;
      }
      /* Report transmit channel bit rate */
      else if ( ((localItem >= ARU_TX_CH01_BIT_RATE) && (localItem <= ARU_TX_CH32_BIT_RATE)) || 
                ((localItem >= ARU_OLD_TX_CH01_BIT_RATE) && (localItem <= ARU_OLD_TX_CH16_BIT_RATE)) ) {
         
         if (localItem >= ARU_TX_CH01_BIT_RATE) {
            channel = (CDEV_CHAN_TYPE)(localItem - ARU_TX_CH01_BIT_RATE);
         }
         else {
            channel = (CDEV_CHAN_TYPE)(localItem - ARU_OLD_TX_CH01_BIT_RATE);
         }
            
         status = ar_get_device_config (board, channel, (CDEV_PARM_SSI_TYPE)ARU_TX_BITRATE, &value);
         
         if (status == ARS_NORMAL) {
            if (value > ARU_SPEED_LOW) {
               if ((deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15) ||
                   (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15XT)) {
                  state = CEIDEV_20MHZ_REF_CLOCK / ((CEI_UINT32)value + 2U);
               }
               else {
                  state = CEIDEV_16MHZ_REF_CLOCK / ((CEI_UINT32)value + 2U);
               }
            }
            else {
               state = (CEI_UINT32)value;
            }
         }
         else {
            state = (CEI_UINT32)status;
         }
      }                                    
      /* Report receive channel parity */
      else if ( ((localItem >= ARU_RX_CH01_PARITY) && (localItem <= ARU_RX_CH32_PARITY)) ||
                ((localItem >= ARU_OLD_RX_CH01_PARITY) && (localItem <= ARU_OLD_RX_CH16_PARITY)) ) {    
         
         if (localItem >= ARU_RX_CH01_PARITY) {
            channel = (CDEV_CHAN_TYPE)(localItem - ARU_RX_CH01_PARITY);
         }
         else {
            channel = (CDEV_CHAN_TYPE)(localItem - ARU_OLD_RX_CH01_PARITY);
         }
            
         status = ar_get_device_config (board, channel, (CDEV_PARM_SSI_TYPE)ARU_RX_PARITY, &value);

         if (status == ARS_NORMAL) {
            state = value == AR_ON ? AR_ODD : AR_OFF;
         }
         else {
            state = (CEI_UINT32)status;
         }
      }                                    
      /* Report transmit channel parity x20 style - requires accessing h/w directly */
      else if ( ((localItem >= ARU_TX_CH01_PARITY) && (localItem <= ARU_TX_CH32_PARITY)) ||
                ((localItem >= ARU_OLD_TX_CH01_PARITY) && (localItem <= ARU_OLD_TX_CH16_PARITY)) ) {    
         
         if (localItem >= ARU_TX_CH01_PARITY) {
            channel = (CDEV_CHAN_TYPE)(localItem - ARU_TX_CH01_PARITY);
         }
         else {
            channel = (CDEV_CHAN_TYPE)(localItem - ARU_OLD_TX_CH01_PARITY);
         }
            
         deviceChannel = (CEI_UINT32)channel + (CEI_UINT32)deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
         status = a429_utl_validate_channel_request (board, deviceChannel, ARU_TX_PARITY);
         if (status == ARS_NORMAL) {
            /* read the device register with respect to the attribute specified */
            memOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (deviceChannel * CEIDEV_CHANNEL_REGION_SIZE);
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &bitFieldValue);
            if (bitFieldValue & CEIDEV_PARITY_ENABLE) {
               state = (bitFieldValue & CEIDEV_PARITY_SELECT_EVEN) ? AR_EVEN : AR_ODD;
            }
            else {
               state = AR_RAW;
            }
         }
         else {
            state = (CEI_UINT32)status;
         }
      }                                    
      /* Report transmit channel enable state */
      else if ( ((localItem >= ARU_TX_CH01_SHUT_OFF) && (localItem <= ARU_TX_CH32_SHUT_OFF)) ||
                ((localItem >= ARU_OLD_TX_CH01_SHUT_OFF) && (localItem <= ARU_OLD_TX_CH16_SHUT_OFF)) ) {    
         
         if (localItem >= ARU_TX_CH01_SHUT_OFF) {
            channel = (CDEV_CHAN_TYPE)(localItem - ARU_TX_CH01_SHUT_OFF);
         }
         else {
            channel = (CDEV_CHAN_TYPE)(localItem - ARU_OLD_TX_CH01_SHUT_OFF);
         }
            
         status = ar_get_device_config (board, channel, (CDEV_PARM_SSI_TYPE)ARU_TX_DISABLE, &value);

         if (status == ARS_NORMAL) {
            state = (CEI_UINT32)value;
         }
         else {
            state = (CEI_UINT32)status;
         }
            
      }                                    
      /* Report transmit channel bit count error injection */
      else if ( ((localItem >= ARU_TX_CH01_HB_INJ) && (localItem <= ARU_TX_CH32_HB_INJ)) ||
                ((localItem >= ARU_OLD_TX_CH01_HB_INJ) && (localItem <= ARU_OLD_TX_CH16_HB_INJ)) ) {    
         
         if (localItem >= ARU_TX_CH01_HB_INJ) {
            channel = (CDEV_CHAN_TYPE)(localItem - ARU_TX_CH01_HB_INJ);
         }
         else {
            channel = (CDEV_CHAN_TYPE)(localItem - ARU_OLD_TX_CH01_HB_INJ);
         }
            
         status = ar_get_device_config (board, channel, (CDEV_PARM_SSI_TYPE)ARU_TX_BIT_ERROR, &value);
         if (status != ARS_NORMAL) {
            state = (CEI_UINT32)status;
         }
         else {
            state = (value == AR_HI) ? AR_ON : AR_OFF;
         }
      }                                    
      /* Report transmit channel bit count error injection */   
      else if ( ((localItem >= ARU_TX_CH01_LB_INJ) && (localItem <= ARU_TX_CH32_LB_INJ)) ||
                ((localItem >= ARU_OLD_TX_CH01_LB_INJ) && (localItem <= ARU_OLD_TX_CH16_LB_INJ)) ) {    
         
         if (localItem >= ARU_TX_CH01_LB_INJ) {
            channel = (CDEV_CHAN_TYPE)(localItem - ARU_TX_CH01_LB_INJ);
         }
         else {
            channel = (CDEV_CHAN_TYPE)(localItem - ARU_OLD_TX_CH01_LB_INJ);
         }
            
         status = ar_get_device_config (board, channel, (CDEV_PARM_SSI_TYPE)ARU_TX_BIT_ERROR, &value);
         if (status != ARS_NORMAL) {
            state = (CEI_UINT32)status;
         }
         else {
            state = (value == AR_LO) ? AR_ON : AR_OFF;
         }
      }                                    
      /* Report transmit channel message gap error injection */   
      else if ( ((localItem >= ARU_TX_CH01_GAP_INJ) && (localItem <= ARU_TX_CH32_GAP_INJ)) ||
                ((localItem >= ARU_OLD_TX_CH01_GAP_INJ) && (localItem <= ARU_OLD_TX_CH16_GAP_INJ)) ) {    
         
         if (localItem >= ARU_TX_CH01_GAP_INJ)  {
            channel = (CDEV_CHAN_TYPE)(localItem - ARU_TX_CH01_GAP_INJ);
         }
         else {
            channel = (CDEV_CHAN_TYPE)(localItem - ARU_OLD_TX_CH01_GAP_INJ);
         }
            
         status = ar_get_device_config (board, channel, (CDEV_PARM_SSI_TYPE)ARU_TX_GAP_ERROR, &value);
         if (status != ARS_NORMAL) {
            state = (CEI_UINT32)status;
         }
         else {
            state = (CEI_UINT32)value;
         }

      }
      /* return IRIG Receiver Raw Time-code Value */
      else if (localItem == ARU_IRIG_INPUT_TIME) {
#ifndef TARGET_EMBEDDED
         /* if multi-thread protection requested, semaphore/mutex-lock IRIG sample time register access */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            status = (CDEV_API_RET_TYPE)cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (status == ARS_NORMAL) {
#endif
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_IRIG_INPUT_TIME], (pCEI_UINT32)&bitFieldValue);  
            bitFieldValue &= 0x3FFFFFFF;
            state = bitFieldValue;
#ifndef TARGET_EMBEDDED
         }
         /* if multi-thread protection is active, free any acquired lock */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
         }
#endif
      }
      /* return the board serial number (currently applies only to RAR-EC) */
      else if (localItem == ARU_SERIAL_NUMBER) {
         state = (CEI_UINT32)deviceGlobalStructPtr->serialNumber;
      }
      /* return the parametric support state (fixed as not supported for x30 boards) */
      else if (localItem == ARU_PARAMETRIC_SUPPORT) {
         state = AR_OFF;
      }
#if defined DEBUG_PROGRAMMING
      else if (localItem ==  0x7995) {
         status = READ_32_BIT (board, &deviceLocalStructPtr->baseAddr[PLX_CONFIG_MODE_DMA_ARB_REG], &state);
      }
#endif
      else if (localItem == ARU_HW_FPGA_TEMPERATURE) {
         if (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_CEI430A) {
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_FPGA_FPGA_TEMPERATURE], (pCEI_UINT32)&bitFieldValue);
            ar_sleep (100);
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_FPGA_FPGA_TEMPERATURE], (pCEI_UINT32)&bitFieldValue);
            ar_sleep (100);
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_FPGA_FPGA_TEMPERATURE], (pCEI_UINT32)&bitFieldValue);
            if (bitFieldValue & 0x80) {
               bitFieldValue &= 0x7F;
               bitFieldValue = 127 - bitFieldValue;
               bitFieldValue *= (CEI_UINT32)-1;
            }
            state = bitFieldValue;
         }
         else if ((deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR_XMC) || (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR_MPCIE)) {
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_FPGA_FPGA_TEMPERATURE], (pCEI_UINT32)&bitFieldValue);
            bitFieldValue &= 0xFFF;
            parametricValue = ((CEI_FLOAT)bitFieldValue * ((CEI_FLOAT)503.975 / (CEI_FLOAT)4096.0)) - (CEI_FLOAT)273.15;
            state = (CEI_UINT32)((CEI_FLOAT)1000 * parametricValue); /* return temperature in millidegrees degrees C */
         }
         else {
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_FPGA_FPGA_TEMPERATURE], (pCEI_UINT32)&bitFieldValue);
            parametricValue = ((CEI_FLOAT)(bitFieldValue & 0xFFF) * ((CEI_FLOAT)503.975 / (CEI_FLOAT)1024.0)) - (CEI_FLOAT)273.15;
            state = (CEI_UINT32)((CEI_FLOAT)1000 * parametricValue); /* return temperature in millidegrees C */
         }
      }
      else { /* all remaining non-x20 legacy items are handled in the routine ar_get_device_config */
         status = ar_get_device_config (board, (CDEV_CHAN_TYPE)channel, (CDEV_PARM_SSI_TYPE)localItem, &value);
         if (status != ARS_NORMAL) {
            state = (CEI_UINT32)status;
         }
         else {
            state = (CEI_UINT32)value;
         }
      }
#ifndef TARGET_EMBEDDED
   }
   else {
      state = ARS_INVBOARD;
   }
#endif
   if (accessStatus) ;
   return state;
}
 

/*===================================================================================================================*
 * ENTRY POINT:                             A R _ G E T _ S T O R A G E _ M O D E
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the board-level receive data storage mode
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board        -- (input)  board index (0 to 127)
 *              CDEV_PARM_SSI_PTR_TYPE mode  -- (output) valid values are:
 *                                                       ARU_BUFFERED individual buffer per channel
 *                                                       ARU_MERGED   one merged circular buffer
 * 
 * RETURN VAL:  ARS_NORMAL        success
 *              ARS_INVBOARD      invalid board parameter or board not initialized
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine returns the current "global" storage mode to the application .
 *
 *              This routine is provided for backward compatibility with the CEI-710 API.  The proper method to 
 *              retrieve the Merged Receive Mode is provided for individual channels via the respective channel 
 *              specific invocation of the ar_get_device_config routine using the item parameter value 
 *              ARU_RX_MERGED_MODE and its ON/OFF selection.
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_get_storage_mode 
(
   CDEV_BOARD_TYPE board,                                      /* (input)  board index (0 to 127)                    */
   CDEV_PARM_SSI_PTR_TYPE mode                                 /* (output) storage mode (ARU_BUFFERED or ARU_MERGED) */
)
{
   CDEV_API_RET_TYPE status = ARS_NORMAL; /* function return status          */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      *mode = (CDEV_PARM_SSI_TYPE)deviceGlobal[board]->receiveMode;
#ifndef TARGET_EMBEDDED
   }
   else {
      status = ARS_INVBOARD;
   }
#endif
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                                A R _ Q U E R Y _ D E V I C E
 *===================================================================================================================*
 *
 * FUNCTION:    Open a session to the board and identify the board type
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board     -- (input)  device number to open (0-127)
 *              CDEV_BOARD_TYPE * boardType -- (output) if supported, the board type
 *
 *                                                  PRODUCT       DEFINE    VALUE
 *                                                  CEI-830     - CEI830      19
 *                                                  CEI-430     - CEI430      21
 *                                                  AMC-830     - AMC830      22
 *                                                  CEI-530     - CEI530      26
 *                                                  R-830RX     - R830RX      27
 *                                                  RAR-CPCI    - RARCPCI     28
 *                                                  RAR-EC      - RAREC       29
 *                                                  RAR-PCIE    - RARPCIE     30
 *                                                  CEI-430A    - CEI430A     31
 *                                                  CEI-830X820 - CEI830X820  34
 *                                                  RAR-XMC     - RARXMC      35
 *                                                  RCEI-830A   - RCEI830A    36
 *                                                  RP-708      - RP_708      37
 *                                                  RAR-MPCIE   - RARMPCIE    38
 *
 *                                               unsupported by the API but 
 *                                               reported includes the following:
 *
 *                                                  CEI-520     - CEI520       6 
 *                                                  CEI-620     - CEI620      10 
 *                                                  CEI-820     - CEI820      11 
 *                                                  CEI-715     - CEI715      16 
 *                                                  PMC-708     - PMC708      20
 *                                                  P-SER       - PSER       104
 *                                                  P-MIO       - PMIO       105
 *                                                  P-DIS       - PDIS       106
 *                                                  R15_MPCIE   - R15-MPCIE  107
 *                                                  AMC-1553    - QD_AMC1553 108
 *                                                  PCI-1553    - PCx1553    109
 *                                                  & cPCI-1553 
 *                                                  QPCI-1553   - QD_QPCI    110
 *                                                  QCP-1553    - QD_QCP     111
 *                                                  QPCX-1553   - QD_QPCX    112
 *
 *                                               if not, "UNSUPPORTED"
 *
 * RETURN VAL:  ARS_NORMAL      - Session was opened successfully
 *              ARS_DRIVERFAIL  - Driver Open/Close failed
 *              ARS_BOARD_MUTEX - failed to aquire or relinquish the global mutex
 *              ARS_FAILURE     - <Windows only> CEI_Install library failed to load in memory
 *
 * GLOBALS:     None
 *                                    
 * DESCRIPTION: This routine opens a session to the device and checks the PCI manufacturer and device ID.  If these 
 *              IDs are of supported x30 products, identify the board; otherwise report it as "unsupported".
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_query_device 
(
   CDEV_BOARD_TYPE board,                                                  /* (input)  device number to open (0-127) */
   CDEV_BOARD_TYPE * boardType                                             /* (output) if supported, the board type  */
)
{
   CEI_INT32  vendorIdent;              /* PCI Vendor ID for opened device   */
   CEI_INT32  deviceIdent;              /* PCI Device ID for opened device   */
   CDEV_API_RET_TYPE status = ARS_NORMAL; /* OS-specific function return status */

#ifndef TARGET_EMBEDDED
   if ( board < MAX_BOARDS ) {
#endif
      /*-------------------------------------------------------------------*
       * We are now ready to map and identify the device.
       *-------------------------------------------------------------------*/
      status = (CDEV_API_RET_TYPE)cei_acquire_device (board, (ARINC_LOCAL_TYPE *)&device[board]);
      
      /* validate the device configuration */
      if (status == ARS_NORMAL) {
         status = (CDEV_API_RET_TYPE)cei_acquire_device_identity ((CEI_INT32)board, &vendorIdent, &deviceIdent);
         if ( (status != ARS_FAILURE) && (vendorIdent == PCI_X30_MANUF_ID) ) {
            switch (deviceIdent) {
               case PCI_DEVICE_ID_830:
                  *boardType = CEI830;
                  break;
               case PCI_DEVICE_ID_830RX:
                  *boardType = R830RX;
                  break;
               case PCI_DEVICE_ID_830X820:
                  *boardType = CEI830X820;
                  break;
               case PCI_DEVICE_ID_RCEI830A:
                  *boardType = RCEI830A;
                  break;
               case PCI_DEVICE_ID_A30:
                  *boardType = AMCA30;
                  break;
               case PCI_DEVICE_ID_430:
                  *boardType = CEI430;
                  break;
               case PCI_DEVICE_ID_430A:
                  *boardType = CEI430A;
                  break;
               case PCI_DEVICE_ID_530:
                  *boardType = CEI530;
                  break;
               case PCI_DEVICE_ID_630:
                  *boardType = RARCPCI;
                  break;
               case PCI_DEVICE_ID_RAREC:
                  *boardType = RAREC;
                  break;
               case PCI_DEVICE_ID_RAR_PCIE:
                  *boardType = RARPCIE;
                  break;
               case PCI_DEVICE_ID_RAR_XMC:
                  *boardType = RARXMC;
                  break;
               case PCI_DEVICE_ID_RAR_MPCIE:
                  *boardType = RARMPCIE;
                  break;
               case PCI_DEVICE_ID_RAR15:
                  *boardType = RAR15;
                  break;
               case PCI_DEVICE_ID_RAR15XT:
                  *boardType = RAR15XT;
                  break;
               case PCI_DEVICE_ID_708 :
                  *boardType = PMC708;
                  break;
               case PCI_DEVICE_ID_RP708 :
                  *boardType = RP_708;
                  break;
               case PCI_DEVICE_ID_520 :
                  *boardType = CEI520;
                  break;
               case PCI_DEVICE_ID_620 :
                  *boardType = CEI620;
                  break;
               case PCI_DEVICE_ID_820 :
                  *boardType = CEI820;
                  break;
               case PCI_DEVICE_ID_PSER:
                  *boardType = PSER;
                  break;
               case PCI_DEVICE_ID_PDIS:
                  *boardType = PDIS;
                  break;
               case PCI_DEVICE_ID_AMC :
                  *boardType = QD_AMC1553;
                  break;
               case PCI_DEVICE_ID_1553:
                  *boardType = PCx1553;
                  break;
               case PCI_DEVICE_ID_QPCI:
                  *boardType = QD_QPCI;
                  break;
               case PCI_DEVICE_ID_QCP :
                  *boardType = QD_QCP;
                  break;
               case PCI_DEVICE_ID_QPCX:
                  *boardType = QD_QPCX;
                  break;
               case PCI_DEVICE_ID_R15EC:
                  *boardType = QD_R15EC;
                  break;
               case PCI_DEVICE_ID_RXMC:
                  *boardType = QD_RXMC;
                  break;
               case PCI_DEVICE_ID_RPCIE:
                  *boardType = QD_RPCIE;
                  break;
               case PCI_DEVICE_ID_LPCIE:
                  *boardType = QD_LCPIE;
                  break;
               case PCI_DEVICE_ID_RXMC2:
                  *boardType = QD_RXMC2;
                  break;
               case PCI_DEVICE_ID_R15MPCIE:
                  *boardType = QD_R15MPCIE;
                  break;
               default:
                  *boardType = UNSUPPORTED;
            }
         }
         else
            *boardType = UNSUPPORTED;

         status = (CDEV_API_RET_TYPE)cei_release_device(board, (ARINC_LOCAL_TYPE *)&device[board]);
      }
#ifndef TARGET_EMBEDDED
   }
   else  {
      status = ARS_INVBOARD;
   }
#endif

   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                                      A R _ V E R S I O N
 *===================================================================================================================*
 *
 * FUNCTION:    Return the software version
 * 
 * PARAMETERS:  pCEI_CHAR verstr -- (output) string containing version.
 * 
 * RETURN VAL:  CEI_VOID 
 *
 * GLOBALS:     none
 * 
 * DESCRIPTION: The software version API_VERSION defined in cdev_api.h is copied to the string parameter passed into 
 *              this function.
 * 
 *===================================================================================================================*/ 
EXPORT32 CEI_VOID DLL_EXPORTED ar_version
(
   pCEI_CHAR verstr
)
{
   strcpy (verstr, API_VERSION);
}                  
    


