/*====================================================================================================================*
 * FILE:                                       C D E V _ W I N . C
 *====================================================================================================================*
 *
 *   COPYRIGHT 2005 - 2019 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED
 *   ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH THE INCLUSION
 *   OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY OTHER COPIES THEREOF
 *   MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE
 *   AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS, INC. 
 *
 *====================================================================================================================*
 *
 * FUNCTION:    Low-level Application Programmer's Interface for a CEI-x30
 *              ARINC device operating with all Windows Operating Systems
 *
 *    Internal CEI-x30 routines implemented herein:
 *
 *    cei_acquire_device          - opens a session with the specified x30 device
 *    cei_acquire_device_identity - supplies the PCI configuration space ID values
 *    cei_acquire_thread_lock     - acquires sempahore/mutex lock
 *    cei_delay_milliseconds      - o/s specific delay utility in milliseconds
 *    cei_delay_seconds           - o/s specific delay utility in fractional seconds
 *    cei_get_host_system_time_x  - acquires host system time-of-year 64-bit uSec     
 *    cei_get_host_system_time    - acquires host system time-of-year 32-bit mSec     
 *    cei_get_irig_initial_time   - acquires host system time-of-year IRIG format
 *    cei_map_shared_memory       - maps memory region for sharing globals across processes
 *    cei_open_shared_lock        - creates a shared semaphore across processes
 *    cei_release_device          - closes a session with the specified x30 device
 *    cei_release_thread_lock     - releases sempahore/mutex lock     
 *    cei_read_32bit              - read a 32-bit value from an x30 device
 *    cei_write_32bit             - write a 32-bit value to an x30 device
 *    cei_init_interrupt          - configures an interrupt and ISR
 *    cei_terminate_interrupt     - terminate an interrupt
 *          
 *    internal routines previously implemented in CEI_LL library:      
 *
 *    WaitForInterrupts           - isr thread awoken by device driver generated event
 *    ceill_initInterrupt         - associates the isr with the above thread, sets up interrupt for CEI_Install
 *    ceill_terminateInterrupt    - disconnects the isr from the driver
 *
 *--------------------------------------------------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date    By   Comments
 * -------- ----- ----------------------------------------------------------------------------------------------------
 * 11/06/03  BDW  Created.
 * 04/11/06  BDW  Removed utility routines cei_read_plx_32bit & cei_write_plx_32bit
 * 08/24/06  BDW  Incorporated CEI_TYPES.H data type definitions in 
 *                   support of 64-bit Linux operating systems. 
 *                Swapped the BAR reference order by the device structure 
 *                   members baseAddr and devicePtr
 * 02/22/07  BDW  Added cei_acquire_device_identity() to supply the PCI 
 *                   Manufacturer and Device IDs for x30 products 
 * 10/29/07  BDW  Added routine for open error handling, cei_get_open_error
 * 02/18/08  BDW  Removed the boardInitialized array, relocated as a structure
 *                   member in the data structure array "deviceGlobal"
 *                Added new routines for multi-process support: 
 *                   cei_open_shared_lock and cei_map_shared_memory
 * 03/10/08  BDW  Removed the declaration of the following global arrays, 
 *                relocated to CDEV_GLB.H:
 *                   bypassWrapTest
 *                   irigDacThreshold
 *                   a717syncWords
 *                Removed declaration of basicConfiguration, it was unused.
 *                Added reference to device[] structure as a parameter to
 *                   cei_acquire_device & cei_release_device so they no longer access 
 *                   the global data structure directly.
 * 04/15/08  BDW  Added support for initiation and termination of h/w interrupts
 *                   via routines cei_init_interrupt and cei_terminate_interrupt.
 * 06/16/09  BDW  Added the function cei_open_preload_lock for multiprocess support.
 * 07/01/09  BDW  Added new function cei_delay_milliseconds (o/s thread suspend).
 * 10/28/10  BDW  Drastic alterations to switch from CEI_LL library to CEI_Install
 *                   for all driver interface operations.
 *                Modified the parameter list for cei_init_interrupt to account for
 *                   interrupt setup requirements in the Windows-only CEI_Install library
 * 01/05/11  BDW  Corrected successful interrupt initialization return status 
 *                   to force the value ARS_NORMAL from ceill_initInterrupt
 * 02/17/12  BDW  Modified the initialization of ceiAccessLock[] to allow MAX_BOARDS of 128
 * 07/16/12  BDW  Modified the cei_init_interrupt argument list to include board type
 * 07/21/12  BDW  Modified ceill_initInterrupt to add support for ARINC/1553 Combo board  
 *                shared PCI Interrupt driver setup.
 * 09/16/13  BDW  Replaced CreateThread with _beginthreadex and ExitThread with
 *                _endthreadex; terminated PCI Interrupt thread on error instead
 *                of just returning to non-existent calling function.
 * 12/13/16  BDW  Modified cei_release_device to allow release of shared memory and blocking 
 *                mutex resources independent of memory+device dependent on successful session open.
 * 10/26/19  BDW  In the routine cei_acquire_device, replaced the return status for a failure to 
 *                load CEI_Install.dll, previously ARS_FAILURE, with ARS_DRIVERFAIL.
 *
 *====================================================================================================================*/ 

/*--------------------------------------------------------------------------------------------------------------------*
 *                                           GENERAL INCLUDE STATEMEMTS
 *--------------------------------------------------------------------------------------------------------------------*/

/* include installation library dynamic link support */
#define CEI_INSTALL_DYNAMIC_LINK_SUPPORT
#include "cei_install.h"

#include "cdev_glb.h"
#include <process.h>
#include <mmsystem.h>
#include <sys/types.h>
#include <sys/timeb.h>

/*========================================================================* 
 * Local data and defines
 *========================================================================*/
#define MAX_DEV_OPEN_ERROR_STR_LEN 512

/* global: pData                                                               */
/* usage:  Structure of information passed to the hardware interrupt thread    */
typedef struct _THREAD_DATA_STRUCT {
   CEI_DEV_HANDLE handle;    /* Handle to current device.                      */
   CEI_DEV_HANDLE hIntThrd;  /* Handle to the interrupt thread                 */
   pCEI_CHAR HWaddress;      /* Save the device BAR0 base address.             */
   pCEI_CHAR configAddr;     /* Save the device BAR2 address.                  */
   CEI_INT32 fExitThread;    /* Set TRUE to cause the threads to exit.         */
   CEI_INT32 DeviceIntOpen;  /* WinRT device Interrupt open count.             */
   CEI_INT32 board;          /* Board number being processed.                  */
   void (_stdcall * interrupt_function)(CEI_INT32 deviceIndex); /* Delayed ISR */
   CEI_UINT32 vendorID;      /* Vendor ID if PCI card                          */
   CEI_UINT32 deviceID;      /* Device ID if PCI card                          */
   CEI_UINT32 statusRegIndex;/* Index into Status Register Array for this dev  */
} THREAD_DATA_STRUCT, *PTHREAD_DATA_STRUCT;

static THREAD_DATA_STRUCT pData[MAX_BOARDS];

/* global: virtualBaseAddress                                */
/* usage:  Local reference to the device's three potentially */
/*         mapped host interface Base Address Regions        */
static CEI_UINT64 virtualBaseAddress[MAX_BOARDS][3];

/* global: ceiAccessLock                         */
/* usage:  Host Interface Access blocking mutex  */
static CEI_DEV_HANDLE ceiAccessLock[MAX_BOARDS] = { 
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
};
   
/* global: deviceOpenErrorDesc                                                */
/* usage:  storage for Windows driver low-level board mapping error messages  */
static CEI_INSTALL_ERR deviceOpenErrorDesc = "No error description available";

/* global: dev_handle                         */
/* usage:  Stores handle to each open device. */
static CEI_DEV_HANDLE dev_handle[MAX_BOARDS];

/* global: cei_dll                                             */
/* usage:  Pointer to CEI-Install library dynamic link module. */
static CEI_INSTALL_MODULE *cei_dll = NULL;


/*========================================================================* 
 * Windows Operating System specific functionality
 *========================================================================*/

/*===========================================================================*
 * INTERNAL ROUTINE:               e r r o r _ b o x 
 *===========================================================================*
 *
 * FUNCTION:    Display error description.
 *
 * RETURN VAL:  none
 *
 * DESCRIPTION: Used to display a low-level error message to the user when 
 *              propagation of the error message to the caller is impractical.
 *              The message text is displayed using the MessageBox routine.
 *
 *===========================================================================*/
CEI_VOID error_box(const CEI_CHAR *format_str,...)
{
    /* receives message string */
    CEI_CHAR text[1024];

    /* store error description */
    va_list argp;
    va_start(argp,format_str);
    _vsnprintf(text,sizeof(text) - 1,format_str,argp);
    text[sizeof(text)-1] = 0;

    /* display a message box  */
    MessageBox(NULL,text,"CEI-x30-SW Error",MB_SYSTEMMODAL|MB_OK|MB_ICONERROR);
}


/*===========================================================================*
 *  PROCEDURE - WaitForInterrupts
 *===========================================================================*
 *
 *  FUNCTION
 *     This thread waits on an interrupt event from WinRT, calling the 
 *     API-supplied interrupt service routine when an interrupt occurs.
 *
 *  PARAMETERS
 *     Param - THREAD_DATA_STRUCT struct pointer.
 *
 *   Notes:  To kill this thread, signal the hCloseEvent event.
 *
 *  RETURNS
 *     Thread/Interrupt Blocking fail status.
 *===========================================================================*/
unsigned __stdcall WaitForInterrupts (void * Param)
{
   PTHREAD_DATA_STRUCT pData;
   CEI_INT32           board;
   CEI_UINT64          wait_on_regs; /* bit field specifying which int status register(s) to wait on */
   CEI_UINT32          wait_result;  /* receives wait operation result */
   CEI_UINT64          int_on_regs;  /* receives bitfield indicating which int status register(s) generated the int */
   CEI_INSTALL_ERR     cei_err;      /* receives error description upon failure */

   pData = (PTHREAD_DATA_STRUCT)Param;  /* Convert the input parameter. */
   board = pData->board;
   wait_on_regs = pData->statusRegIndex;

   /* handle interrupts until the process calls LL_UnhookInterrupt 22 */
   while (TRUE) {
      /* wait for a pending interrupt */
      if (cei_dll->LL_WaitForInterrupt (pData->handle, wait_on_regs, &wait_result, &int_on_regs, NULL, cei_err) == FALSE) {
         _endthreadex (ARS_INT_ISR);
      }
      
      /* bail out if the process called LL_UnhookInterrupt */
      if (wait_result == CEI_WIR_INT_DISABLED) {
         break;
      }
      else if (wait_result != CEI_WIR_INT_OCCURRED) {
         _endthreadex (ARS_INT_ISR);
      }

      /* an interrupt has occurred, handle it  */
      pData->interrupt_function(board);
   }
   
   /* signal the main thread this thread is done */
   _endthreadex(0);
   return 0;
}


/*===========================================================================*
 * INTERNAL ROUTINE: c e i l l _ i n i t I n t e r r u p t
 *===========================================================================*
 *
 * FUNCTION:    Provides the method to associate an ISR with an interrupt
 * 
 * PARAMETERS:  CEI_INT32   deviceId       -- Device Number to map
 *              CEI_INT32   _stdcall isr   -- Address of interrupt function to be called
 *              CEI_INT32   priority       -- Interrupt Service Routine thread priority
 *              CEI_INT32   memRegion      -- PCI BAR region/index for interrupt clear bit (0 - BAR0 or 1 - BAR2)
 *              CEI_UINT32  disOffset      -- PCI Interrupt Disable register (PLX reg offset, 0 for no PLX interface)
 *              CEI_UINT32  comboChanCount -- 1553/ARINC Combo Board configuration
 *              pCEI_UINT32 isrEnabled     -- Thread successfully defined state (TRUE, FALSE)
 * 
 * RETURN VAL:  CEI_INT32 STATUS -        ARS_DRIVERFAIL
 *                                        ARS_NO_INT_SUPPORT
 *                                        ARS_NO_INT_ENABLED
 *                                        ARS_INT_ISR
 *
 * GLOBALS:     none
 * 
 * DESCRIPTION: Connects the user-specified interrupt function to the  
 *              interrupt assigned to the respective device.
 * 
 *===========================================================================*/
CEI_INT32 ceill_initInterrupt 
(
   CEI_INT32   deviceId,
   CEI_INT32   (_stdcall *isr)(CEI_INT32), 
   CEI_INT32   priority,
   CEI_INT32   memRegion,
   CEI_UINT32  disOffset,
   CEI_UINT32  comboChanCount,
   pCEI_UINT32 isrEnabled
)
{
   CEI_INT32 status = ARS_NORMAL;
   unsigned  ThreadID;                  /* Unique id of the interrupt thread */
   HANDLE    hThread;                   /* Handle to the interrupt thread    */
   CEI_UINT32 ints_supported;     /* local hardware interrupt support status */
   CEI_INT_CONTROL_V1 int_cntrl; /* low-level interrupt management structure */
   CEI_UINT32 preexisted;     /* device interrupt in use by existing process */
   CEI_INSTALL_ERR cei_err;                /* error description upon failure */
   CEI_DEV_HANDLE devHandle = dev_handle[deviceId]; /* ref to global dev handle */
   CEI_UINT32 chan, i;        /* channel indices for combo 1553/ARINC boards */
   CEI_UINT32 cntrl_reg_len;   /* temp variables for combo 1553/ARINC boards */
   CEI_UINT32 cntrl_reg_int_chk_offset;
   CEI_UINT32 cntrl_reg_int_dis_offset;
   CEI_UINT32 cntrl_reg_int_disable_mask;
   CEI_UINT32 cntrl_reg_int_clear_offset;
   CEI_UINT32 cntrl_reg_int_clear_mask;
   CEI_UINT64 cap_reg_mask;


   /* query hardware interrupt support capability */
   if (cei_dll->LL_GetCapability(devHandle, CEI_CAP_HW_INTS, &ints_supported, cei_err) == FALSE) {
      error_box("Failed to acquire capabilities for Device '%ld' (details: %s).", deviceId, cei_err);
      status = ARS_DRIVERFAIL;
   }
   
   /* bail out if hardware interrupts are not supported */
   else if (!ints_supported) {
      error_box("Hardware interrupts are not supported for Device '%ld'.", deviceId);
      status = ARS_NO_INT_SUPPORT;
   }
   else {
      /* initialize the interrupt control structure */
      if (cei_dll->LL_InitInterruptControlStruct(devHandle, CEI_INT_CONTROL_VER_V1, sizeof(int_cntrl), &int_cntrl, cei_err) == FALSE) {
         error_box("Failed to initialize an interrupt for Device '%ld' (details: %s).", deviceId, cei_err);
         status = ARS_NO_INT_ENABLED;
      }
   }

   if (status == ARS_NORMAL) {
      /* any combo 1553/ARINC board requires interrupt setup for all 1553 channels in addition to the 
       * ARINC 429 interrupt AND it must match identically to the respective 1553 API interrupt setup. */
      if (comboChanCount > 0) {
         int_cntrl.chk_cap_reg = TRUE;
         int_cntrl.cap_reg.rgn_usr_addr = (CEI_UINT64)virtualBaseAddress[deviceId][memRegion];
         int_cntrl.cap_reg.offset_bytes = 0;
         int_cntrl.num_status_reg = comboChanCount + 1;
         int_cntrl.num_disable_reg = (comboChanCount * 4) + 1;

         cntrl_reg_len = 4;
         cntrl_reg_int_disable_mask = 0x40004000;
         cntrl_reg_int_clear_mask = 0x40000200;  /* operation: "clear" UCA interrupt pending bit "9" */

         for (chan = 0; chan < comboChanCount; chan++) {
            if (chan == 0)
               cap_reg_mask = 0x1C0;  /* channels 1 thru 4 */
            else if(chan == 1)
               cap_reg_mask = 0x180;  /* channels 2 thru 4 */
            else
               cap_reg_mask = 0x100;  /* channel 4 */
   
            cntrl_reg_int_chk_offset = 0x00030000 + (chan * 0x4000) + 0x58;
            cntrl_reg_int_dis_offset = 0x00030000 + (chan * 0x4000) + 0x58;
            /* UCA32: "1553_control_reg:9" = clear HWINT on specified channel */
            cntrl_reg_int_clear_offset = cntrl_reg_int_chk_offset;

            int_cntrl.cap_reg.len_bytes = cntrl_reg_len;
            /* "1553_control_reg:9" = check for HWINT on specified channel */
            int_cntrl.status_reg[chan].cap_reg_mask          = cap_reg_mask;
            int_cntrl.status_reg[chan].chk_reg_mask          = 0x200;
            int_cntrl.status_reg[chan].chk_reg.offset_bytes  = cntrl_reg_int_chk_offset; 
            int_cntrl.status_reg[chan].chk_reg.len_bytes     = cntrl_reg_len;
            int_cntrl.status_reg[chan].chk_reg.rgn_usr_addr  = int_cntrl.cap_reg.rgn_usr_addr;
            int_cntrl.status_reg[chan].chk_reg_opt           = CEI_XRO_PROP_CHK_EN | ((CEI_UINT64)0x4000 << 32);
            /* clear HWINT on specified channel  */
            int_cntrl.status_reg[chan].clr_reg_mask          = cntrl_reg_int_clear_mask;
            int_cntrl.status_reg[chan].clr_reg.offset_bytes  = cntrl_reg_int_clear_offset;
            int_cntrl.status_reg[chan].clr_reg.len_bytes     = cntrl_reg_len;
            int_cntrl.status_reg[chan].clr_reg.rgn_usr_addr  = int_cntrl.cap_reg.rgn_usr_addr;
            int_cntrl.status_reg[chan].clr_reg_opt           = CEI_CRO_SET;
            
            /* configure "CEI_INT_DISABLE_REG_V1", need 4 registers for 4 transactions per channel */
            /* "1553_control_reg:14" = disable hardware interrupts for the specified channel */
            i = (chan*4);
            int_cntrl.disable_reg[i].cap_reg_mask          = cap_reg_mask;
            int_cntrl.disable_reg[i].dis_reg_mask          = (CEI_UINT64)cntrl_reg_int_disable_mask;
            int_cntrl.disable_reg[i].dis_reg.offset_bytes  = cntrl_reg_int_dis_offset; 
            int_cntrl.disable_reg[i].dis_reg.len_bytes     = cntrl_reg_len;
            int_cntrl.disable_reg[i].dis_reg.rgn_usr_addr  = int_cntrl.cap_reg.rgn_usr_addr;
            int_cntrl.disable_reg[i].dis_reg_opt           = CEI_DRO_SET;
            /* "write_interrupt_bit:9" = clear HWINT on specified channel  */
            i = (chan*4) + 1;
            int_cntrl.disable_reg[i].cap_reg_mask          = cap_reg_mask;
            int_cntrl.disable_reg[i].dis_reg_mask          = (CEI_UINT64)cntrl_reg_int_clear_mask;
            int_cntrl.disable_reg[i].dis_reg.offset_bytes  = cntrl_reg_int_clear_offset;
            int_cntrl.disable_reg[i].dis_reg.len_bytes     = cntrl_reg_len;
            int_cntrl.disable_reg[i].dis_reg.rgn_usr_addr  = int_cntrl.cap_reg.rgn_usr_addr;
            int_cntrl.disable_reg[i].dis_reg_opt           = CEI_CRO_SET;
            /* "1553_control_reg:14" = disable hardware interrupts for the specified channel */
            i = (chan*4) + 2;
            int_cntrl.disable_reg[i].cap_reg_mask          = cap_reg_mask;
            int_cntrl.disable_reg[i].dis_reg_mask          = (CEI_UINT64)cntrl_reg_int_disable_mask;
            int_cntrl.disable_reg[i].dis_reg.offset_bytes  = cntrl_reg_int_dis_offset; 
            int_cntrl.disable_reg[i].dis_reg.len_bytes     = cntrl_reg_len;
            int_cntrl.disable_reg[i].dis_reg.rgn_usr_addr  = int_cntrl.cap_reg.rgn_usr_addr;
            int_cntrl.disable_reg[i].dis_reg_opt           = CEI_DRO_SET;
            /* "write_interrupt_bit:9" = clear HWINT on specified channel  */
            i = (chan*4) + 3;
            int_cntrl.disable_reg[i].cap_reg_mask          = cap_reg_mask;
            int_cntrl.disable_reg[i].dis_reg_mask          = (CEI_UINT64)cntrl_reg_int_clear_mask;
            int_cntrl.disable_reg[i].dis_reg.offset_bytes  = cntrl_reg_int_clear_offset;
            int_cntrl.disable_reg[i].dis_reg.len_bytes     = cntrl_reg_len;
            int_cntrl.disable_reg[i].dis_reg.rgn_usr_addr  = int_cntrl.cap_reg.rgn_usr_addr;
            int_cntrl.disable_reg[i].dis_reg_opt           = CEI_CRO_SET;
         }

         /* The ARINC 429 "channel" has h/w interrupt support via the single 4-byte global enable */
         /* register at byte offset 0 in the memory region specified in the priority parameter. */
         /* Global Enable reg Bit 4 is non-zero upon interrupt, no chk_reg_opt option flags needed */
         int_cntrl.status_reg[chan].cap_reg_mask          = 0x1C0;  /* channels 1 thru 4 */
         int_cntrl.status_reg[chan].chk_reg.rgn_usr_addr  = virtualBaseAddress[deviceId][memRegion];
         int_cntrl.status_reg[chan].chk_reg.offset_bytes  = CEIDEV_RAR15_429_BYTE_OFFSET;
         int_cntrl.status_reg[chan].chk_reg.len_bytes     = 4;
         int_cntrl.status_reg[chan].chk_reg_opt           = CEI_XRO_NONE;
         int_cntrl.status_reg[chan].chk_reg_mask          = CEIDEV_INTERRUPT_PENDING;
      
         /* upon interrupt, clear the interrupt pending bit in the same register */
         int_cntrl.status_reg[chan].clr_reg.rgn_usr_addr  = virtualBaseAddress[deviceId][memRegion];
         int_cntrl.status_reg[chan].clr_reg.offset_bytes  = CEIDEV_RAR15_429_BYTE_OFFSET;
         int_cntrl.status_reg[chan].clr_reg.len_bytes     = 4;
         int_cntrl.status_reg[chan].clr_reg_opt           = CEI_CRO_SET;
         int_cntrl.status_reg[chan].clr_reg_mask          = CEIDEV_GLOBAL_ENABLE_AUTON_CLEAR + CEIDEV_INTERRUPT_PENDING;

         /* upon abnormal application termination, clear the interrupt enable bit in the Global Enable reg */
         i = (CEI_UINT32)int_cntrl.num_disable_reg - 1;
         int_cntrl.disable_reg[i].cap_reg_mask         = 0x1C0;  /* channels 1 thru 4 */;
         int_cntrl.disable_reg[i].dis_reg.rgn_usr_addr = virtualBaseAddress[deviceId][memRegion];
         int_cntrl.disable_reg[i].dis_reg.offset_bytes = CEIDEV_RAR15_429_BYTE_OFFSET;
         int_cntrl.disable_reg[i].dis_reg.len_bytes    = 4;
         int_cntrl.disable_reg[i].dis_reg_opt          = CEI_DRO_SET;
         int_cntrl.disable_reg[i].dis_reg_mask         = CEIDEV_GLOBAL_ENABLE_AUTON_CLEAR + CEIDEV_INTERRUPT_PENDING + CEIDEV_INTERRUPTS_ENABLED; /* clear int enable and pending bits */
         
         pData[deviceId].statusRegIndex = 1 << chan;
      }
      else {  /* all other CEI-x30 boards (non-combo 1553 products) */
         pData[deviceId].statusRegIndex = 1;
         
         /* set main interrupt control parameters */
         int_cntrl.chk_cap_reg = FALSE; /* no capability register checking           */
         int_cntrl.num_status_reg = 1;  /* device uses a single int status register  */
         int_cntrl.num_disable_reg = 1; /* device uses a single int disable register */
      
         /* All x30 devices have h/w interrupt support via the single 4-byte global enable 
          * register at byte offset 0 in the memory region specified in the priority parameter.
          * Global Enable reg Bit 4 is non-zero upon interrupt, no chk_reg_opt option flags needed.
          */
         int_cntrl.status_reg[0].chk_reg.rgn_usr_addr = virtualBaseAddress[deviceId][memRegion];
         int_cntrl.status_reg[0].chk_reg.offset_bytes = 0;
         int_cntrl.status_reg[0].chk_reg.len_bytes = 4;
         int_cntrl.status_reg[0].chk_reg_opt = CEI_XRO_NONE;
         int_cntrl.status_reg[0].chk_reg_mask = CEIDEV_INTERRUPT_PENDING;
      
         /* upon interrupt, clear the interrupt pending bit in the same register */
         int_cntrl.status_reg[0].clr_reg_opt = CEI_CRO_USE_CHK_REG | CEI_CRO_SET ;
         int_cntrl.status_reg[0].clr_reg_mask = CEIDEV_GLOBAL_ENABLE_AUTON_CLEAR + CEIDEV_INTERRUPT_PENDING;
      
         /* If an x30 device has a PLX PCI interface, use the PLX PCI Interrupt Control to  
          * disable interrupts upon improper termination of the process and loss of the     
          * handle in the driver; otherwise reset the Global Enable "Interrupt Enable" bit. 
          */
         int_cntrl.disable_reg[0].dis_reg.len_bytes = 4;
         if (disOffset) {
            int_cntrl.disable_reg[0].dis_reg.rgn_usr_addr = virtualBaseAddress[deviceId][0];
            int_cntrl.disable_reg[0].dis_reg.offset_bytes = disOffset;
            int_cntrl.disable_reg[0].dis_reg_opt = CEI_DRO_AND;
            int_cntrl.disable_reg[0].dis_reg_mask = (CEI_UINT64)0xFFFFFFFF;
         }
         else {
            int_cntrl.disable_reg[0].dis_reg.rgn_usr_addr = virtualBaseAddress[deviceId][memRegion];
            int_cntrl.disable_reg[0].dis_reg.offset_bytes = 0;
            int_cntrl.disable_reg[0].dis_reg_opt = CEI_DRO_SET;
            int_cntrl.disable_reg[0].dis_reg_mask = (CEI_UINT64)(CEIDEV_GLOBAL_ENABLE_AUTON_CLEAR + CEIDEV_INTERRUPT_PENDING + CEIDEV_INTERRUPTS_ENABLED);
         }
      }
      
      int_cntrl.add_chk_cnt = 0;     /* no special ISA processing */
      
      /* install the interrupt handler */
      if (cei_dll->LL_HookInterrupt(devHandle, &int_cntrl, &preexisted, cei_err) == FALSE) {
         error_box("Failed to install the interrupt handler for Device '%ld' (details: %s).", deviceId, cei_err);
         status = ARS_INT_ISR;
      }
      else {
         pData[deviceId].board = deviceId;
         pData[deviceId].handle = devHandle;
         pData[deviceId].interrupt_function = isr;
      
         /* Create the interrupt thread suspended so we can set its priority */
         hThread = (HANDLE)_beginthreadex ( NULL,
                                            0, 
                                            &WaitForInterrupts, 
                                            (void *)&pData[deviceId],
                                            CREATE_SUSPENDED, 
                                            &ThreadID );
      
         /* Make sure the thread was created properly */
         if ( hThread != NULL ) {
            pData[deviceId].hIntThrd = hThread;
         
            /* Set the thread priority to user-specified value */
            status = SetThreadPriority(hThread, priority);
      
            /* Everything's set, let it run.  It will execute "startup". */
            status = ResumeThread(hThread);
            status = ARS_NORMAL;
            *isrEnabled = TRUE;
         }
         else {
            error_box("Failed to create the interrupt handler thread for Device '%ld'.", deviceId);
            status = ARS_INT_ISR;
            *isrEnabled = FALSE;
         }
      }
   }
   return status;
}


/*===========================================================================*
 * INTERNAL ROUTINE: c e i l l _ t e r m i n a t e I n t e r r u p t
 *===========================================================================*
 *
 * FUNCTION:    Provides the method to terminate an interrupt association to 
 *              an ISR
 * 
 * PARAMETERS:  devHandle   -- Handle for the Device Number whose interrupt 
 *                             is to be disabled
 * 
 * RETURN VAL:  ARS_INT_ISR -- Interrupt was previously disabled
 *              ARS_NORMAL  -- Interrupt was successfully relinquished
 *
 * GLOBALS:     none
 * 
 * DESCRIPTION: terminates the specified device interrupt function
 * 
 *===========================================================================*/
CEI_INT32 ceill_terminateInterrupt (CEI_DEV_HANDLE devHandle)
{
   CEI_UINT32 persists;       /* receives true if another process is still using the devices interrupt */
   CEI_INSTALL_ERR cei_err;   /* receives error description upon failure */
   CEI_INT32 status;

   /* indicate that this process is done monitoring the device interrupt - */
   /* any pending LL_WaitForInterrupt calls associated with threads */
   /* belonging to this process return with the result parameter set to */
   /* CEI_WIR_INT_DISABLED - if this is the last process to unhook, device */
   /* interrupts are disabled and the interrupt handler is uninstalled */
   if (cei_dll->LL_UnhookInterrupt(devHandle, &persists, cei_err) == FALSE) {
      status = ARS_INT_ISR;
   }
   else {
      status = ARS_NORMAL;
   }
   return status;   
}


/*===================================================================================================================*
 * ENTRY POINT:                         C E I _ A C Q U I R E _ D E V I C E
 *===================================================================================================================*
 *
 * FUNCTION:    Open a session with a device
 *              
 * PARAMETERS:  CDEV_BOARD_TYPE deviceIndex  -- (input)  device to reference (0 to 15, 127 for Windows)
 *              ARINC_LOCAL_TYPE * device    -- (output) reference to ARINC_LOCAL_TYPE struct
 * 
 * RETURN VAL:  ARS_NORMAL     -- Session was opened successfully
 *              ARS_DRIVERFAIL -- Under Windows, Linux, Visa, and LabVIEW R/T, the respective driver 'Open Session' 
 *                                or 'Resource Acquisition' operations failed.
 *              ARS_FAILURE    -- Under VxWorks and Integrity, the driver 'Open Session' opreation failed.
 *                                Under Windows, the CEI_Install.DLL cannot be accessed
 *
 * DESCRIPTION: This routine searches for the instance of a CEI-x30 device matching the supplied device ID, with the 
 *              resource acquisition for each instance of a device found limited to the first execution of this 
 *              routine.  In most cases subsequent invocations will simply retrieve the resources allocated for the 
 *              respective device index from global/static variables, (with the exception of some PPC SBCs such as  
 *              Thales where the "sysPciMemToLocalAdrs" translation will be applied to previously mapped addresses).
 *
 *===================================================================================================================*/
CEI_INT32 cei_acquire_device (CDEV_BOARD_TYPE deviceIndex, ARINC_LOCAL_TYPE * device)
{
   TIMECAPS tc;
   CEI_DEV_HANDLE new_dev_handle;  /* receives new device handle */
   CEI_DEVICE_MEM dev_mem;         /* receives device memory region info */
   CEI_INSTALL_ERR cei_err_ignore; /* ignored installation library error     */
   CEI_UINT32 preexisted;          /* receives device mapping preexistence */
   CEI_UINT32 board = (CEI_UINT32)deviceIndex;
   CEI_INT32 status = ARS_NORMAL;
#ifndef TARGET_EMBEDDED
   CEI_INT32 sharedLockState;
#endif

   /* dynamically link to the installation library and open the device */
   cei_dll = CEI_DLL_GetDynamicLink (0, deviceOpenErrorDesc);
   if (cei_dll == NULL) {
      status = ARS_DRIVERFAIL;
   }
   else {
      /* attempt to open the specified device */
      if (cei_dll->LL_OpenDeviceByLogicalId (board, &new_dev_handle, deviceOpenErrorDesc) == FALSE) {
         status = ARS_DRIVERFAIL;
      }
   
      /* attempt to map device memory */
      else if (cei_dll->LL_MapMemoryRegions (new_dev_handle, CEI_DEVICE_MEM_VER, 0, &dev_mem, &preexisted, deviceOpenErrorDesc) == FALSE) {
         /* close the device and bail out */
         cei_dll->LL_CloseDevice (new_dev_handle, cei_err_ignore);
         status = ARS_DRIVERFAIL;
      }
   
      /* store host addresses of all memory regions */
      else {
         virtualBaseAddress[board][0] = (CEI_UINT64)dev_mem.rgn_usr_addr[0];
         virtualBaseAddress[board][1] = (CEI_UINT64)dev_mem.rgn_usr_addr[1];
         device->baseAddr = (pCEI_UINT32)dev_mem.rgn_usr_addr[0];
         device->devicePtr = (pCEI_UINT32)dev_mem.rgn_usr_addr[1];
         /* store the device handle for later */
         dev_handle[board] = new_dev_handle;
      
#ifndef TARGET_EMBEDDED
         /* if not already defined, request multi-thread/process protection */
         if (ceiAccessLock[board] == NULL) {
            status = cei_open_shared_lock (deviceIndex, &sharedLockState);
         }
#endif
   
         /* Attempt to set the Windows timer resolution to 1 millisecond */
         if (timeGetDevCaps(&tc, sizeof(TIMECAPS)) == TIMERR_NOERROR) {
            timeBeginPeriod(min(max(tc.wPeriodMin, 1), tc.wPeriodMax));
         }
      }
   }
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                    C E I _ A C Q U I R E _ D E V I C E _ I D E N T I T Y
 *===================================================================================================================*
 *
 * FUNCTION:    Open a session with a device
 *              
 * PARAMETERS:  CEI_INT32  deviceIndex -- (input)  device to reference (0 to 15, 127 for Windows)
 *              pCEI_INT32 vendorId    -- (output) ptr to storage of vendor PCI ID
 *              pCEI_INT32 deviceId    -- (output) ptr to storage of device PCI ID
 * 
 * RETURN VAL:  ARS_NORMAL     -- Successful acquisition of the device ID's
 *              ARS_FAILURE    -- Under Visa, LabVIEW R/T, VxWorks and Linux, the driver session is not open
 *                                Under Windows, the CEI_Install DLL cannot be accessed
 *              ARS_INVBOARD   -- Under Windows only, device configuration space is inaccessible (session not opened)
 *
 * DESCRIPTION: This routine reports the specified device PCI Configuration Space Vendor ID and Device ID values 
 *              either read directly from PCI Configuration Space or acquired and saved during initialization.
 *
 *===================================================================================================================*/
CEI_INT32 cei_acquire_device_identity 
(
   CEI_INT32  deviceIndex,                                                        /* device number to test (0-127)   */
   pCEI_INT32 vendorId,                                                           /* ptr to storage of vendor PCI ID */
   pCEI_INT32 deviceId                                                            /* ptr to storage of device PCI ID */
)
{
   CEI_INT32 status = ARS_FAILURE;                                                /* local function return status    */
   CEI_UINT32 idValue;
   CEI_INSTALL_ERR err;

   if (cei_dll != NULL) {
      if ( (status = cei_dll->LL_ReadConfigSpace(dev_handle[deviceIndex], 0, 2, &idValue, err)) == FALSE) {
         status = ARS_INVBOARD;
      }
      else {
         *vendorId = (CEI_INT32)idValue;
         status = cei_dll->LL_ReadConfigSpace(dev_handle[deviceIndex], 2, 2, &idValue, err);
         *deviceId = (CEI_INT32)idValue;
         status = ARS_NORMAL;
      }
   }
   else {
      status = ARS_FAILURE;
   }
   return status;
}


#ifndef TARGET_EMBEDDED
/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ A C Q U I R E _ T H R E A D _ L O C K    
 *===================================================================================================================*
 *
 * FUNCTION:    Block on a mutex or semaphore previously reserved for the specific device.  
 *              WINDOWS:    A mutex is used for single process multi-threading and multi-process applications.
 *              INTEGRITY:  A Posix mutex is used for single process multi-threading applications.
 *              VXWORKS:    A global semaphore is used for multi-process applications.
 *              LINUX:      A Posix mutex is used for single process multi-threading applications.
 *                          A global semaphore is used for multi-process applications.
 *              VISA:       A Visa Lock mechanism is used for multi-threading single and multi-process applications.
 *              LABVIEW RT: A Visa Lock mechanism is used for multi-threading single and multi-process applications.
 *
 * PARAMETERS:  deviceIndex -- (input) device to reference (0 to 15, 127 for Windows)
 * 
 * RETURN VAL:  ARS_NORMAL      -- The mutex/semaphore was acquired successfully
 *              ARS_FAILURE     -- Acquisition of the lock mechanism timed-out/failed (Linux, VxWorks and Integrity)          
 *              ARS_BOARD_MUTEX -- Acquisition of the lock mechanism timed-out/failed (Windows, VISA, LabVIEW R/T)
 *
 * DESCRIPTION: This routine blocks execution, waiting for access to the locking mechanism previously allocated for  
 *              the specific device.  Some operating systems support a time-out error, others do not; for those that
 *              do, the ARS_BOARD_MUTEX status code is used in place of the generic ARS_FAILURE status code.
 *
 *===================================================================================================================*/
CEI_INT32 cei_acquire_thread_lock (CDEV_BOARD_TYPE deviceIndex)
{
   DWORD status;
   status = WaitForSingleObject(ceiAccessLock[deviceIndex], INFINITE);
   if (status == WAIT_FAILED) {
      status = ARS_BOARD_MUTEX;
   }
   else {
      status = ARS_NORMAL;
   }
   return status;
}
#endif


/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ D E L A Y _ M I L L I S E C O N D S
 *===================================================================================================================*
 *
 * FUNCTION:    Delay for a specified number of milliseconds
 *
 * PARAMETERS:  CEI_UINT32 numberOfMillieconds -- (input) the number of milliseconds to delay execution of the 
 *                                                        calling thread
 * 
 * RETURN VAL:  ARS_FAILURE -- For Integrity only, indicates a nanosleep invocation failure
 *              ARS_NORMAL  -- Success
 *
 * DESCRIPTION: This routine utilizes the following o/s-specific mechanism for delaying thread execution:
 *
 *              WINDOWS:    The C standard library Sleep() function is used.
 *              INTEGRITY:  Delay method is optional - for Posix, nanosleep is used; otherwise, usleep is used.
 *              VXWORKS:    Delay method is BSP dependent - either sysUsDelay, sysMsDelay, or taskDelay is used.
 *              LINUX:      A delay-polling method is used based on repeated invocations of clock()
 *              VISA:       The C standard library Sleep() function is used.
 *              LABVIEW RT: The o/s function Delay() is used
 *
 *===================================================================================================================*/
CEI_INT32 cei_delay_milliseconds (CEI_UINT32 numberOfMilliseconds)
{
   Sleep(numberOfMilliseconds);
   return ARS_NORMAL;
}


/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ D E L A Y _ S E C O N D S
 *===================================================================================================================*
 *
 * FUNCTION:    Delay for a specified (fractional) number of seconds
 *
 * PARAMETERS:  CEI_FLOAT numberOfSeconds -- (input) a floating point representation of the number (or fraction) of
 *                                                   seconds to delay execution of the calling thread
 * 
 * RETURN VAL:  ARS_FAILURE -- For Integrity only, indicates a nanosleep invocation failure
 *              ARS_NORMAL  -- Success
 *
 * DESCRIPTION: This routine utilizes the following o/s-specific mechanism for delaying thread execution:
 *
 *              WINDOWS:    A delay-polling method is used based on repeated invocations of timeGetTime, having a 
 *                          typical resolution of 1 millisecond.
 *              INTEGRITY:  Delay method is optional - for Posix, nanosleep is used; otherwise, usleep is used.
 *              VXWORKS:    Delay method is BSP dependent - either sysUsDelay, sysMsDelay, or taskDelay is used.
 *              LINUX:      A delay-polling method is used based on repeated invocations of clock()
 *              VISA:       The C standard library Sleep() function is used.
 *              LABVIEW RT: The o/s function Delay() is used
 *
 *===================================================================================================================*/
CEI_INT32 cei_delay_seconds (CEI_FLOAT numberOfSeconds)
{
   CEI_UINT32 start, now, delay;

   delay = (CEI_UINT32)((float)TIMER_CLOCK_TICKS_PER_SECOND * numberOfSeconds);
   start = timeGetTime();
   now = start;
   while ((now - start) < delay) {
      now = timeGetTime();
   }
   return ARS_NORMAL;
}


/*===================================================================================================================*
 * ENTRY POINT:                     C E I _ G E T _ H O S T _ S Y S T E M _ T I M E _ X
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the host 64-bit 1us calendar time value, supplied via operating system as time since Jan 1.
 *
 * PARAMETERS:  TIME_TAG_TYPE hostTime -- (output) 64-bit 1us millisecond resolution time
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current 64-bit 1us resolution calendar time value
 *
 *              WINDOWS:    Calendar time is acquired via combination of localtime() and mktime() invocations.
 *              INTEGRITY:  Calendar time is acquired via combination of localtime() and mktime() invocations (Posix only).
 *              VXWORKS:    Calendar time is acquired via combination of localtime() and mktime() invocations.
 *              LINUX:      Calendar time is acquired via combination of localtime() and mktime() invocations.
 *              VISA:       Calendar time is acquired via combination of time() and localtime() inocations.
 *              LABVIEW RT: Calendar time is acquired via combination of time() and localtime() inocations.
 *
 *===================================================================================================================*/
CEI_VOID  cei_get_host_system_time_x (TIME_TAG_TYPE * hostTime)
{
   time_t now;           /* Current time, modify to get Time Reference Point. */
   struct   tm start;    /* Formatted "now" current time, modified for TRP.   */
   time_t   then;        /* Time Reference Point, Midnight or Jan 1.          */
   struct   timeb ts;    /* Current time, use to get milliseconds.            */
   CEI_UINT32 last_milli;/* This is time we use to wait for time to change.   */
   TIME_TAG_TYPE diff;   /* This is difference time we return in microseconds */

   time(&now);               /* Time in seconds since midnight Jan 1,1970 GMT */
   start = *localtime(&now); /* Local                                         */
   start.tm_sec  = 0;        /*   time                                        */
   start.tm_min  = 0;        /*     at                                        */
   start.tm_hour = 0;        /*       Jan 1                                   */
   start.tm_mday = 1;        /*         midnight (This is day 0)              */
   start.tm_mon  = 0;        /*           this                                */
   start.tm_yday = 0;        /*             year                              */
   then = mktime(&start);    /* local Jan 1 to time in seconds since Jan 1    */
                             /* We do IRIG time, which makes Jan 1 "day 0"    */ 

   ftime(&ts);                             /* Current time                    */
   last_milli = ts.millitm;                /*  save milliseconds              */
   while (TRUE) {                          /* Wait for milliseconds to change */
      ftime(&ts);                          /* Current time                    */
      if (ts.millitm != last_milli)
         break;                            /* Break at exact time tick change */
   }
   diff = ts.time-then;       /* We only want the time since Midnight on Jan 1*/
   diff *= 1000000;           /* Number of Seconds converted to microseconds  */
   diff += ts.millitm * 1000; /* plus number of Milliseconds converted to     */
   *hostTime = diff;          /* microseconds.  Passed back to the caller.    */
}

/*===================================================================================================================*
 * ENTRY POINT:                     C E I _ G E T _ H O S T _ S Y S T E M _ T I M E
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the host 32-bit 1ms reference time value, supplied via operating system "System Time"
 *
 * PARAMETERS:  pCEI_UINT32 hostTime -- (output) 32-bit 1 millisecond resolution time
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current System Time in milliseconds
 *
 *              WINDOWS:    System time is acquired via the o/s function timeGetTime().
 *              INTEGRITY:  System time is acquired via the o/s function clock_gettime (Posix), GetClockTime otherwise.
 *              VXWORKS:    System time is acquired via combination of localtime() and mktime() invocations.
 *              LINUX:      System time is acquired via the o/s function clock().
 *              VISA:       Invokes cei_get_host_system_time_x to acquire system time.
 *              LABVIEW RT: Invokes cei_get_host_system_time_x to acquire system time.
 *
 *===================================================================================================================*/
CEI_INT32 cei_get_host_system_time (pCEI_UINT32 hostTime)
{
   *hostTime = timeGetTime();
   return ARS_NORMAL;
}

/*===================================================================================================================*
 * ENTRY POINT:                  C E I _ G E T _ I R I G _ I N I T I A L _ T I M E
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the host reference calendar time/date value for CEI-x30 IRIG generator initialization
 *
 * PARAMETERS:  time -- (output) 32-bit packed IRIG time value
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current System Time in calendar time/date and formats it into an 
 *              IRIG bit-encoded format value.
 *
 *===================================================================================================================*/
CEI_VOID  cei_get_irig_initial_time (pCEI_UINT32 time)
{
   TIME_TAG_TYPE system_time;
   CEI_NATIVE_UINT seconds;
   CEI_NATIVE_UINT tens_of_seconds;
   CEI_NATIVE_UINT minutes;
   CEI_NATIVE_UINT tens_of_minutes;
   CEI_NATIVE_UINT hours;
   CEI_NATIVE_UINT tens_of_hours;
   CEI_NATIVE_UINT days;
   CEI_NATIVE_UINT tens_of_days;
   CEI_NATIVE_UINT hundreds_of_days;

   union
   {
      struct irigTimeFormat
           {
              CEI_NATIVE_UINT seconds          :4;
              CEI_NATIVE_UINT tens_of_seconds  :3;
              CEI_NATIVE_UINT minutes          :4;
              CEI_NATIVE_UINT tens_of_minutes  :3;
              CEI_NATIVE_UINT hours            :4;
              CEI_NATIVE_UINT tens_of_hours    :2;
              CEI_NATIVE_UINT days             :4;
              CEI_NATIVE_UINT tens_of_days     :4;
              CEI_NATIVE_UINT hundreds_of_days :2;
           } IRIG_TIME_FORMAT;
      CEI_UINT32 timeValue;
   } timeConvert;

   cei_get_host_system_time_x (&system_time);

   seconds = (CEI_UINT32)(system_time / (TIME_TAG_TYPE)1000000);
   if ( (system_time % (TIME_TAG_TYPE)1000000) > (TIME_TAG_TYPE)500000) {
      seconds++;
   }

   /* Extract Minutes from Seconds */
   minutes = seconds/60;
   seconds = seconds%60;
   tens_of_seconds = seconds/10;
   seconds = seconds%10;

   /* Extract Hours from Minutes */
   hours = minutes/60;  
   minutes = minutes%60;
   tens_of_minutes = minutes/10;
   minutes = minutes%10;

   /* Extract Days from Hours */
   days = hours/24;     
   hours = hours%24;     
   tens_of_hours = hours/10;    
   hours = hours%10;    
   days++;                       /* account for no such thing as day 0 */
   hundreds_of_days = days/100;
   days = days%100;
   tens_of_days = days/10;
   days = days%10;

   timeConvert.IRIG_TIME_FORMAT.seconds          = seconds;
   timeConvert.IRIG_TIME_FORMAT.tens_of_seconds  = tens_of_seconds;
   timeConvert.IRIG_TIME_FORMAT.minutes          = minutes;
   timeConvert.IRIG_TIME_FORMAT.tens_of_minutes  = tens_of_minutes;
   timeConvert.IRIG_TIME_FORMAT.hours            = hours;
   timeConvert.IRIG_TIME_FORMAT.tens_of_hours    = tens_of_hours;
   timeConvert.IRIG_TIME_FORMAT.days             = days;
   timeConvert.IRIG_TIME_FORMAT.tens_of_days     = tens_of_days;
   timeConvert.IRIG_TIME_FORMAT.hundreds_of_days = hundreds_of_days;
   
   *time = timeConvert.timeValue;
}


/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ G E T _ O P E N _ E R R O R
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the error description for a device open error
 *
 * PARAMETERS:  openErrorString -- (output) reference to assign to error string
 * 
 * RETURN VAL:  ARS_NORMAL  -- Mutex was acquired successfully
 *
 * GLOBALS:     deviceOpenErrorDesc
 *
 * DESCRIPTION: This routine retrieves the error description previously saved when a device open error is encountered.
 *
 *===================================================================================================================*/
pCEI_CHAR cei_get_open_error (CEI_VOID)
{
  return deviceOpenErrorDesc;
}


/*===================================================================================================================*
 * ENTRY POINT:                         C E I _ I N I T _ I N T E R R U P T
 *===================================================================================================================*
 *
 * FUNCTION:    Configures the device PCI interrupt enable and associates the supplied ISR function with it.
 *              
 * PARAMETERS:  CEI_INT32   deviceIndex       -- (input)  device to reference (0 to 127)
 *              CEI_INT32   _stdcall isr      -- (input)  ptr to the interrupt service routine
 *              CEI_INT32   priority          -- (input)  desired priority of the ISR thread
 *              pCEI_UINT32 deviceBaseAddress -- (input)  virtual address of BAR0/BAR2 region
 *              CEI_INT32   memRegion         -- (input)  PCI BAR region/index for interrupt clear bit (0 - BAR0 or 1 - BAR2)
 *              CEI_UINT32  disOffset         -- (input)  PCI Interrupt Disable register (PLX reg offset, 0 for no PLX interface)
 *              CEI_UINT32  interruptBit      -- (input)  Local Interrupt Input Enable bit for PLX interface control register
 *              CEI_UINT32  comboChannelCount -- (input)  1553/ARINC Combo Board configuration
 *              pCEI_UINT32 isrEnabled        -- (output) thread handle if successfully defined
 * 
 * RETURN VAL:  ARS_NORMAL         - Interrupt was initialized successfully
 *              ARS_DRIVERFAIL     - Device driver session never opened
 *              ARS_INT_ISR        - Interrupt was not initialized successfully
 *                                 - Driver interrupt event creation failure
 *                                 - Windows thread create failure
 *              ARS_NO_INT_SUPPORT - Driver interrupt handling not supported for this board
 *              ARS_NO_INT_ENABLED - Driver Interrupt handling not enabled for this board
 *
 * GLOBALS:     None
 *
 * DESCRIPTION: This routine enables the PCI interrupt, then connects the user-specified interrupt function   
 *              to the interrupt assigned to the respective device by the host o/s.
 *
 *===================================================================================================================*/
CEI_INT32 cei_init_interrupt 
( 
   CEI_INT32   deviceIndex,                /* (input)  device to reference (0 to 15, for Windows 127)                */
   CEI_INT32   (_stdcall *isr)(CEI_INT32), /* (input)  ptr to the interrupt service routine                          */
   CEI_INT32   priority,                   /* (input)  desired priority of the ISR thread                            */         
   pCEI_UINT32 deviceBaseAddress,          /* (input)  virtual address of BAR0/BAR2 region                           */         
   CEI_INT32   memRegion,                  /* (input)  PCI BAR region/index for interrupt clear bit                  */
   CEI_UINT32  disOffset,                  /* (input)  PCI Interrupt Disable register                                */
   CEI_UINT32  interruptBit,               /* (input)  Local Interrupt Input Enable bit for PLX control register     */
   CEI_UINT32  comboChanCount,             /* (input)  1553/ARINC Combo Board configuration                          */         
   pCEI_UINT32 isrEnabled                  /* (output) thread handle if successfully defined                         */         
)
{
   CEI_UINT32 intCtrlStatReg;             /* PLX Int Control/Status Reg data */
   CEI_INT32 status = ARS_NORMAL;         /* local function return status    */
   /* local reference to the low-level global virtual base address array     */
   pCEI_UINT32 virtualBaseAddressPtr = deviceBaseAddress;
   
   if (interruptBit != 0) {
      status = cei_read_32bit ((CEI_INT16)deviceIndex, virtualBaseAddressPtr, &intCtrlStatReg);
      intCtrlStatReg &= ~interruptBit;
      status = cei_write_32bit ((CEI_INT16)deviceIndex, virtualBaseAddressPtr, intCtrlStatReg);
   }

   /* initialize the interrupt */
   status = ceill_initInterrupt (deviceIndex, isr, priority, memRegion, disOffset, comboChanCount, isrEnabled);   
   
   if (status == ARS_NORMAL) {
      /* Set either the PLX PCI9056 or PCI9030 Interrupt Control/Status Register
       * Local Interrupt Input Enable, High to pass local bus interrupts
       */
      if (interruptBit != 0) {
         status = cei_read_32bit ((CEI_INT16)deviceIndex, virtualBaseAddressPtr, &intCtrlStatReg);
         intCtrlStatReg |= interruptBit;
         status = cei_write_32bit ((CEI_INT16)deviceIndex, virtualBaseAddressPtr, intCtrlStatReg);
      }
   }
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                      C E I _ M A P _ S H A R E D _ M E M O R Y
 *===================================================================================================================*
 *
 *              NOTE: IMPLEMENTED FOR WINDOWS AND LINUX TARGET OPERATING SYSTEMS ONLY!
 *
 * FUNCTION:    Map the desired shared global memory block into the calling process's address space, allocating 
 *              the memory block if it doesn't already exist.  
 *
 * PARAMETERS:  CEI_UCHAR  blockId        -- (input)  unique global block id (0-15)
 *              CEI_UINT32 blockSize      -- (input)  memory block size (in bytes)
 *              CEI_VOID   **blockPtr     -- (output) receives pointer to memory block
 *              CEI_VOID   **blockHandle  -- (output) receives platform-specific handle to the shared memory block
 *              CEI_UINT32 *alreadyExists -- (output) receives true if the block already exists, false otherwise
 *              CEI_UINT32 *isExclusive   -- (output) receives true if no other processes currently have the block 
 *                                                    mapped, false otherwise
 *
 * RETURN VAL:  ARS_NORMAL          - operation completed successfully
 *              ARS_ERR_SH_MEM_MAP  - failed to map shared memory block
 *              ARS_ERR_SH_MEM_OBJ  - failed to allocate shared memory block
 *              ARS_INVARG          - board index is invalid
 *              ARS_FAILURE         - Windows only - the CEI_Install library is not loaded in memory
 *
 * DESCRIPTION: If a shared global memory block associated with the given application-specific id has already been 
 *              allocated on the target system, map the memory block into the calling process's address space and 
 *              set 'already_exists' to true.  Otherwise, allocate and map the specified global memory block
 *              and set 'already_exists' to false.
 *
 *              If the memory block is being allocated for the first time, the caller should zero-fill the block.
 *
 *===================================================================================================================*/
CEI_INT32 cei_map_shared_memory 
( 
   CEI_UCHAR    blockId,                                      /* (input)  unique global block id (0-15)              */
   CEI_UINT32   blockSize,                                    /* (input)  memory block size (in bytes)               */
   CEI_VOID  ** blockPtr,                                     /* (output) receives pointer to memory block           */
   CEI_VOID  ** blockHandle,                                  /* (output) handle to the shared memory block          */
   CEI_UINT32 * alreadyExists,                                /* (output) true - block already exists, false - new   */  
   CEI_UINT32 * isExclusive                                   /* (output) true = no other owners, false - in use     */                                                         
)                                                              
{
   CHAR blockName[32];                            /* build unique block name */
   CEI_INT32 status = ARS_NORMAL;                 /* function return status  */
   
   if (blockId >= MAX_BOARDS) {  /* assure the block id is valid */
      status = ARS_INVARG;
   }
   else if (cei_dll == NULL) {   /* assure we've loaded CEI_Install library  */
      status = ARS_FAILURE;
   }
   else {
      sprintf(blockName, "ceix30_shmem_block_%d", blockId);
   
      /* initialize the block pointer and handle */
      *blockPtr = NULL;
      *blockHandle = NULL;
   
      /* create a file mapping object of the desired size */
      if (cei_dll->OpenGlobalNamedObject (CEI_GSO_SHMEM, (CEI_UINT64)blockSize, blockName, alreadyExists, blockPtr, blockHandle, deviceOpenErrorDesc) == FALSE) {
         status = ARS_ERR_SH_MEM_OBJ;
      }
      else if (*blockPtr == NULL) {
         /* failed to obtain a valid pointer to the shared memory block */
         status = ARS_ERR_SH_MEM_MAP;
      }
      else {
         /* indicate to the calling routine if any other processes have this shared memory block already mapped */
         *isExclusive = (*alreadyExists == FALSE);
      }
   }
   return status;
}


#ifndef TARGET_EMBEDDED
/*===================================================================================================================*
 * ENTRY POINT:                       C E I _ O P E N _ P R E L O A D _ L O C K
 *===================================================================================================================*
 *
 * FUNCTION:    Requests the initiation of a shared (global) lock for multiprocess application usage
 *
 * PARAMETERS:  CDEV_BOARD_TYPE device -- (input) index of the device for which to create the lock 
 *                                                (0 to 15, 127 for Windows)
 *
 * RETURN VAL:  ARS_NORMAL      - global mutex was opened successfully
 *              ARS_INVARG      - global mutex id too large
 *              ARS_BOARD_MUTEX - failed to create/open the global mutex
 *              ARS_FAILURE     - CEI_Install library is not loaded in memory
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: Create the device-specific shared (global) lock for use in multiple process applications for select
 *              operating systems.
 *
 *              WINDOWS:    Invokes cei_open_shared_lock.
 *              INTEGRITY:  Not implemented.
 *              VXWORKS:    Not implemented.
 *              LINUX:      Invokes cei_open_shared_lock.
 *              VISA:       Invokes viLock to create a shared (global) lock
 *              LABVIEW RT: Invokes viLock to create a shared (global) lock
 *
 *===================================================================================================================*/
CEI_INT32 cei_open_preload_lock (CDEV_BOARD_TYPE device)
{
   CEI_INT32 status;
   CEI_INT32 temp;
   CEI_INSTALL_ERR cei_err;        /* receives installation library error     */

   /* dynamically link to the installation library and open the device */
   cei_dll = CEI_DLL_GetDynamicLink (0, cei_err);
   if (cei_dll == NULL) {
      status = ARS_FAILURE;
   }
   else {
      status = cei_open_shared_lock(device, &temp);
   }
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                         C E I _ O P E N _ S H A R E D _ L O C K
 *===================================================================================================================*
 *
 *              NOTE: IMPLEMENTED FOR WINDOWS AND LINUX TARGET OPERATING SYSTEMS ONLY.
 *                    FOR THE VISA DRIVER IMPLEMENTED AS AN BLOCK ON LOCK ACCESS
 *
 * FUNCTION:    Opens a shared (global) mutex/semaphore lock
 *
 * PARAMETERS:  CDEV_BOARD_TYPE sharedLockId -- (input) device to reference (0 to 15, 127 for Windows)
 *              pCEI_INT32 sharedLockState   -- (input) location to store the mutex/semaphore handle
 *
 * RETURN VAL:  ARS_NORMAL      -- global mutex/semaphore was opened successfully
 *              ARS_INVARG      -- global mutex/semaphore id too large
 *              ARS_BOARD_MUTEX -- failed to create/open the global mutex/semaphore
 *              ARS_FAILURE     -- Windows only, CEI_Install library is not loaded in memory
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: Opens the specified shared (global) mutex/semaphore for use in multiple process applications
 *
 *===================================================================================================================*/
CEI_INT32 cei_open_shared_lock (CDEV_BOARD_TYPE sharedLockId, pCEI_INT32 sharedLockState)
{
   CHAR mutexName[32];                /* unique mutex name                   */
   CEI_UINT32 preexisted_ignore;      /* whether the lock already existed    */
   CEI_INSTALL_ERR cei_err;           /* receives installation library error */
   CEI_INT32 status = ARS_NORMAL;     /* function return status              */
   
   /* build unique mutex name */
   if (sharedLockId >= MAX_BOARDS) {
      status = ARS_INVARG;
   }
   else if (cei_dll == NULL) {
      status = ARS_FAILURE;
   }
   else {
      sprintf (mutexName, "ceix30_shared_lock_%d", sharedLockId);
   
      cei_dll->OpenGlobalNamedObject (CEI_GSO_MUTEX, FALSE, mutexName, &preexisted_ignore, NULL, (HANDLE*)&ceiAccessLock[sharedLockId], cei_err);
      if (ceiAccessLock[sharedLockId] == NULL) {
         status = ARS_BOARD_MUTEX;
      }
   }
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ R E L E A S E _ T H R E A D _ L O C K    
 *===================================================================================================================*
 *
 * FUNCTION:    Release a semaphore previously reserved/blocked by the application/API for the specific device.
 *
 *              WINDOWS:    A mutex is used for single process multi-threading and multi-process applications.
 *              INTEGRITY:  A Posix mutex is used for single process multi-threading applications.
 *              VXWORKS:    A global semaphore is used for multi-process applications.
 *              LINUX:      A Posix mutex is used for single process multi-threading applications.
 *                          A global semaphore is used for multi-process applications.
 *              VISA:       A Visa Lock mechanism is used for multi-threading single and multi-process applications.
 *              LABVIEW RT: A Visa Lock mechanism is used for multi-threading single and multi-process applications.
 *
 * PARAMETERS:  deviceIndex -- (input) device to reference (0 to 15, 127 for Windows)
 * 
 * RETURN VAL:  ARS_NORMAL      -- The mutex/semaphore was acquired successfully
 *              ARS_FAILURE     -- Acquisition of the lock mechanism timed-out/failed (Linux, VxWorks and Integrity)          
 *              ARS_BOARD_MUTEX -- Acquisition of the lock mechanism timed-out/failed (Windows, VISA, LabVIEW R/T)
 *
 * DESCRIPTION: This routine releases the resource access previously allocated as a locking mechanism for
 *              the specific device.  Some operating systems support a time-out error, others do not; for those that
 *              do, the ARS_BOARD_MUTEX status code is used in place of the generic ARS_FAILURE status code.
 *
 *===================================================================================================================*/
CEI_INT32 cei_release_thread_lock (CDEV_BOARD_TYPE deviceIndex)
{
   CEI_INT32 status = ARS_NORMAL;                  /* function return status */
   if (ReleaseMutex(ceiAccessLock[deviceIndex]) == FALSE) {
      status = ARS_FAILURE;
   }
   return status;
}
#endif


/*===========================================================================*
 * ENTRY POINT:             R E L E A S E   D E V I C E
 *===========================================================================*
 *
 * FUNCTION:    Close a session with a device
 *              
 * PARAMETERS:  CDEV_BOARD_TYPE deviceIndex  -- (input)  device to reference (0 to 15, 127 for Windows)
 *              ARINC_LOCAL_TYPE * device    -- (output) reference to ARINC_LOCAL_TYPE struct
 * 
 * RETURN VAL:  ARS_NORMAL         - session was closed successfully
 *              ARS_BOARD_MUTEX    - failed to release the host interface lock
 *              ARS_ERR_SH_MEM_OBJ - failed to release shared memory resource
 *              ARS_DRIVERFAIL     - could not access the CEI_Install library or
 *                                   the Driver Close Session invocation failed
 *
 * GLOBALS:     ceiAccessLock[]
 *              dev_handle[]
 *              cei_dll
 *
 * DESCRIPTION: This routine releases a driver handle and unmaps the two PCI 
 *              memory regions previously mapped for the specific device.
 *
 *===========================================================================*/
CEI_INT32 cei_release_device (CDEV_BOARD_TYPE deviceIndex, ARINC_LOCAL_TYPE * device)
{
   CEI_INSTALL_ERR cei_err;           /* receives installation library error */
   CEI_UINT32 persists_ignore;        /* receives device mapping persistence */
   TIMECAPS tc;
   CEI_UINT32 board = (CEI_UINT32)deviceIndex;
   CEI_INT32 status = ARS_NORMAL;                  /* function return status */

   if (cei_dll != NULL) {
      if (dev_handle[board] != NULL) {
         /* unmap device memory */
         if (cei_dll->LL_UnmapMemoryRegions (dev_handle[board], &persists_ignore, cei_err) == FALSE) {
            error_box("Failed to unmap device memory on board '%ld' (details: %s).",board,cei_err);
         }
      
         /* close the device and flag the handle as closed */
         if (cei_dll->LL_CloseDevice(dev_handle[board],cei_err) == FALSE) {
            error_box("Failed to close board '%ld' (details: %s).",board,cei_err);
         }
      
         /* indicate no regions are still mapped and flag the handle as closed */
         dev_handle[board] = INVALID_HANDLE_VALUE;
         
         /* restore the Windows tick timer resolution to normal system resolution */
         if (timeGetDevCaps(&tc, sizeof(TIMECAPS)) == TIMERR_NOERROR) {
            timeEndPeriod(min(max(tc.wPeriodMin, 1), tc.wPeriodMax));
         }
      }
#ifndef TARGET_EMBEDDED
      /* release multi-thread mutex protection */
      if (ceiAccessLock[board] != NULL) {
         if (cei_dll->CloseGlobalNamedObject(ceiAccessLock[board], &persists_ignore, cei_err) == FALSE) {
            status = ARS_BOARD_MUTEX;
         }
         ceiAccessLock[board] = NULL;
      }
   
      /* relinquish any multi-process shared memory resource */
      if ((device->multiThreadProtect == AR_CONC_MULTIPROC) && (device->sharedMemoryHandle != NULL)) {
         if (cei_dll->CloseGlobalNamedObject(device->sharedMemoryHandle, &persists_ignore, cei_err) == FALSE) { 
            status = ARS_ERR_SH_MEM_OBJ;
         }
      }
#endif
   }
   else {
      status = ARS_DRIVERFAIL;
   }
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                        C E I _ T E R M I N A T E _ I N T E R R U P T
 *===================================================================================================================*
 *
 * FUNCTION:    Provides the method to terminate an interrupt association to an ISR and clear the interrupt enable.
 * 
 * PARAMETERS:  CEI_INT32   deviceIndex       -- (input)  device to reference (0 to 127)
 *              CEI_UINT32  interruptBit      -- (input)  Local Interrupt Input Enable bit for PLX interface control register
 *              pCEI_UINT32 deviceBaseAddress -- (input)  virtual address of BAR2 region
 * 
 * RETURN VAL:  ARS_NORMAL   -- Interrupt was terminated successfully
 *              ARS_INT_ISR  -- Interrupt was not terminated successfully
 *
 * GLOBALS:     None
 * 
 * DESCRIPTION: This routine terminates the specified device interrupt function and clears the PCI interrupt enable.
 * 
 *===================================================================================================================*/
CEI_INT32 cei_terminate_interrupt
(
   CEI_INT32   deviceIndex,                /* (input)  device to reference (0 to 15, for Windows 127)                */
   CEI_UINT32  interruptBit,               /* (input)  Local Interrupt Input Enable bit for PLX control register     */
   pCEI_UINT32 deviceBaseAddress           /* (input)  virtual address of BAR0/BAR2 region                           */         
)
{
   CEI_UINT32 intCtrlStatReg;             /* PLX Int Control/Status Reg data */
   CEI_INT32 status = ARS_NORMAL;         /* local function return status    */
   /* local reference to the supplied device BAR0 virtual base address       */
   pCEI_UINT32 virtualBaseAddressPtr = deviceBaseAddress;
   
   /* Reset either the PLX PCI9056 or PCI9030 Interrupt Control/Status Register
    * Local Interrupt Input Enable, Low to disable local bus interrupts
    */
   if (interruptBit != 0) {
      status = cei_read_32bit ((CEI_INT16)deviceIndex, virtualBaseAddressPtr, &intCtrlStatReg);
      if (status == ARS_NORMAL) {
         intCtrlStatReg &= ~interruptBit;
         status = cei_write_32bit ((CEI_INT16)deviceIndex, virtualBaseAddressPtr, intCtrlStatReg);
      }
   }

   if (status == ARS_NORMAL) {
      status = ceill_terminateInterrupt(dev_handle[deviceIndex]);
   }

   return status;
}


/*========================================================================* 
 * Device Access Routines for Windows Operating Systems 
 *========================================================================*/

/*===================================================================================================================*
 * ENTRY POINT:                           C E I _ R E A D _ 3 2 _ B I T
 *===================================================================================================================*
 *
 * FUNCTION:    Perform a 32-bit read from the specified device
 *
 * PARAMETERS:  CDEV_BOARD_TYPE deviceIndex  -- (input) device to reference (0 to 15, 127 for Windows)
 *              pCEI_UINT32 address          -- (input) location in host virtual memory to read
 *              pCEI_UINT32 data             -- (input) pointer in which to store the data value read
 * 
 * RETURN VAL:  For Visa and LabVIEW R/T: -- Visa 'viIn32' invocation return status, if not VI_SUCCESS
 *              ARS_NORMAL                -- Success
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine performs a 32-bit read access from a device PCI memory region
 *
 *===================================================================================================================*/
CEI_INT32 cei_read_32bit (CDEV_BOARD_TYPE deviceIndex, pCEI_UINT32 address, pCEI_UINT32 data)
{
   *data = *address;
   return ARS_NORMAL;
}

/*===================================================================================================================*
 * ENTRY POINT:                           C E I _ W R I T E _ 3 2 _ B I T
 *===================================================================================================================*
 *
 * FUNCTION:    Perform a 32-bit write to the specified device
 *
 * PARAMETERS:  CDEV_BOARD_TYPE deviceIndex  -- (input) device to reference (0 to 15, 127 for Windows)
 *              pCEI_UINT32 address          -- (input) location in host virtual memory to write
 *              CEI_UINT32 data              -- (input) the data value to write
 * 
 * RETURN VAL:  For Visa and LabVIEW R/T: -- Visa 'viOut32' invocation return status, if not VI_SUCCESS
 *              ARS_NORMAL                -- Success
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine performs a 32-bit write access to a device PCI memory region
 *
 *===================================================================================================================*/
CEI_INT32 cei_write_32bit (CDEV_BOARD_TYPE deviceIndex, pCEI_UINT32 address, CEI_UINT32 data)
{
   *address = data;
   return ARS_NORMAL;
}



