/*====================================================================================================================*
 * FILE:                                C O N F I G _ F R O M _ F I L E S . C
 *====================================================================================================================*
 *
 *   COPYRIGHT 2012 - 2019 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED
 *   ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH THE INCLUSION
 *   OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY OTHER COPIES THEREOF
 *   MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE
 *   AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS, INC. 
 *
 *====================================================================================================================*
 *
 * FUNCTION:    Demonstration program for the CEI-x30 ARINC products, showing
 *              methods to configure channels and define scheduled messages
 *              from the respective text and/or XML files.
 *
 * ENTRY POINTS:
 *              main/wrap                 - main program
 *              error_exit                - print a message and error exit.
 *
 * DESCRIPTION: The main routine in this file is a basic example test program
 *              demonstrating how to configure transmit and receive channels
 *              and populate the message scheduler table using the API 
 *              routines that perform those functions based on the content of 
 *              text and XML configuration/definition files.
 *
 *====================================================================================================================*
 *          
 * HISTORY:   
 *
 *   Date   Initials Comments
 * -------- -------- --------------------------------------------------------------------------------------------
 * 05/10/12   BDW    Created.
 * 10/18/16   BDW    Introduced use of CEIx30 API-specific type defines via the file CEIX30_TYPES.H allowing
 *                   for easy build to common use of only signed integer return types and unsigned integer
 *                   parameter types (with the exception of 64-bit types); specifically replacing use of
 *                   CEI_INT16 with parameter specific types for functions using short integer parameters.
 *
 *====================================================================================================================*/ 

/*---------------------------------------------------------------------------*
 *          INCLUDES, DEFINES and GLOBALS
 *---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC)
#include <conio.h>
#endif
#include <time.h>
//#include "cdev_hw.h"  
#include "cdev_api.h"  

/*===========================================================================*
 *              G L O B A L   C O N S T A N T S   A N D   D A T A
 *===========================================================================*/
#define SCHEDULED_MESSAGE_DEFINITION_TEXT_FILE  "8 msg def.txt"
#define CHANNEL_CONFIGURATION_TEXT_FILE         "2 chan cfg.txt"   

#define SCHEDULED_MESSAGE_DEFINITION_XML_FILE   "8 msg def.xml"
#define CHANNEL_CONFIGURATION_XML_FILE          "2 chan cfg.xml"   

#define MAX_SCHEDULED_MESSAGE_FILE_ENTRIES      10   /* should exceed file contents */
#define SCHEDULED_MESSAGE_FILE_ENTRIES_TO_READ  8    /* should match file contents */

static CEI_CHAR configurationString[40];   /* Board name string storage   */
static pCEI_CHAR pConfigurationString;     /* Board name string reference */
static pCEI_CHAR altConfigurationString[CEIDEV_NUM_ALT_CONFIGURATIONS] = {" ", "-J", " "};
static CDEV_CHAN_TYPE numTxChan;           /* Number of transmitters installed      */
static CDEV_CHAN_TYPE numRxChan;           /* Number of receivers installed         */

AR_SCHEDULED_MSG_ENTRY_TYPE msgEntry[MAX_SCHEDULED_MESSAGE_FILE_ENTRIES];

char * parityString[3]        = {"ODD ",  "EVEN",  "NONE"};
char * onOffString[2]         = {"ENABLED ",       "DISABLED" };
char * mergedString[2]        = {"MERGED    ",     "INDIVIDUAL" };
char * protocolErrorString[3] = {"BIT COUNT LOW ", "BIT COUNT HIGH", "SHORT GAP     "};

/*===========================================================================*
 * ENTRY POINT:                      M A I N
 *===========================================================================*
 *
 * FUNCTION:    Main entry point for the example internal wrap test.
 *
 * DESCRIPTION: This is a simple test program that sends out some data on a
 *              transmit channel and reads it back in on the respective  
 *              receive channel(s).
 *
 *===========================================================================*/
#ifdef __TURBOC__
   #pragma argsused
#endif

#if defined (VXW_PCI_X86) || defined (VXW_PCI_PPC)
CEI_VOID wrap(CEI_VOID)
#elif defined (_LINUX_X86_) || defined (_LINUX_PPC_)
CEI_NATIVE_INT main(CEI_NATIVE_INT argc, pCEI_CHAR argv[])
#else
__cdecl main(CEI_NATIVE_INT argc, pCEI_CHAR argv[])
#endif
{
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (INTEGRITY_PCI_PPC)
#if !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) 
   CEI_NATIVE_INT result;           /* status from scanf operator prompt     */
   CEI_NATIVE_INT argIndex;         /* Index of arguments passed to main()   */
#endif   
   CEI_NATIVE_INT deviceNum;        /* User-entered WinRT/Condor Device ID   */
#endif   
   CDEV_PARM_SSI_TYPE configuration; /* the device model/configuration       */
   CDEV_BOARD_TYPE board    = MAX_BOARDS; /* User-entered Device ID          */
   CDEV_API_RET_TYPE status = ARS_NORMAL; /* API function return status      */
   CDEV_PARM_SSI_TYPE itemGet;      /* General purpose chan cfg item state   */
   CDEV_CHAN_TYPE channel;          /* General purpose channel index         */
   CEI_NATIVE_INT fileSelect = 2;   /* File type to use: 0 - txt  1 - XML    */

   printf ("\n");
   printf ("     --------------------------------------------------------------\n");
   printf ("                  CEI-x30 Configuration File Usage Example         \n");
   printf ("     --------------------------------------------------------------\n");
   printf ("     This program demonstrates the use of the Channel Configuration\n");
   printf ("     and Scheduled Message Definition File features provided with  \n");
   printf ("     the CEI-x30 product distribution and AutoConfig ARINC tool.   \n");
   printf ("     File names used by this program are fixed, with optional file \n");
   printf ("     extensions depending on the file type selected, text or XML.  \n");
   printf ("                                                                   \n");
   printf ("     Those fixed file names are:  8 msg def.***                    \n");
   printf ("                                  2 chan cfg.***                   \n");
   printf ("                                                                   \n");
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (INTEGRITY_PCI_PPC)
   printf ("       Optional program arguments include:                         \n");
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_) 
   printf ("       Any Numeric Value - defines the device ID to reference      \n");
   printf ("       Any other character following the device ID is evaluated as:\n");
   printf ("          T - indicates use of the text file format                \n");
   printf ("          X - indicates use of the XML file format                 \n");
#else
   printf ("         /BD   - provides for an optional device ID to reference   \n");
   printf ("         /TXT  - indicates use of the text file format (default)   \n");
   printf ("         /XML  - indicates use of the XML file format              \n");
#endif   
#endif   
   printf ("     --------------------------------------------------------------\n");

   /*------------------------------------------------------------------------*
    * Process the possible argument values (prompt operator if not specified):
    * 1) Alternate Board Number/Device ID 
    * 2) Text configuration and msg definition file selection
    * 3) XML configuration and msg definition file selection
    *------------------------------------------------------------------------*/
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (INTEGRITY_PCI_PPC)
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
   /* Process Linux command line parameters */
   if (argc == 2) {
      sscanf (argv[1], "%d", &deviceNum);
      board = (CDEV_BOARD_TYPE)deviceNum;
   }
   else if (argc == 3) {
      char entry;
      sscanf (argv[1], "%d", &deviceNum);
      board = (CDEV_BOARD_TYPE)deviceNum;
      sscanf (argv[2], "%c", &entry);
      if (entry == 'T') {
         fileSelect = 0;
      }
      else if (entry == 'X') {
         fileSelect = 1;
      }
   }
   else
      board = 0;
#else
   /* Process Windows command line parameters */
   for (argIndex = 1; argIndex < argc; argIndex++) {
      if (strnicmp ("/BD", argv[argIndex], 3) == 0) {
         sscanf (argv[argIndex]+3, "%d", &board);
         if (board >= MAX_BOARDS) {
            printf ("\n\nERROR on Board ID parameter selection /BD%d\n\n", board);
         }
      }      
      else if (strnicmp ("/TXT", argv[argIndex], 4) == 0) {
         fileSelect = 0;
      }
      else if (strnicmp ("/XML", argv[argIndex], 4) == 0) {
         fileSelect = 1;
      }
   }

   /*------------------------------------------------------------------------*
    *  If not defined via argument, ask the user for the board/device number.
    *------------------------------------------------------------------------*/
   if (board >= MAX_BOARDS) {
      printf ("Enter the board number (installation device number): ");
      result = scanf("%d", &deviceNum);
      board = (CDEV_BOARD_TYPE)deviceNum;
      fflush(stdin);
   }

   if (fileSelect == 2) {
      printf ("Select the type of files to use, 0 for text or 1 for XML: ");
      result = scanf("%d", &fileSelect);
      fflush(stdin);
   }

#endif
#else
   fileSelect = 0;               /* fixed values for embedded RTOS platforms */
   board = (CDEV_BOARD_TYPE)0;
#endif

   /*------------------------------------------------------------------------*
    *  Initialize the board, exit if not found or it fails initialization.
    *------------------------------------------------------------------------*/
   status = ar_open (board);             /* Device Number of the board       */

   if (status != ARS_NORMAL) {
      printf ("\n\nFATAL ERROR: BOARD FAILED INITIALIZATION!\n");
      printf ("Error reported: \'%s\'\n", ar_get_error(status));
   #if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC)
      printf ("Press any key to exit\n");
      getch();
   #endif
      exit (0);
   }

   /*------------------------------------------------------------------------*
    *  Read the board type and number of channels installed
    *------------------------------------------------------------------------*/
   status = ar_get_device_config (board, 0, ARU_CONFIGURATION, &configuration);

   numTxChan = ar_num_xchans ( board );     /* Total number of transmitters  */
   numRxChan = ar_num_rchans ( board );     /* Total number of receivers     */
   
   /*------------------------------------------------------------------------*
    *  Display the detected CEI-x30 board type and channel configuration
    *------------------------------------------------------------------------*/
   if (configuration & CEIDEV_ALT_CONFIGURATION_717) {  
      numTxChan--;              /* -J module part # doesn't include 717 chan */
      numRxChan--;
   }
   
   pConfigurationString = ar_get_boardname (board, configurationString);
   
   printf ( "This program has detected a %s-%02d%02d%s\n\n", 
            configurationString, (CEI_NATIVE_INT)numRxChan, (CEI_NATIVE_INT)numTxChan, 
            altConfigurationString[(configuration & CEIDEV_ALT_CONFIGURATION_717) >> CEIDEV_ALT_CONFIG_717_SHIFT]);

   /*------------------------------------------------------------------------*
    *  Populate the Message Scheduler Table from the respective file type
    *------------------------------------------------------------------------*/
   if (fileSelect) {
      status = ar_define_messages_from_xml_file(SCHEDULED_MESSAGE_DEFINITION_XML_FILE);
   }
   else {
      status = ar_define_messages_from_txt_file(SCHEDULED_MESSAGE_DEFINITION_TEXT_FILE);
   }
   if (status == ARS_NORMAL) {
      /*------------------------------------------------------------------------*
       *  If all is well, read back the the first eight messages from the 
       *  message scheduler table and display the contents in the console window
       *------------------------------------------------------------------------*/
      printf ("\nMessage Scheduler Table Definition for Device #: %2d\n", board );
      status = ar_read_scheduled_msg_block (board, 0, SCHEDULED_MESSAGE_FILE_ENTRIES_TO_READ, msgEntry);
      if (status == ARS_NORMAL) {
           printf ("\n Msg   Board  Channel   Message");
           printf ("\nIndex  Index   Index     Value       Rate  Start Offset  Transmit Count");
         for (channel = 0; channel < 8; channel++) {
            printf ("\n%3d     %2d      %2d    0x%08X    %4d     %4d       %8d",
                    msgEntry[channel].messageIndex, msgEntry[channel].board, msgEntry[channel].channel, 
                    msgEntry[channel].data, msgEntry[channel].rate, msgEntry[channel].start, msgEntry[channel].txCount );
         }
      }
   }

   /*--------------------------------------------------------------------------*
    *  Configure the transmit and receive channels via the respective file type
    *--------------------------------------------------------------------------*/
   if (fileSelect) {
      status = ar_channel_configuration_from_xml_file(CHANNEL_CONFIGURATION_XML_FILE, board);
   }
   else {
      status = ar_config_channels_from_txt_file(board, CHANNEL_CONFIGURATION_TEXT_FILE);
   }

   if (status == ARS_NORMAL) {
      /*------------------------------------------------------------------------*
       *  If all is well, read back the the channel attribute settings and
       *  display the channel configurations in the console window
       *------------------------------------------------------------------------*/
      printf ("\n\n\nChannel Configuration for Device #: %2d\n", board );
      printf ("\nRX Chan FIFO Enable  Bit Rate   Parity      Buffer Mode   Internal Wrap");
      for (channel = 0; channel < numRxChan; channel++) {
         status = ar_get_device_config (board, channel, ARU_RX_FIFO_ENABLE, &itemGet);
         printf ("\n %2d      %s", channel, onOffString[itemGet-AR_ON] );
         
         status = ar_get_device_config (board, channel, ARU_RX_BITRATE, &itemGet);
         if (itemGet == ARU_SPEED_HIGH) 
            printf ("     100000");
         else if (itemGet == ARU_SPEED_LOW)
            printf ("      12500");
         else {
            itemGet = 16000000 / (itemGet+2);
            printf ("    %7d", itemGet);
         }
         
         status = ar_get_device_config (board, channel, ARU_RX_PARITY, &itemGet);
         printf ("    %s", onOffString[itemGet-AR_ON] );
         
         status = ar_get_device_config (board, channel, ARU_RX_MERGED_MODE, &itemGet);
         printf ("    %s ", mergedString[itemGet-AR_ON] );
         
         status = ar_get_device_config (board, channel, ARU_RECV_MODE, &itemGet);
         printf ("      %s", onOffString[itemGet] );
      }
   
      printf ("\n\nTX Chan FIFO Enable  Bit Rate  Parity Protocol Error Ext Disable Fast Slew Rate");
      for (channel = 0; channel < numTxChan; channel++) {
         status = ar_get_device_config (board, channel, ARU_TX_FIFO_ENABLE, &itemGet);
         printf ("\n %2d      %s", channel, onOffString[itemGet-AR_ON] );
         
         status = ar_get_device_config (board, channel, ARU_TX_BITRATE, &itemGet);
         if (itemGet == ARU_SPEED_HIGH) 
            printf ("     100000");
         else if (itemGet == ARU_SPEED_LOW)
            printf ("      12500");
         else {
            itemGet = 16000000 / (itemGet+2);
            printf ("    %7d", itemGet);
         }
         
         status = ar_get_device_config (board, channel, ARU_TX_PARITY, &itemGet);
         printf ("    %s   ", parityString[itemGet-ARU_PARITY_ODD] );
         
         status = ar_get_device_config (board, channel, ARU_TX_BIT_ERROR, &itemGet);
         if (itemGet == AR_OFF) {
            status = ar_get_device_config (board, channel, ARU_TX_GAP_ERROR, &itemGet);
            if (itemGet == AR_OFF)
               printf ("%s       ", onOffString[1] );
            else
               printf ("%s ", protocolErrorString[2] );
         }
         else 
            printf ("%s ", protocolErrorString[itemGet-AR_LO] );

         status = ar_get_device_config (board, channel, ARU_TX_DISABLE, &itemGet);
         printf ("%s", onOffString[itemGet-AR_ON] );
   
         status = ar_get_device_config (board, channel, ARU_FAST_SLEW_RATE, &itemGet);
         printf ("     %s", onOffString[itemGet-AR_ON] );
      }
   }

   /*------------------------------------------------------------------------*
    *  Initiate board message processing.
    *------------------------------------------------------------------------*/
   status = ar_go (board);
   
   /*------------------------------------------------------------------------*
    *  Now that we are done, print out final status.
    *------------------------------------------------------------------------*/
   if (status != ARS_NORMAL)
      printf("\nLast Error detected: %s\n", ar_get_error (status));
   else
      printf("\nTest PASSED, no errors were detected.\n");

#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC)
   printf("Press any key to exit\n");
   getch();
#endif

   /*------------------------------------------------------------------------*
    *  Release all resources acquired when this board was first accessed.
    *------------------------------------------------------------------------*/
   status = ar_stop (board);
   status = ar_close (board);

#if defined (_LINUX_X86_) || defined (_LINUX_PPC_) 
   return 0;
#endif
}

