/*====================================================================================================================*
 * FILE:                                 M U L T P R O C R O C E S _ T E S T . C 
 *====================================================================================================================*
 *
 *   COPYRIGHT 2005 - 2019 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED
 *   ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH THE INCLUSION
 *   OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY OTHER COPIES THEREOF
 *   MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE
 *   AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS, INC. 
 *
 *====================================================================================================================*
 *
 * FUNCTION:
 *  
 *      This program is a demonstration illustrating the use of a single 
 *      CEI-x30 device in conjunction with multiple processes.
 *
 * DESCRIPTION: 
 *  
 *      This example application is based on separate processes, one for ARINC 
 *      429 scheduled message transmission on multiple transmit channels and
 *      multiple others for ARINC 429 message reception on individual receive
 *      channels.
 *
 *      A single "transmit" process should be launched first, with subsequent  
 *      invocation of one or more "receive" processes.  The "transmit" process 
 *      will load the board via AR_LOADSLV, configure board and channel
 *      specific parameters via AR_SET_DEVICE_CONFIG, and activate ARINC 429
 *      processing via AR_GO.  After the "transmit" process invokes AR_GO, it
 *      is then permissible to launch one or more "receive" processes.  A 
 *      "receive" process will first attach to the board via AR_LOADSLV, 
 *      then execute operations strictly confined to the particular channel(s)
 *      to which that process is associated (e.g. AR_SET_DEVICE_CONFIG,
 *      AR_GETWORD, etc.).  When finished, each "receive" process should 
 *      invoke AR_CLOSE and terminate.  The "transmit" process should remain 
 *      running until all "receive" processes have terminated; upon termination
 *      the "transmit" process should invoke AR_STOP and AR_CLOSE and terminate.
 *
 *      Note that while the previously discussed application process invocation
 *      order is recommended, it is not strictly required.  You may actually
 *      invoke a "receive" process first and terminate it last; however,
 *      multi-process applications should rely on a primary board-control
 *      process as the focus for board initialization, BIT functionality, and 
 *      active h/w message processing control.
 *
 *      Execute this sample program with the command-line parameter "board" to
 *      launch the board process, or specify "channel" to launch a channel 
 *      process.
 *     
 *      Note that multi-process support is only available for certain CEI-x30
 *      boards and operating systems.  Refer to the documentation for the
 *      AR_SET_PRELOAD_CONFIG routine for a list of supported boards/platforms
 *      as well as other special programming considerations when using 
 *      multiple processes in conjunction with the CEI-x30 API.
 *
 * HISTORY:
 *
 *   Date     By   Vers                        Comments
 * --------  ----  ----  --------------------------------------------------------------------------------------------
 * 03/03/08  BDW   1.50  Modified for use with the CEI-x30 API
 * 07/31/12  BDW   4.00  Reduced the length of the program arguments
 * 07/31/12  BDW   4.00  Reduced the length of the program arguments
 * 10/18/16  BDW   4.90  Introduced use of CEIx30 API-specific type defines via the file CEIX30_TYPES.H allowing
 *                       for easy build to common use of only signed integer return types and unsigned integer
 *                       parameter types (with the exception of 64-bit types); specifically replacing use of
 *                       CEI_INT16 with parameter specific types for functions using short integer parameters.
 *
 *====================================================================================================================*/ 
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC)
#include <conio.h>
#endif
#include <time.h>
#include <ctype.h>
#include <assert.h>
#include "cdev_api.h"


/* global: boardIsOpen                                                       */
/* usage:  True if the board has been opened with a call to AR_OPEN,         */
/*         false otherwise.                                                  */
CDEV_PARM_SSI_TYPE boardIsOpen = FALSE;

/* global: board                                                             */
/* usage:  CEI-x30 device index associated with this process.                */
CDEV_BOARD_TYPE board;

static pCEI_CHAR altConfigurationString[CEIDEV_NUM_ALT_CONFIGURATIONS] = {" ", "-J", " "};

/*===========================================================================*
 * INTERNAL ROUTINE:              A P P _ E X I T
 *===========================================================================*
 *
 * FUNCTION:    Execute application shutdown.
 *
 * RETURN VAL:  none
 *
 * DESCRIPTION: Display the given message (if any), close the board (if it's
 *              open), prompt for the user to hit return, and terminate the
 *              application.
 *
 *===========================================================================*/
static void app_exit (CEI_CHAR *msg)
{
   CDEV_API_RET_TYPE status;

   /* display error message (if any)  */
   if (msg != NULL) {
      printf("%s\n",msg);
   }

   /* check if the board is currently open  */
   if (boardIsOpen) {
      boardIsOpen = FALSE;
      
      /* attempt to close the board  */
      status = ar_close(board);
      if (status != ARS_NORMAL) {
         /* display error description  */
         printf("Error while closing board %d:\n",board);
         printf("  Error reported:  \'%s\'\n",ar_get_error(status));
         printf("  Additional info: \'%s\'\n",ar_get_error(ARS_LAST_ERROR));
      }
   }

   /* prompt the user to hit return  */
   printf("\nPress <Enter> to exit...\n");
   getchar();

   /* terminate the application  */
   exit(0);
}


/*===========================================================================*
 * INTERNAL ROUTINE:           E R R O R _ E X I T
 *===========================================================================*
 *
 * FUNCTION:    Display description of given error and call app_exit.
 *
 * RETURN VAL:  none
 *
 * DESCRIPTION: Display text describing the given CEI-x30 API status code. 
 *              Also display the last error description (if any), then call
 *              app_exit.
 *
 *===========================================================================*/
static void error_exit (CDEV_BOARD_TYPE board, CDEV_API_RET_TYPE status)
{
    /* display the error message  */
    printf("Error while testing board %d:\n",board);
    printf("  Error reported:  \'%s\'\n",ar_get_error(status));

    /* handle process shut down  */
    app_exit(NULL);
}


/*===========================================================================*
 * ENTRY POINT:                     M A I N
 *===========================================================================*
 *
 * FUNCTION:    Sample application entry point.
 *
 * DESCRIPTION: Execute sample board or channel process.
 * 
 *===========================================================================*/
int main(int argc,char *argv[])
{

   CEI_CHAR api_ver[150];            /* API version being used                    */
   CEI_CHAR configurationString[20]; /* board name string                         */
   CDEV_API_RET_TYPE status;         /* API status value                          */
   CDEV_CHAN_TYPE channel = 0;       /* channel being used                        */
   CDEV_CHAN_TYPE numTransmitters;   /* number of transmitters purchased          */
   CDEV_CHAN_TYPE numReceivers;      /* number of receivers purchased             */
   CDEV_CHAN_TYPE numChanMatch;      /* number of xmtr/rcvr channel pairs         */
   CEI_INT32 isTransmitProcess;      /* true if board proc, false if receive proc */
   CDEV_PARM_SSI_TYPE configuration; /* the device model/configuration            */
   CEI_UINT32 msgIndex;              /* the received ARINC 429 message            */
   CEI_UINT32 rxWord;                /* the received ARINC 429 message            */
   CEI_UINT32 rxTimeTag;             /* timetag of the received ARINC 429 message */
   CEI_NATIVE_INT terminateFlag;     /* flag indicating operator terminated test  */

   /* obtain the API version string  */
   ar_version (api_ver);

   /* display the program header  */
   printf("                                                                      \n");
   printf("     -----------------------------------------------------------------\n");
   printf("     CEI-x30 Multi-Process Example Application                        \n");
   printf("     API Version %s\n", api_ver                                          );
   printf("     -----------------------------------------------------------------\n");
   printf("     Execute either a transmit-all or receive-one process.  Refer to  \n");
   printf("     the comment header in multiprocess_test.c for details.           \n");
   printf("     -----------------------------------------------------------------\n");
   printf("                                                                      \n");

   /* check if this is a board process or a channel process  */
   if ((argc == 3) && (strcmp(argv[1],"tx") == 0)) {
      /* executing a board process  */
      isTransmitProcess = TRUE;
      board = (CDEV_BOARD_TYPE)atoi(argv[2]);
      printf("Executing 'transmit' process.\n\n");
   }
   else if ((argc == 4) && (strcmp(argv[1],"rx") == 0)) {
      /* executing a channel process  */
      isTransmitProcess = FALSE;
      board = (CDEV_BOARD_TYPE)atoi(argv[2]);
      channel = (CDEV_CHAN_TYPE)atoi(argv[3]);
      printf("Executing 'receive' process.\n\n");
   }
   else {
      /* display program usage  */
      printf("CEI-x30 Multi-Process Example Program command-line usage:\n\n");
      printf("  'multiproc tx <b>'     to invoke a transmit process for board <b>\n");
      printf("  'multiproc rx <b> <c>' to invoke a receive process for board <b> channel <c>\n\n");
      printf("Use 'multiproc64' under 64-bit Windows and Linux operating systems.\n");
      printf("All board and channel indices are zero-based.  That is, pass zero as <b> to\n");
      printf("select the first board; pass zero as <c> to select the first receive channel.\n");
      printf("\nPress <Enter> to exit...\n");
      getchar();
      return 0;
   }

   /* enable multi-process support  */
   status = ar_set_preload_config (board, ARU_CONCURRENCY_MODE, AR_CONC_MULTIPROC);
   if (status != ARS_NORMAL) {
      error_exit (board, status);
   }

   /* open a session and load the board (address, port, and size parameters are not used)  */
   status = ar_open (board);                           /* Device Number of the board       */
   if (status != ARS_NORMAL) {
      error_exit (board, status);
   }

   /* the board has been loaded  */
   boardIsOpen = TRUE;

   /* display the configuration with the number of receivers and transmitters  */
   numTransmitters = ar_num_xchans (board);
   numReceivers = ar_num_rchans (board);

   status = ar_get_device_config (board, 0, ARU_CONFIGURATION, &configuration);
   if (configuration & CEIDEV_ALT_CONFIGURATION_717) {  
      numTransmitters--;   /* -J module part # doesn't include 717 channel */
      numReceivers--;
   }
   
   status = (CDEV_API_RET_TYPE)ar_get_boardnameLV ((CEI_UINT32)board, configurationString);

   if (configuration == CEIDEV_CONFIG_R830RX) {
      printf ("This program has detected a %s\n\n", configurationString);
      app_exit ("This board configuration is not supported by this example.");
   }
   else {
      printf ( "This program has detected a %s-%02d%02d%s\n\n", 
               configurationString, (CEI_NATIVE_INT)numReceivers, (CEI_NATIVE_INT)numTransmitters, 
               altConfigurationString[(configuration & CEIDEV_ALT_CONFIGURATION_717) >> CEIDEV_ALT_CONFIG_717_SHIFT]);
   }

   numChanMatch = (CDEV_CHAN_TYPE)((numTransmitters < numReceivers) ? numTransmitters : numReceivers);

   /* check if this is a board or channel process  */
   if (isTransmitProcess) {

      /* define a single scheduled message on each transmit channel  */
      for (channel = 0; channel < numChanMatch; channel++)
      {
         /* transmit @ 12.5kbs */
         status = ar_set_device_config (board, channel, ARU_TX_BITRATE, ARU_SPEED_LOW);
         
         /* sample message rate is 1.0 sec for ch 0, 1.1 sec for ch 1, 1.2 sec for ch 2, etc.  */
         /* message content is the board number in the upper 16 bits and 
          * the transmit channel number in the lower 16 bits */
         status = ar_define_msg ( board,
                                  channel,
                                  (CDEV_PARM_SSI_TYPE)(1000 + (channel*100)),
                                  0,
                                  (CDEV_PARM_SI_TYPE)(board << 16) | channel );

         if (status == ARS_FAILURE)
            error_exit (board, status);
      }

      /* initiate active message processing on the board  */
      status = ar_go (board);

      if (status != ARS_NORMAL) {
         error_exit (board, status);
      }

      /* wait for <enter> indicating all channel processes have completed  */
      printf ("Board initialization and launch is complete.  One or more channel     \n");
      printf ("processes may now be created.  After all channel processes have       \n");
      printf ("terminated, press <enter> to terminate this board process.            \n");
      getchar();

      /* close the board and terminate the program  */
      status = ar_stop (board);
      status = ar_reset (board);
      app_exit ("Shutting down board process.");
   }
   else {
      /* make sure the channel index is valid  */
      if (channel >= numChanMatch)
         app_exit ("Error: invalid channel index was supplied.");

      /* receive @ 12.5kbs */
      status = ar_set_device_config (board, channel, ARU_RECV_MODE, AR_WRAP_ON);
      status = ar_set_device_config (board, channel, ARU_RX_BITRATE, ARU_SPEED_LOW);
      
      /* display the process start time  */
      printf ("process (board %d, channel %d): process launched at board time %d\n", board,channel, ar_get_timercntl(board));

      /* flush the receive buffer for this channel  */
      while (ar_getwordt (board, channel, &rxWord, &rxTimeTag) == ARS_GOTDATA);

      /* loop until 25 words have been received - display each received ARINC 429 message  */
      for (msgIndex = 0, terminateFlag = 0; (msgIndex < 25) && (terminateFlag == 0); msgIndex++) {

         /* attempt to read a message from the receiver, using the API routine with a built in delay */
         while ((status = ar_getnextt (board, channel, &rxWord, &rxTimeTag)) == ARS_CHAN_TIMEOUT) {
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC)
            if (kbhit()) {
               terminateFlag = 1;
               break;
            }
#endif
         }
         if (status != ARS_GOTDATA) {
            error_exit (board, status);
         }
         /* display the ARINC 429 message and timetag  */
         printf ("Receive process (board %d, channel %d): received 0x%08x at time %d\n", board, channel, rxWord, rxTimeTag);
      }

      /* close the board and terminate the program  */
      app_exit ("Sample channel process complete.");
   }

   /* control should never reach here  */
   return 0;
}


