/*====================================================================================================================*
 * FILE:                                     C D E V _ A P I _ A 7 1 7 . C                    
 *====================================================================================================================*
 *
 *   COPYRIGHT 2016 - 2019 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED
 *   ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH THE INCLUSION
 *   OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY OTHER COPIES THEREOF
 *   MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE
 *   AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS, INC. 
 *
 *====================================================================================================================*
 *
 * FUNCTION:    ARINC 573/717 Specific Application Programmer's Interface functionality for a CEI-x30 ARINC device.
 *
 * DESCRIPTION: These routines are provided to demonstrate the ARINC 717-specific programming primitives for the 
 *              CEI-x30 API using the C programming language.  Each routine implements a particular API function; 
 *              however, a user application program may need to combine functions in different ways to perform the 
 *              desired operation.
 *
 *              Most routines utilize a board number parameter, zero-based on the Device ID/Board Number assigned 
 *              during the installation of CEI-x30-SW distribution.  Channel parameter values are all zero-based.
 *
 * EXTERNALLY DEFINED ROUTINE ENTRY POINTS:         
 *
 *    CEI-x30 ARINC 717-specific API routines:
 *
 *    ar_get_573_config           - returns 573 config register bit states
 *    ar_get_573_frame            - retrieves an ARINC 573 data frame
 *    ar_get_transmitter_mode     - returns current 573/CSDB transmitter mode
 *    ar_set_transmitter_mode     - set ARINC 573/CSDB encoding (BPRZ vs. HBP)
 *    ar_put_573_frame            - transmit an ARINC 573 frame
 *    ar_set_573_config           - set ARINC 573-specific chan attributes  
 *
 * INTERNAL UTILITY ROUTINES:
 *
 *    a717_utl_initialize_global_data
 *
 *--------------------------------------------------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date    By   Ver. Comments
 * -------- ----- ---- -----------------------------------------------------------------------------------------------
 * 10/18/16  BDW  4.90 Initial version.  For modification history related to the routines contained herein prior to
 *                     and including v4.90, see the history section in the header of the file CDEV_API.C.
 *
 *====================================================================================================================*/ 
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
   #include <unistd.h>
   #include <fcntl.h>
   #include <string.h>
   #include <pthread.h>
   #include <errno.h>
   #include <sys/stat.h> 
   #include <sys/shm.h>
   #include <sys/sem.h>
   #include <sys/syscall.h>
   #define VOLATILE volatile
#elif defined (VXW_PCI_X86) || defined (VXW_PCI_PPC)
   #include <stdio.h>
   #include <stdlib.h>
   #include <unistd.h>
   #define VOLATILE 
#elif defined (NIVISA_PXI)
   #ifdef LABVIEW_RT
      #include <ansi_c.h> 
   #endif
   #include <visa.h>  
   #define VOLATILE 
#elif defined (INTEGRITY_PCI_PPC)
   #include <string.h>
   #define VOLATILE 
#else
   #define VOLATILE volatile
#endif

#include "cdev_glb.h"

extern VOLATILE ARINC_LOCAL_TYPE device[MAX_BOARDS];
extern ARINC_GLOBAL_TYPE * deviceGlobal[MAX_BOARDS];

/* Local Structure Array "a717info"
 * used in ar_get_573_config() to translate config register bit field values to parameter constant values,
 * bus-speed & frame-size delay and retry counts specific to ar_get_573_frame frame retrieval algorithm,
 * indexed by config reg BAUD RATE bit field value */
static AR_717_API_LOCAL_CONFIG_AND_CONTROL_TYPE a717info[A717_NUM_SUPPORTED_BUS_SPEEDS] = {
  /* Bus Speed/Sub-Frame Size         Sub-Frame Size              Rx Delay           Retry Count
   * a717busSpeedOption             a717frameSizeOption     a717RxDelayDuration  a717RxTimeOutCount; */
   {ARU_573_RATE_SIZE_384_32,      ARU_573_SUB_FRAME_32,            32,                  128},
   {ARU_573_RATE_SIZE_768_64,      ARU_573_SUB_FRAME_64,            16,                  256},
   {ARU_573_RATE_SIZE_1536_128,    ARU_573_SUB_FRAME_128,            8,                  512},
   {ARU_573_RATE_SIZE_3072_256,    ARU_573_SUB_FRAME_256,            4,                 1024},
   {ARU_573_RATE_SIZE_6144_512,    ARU_573_SUB_FRAME_512,            2,                 2048},
   {ARU_573_RATE_SIZE_12288_1024,  ARU_573_SUB_FRAME_1024,           1,                 4096},
   {ARU_573_RATE_SIZE_24576_2048,  ARU_573_SUB_FRAME_2048,           1,                 4096},
   {ARU_573_RATE_SIZE_49152_4096,  ARU_573_SUB_FRAME_4096,           1,                 4096}
};

/* local storage of ARINC 717 sync words for app retrieval */
static CEI_UINT32 a717syncWords[MAX_BOARDS][NUM_SUBFRAME_SYNC_WORDS];

/*===================================================================================================================*
 * ENTRY POINT:                A 7 1 7 _ U T L _ I N I T I A L I Z E _ G L O B A L _ D A T A
 *===================================================================================================================*
 *
 * FUNCTION:    Function that initializes the ARINC 717 API stored sync words
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board -- (input) board to reference (0 to 127)
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     a717syncWords[][]
 *
 * DESCRIPTION: This routine initializes the internal API sub-frame sync word set (4) for the specified device.
 *
 *===================================================================================================================*/ 
CEI_VOID  a717_utl_initialize_global_data 
(
   CDEV_BOARD_TYPE board                                                        /* board to reference (0 to 127)     */
)
{
   CEI_INT32 index;                                                                             /* general use index */
   for (index = 0; index < NUM_SUBFRAME_SYNC_WORDS; index++) {
      a717syncWords[board][index] = 0U;
   }
}



/*===================================================================================================================*
 * ENTRY POINT:                               A R _ G E T _ 5 7 3 _ C O N F I G
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve an ARINC 573 channel configuration register bit field def.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input) board to reference (0 to 127)
 *              CDEV_PARM_SSI_TYPE item -- (input) item to process, valid items are:
 *                                          ARU_RECV_MODE           
 *                                          ARU_RX_BITRATE          
 *                                          ARU_TX_BITRATE          
 *                                          ARU_RX_FIFO_ENABLE
 *                                          ARU_TX_FIFO_ENABLE
 *                                          ARU_573_RX_ENABLE
 *                                          ARU_573_RX_SUBFRAME_SIZE       
 *                                          ARU_573_RX_AUTO_DETECT  
 *                                          ARU_573_RX_BPRZ_SELECT  
 *                                          ARU_573_TX_BPRZ_SELECT
 *                                          ARU_573_TX_HBP_SELECT
 *                                          ARU_573_TX_SLEW_RATE    
 *                                          ARU_573_SYNC_WORD1
 *                                          ARU_573_SYNC_WORD2
 *                                          ARU_573_SYNC_WORD3
 *                                          ARU_573_SYNC_WORD4
 * 
 *              CDEV_PARM_SI_PTR_TYPE value  -- (output) the state of the item specified
 *                                                       with valid item/value combinations:
 *
 *                  item                     valid values
 * 
 *                 ARU_RECV_MODE            AR_WRAP_OFF, AR_WRAP_ON
 *                 ARU_RX_FIFO_ENABLE       AR_OFF, AR_ON
 *                 ARU_573_RX_ENABLE        AR_OFF, AR_ON (same as ARU_RX_FIFO_ENABLE)
 *                 ARU_573_RX_AUTO_DETECT   AR_OFF, AR_ON
 *                 ARU_573_RX_BPRZ_SELECT   AR_OFF, AR_ON
 *                 ARU_573_TX_BPRZ_SELECT   AR_OFF, AR_ON
 *                 ARU_573_TX_HBP_SELECT    AR_OFF, AR_ON
 *                 ARU_RX_BITRATE &         ARU_573_RATE_SIZE_384_32    
 *                 ARU_TX_BITRATE           ARU_573_RATE_SIZE_768_64    
 *                                          ARU_573_RATE_SIZE_1536_128  
 *                                          ARU_573_RATE_SIZE_3072_256  
 *                                          ARU_573_RATE_SIZE_6144_512  
 *                                          ARU_573_RATE_SIZE_12288_1024
 *                                          ARU_573_RATE_SIZE_24576_2048
 *                                          ARU_573_RATE_SIZE_49152_4096
 *                 ARU_573_RX_SUBFRAME_SIZE ARU_573_SUB_FRAME_64  
 *                                          ARU_573_SUB_FRAME_128 
 *                                          ARU_573_SUB_FRAME_256 
 *                                          ARU_573_SUB_FRAME_512 
 *                                          ARU_573_SUB_FRAME_1024
 *                                          ARU_573_SUB_FRAME_2048
 *                                          ARU_573_SUB_FRAME_4096
 *                 ARU_TX_FIFO_ENABLE       AR_OFF, AR_ON
 *                 ARU_TX_DISABLE           AR_OFF, AR_ON (inverse of ARU_TX_FIFO_ENABLE)
 *                 ARU_573_TX_SLEW_RATE     ARU_573_TX_SLEW_1PT5
 *                                          ARU_573_TX_SLEW_10PT0
 *                 ARU_RX_MERGED_MODE       AR_ON, AR_OFF
 *                 ARU_573_SYNC_WORD1       value from 0x0000 to 0x0fff
 *                 ARU_573_SYNC_WORD2       value from 0x0000 to 0x0fff
 *                 ARU_573_SYNC_WORD3       value from 0x0000 to 0x0fff
 *                 ARU_573_SYNC_WORD4       value from 0x0000 to 0x0fff
 *                 ARU_TX_FIFO_COUNT        value from 0 to 8191
 *      
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL      Normal succesful completion.
 *                                ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                                ARS_INVARG      "item" parameter value is not supported by this function
 *                                ARS_INVHARVAL   Item was invalid or channel not supported on this device
 *                                ARS_INVBOARD    Board parameter value invalid or board has not been initialized
 * 
 * GLOBALS:     device[]
 *              deviceGlobal[]
 *              
 * DESCRIPTION: Supplies the current state of the "item" requested from the device.
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_get_573_config 
( 
   CDEV_BOARD_TYPE board,                                                  /* (input)  board to reference (0 to 127) */
   CDEV_PARM_SSI_TYPE item,                                                /* (input)  item to process               */
   CDEV_PARM_SI_PTR_TYPE value                                             /* (output) state of the item             */
)
{
   CEI_INT32 accessStatus = ARS_NORMAL;                                        /* status of device read/write access */
   CDEV_PARM_SI_TYPE state = 0;                                              /* temporary config reg bit field state */
   CEI_UINT32 deviceChannel;                                                 /* temp device-specific channel number  */
   CEI_UINT32 bitFieldValue = 0;                                             /* temp for device register read content*/
   CEI_UINT32 memOffset;                                                      /* local offset to channel memory area */
   CEI_UINT32 offsetIntoChannelWindow = 0;                                      /* offset into chan mem region       */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                       /* function return status            */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr = NULL;                     /* pointer to board local structure  */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr = NULL;                            /* pointer to board global structure */

   /* return state is based on the value of the respective shadow 
    * register's bit field 
    */
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   else {
#endif
      deviceLocalStructPtr = &device[board];
      deviceGlobalStructPtr = deviceGlobal[board];
      deviceChannel = (CEI_INT32)0;  /* here for future growth w/channel parm */ 

      /* offset the first installed CSDB channel by the specified channel no. */
      switch (item) {
         case ARU_RX_BITRATE:
         case ARU_RX_PARITY:
         case ARU_RECV_MODE:
         case ARU_RX_DISABLE:
         case ARU_RX_FIFO_ENABLE:
         case ARU_RX_MERGED_MODE:
         case ARU_573_RX_ENABLE:
         case ARU_573_RX_SUBFRAME_SIZE:
         case ARU_573_RX_AUTO_DETECT:
         case ARU_573_RX_BPRZ_SELECT:
            deviceChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX];
            offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
            break;
   
         case ARU_TX_BITRATE:
         case ARU_TX_DISABLE:
         case ARU_TX_FIFO_ENABLE:
         case ARU_573_TX_SLEW_RATE:
         case ARU_573_TX_BPRZ_SELECT:
         case ARU_573_TX_HBP_SELECT:
            deviceChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX];
            offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
            break;

         case ARU_TX_FIFO_COUNT:
            deviceChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX];
            offsetIntoChannelWindow = CEIDEV_CHANNEL_STATUS;
            break;
   
         case ARU_RX_FIFO_COUNT:
            offsetIntoChannelWindow = CEIDEV_CHANNEL_STATUS;
            deviceChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX];
            break;

         case ARU_573_SYNC_WORD1:
         case ARU_573_SYNC_WORD2:
         case ARU_573_SYNC_WORD3:
         case ARU_573_SYNC_WORD4:
            *value = (CDEV_PARM_SI_TYPE)a717syncWords[board][item-ARU_573_SYNC_WORD1];
            return status;
   
         default:
            status = ARS_INVARG;
      }

#ifndef TARGET_EMBEDDED
      /* validate the specified channel */
      status = a429_utl_validate_channel_request (board, deviceChannel, item);

      if (status == ARS_NORMAL) {
         /* if multi-thread protection requested, semaphore/mutex-lock IRIG sample time register access */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (accessStatus == ARS_NORMAL) {
#endif
            /* read the device register with respect to the attribute specified */
            memOffset = CEIDEV_CHANNEL_REGION_START + offsetIntoChannelWindow + (deviceChannel * CEIDEV_CHANNEL_REGION_SIZE);
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &bitFieldValue);
#ifndef TARGET_EMBEDDED
         }
         /* if multi-thread protection is active, free any acquired lock */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
         }
#endif
         if (accessStatus == ARS_NORMAL) {

            switch (item) {
               case ARU_RECV_MODE:
                  state = (bitFieldValue & CEIDEV_INTERNAL_WRAP_ENABLE) ? AR_WRAP_ON : AR_WRAP_OFF;
                  break;
         
               case ARU_573_RX_ENABLE:
               case ARU_RX_FIFO_ENABLE:
               case ARU_TX_FIFO_ENABLE:
                  state = (bitFieldValue & CEIDEV_CHANNEL_ENABLE) ? AR_ON : AR_OFF;
                  break;

               case ARU_RX_DISABLE:
                  /* invert the state for CEI-710 legacy ARU_RX_DISABLE option support */
                  state = (bitFieldValue & CEIDEV_CHANNEL_ENABLE) ? AR_OFF : AR_ON;
                  break;
         
               case ARU_573_RX_AUTO_DETECT:  
                  state = (bitFieldValue & CEIDEV_573_RX_MODE_RAW) ? AR_OFF : AR_ON;
                  break;
         
               case ARU_573_RX_BPRZ_SELECT:  
                  state = (bitFieldValue & CEIDEV_573_RX_BPRZ_SELECT) ? AR_ON : AR_OFF;
                  break;
         
               case ARU_RX_MERGED_MODE:
                  state = (bitFieldValue & CEIDEV_MERGE_MODE_ENABLE) ? AR_ON : AR_OFF;
                  break;
   
               case ARU_RX_BITRATE:
                  state = bitFieldValue & CEIDEV_573_FRAME_49152_4096;
                  state = state >> CEIDEV_573_FRAME_SHIFT;
                  state = a717info[state].a717busSpeedOption;
                  break;
         
               /* the following case is only provided for CEI-710 backward compatibility */
               case ARU_573_RX_SUBFRAME_SIZE:
                  state = bitFieldValue & CEIDEV_573_FRAME_49152_4096;
                  state = state >> CEIDEV_573_FRAME_SHIFT;
                  state = a717info[state].a717frameSizeOption;
                  break;
         
               case ARU_TX_BITRATE:
                  state = bitFieldValue & CEIDEV_573_FRAME_49152_4096;
                  state = state >> CEIDEV_573_FRAME_SHIFT;
                  state = a717info[state].a717busSpeedOption;
                  break;
         
               case ARU_573_TX_SLEW_RATE:    
                  state = (bitFieldValue & CEIDEV_573_SLEW_1PT5_SELECT) ? ARU_573_TX_SLEW_1PT5 : ARU_573_TX_SLEW_10PT0;
                  break;
         
               case ARU_573_TX_BPRZ_SELECT:
                  state = (bitFieldValue & CEIDEV_573_TX_BPRZ_SELECT) ? AR_ON : AR_OFF;
                  break;
         
               case ARU_573_TX_HBP_SELECT:
                  state = (bitFieldValue & CEIDEV_573_TX_HBP_SELECT) ? AR_ON : AR_OFF;
                  break;
         
               case ARU_TX_DISABLE:
                  /* invert the state for CEI-710 legacy ARU_TX_DISABLE option support */
                  state = (bitFieldValue & CEIDEV_CHANNEL_ENABLE) ? AR_OFF : AR_ON;
                  break;
         
               case ARU_RX_FIFO_COUNT:
                  memOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX] * CEIDEV_CHANNEL_REGION_SIZE);
                  accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &bitFieldValue);
                  if (bitFieldValue & CEIDEV_MERGE_MODE_ENABLE) {
                     memOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS + (deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_MERGED_RX] * CEIDEV_CHANNEL_REGION_SIZE);
                     accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &bitFieldValue);
                  }
                  else {
                     memOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX] * CEIDEV_CHANNEL_REGION_SIZE);
                     accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &bitFieldValue);
                  }
                  state = (CEI_INT32)(bitFieldValue & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT;
                  break;

               case ARU_TX_FIFO_COUNT:
                  state = (CEI_INT32)(bitFieldValue & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT;
                  break;
   
               default:
                  status = ARS_INVARG;
            }
         }
#ifndef TARGET_EMBEDDED
         else
            status = (CDEV_API_RET_TYPE)accessStatus;
      }
   }
#endif
   *value = (CDEV_PARM_SI_TYPE)state;
   
   return status;
}
 

/*===================================================================================================================*
 * ENTRY POINT:                               A R _ G E T _ 5 7 3 _ F R A M E
 *===================================================================================================================*
 *
 * FUNCTION:    Read an ARINC 573/717 frame from the ARINC 717 receive buffer. 
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input)  Board number of interest (0 - 127)                          
 *              pCEI_UINT32 numberWords -- (input)  The number of ARINC 717 messages to read from the buffer
 *                                         (output) The number of ARINC 717 messages actually read from the buffer
 *              pCEI_UINT16 arincData   -- (output) Pointer to the ARINC 573 buffer
 * 
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NODATA      No frame data was available
 *                                ARS_GOTDATA     Some frame data was retrieved
 *                                ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                                ARS_INVBOARD    Invalid board parameter or board not initialized
 *                                ARS_INVHARVAL   Protocol not available on device
 *                                ARS_INVARG      Invalid numberWords or arincData parameter
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: When the ARINC 573/717 receiver is configured for auto-sync, this routine will read an entire data 
 *              frame based on the configured receiver sub-frame size (from the configuration register speed setting).
 *              If the receiver is configured for raw reception, all available data will be retrieved from the 
 *              receive channel buffer.  If the receive buffer doesn't contain any data, it will return a zero; 
 *              otherwise, it will read and return what is available.
 * 
 *              The ARINC 573/717 frame (message) data is returned as an array of 16-bit unsigned values.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_get_573_frame 
( 
   CDEV_BOARD_TYPE board,                                                /* (input)  board to reference (0 to 127)   */
   pCEI_UINT32 numberWords,                                              /* (in/output) number words to read/actual  */
   pCEI_UINT16 arincData                                                 /* (output) data read from 717 buffer       */
)
{
   CEI_INT32 rxFifoCount;                                                      /* Receive Buffer FIFO Count value    */
   CEI_INT32 timeOutLimit = 0;                                                 /* local time-out duration selection  */
   pCEI_UINT16 arincDataPtr = NULL;                                            /* local ptr to applic. data array    */
   CEI_UINT32 localChannel = 0U;                                               /* merged/individual channel select   */
   CEI_UINT32 numberToRead = 0U;                                               /* local value for # of words to read */
   CEI_UINT32 localData;                                                       /* temp storage for received 717 data */
   CEI_UINT32 statusReg;                                                       /* temp storage for 717 status reg    */
   CEI_UINT32 index = 0U;                                                      /* local # of words read counter      */
   CEI_UINT32 autoSyncDisable = 0U;                                            /* current 573 Rx autosync setting    */
   CEI_UINT32 a717rxCfgReg = 0U;                                               /* current 573 Rx config register     */
   CEI_UINT32 progFrameSize = 0U;                                              /* current 573 frame size setting     */
   CEI_UINT32 configOffset;                                                 /* local offset to channel memory area   */
   CEI_UINT32 bufferOffset;                                                 /* local offset to channel memory area   */
   CEI_INT32 delayDuration = 0;                                                /* receive word time-out duration     */
   CEI_INT32 timeOutCount = 0;                                                 /* receive frame time-out counter     */
   CEI_INT32 subframeCount = 0;                                                /* local sub-frame count index        */
   CEI_INT32 nSync = FALSE;                                                    /* local sync-detect flag             */
   CEI_INT32 accessStatus = ARS_NORMAL;                                        /* status of device read/write access */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                       /* function return status            */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr = NULL;                     /* pointer to board local structure  */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr = NULL;                            /* pointer to board global structure */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   else if ((arincData == NULL) || (numberWords == NULL)) {
      status = ARS_INVARG;
   }
   else {
#endif
      deviceGlobalStructPtr = deviceGlobal[board];
      deviceLocalStructPtr = &device[board];

      /* future support: translate the channel using the device channel 
       * allocation instead a the fixed single ARINC 573 channel */
      localChannel = deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX];

      /* read the current ARINC 717 Receiver configuration register */
      configOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE);
      accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[configOffset], &a717rxCfgReg);
      autoSyncDisable = a717rxCfgReg & CEIDEV_573_RX_MODE_RAW;
      
      /* If the Merged Receive Mode is active for this channel, use the Merged Rx FIFO */
      if ((a717rxCfgReg & CEIDEV_MERGE_MODE_ENABLE)) {
         localChannel = deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_MERGED_RX];
      }
#ifndef TARGET_EMBEDDED
      if (accessStatus == ARS_NORMAL) {
         status = a429_utl_validate_channel_request (board, localChannel, ARU_GET_DATA);
      }

      if ((accessStatus == ARS_NORMAL) && (status == ARS_NORMAL)) {
         /* if multi-thread protection is active, acquire a lock on the host interface */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
         }
#endif
         /* setup pointers to the ARINC 717 channel status and data buffer */
         if (accessStatus == ARS_NORMAL) {
            /* If auto-sync is selected, setup the time-out logic to wait for an 
             * entire ARINC 573/717 frame to be received.  In the raw mode, return
             * only what is available in the FIFO.
             */
            if (autoSyncDisable) {
               timeOutLimit = A573_RAW_TIMEOUT;
            }
            else {
               /* extract the current subframe size value from the configuration register and vary delay and count respectively */
               progFrameSize = (autoSyncDisable & CEIDEV_573_FRAME_49152_4096) >> CEIDEV_573_FRAME_SHIFT;
               timeOutLimit  = a717info[progFrameSize].a717RxTimeOutCount;
               delayDuration = a717info[progFrameSize].a717RxDelayDuration;
               progFrameSize = a717info[progFrameSize].a717frameSizeOption;
            }
            
            status = ARS_NODATA;
            numberToRead = *numberWords;
            arincDataPtr = arincData;

            /* setup pointers to the ARINC 717 channel status and data buffer */
            configOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS + (localChannel * CEIDEV_CHANNEL_REGION_SIZE);
            bufferOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_BUFFER_3 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE);
   
            while ((index < numberToRead) && (accessStatus == ARS_NORMAL)) {
      
               if (timeOutCount > timeOutLimit) { /* if we get "no data" too many times */
                  if (!autoSyncDisable) {         /* for raw mode return active status setting */
                     status = ARS_NODATA;         /* for autosync return NO DATA when frame is incomplete */
                  }
                  break;
               }

               /* check for messages present in the buffer, if there are any, read them */
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[configOffset], &statusReg);
               if ((statusReg & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) && (accessStatus == ARS_NORMAL)) {
                  accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[bufferOffset], &localData);
         
                  if (accessStatus == ARS_NORMAL) {
                     if (!autoSyncDisable) {
      
                        /* if h/w indicates this is a subframe-sync word, 
                         * check it against what is programmed by the API */
                        if ( (localData & 0x8000) == 0x8000 ) {
                           if ( (localData & 0x0FFF) == a717syncWords[board][subframeCount] ) {
                              nSync = TRUE;
                              subframeCount++;
                              if (subframeCount == 4) {
                                 subframeCount = 0;
                              }
                              /* if the selected receive buffer contains more than a full frame of data, 
                               * don't wait for the full frame to be received... flush instantly.
                               */
                              rxFifoCount = (statusReg & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT;
                              if (rxFifoCount < (CEI_INT32)(progFrameSize*4)) {
                                 accessStatus = cei_delay_milliseconds(1000U);
                              }
                           }
                           else {   /* subframe sync word doesn't match expected order */
                              nSync = FALSE;
                           }
                        }
      
                        if (nSync) {    /* still in valid sync, log data and continue */
                           *arincDataPtr = (CEI_UINT16)localData;
                           arincDataPtr++;
                           index++;
                           status = ARS_GOTDATA;
                        }
                     }
                     else {    /* autoSyncDisable = AR_OFF, raw mode recording only */
                        *arincDataPtr = (CEI_UINT16)localData;
                        arincDataPtr++;
                        index++;
                        status = ARS_GOTDATA;
                     }
                  }
               }
               else {    /* if the FIFO status indicates no messages present, delay and increment the time-out count */
                  /* if multi-thread protection is active, free any acquired lock */
#ifndef TARGET_EMBEDDED
                  if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
                     accessStatus = cei_release_thread_lock(LOCK_IDENT);
                  }
                  if (accessStatus == ARS_NORMAL) {
#endif
                     ar_sleep(delayDuration);
                     timeOutCount++;
#ifndef TARGET_EMBEDDED
                     /* if multi-thread protection is active, acquire a lock on the host interface */
                     if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
                        accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
                     }
                  }
#endif
               }
            }
         }
#ifndef TARGET_EMBEDDED
         /* if multi-thread protection is active, free any acquired lock */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
         }
      }
   }
#endif

   if ( ((status == ARS_NORMAL) || (status == ARS_NODATA)) && (accessStatus != ARS_NORMAL) ) {
      status = (CDEV_API_RET_TYPE)accessStatus;
   }
   /* return the number of words read */
   *numberWords = index;
   
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                           A R _ G E T _ T R A N S M I T T E R _ M O D E
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the ARINC 573 multi-protocol transmitter mode
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board       -- (input) board index (0 to 127)
 *              CDEV_PARM_SSI_PTR_TYPE mode -- (input) valid values are:
 *                                                     ARU_XMIT_573_HBP
 *                                                     ARU_XMIT_573_BPRZ
 *
 * RETURN VAL:  ARS_NORMAL        success
 *              ARS_INVBOARD      invalid board parameter or board not 
 *                                initialized
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine returns the currently enabled ARINC 573/717 transmitter configuration to the 
 *              application, which applies to a limited subset of CEI-x30 products (most do both simultaneously).
 *
 *              This routine is provided for backward compatibility with the CEI-710 API.  The proper method to 
 *              retrieve the ARINC 573/717 Encoding as BPRZ or HBP is provided for individual channel(s) via the 
 *              ar_get_573_config routine using the item parameter values ARU_573_TX_HBP_SELECT and 
 *              ARU_573_TX_BPRZ_SELECT.
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_get_transmitter_mode 
(
   CDEV_BOARD_TYPE board,                                         /* (input)  board index (0 to 127)                 */
   CDEV_PARM_SSI_PTR_TYPE mode                                    /* (output) transmitter mode (717-HBP or 717-BPRZ) */
)
{
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                               /* function return status    */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      *mode = (CDEV_PARM_SSI_TYPE)deviceGlobal[board]->transmitMode;
#ifndef TARGET_EMBEDDED
   }
   else {
      status = ARS_INVBOARD;
   }
#endif
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                        A R _ S E T _ T R A N S M I T T E R _ M O D E
 *===================================================================================================================*
 *
 * FUNCTION:    Set the ARINC 717 transmitter mode
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board    -- (input) board index (0 to 127)
 *              CDEV_PARM_SSI_TYPE mode  -- (input) valid values are:
 *                                                  ARU_XMIT_573_HBP 
 *                                                  ARU_XMIT_573_BPRZ
 *
 * RETURN VAL:  ARS_NORMAL       Success
 *              ARS_INVHARVAL    "mode" parameter value is not supported or selected protocol is not available
 *              ARS_INVBOARD     Invalid board parameter or board not initialized
 *              ARS_BOARD_MUTEX  Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: This routine enables the respective transmitter encoding for the ARINC 573/717 transmit channel. 
 *
 *              This routine is provided for backward compatibility with the CEI-710 API.  The proper method to 
 *              assign ARINC 573 Encoding as BPRZ or HBP is provided for individual channel(s) via ar_set_573_config
 *              using the item parameter values ARU_573_TX_HBP_SELECT and ARU_573_TX_BPRZ_SELECT.
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_set_transmitter_mode 
(
   CDEV_BOARD_TYPE board,                                     /* (input)  board index (0 to 127)                     */
   CDEV_PARM_SSI_TYPE mode                                    /* (input)  transmitter mode (CSDB 717-HBP or 717-BPRZ */
) 
{
   CEI_UINT32  bitFieldValue;                                                   /* temp var for device register data */
   CEI_UINT32  memOffset;                                                    /* local offset to channel memory area  */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                       /* function return status            */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr = NULL;                     /* pointer to board local structure  */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr = NULL;                            /* pointer to board global structure */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      if ((mode == ARU_XMIT_573_HBP) || (mode == ARU_XMIT_573_BPRZ)) {
#endif
         deviceLocalStructPtr = &device[board];
         deviceGlobalStructPtr = deviceGlobal[board];
         memOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX] * CEIDEV_CHANNEL_REGION_SIZE);
#ifndef TARGET_EMBEDDED
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            status = (CDEV_API_RET_TYPE)cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (status == ARS_NORMAL) {
#endif
            status = (CDEV_API_RET_TYPE)cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &bitFieldValue);
      
            if (status == ARS_NORMAL) {
               if (mode == ARU_XMIT_573_HBP) {
                  bitFieldValue &= CEIDEV_573_TX_BPRZ_MASK;
                  bitFieldValue |= CEIDEV_573_TX_HBP_SELECT + CEIDEV_CHANNEL_ENABLE;
               }
               else if (mode == ARU_XMIT_573_BPRZ) {
                  bitFieldValue &= CEIDEV_573_TX_HBP_MASK;
                  bitFieldValue |= CEIDEV_573_TX_BPRZ_SELECT + CEIDEV_CHANNEL_ENABLE;
               }
               status = (CDEV_API_RET_TYPE)cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], bitFieldValue);
               deviceGlobalStructPtr->transmitMode = mode;
            }
#ifndef TARGET_EMBEDDED
            if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
               status = (CDEV_API_RET_TYPE)cei_release_thread_lock(LOCK_IDENT);
            }
         }
      }
      else {
         status = ARS_INVHARVAL;
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                               A R _ P U T _ 5 7 3 _ F R A M E
 *===================================================================================================================*
 *
 * FUNCTION:    Transmit ARINC 573/717 frame data on the ARINC 573/717 transmit channel. 
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board     -- (input) Board number of interest (0-127)
 *              CEI_UINT32  numberWords   -- (input) no. of ARINC 573/717 words to send NOTE: must be an even number
 *              pCEI_UINT32 transmitCount -- (input) number of words transmitted, a return value of 0 indicates 
 *                                                   a failure to transmit the supplied frame data
 *              pCEI_UINT16 arincData     -- (input) Pointer to the frame data
 * 
 * RETURN VAL:  ARS_NORMAL        success
 *              ARS_INVBOARD      invalid board parameter or  
 *                                board not initialized
 *              ARS_INVHARVAL     invalid channel parameter
 *                              
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: This routine will insert the ARINC 573/717 frame data into the respective transmit buffer.  Once the 
 *              buffer is full or the specified number of elements have been copied to the buffer, the number of 
 *              elements "transmitted" will be returned.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_put_573_frame 
( 
   CDEV_BOARD_TYPE board,                                                /* (input) Board number of interest (0-127) */
   CEI_UINT32  numberWords,                                              /* (input) no. of ARINC 573 words to send   */
   pCEI_UINT32 transmitCount,                                            /* (input) number of words transmitted      */
   pCEI_UINT16 arincData                                                 /* (input) Pointer to the frame data        */
) 
{
   CEI_UINT32 localChannel = 0U;                                                /* merged/individual channel select  */
   CEI_UINT32 channelStatus;                                                    /* Rx channel FIFO data avail status */
   CEI_UINT32 index = 0U;                                                       /* local # of words written counter  */
   CEI_UINT32 messageCount;                                                     /* local number of words to transmit */
   CEI_UINT16 transmitData;                                                     /* local temp transmit data storage  */
   pCEI_UINT16 localDataPtr;                                                    /* local ptr to applic. data array   */
   CEI_UINT32 channelOffset;                                                 /* local offset to channel memory area  */
   CEI_UINT32 bufferOffset;                                                   /* local offset to buffer memory area  */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                               /* function return status    */
   CEI_INT32 accessStatus = ARS_NORMAL;                                        /* status of device read/write access */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr = NULL;                          /* ptr to board local structure */

   *transmitCount = 0;

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      /* future support: translate the channel using the device channel 
       * allocation instead a the fixed single ARINC 573 channel
       */
      localChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX];
      
#ifndef TARGET_EMBEDDED
      status = a429_utl_validate_channel_request (board, localChannel, ARU_PUT_DATA);
   }
   else {
      status = ARS_INVBOARD;
   }

   if (status == ARS_NORMAL) {
#endif
      deviceLocalStructPtr = &device[board];
      
      /* Message count is half the requested word count (2 words/message) */
      localDataPtr = arincData;

      channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS + (localChannel * CEIDEV_CHANNEL_REGION_SIZE);
      bufferOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_BUFFER_3 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE);

#ifndef TARGET_EMBEDDED
      /* if multi-thread protection requested, acquire semaphore/mutex-locking */
      if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
         accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
      }
      if (accessStatus == ARS_NORMAL) {
#endif

         accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &channelStatus);
         if (accessStatus == ARS_NORMAL) {
            if (channelStatus & CEIDEV_CSR_XMIT_FIFO_FULL) {
               status = ARS_XMITOVRFLO;
            }
            else {
               /* limit the number of messages allowed to transmit based on available space in the buffer */
               channelStatus = (channelStatus & CEIDEV_CSR_TX_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT;
               messageCount = numberWords <= (CEIDEV_CIRCULAR_BUFFER_SIZE - channelStatus) ? numberWords : CEIDEV_CIRCULAR_BUFFER_SIZE - channelStatus; 

               while ((index < messageCount) && (accessStatus == ARS_NORMAL)) {
                  transmitData = *localDataPtr;
                  accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[bufferOffset], (CEI_INT32)transmitData);
                  localDataPtr++;
                  index++;
               }
            }
         }
#ifndef TARGET_EMBEDDED
         /* if multi-thread protection is active, free any acquired lock */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
         }
      }
   }
#endif
   if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL)) {
      status = (CDEV_API_RET_TYPE)accessStatus;
   }
   *transmitCount = index;

   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                               A R _ S E T _ 5 7 3 _ C O N F I G
 *===================================================================================================================*
 *
 * FUNCTION:    Set ARINC 573/717 channel configuration register attributes
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input) board to reference (0 to 127)
 *              CDEV_PARM_SSI_TYPE item -- (input) config register attribute to set
 *              CDEV_PARM_SI_TYPE value -- (input) value applied to item attribute
 * 
 *                  item                 valid values
 * 
 *              ARU_RECV_MODE            AR_WRAP_OFF, AR_WRAP_ON
 *              ARU_573_RX_AUTO_DETECT   AR_OFF, AR_ON
 *              ARU_573_RX_BPRZ_SELECT   AR_OFF, AR_ON
 *              ARU_573_TX_BPRZ_SELECT   AR_OFF, AR_ON
 *              ARU_573_TX_HBP_SELECT    AR_OFF, AR_ON
 *
 *              ARU_RX_BITRATE &         ARU_573_RATE_SIZE_384_32    
 *              ARU_TX_BITRATE           ARU_573_RATE_SIZE_768_64    
 *                                       ARU_573_RATE_SIZE_1536_128  
 *                                       ARU_573_RATE_SIZE_3072_256  
 *                                       ARU_573_RATE_SIZE_6144_512  
 *                                       ARU_573_RATE_SIZE_12288_1024
 *                                       ARU_573_RATE_SIZE_24576_2048
 *                                       ARU_573_RATE_SIZE_49152_4096
 * 
 *              ARU_573_TX_SLEW_RATE     ARU_573_TX_SLEW_1PT5
 *                                       ARU_573_TX_SLEW_10PT0
 * 
 *              ARU_TX_DISABLE           AR_OFF, AR_ON
 *              ARU_TX_FIFO_ENABLE       AR_OFF, AR_ON
 *              ARU_RX_FIFO_ENABLE       AR_OFF, AR_ON
 *              ARU_573_RX_ENABLE        AR_OFF, AR_ON
 * 
 *              ARU_573_SYNC_WORD1       value from 0x0000 to 0x0fff
 *              ARU_573_SYNC_WORD2       value from 0x0000 to 0x0fff
 *              ARU_573_SYNC_WORD3       value from 0x0000 to 0x0fff
 *              ARU_573_SYNC_WORD4       value from 0x0000 to 0x0fff
 *              ARU_RX_MERGED_MODE       AR_ON, AR_OFF
 * 
 *              (Following supported for legacy CEI-710 API only)
 *              ARU_RX_BITRATE &         ARU_573_SPEED_384  
 *              ARU_TX_BITRATE           ARU_573_SPEED_768  
 *                                       ARU_573_SPEED_1536 
 *                                       ARU_573_SPEED_3072 
 *                                       ARU_573_SPEED_6144 
 *                                       ARU_573_SPEED_12288
 *                                       ARU_573_SPEED_24576
 * 
 *              ARU_573_RX_SUBFRAME_SIZE ARU_573_SUB_FRAME_64  
 *                                       ARU_573_SUB_FRAME_128 
 *                                       ARU_573_SUB_FRAME_256 
 *                                       ARU_573_SUB_FRAME_512 
 *                                       ARU_573_SUB_FRAME_1024
 *                                       ARU_573_SUB_FRAME_2048
 *                                       ARU_573_SUB_FRAME_4096
 * 
 * RETURN VAL:  ARS_NORMAL      Normal succesful completion.
 *              ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *              ARS_INVARG      "item" parameter value is not supported by this function or device configuration. 
 *              ARS_INVHARVAL   "value" parameter value is not supported for "item"
 *              ARS_INVBOARD    board parameter value invalid, board not initialized, or the ARINC 573/717 protocol 
 *                              is not supported on this device
 * 
 * GLOBALS:     device[]
 *              deviceGlobal[]
 *
 * DESCRIPTION: This function assigns the respective device configuration register attributes based on the combined 
 *              item/value parameter selections.  It is designed to support the ARINC 573/717 protocol configuration 
 *              register attributes only.
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_set_573_config 
( 
   CDEV_BOARD_TYPE board,                                                  /* (input)  board to reference (0 to 127) */
   CDEV_PARM_SSI_TYPE item,                                                /* (input)  item to process               */
   CDEV_PARM_SI_TYPE value                                                 /* (input)  state of the item             */
)
{
   CEI_UINT32 deviceChannel = 0U;                                           /* current channel to select             */
   CEI_UINT32 bitFieldValue;                                                /* temp for device register read content */
   CEI_INT32 accessStatus = ARS_NORMAL;                                        /* status of device read/write access */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                       /* function return status            */
#ifndef TARGET_EMBEDDED                                                                                              
   CEI_INT32 tempStatus = ARS_NORMAL;                                                  /* local temp return status   */
#endif                                                                                                               
   CEI_UINT32 offsetIntoChannelWindow = 0U;                                        /* offset into chan mem region    */
   CEI_UINT32 memOffset;                                                    /* local offset to channel memory area   */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr = NULL;                     /* pointer to board local structure  */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr = NULL;                            /* pointer to board global structure */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   else {
#endif
      /* assign references to this board's device and deviceGlobal structure-array elements */
      deviceLocalStructPtr = &device[board];
      deviceGlobalStructPtr = deviceGlobal[board];
      deviceChannel = 0U;  /* here for future growth w/channel parm */
   
#ifndef TARGET_EMBEDDED
      /* if multi-thread protection is active, acquire a lock on the host interface */
      if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
         status = (CDEV_API_RET_TYPE)cei_acquire_thread_lock(LOCK_IDENT);
      }

      if (status == ARS_NORMAL) {
#endif
         /* Assign the h/w-respective channel number and type, then validate 
          * the channel referenced for the specific config item parameter value */
         switch (item) {
            case ARU_RECV_MODE:
            case ARU_RX_BITRATE:
            case ARU_RX_DISABLE:
            case ARU_RX_FIFO_ENABLE:
            case ARU_RX_MERGED_MODE:
            case ARU_573_RX_ENABLE:
            case ARU_573_RX_AUTO_DETECT:
            case ARU_573_RX_BPRZ_SELECT:
            case ARU_573_RX_SUBFRAME_SIZE:
               deviceChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX];
               offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
               break;
         
            case ARU_TX_BITRATE:
            case ARU_TX_DISABLE:
            case ARU_TX_FIFO_ENABLE:
            case ARU_573_TX_SLEW_RATE:
            case ARU_573_TX_BPRZ_SELECT:
            case ARU_573_TX_HBP_SELECT:
               deviceChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX];
               offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
               break;
      
            case ARU_573_SYNC_WORD1:
            case ARU_573_SYNC_WORD2:
               offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_2;
               deviceChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX];
               break;
   
            case ARU_573_SYNC_WORD3:
            case ARU_573_SYNC_WORD4:
               offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_3;
               deviceChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX];
               break;
   
            case 0x7996: 
               #if defined DEBUG_PROGRAMMING
                  #include "dbg_mp.h"
               #else
               status = ARS_INVARG; 
               #endif
               break;

            case 0x7997: 
               #if defined DEBUG_PROGRAMMING
                  #include "dbg_rst.h"
               #else
               status = ARS_INVARG; 
               #endif
               break;

            case 0x7998: 
               #if defined DEBUG_PROGRAMMING
                  #include "dbg_xmc.h"
               #else
               status = ARS_INVARG; 
               #endif
               break;

            case 0x7999: 
            default:
               #if defined DEBUG_PROGRAMMING
                  #include "dbg_pcie.h"
               #else
               status = ARS_INVARG; 
               #endif
               break;
         }
         
#ifndef TARGET_EMBEDDED
         /* validate the specified channel */
         if (status == ARS_NORMAL) {
            status = a429_utl_validate_channel_request (board, deviceChannel, item);
         }
      
         if (status == ARS_NORMAL) {
#endif
            /* for all control register attribute assignments, first read the 
             * device register with respect to the attribute specified 
             */
            memOffset = CEIDEV_CHANNEL_REGION_START + offsetIntoChannelWindow + (deviceChannel * CEIDEV_CHANNEL_REGION_SIZE);
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &bitFieldValue);
            
            if (accessStatus == ARS_NORMAL) {
               switch (item) {
                  case ARU_RECV_MODE:
                  case ARU_RX_BITRATE:
                  case ARU_RX_MERGED_MODE:
                  case ARU_573_RX_SUBFRAME_SIZE:
                  case ARU_573_RX_AUTO_DETECT:
                  case ARU_573_RX_BPRZ_SELECT:
                     /* When assigning Receive Configuration Register attributes, force
                      * a Rx channel FIFO disable prior to modification of attribute, 
                      * then restore the FIFO enable state afterwards
                      */
                     accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], (bitFieldValue & CEIDEV_CHANNEL_ENABLE_MASK));
                     
                     if (item == ARU_RECV_MODE) {
                        if (value == AR_WRAP_ON) {
                           bitFieldValue |= CEIDEV_INTERNAL_WRAP_ENABLE;
                        }
                        else if (value == AR_WRAP_OFF) {
                           bitFieldValue &= CEIDEV_INTERNAL_WRAP_MASK;
                        }
                        else {
                           status = ARS_INVHARVAL;
                        }
                     }
                     else if (item == ARU_RX_BITRATE) {
                        bitFieldValue &= CEIDEV_573_FRAME_SPEC_MASK;
                        /* assign the ARINC 717 h/w bus speed selection,  
                         * supporting legacy CEI-7x0 API options for bus speed */
                        switch (value) {
                           case ARU_573_RATE_SIZE_384_32:
                           case ARU_573_SPEED_384:
                              bitFieldValue |= CEIDEV_573_FRAME_384_32;
                              break;
                           case ARU_573_RATE_SIZE_768_64:
                           case ARU_573_SPEED_768:
                              bitFieldValue |= CEIDEV_573_FRAME_768_64;
                              break;
                           case ARU_573_RATE_SIZE_1536_128:
                           case ARU_573_SPEED_1536:
                              bitFieldValue |= CEIDEV_573_FRAME_1536_128;
                              break;
                           case ARU_573_RATE_SIZE_3072_256:
                           case ARU_573_SPEED_3072:
                              bitFieldValue |= CEIDEV_573_FRAME_3072_256;
                              break;
                           case ARU_573_RATE_SIZE_6144_512:
                           case ARU_573_SPEED_6144:
                              bitFieldValue |= CEIDEV_573_FRAME_6144_512;
                              break;
                           case ARU_573_RATE_SIZE_12288_1024:
                           case ARU_573_SPEED_12288:
                              bitFieldValue |= CEIDEV_573_FRAME_12288_1024;
                              break;
                           case ARU_573_RATE_SIZE_24576_2048:
                           case ARU_573_SPEED_24576:
                              bitFieldValue |= CEIDEV_573_FRAME_24576_2048;
                              break;
                           case ARU_573_RATE_SIZE_49152_4096:
                              bitFieldValue |= CEIDEV_573_FRAME_49152_4096;
                              break;
                           default:
                              status = ARS_INVHARVAL;
                        }
                     }
                     /* The following values are supported for legacy CEI-710 API only */
                     else if (item == ARU_573_RX_SUBFRAME_SIZE) {
                        bitFieldValue &= CEIDEV_573_FRAME_SPEC_MASK;
                        switch (value) {
                           case ARU_573_SUB_FRAME_64:
                              bitFieldValue |= CEIDEV_573_FRAME_768_64;
                              break;
                           case ARU_573_SUB_FRAME_128:
                              bitFieldValue |= CEIDEV_573_FRAME_1536_128;
                              break;
                           case ARU_573_SUB_FRAME_256:
                              bitFieldValue |= CEIDEV_573_FRAME_3072_256;
                              break;
                           case ARU_573_SUB_FRAME_512:
                              bitFieldValue |= CEIDEV_573_FRAME_6144_512;
                              break;
                           case ARU_573_SUB_FRAME_1024:
                              bitFieldValue |= CEIDEV_573_FRAME_12288_1024;
                              break;
                           case ARU_573_SUB_FRAME_2048:
                              bitFieldValue |= CEIDEV_573_FRAME_24576_2048;
                              break;
                           default:
                              status = ARS_INVHARVAL;
                        }
                     }
                     else if (item == ARU_573_RX_AUTO_DETECT) {
                        if (value == AR_OFF) {
                           bitFieldValue |= CEIDEV_573_RX_MODE_RAW;
                        }
                        else if (value == AR_ON) {
                           bitFieldValue &= CEIDEV_573_RX_MODE_AUTO_MASK;
                        }
                        else {
                           status = ARS_INVHARVAL;
                        }
                     }
                     else if (item == ARU_573_RX_BPRZ_SELECT) {
                        if (value == AR_ON) {
                           bitFieldValue |= CEIDEV_573_RX_BPRZ_SELECT;
                        }
                        else if (value == AR_OFF) {
                           bitFieldValue &= CEIDEV_573_RX_ENCODING_MASK;
                        }
                        else {
                           status = ARS_INVHARVAL;
                        }
                     } 
                     else if (item == ARU_RX_MERGED_MODE) {
                        if (value == AR_ON) {
                           bitFieldValue |= CEIDEV_MERGE_MODE_ENABLE;
                        }
                        else if (value == AR_OFF) {
                           bitFieldValue &= CEIDEV_MERGE_MODE_ENABLE_MASK;
                        }
                        else {
                           status = ARS_INVHARVAL;
                        }
                     }
                     break;
         
                  case ARU_RX_DISABLE:
                     if (value == AR_ON) {
                        deviceGlobalStructPtr->receiveChannelBufferMode[deviceChannel] = ARU_MERGED;
                        bitFieldValue &= CEIDEV_CHANNEL_ENABLE_MASK;
                     }
                     else if (value == AR_OFF) {
                        deviceGlobalStructPtr->receiveChannelBufferMode[deviceChannel] = ARU_BUFFERED;
                        bitFieldValue |= CEIDEV_CHANNEL_ENABLE;
                     }
                     else {
                        status = ARS_INVHARVAL;
                     }
                     break;
               
                  case ARU_TX_BITRATE:
                  case ARU_573_TX_BPRZ_SELECT:
                  case ARU_573_TX_HBP_SELECT:
                  case ARU_573_TX_SLEW_RATE:
                     /* When assigning Transmit Configuration Register attributes, force
                      * a Tx channel FIFO disable prior to modification of attribute, 
                      * then restore the FIFO enable state afterwards
                      */
                     accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], (bitFieldValue & CEIDEV_CHANNEL_ENABLE_MASK));
                     if (accessStatus != ARS_NORMAL) {
                        break;
                     }
                     
                     if (item == ARU_TX_BITRATE) {
                        bitFieldValue &= CEIDEV_573_FRAME_SPEC_MASK;
                        /* assign the ARINC 717 h/w bus speed selection,
                         * supporting legacy CEI-7x0 API options for bus speed */
                        switch (value) {
                           case ARU_573_RATE_SIZE_384_32:
                           case ARU_573_SPEED_384:
                              bitFieldValue |= CEIDEV_573_FRAME_384_32;
                              break;
                           case ARU_573_RATE_SIZE_768_64:
                           case ARU_573_SPEED_768:
                              bitFieldValue |= CEIDEV_573_FRAME_768_64;
                              break;
                           case ARU_573_RATE_SIZE_1536_128:
                           case ARU_573_SPEED_1536:
                              bitFieldValue |= CEIDEV_573_FRAME_1536_128;
                              break;
                           case ARU_573_RATE_SIZE_3072_256:
                           case ARU_573_SPEED_3072:
                              bitFieldValue |= CEIDEV_573_FRAME_3072_256;
                              break;
                           case ARU_573_RATE_SIZE_6144_512:
                           case ARU_573_SPEED_6144:
                              bitFieldValue |= CEIDEV_573_FRAME_6144_512;
                              break;
                           case ARU_573_RATE_SIZE_12288_1024:
                           case ARU_573_SPEED_12288:
                              bitFieldValue |= CEIDEV_573_FRAME_12288_1024;
                              break;
                           case ARU_573_RATE_SIZE_24576_2048:
                           case ARU_573_SPEED_24576:
                              bitFieldValue |= CEIDEV_573_FRAME_24576_2048;
                              break;
                           case ARU_573_RATE_SIZE_49152_4096:
                              bitFieldValue |= CEIDEV_573_FRAME_49152_4096;
                              break;
                           default:
                              status = ARS_INVHARVAL;
                        }
                     }
                     else if (item == ARU_573_TX_SLEW_RATE) {
      
                        if (value == ARU_573_TX_SLEW_1PT5) {
                           bitFieldValue |= CEIDEV_FAST_SLEW_RATE;
                        }
                        else if (value == ARU_573_TX_SLEW_10PT0) {
                           bitFieldValue &= CEIDEV_SLEW_RATE_MASK;
                        }
                        else {
                           status = ARS_INVHARVAL;
                        }
                     }
                     else if (item == ARU_573_TX_HBP_SELECT) {
                        if (value == AR_ON) {
                           deviceGlobalStructPtr->transmitMode = ARU_XMIT_573_HBP;
                           bitFieldValue |= CEIDEV_573_TX_HBP_SELECT;
                        }
                        else if (value == AR_OFF) {
                           bitFieldValue &= CEIDEV_573_TX_HBP_MASK;
                        }
                        else {
                           status = ARS_INVHARVAL;
                        }
                     }
                     else if (item == ARU_573_TX_BPRZ_SELECT) {
                        if (value == AR_ON) {
                           deviceGlobalStructPtr->transmitMode = ARU_XMIT_573_BPRZ;
                           bitFieldValue |= CEIDEV_573_TX_BPRZ_SELECT;
                        }
                        else if (value == AR_OFF) {
                           bitFieldValue &= CEIDEV_573_TX_BPRZ_MASK;
                        }
                        else {
                           status = ARS_INVHARVAL;
                        }
                     }
                     break;
            
                  case ARU_TX_DISABLE:
                     if (value == AR_OFF) {
                        bitFieldValue &= CEIDEV_TRANSMIT_DISABLE_MASK;
                     }
                     else if (value == AR_ON) {
                        bitFieldValue |= CEIDEV_TRANSMIT_DISABLE;
                     }
                     else {
                        status = ARS_INVHARVAL;
                     }
                     break;
               
                  case ARU_573_RX_ENABLE:
                  case ARU_RX_FIFO_ENABLE:
                  case ARU_TX_FIFO_ENABLE:
                     if (value == AR_OFF) {
                        bitFieldValue &= CEIDEV_CHANNEL_ENABLE_MASK;
                     }
                     else if (value == AR_ON) {
                        bitFieldValue |= CEIDEV_CHANNEL_ENABLE;
                     }
                     else {
                        status = ARS_INVHARVAL;
                     }
                     break;
            
                  case ARU_573_SYNC_WORD1:
                  case ARU_573_SYNC_WORD2:
                  case ARU_573_SYNC_WORD3:
                  case ARU_573_SYNC_WORD4:
                     a717syncWords[board][item-ARU_573_SYNC_WORD1] = (CEI_UINT32)value;
         
                     if ((item == ARU_573_SYNC_WORD2) || (item == ARU_573_SYNC_WORD4)) {
                        bitFieldValue &= CEIDEV_573_DATA_MASK;
                        bitFieldValue += (CEI_UINT32)(value & CEIDEV_573_DATA_MASK) << CEIDEV_573_SYNCH_SHIFT;
                     }
                     else {
                        bitFieldValue &= CEIDEV_573_DATA_MASK << CEIDEV_573_SYNCH_SHIFT;
                        bitFieldValue += (CEI_UINT32)(value & CEIDEV_573_DATA_MASK);
                     }
                     break;
         
                  default: 
                     status = ARS_INVHARVAL; 
                     break;
               }
            }
   
            /* update the configuration register with the specified attribute setting */
            if (accessStatus == ARS_NORMAL) {
               accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], bitFieldValue);
            }
            
            if (status == EXIT_ROUTINE) {
               status = ARS_NORMAL;
            }
         
            if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL)) {
               status = (CDEV_API_RET_TYPE)accessStatus;
            }
#ifndef TARGET_EMBEDDED
         }
   
         /* if multi-thread protection is active, free any acquired lock */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            if (status == ARS_NORMAL) {
               status = (CDEV_API_RET_TYPE)cei_release_thread_lock(LOCK_IDENT);
            }
            else {
               tempStatus = cei_release_thread_lock(LOCK_IDENT);
            }
         }
      }
   }
   if (tempStatus) ;
#endif
   return status;
}

