/*====================================================================================================================*
 * FILE:                                 C D E V _ A P I _ C F G _ F I L E. C                    
 *====================================================================================================================*
 *
 *   COPYRIGHT 2016 - 2019 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED
 *   ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH THE INCLUSION
 *   OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY OTHER COPIES THEREOF
 *   MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE
 *   AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS, INC. 
 *
 *====================================================================================================================*
 *
 * FUNCTION:    Configuration File Input Interface for a CEI-x30 device.
 *
 * DESCRIPTION: These routines are provided to demonstrate the configuration
 *              text file usage for channel and schedule message definition
 *              for applications interfacing with CEI-x30 ARINC 429 devices.
 *              
 * EXTERNALLY DEFINED ROUTINE ENTRY POINTS:         
 *
 *              ar_config_channels_from_txt_file
 *              ar_define_messages_from_txt_file
 *              ar_channel_configuration_from_xml_file
 *              ar_define_messages_from_xml_file
 *
 *--------------------------------------------------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date    By   Ver. Comments
 * -------- ----- ---- -----------------------------------------------------------------------------------------------*
 * 02/22/12  BDW       Created.
 * 10/18/16  BDW  4.90 Introduced use of CEIx30 API-specific type defines via the file CEIX30_TYPES.H allowing
 *                     for easy build to common use of only signed integer return types and unsigned integer
 *                     parameter types (with the exception of 64-bit types).
 *
 *====================================================================================================================*/ 

/*---------------------------------------------------------------------------*
 *                      GENERAL INCLUDE STATEMEMTS
 *---------------------------------------------------------------------------*/
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
   #include <ctype.h>
   #include <unistd.h>
   #include <fcntl.h>
   #include <string.h>
   #include <errno.h>
   #include <sys/stat.h> 
   #include <sys/syscall.h>
#elif defined (VXW_PCI_X86) || defined (VXW_PCI_PPC)
   #include <stdio.h>
   #include <stdlib.h>
   #include <unistd.h>
#elif defined (NIVISA_PXI)
   #ifdef LABVIEW_RT
      #include <ansi_c.h> 
   #endif
   #ifdef CVI
      #include <ansi_c.h> 
   #endif
   #include <visa.h>  
#elif defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
   #include <string.h>
#else
   #include <stdio.h>
   #include <string.h>
   #include <stdlib.h>
   #include <windows.h>
#endif

/* include local header files, CDEV_API.H included in CDEV_GLB.H */
#include "cdev_glb.h"

#ifdef INCLUDE_MSG_SCHEDULER
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_define_msg_block (CEI_INT32 numberOfEntries, pAR_SCHEDULED_MSG_ENTRY_TYPE messageEntry);
#endif


/*---------------------------------------------------------------------------*
 *                      DEFINES, DATA TYPES, AND GLOBALS
 *---------------------------------------------------------------------------*/

/* Configuration file string-tags for channel attributes */
#define BOARD_TAG          "Board Index:"
#define RX_CH_COUNT_TAG    "RxChannel Number:"
#define TX_CH_COUNT_TAG    "TxChannel Number:"
#define RX_CH_START_TAG    "<RxChannel"
#define RX_CH_END_TAG      "RxChannel>"
#define TX_CH_START_TAG    "<TxChannel"
#define TX_CH_END_TAG      "TxChannel>"
#define CH_INDEX_TAG       "ChannelIndex"
#define CH_ENABLE_TAG      "Enable"
#define CH_BITRATE_TAG     "BitRate"
#define CH_PARITY_TAG      "ParityCheck"
#define CH_BUF_MODE_TAG    "BufferMode"
#define CH_WRAP_TAG        "InternalWrap"
#define CH_PROT_ERR_TAG    "ProtocolError"

/* Configuration file options strings for channel attributes */
#define OPT_ENABLED        "ENABLED"
#define OPT_DISABLED       "DISABLED"
#define OPT_INDIVIDUAL     "INDIVIDUAL"
#define OPT_MERGED         "MERGED"
#define OPT_NONE           "NONE"
#define OPT_ODD            "ODD"
#define OPT_EVEN           "EVEN"
#define OPT_BIT_COUNT_HIGH "BIT COUNT HIGH"
#define OPT_BIT_COUNT_LOW  "BIT COUNT LOW"
#define OPT_SHORT_GAP      "SHORT GAP"


/* Configuration file string-tags for scheduled message attributes */
#define MSG_COUNT_TAG      "MessageNumber:"
#define MSG_INDEX_TAG      "MessageIndex"
#define MSG_GRP_START_TAG  "<MessageGroup"
#define MSG_GRP_END_TAG    "MessageGroup>"
#define BOARD_INDEX_TAG    "BoardIndex"
#define MSG_VALUE_TAG      "MessageContext"
#define MSG_RATE_TAG       "Rate"
#define MSG_OFFSET_TAG     "StartOffset"
#define MSG_TXCOUNT_TAG    "TransmitCount"
#define MSG_EQ_ID          "quipmentID"


/* Generic Text File Field Delimiters */
#define TOKEN_VALUE_FIELD_DELIMITER    ":"
#define ATTRIBUTE_ENTRY_DELIMITER      "_"
#define ATTRIBUTE_ENTRY_DELIMITER_CHAR '_'
#define EQUIP_ID_FIELD_START_CHAR      "E"
#define TRANSMIT_CONTINUOUS_START_CHAR '-'

#define LEFT_BRACKET_DELIMITER  '<'
#define RIGHT_BRACKET_DELIMITER '>'
#define LEFT_BRACKET  10
#define RIGHT_BRACKET 11
#define NO_BRACKET    12
#define True  20
#define False 21

#define RX    30
#define TX    31

/* Channel attribute structures for local storage of configuration file settings */
typedef struct RX_CHANNEL_PARMS {
   CDEV_CHAN_TYPE channel;
   CDEV_PARM_SSI_TYPE enable;
   CDEV_PARM_SSI_TYPE bitRate;
   CDEV_PARM_SSI_TYPE parity;
   CDEV_PARM_SSI_TYPE bufferMode;
   CDEV_PARM_SSI_TYPE internalWrap;
} RX_CHANNEL_PARMS_TYPE;

typedef struct TX_CHANNEL_PARMS {
   CDEV_CHAN_TYPE channel;
   CDEV_PARM_SSI_TYPE enable;
   CDEV_PARM_SSI_TYPE bitRate;
   CDEV_PARM_SSI_TYPE parity;
   CDEV_PARM_SSI_TYPE bitError;
   CDEV_PARM_SSI_TYPE gapError;
} TX_CHANNEL_PARMS_TYPE;

/*XML item struct*/
typedef struct XMLSEC
{
   CEI_CHAR  accTemp[128];
   CDEV_PARM_SSI_TYPE cintLength;
   CEI_VOID  *pNext;
} XMLSEC;

typedef enum {BOARD, COUNT, RXATT, TXATT, QUERY, MSGDEF, CFG_ERR, DONE} CONFIG_FILE_SECTION_TYPE;

AR_SCHEDULED_MSG_ENTRY_TYPE scheduledMessages[1000];

/*Channel configuration XML items*/
typedef enum {CCST = 0,CCED, 
              BRDIDXST,BRDIDXED,BRDIDXVAL,RXNUMST,RXNUMED,RXNUMVAL,TXNUMST,TXNUMED,TXNUMVAL,               /*10*/
              RXGRPST,RXGRPED,TXGRPST,TXGRPED,                                                             /*14*/
              RXCHST,RXCHED,TXCHST,TXCHED,                                                                 /*18*/
              CHIDXST,CHIDXED,CHIDXVAL,BRST,BRED,BRVAL,ENST,ENED,ENVAL,RXPST,RXPED,RXPVAL,BMST,BMED,BMVAL, /*33*/
              IWST,IWED,IWVAL,TXPST,TXPED,TXPVAL,PEST,PEED,PEVAL,XML_CC_ITEM_END} CC_XML_ITEM;             /*43*/
 
/*Schedule message XML items*/
typedef enum {SMST = 0, SMED,
              MSGNUMST,MSGNUMED,MSGNUMVAL,                                                          /*4*/
              MSGGRPST,MSGGRPED,                                                                    /*6*/
              MSGST,MSGED,                                                                          /*8*/
              SBRDIDXST,SBRDIDXED,SBRDIDXVAL,SCHIDXST,SCHIDXED,SCHIDXVAL,MSGCTST,MSGCTED,MSGCTVAL,  /*17*/
              RTST,RTED,RTVAL,STOFFST,STOFFED,STOFFVAL,TRCNST,TRCNED,TRCNVAL,ETIDST,ETIDED,ETIDVAL, /*29*/
              XML_SM_ITEM_END                                                                       /*30*/
             }SM_XML_ITEM;

/*Channel configuration XML item switch, first choice*/
CDEV_PARM_SSI_TYPE XML_CC_ITEM_REL_TAB[XML_CC_ITEM_END] = 
{
   BRDIDXST,          /*After channel configuration element start, it should be board index element start*/
   XML_CC_ITEM_END,   /*Nothing after channel configuration element end*/
   BRDIDXVAL,         /*After board index element start, it should be board index element value*/
   RXNUMST,           /*After board index element end,it should be rx channel num element start*/
   BRDIDXED,          /*After board index element value,it should be board index element end*/
   RXNUMVAL,          /*After rx channel number element start, it should be rx channel number element value*/
   TXNUMST,           /*After rx channle number element end, it should be tx channel number element start*/
   RXNUMED,           /*After rx channel number element value, it should be rx channel number element end*/
   TXNUMVAL,          /*After tx channel number element start, it should be tx channel number element value*/
   RXGRPST,           /*After tx channel number element end, it should be rx channel group start*/
   TXNUMED,           /*After tx channel number element value, it should be tx channel number element end*/
   RXCHST,            /*After rx channel group element start, it should be rx channel element start*/
   TXGRPST,           /*After rx channel group element end, it should be tx channel group element start*/
   TXCHST,            /*After tx channel group element start, it should be tx channel element start*/
   CCED,              /*After tx channel group element end, it should be channel configuration element start*/
   CHIDXST,           /*After rx channel element start, it should be channel index element start*/
   RXCHST,            /*After rx channel element end, it should be rx channel element start*/
   CHIDXST,           /*After tx channel element start, it should be channel index element start*/
   TXCHST,            /*After tx channel element end, it should be tx channel element start*/
   CHIDXVAL,          /*After channel index element start, it should be channel index element value*/
   BRST,              /*After channel index element end, it should be bit rate element start*/
   CHIDXED,           /*After channel index element value, it should be channel index element end*/
   BRVAL,             /*After bit rate element start, it should be bit rate element value*/
   ENST,              /*After bit rate element end, it should be enable element start*/
   BRED,              /*After bit rate element value, it should be bit rate element end*/
   ENVAL,             /*After enable element start, it should be enable element value*/
   RXPST,             /*After enable element end, it should be rx parity check element start*/
   ENED,              /*After enable element value, it should be enable element end*/
   RXPVAL,            /*After rx parity check element start, it should be rx parity check element value*/
   BMST,              /*After rx parity check element end, it should be buffer mode element start*/
   RXPED,             /*After rx parity check element value, it should be rx parity check element end*/
   BMVAL,             /*After buffer mode element start, it should be buffer mode element value*/
   IWST,              /*After buffer mode element end, it should be internal wrap element start*/
   BMED,              /*After buffer mode element value, it should be buffer mode element end*/
   IWVAL,             /*After internal wrap element start, it should be internal wrap element value*/
   RXCHED,            /*After internal wrap element end, it should be rx channel element start*/
   IWED,              /*After internal wrap element value, it should be internal wrap element end*/
   TXPVAL,            /*After tx parity check element start, it should be tx parity check element value*/
   PEST,              /*After tx parity check element end, it should be protocol element start*/
   TXPED,             /*After tx parity check element value, it should be tx parity check element end*/
   PEVAL,             /*After protocol error element start, it should be protocol error element value*/
   TXCHED,            /*After protocol error element end, it should be tx channel element start*/
   PEED               /*After protocol error element value, it should be protocol error element end*/    
};

/*Schedule message XML item switch, first choice*/
CDEV_PARM_SSI_TYPE XML_SM_ITEM_REL_TAB[XML_SM_ITEM_END] = 
{
   MSGNUMST,        /*After schedule message element start, it should be message number element start*/
   XML_SM_ITEM_END, /*Nothing after schedule message element end*/
   MSGNUMVAL,       /*After message number element start, it should be message number element value*/
   MSGGRPST,        /*After message number element end, it should be message group element start*/
   MSGNUMED,        /*After message number element value, it should be message number element end*/
   MSGST,           /*After message group element start, it should be message element start*/
   SMED,            /*After message group element end, it should be schedule message element end*/
   SBRDIDXST,       /*After message element start, it should be board index element start*/
   MSGST,           /*After message element end, it should be message element start*/
   SBRDIDXVAL,      /*After board index element start, it should be board index element value*/
   SCHIDXST,        /*After board index element end, it should be channel index element start*/
   SBRDIDXED,       /*After board index element value, it should be board index element end*/
   SCHIDXVAL,       /*After channel index element start, it should be channel index element value*/
   MSGCTST,         /*After channel index element end, it should be message content element start*/
   SCHIDXED,        /*After channel index element value, it should be channel index element end*/
   MSGCTVAL,        /*After message content element start, it should be message content element value*/
   RTST,            /*After message content element end, it should be rate element start*/ 
   MSGCTED,         /*After message content element value, it should be message content element end*/
   RTVAL,           /*After rate element start, it should be rate element value*/
   STOFFST,         /*After rate element end, it should be start offset element start*/
   RTED,            /*After rate element value, it should be rate element end*/
   STOFFVAL,        /*After start offset element start, it should be start offset element value*/
   TRCNST,          /*After start offset element end, it should be transmit count element start*/     
   STOFFED,         /*After start offset element value, it should be start offset element end*/
   TRCNVAL,         /*After transmit count element start, it should be transmit count element value*/
   ETIDST,          /*After transmit count element end, it should be equipment id elemenet start*/ 
   TRCNED,          /*After transmit count element value, it should be transmit count element end*/        
   ETIDVAL,         /*After equipmetn id element start, it should be equipment id element value*/
   MSGED,           /*After equipmetn id element end, it should be message element end*/
   ETIDED           /*After equipmetn id element value, it should be equipmetn id element end*/
};

/*Channel configuration XML element tag items*/
#define XML_CC_ITEMNAME_NUM 32

CDEV_PARM_SSI_TYPE XML_CC_ITEMNAME_IDX_TAB[XML_CC_ITEMNAME_NUM] = {CCST,CCED,BRDIDXST,BRDIDXED,RXNUMST,RXNUMED,TXNUMST,TXNUMED, 
                                                                  RXGRPST,RXGRPED,TXGRPST,TXGRPED,RXCHST,RXCHED,TXCHST,TXCHED,   
                                                                  CHIDXST,CHIDXED,BRST,BRED,ENST,ENED,RXPST,RXPED,BMST,BMED,   
                                                                  IWST,IWED,TXPST,TXPED,PEST,PEED};

pCEI_CHAR XML_CC_ITEMNAME_TAB[XML_CC_ITEMNAME_NUM] = {"x30boardconfig",
                                                      "/x30boardconfig",
                                                      "boardindex",
                                                      "/boardindex",
                                                      "rxchannelnum",
                                                      "/rxchannelnum",
                                                      "txchannelnum",
                                                      "/txchannelnum",
                                                      "rxchannelgroup",
                                                      "/rxchannelgroup",
                                                      "txchannelgroup",
                                                      "/txchannelgroup",
                                                      "rxchannel",
                                                      "/rxchannel",
                                                      "txchannel",
                                                      "/txchannel",
                                                      "channelindex",
                                                      "/channelindex",
                                                      "bitrate",
                                                      "/bitrate",
                                                      "enable",
                                                      "/enable",
                                                      "rxparity",
                                                      "/rxparity",
                                                      "buffermode",
                                                      "/buffermode",
                                                      "internalwrap",
                                                      "/internalwrap",
                                                      "txparity",
                                                      "/txparity",
                                                      "protocolerror",
                                                      "/protocolerror"};

/*Channel configuration XML element value items*/
#define XML_CC_VAL_ITEM_NUM 10

CDEV_PARM_SSI_TYPE XML_CC_VAL_ITEM_IDX_TAB[XML_CC_VAL_ITEM_NUM] = {RXNUMVAL,TXNUMVAL,CHIDXVAL,BRVAL,ENVAL,RXPVAL,BMVAL,
                                                                  IWVAL,TXPVAL,PEVAL};

/*Schedule message XML element tag items*/
#define XML_SM_ITEMNAME_NUM 22

CDEV_PARM_SSI_TYPE XML_SM_ITEMNAME_IDX_TAB[XML_SM_ITEMNAME_NUM] = {SMST, SMED,MSGNUMST,MSGNUMED,MSGGRPST,MSGGRPED,      
                                                                  MSGST,MSGED,SBRDIDXST,SBRDIDXED,SCHIDXST,SCHIDXED,MSGCTST,
                                                                  MSGCTED,RTST,RTED,STOFFST,STOFFED,TRCNST,TRCNED,ETIDST,ETIDED};

/*Schedule message XML item tag string*/
pCEI_CHAR XML_SM_ITEMNAME_TAB[XML_SM_ITEMNAME_NUM] = {"x30schedulemessage",
                                                      "/x30schedulemessage",
                                                      "messagenumber",
                                                      "/messagenumber",
                                                      "messagegroup",
                                                      "/messagegroup",
                                                      "message",
                                                      "/message",
                                                      "boardindex",
                                                      "/boardindex",
                                                      "channelindex",
                                                      "/channelindex",
                                                      "messagecontent",
                                                      "/messagecontent",
                                                      "rate",
                                                      "/rate",
                                                      "startoffset",
                                                      "/startoffset",
                                                      "transmitcount",
                                                      "/transmitcount",
                                                      "equipementid",
                                                      "/equipementid"};

/*Schedule message XML element value items*/
#define XML_SM_VAL_ITEM_NUM 7

CDEV_PARM_SSI_TYPE XML_SM_VAL_ITEM_IDX_TAB[XML_SM_VAL_ITEM_NUM] = {SBRDIDXVAL,SCHIDXVAL,MSGCTVAL,RTVAL,STOFFVAL,
                                                                  TRCNVAL,ETIDVAL};

#define CEI_INT32_LEN 10

CEI_INT32 maxCINT32[10] = {4,2,9,4,9,6,7,2,9,5};
CDEV_PARM_SSI_TYPE configuration;         /* The device model/configuration        */

CDEV_API_RET_TYPE ar_is_number(pCEI_CHAR pStr);
CDEV_API_RET_TYPE ar_is_32_ucint(pCEI_CHAR pStr);

/*Sub functions for channel configuration XML decoding*/
CEI_VOID ar_initialize_channel_configuration_struct(CDEV_CHAN_TYPE cintRxChNum,
                                                    CDEV_CHAN_TYPE cintTxChNum,
                                                    RX_CHANNEL_PARMS_TYPE *prxChanConfig,
                                                    TX_CHANNEL_PARMS_TYPE *ptxChanConfig);

CDEV_API_RET_TYPE ar_set_channel_configuration(CDEV_BOARD_TYPE cintBrdIdx,
                                               CDEV_CHAN_TYPE cintRxChNum,
                                               CDEV_CHAN_TYPE cintTxChNum,
                                               RX_CHANNEL_PARMS_TYPE *prxChanConfig,
                                               TX_CHANNEL_PARMS_TYPE *ptxChanConfig);

CDEV_API_RET_TYPE ar_get_item_from_xml_file(FILE *fConfigFilePtr, XMLSEC *pHead);

CEI_VOID ar_release_xmlsec_list(XMLSEC *pHead);

CDEV_API_RET_TYPE ar_channel_config_from_xml_list(XMLSEC    *pHead,
                                                  CDEV_BOARD_TYPE *pcintBoardIndex,
                                                  CDEV_CHAN_TYPE *pcintRxChNum,
                                                  CDEV_CHAN_TYPE *pcintTxChNum,
                                                  RX_CHANNEL_PARMS_TYPE *prxChanConfig,
                                                  CDEV_PARM_SSI_TYPE cintRxChConfigSize,
                                                  TX_CHANNEL_PARMS_TYPE *ptxChanConfig,
                                                  CDEV_PARM_SSI_TYPE cintTxChConfigSize);

CDEV_API_RET_TYPE ar_check_cc_xml_list_input(XMLSEC    *pHead,
                                             CDEV_BOARD_TYPE *pcintBoardIndexTemp,
                                             CDEV_CHAN_TYPE *pcintRxChNum,
                                             CDEV_CHAN_TYPE *pcintTxChNum,
                                             RX_CHANNEL_PARMS_TYPE *prxChanConfig,
                                             TX_CHANNEL_PARMS_TYPE *ptxChanConfig);

CDEV_API_RET_TYPE ar_xml_check_title(XMLSEC *pHead, XMLSEC **pReturn);

CDEV_API_RET_TYPE ar_xml_get_cc_item_name_idx(CDEV_PARM_SSI_TYPE cintCurrentItemType,CDEV_PARM_SSI_TYPE *pcintIndex);

CDEV_API_RET_TYPE ar_xml_get_cc_val_item_idx(CDEV_PARM_SSI_TYPE cintCurrentItemType, CDEV_PARM_SSI_TYPE *pcintIndex);

CDEV_API_RET_TYPE ar_xml_cc_save_config(CDEV_PARM_SSI_TYPE cintCurrentItemType,
                                        CDEV_CHAN_TYPE *pcintRxChNum,
                                        CDEV_CHAN_TYPE *pcintTxChNum,
                                        CDEV_CHAN_TYPE cintCurRxIndex,
                                        CDEV_CHAN_TYPE cintCurTxIndex,
                                        CDEV_CHAN_TYPE cintRxTx,
                                        RX_CHANNEL_PARMS_TYPE *prxChanConfig,
                                        CDEV_PARM_SSI_TYPE cintRxChConfigSize,
                                        TX_CHANNEL_PARMS_TYPE *ptxChanConfig,
                                        CDEV_PARM_SSI_TYPE cintTxChConfigSize,
                                        CDEV_PARM_SSI_TYPE cintValue);

CDEV_API_RET_TYPE ar_channel_config_xml_next_item_type(CDEV_PARM_SSI_TYPE cintCurrentItemType,
                                                       CDEV_CHAN_TYPE cintCurRxChIndex,
                                                       CDEV_CHAN_TYPE cintTotalRxChNum,
                                                       CDEV_CHAN_TYPE cintCurTxChIndex,
                                                       CDEV_CHAN_TYPE cintTotalTxChNum,
                                                       CDEV_CHAN_TYPE cintRxTx);

CDEV_API_RET_TYPE ar_check_rx_channel_num(pCEI_CHAR pVal, CDEV_CHAN_TYPE *pcintRxChNum);
CDEV_API_RET_TYPE ar_check_tx_channel_num(pCEI_CHAR pVal, CDEV_CHAN_TYPE *pcintTxChNum);
CDEV_API_RET_TYPE ar_check_channel_idx(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE *pcintChIndex);
CDEV_API_RET_TYPE ar_check_bit_rate(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE *pcintBitRate);
CDEV_API_RET_TYPE ar_check_enable(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE *pcintEnable);
CDEV_API_RET_TYPE ar_check_rx_parity_check(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE *pcintRxParityChk);
CDEV_API_RET_TYPE ar_check_buffer_mode(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE *pcintBufMode);
CDEV_API_RET_TYPE ar_check_internal_wrap(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE *pcintIntWrap);
CDEV_API_RET_TYPE ar_check_tx_parity_check(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE *pcintTxParityChk);
CDEV_API_RET_TYPE ar_check_protocol_error(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE *pcintProtErr);
/*Sub functions for channel configuration XML decoding*/

/*Sub functions for schedule messge XML decoding*/
CDEV_API_RET_TYPE ar_schedule_message_from_xml_list(XMLSEC *pHead, CDEV_PARM_SSI_TYPE *pcintMsgNum);
CDEV_API_RET_TYPE ar_xml_get_sm_item_name_idx(CDEV_PARM_SSI_TYPE cintCurrentItemType,CDEV_PARM_SSI_TYPE *pcintIndex);
CDEV_API_RET_TYPE ar_xml_get_sm_val_item_idx(CDEV_PARM_SSI_TYPE cintCurrentItemType, CDEV_PARM_SSI_TYPE *pcintIndex);
CDEV_API_RET_TYPE ar_schedule_message_xml_next_item_type(CDEV_PARM_SSI_TYPE cintCurrentItemType,CDEV_PARM_SSI_TYPE cintMsgNum,CDEV_PARM_SSI_TYPE cintMsgIndex);
CDEV_API_RET_TYPE ar_sm_check_save_board_index(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE cintMsgIndex);
CDEV_API_RET_TYPE ar_sm_check_save_channel_index(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE cintMsgIndex);
CDEV_API_RET_TYPE ar_sm_check_save_rate(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE cintMsgIndex);
CDEV_API_RET_TYPE ar_sm_check_save_start_offset(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE cintMsgIndex);
CDEV_API_RET_TYPE ar_sm_check_save_transmit_count(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE cintMsgIndex);
CDEV_API_RET_TYPE ar_sm_check_save_data(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE cintMsgIndex);
/*Sub functions for schedule messge XML decoding*/

/*Function list to procss channel configuration XML element values*/
CDEV_API_RET_TYPE (*XML_CC_VAL_FUN_TAB[XML_CC_VAL_ITEM_NUM])(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE *pcintReturn) = 
   {ar_check_rx_channel_num,
    ar_check_tx_channel_num,
    ar_check_channel_idx,
    ar_check_bit_rate,
    ar_check_enable,
    ar_check_rx_parity_check,
    ar_check_buffer_mode,
    ar_check_internal_wrap,
    ar_check_tx_parity_check,
    ar_check_protocol_error};


/*Function list to process schedule message XML element values*/
CDEV_API_RET_TYPE (*XML_SM_VAL_FUN_TAB[XML_SM_VAL_ITEM_NUM])(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE cintMsgIndex) = 
   {ar_sm_check_save_board_index,
    ar_sm_check_save_channel_index,
    ar_sm_check_save_data,
    ar_sm_check_save_rate,
    ar_sm_check_save_start_offset,
    ar_sm_check_save_transmit_count,
    NULL};    

/*===========================================================================*
 * ENTRY:   A R _ C O N F I G _ C H A N N E L S _ F R O M _ T X T _ F I L E       
 *===========================================================================*
 *
 * FUNCTION:    This routine configures a CEI-x30 channel set from file
 *              
 * PARAMETERS:  CDEV_BOARD_TYPE board -- (input) board to reference (0 to 127)
 *              pCEI_CHAR cfgFileName (input) channel configuration file name
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_BADLOAD     File not found
 *                                 ARS_INVHARCMD   Invalid File entry
 *
 *              from ar_set_device_config:
 *
 *                         ARS_INVARG      "item" parameter value is not
 *                                            supported by this function
 *                         ARS_INVHARVAL   "value" was invalid or channel
 *                                            not supported on this device
 *                         ARS_INVBOARD    Board parameter value invalid or
 *                                            board has not been initialized
 *                         ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                         ARS_BADLOAD     The specified configuration file 
 *                                         path/name could not be opened
 *
 * GLOBALS:     none
 * 
 * DESCRIPTION: This routine reads the contents of the channel configuration
 *              text file (ASCII only format) and configures the respective
 *              CEI-x30 device ARINC 429 channels based on the contents of
 *              that file.
 * 
 *===========================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_config_channels_from_txt_file
(
   CDEV_BOARD_TYPE BoardIdx,           /* (input) Board Index for channel configuration*/
   pCEI_CHAR cfgFileName         /* (input) channel configuration file name */
)
{
   CDEV_API_RET_TYPE status = ARS_NORMAL;
   CEI_CHAR fileEntry[160];
   FILE * inputConfigFilePtr;
   CONFIG_FILE_SECTION_TYPE whereAmI;
   CEI_INT32 deviceIndex = 0;
   CEI_INT32 rxChannelCount = 0;
   CEI_INT32 txChannelCount = 0;
   CEI_INT32 channelIndex;
   CEI_INT32 j;
   CEI_INT32 passesThruParser;
   CEI_INT32 busSpeed;
   CEI_INT32 refClockFreq;
   CDEV_BOARD_TYPE board;
   RX_CHANNEL_PARMS_TYPE rxChanConfig[CEIDEV_MAX_RX_CHAN];
   TX_CHANNEL_PARMS_TYPE txChanConfig[CEIDEV_MAX_RX_CHAN];
   pCEI_CHAR toss;
   pCEI_CHAR tempStrPtr;
   
   if ( (inputConfigFilePtr = fopen(cfgFileName, "rb")) == NULL ) {
      status = ARS_BADLOAD;
   }
   else {
      for (j = 0; j < CEIDEV_MAX_RX_CHAN; j++) {
         rxChanConfig[j].channel      = CEIDEV_MAX_RX_CHAN;
         rxChanConfig[j].enable       = AR_OFF;
         rxChanConfig[j].bitRate      = ARU_SPEED_HIGH;
         rxChanConfig[j].parity       = AR_OFF;
         rxChanConfig[j].bufferMode   = AR_OFF;
         rxChanConfig[j].internalWrap = AR_WRAP_OFF;
         txChanConfig[j].channel      = CEIDEV_MAX_RX_CHAN;
         txChanConfig[j].enable       = AR_OFF;
         txChanConfig[j].bitRate      = ARU_SPEED_HIGH;
         txChanConfig[j].parity       = ARU_PARITY_ODD;
         txChanConfig[j].bitError     = AR_OFF;
         txChanConfig[j].gapError     = AR_OFF;
      }

      status = ar_get_device_config (BoardIdx, 0, ARU_CONFIGURATION, &configuration);
      configuration &= CEIDEV_CONFIGURATION_MASK;
      if ((configuration == CEIDEV_CONFIG_RAR15XT) || (configuration == CEIDEV_CONFIG_RAR15XT)) {
         refClockFreq = CEIDEV_20MHZ_REF_CLOCK;
      }
      else {
         refClockFreq = CEIDEV_16MHZ_REF_CLOCK;
      }

      whereAmI = BOARD;
      while ((fgets(fileEntry, 159, inputConfigFilePtr) != NULL) && (whereAmI != DONE)) {
         switch (whereAmI) {
            case BOARD:
               /* Board Index attribute entry format is:
               
                     Board Index:0
               */
               if (strncmp (fileEntry, BOARD_TAG, strlen(BOARD_TAG)) == 0) {
                  toss = strtok (fileEntry, TOKEN_VALUE_FIELD_DELIMITER);
                  tempStrPtr = strtok (NULL, "");
                  deviceIndex = atoi (tempStrPtr);
                  if ((deviceIndex < 0) || (deviceIndex > CEIDEV_MAX_BRD_IDX)) {
                     return ARS_INVBOARD;
                  }
                  whereAmI = QUERY;
               }
               else
                  whereAmI = DONE;
               break;

            case QUERY:
               /* Receive and Transmit Channel Count attribute entry format is:
               
                     RxChannel Number:2
                     TxChannel Number:2

                  Receive and Transmit Channel Entry Start/Stop format is:                  
                  
                     <RxChannel
                     RxChannel>
                     <TxChannel
                     TxChannel>
               */
               j = 0;
               if (strncmp (fileEntry, RX_CH_START_TAG, strlen(RX_CH_START_TAG)) == 0) {
                  whereAmI = RXATT;
               }
               else if (strncmp (fileEntry, TX_CH_START_TAG, strlen(TX_CH_START_TAG)) == 0) {
                  whereAmI = TXATT;
               }
               else if (strncmp (fileEntry, RX_CH_COUNT_TAG, strlen(RX_CH_COUNT_TAG)) == 0) {
                  toss = strtok (fileEntry, TOKEN_VALUE_FIELD_DELIMITER);
                  tempStrPtr = strtok (NULL, "");
                  rxChannelCount = atoi (tempStrPtr);
                  if ((rxChannelCount < 0) || (rxChannelCount > CEIDEV_MAX_RX_CHAN)) {
                     return ARS_INVHARCMD;
                  }
               }
               else if (strncmp (fileEntry, TX_CH_COUNT_TAG, strlen(TX_CH_COUNT_TAG)) == 0) {
                  toss = strtok (fileEntry, TOKEN_VALUE_FIELD_DELIMITER);
                  tempStrPtr = strtok (NULL, "");
                  txChannelCount = atoi (tempStrPtr);
                  if ((txChannelCount < 0) || (txChannelCount > CEIDEV_MAX_TX_CHAN)) {
                     return ARS_INVHARCMD;
                  }
               }
               else {
                  whereAmI = DONE;
                  status = ARS_INVHARCMD;
               }
               break;

            case RXATT:
               /* Receive Channel Attribute Entries in the configuration file look like this:
               
                  ChannelIndex:0___Enable:ENABLED___BitRate:100000___ParityCheck:ENABLED___BufferMode:INDIVIDUAL___InternalWrap:ENABLED
                  ChannelIndex:1___Enable:DISABLED___BitRate:12500___ParityCheck:DISABLED___BufferMode:MERGED___InternalWrap:DISABLED
                  
                  The combined use of strtok (NULL, ":") and strtok (NULL, "_") parses only the attribute settings from each entry,
                  so it is always assumed the format of the channel attribute string will be fully populated using the order and attribute
                  format demonstrated above.  If a receive channel entry deviates from what is shown above, a failure will occur.
               */
               toss = strtok (fileEntry, TOKEN_VALUE_FIELD_DELIMITER);
               if ((toss != NULL) && (strncmp (toss, CH_INDEX_TAG, strlen(CH_INDEX_TAG)) == 0)) {
                  /* extract the receive channel number */
                  tempStrPtr = strtok (NULL, ATTRIBUTE_ENTRY_DELIMITER);

                  /* assure the string is a number */
                  if (False == ar_is_number(tempStrPtr)) {
                     return ARS_INVHARVAL;
                  }
                  
                  channelIndex = atoi (tempStrPtr);

                  if ((channelIndex < CEIDEV_MAX_RX_CHAN) && (channelIndex >= 0)) {
                     rxChanConfig[j].channel = (CDEV_PARM_SSI_TYPE)channelIndex; 
                     
                     tempStrPtr = strtok (NULL, TOKEN_VALUE_FIELD_DELIMITER);
                     if (tempStrPtr == NULL)
                        break;
                     while (tempStrPtr[0] == ATTRIBUTE_ENTRY_DELIMITER_CHAR)
                        tempStrPtr++;

                     passesThruParser = 0;
                     while (tempStrPtr != NULL) {
                        /* extract the receive channel enable state */
                        if (strncmp (tempStrPtr, CH_ENABLE_TAG, strlen(CH_ENABLE_TAG)) == 0) {
                           tempStrPtr = strtok (NULL, ATTRIBUTE_ENTRY_DELIMITER);
   
                           if (strncmp (tempStrPtr, OPT_ENABLED, strlen(OPT_ENABLED)) == 0) 
                              rxChanConfig[j].enable = AR_ON;
                           else if (strncmp (tempStrPtr, OPT_DISABLED, strlen(OPT_DISABLED)) == 0) 
                              rxChanConfig[j].enable = AR_OFF;
                           else {
                              whereAmI = CFG_ERR;                             /* channel entry format error detected */
                              break;
                           }
   
                           tempStrPtr = strtok (NULL, TOKEN_VALUE_FIELD_DELIMITER);
                           if (tempStrPtr == NULL)
                              break;
                           while (tempStrPtr[0] == ATTRIBUTE_ENTRY_DELIMITER_CHAR)
                              tempStrPtr++;
                        }
   
                        /* extract the bus speed & convert from frequency to divisor */
                        if (strncmp (tempStrPtr, CH_BITRATE_TAG, strlen(CH_BITRATE_TAG)) == 0) {
                           tempStrPtr = strtok (NULL, ATTRIBUTE_ENTRY_DELIMITER);
      
                           if ( (True == ar_is_number(tempStrPtr))  && (atoi(tempStrPtr) >= CEIDEV_MIN_BIT_RATE) &&
                                (atoi(tempStrPtr) <= CEIDEV_MAX_BIT_RATE) ) {                      
                              sscanf (tempStrPtr, "%d", &busSpeed);
                              rxChanConfig[j].bitRate = (CDEV_PARM_SSI_TYPE)((refClockFreq / busSpeed) - 2);
                           }
                           else {
                              whereAmI = CFG_ERR;                             /* bus speed entry format error detected */
                              break;
                           }
   
                           tempStrPtr = strtok (NULL, TOKEN_VALUE_FIELD_DELIMITER);
                           if (tempStrPtr == NULL)
                              break;
                           while (tempStrPtr[0] == ATTRIBUTE_ENTRY_DELIMITER_CHAR)
                              tempStrPtr++;
                        }
                        
                        /* extract the parity enable selection */
                        if (strncmp (tempStrPtr, CH_PARITY_TAG, strlen(CH_PARITY_TAG)) == 0) {
                           tempStrPtr = strtok (NULL, ATTRIBUTE_ENTRY_DELIMITER);
      
                           if (strncmp (tempStrPtr, OPT_ENABLED, strlen(OPT_ENABLED)) == 0) 
                              rxChanConfig[j].parity = AR_ON;
                           else if (strncmp (tempStrPtr, OPT_DISABLED, strlen(OPT_DISABLED)) == 0) 
                              rxChanConfig[j].parity = AR_OFF;
                           else {
                              whereAmI = CFG_ERR;                             /* channel entry format error detected */
                              break;
                           }
   
                           tempStrPtr = strtok (NULL, TOKEN_VALUE_FIELD_DELIMITER);
                           if (tempStrPtr == NULL)
                              break;
                           while (tempStrPtr[0] == ATTRIBUTE_ENTRY_DELIMITER_CHAR)
                              tempStrPtr++;
                        }
   
                        /* extract the merged mode buffer enable selection */
                        if (strncmp (tempStrPtr, CH_BUF_MODE_TAG, strlen(CH_BUF_MODE_TAG)) == 0) {
                           tempStrPtr = strtok (NULL, ATTRIBUTE_ENTRY_DELIMITER);
      
                           if (strncmp (tempStrPtr, OPT_MERGED, strlen(OPT_MERGED)) == 0) 
                              rxChanConfig[j].bufferMode = AR_ON;
                           else if (strncmp (tempStrPtr, OPT_INDIVIDUAL, strlen(OPT_INDIVIDUAL)) == 0) 
                              rxChanConfig[j].bufferMode = AR_OFF;
                           else {
                              whereAmI = CFG_ERR;                             /* channel entry format error detected */
                              break;
                           }
   
                           tempStrPtr = strtok (NULL, TOKEN_VALUE_FIELD_DELIMITER);
                           if (tempStrPtr == NULL)
                              break;
                           while (tempStrPtr[0] == ATTRIBUTE_ENTRY_DELIMITER_CHAR)
                              tempStrPtr++;
                        }
   
                        /* extract the internal wrap selection */
                        if (strncmp (tempStrPtr, CH_WRAP_TAG, strlen(CH_WRAP_TAG)) == 0) {
                           tempStrPtr = strtok (NULL, "");
      
                           if (strncmp (tempStrPtr, OPT_ENABLED, strlen(OPT_ENABLED)) == 0) 
                              rxChanConfig[j].internalWrap = AR_WRAP_ON;      
                           else if (strncmp (tempStrPtr, OPT_DISABLED, strlen(OPT_DISABLED)) == 0) 
                              rxChanConfig[j].internalWrap = AR_WRAP_OFF;
                           else {
                              whereAmI = CFG_ERR;                             /* channel entry format error detected */
                              break;
                           }
                        
                           tempStrPtr = strtok (NULL, TOKEN_VALUE_FIELD_DELIMITER);
                           if (tempStrPtr == NULL)
                              break;
                           while (tempStrPtr[0] == ATTRIBUTE_ENTRY_DELIMITER_CHAR)
                              tempStrPtr++;
                        }
                     
                        if (++passesThruParser > 10) {
                           whereAmI = CFG_ERR;                             /* unrecognized tag in channel cfg entry */
                           break;
                        }
                     }
                  }
                  else {
                      return ARS_INVHARVAL;
                  }
                  j++;
               }
               else if (strncmp (toss, RX_CH_END_TAG, strlen(RX_CH_END_TAG)) == 0) {
                  whereAmI = QUERY;
               }
               else {
                  whereAmI = CFG_ERR;                             /* ChannelIndex tag format error  */
                  break;            
               }
               break;

            case TXATT:
               /* Transmit Channel Attribute Entries in the configuration file look like this:
               
                  ChannelIndex:0___Enable:ENABLED___BitRate:100000___ParityCheck:NONE___ProtocolError:BIT COUNT HIGH
                  ChannelIndex:1___Enable:DISABLED___BitRate:100000___ParityCheck:ODD___ProtocolError:BIT COUNT LOW
                  ChannelIndex:2___Enable:ENABLED___BitRate:100000___ParityCheck:EVEN___ProtocolError:SHORT GAP
                  
                  The combined use of strtok (NULL, ":") and strtok (NULL, "_") parses only the attribute settings from each entry,
                  so it is always assumed the format of the channel attribute string will be fully populated using the order and attribute
                  format demonstrated above.  If a receive channel entry deviates from what is shown above, a failure will occur.
               */
               toss = strtok (fileEntry, TOKEN_VALUE_FIELD_DELIMITER);
               if ((toss != NULL) && (strncmp (toss, CH_INDEX_TAG, strlen(CH_INDEX_TAG)) == 0)) 
               {
                  /* extract the transmit channel number */
                  tempStrPtr = strtok (NULL, ATTRIBUTE_ENTRY_DELIMITER);
                  
                  /* assure the string is a number */
                  if (False == ar_is_number(tempStrPtr)) {
                     return ARS_INVHARVAL;
                  }

                  channelIndex = atoi (tempStrPtr);

                  if ((channelIndex < CEIDEV_MAX_TX_CHAN) && (channelIndex >= 0)) {
                     txChanConfig[j].channel = (CDEV_PARM_SSI_TYPE)channelIndex; 
                     
                     /* extract the transmit channel enable state */
                     tempStrPtr = strtok (NULL, TOKEN_VALUE_FIELD_DELIMITER);
                     if (tempStrPtr == NULL)
                        break;
                     while (tempStrPtr[0] == ATTRIBUTE_ENTRY_DELIMITER_CHAR)
                        tempStrPtr++;

                     passesThruParser = 0;
                     while (tempStrPtr != NULL) {
                        if (strncmp (tempStrPtr, CH_ENABLE_TAG, strlen(CH_ENABLE_TAG)) == 0) {
                           tempStrPtr = strtok (NULL, ATTRIBUTE_ENTRY_DELIMITER);
   
                           if (strncmp (tempStrPtr, OPT_ENABLED, strlen(OPT_ENABLED)) == 0) 
                              txChanConfig[j].enable = AR_ON;
                           else if (strncmp (tempStrPtr, OPT_DISABLED, strlen(OPT_DISABLED)) == 0) 
                              txChanConfig[j].enable = AR_OFF;
                           else {
                              whereAmI = CFG_ERR;                             /* channel entry format error detected */
                              break;
                           }
                        
                           tempStrPtr = strtok (NULL, TOKEN_VALUE_FIELD_DELIMITER);
                           if (tempStrPtr == NULL)
                              break;
                           while (tempStrPtr[0] == ATTRIBUTE_ENTRY_DELIMITER_CHAR)
                              tempStrPtr++;
                        }
   
                        /* extract the bus speed & convert from frequency to divisor */
                        if (strncmp (tempStrPtr, CH_BITRATE_TAG, strlen(CH_BITRATE_TAG)) == 0) {
                           tempStrPtr = strtok (NULL, ATTRIBUTE_ENTRY_DELIMITER);
      
                           if ( (True == ar_is_number(tempStrPtr)) && (atoi(tempStrPtr) >= CEIDEV_MIN_BIT_RATE) &&
                                (atoi(tempStrPtr) <= CEIDEV_MAX_BIT_RATE) ) { /* assure the string is a number */
                              sscanf (tempStrPtr, "%d", &busSpeed);
                              txChanConfig[j].bitRate = (CDEV_PARM_SSI_TYPE)((refClockFreq / busSpeed) - 2);
                           }
                           else {
                              whereAmI = CFG_ERR;                             /* bus speed entry format error detected */
                              break;
                           }
   
                           tempStrPtr = strtok (NULL, TOKEN_VALUE_FIELD_DELIMITER);
                           if (tempStrPtr == NULL)
                              break;
                           while (tempStrPtr[0] == ATTRIBUTE_ENTRY_DELIMITER_CHAR)
                              tempStrPtr++;
                        }
                        
                        /* extract the parity enable selection */
                        if (strncmp (tempStrPtr, CH_PARITY_TAG, strlen(CH_PARITY_TAG)) == 0) {
                           tempStrPtr = strtok (NULL, ATTRIBUTE_ENTRY_DELIMITER);
      
                           if (strncmp (tempStrPtr, OPT_ODD, strlen(OPT_ODD)) == 0) 
                              txChanConfig[j].parity = ARU_PARITY_ODD;        /* extract the parity enable selection */
                           else if (strncmp (tempStrPtr, OPT_EVEN, strlen(OPT_EVEN)) == 0) 
                              txChanConfig[j].parity = ARU_PARITY_EVEN;
                           else if (strncmp (tempStrPtr, OPT_NONE, strlen(OPT_NONE)) == 0) 
                              txChanConfig[j].parity = ARU_PARITY_NONE;
                           else {
                              whereAmI = CFG_ERR;                             /* channel entry format error detected */
                              break;
                           }
   
                           tempStrPtr = strtok (NULL, TOKEN_VALUE_FIELD_DELIMITER);
                           if (tempStrPtr == NULL)
                              break;
                           while (tempStrPtr[0] == ATTRIBUTE_ENTRY_DELIMITER_CHAR)
                              tempStrPtr++;
                        }
                        
                        /* extract the bit count & gap error selection */
                        if (strncmp (tempStrPtr, CH_PROT_ERR_TAG, strlen(CH_PROT_ERR_TAG)) == 0) {
                           tempStrPtr = strtok (NULL, ATTRIBUTE_ENTRY_DELIMITER);
                        
                           if (strncmp (tempStrPtr, OPT_BIT_COUNT_HIGH, strlen(OPT_BIT_COUNT_HIGH)) == 0) { 
                              txChanConfig[j].bitError = AR_HI;
                              txChanConfig[j].gapError = AR_OFF;
                           }
                           else if (strncmp (tempStrPtr, OPT_BIT_COUNT_LOW, strlen(OPT_BIT_COUNT_LOW)) == 0) { 
                              txChanConfig[j].bitError = AR_LO;
                              txChanConfig[j].gapError = AR_OFF;
                           }
                           else if (strncmp (tempStrPtr, OPT_SHORT_GAP, strlen(OPT_SHORT_GAP)) == 0) {
                              txChanConfig[j].bitError = AR_OFF;
                              txChanConfig[j].gapError = AR_ON;
                           }
                           else if (strncmp (tempStrPtr, OPT_NONE, strlen(OPT_NONE)) == 0) {
                              txChanConfig[j].bitError = AR_OFF;
                              txChanConfig[j].gapError = AR_OFF;
                           }
                           else
                           {
                              whereAmI = CFG_ERR;                       
                              break;
                           }
                                           
                           tempStrPtr = strtok (NULL, TOKEN_VALUE_FIELD_DELIMITER);
                           if (tempStrPtr == NULL)
                              break;
                           while (tempStrPtr[0] == ATTRIBUTE_ENTRY_DELIMITER_CHAR)
                              tempStrPtr++;
                        }
                     
                        if (++passesThruParser > 10) {
                           whereAmI = CFG_ERR;                             /* unrecognized tag in channel cfg entry */
                           break;
                        }
                     }
                  }
                  else {
                     return ARS_INVHARVAL;
                  }
                  j++;
               }
               else if (strncmp (toss, TX_CH_END_TAG, strlen(TX_CH_END_TAG)) == 0) {
                  whereAmI = QUERY;
               }
               else
               {
                  whereAmI = CFG_ERR;                             /* ChannelIndex tag format error  */
                  break;            
               }
               break;

            case CFG_ERR:
               status = ARS_INVHARCMD;
               whereAmI = DONE;
               break;

            default:
               whereAmI = DONE;
               break;
         }
      }

      fclose (inputConfigFilePtr);
      
      if ((BoardIdx >= 0) && (BoardIdx <= CEIDEV_MAX_BRD_IDX)) {
         board = BoardIdx;
      }
      else {
         return ARS_INVBOARD;
      }
      
      /*  Now update the specified board's channel attributes with what was read from the config file. */
      
     if ( (status == ARS_NORMAL) && ((txChannelCount > 0) || (rxChannelCount > 0)) ) {
         for (channelIndex = 0; (channelIndex < txChannelCount) &&  (status == ARS_NORMAL); channelIndex++) {
            if (txChanConfig[channelIndex].channel < CEIDEV_MAX_RX_CHAN) {
               status = ar_set_device_config (board, txChanConfig[channelIndex].channel, ARU_TX_BITRATE, txChanConfig[channelIndex].bitRate);
               if (status == ARS_NORMAL) {
                  status = ar_set_device_config (board, txChanConfig[channelIndex].channel, ARU_TX_PARITY, txChanConfig[channelIndex].parity);
               }
               if (status == ARS_NORMAL) {
                  status = ar_set_device_config (board, txChanConfig[channelIndex].channel, ARU_TX_BIT_ERROR, txChanConfig[channelIndex].bitError);
               }
               if (status == ARS_NORMAL) {
                  status = ar_set_device_config (board, txChanConfig[channelIndex].channel, ARU_TX_GAP_ERROR, txChanConfig[channelIndex].gapError);
               }
               if (status == ARS_NORMAL) {
                  status = ar_set_device_config (board, txChanConfig[channelIndex].channel, ARU_TX_FIFO_ENABLE, txChanConfig[channelIndex].enable);
               }
         }
       }
       for (channelIndex = 0; (channelIndex < rxChannelCount) &&  (status == ARS_NORMAL); channelIndex++) {
            if (rxChanConfig[channelIndex].channel < CEIDEV_MAX_RX_CHAN) {
               status = ar_set_device_config (board, rxChanConfig[channelIndex].channel, ARU_RX_BITRATE, rxChanConfig[channelIndex].bitRate);
               if (status == ARS_NORMAL) {
                  status = ar_set_device_config (board, rxChanConfig[channelIndex].channel, ARU_RX_PARITY, rxChanConfig[channelIndex].parity);
               }
               if (status == ARS_NORMAL) {
                  status = ar_set_device_config (board, rxChanConfig[channelIndex].channel, ARU_RX_MERGED_MODE, rxChanConfig[channelIndex].bufferMode);
               }
               if (status == ARS_NORMAL) {
                  status = ar_set_device_config (board, rxChanConfig[channelIndex].channel, ARU_RECV_MODE, rxChanConfig[channelIndex].internalWrap);
               }
               if (status == ARS_NORMAL) {
                  status = ar_set_device_config (board, rxChanConfig[channelIndex].channel, ARU_RX_FIFO_ENABLE, rxChanConfig[channelIndex].enable);
               }
            }
         }
      }
   }
   return status;
}


#ifdef INCLUDE_MSG_SCHEDULER
/*===========================================================================*
 * ENTRY:  A R _ D E F I N E _ M E S S A G E S _ F R O M _ T X T _ F I L E        
 *===========================================================================*
 *
 * FUNCTION:    This routine definess a scheduled message set from file
 *              
 * PARAMETERS:  pCEI_CHAR cfgFileName (input) channel configuration file name
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_BADLOAD     File not found
 *                                 ARS_INVHARCMD   Invalid File entry
 *
 *              from ar_define_msg_block:
 *
 *                         ARS_NORMAL      Success                                  
 *                         ARS_INVBOARD    Invalid or uninitialized board           
 *                         ARS_INVHARCMD   Invalid transmit channel entry           
 *                         ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                         ARS_FAILURE     The number of entries requested          
 *                                         exceeds the available number             
 *                                         of remaining table entries               
 * GLOBALS:     none
 * 
 * DESCRIPTION: This routine reads the contents of the scheduled message
 *              definition text file (ASCII only format) and defines the
 *              message scenario in the specified CEI-x30 board's message
 *              scheduler table.
 * 
 *===========================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_define_messages_from_txt_file
(
   pCEI_CHAR cfgFileName          /* (input) channel configuration file name */
)
{
   CDEV_API_RET_TYPE status = ARS_NORMAL;
   CEI_CHAR fileEntry[160];
   FILE * inputConfigFilePtr;
   CONFIG_FILE_SECTION_TYPE whereAmI;
   CEI_INT32 messageCount = 0;
   CEI_INT32 messageIndex;
   CEI_INT32 j;
   pCEI_CHAR toss;
   pCEI_CHAR tempStrPtr;
   CEI_CHAR tempEntry[160];
   
   if ( (inputConfigFilePtr = fopen(cfgFileName, "rb")) == NULL ) {
      status = ARS_BADLOAD;
   }
   else {
      whereAmI = COUNT;
      while ((fgets(fileEntry, 159, inputConfigFilePtr) != NULL) && (whereAmI != DONE)) {
         switch (whereAmI) {
            case COUNT:
               if (strncmp (fileEntry, MSG_COUNT_TAG, strlen(MSG_COUNT_TAG)) == 0) {
                  toss = strtok (fileEntry, ":");
                  tempStrPtr = strtok (NULL, "");
                  messageCount = atoi (tempStrPtr);

                  if (messageCount > 1000) {
                     return ARS_INVHARCMD;
                  }
                  whereAmI = QUERY;
               }
               else {
                  status = ARS_INVHARCMD;
                  whereAmI = DONE;
               }
               break;

            case QUERY:
               if (strncmp (fileEntry, MSG_GRP_START_TAG, strlen(MSG_GRP_START_TAG)) == 0) {
                  whereAmI = MSGDEF;
               }
               else {
                  whereAmI = DONE;
                  status = ARS_INVHARCMD;
               }
               break;

            case MSGDEF:
               /* Scheduled Message Definition Entries in the define messages file look like this:
               
                  MessageIndex:0___BoardIndex:0___ChannelIndex:2___MessageContext:12345678___Rate:200___StartOffset:5___TransmitCount:-1___EquipmentID:0
                  MessageIndex:1___BoardIndex:0___ChannelIndex:0___MessageContext:0x56cc04___Rate:50___StartOffset:0___TransmitCount:-1___EquipmentID:1
                  MessageIndex:2___BoardIndex:0___ChannelIndex:1___MessageContext:0xc8640a___Rate:20___StartOffset:0___TransmitCount:-1___EquipmentID:4
                  
                  The combined use of strtok (NULL, ":") and strtok (NULL, "_") parses only the scheduled message attributes from each entry,
                  so it is always assumed the format of the channel attribute string will be fully populated using the order and attribute
                  format demonstrated above.  If a scheduled message file entry deviates from what is shown above, a failure will occur.
               */
               toss = strtok (fileEntry, TOKEN_VALUE_FIELD_DELIMITER);
               if ((toss != NULL) && (strncmp (toss, MSG_INDEX_TAG, strlen(MSG_INDEX_TAG)) == 0)) {
                  tempStrPtr = strtok (NULL, ATTRIBUTE_ENTRY_DELIMITER);
                  
                  /* assure the message index string is the first item in this entry and it is a number */
                  if (True == ar_is_number(tempStrPtr)) {
                     messageIndex = atoi (tempStrPtr);

                     /* assure the message index is valid */
                     if (messageIndex < messageCount) {
                        tempStrPtr = strtok (NULL, TOKEN_VALUE_FIELD_DELIMITER);
                        if (tempStrPtr == NULL)
                           break;
                        while (tempStrPtr[0] == ATTRIBUTE_ENTRY_DELIMITER_CHAR)
                           tempStrPtr++;
   
                        /* extract the board number */
                        if (strncmp (tempStrPtr, BOARD_INDEX_TAG, strlen(BOARD_INDEX_TAG)) == 0) {
                           tempStrPtr = strtok (NULL, ATTRIBUTE_ENTRY_DELIMITER);
                           if ( (True == ar_is_number(tempStrPtr)) &&
                                (atoi(tempStrPtr) >= 0)            &&
                                (atoi(tempStrPtr) <= CEIDEV_MAX_BRD_IDX) ) {
                              scheduledMessages[messageIndex].board = (CDEV_PARM_SSI_TYPE)atoi(tempStrPtr);   
                              scheduledMessages[messageIndex].messageIndex = 0;         /* zero the message index   */
                              scheduledMessages[messageIndex].data = 0;                 /* allow no message value, start offset or transmit count entries */
                              scheduledMessages[messageIndex].start = 0;
                              scheduledMessages[messageIndex].rate = 0;
                              scheduledMessages[messageIndex].txCount = (CEI_UINT32)-1;  /* -1 is "infinite transmission" */
                           }
                           else {
                              whereAmI = CFG_ERR;
                              break;
                           }
                        
                           tempStrPtr = strtok (NULL, TOKEN_VALUE_FIELD_DELIMITER);
                           if (tempStrPtr == NULL)
                              break;
                           while (tempStrPtr[0] == ATTRIBUTE_ENTRY_DELIMITER_CHAR)
                              tempStrPtr++;
                        }
                        else {
                           whereAmI = CFG_ERR;
                           break;
                        }
                        
                        /* extract the transmit channel number, assuring it is the next item in this entry and is a number */
                        if (strncmp (tempStrPtr, CH_INDEX_TAG, strlen(CH_INDEX_TAG)) == 0) {
                           tempStrPtr = strtok (NULL, ATTRIBUTE_ENTRY_DELIMITER);
                           if ( (True == ar_is_number(tempStrPtr)) && 
                                (atoi(tempStrPtr) >=0 ) && 
                                (atoi(tempStrPtr) < CEIDEV_MAX_TX_CHAN) ) 
                              scheduledMessages[messageIndex].channel = (CDEV_PARM_SSI_TYPE)atoi(tempStrPtr); 
                           else {
                              return ARS_INVHARVAL;
                           }
                        
                           tempStrPtr = strtok (NULL, TOKEN_VALUE_FIELD_DELIMITER);
                           if (tempStrPtr == NULL)
                              break;
                           while (tempStrPtr[0] == ATTRIBUTE_ENTRY_DELIMITER_CHAR)
                              tempStrPtr++;
                        }
                        else {
                           whereAmI = CFG_ERR;
                           break;
                        }
                           
                        /* extract the raw message value */ 
                        if (strncmp (tempStrPtr, MSG_VALUE_TAG, strlen(MSG_VALUE_TAG)) == 0) {
                           tempStrPtr = strtok (NULL, ATTRIBUTE_ENTRY_DELIMITER);
                           if ( (tempStrPtr[0] == '0') && 
                                ((tempStrPtr[1] == 'x') || (tempStrPtr[1] == 'X')) &&
                                (strlen(tempStrPtr) > 2) && 
                                (strlen(tempStrPtr) <= 10) ) {
                              /*Must be 0~9 or a~f or A~F*/
                              for (j = 2;j < (CEI_INT32)strlen(tempStrPtr);j++) {
                                 if ( ((tempStrPtr[j] >= '0') && (tempStrPtr[j] <= '9')) ||
                                      ((tempStrPtr[j] >= 'a') && (tempStrPtr[j] <= 'f')) || 
                                      ((tempStrPtr[j] >= 'A') && (tempStrPtr[j] <= 'F')) ) {
                                    status = ARS_NORMAL;
                                 }
                                 else {
                                    return ARS_INVHARCMD;
                                 }
                              }
                              sscanf (tempStrPtr, "%x", &scheduledMessages[messageIndex].data);
                           }
                           else {
                              whereAmI = CFG_ERR;
                              break;
                           }
                        
                           tempStrPtr = strtok (NULL, TOKEN_VALUE_FIELD_DELIMITER);
                           if (tempStrPtr == NULL)
                              break;
                           while (tempStrPtr[0] == ATTRIBUTE_ENTRY_DELIMITER_CHAR)
                              tempStrPtr++;
                        }
                        else {
                           whereAmI = CFG_ERR;
                           break;
                        }
                           
                        /* extract the message rate */
                        if (strncmp (tempStrPtr, MSG_RATE_TAG, strlen(MSG_RATE_TAG)) == 0) {
                           tempStrPtr = strtok (NULL, ATTRIBUTE_ENTRY_DELIMITER);
                           if (True == ar_is_32_ucint(tempStrPtr)) {
                              scheduledMessages[messageIndex].rate = (CDEV_PARM_SSI_TYPE)atoi(tempStrPtr);
                           }
                           else {
                              whereAmI = CFG_ERR;
                              break;
                           }

                           tempStrPtr = strtok (NULL, TOKEN_VALUE_FIELD_DELIMITER);
                           if (tempStrPtr == NULL)
                              break;
                           while (tempStrPtr[0] == ATTRIBUTE_ENTRY_DELIMITER_CHAR)
                              tempStrPtr++;
                        }
                        else {
                           whereAmI = CFG_ERR;
                           break;
                        }
                           
                        /* extract the starting offset value */
                        if (strncmp (tempStrPtr, MSG_OFFSET_TAG, strlen(MSG_OFFSET_TAG)) == 0) {
                           tempStrPtr = strtok (NULL, ATTRIBUTE_ENTRY_DELIMITER);
                           if (True == ar_is_32_ucint(tempStrPtr)) {
                              scheduledMessages[messageIndex].start = (CDEV_PARM_SSI_TYPE)atoi(tempStrPtr);
                           }
                           else {
                              whereAmI = CFG_ERR;
                              break;
                           }
                           
                           tempStrPtr = strtok (NULL, TOKEN_VALUE_FIELD_DELIMITER);
                           if (tempStrPtr == NULL)
                              break;
                           while (tempStrPtr[0] == ATTRIBUTE_ENTRY_DELIMITER_CHAR)
                              tempStrPtr++;
                        }
                              
                        /* extract the transmit count */
                        if (strncmp (tempStrPtr, MSG_TXCOUNT_TAG, strlen(MSG_TXCOUNT_TAG)) == 0) {
                           tempStrPtr = strtok (NULL, EQUIP_ID_FIELD_START_CHAR);                               
                           if (((char)tempStrPtr[0] == TRANSMIT_CONTINUOUS_START_CHAR) && ((char)tempStrPtr[1] == '1')) {
                              if ('_' != tempStrPtr[2]) {
                                 return ARS_INVHARCMD;
                              }
                              scheduledMessages[messageIndex].txCount = (CEI_UINT32)-1;   /* -1 is "infinite transmission" */
                           }
                           else {
                              j = 0;
                              while (j < 10) {
                                 if ((isdigit(tempStrPtr[j]) == FALSE ) || (j == 9)) {  /* extract limited transmit count */
                                    tempEntry[j] = '\0';
                                    scheduledMessages[messageIndex].txCount = (CDEV_PARM_SSI_TYPE)atoi(tempEntry); 
                                    break;
                                 }
                                 else {
                                    tempEntry[j] = tempStrPtr[j];
                                    j++;
                                 }
                              }
                           }

                           tempStrPtr = strtok (NULL, TOKEN_VALUE_FIELD_DELIMITER);
                           if (tempStrPtr == NULL)
                              break;
                           while (tempStrPtr[0] == ATTRIBUTE_ENTRY_DELIMITER_CHAR)
                              tempStrPtr++;
                        }

                        /* The attribute "EquipmentID" is not used */
                        if (strncmp (tempStrPtr, MSG_EQ_ID, strlen(MSG_EQ_ID)) == 0) {
                           tempStrPtr = strtok (NULL, TOKEN_VALUE_FIELD_DELIMITER);
                           tempStrPtr++;
                           if (tempStrPtr[0] == (char)0x0d)
                              break;
                           while (tempStrPtr[0] == ATTRIBUTE_ENTRY_DELIMITER_CHAR)
                              tempStrPtr++;
                        }  
                        else {
                           break;
                        }
                     }   /* if (messageIndex < messageCount)  */
                     else {
                        whereAmI = CFG_ERR;
                     }

                     if (scheduledMessages[messageIndex].rate == 0) {
                        whereAmI = CFG_ERR;
                     }

                  }  /* True == ar_is_number(tempStrPtr) */
                  else if (strncmp (toss, MSG_GRP_END_TAG, strlen(MSG_GRP_END_TAG)) == 0) {
                     whereAmI = DONE;
                  }
               }
               break;

            case CFG_ERR:
               status = ARS_INVHARCMD;
               whereAmI = DONE;
               break;

            default:
               whereAmI = DONE;
               break;
         }
      }
      fclose (inputConfigFilePtr);

      /*  Now update the specified board's message scheduler table(s) with what was read from the definition file. */
      if ( (status == ARS_NORMAL) && (messageCount > 0) ) {
         status = ar_define_msg_block (messageCount, scheduledMessages);
      }
   }
   return status;
}
#endif


/*===========================================================================*
 * ENTRY POINT:   A R _ CHANNEL _ CONFIGURATION _ F R O M _ XML _ F I L E        
 *===========================================================================*
 *
 * FUNCTION:    This routine definess a channel configuration set from XML file
 *              
 * PARAMETERS:  pCEI_CHAR pcCfgFileName (input) channel configuration file name
 *              CDEV_BOARD_TYPE cintBrdIdx  (input) board index
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_BADLOAD     File not found
 *                                 ARS_INVHARCMD   Invalid File entry
 *
 *              from ar_set_device_config:
 *
 *                         ARS_INVARG      "item" parameter value is not
 *                                          supported by this function
 *                         ARS_INVHARVAL   "value" was invalid or channel
 *                                          not supported on this device
 *                         ARS_INVBOARD    Board parameter value invalid or
 *                                         board has not been initialized
 *                         ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed               
 * GLOBALS:     none
 * 
 * DESCRIPTION: This routine reads the contents of the channel configuration
 *              definition XML file and defines the message scenario in the 
 *              specified CEI-x30 board's message scheduler table.
 * 
 *===========================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_channel_configuration_from_xml_file
(
   pCEI_CHAR pcCfgFileName, /* (input) channel configuration file name */
   CDEV_BOARD_TYPE cintBrdIdx   /* (input) board index */
)
{
   CDEV_API_RET_TYPE cintStatus = ARS_NORMAL;
   CDEV_BOARD_TYPE cintBoardIndexTemp;
   CDEV_CHAN_TYPE cintRxChNum;
   CDEV_CHAN_TYPE cintTxChNum;
   FILE      *fConfigFilePtr;
   XMLSEC    *pHead = NULL;
   RX_CHANNEL_PARMS_TYPE rxChanConfig[CEIDEV_MAX_RX_CHAN];
   TX_CHANNEL_PARMS_TYPE txChanConfig[CEIDEV_MAX_TX_CHAN];
 
   /*Read channel configuration XML file*/  
   if (NULL == (fConfigFilePtr = fopen(pcCfgFileName, "r"))) {
      cintStatus = ARS_BADLOAD;
   }
   /*If borad index value as input is valid, use it to configure hardware*/
   else if ((cintBrdIdx < 0) || (cintBrdIdx > CEIDEV_MAX_BRD_IDX)) {
      return ARS_INVBOARD;
   }
   else {
      cintBoardIndexTemp = cintBrdIdx;
      cintStatus = ar_get_device_config (cintBrdIdx, 0, ARU_CONFIGURATION, &configuration);
      configuration &= CEIDEV_CONFIGURATION_MASK;

      /*Initialize channel configuration settings to default values*/
      ar_initialize_channel_configuration_struct(CEIDEV_MAX_RX_CHAN,
                                                 CEIDEV_MAX_TX_CHAN,
                                                 rxChanConfig,
                                                 txChanConfig);

      /*Allocate the head node*/
      if (NULL == (pHead = (XMLSEC*)calloc(1, sizeof(XMLSEC)))) {
         fclose(fConfigFilePtr);
         return ARS_FAILURE;
      }   

      /*Initialize head node*/
      pHead->cintLength = 0;
      pHead->pNext     = NULL;

      /*Get XML file items*/
      if (ARS_NORMAL != (cintStatus = ar_get_item_from_xml_file(fConfigFilePtr, pHead))) {
         fclose(fConfigFilePtr);
         ar_release_xmlsec_list(pHead);
         return cintStatus;
      }

      /*Get channel configuration parameters*/
      if (ARS_NORMAL != (cintStatus = ar_channel_config_from_xml_list(pHead,
                                                                        &cintBoardIndexTemp,
                                                                        &cintRxChNum,
                                                                        &cintTxChNum,
                                                                        rxChanConfig,
                                                                        CEIDEV_MAX_RX_CHAN,
                                                                        txChanConfig,
                                                                        CEIDEV_MAX_RX_CHAN))) {
         fclose(fConfigFilePtr);
         ar_release_xmlsec_list(pHead);
         return cintStatus;
      }

      fclose(fConfigFilePtr);
      ar_release_xmlsec_list(pHead);

      /*Load the settings to hardware*/
      cintStatus = ar_set_channel_configuration(cintBoardIndexTemp,
                                                  cintRxChNum,
                                                  cintTxChNum,
                                                  rxChanConfig,
                                                  txChanConfig);
   }
 
   return cintStatus;
}


/*===========================================================================*
 * ENTRY POINT:  AR_INITIALIZE_CHANNEL_CONFIGURATION_STRUCT       
 *===========================================================================*
 *
 * FUNCTION:    This routine initialize RX channel and TX channels' settings
 *              
 * PARAMETERS:  CDEV_CHAN_TYPE cintRxChNum(input) Rx channel number
 *              CDEV_CHAN_TYPE cintTxChNum(input) Tx channel number
 *              RX_CHANNEL_PARMS_TYPE *prxChanConfig(input) Rx channels settings
 *              TX_CHANNEL_PARMS_TYPE *ptxChanConfig(input) Tx channels settings
 *
 * RETURN VAL:               
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CEI_VOID ar_initialize_channel_configuration_struct
(
   CDEV_CHAN_TYPE cintRxChNum,
   CDEV_CHAN_TYPE cintTxChNum,
   RX_CHANNEL_PARMS_TYPE *prxChanConfig,
   TX_CHANNEL_PARMS_TYPE *ptxChanConfig
)
{
   CDEV_CHAN_TYPE chanIndex;

   for (chanIndex = 0;chanIndex < cintRxChNum; chanIndex++) {
      prxChanConfig[chanIndex].channel      = CEIDEV_MAX_RX_CHAN;
      prxChanConfig[chanIndex].enable       = AR_OFF;
      prxChanConfig[chanIndex].bitRate      = ARU_SPEED_HIGH;
      prxChanConfig[chanIndex].parity       = AR_OFF;
      prxChanConfig[chanIndex].bufferMode   = AR_OFF;
      prxChanConfig[chanIndex].internalWrap = AR_WRAP_OFF;
   }

   for (chanIndex = 0;chanIndex < cintTxChNum;chanIndex++) {
      ptxChanConfig[chanIndex].channel  = CEIDEV_MAX_RX_CHAN;
      ptxChanConfig[chanIndex].enable   = AR_OFF;
      ptxChanConfig[chanIndex].bitRate  = ARU_SPEED_HIGH;
      ptxChanConfig[chanIndex].parity   = ARU_PARITY_ODD;
      ptxChanConfig[chanIndex].bitError = AR_OFF;
      ptxChanConfig[chanIndex].gapError = AR_OFF;
   }
}


/*===========================================================================*
 * ENTRY POINT:   AR_SET_CHANNEL_CONFIGURATION     
 *===========================================================================*
 *
 * FUNCTION:    This routine downloads settings to hardware
 *              
 * PARAMETERS:  CDEV_BOARD_TYPE cintBrdIdx,
 *              CDEV_CHAN_TYPE cintRxChNum,
 *              CDEV_CHAN_TYPE cintTxChNum,
 *              RX_CHANNEL_PARMS_TYPE *prxChanConfig,
 *              TX_CHANNEL_PARMS_TYPE *ptxChanConfig
 *
 * RETURN VAL:  CDEV_API_RET_TYPE
 *                         ARS_NORMAL      Success
 *                         ARS_BADLOAD     File not found
 *                         ARS_INVHARCMD   Invalid File entry
 *
 *
 *                         ARS_INVARG      "item" parameter value is not
 *                                          supported by this function
 *                         ARS_INVHARVAL   "value" was invalid or channel
 *                                          not supported on this device
 *                         ARS_INVBOARD    Board parameter value invalid or
 *                                         board has not been initialized
 *                         ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed               
 * GLOBALS:     none
 * 
 * DESCRIPTION: 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_set_channel_configuration
(
   CDEV_BOARD_TYPE cintBrdIdx,
   CDEV_CHAN_TYPE cintRxChNum,
   CDEV_CHAN_TYPE cintTxChNum,
   RX_CHANNEL_PARMS_TYPE *prxChanConfig,
   TX_CHANNEL_PARMS_TYPE *ptxChanConfig
)
{
   CDEV_API_RET_TYPE status = ARS_NORMAL;
   CDEV_CHAN_TYPE chanIndex;

   for (chanIndex = 0; (chanIndex < cintTxChNum) &&  (status == ARS_NORMAL); chanIndex++) {
      if (ptxChanConfig[chanIndex].channel < CEIDEV_MAX_RX_CHAN) {
         status = ar_set_device_config (cintBrdIdx, ptxChanConfig[chanIndex].channel, ARU_TX_BITRATE, ptxChanConfig[chanIndex].bitRate);

         if (status == ARS_NORMAL) {
            status = ar_set_device_config (cintBrdIdx, ptxChanConfig[chanIndex].channel, ARU_TX_PARITY, ptxChanConfig[chanIndex].parity);
         }
         if (status == ARS_NORMAL) {
            status = ar_set_device_config (cintBrdIdx, ptxChanConfig[chanIndex].channel, ARU_TX_BIT_ERROR, ptxChanConfig[chanIndex].bitError);
         }
         if (status == ARS_NORMAL) {
            status = ar_set_device_config (cintBrdIdx, ptxChanConfig[chanIndex].channel, ARU_TX_GAP_ERROR, ptxChanConfig[chanIndex].gapError);
         }
         if (status == ARS_NORMAL) {
            status = ar_set_device_config (cintBrdIdx, ptxChanConfig[chanIndex].channel, ARU_TX_FIFO_ENABLE, ptxChanConfig[chanIndex].enable);
         }        
      }
   }

   for (chanIndex = 0; (chanIndex < cintRxChNum) &&  (status == ARS_NORMAL); chanIndex++)
   {
      if (prxChanConfig[chanIndex].channel < CEIDEV_MAX_RX_CHAN) {
         status = ar_set_device_config (cintBrdIdx, prxChanConfig[chanIndex].channel, ARU_RX_BITRATE, prxChanConfig[chanIndex].bitRate);
         
         if (status == ARS_NORMAL)  {
            status = ar_set_device_config (cintBrdIdx, prxChanConfig[chanIndex].channel, ARU_RX_PARITY, prxChanConfig[chanIndex].parity);
         }
         if (status == ARS_NORMAL) {
            status = ar_set_device_config (cintBrdIdx, prxChanConfig[chanIndex].channel, ARU_RX_MERGED_MODE, prxChanConfig[chanIndex].bufferMode);
         }
         if (status == ARS_NORMAL)  {
            status = ar_set_device_config (cintBrdIdx, prxChanConfig[chanIndex].channel, ARU_RECV_MODE, prxChanConfig[chanIndex].internalWrap);
         }
         if (status == ARS_NORMAL)  {
            status = ar_set_device_config (cintBrdIdx, prxChanConfig[chanIndex].channel, ARU_RX_FIFO_ENABLE, prxChanConfig[chanIndex].enable);
         }
      }
   }
   return status;
}


/*===========================================================================*
 * ENTRY POINT:  AR_GET_ITEM_FROM_XML_FILE
 *===========================================================================*
 *
 * FUNCTION:    Read back from channel configuration XML file, and save items
 *              in a list structure
 *
 * PARAMETERS:  FILE      *fConfigFilePtr(input) XML file handle
 *              XMLSEC    *pHead(input) XML item list
 *
 * RETURN VAL:  CDEV_PARM_SSI_TYPE ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry             
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_get_item_from_xml_file
(
   FILE      *fConfigFilePtr, /*The xml file to get items back, calling funtion in charge of valid file handle*/
   XMLSEC    *pHead           /*The list to store the handles, calling function in charge of valid file*/
)
{
   CEI_INT32 cint32Cur;
   CDEV_PARM_SSI_TYPE cintNewSec      = False;
   CDEV_PARM_SSI_TYPE cintStart       = False;
   CDEV_PARM_SSI_TYPE cintLastBracket = NO_BRACKET;
   XMLSEC    *pCur             = NULL;
   XMLSEC    *pTemp            = NULL;

   /*Save the tags and values of XML elements into the XMLSEC list*/
   while (EOF != (cint32Cur = getc(fConfigFilePtr))) {
      if (LEFT_BRACKET_DELIMITER == cint32Cur) {
         /*Format check*/
         if (LEFT_BRACKET == cintLastBracket) {
            return ARS_INVHARCMD;
         }

         /*To start decoding*/
         if (False == cintStart) {
            /*Head node as the current node*/
            pCur = pHead;
            cintStart = True;
         }

         cintLastBracket = LEFT_BRACKET;
         cintNewSec = True;
      }
      else if (RIGHT_BRACKET_DELIMITER == cint32Cur) {
         /*Format check*/
         if (RIGHT_BRACKET == cintLastBracket) {
            return ARS_INVHARCMD;
         }           
         cintLastBracket = RIGHT_BRACKET;
         cintNewSec = True;
      }
      else {
         /*Add current character to current item node*/
         cintNewSec = False;
      }

      if (True == cintNewSec) {
         if (NULL == (pTemp = (XMLSEC*)calloc(1, sizeof(XMLSEC)))) {
            return ARS_FAILURE;
         }

           /*Add end mark*/
         if (128 > pCur->cintLength) {
            pCur->accTemp[pCur->cintLength] = '\0';
         }

         /*Add the new item node to the list*/
         pCur->pNext = pTemp;
         pCur = pTemp;
         pCur->pNext = NULL;

         /*Initialize new item node length*/
         pCur->cintLength = 0;
      }
      else {
         /* 0 ,NUT, 10, new line, 255, null, 13 CR, 32, space, if such items,jump to the next character*/
         if ((0 == cint32Cur) || (10 == cint32Cur) || (255 == cint32Cur) || (13 == cint32Cur) || 
             (False == cintStart) || (NULL == pCur) || (127 <= pCur->cintLength) || 
             ((32 == cint32Cur) && (LEFT_BRACKET != cintLastBracket) && (0 == pCur->cintLength))) {
            continue;
         }
         else {
            /*Transfer from ascii code to character*/
            CEI_CHAR accTemp[2];
            sprintf(accTemp,"%c",cint32Cur);

            pCur->accTemp[pCur->cintLength] = accTemp[0];
            pCur->cintLength++;
         }
      }
   }
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:  AR_RELEASE_XMLSEC_LIST
 *===========================================================================*
 *
 * FUNCTION:    Release XML item list
 *
 * PARAMETERS:  XMLSEC *pHead(input) XML item list
 *
 * RETURN VAL:          
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
void ar_release_xmlsec_list(XMLSEC *pHead)
{
   XMLSEC *pTemp;
   XMLSEC *pCur = pHead;

   while (NULL != pCur) {
      pTemp = pCur;
      pCur  = pCur->pNext;

      /*Relase the last XMLSEC node*/
      free(pTemp);
   }
}


/*===========================================================================*
 * ENTRY POINT:  AR_CHANNEL_CONFIG_FROM_XML_LIST
 *===========================================================================*
 *
 * FUNCTION:    Decode item information into channel configuration struct
 *
 * PARAMETERS:  XMLSEC    *pHead(input) XML item list
 *              CDEV_API_RET_TYPE *pcintBoardIndex(output)  return value of board index
 *              CDEV_CHAN_TYPE *pcintRxChNum(output) return value of Rx channel number
 *              CDEV_CHAN_TYPE *pcintTxChNum(output) return value of Tx channel number
 *              RX_CHANNEL_PARMS_TYPE *prxChanConfig(output) return settings of Rx channels
 *              CDEV_PARM_SSI_TYPE             cintRxChConfigSize(input) Rx channel struct size
 *              TX_CHANNEL_PARMS_TYPE *ptxChanConfig(output) return settings of Tx channels
 *              CDEV_PARM_SSI_TYPE             cintTxChConfigSize(input) Tx channel struct size
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry             
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_channel_config_from_xml_list
(
   XMLSEC    *pHead,
   CDEV_BOARD_TYPE *pcintBoardIndex,
   CDEV_CHAN_TYPE *pcintRxChNum,
   CDEV_CHAN_TYPE *pcintTxChNum,
   RX_CHANNEL_PARMS_TYPE *prxChanConfig,
   CDEV_PARM_SSI_TYPE cintRxChConfigSize,
   TX_CHANNEL_PARMS_TYPE *ptxChanConfig,
   CDEV_PARM_SSI_TYPE cintTxChConfigSize
)
{
   XMLSEC    *pCur = NULL;
   CDEV_PARM_SSI_TYPE cintIndex;
   CDEV_PARM_SSI_TYPE cintIndexSec;
   CDEV_PARM_SSI_TYPE cintCurrentItemType = CCST;
   CDEV_CHAN_TYPE cintCurRxChIndex = 0;
   CDEV_CHAN_TYPE cintCurTxChIndex = 0;
   CDEV_CHAN_TYPE cintRxTx         = RX;
   CDEV_PARM_SSI_TYPE cintReturn   = 0;
   CDEV_PARM_SSI_TYPE cintEnd      = False;
   CDEV_PARM_SSI_TYPE cintStatus;
   CDEV_PARM_SSI_TYPE cintLength;

   *(CDEV_PARM_SSI_TYPE*)pcintRxChNum = CEIDEV_MAX_RX_CHAN;
   *(CDEV_PARM_SSI_TYPE*)pcintTxChNum = CEIDEV_MAX_TX_CHAN;

   /*Check input validity*/
   if ( (ARS_NORMAL != ar_xml_check_title(pHead, &pCur)) ||
        (ARS_NORMAL != ar_check_cc_xml_list_input(pHead, pcintBoardIndex, pcintRxChNum, pcintTxChNum, prxChanConfig, ptxChanConfig)) ) {
      return ARS_INVHARCMD;
   }  

   if (NULL != pCur) {
      pCur = pCur->pNext;
   }
   else {
      return ARS_INVHARCMD;
   }

   /*Go through the xml item list*/
   while (NULL != pCur) {
      /*If empty node, skip*/
      if (0 != pCur->cintLength) {       
         /*If element name, check whether match*/
         if (True == ar_xml_get_cc_item_name_idx(cintCurrentItemType, &cintIndex)) {
            cintLength = (CDEV_PARM_SSI_TYPE)strlen(XML_CC_ITEMNAME_TAB[cintIndex]);
            
            if (cintLength > pCur->cintLength) {
               return ARS_INVHARCMD;
            }
            else {
               for (cintIndexSec = 0;cintIndexSec < cintLength;cintIndexSec++) {
                  if (XML_CC_ITEMNAME_TAB[cintIndex][cintIndexSec] != pCur->accTemp[cintIndexSec]) {
                     return ARS_INVHARCMD;
                  }
               }

               if (cintLength < pCur->cintLength) {
                  if (pCur->accTemp[cintLength] != ' ') {
                     return ARS_INVHARCMD;
                  }
               }
            }
            if (RXCHED == cintCurrentItemType) {
               cintCurRxChIndex++;
            }
            if (TXCHED == cintCurrentItemType) {
               cintCurTxChIndex++;
            }
            if (RXGRPED == cintCurrentItemType) {
               cintRxTx = TX;
            }
         }
         /*If element value, check range and (map string to CDEV_PARM_SSI_TYPE value, of hardware)*/
         else {
            if (BRDIDXVAL == cintCurrentItemType) {
               *(CDEV_BOARD_TYPE*)pcintBoardIndex = (CDEV_BOARD_TYPE)atoi(pCur->accTemp);
               if ((False == ar_is_number(pCur->accTemp)) || (*(CDEV_BOARD_TYPE*)pcintBoardIndex < 0) || (*(CDEV_BOARD_TYPE*)pcintBoardIndex > CEIDEV_MAX_BRD_IDX)) {
                  return ARS_INVHARCMD;
               }
            }
            else if (True == ar_xml_get_cc_val_item_idx(cintCurrentItemType, &cintIndex)) {
               if (ARS_NORMAL != (cintStatus = XML_CC_VAL_FUN_TAB[cintIndex](pCur->accTemp, &cintReturn))) {
                  return cintStatus;
               }

               if (ARS_NORMAL != ar_xml_cc_save_config(cintCurrentItemType,
                                                       pcintRxChNum,
                                                       pcintTxChNum,
                                                       cintCurRxChIndex,
                                                       cintCurTxChIndex,
                                                       cintRxTx,
                                                       prxChanConfig,
                                                       cintRxChConfigSize,
                                                       ptxChanConfig,
                                                       cintTxChConfigSize,
                                                       cintReturn)) {
                  return ARS_INVHARCMD;
               }
               if (0 == *(CDEV_PARM_SSI_TYPE*)pcintRxChNum) {
                  cintRxTx = TX;
               }
            }
            else {
               return ARS_INVHARCMD;
            }
         }

         /*Get next item type*/
          if (XML_CC_ITEM_END == (cintCurrentItemType = ar_channel_config_xml_next_item_type(cintCurrentItemType,
                                                                                           cintCurRxChIndex,
                                                                                           *(CDEV_PARM_SSI_TYPE*)pcintRxChNum,
                                                                                           cintCurTxChIndex,
                                                                                           *(CDEV_PARM_SSI_TYPE*)pcintTxChNum,
                                                                                           cintRxTx))) {
            return ARS_INVHARCMD;
         }
         /*Mark valid end*/
         else if (CCED == cintCurrentItemType) {
            cintEnd = True;
            break;
         }
      }  

      /*Get next item*/
      pCur = pCur->pNext;
   }

   /*Check proper end*/
   if (False == cintEnd) {
      return ARS_INVHARCMD;
   }
   else {
      return ARS_NORMAL;
   }
}


/*===========================================================================*
 * ENTRY POINT:  AR_CHECK_CC_XML_LIST_INPUT
 *===========================================================================*
 *
 * FUNCTION:    Guarantee no NULL pointers
 *
 * PARAMETERS:  XMLSEC    *pHead(input) XML item list
 *              CDEV_BOARD_TYPE *pcintBoardIndex(output)  return value of board index
 *              CDEV_CHAN_TYPE *pcintRxChNum(output) return value of Rx channel number
 *              CDEV_CHAN_TYPE *pcintTxChNum(output) return value of Tx channel number
 *              RX_CHANNEL_PARMS_TYPE *prxChanConfig(output) return settings of Rx channels
 *              TX_CHANNEL_PARMS_TYPE *ptxChanConfig(output) return settings of Tx channels
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_FAILURE     Failed         
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_check_cc_xml_list_input(XMLSEC    *pHead,
                                     CDEV_BOARD_TYPE *pcintBoardIndexTemp,
                                     CDEV_CHAN_TYPE *pcintRxChNum,
                                     CDEV_CHAN_TYPE *pcintTxChNum,
                                     RX_CHANNEL_PARMS_TYPE *prxChanConfig,
                                     TX_CHANNEL_PARMS_TYPE *ptxChanConfig)
{
   /*Check validity of inputs*/
   if ((NULL == pHead)             ||
       (NULL == pcintBoardIndexTemp) ||
       (NULL == pcintRxChNum)    ||
       (NULL == pcintTxChNum)    ||
       (NULL == prxChanConfig)     ||
       (NULL == ptxChanConfig)) {
      return ARS_FAILURE;
   }
   else {
      return ARS_NORMAL;
   }
}


/*===========================================================================*
 * ENTRY POINT:  AR_XML_CHECK_TITLE
 *===========================================================================*
 *
 * FUNCTION:    Guarantee corret XML file title
 *
 * PARAMETERS:  XMLSEC *pHead(input) XML item list
 *              XMLSEC **pReturn(output) pointer of XML file title item
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry         
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_xml_check_title(XMLSEC *pHead, XMLSEC **pReturn)
{
   XMLSEC *pCur = pHead;

   /*Get XML title, format check*/
   if (NULL == pCur) {
      return ARS_INVHARCMD;
   }

   /*Find first valid item node*/
   while (NULL != pCur) {
      if (0 != pCur->cintLength) {
         break;
      }
      pCur = pCur->pNext;
   }

   if (NULL == pCur) {
      *pReturn  = NULL;
      return ARS_INVHARCMD;
   }

   if ('?' != pCur->accTemp[0]) {
      *pReturn  = NULL;
      return ARS_INVHARCMD;
   }
   else if ('?' != pCur->accTemp[(pCur->cintLength) - 1]) {
      *pReturn  = NULL;
      return ARS_INVHARCMD;
   }
   else {
      *pReturn = pCur;
      return ARS_NORMAL;
   }
}


/*===========================================================================*
 * ENTRY POINT:  AR_XML_GET_CC_ITEM_NAME_IDX
 *===========================================================================*
 *
 * FUNCTION:    Find channel configuration XML element tag item position
 *              in XML_CC_ITEMNAME_IDX_TAB 
 *
 * PARAMETERS:  CDEV_PARM_SSI_TYPE cintCurrentItemType(input) current item type
 *              CDEV_PARM_SSI_TYPE *pcintIndex(output) position of item
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  True item found 
 *                                 False item not found        
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_xml_get_cc_item_name_idx(CDEV_PARM_SSI_TYPE cintCurrentItemType,CDEV_PARM_SSI_TYPE *pcintIndex)
{
   CDEV_PARM_SSI_TYPE cintIndex;

   /*To check whether element name*/
   for (cintIndex = 0;cintIndex < XML_CC_ITEMNAME_NUM;cintIndex++) {
      if (XML_CC_ITEMNAME_IDX_TAB[cintIndex] == cintCurrentItemType) {
         *(CDEV_PARM_SSI_TYPE*)pcintIndex = cintIndex;
         return True;
      }
   }
   return False;
}


/*===========================================================================*
 * ENTRY POINT:  AR_XML_GET_CC_VAL_ITEM_IDX
 *===========================================================================*
 *
 * FUNCTION:    Find channel configuration XML element value item position
 *              in XML_CC_VAL_ITEM_IDX_TAB
 *
 * PARAMETERS:  CDEV_PARM_SSI_TYPE cintCurrentItemType(input) current item type
 *              CDEV_PARM_SSI_TYPE *pcintIndex(output) position of item
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  True item found 
 *                                 False item not found        
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_xml_get_cc_val_item_idx(CDEV_PARM_SSI_TYPE cintCurrentItemType, CDEV_PARM_SSI_TYPE *pcintIndex)
{
   CDEV_API_RET_TYPE cintIndex;

   for (cintIndex = 0;cintIndex < XML_CC_VAL_ITEM_NUM;cintIndex++) {
      if (cintCurrentItemType == XML_CC_VAL_ITEM_IDX_TAB[cintIndex]) {
         *(CDEV_PARM_SSI_TYPE*)pcintIndex = cintIndex;
         return True;
      }
   }
   return False;
}


/*===========================================================================*
 * ENTRY POINT:  AR_XML_CC_SAVE_CONFIG
 *===========================================================================*
 *
 * FUNCTION:    Get hardware settings and save to Rx/Tx struct
 *
 * PARAMETERS:  CDEV_PARM_SSI_TYPE cintCurrentItemType(input) current item type
 *              CDEV_PARM_SSI_TYPE *pcintBoardIndex(output)  return value of board index
 *              CDEV_CHAN_TYPE *pcintRxChNum(output) return value of Rx channel number
 *              CDEV_CHAN_TYPE *pcintTxChNum(output) return value of Tx channel number
 *              RX_CHANNEL_PARMS_TYPE *prxChanConfig(output) return settings of Rx channels
 *              CDEV_PARM_SSI_TYPE             cintRxChConfigSize(input) Rx channel struct size
 *              TX_CHANNEL_PARMS_TYPE *ptxChanConfig(output) return settings of Tx channels
 *              CDEV_PARM_SSI_TYPE             cintTxChConfigSize(input) Tx channel struct size
 *              CDEV_PARM_SSI_TYPE             cintValue(input) value of current item
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry         
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_xml_cc_save_config
(
   CDEV_PARM_SSI_TYPE cintCurrentItemType,
   CDEV_CHAN_TYPE *pcintRxChNum,
   CDEV_CHAN_TYPE *pcintTxChNum,
   CDEV_CHAN_TYPE cintCurRxIndex,
   CDEV_CHAN_TYPE cintCurTxIndex,
   CDEV_CHAN_TYPE cintRxTx,
   RX_CHANNEL_PARMS_TYPE *prxChanConfig,
   CDEV_PARM_SSI_TYPE             cintRxChConfigSize,
   TX_CHANNEL_PARMS_TYPE *ptxChanConfig,
   CDEV_PARM_SSI_TYPE             cintTxChConfigSize,
   CDEV_PARM_SSI_TYPE             cintValue
)
{
   switch (cintCurrentItemType) {
      case RXNUMVAL: 
         *(CDEV_PARM_SSI_TYPE*)pcintRxChNum = cintValue;
         break;
      case TXNUMVAL:
         *(CDEV_PARM_SSI_TYPE*)pcintTxChNum = cintValue;
         break;
      case CHIDXVAL:
         /*Check struct range*/
         if ((RX == cintRxTx) && (cintRxChConfigSize > cintCurRxIndex)) {
            prxChanConfig[cintCurRxIndex].channel = cintValue;
         }
         else if ((TX == cintRxTx) && (cintTxChConfigSize > cintCurTxIndex)) {
            if (cintValue >= CEIDEV_MAX_TX_CHAN) {
               return ARS_INVHARCMD;
            }

            ptxChanConfig[cintCurTxIndex].channel = cintValue;
         }  
         else {
            return ARS_INVHARCMD;
         }
         break;
      case BRVAL:
         /*Check struct range*/
         if ((RX == cintRxTx) && (cintRxChConfigSize > cintCurRxIndex)) {
            prxChanConfig[cintCurRxIndex].bitRate = cintValue;
         }
         else if ((TX == cintRxTx) && (cintTxChConfigSize > cintCurTxIndex)) {
            ptxChanConfig[cintCurTxIndex].bitRate = cintValue;
         }  
         else {
            return ARS_INVHARCMD;
         }
         break;
      case ENVAL:
         /*Check struct range*/
         if ((RX == cintRxTx) && (cintRxChConfigSize > cintCurRxIndex)) {
            prxChanConfig[cintCurRxIndex].enable = cintValue;
         }
         else if ((TX == cintRxTx) && (cintTxChConfigSize > cintCurTxIndex)) {
            ptxChanConfig[cintCurTxIndex].enable = cintValue;
         }  
         else {
            return ARS_INVHARCMD;
         }
         break;
      case RXPVAL:
         /*Check struct range*/
         if (cintRxChConfigSize > cintCurRxIndex) {
            prxChanConfig[cintCurRxIndex].parity = cintValue;
         }
         else {
            return ARS_INVHARCMD;
         }
         break;
      case BMVAL:
         /*Check struct range*/
         if (cintRxChConfigSize > cintCurRxIndex) {
            prxChanConfig[cintCurRxIndex].bufferMode = cintValue;
         }
         else {
            return ARS_INVHARCMD;
         }
         break;
      case IWVAL:
         /*Check struct range*/
         if (cintRxChConfigSize > cintCurRxIndex) {
            prxChanConfig[cintCurRxIndex].internalWrap = cintValue;
         }
         else {
            return ARS_INVHARCMD;
         }
         break;
      case TXPVAL:
         /*Check struct range*/
         if (cintTxChConfigSize > cintCurTxIndex) {
            ptxChanConfig[cintCurTxIndex].parity = cintValue;
         }
         else {
            return ARS_INVHARCMD;
         }
         break;
      case PEVAL:
         /*Check struct range*/
         if (cintTxChConfigSize > cintCurTxIndex) {
            switch (cintValue) {
               case AR_BIT_HIGH:
                  ptxChanConfig[cintCurTxIndex].bitError = AR_HI;
                  ptxChanConfig[cintCurTxIndex].gapError = AR_OFF;
                  break;
               case AR_BIT_LOW:
                  ptxChanConfig[cintCurTxIndex].bitError = AR_LO;
                  ptxChanConfig[cintCurTxIndex].gapError = AR_OFF;
                  break;
               case AR_SHORT_GAP:
                  ptxChanConfig[cintCurTxIndex].bitError = AR_OFF;
                  ptxChanConfig[cintCurTxIndex].gapError = AR_ON;
                  break;
               case AR_PE_NONE:
                  ptxChanConfig[cintCurTxIndex].bitError = AR_OFF;
                  ptxChanConfig[cintCurTxIndex].gapError = AR_OFF;
                  break;
               default:
                  return ARS_INVHARCMD;
            }
         }
         else {
            return ARS_INVHARCMD;
         }
         break;
      default:
         return ARS_INVHARCMD;
   }

   return ARS_NORMAL;
}
/*===========================================================================*
 * ENTRY POINT:  AR_CHANNEL_CONFIG_XML_NEXT_ITEM_TYPE
 *===========================================================================*
 *
 * FUNCTION:    Get next item type
 *
 * PARAMETERS:  CDEV_PARM_SSI_TYPE cintCurrentItemType(input)
 *              CDEV_CHAN_TYPE cintCurRxChIndex(input)
 *              CDEV_CHAN_TYPE cintTotalRxChNum(input)
 *              CDEV_CHAN_TYPE cintCurTxChIndex(input)
 *              CDEV_CHAN_TYPE cintTotalTxChNum(input)
 *              CDEV_CHAN_TYPE cintRxTx(input) Rx or Tx section
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry         
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_channel_config_xml_next_item_type
(
   CDEV_PARM_SSI_TYPE cintCurrentItemType,
   CDEV_CHAN_TYPE cintCurRxChIndex,
   CDEV_CHAN_TYPE cintTotalRxChNum,
   CDEV_CHAN_TYPE cintCurTxChIndex,
   CDEV_CHAN_TYPE cintTotalTxChNum,
   CDEV_CHAN_TYPE cintRxTx
)
{
   CDEV_API_RET_TYPE cintNextItemType = XML_CC_ITEM_END;

   /*If no rx channel record, RXGRPEND directly after RXGRPST*/
   if ((0 == cintTotalRxChNum) && (RXGRPST == cintCurrentItemType)) {
      return TXGRPST;
   }

   /*If no tx channel record, TXGRPEND directly after TXGRPST*/
   if ((0 == cintTotalTxChNum) && (TXGRPST == cintCurrentItemType)) {
      return CCED;
   }

   /*If last rx channel record, RXGRPED after RXCHED*/
   if ((cintCurRxChIndex == cintTotalRxChNum) && (RXCHED == cintCurrentItemType)) {
      return RXGRPED;
   }

   /*If last tx channel record, TXGRPED after TXCHED*/
   if ((cintCurTxChIndex == cintTotalTxChNum) && (TXCHED == cintCurrentItemType)) {
      return TXGRPED;
   }

   /*Find the default next item type*/
   if (cintCurrentItemType > PEVAL) {
      return XML_CC_ITEM_END;
   }
   else {
      cintNextItemType = XML_CC_ITEM_REL_TAB[cintCurrentItemType];
   }
   
   /*After "enable" element,could be rx/tx parity check, change the default one to TXPST*/
   if ((TX == cintRxTx)&& (RXPST == cintNextItemType)) {
      cintNextItemType = TXPST;
   }

   return cintNextItemType;
}


/*===========================================================================*
 * ENTRY POINT:  AR_CHECK_RX_CHANNEL_NUM
 *===========================================================================*
 *
 * FUNCTION:    Check and return configuration setting of Rx channel number
 *
 * PARAMETERS:  pCEI_CHAR pVal(input) current configuration input
 *              CDEV_CHAN_TYPE *pcintRxChNum(output) configuration setting return
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry         
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_check_rx_channel_num(pCEI_CHAR pVal, CDEV_CHAN_TYPE *pcintRxChNum)
{
   CDEV_CHAN_TYPE cintRxChNum = (CDEV_PARM_SSI_TYPE)atoi(pVal);

   if ((False == ar_is_number(pVal)) || (cintRxChNum > CEIDEV_MAX_RX_CHAN) || (cintRxChNum < 0)) {
      return ARS_INVHARCMD;
   }
   else {
      *(CDEV_PARM_SSI_TYPE*)pcintRxChNum = cintRxChNum;
   }
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:  AR_CHECK_TX_CHANNEL_NUM
 *===========================================================================*
 *
 * FUNCTION:    Check and return configuration setting of Tx channel number
 *
 * PARAMETERS:  pCEI_CHAR pVal(input) current configuration input
 *              CDEV_CHAN_TYPE *pcintTxChNum(output) configuration setting return
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry         
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_check_tx_channel_num(pCEI_CHAR pVal, CDEV_CHAN_TYPE *pcintTxChNum)
{
   CDEV_CHAN_TYPE cintTxChNum = (CDEV_PARM_SSI_TYPE)atoi(pVal);

   if ((False == ar_is_number(pVal)) || (cintTxChNum > CEIDEV_MAX_TX_CHAN) || (cintTxChNum < 0)) {
      return ARS_INVHARCMD;
   }
   else {
      *(CDEV_PARM_SSI_TYPE*)pcintTxChNum = cintTxChNum;
   }
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:  AR_CHECK_CHANNEL_IDX
 *===========================================================================*
 *
 * FUNCTION:    Check and return configuration setting of channel index
 *
 * PARAMETERS:  pCEI_CHAR pVal(input) current configuration input
 *              CDEV_PARM_SSI_TYPE *pcintChIndex(output) configuration setting return
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry         
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_check_channel_idx(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE *pcintChIndex)
{
   CDEV_PARM_SSI_TYPE cintChIndex = (CDEV_PARM_SSI_TYPE)atoi(pVal);

   if ((False == ar_is_number(pVal)) || (cintChIndex >= CEIDEV_MAX_RX_CHAN) || (cintChIndex < 0)) {
      return ARS_INVHARVAL;
   }
   else {
      *(CDEV_PARM_SSI_TYPE*)pcintChIndex = cintChIndex;
   }
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:  AR_CHECK_BIT_RATE
 *===========================================================================*
 *
 * FUNCTION:    Check and return configuration setting of bit rate
 *
 * PARAMETERS:  pCEI_CHAR pVal(input) current configuration input
 *              CDEV_PARM_SSI_TYPE *pcintBitRate(output) configuration setting return
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry         
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_check_bit_rate(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE *pcintBitRate)
{
   CEI_INT32 cint32BR = atoi(pVal);
   CEI_INT32 refClockFreq;

   if ((False == ar_is_number(pVal)) || (cint32BR < CEIDEV_MIN_BIT_RATE) || (cint32BR > CEIDEV_MAX_BIT_RATE)) {
      return ARS_INVHARCMD;
   }
   else {
      if ((configuration == CEIDEV_CONFIG_RAR15XT) || (configuration == CEIDEV_CONFIG_RAR15XT)) {
         refClockFreq = CEIDEV_20MHZ_REF_CLOCK;
      }
      else {
         refClockFreq = CEIDEV_16MHZ_REF_CLOCK;
      }
      *(CDEV_PARM_SSI_TYPE*)pcintBitRate = (CDEV_PARM_SSI_TYPE)((refClockFreq / cint32BR) - 2);     
   }
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:  AR_CHECK_ENABLE
 *===========================================================================*
 *
 * FUNCTION:    Check and return configuration setting of enable
 *
 * PARAMETERS:  pCEI_CHAR pVal(input) current configuration input
 *              CDEV_PARM_SSI_TYPE *pcintEnable(output) configuration setting return
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry         
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_check_enable(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE *pcintEnable)
{
   if (0 == strncmp(pVal, OPT_ENABLED, sizeof(OPT_ENABLED))) {
      *(CDEV_PARM_SSI_TYPE*)pcintEnable = (CDEV_PARM_SSI_TYPE)AR_ON;
   }
   else if (0 == strncmp(pVal, OPT_DISABLED, sizeof(OPT_DISABLED))) {
      *(CDEV_PARM_SSI_TYPE*)pcintEnable = (CDEV_PARM_SSI_TYPE)AR_OFF;
   }
   else {
      return ARS_INVHARCMD;
   }
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:  AR_CHECK_RX_PARITY_CHECK
 *===========================================================================*
 *
 * FUNCTION:    Check and return configuration setting of Rx parity check
 *
 * PARAMETERS:  pCEI_CHAR pVal(input) current configuration input
 *              CDEV_PARM_SSI_TYPE *pcintRxParityChk(output) configuration setting return
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry         
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_check_rx_parity_check(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE *pcintRxParityChk)
{
   if (0 == strncmp(pVal, OPT_ENABLED, sizeof(OPT_ENABLED))) {
      *(CDEV_PARM_SSI_TYPE*)pcintRxParityChk = (CDEV_PARM_SSI_TYPE)AR_ON;
   }
   else if (0 == strncmp(pVal, OPT_DISABLED, sizeof(OPT_DISABLED))) {
      *(CDEV_PARM_SSI_TYPE*)pcintRxParityChk = (CDEV_PARM_SSI_TYPE)AR_OFF;
   }
   else {
      return ARS_INVHARCMD;
   }
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:  AR_CHECK_BUFFER_MODE
 *===========================================================================*
 *
 * FUNCTION:    Check and return configuration setting of buffer mode
 *
 * PARAMETERS:  pCEI_CHAR pVal(input) current configuration input
 *              CDEV_PARM_SSI_TYPE *pcintBufMode(output) configuration setting return
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry         
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_check_buffer_mode(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE *pcintBufMode)
{
   if (0 == strncmp(pVal, OPT_MERGED, sizeof(OPT_MERGED))) {
      *(CDEV_PARM_SSI_TYPE*)pcintBufMode = (CDEV_PARM_SSI_TYPE)AR_ON;
   }
   else if (0 == strncmp(pVal, OPT_INDIVIDUAL, sizeof(OPT_INDIVIDUAL))) {
      *(CDEV_PARM_SSI_TYPE*)pcintBufMode = (CDEV_PARM_SSI_TYPE)AR_OFF;
   }
   else {
      return ARS_INVHARCMD;
   }
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:  AR_CHECK_INTERNAL_WRAP
 *===========================================================================*
 *
 * FUNCTION:    Check and return configuration setting of internal wrap
 *
 * PARAMETERS:  pCEI_CHAR pVal(input) current configuration input
 *              CDEV_PARM_SSI_TYPE *pcintIntWrap(output) configuration setting return
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry         
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_check_internal_wrap(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE *pcintIntWrap)
{
   if (0 == strncmp(pVal, OPT_ENABLED, sizeof(OPT_ENABLED))) {
      *(CDEV_PARM_SSI_TYPE*)pcintIntWrap = (CDEV_PARM_SSI_TYPE)AR_WRAP_ON;
   }
   else if (0 == strncmp(pVal, OPT_DISABLED, sizeof(OPT_DISABLED))) {
      *(CDEV_PARM_SSI_TYPE*)pcintIntWrap = (CDEV_PARM_SSI_TYPE)AR_WRAP_OFF;
   }
   else {
      return ARS_INVHARCMD;
   }
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:  AR_CHECK_TX_PARITY_CHECK
 *===========================================================================*
 *
 * FUNCTION:    Check and return configuration setting of Tx parity check
 *
 * PARAMETERS:  pCEI_CHAR pVal(input) current configuration input
 *              CDEV_PARM_SSI_TYPE *pcintTxParityChk(output) configuration setting return
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry         
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_check_tx_parity_check(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE *pcintTxParityChk)
{
   if (0 == strncmp(pVal, OPT_ODD, sizeof(OPT_ODD))) {
      *(CDEV_PARM_SSI_TYPE*)pcintTxParityChk = (CDEV_PARM_SSI_TYPE)ARU_PARITY_ODD;
   }
   else if (0 == strncmp(pVal, OPT_EVEN, sizeof(OPT_EVEN))) {
      *(CDEV_PARM_SSI_TYPE*)pcintTxParityChk = (CDEV_PARM_SSI_TYPE)ARU_PARITY_EVEN;
   }
   else if (0 == strncmp(pVal, OPT_NONE, sizeof(OPT_NONE))) {
      *(CDEV_PARM_SSI_TYPE*)pcintTxParityChk = (CDEV_PARM_SSI_TYPE)ARU_PARITY_NONE;
   }
   else {
      return ARS_INVHARCMD;
   }
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:  AR_CHECK_PROTOCOL_ERROR
 *===========================================================================*
 *
 * FUNCTION:    Check and return configuration setting of protocol error
 *
 * PARAMETERS:  pCEI_CHAR pVal(input) current configuration input
 *              CDEV_PARM_SSI_TYPE *pcintProtErr(output) configuration setting return
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry         
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_check_protocol_error(pCEI_CHAR pVal, CDEV_PARM_SSI_TYPE *pcintProtErr)
{
   if (0 == strncmp(pVal, OPT_BIT_COUNT_HIGH, sizeof(OPT_BIT_COUNT_HIGH))) {
      *(CDEV_PARM_SSI_TYPE*)pcintProtErr = (CDEV_PARM_SSI_TYPE)AR_BIT_HIGH;
   }
   else if (0 == strncmp(pVal, OPT_BIT_COUNT_LOW, sizeof(OPT_BIT_COUNT_LOW))) {
      *(CDEV_PARM_SSI_TYPE*)pcintProtErr = (CDEV_PARM_SSI_TYPE)AR_BIT_LOW;
   }
   else if (0 == strncmp(pVal, OPT_SHORT_GAP,sizeof(OPT_SHORT_GAP))) {
      *(CDEV_PARM_SSI_TYPE*)pcintProtErr = (CDEV_PARM_SSI_TYPE)AR_SHORT_GAP;
   }
   else if (0 == strncmp(pVal, OPT_NONE,sizeof(OPT_NONE))) {
      *(CDEV_PARM_SSI_TYPE*)pcintProtErr = (CDEV_PARM_SSI_TYPE)AR_PE_NONE;
   }
   else {
      return ARS_INVHARCMD;
   }
   return ARS_NORMAL;
}


#ifdef INCLUDE_MSG_SCHEDULER
/*===========================================================================*
 * ENTRY POINT:   A R _ DEFINE _ MESSAGES _ F R O M _ XML _ F I L E        
 *===========================================================================*
 *
 * FUNCTION:    This routine definess a schedule message set from XML file
 *              
 * PARAMETERS:  pCEI_CHAR pcSMFileName (input) schedule message file name
 *
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_BADLOAD     File not found
 *                                 ARS_INVHARCMD   Invalid File entry
 *
 *              from ar_set_device_config:
 *
 *                         ARS_INVARG      "item" parameter value is not
 *                                          supported by this function
 *                         ARS_INVHARVAL   "value" was invalid or channel
 *                                          not supported on this device
 *                         ARS_INVBOARD    Board parameter value invalid or
 *                                         board has not been initialized
 *                         ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed               
 * GLOBALS:     none
 * 
 * DESCRIPTION: This routine reads the contents of the schedulem message
 *              definition XML file and defines the message scenario in the 
 *              specified CEI-x30 board's message scheduler table.
 * 
 *===========================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_define_messages_from_xml_file
(
   pCEI_CHAR pcSMFileName /* (input) schedule message file name */
)
{
   CDEV_API_RET_TYPE cintStatus = ARS_NORMAL;
   CDEV_PARM_SSI_TYPE cintMsgNum = 0;
   FILE      *fSMFilePtr;
   XMLSEC    *pHead = NULL;
 
   /*Read schedule message XML file*/  
   if (NULL == (fSMFilePtr = fopen(pcSMFileName, "r"))) {
      cintStatus = ARS_BADLOAD;
   }
   else {
      /*Allocate the head node*/
      if (NULL == (pHead = (XMLSEC*)calloc(1, sizeof(XMLSEC)))) {
         fclose(fSMFilePtr);
         return ARS_FAILURE;
      }   

      /*Initialize head node*/
      pHead->cintLength = 0;
      pHead->pNext     = NULL;

      /*Get XML file items*/
      if (ARS_NORMAL != (cintStatus = ar_get_item_from_xml_file(fSMFilePtr, pHead))) {
         fclose(fSMFilePtr);
         ar_release_xmlsec_list(pHead);
         return cintStatus;
      }

      if (ARS_NORMAL != (cintStatus = ar_schedule_message_from_xml_list(pHead, &cintMsgNum))) {
         fclose(fSMFilePtr);
         ar_release_xmlsec_list(pHead);
         return cintStatus;
      }

      fclose(fSMFilePtr);
      ar_release_xmlsec_list(pHead);

      /*Download to hardware*/
      cintStatus = ar_define_msg_block (cintMsgNum, scheduledMessages);
   }
   return cintStatus;
}
#endif


/*===========================================================================*
 * ENTRY POINT:  AR_SCHEDULE_MESSAGE_FROM_XML_LIST
 *===========================================================================*
 *
 * FUNCTION:    Decode the schedule message XML items from list structure
 *
 * PARAMETERS:  XMLSEC *pHead(input), the list structure head item
 *              CDEV_PARM_SSI_TYPE *pcintMsgNum(output), the number of schedule message
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry          
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_schedule_message_from_xml_list
(
   XMLSEC *pHead,
   CDEV_PARM_SSI_TYPE *pcintMsgNum
)
{
   XMLSEC    *pCur = NULL;
   CDEV_PARM_SSI_TYPE cintMsgIndex = 0;
   CDEV_PARM_SSI_TYPE cintCurrentItemType = SMST;
   CDEV_PARM_SSI_TYPE cintIndex;
   CDEV_PARM_SSI_TYPE cintIndexSec;
   CDEV_PARM_SSI_TYPE cintEnd = False;
   CDEV_PARM_SSI_TYPE cintMsgNum = 1000; /*The size of maximum message number*/
   CDEV_API_RET_TYPE  cintStatus;
   CDEV_PARM_SSI_TYPE cintLength;

   /*Check input validity*/
   if ((NULL == pHead) || (NULL == pcintMsgNum)) {
      return ARS_INVHARCMD;
   }
   /*Check title format*/
   if (ARS_NORMAL != ar_xml_check_title(pHead, &pCur)) {
      return ARS_INVHARCMD;
   }
   /*Get first XML element*/
   if (NULL != pCur) {
      pCur = pCur->pNext;
   }
   else {
      return ARS_INVHARCMD;
   }

   /*Go through the item list*/
   while (NULL != pCur) {
      /*If empty node, skip*/
      if (0 != pCur->cintLength) {
         /*If element name, check whether match*/
         if (True == ar_xml_get_sm_item_name_idx(cintCurrentItemType, &cintIndex)) {
            cintLength = (CDEV_PARM_SSI_TYPE)strlen(XML_SM_ITEMNAME_TAB[cintIndex]);
            
            if (cintLength > pCur->cintLength) {
               return ARS_INVHARCMD;
            }
            else {
               for (cintIndexSec = 0;cintIndexSec < cintLength;cintIndexSec++) {
                  if (XML_SM_ITEMNAME_TAB[cintIndex][cintIndexSec] != pCur->accTemp[cintIndexSec]) {
                     return ARS_INVHARCMD;
                  }
               }

               if (cintLength < pCur->cintLength) {
                  if (pCur->accTemp[cintLength] != ' ') {
                     return ARS_INVHARCMD;
                  }
               }
            }

            /*Initialize message index*/
            if (MSGST == cintCurrentItemType) {
               scheduledMessages[cintMsgIndex].messageIndex = 0;
            }
            /*Go to the next message storage*/
            if (MSGED == cintCurrentItemType) {
               cintMsgIndex++;
            }
         }
         /*If element value, check range and (map string to CDEV_PARM_SSI_TYPE value, of hardware)*/
         else {
            if (MSGNUMVAL == cintCurrentItemType) {
               if (((cintMsgNum = (CDEV_PARM_SSI_TYPE)atoi(pCur->accTemp)) > 1000) || (cintMsgNum < 0) || (!isdigit(pCur->accTemp[0]))) {
                  return ARS_INVHARCMD;
               }
            }
            else if (True == ar_xml_get_sm_val_item_idx(cintCurrentItemType, &cintIndex)) {
               if (NULL != XML_SM_VAL_FUN_TAB[cintIndex]) {
                  if (ARS_NORMAL != (cintStatus = XML_SM_VAL_FUN_TAB[cintIndex](pCur->accTemp, cintMsgIndex))) {
                     return cintStatus;
                  }
               }
            }
            else {
               return ARS_INVHARCMD;
            }
         }

         if (XML_SM_ITEM_END == (cintCurrentItemType = ar_schedule_message_xml_next_item_type(cintCurrentItemType, cintMsgNum, cintMsgIndex))) {
            return ARS_INVHARCMD;
         }
         else if (SMED == cintCurrentItemType) {
            cintEnd = True;
            break;
         }
      }

      /*Get next item*/
      pCur = pCur->pNext;
   }

   /*Check proper end*/
   if (False == cintEnd) {
      return ARS_INVHARCMD;
   }
   else {
      *(CDEV_PARM_SSI_TYPE*)pcintMsgNum = cintMsgNum;
      return ARS_NORMAL;
   }
}


/*===========================================================================*
 * ENTRY POINT:  AR_XML_GET_SM_ITEM_NAME_IDX
 *===========================================================================*
 *
 * FUNCTION:    Get the index of schedule message element tag in storage
 *
 * PARAMETERS:  CDEV_PARM_SSI_TYPE cintCurrentItemType(input), current item type
 *              CDEV_PARM_SSI_TYPE *pcintIndex(output), tag position in storage
 * RETURN VAL:  CDEV_API_RET_TYPE  True Tag found
 *                                 False Tag not found          
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_xml_get_sm_item_name_idx
(
   CDEV_PARM_SSI_TYPE cintCurrentItemType,
   CDEV_PARM_SSI_TYPE *pcintIndex
)
{
   CDEV_PARM_SSI_TYPE cintIndex;

   /*To check whether element name*/
   for (cintIndex = 0;cintIndex < XML_SM_ITEMNAME_NUM;cintIndex++) {
      if (XML_SM_ITEMNAME_IDX_TAB[cintIndex] == cintCurrentItemType) {
         *(CDEV_PARM_SSI_TYPE*)pcintIndex = cintIndex;
         return True;
      }
   }
   return False;
}


/*===========================================================================*
 * ENTRY POINT:  AR_XML_GET_SM_VAL_ITEM_IDX
 *===========================================================================*
 *
 * FUNCTION:    Get the index of schedule message element value in storage
 *
 * PARAMETERS: CDEV_PARM_SSI_TYPE cintCurrentItemType(input), current item type
 *             CDEV_PARM_SSI_TYPE *pcintIndex(output), tag position in storage
 * RETURN VAL: CDEV_API_RET_TYPE  True Tag found
 *                         False Tag not found          
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_xml_get_sm_val_item_idx
(
   CDEV_PARM_SSI_TYPE cintCurrentItemType, 
   CDEV_PARM_SSI_TYPE *pcintIndex
)
{
   CDEV_PARM_SSI_TYPE cintIndex;

   for (cintIndex = 0;cintIndex < XML_SM_VAL_ITEM_NUM;cintIndex++) {
      if (cintCurrentItemType == XML_SM_VAL_ITEM_IDX_TAB[cintIndex]) {
         *(CDEV_PARM_SSI_TYPE*)pcintIndex = cintIndex;
         return True;
      }
   }
   return False;
}


/*===========================================================================*
 * ENTRY POINT:  AR_SCHEDULE_MESSAGE_XML_NEXT_ITEM_TYPE
 *===========================================================================*
 *
 * FUNCTION:    Get the item type of next item
 *
 * PARAMETERS:  CDEV_PARM_SSI_TYPE cintCurrentItemType(input), current item type
 *              CDEV_PARM_SSI_TYPE cintMsgNum(input), total message number
 *              CDEV_PARM_SSI_TYPE cintMsgIndex(input), current message index
 * RETURN VAL:  CDEV_API_RET_TYPE  XML_SM_ITEM_END Error
 *                                 SMST ~ ETID     Right return value       
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_schedule_message_xml_next_item_type
(
   CDEV_PARM_SSI_TYPE cintCurrentItemType,
   CDEV_PARM_SSI_TYPE cintMsgNum,
   CDEV_PARM_SSI_TYPE cintMsgIndex
)
{
   if (((0 == cintMsgNum) && (MSGGRPST == cintCurrentItemType)) || ((cintMsgNum == cintMsgIndex) && (MSGED == cintCurrentItemType))) {
      return SMED;
   }
   else if (cintCurrentItemType > ETIDVAL) {
      return XML_SM_ITEM_END;
   }
   else {
      return XML_SM_ITEM_REL_TAB[cintCurrentItemType];
   }
}


/*===========================================================================*
 * ENTRY POINT:  AR_SM_CHECK_SAVE_BOARD_INDEX
 *===========================================================================*
 *
 * FUNCTION:    Check and save the value of schedule message board index
 *
 * PARAMETERS:  pCEI_CHAR pVal(input), current value string
 *              CDEV_PARM_SSI_TYPE cintMsgIndex(input), current message index
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry          
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_sm_check_save_board_index
(
   pCEI_CHAR pVal, 
   CDEV_PARM_SSI_TYPE cintMsgIndex
)
{
   CEI_INT32 cint32BrdIdx = (CEI_UINT32)atoi(pVal);

   if ((False == ar_is_number(pVal)) || (cint32BrdIdx > CEIDEV_MAX_BRD_IDX) || (cint32BrdIdx < 0) || (cintMsgIndex > 1000) || (cintMsgIndex < 0)) {
      return ARS_INVBOARD;
   }
   else {
      scheduledMessages[cintMsgIndex].board = (CEI_UINT32)cint32BrdIdx;
   }
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:  AR_SM_CHECK_SAVE_CHANNEL_INDEX
 *===========================================================================*
 *
 * FUNCTION:    Check and save the value of schedule message channel index
 *
 * PARAMETERS:  pCEI_CHAR pVal(input), current value string
 *              CDEV_PARM_SSI_TYPE cintMsgIndex(input), current message index
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry          
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_sm_check_save_channel_index
(
   pCEI_CHAR pVal,
   CDEV_PARM_SSI_TYPE cintMsgIndex
)
{
   CEI_INT32 cint32ChIdx = (CEI_INT32)atoi(pVal);

   if ((False == ar_is_number(pVal)) || (cint32ChIdx >= CEIDEV_MAX_TX_CHAN) || (cint32ChIdx < 0) || (cintMsgIndex > 1000) || (cintMsgIndex < 0)) {
      return ARS_INVHARVAL;
   }
   else {
      scheduledMessages[cintMsgIndex].channel  = (CEI_UINT32)cint32ChIdx;
   }
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:  AR_SM_CHECK_SAVE_RATE
 *===========================================================================*
 *
 * FUNCTION:    Check and save the value of schedule message rate
 *
 * PARAMETERS:  pCEI_CHAR pVal(input), current value string
 *              CDEV_PARM_SSI_TYPE cintMsgIndex(input), current message index
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry          
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_sm_check_save_rate
( 
   pCEI_CHAR pVal, 
   CDEV_PARM_SSI_TYPE cintMsgIndex
)
{
   if ((False == ar_is_32_ucint(pVal)) || (cintMsgIndex > 1000) || (cintMsgIndex < 0)) {
      return ARS_INVHARCMD;
   }
   else {
      scheduledMessages[cintMsgIndex].rate  = (CEI_UINT32)atoi(pVal);
   }
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:  AR_SM_CHECK_SAVE_START_OFFSET
 *===========================================================================*
 *
 * FUNCTION:    Check and save the value of schedule message start offset
 *
 * PARAMETERS:  pCEI_CHAR pVal(input), current value string
 *              CDEV_PARM_SSI_TYPE cintMsgIndex(input), current message index
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry          
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_sm_check_save_start_offset
(
   pCEI_CHAR pVal, 
   CDEV_PARM_SSI_TYPE cintMsgIndex
)
{
   if ((False == ar_is_32_ucint(pVal)) || (cintMsgIndex > 1000) || (cintMsgIndex < 0)) {
      return ARS_INVHARCMD;
   }
   else {
      scheduledMessages[cintMsgIndex].start  = (CEI_UINT32)atoi(pVal);
   }
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:  AR_SM_CHECK_SAVE_TRANSMIT_COUNT
 *===========================================================================*
 *
 * FUNCTION:    Check and save the value of schedule message tansmit count
 *
 * PARAMETERS:  pCEI_CHAR pVal(input), current value string
 *              CDEV_PARM_SSI_TYPE cintMsgIndex(input), current message index
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry          
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_sm_check_save_transmit_count
(
   pCEI_CHAR pVal, 
   CDEV_PARM_SSI_TYPE cintMsgIndex
)
{
   CEI_INT32 cint32TrCn = (CEI_INT32)atoi(pVal);

   if (!isdigit(pVal[0]) && (('-' != pVal[0]) || (!isdigit(pVal[1]) || (strlen(pVal) != 2)))) {
      return ARS_INVHARCMD;   
   }
   else if (((cint32TrCn >= 0) || (-1 == cint32TrCn)) && (cintMsgIndex <= 1000) && (cintMsgIndex >= 0)) {
      if ((-1 != cint32TrCn) && (False == ar_is_32_ucint(pVal))) {
         return ARS_INVHARCMD;   
      }
      scheduledMessages[cintMsgIndex].txCount = (CEI_UINT32)atoi(pVal);
      return ARS_NORMAL;
   }
   else {
      return ARS_INVHARCMD;      
   }  
}


/*===========================================================================*
 * ENTRY POINT:  AR_SM_CHECK_SAVE_DATA
 *===========================================================================*
 *
 * FUNCTION:    Check and save the value of schedule message data
 *
 * PARAMETERS:  pCEI_CHAR pVal(input), current value string
 *              CDEV_PARM_SSI_TYPE cintMsgIndex(input), current message index
 * RETURN VAL:  CDEV_API_RET_TYPE  ARS_NORMAL      Success
 *                                 ARS_INVHARCMD   Invalid File entry          
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_sm_check_save_data
(
   pCEI_CHAR pVal, 
   CDEV_PARM_SSI_TYPE cintMsgIndex
)
{
   CDEV_PARM_SSI_TYPE cintIndex;
   CDEV_PARM_SSI_TYPE cintLength = (CDEV_PARM_SSI_TYPE)strlen(pVal);
   CDEV_API_RET_TYPE  status = ARS_NORMAL;

   if ( ('0' == pVal[0]) && (('x' == pVal[1]) || ('X' == pVal[1])) && (cintMsgIndex <= 1000) && (cintMsgIndex >= 0) && (cintLength > 2) && (cintLength <= 10) ) {

      /*Must be 0~9 or a~f or A~F*/
      for (cintIndex = 2; (cintIndex < cintLength) && (status == ARS_NORMAL); cintIndex++) {
         if ((pVal[cintIndex] >= '0') && (pVal[cintIndex] <= '9')) {
            status = ARS_NORMAL;
         }
         else if (((pVal[cintIndex] >= 'a') && (pVal[cintIndex] <= 'f')) || ((pVal[cintIndex] >= 'A') && (pVal[cintIndex] <= 'F'))) {
            status = ARS_NORMAL;
         }
         else {
            status = ARS_INVHARCMD;
         }
      }

      sscanf(pVal, "%x", &scheduledMessages[cintMsgIndex].data);  
   }
   else {
      status = ARS_INVHARCMD;
   }
   return status;
}


/*===========================================================================*
 * ENTRY POINT:  ar_is_number
 *===========================================================================*
 *
 * FUNCTION:    Check string all digit
 *
 * PARAMETERS:  pCEI_CHAR pStr(input), current value string
 * RETURN VAL:  CDEV_API_RET_TYPE False    
 *                                True        
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_is_number(pCEI_CHAR pStr)
{
   CDEV_PARM_SSI_TYPE cintIndex;
   CDEV_PARM_SSI_TYPE cintLength = (CDEV_PARM_SSI_TYPE)strlen(pStr);

   for (cintIndex = 0; cintIndex < cintLength; cintIndex++) {
      if ( ((pStr[cintIndex] < '0') || (pStr[cintIndex] > '9')) && 
           (pStr[cintIndex] != '\n') && (pStr[cintIndex] != '\0') && (pStr[cintIndex] != 13) && (pStr[cintIndex] != ' ') ) {
         return False;
      }
   }
   return True;
}


/*===========================================================================*
 * ENTRY POINT:  ar_is_32_cint
 *===========================================================================*
 *
 * FUNCTION:    Check string CEI_INT32 
 *
 * PARAMETERS:  pCEI_CHAR pStr(input), current value string
 * RETURN VAL:  CDEV_API_RET_TYPE False    
 *                                True        
 * GLOBALS:     none 
 * DESCRIPTION: 
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE ar_is_32_ucint(pCEI_CHAR pStr)
{
   CDEV_PARM_SSI_TYPE cintIndex;
   CDEV_PARM_SSI_TYPE cintLength = (CDEV_PARM_SSI_TYPE)strlen(pStr);

   if (False == ar_is_number(pStr)) {
      return False;
   }
   else if (cintLength > CEI_INT32_LEN) {
      return False;
   }
   else if (cintLength < CEI_INT32_LEN) {
      return True;
   }
   else {
      for (cintIndex = 0; cintIndex < CEI_INT32_LEN; cintIndex++) {
         if (((CEI_INT32)pStr[cintIndex] - 48) > maxCINT32[cintIndex]) {
            return False;
         }
         else if (((CEI_INT32)pStr[cintIndex] - 48) < maxCINT32[cintIndex]) {
            return True;
         }
      }
   }
   return True;
}

