/*====================================================================================================================*
 * FILE:                                     C D E V _ A P I _ E X P _ T X. C                    
 *====================================================================================================================*
 *
 *   COPYRIGHT 2016 - 2019 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED
 *   ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH THE INCLUSION
 *   OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY OTHER COPIES THEREOF
 *   MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE
 *   AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS, INC. 
 *
 *====================================================================================================================*
 *
 * FUNCTION:    The Expanded Transmit Message portion of the Application Programmer's Interface for CEI-x30 devices.
 *
 * DESCRIPTION: These routines are provided to demonstrate the optional transmit message-specific programming primitives
 *              for the CEI-x30 API using the C programming language.  Each routine implements a particular API function; 
 *              however, a user application program may need to combine functions in different ways to perform the 
 *              desired operation.
 *
 *              Most routines utilize a board number parameter, zero-based on the Device ID/Board Number assigned 
 *              during the installation of CEI-x30-SW distribution.  Channel parameter values are all zero-based.
 *
 * EXTERNALLY DEFINED ROUTINE ENTRY POINTS:         
 *
 *    CEI-x30 ARINC 429 Message Transmit API routines:
 *
 *    ar_putblock_multi_chan      - write a block of messages to multiple transmitter FIFOs
 *    ar_put_429_message          - send an ARINC word on a channel   
 *
 *--------------------------------------------------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date    By   Ver. Comments
 * -------- ----- ---- -----------------------------------------------------------------------------------------------
 * 10/18/16  BDW  4.90 Initial version.  For modification history related to the routines contained herein prior to
 *                     and including v4.90, see the history section in the header of the file CDEV_API.C.
 *
 *====================================================================================================================*/ 
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
   #include <unistd.h>
   #include <fcntl.h>
   #include <string.h>
   #include <pthread.h>
   #include <errno.h>
   #include <sys/stat.h> 
   #include <sys/shm.h>
   #include <sys/sem.h>
   #include <sys/syscall.h>
   #define VOLATILE volatile
#elif defined (VXW_PCI_X86) || defined (VXW_PCI_PPC)
   #include <stdio.h>
   #include <stdlib.h>
   #include <unistd.h>
   #define VOLATILE 
#elif defined (NIVISA_PXI)
   #ifdef LABVIEW_RT
      #include <ansi_c.h> 
   #endif
   #include <visa.h>  
   #define VOLATILE 
#elif defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
   #include <string.h>
   #define VOLATILE 
#else
   #define VOLATILE volatile
#endif

#include "cdev_glb.h"

extern VOLATILE ARINC_LOCAL_TYPE device[MAX_BOARDS];
extern ARINC_GLOBAL_TYPE * deviceGlobal[MAX_BOARDS];

/*===================================================================================================================*
 * ENTRY POINT:                             A R _ P U T _ 4 2 9 _ M E S S A G E
 *===================================================================================================================*
 *
 * FUNCTION:    Insert an ARINC 429 message in the specified ARINC 429 transmit buffer (for legacy compatibility with
 *              the CEI-710 API).
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board  -- (input) board to reference (0 to 127)
 *              CDEV_CHAN_TYPE channel -- (input) 429 transmit channel to use
 *              CDEV_PARM_SI_TYPE data -- (input) 32-bit ARINC 429 message to transmit 
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL      Success
 *                                ARS_INVBOARD    Invalid board or board not initialized
 *                                ARS_INVHARVAL   Invalid transmit channel argument
 *                                ARS_XMITOVRFLO  Transmit buffer overrun has occurred
 *                                ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 * GLOBALS:     none
 * 
 * DESCRIPTION: This routine utilizes the ar_putword routine to insert a message into the respective ARINC 429 
 *              transmit buffer.
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_put_429_message  
( 
   CDEV_BOARD_TYPE board,                                               /* (input) board to reference (0 to 127)     */
   CDEV_CHAN_TYPE  channel,                                             /* (input) 429 transmit channel to use       */
   CDEV_PARM_SI_TYPE data                                               /* (input) 32-bit ARINC 429 data to transmit */
)
{
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                                 /* function return status  */
   status = ar_putword ( board, channel, data );
   return (status);
}
           

/*===================================================================================================================*
 * ENTRY POINT:                          A R _ P U T B L O C K _ M U L T I _ C H A N
 *===================================================================================================================*
 *
 * FUNCTION:    Insert multiple transmit messages into the multiple channel's transmit buffers.
 *
 * PARAMETERS:  CEI_UINT32 board                  -- (input)  board to reference (0 to 127)
 *              CDEV_PARM_SI_TYPE maxMessages     -- (input)  number of messages to transmit
 *              pCEI_UINT32 channels              -- (input)  array of ARINC 429 transmit channels (0 to 31)
 *                                                            having a one-to-one correspondence to the message array
 *              CDEV_PARM_SI_PTR_TYPE data        -- (input)  array of ARINC 429 messages to insert into tx buffers
 *              CDEV_PARM_SI_PTR_TYPE actualCount -- (output) number of messages inserted into transmit buffers
 *
 * RETURN VAL:  ARS_NORMAL        Success
 *              ARS_INVBOARD      Invalid board number
 *              ARS_INVHARVAL     Invalid receive channel argument
 *              ARS_INVARG        Invalid argument value
 *              ARS_BOARD_MUTEX   Access to the Board Lock timed-out/failed
 *              ARS_XMITOVRFLO    Transmit buffer overflow - indicates that there was not enough room in one or  
 *                                more transmit buffers for all of the ARINC 429 messages specified
 *                                ('ActualCount' receives the number of words actually inserted into tx buffers)
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 *
 * DESCRIPTION: This routine inserts multiple ARINC 429 messages in the transmit queues for multiple transmit
 *              channels. When this routine returns, the messages have not necessarily been transmitted, they have 
 *              only been placed in the respective transmit buffer(s). If other messages are present in the transmit
 *              buffer ahead of these messages, they will be transmitted in turn.
 *
 *              Each element of the message data array will be placed in the transmitter buffer designated by the 
 *              same "channels" array element, on an element-by-element basis.
 *
 *===================================================================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_putblock_multi_chan
(
 CEI_UINT32  board,                                                       /* (input)  board to reference (0 to 127)  */
 CDEV_PARM_SI_TYPE maxMessages,                                           /* (input)  number of messages to transmit */
 pCEI_UINT32 channels,                                                    /* (input)  tx channel to use (0 to 31)    */
 CDEV_PARM_SI_PTR_TYPE data,                                              /* (input)  source of messages             */
 CDEV_PARM_SI_PTR_TYPE actualCount                                        /* (output) number of messages transmitted */
)
{
   CDEV_PARM_SI_TYPE messageCount = 0;                                    /* receiver buffer message counter         */
   CEI_UINT32 channelStatus = 0;                                          /* Rx channel FIFO data avail status       */
   CEI_UINT32 channelOffset;                                              /* chan reg region offset in host intfc    */
   CEI_UINT32 bufferOffset;                                               /* chan buffer region offset in host intfc */
   CDEV_PARM_SI_TYPE arincdata;                                           /* temp ARINC message data storage         */
   pCEI_UINT32 localChanPtr;                                              /* pointer to app's channel array          */
   CDEV_PARM_SI_PTR_TYPE localDataPtr;                                    /* pointer to app's message array          */
   CDEV_PARM_SI_TYPE localChannel = 0;                                    /* temp transmit channel storage           */
   CEI_INT32 status = ARS_NORMAL;                                         /* function return status                  */
   CEI_INT32 accessStatus = ARS_NORMAL;                                   /* status of device read/write access      */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                      /* pointer to board local structure        */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                             /* pointer to board global structure       */


#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      /* verify required parameters */
      if ((channels == NULL) || (data == NULL) || (actualCount == NULL)) {
         status = ARS_INVARG;
      }

      if (status == ARS_NORMAL) {
#endif
         /* assign references to this board's device and deviceGlobal structure-array elements */
         deviceLocalStructPtr = &device[board];
         deviceGlobalStructPtr = deviceGlobal[board];

         /* attempt to transmit as many messages as requested */
         messageCount = 0;
         localDataPtr = data;
         localChanPtr = channels;
#ifndef TARGET_EMBEDDED
         /* if multi-thread protection requested, acquire semaphore/mutex-locking */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (accessStatus == ARS_NORMAL) {
#endif
            while ((messageCount < maxMessages) && (accessStatus == ARS_NORMAL) && (status == ARS_NORMAL)) {
   
               /* offset 0-basis and switch the channel select for each "channels" array entry */
               localChannel = (CDEV_PARM_SI_TYPE)*localChanPtr;
               localChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
#ifndef TARGET_EMBEDDED
               status = a429_utl_validate_channel_request ((CDEV_BOARD_TYPE)board, localChannel, ARU_PUT_DATA);
               if (status == ARS_NORMAL) {
#endif
                  channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS + (localChannel * CEIDEV_CHANNEL_REGION_SIZE);
                  bufferOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_BUFFER_3 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE);
      
                  /* check the transmit channel status, if it indicates "NOT FULL"
                   * then stuff the data into the queue; otherwise, it's an overflow.
                   */
                  if (accessStatus == ARS_NORMAL) {
                     accessStatus = cei_read_32bit ((CDEV_BOARD_TYPE)board, &deviceLocalStructPtr->devicePtr[channelOffset], &channelStatus);
                  }
      
                  if (accessStatus == ARS_NORMAL) {
                     if (channelStatus & CEIDEV_CSR_XMIT_FIFO_FULL) {
                        status = ARS_XMITOVRFLO;
                        break;
                     }
                     else {
                        arincdata = *localDataPtr;
                        localDataPtr++;
                        localChanPtr++;
         
                        /* stuff the data into the transmit FIFO */
                        accessStatus = cei_write_32bit ((CDEV_BOARD_TYPE)board, &deviceLocalStructPtr->devicePtr[bufferOffset], (CEI_UINT32)arincdata);
                        if (accessStatus == ARS_NORMAL) {
                           messageCount++;
                        }
                     }
                  }
#ifndef TARGET_EMBEDDED
               }
#endif
            }
#ifndef TARGET_EMBEDDED
            /* if multi-thread protection is active, free any acquired lock */
            if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
               accessStatus = cei_release_thread_lock(LOCK_IDENT);
            }
         }
#endif
      
         *actualCount = (CDEV_PARM_SI_TYPE)messageCount;

#ifndef TARGET_EMBEDDED
      }
#endif
   
#ifndef TARGET_EMBEDDED
      if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL)) {
         status = accessStatus;
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   return status;
}



