/*====================================================================================================================*
 * FILE:                                     C D E V _ A P I _ I N T R P T. C                    
 *====================================================================================================================*
 *
 *   COPYRIGHT 2016 - 2019 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED
 *   ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH THE INCLUSION
 *   OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY OTHER COPIES THEREOF
 *   MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE
 *   AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS, INC. 
 *
 *====================================================================================================================*
 *
 * FUNCTION:    The PCI Interrupt Related Application Programmer's Interface for a CEI-x30 ARINC device.
 *
 * DESCRIPTION: These routines are provided to demonstrate the PCI interrupt-specific programming primitives for the 
 *              CEI-x30 API using the C programming language.  Each routine implements a particular API function; 
 *              however, a user application program may need to combine functions in different ways to perform the 
 *              desired operation.
 *
 *              Most routines utilize a board number parameter, zero-based on the Device ID/Board Number assigned 
 *              during the installation of CEI-x30-SW distribution.  Channel parameter values are all zero-based.
 *
 * EXTERNALLY DEFINED ROUTINE ENTRY POINTS:         
 *
 *    CEI-x30 PCI Interrupt related API routines:
 *
 *    ar_hw_interrupt_buffer_read - reads the API h/w interrupt buffer
 *    ar_interrupt_queue_read     - reads the device interrupt queue
 *    ar_set_isr_function         - reassigns the default ISR to host app
 *    ar_int_control              - legacy routine
 *
 * INTERNAL UTILITY ROUTINES:
 *
 *    a429_utl_initialize_int_queue
 *    a429_utl_process_hw_int
 *    cei_utl_interrupt_queue_read
 *    cei_utl_interrupt_handler   
 *
 *--------------------------------------------------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date    By   Ver. Comments
 * -------- ----- ---- -----------------------------------------------------------------------------------------------
 * 10/18/16  BDW  4.90 Initial version.  For modification history related to the routines contained herein prior to
 *                     and including v4.90, see the history section in the header of the file CDEV_API.C.
 *
 *====================================================================================================================*/ 
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
   #include <unistd.h>
   #include <fcntl.h>
   #include <string.h>
   #include <pthread.h>
   #include <errno.h>
   #include <sys/stat.h> 
   #include <sys/shm.h>
   #include <sys/sem.h>
   #include <sys/syscall.h>
   #define ARU_OS_THREAD_PRIORITY 2
   #define VOLATILE volatile
#elif defined (VXW_PCI_X86) || defined (VXW_PCI_PPC)
   #include <stdio.h>
   #include <stdlib.h>
   #include <unistd.h>
   #define ARU_OS_THREAD_PRIORITY 0
   #define VOLATILE 
#elif defined (NIVISA_PXI)
   #ifdef LABVIEW_RT
      #include <ansi_c.h> 
   #endif
   #include <visa.h>  
   #define ARU_OS_THREAD_PRIORITY 0
   #define VOLATILE 
#elif defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
   #include <string.h>
   #define ARU_OS_THREAD_PRIORITY 0
   #define VOLATILE 
#else
   #define ARU_OS_THREAD_PRIORITY THREAD_PRIORITY_TIME_CRITICAL 
   #define VOLATILE volatile
#endif

#include "cdev_glb.h"

extern VOLATILE ARINC_LOCAL_TYPE device[MAX_BOARDS];
extern ARINC_GLOBAL_TYPE * deviceGlobal[MAX_BOARDS];

/*--------------------------------------------------------------------------------------------------------------------*
 * The Interrupt Service Routine function prototype differs for different operating systems
 *--------------------------------------------------------------------------------------------------------------------*/
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
   /* All Linux distributions */
   void cei_utl_interrupt_handler ( CDEV_BOARD_TYPE deviceIndex );
   void (*x30InterruptFunctionPtr[MAX_BOARDS])(CDEV_BOARD_TYPE) = {
      cei_utl_interrupt_handler, cei_utl_interrupt_handler, cei_utl_interrupt_handler, cei_utl_interrupt_handler, 
      cei_utl_interrupt_handler, cei_utl_interrupt_handler, cei_utl_interrupt_handler, cei_utl_interrupt_handler
   };
#elif defined (VXW_PCI_X86) || defined (VXW_PCI_PPC) || defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
   /* VxWorks and Integrity */
   void cei_utl_interrupt_handler (CDEV_BOARD_TYPE deviceIndex, pCEI_UINT32 data);

   void (*x30InterruptFunctionPtr[MAX_BOARDS])(CDEV_BOARD_TYPE, pCEI_UINT32) = {
      cei_utl_interrupt_handler, cei_utl_interrupt_handler, cei_utl_interrupt_handler, cei_utl_interrupt_handler, 
      cei_utl_interrupt_handler, cei_utl_interrupt_handler, cei_utl_interrupt_handler, cei_utl_interrupt_handler,
      cei_utl_interrupt_handler, cei_utl_interrupt_handler, cei_utl_interrupt_handler, cei_utl_interrupt_handler, 
      cei_utl_interrupt_handler, cei_utl_interrupt_handler, cei_utl_interrupt_handler, cei_utl_interrupt_handler
   };
#elif defined (NIVISA_PXI)
   /* VISA driver under Windows and LabVIEW R/T */
   void (*x30InterruptFunctionPtr[MAX_BOARDS])(CEI_INT32) = {
      NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
   };
#else
   /* All Windows o/s */
   void (*x30InterruptFunctionPtr[MAX_BOARDS])(CDEV_BOARD_TYPE) = {
      NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
      NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
      NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
      NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
      NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
      NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
      NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
      NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
   };
#endif

/* API local PCI interrupt entry storage & controls, maintaining a copy of the h/w interrupt info, 
 * used by the default API-supplied ISR cei_utl_interrupt_handler AND host via ar_interrupt_queue_read.
 */
CEI_INT_QUEUE_TYPE interruptQueue[MAX_BOARDS];


/*===================================================================================================================*
 * ENTRY POINT:                      C E I _ I N T E R R U P T _ Q U E U E _ R E A D
 *===================================================================================================================*
 *
 * FUNCTION:    Reads the hardware interrupt queue filled buffer
 *
 * PARAMETERS:  deviceIndex   -- (input)  device number to reference (0-127)
 *              numberOfWords -- (in/output) number of words to read & no. read
 *              data          -- (output) the queue entry value(s) read
 *
 * RETURN VAL:  ARS_GOTDATA    -- Data was successfully retrieved
 *              ARS_NODATA     -- No data was retrieved
 *
 * GLOBALS:     device[].devicePtr
 *              interruptQueue[].lastReadIntQueueLocation
 *
 * DESCRIPTION: This routine provides read access to the interrupt queue filled by processing h/w interrupts using 
 *              the default API ISR.  For each invocation, the specified number of queue entries will be read from 
 *              the buffer region starting at the location last referenced by the API and ending at the location
 *              referenced by the interrupt queue pointer.
 *
 *===================================================================================================================*/
CEI_INT32 cei_utl_interrupt_queue_read (
   CDEV_BOARD_TYPE deviceIndex,                                                      /* device number (0-15)         */
   pCEI_UINT32 numberOfWords,                                                        /* number of words to read      */
   pCEI_UINT32 data                                                                  /* data read from the buffer    */
)
{
   CEI_UINT32  bufferHead = 0;                                                       /* temp register data parm      */
   CEI_UINT32  bufferTail = 0;                                                       /* temp API tail ptr tracker    */
   CEI_UINT32  unreadBufferSize = 0;                                                 /* unused space in interrupt q  */
   CEI_UINT32  wordsRead = 0;                                                        /* words to write index         */
   CEI_INT32   rxStatus = ARS_NODATA;                                                /* API routine status           */
   CEI_INT32   status = ARS_NORMAL;                                                  /* local API routine status     */
   pCEI_UINT32 appDataPtr;                                                           /* ptr to the app data array    */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                                 /* ptr to board local structure */
   
   deviceLocalStructPtr = &device[deviceIndex];

   /* Acquire the current interrupt queue head pointer */
   status = cei_read_32bit(deviceIndex, &deviceLocalStructPtr->devicePtr[CEIDEV_INT_QUEUE_HEAD_PTR], &bufferHead);
   bufferTail = interruptQueue[deviceIndex].lastReadIntQueueLocation;

   /* Determine how many unread queue entries are available in the h/w interrupt queue */
   if (status == ARS_NORMAL) {
      if (bufferTail != bufferHead) {
         if (bufferTail < bufferHead) {
            unreadBufferSize = bufferHead - bufferTail;
         }
         else {
            unreadBufferSize = (CEIDEV_INT_QUEUE_SIZE - bufferTail) + bufferHead;
         }

         if (unreadBufferSize > 0) {
            rxStatus = ARS_GOTDATA;
         }
      }
   }

   if (rxStatus == ARS_GOTDATA) {
      appDataPtr = data;
      /* Read the lesser of the number of entries requested or the number of
       * unread entries available in the queue, starting from the tail ptr */
      for (wordsRead = 0; (wordsRead < *numberOfWords) && (wordsRead < unreadBufferSize); wordsRead++) {
         status = cei_read_32bit (deviceIndex, &deviceLocalStructPtr->devicePtr[CEIDEV_INT_QUEUE_START + bufferTail], appDataPtr++);
         /* If we reach the end of the h/w interrupt queue, wrap around to the start */
         if (bufferTail == (CEIDEV_INT_QUEUE_SIZE-1)) {
            bufferTail = 0;
         }
         else {   /* increment to the next queue entry */
            bufferTail++;
         }
      }

      /* Update the API-tracked interrupt queue tail ptr to upon completion */
      interruptQueue[deviceIndex].lastReadIntQueueLocation = bufferTail;
   }

   *numberOfWords = wordsRead;

   if (status != ARS_NORMAL) {
      rxStatus = status;
   }

   return rxStatus;
}


/*===================================================================================================================*
 * ENTRY POINT:                     C E I _ U T L _ I N T E R R U P T _ H A N D L E R
 *===================================================================================================================*
 *
 * FUNCTION:    The default hardware interrupt delayed processing routine
 *
 * PARAMETERS:  deviceIndex   -- (input)  device number to reference (0-127)
 *
 * RETURN VAL:  None
 *
 * GLOBALS:     interruptQueue[] - all members
 *
 * DESCRIPTION: This API local example interrupt processing routine provides a demonstration of how the interrupt 
 *              queue can be processed when hardware interrupts are encountered with the x30 board.  When invoked, 
 *              this routine will flush the interrupt queue and store the entries in an API local interrupt entry 
 *              buffer.  The local buffer is treated as a circular buffer, and requires constant flushing by the 
 *              application in order to eliminate any potential loss of interrupt entry information.
 *
 *              The processing in this routine as described above may be superceeded by a user-defined function, 
 *              defined as follows:
 *
 *                Windows:
 *
 *                   void _stdcall function ( CEI_INT32 argument )
 *
 *                Linux:
 *
 *                   void function ( CEI_INT32 argument )
 *
 *                VxWorks/Integrity:
 *
 *                   void function ( CEI_INT32 argument1, CEI_UINT32 * argument2 )
 *
 *              A user-defined interrupt processing routine can be specified ia invocation of ar_set_isr_function().
 *              Once ar_set_isr_function is invoked, the replacement routine will remain as the invoked
 *              interrupt handler for the duration of the application session.
 *
 *===================================================================================================================*/
#if defined (NIVISA_PXI)
   void cei_utl_interrupt_handler ( CEI_INT32 deviceIndex )
#elif defined (_WIN32)
   void _stdcall cei_utl_interrupt_handler ( CDEV_BOARD_TYPE deviceIndex )
#elif defined (_LINUX_X86_) || defined (_LINUX_PPC_)
   void cei_utl_interrupt_handler ( CDEV_BOARD_TYPE deviceIndex )
#else
   void cei_utl_interrupt_handler ( CDEV_BOARD_TYPE deviceIndex, pCEI_UINT32 data )
#endif
{
   CEI_UINT32 numIntQueueEntriesToRead;                                                 /* the amt of space in the   */
   CEI_UINT32 numIntQueueEntriesToFill;                                                 /* buffer available to fill  */
   CEI_UINT32 tempTailPtr;                                                              /* API int buffer tail ptr   */
   CEI_UINT32 tempHeadPtr;                                                              /* API int buffer head ptr   */
   CEI_INT32  status = ARS_NORMAL;                                                      /* local API routine status  */
   CEI_INT_QUEUE_TYPE * interruptQueuePtr;                                              /* local pointer to int bufr */
   CEI_UINT32 * intQueEntryArrayPtr;
   CEI_UINT32 * intQueEntryPtr;
#if defined (_WIN32)
   /* for Windows, a user-defined ISR is invoked here - for VxWorks the user
    * replaces this function "cei_utl_interrupt_handler" with their own
    */
   if (x30InterruptFunctionPtr[deviceIndex] != NULL) {
      x30InterruptFunctionPtr[deviceIndex] ( deviceIndex );
   }
   else {
#endif
      /* get the API local interrupt entry buffer head/tail pointers, then determine how
       * much room is available in the local interrupt entry buffer for storage of unread 
       * interrupt queue entries that have been added by the f/w since the last interrupt
       * was serviced.
       */
      tempTailPtr = interruptQueue[deviceIndex].intQueueTailPtr;
      tempHeadPtr = interruptQueue[deviceIndex].intQueueHeadPtr;
   
      if (tempTailPtr == tempHeadPtr) {
         numIntQueueEntriesToRead = CEIDEV_INT_QUEUE_SIZE - 1;     /* empty buffer */
      }
      else if (tempTailPtr > (tempHeadPtr+1) ) {
         numIntQueueEntriesToRead = tempTailPtr - tempHeadPtr - 1; /* buffer wrap around */
      }
      else if (tempTailPtr < tempHeadPtr) {                        /* typical scenario */
         numIntQueueEntriesToRead = CEIDEV_INT_QUEUE_SIZE - 1 - tempHeadPtr + tempTailPtr;
      }
      else {
         numIntQueueEntriesToRead = 0;                             /* API local entry buffer is full */
      }
      
      /* If there is room in the API lcoal interrupt entry buffer, request to read as much as possible */
      if (numIntQueueEntriesToRead) {
         
         interruptQueuePtr = &interruptQueue[deviceIndex];
         intQueEntryArrayPtr = interruptQueuePtr->intQueueEntries;
         intQueEntryPtr = &intQueEntryArrayPtr[tempHeadPtr];

         /* If the capacity of the API local buffer will result in a buffer wrap when flushing the board's 
          * interrupt queue, first fill the buffer to the end with one read, then fill from the beginning
          * by flushing remaining entries 
          */
         if ((tempHeadPtr + numIntQueueEntriesToRead) > CEIDEV_INT_QUEUE_SIZE - 1) {
            numIntQueueEntriesToFill = CEIDEV_INT_QUEUE_SIZE - tempHeadPtr;
            status = cei_utl_interrupt_queue_read ( (CDEV_BOARD_TYPE)deviceIndex,
                                                    &numIntQueueEntriesToFill,
                                                    intQueEntryPtr );
            tempHeadPtr += numIntQueueEntriesToFill;
            intQueEntryPtr = &intQueEntryArrayPtr[0];
            if (tempHeadPtr >= CEIDEV_INT_QUEUE_SIZE) {
               numIntQueueEntriesToFill = numIntQueueEntriesToRead - numIntQueueEntriesToFill;
               status = cei_utl_interrupt_queue_read ( (CDEV_BOARD_TYPE)deviceIndex,
                                                       &numIntQueueEntriesToFill,
                                                       intQueEntryPtr );
               tempHeadPtr = numIntQueueEntriesToFill;
            }
         }
         else {  /* The capacity of the local buffer won't require a buffer wrap to fill, a single read required */
            status = cei_utl_interrupt_queue_read ( (CDEV_BOARD_TYPE)deviceIndex,
                                                    &numIntQueueEntriesToRead,
                                                    intQueEntryPtr );
      
            tempHeadPtr += numIntQueueEntriesToRead;
         }
   
         /* Set the API local buffer head pointer to the last copied entry from the hardware interrupt queue */
         interruptQueuePtr->intQueueHeadPtr = tempHeadPtr;
      }

#if defined (_WIN32)
   }
#endif
   if (status) ;
   return;
}


/*===================================================================================================================*
 * ENTRY POINT:                 A 4 2 9 _ U T L _ I N I T I A L I Z E _ I N T _ Q U E U E
 *===================================================================================================================*
 *
 * FUNCTION:    Function that initializes the internal API interrupt queue
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board -- (input) board to reference (0 to 127)
 *              CEI_UINT32 initSource -- (input) source/calling function
 *                                               TRUE  - a429_utl_initialize_global_data (session init processing)
 *                                               FALSE - a429_utl_initialize_device      (API reset processing)
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     interruptQueue
 *              device
 *
 * DESCRIPTION: This routine initializes the internal API interrupt queue for the specified device.
 *
 *===================================================================================================================*/
CEI_VOID  a429_utl_initialize_int_queue 
(
   CDEV_BOARD_TYPE board,                                                       /* board to reference (0 to 127)     */
   CEI_UINT32      initSource                                                   /* source/calling function           */
)
{
   interruptQueue[board].intQueueTailPtr = 0U;
   interruptQueue[board].intQueueHeadPtr = 0U;
   if (initSource) {
      /* this is a first-time/power-up initialization invocation */
      interruptQueue[board].lastReadIntQueueLocation = 0U;
      interruptQueue[board].lastApplicationReadIntQueueLocation = 0U;
   }
   else {
      CEI_INT32 accessStatus;                                                        /* status of device r/w access  */
      VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                              /* ptr to board local structure */
      CEI_INT_QUEUE_TYPE * interruptQueuePtr;                                        /* local pointer to int buffer  */
      CEI_UINT32 * intQuePtr;
      
      /* this is a board/API reset initialization invocation, so we reset API tracked    */
      /* queue locations to the current firmware interrupt queue head pointer            */
      deviceLocalStructPtr = &device[board];
      interruptQueuePtr = &interruptQueue[board];
      intQuePtr = &interruptQueuePtr->lastReadIntQueueLocation;
      accessStatus = cei_read_32bit(board, &deviceLocalStructPtr->devicePtr[CEIDEV_INT_QUEUE_HEAD_PTR], intQuePtr );
      interruptQueuePtr->lastApplicationReadIntQueueLocation = interruptQueuePtr->lastReadIntQueueLocation;
   }
}


/*===================================================================================================================*
 * ENTRY POINT:                       A 4 2 9 _ U T L _ P R O C E S S _ H W _ I N T
 *===================================================================================================================*
 *
 * FUNCTION:    Assign PCI Interrupt Control to the Board 
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board    -- (input) Board number (0 to 127) 
 *
 *              CDEV_PARM_SSI_TYPE value -- (input) interrupt enable/disable selection                                
 *                                                  AR_ON  :  enable PCI interrupts   
 *                                                  AR_OFF : disable PCI interrupts   
 *                                              
 * RETURN VAL:  CDEV_API_RET_TYPE EXIT_ROUTINE    Success
 *                                ARS_INVHARVAL   Invalid value parameter option
 *
 * GLOBALS:     device[]
 * 
 * DESCRIPTION: This routine assigns the state of PCI Interrupt Control to the Board, by set/clear of the PLX or 
 *              PCIe Core interrupt enable, and setup or removal of the interrupt task/thread from the 
 *              host-specific o/s (common named) interrupt routines.
 * 
 *===================================================================================================================*/
CDEV_API_RET_TYPE a429_utl_process_hw_int 
( 
   CDEV_BOARD_TYPE board,                                            /* (input) board to reference (0 to 127)        */
   CDEV_PARM_SSI_TYPE value                                          /* (input) interrupt enable/disable selection   */
)
{
   CEI_UINT32 bitFieldValue = 0U;                                           /* temp for device register read content */
   CEI_INT32  accessStatus = ARS_NORMAL;                                        /* status of read/write access       */
   CDEV_PARM_SSI_TYPE status = ARS_NORMAL;                                             /* function return status     */
   CEI_UINT32 localIntEnableFlag = FALSE;                                                                            
   pCEI_UINT32 intCtrlRegPtr = NULL;                                           /* ptr to the PLX Interrupt Ctrl Reg  */
   CEI_UINT32 pciIntCtrlRegOffset = 0U;                                        /* the PLX Interrupt Ctrl Reg Offset  */
   CEI_UINT32 interruptBit;                                                    /* the respective PLX Int Enable bit  */
   CEI_INT32  memoryRegion = 0;                                                /* board's BAR # for processing ISR   */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                           /* pointer to board local structure   */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                                  /* pointer to board global structure  */
   
   /* assign references to this board's device and deviceGlobal structure-array elements */
   deviceLocalStructPtr = &device[board];
   deviceGlobalStructPtr = deviceGlobal[board];

   switch (deviceGlobalStructPtr->configuration) {
      /* PLX PCI9030 Interrupt setup */
      case CEIDEV_CONFIG_CEI430:
      case CEIDEV_CONFIG_RAR_EC:
      case CEIDEV_CONFIG_CEI430A:
         pciIntCtrlRegOffset = PLX9030_CFG_INT_CTRL_STAT_REG;
         intCtrlRegPtr = deviceLocalStructPtr->baseAddr + PLX9030_CFG_INT_CTRL_STAT_REG;
         interruptBit = PLX9030_LOCAL_INT_INP_ENABLE;
         memoryRegion = 1;
         break;

      /* PLX PCI9056 Interrupt setup */
      case CEIDEV_CONFIG_CEI830:
      case CEIDEV_CONFIG_AMCA30:
      case CEIDEV_CONFIG_CEI530:
      case CEIDEV_CONFIG_R830RX:
      case CEIDEV_CONFIG_RAR_CPCI:
      case CEIDEV_CONFIG_R830X820:
      case CEIDEV_CONFIG_RCEI830A:
         pciIntCtrlRegOffset = PLX_CONFIG_INT_CTRL_STAT_REG;
         intCtrlRegPtr = deviceLocalStructPtr->baseAddr + PLX_CONFIG_INT_CTRL_STAT_REG;
         interruptBit = PLX9056_LOCAL_INT_INP_ENABLE;
         memoryRegion = 1;
         break;

      /* No PLX component for native PCI Express - RAR-PCIE, RAR15-XMC... */
      case CEIDEV_CONFIG_RAR_PCIE: 
      case CEIDEV_CONFIG_RAR15:    
      case CEIDEV_CONFIG_RAR15XT:  
      case CEIDEV_CONFIG_RAR_XMC:
      default:  
         interruptBit = 0;
         memoryRegion = 0;
         pciIntCtrlRegOffset = 0;
         break;
   }
   if (value == AR_ON) {
      /* first disable interrupts before initiating the ISR */
      bitFieldValue = CEIDEV_GLOBAL_ENABLE_AUTON_CLEAR + CEIDEV_INTERRUPTS_ENABLED;
      accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
      /* assign the ISR to the h/w interrupt, then enable PLX local interrupt */
      accessStatus = cei_init_interrupt ( board, 
                                          (pCEI_VOID)cei_utl_interrupt_handler, 
                                          ARU_OS_THREAD_PRIORITY, 
                                          intCtrlRegPtr,
                                          memoryRegion,
                                          pciIntCtrlRegOffset,
                                          interruptBit,
                                          deviceGlobalStructPtr->comboBoard1553chanCount,
                                          &localIntEnableFlag );
      if (accessStatus == ARS_NORMAL) {
         deviceGlobalStructPtr->ceiPciInterruptEnabled = localIntEnableFlag;
         /* enable interrupts after initiating the ISR */
         bitFieldValue = CEIDEV_GLOBAL_ENABLE_AUTON_SET + CEIDEV_INTERRUPTS_ENABLED;
         accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
      }
      else {
         status = (CDEV_API_RET_TYPE)accessStatus;
      }
   }
   else if (value == AR_OFF) {
      /* disable interrupts */
      bitFieldValue = CEIDEV_GLOBAL_ENABLE_AUTON_CLEAR + CEIDEV_INTERRUPTS_ENABLED;
      accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
      if (deviceGlobalStructPtr->ceiPciInterruptEnabled) {
         accessStatus = cei_terminate_interrupt(board, interruptBit, intCtrlRegPtr);
         deviceGlobalStructPtr->ceiPciInterruptEnabled = FALSE;
         if (accessStatus != ARS_NORMAL) {
            status = (CDEV_API_RET_TYPE)accessStatus;
         }
      }
   }
   else {
      status = ARS_INVHARVAL;
   }
   if (status == ARS_NORMAL) {
      status = EXIT_ROUTINE;
   }
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                  A R _ H W _ I N T E R R U P T _ B U F F E R _ R E A D
 *===================================================================================================================*
 *
 * FUNCTION:    Read entries from the local API buffer copy of the device h/w interrupt queue
 *
 * PARAMETERS:  board         -- (input)  device number to reference (0-127)
 *              numberOfWords -- (in/output) number of words to read & no. read
 *              data          -- (output) the queue entry value(s) read
 *
 * RETURN VAL:  ARS_GOTDATA   -- Data was successfully retrieved
 *              ARS_NODATA    -- No data was retrieved
 *              ARS_INVBOARD  -- Invalid device ID parameter
 *              ARS_INVARG    -- Invalid data buffer parameter
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 *              lastApplicationReadIntQueueLocation[]
 *
 * DESCRIPTION: This routine provides read access to the local API buffer populated from the device interrupt queue 
 *              via the interrupt queue processing within the default API interrupt service routine (ISR function
 *              name is cei_utl_interrupt_handler).  If the host application replaces the default API ISR with a
 *              custom ISR, via invocation of ar_set_isr_function, this routine will not be usable.
 *
 *              For each invocation of this routine, the specified number of queue entries will be read from the 
 *              buffer region starting at the location last referenced by the API, limited to the location 
 *              referenced by the interrupt queue pointer (last entry acquired from the board).
 *
 *===================================================================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_hw_interrupt_buffer_read 
(
   CEI_UINT32  board,                                             /* (input)  board to reference (0 to 127)          */
   pCEI_UINT32 numberOfWords,                                     /* (input)  number of entries to read              */
   pCEI_UINT32 data                                               /* (output) entries read from the interrupt buffer */
)
{
   CEI_UINT32 tempTailPtr;                                                              /* API int buffer tail ptr   */
   CEI_UINT32 tempHeadPtr;                                                              /* API int buffer head ptr   */
   CEI_UINT32 wordsRead;                                                                /* words to read index       */
   CEI_UINT32 unreadBufferSize = 0;                                                  /* unused space in interrupt q  */
   CEI_INT32  status = ARS_NODATA;                                                      /* local API routine status  */
   pCEI_UINT32 dataPtr;                                                              /* pointer to host data buffer  */
   
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      if (data == NULL) {
         status = ARS_INVARG;
      }
      else {
#endif
         tempTailPtr = interruptQueue[board].intQueueTailPtr;
         tempHeadPtr = interruptQueue[board].intQueueHeadPtr;

         if (tempTailPtr != tempHeadPtr) {
            if (tempTailPtr < tempHeadPtr) {
               unreadBufferSize = tempHeadPtr - tempTailPtr;
            }
            else {
               unreadBufferSize = (INTERRUPT_QUEUE_SIZE - tempTailPtr) + tempHeadPtr;
            }
      
            if (unreadBufferSize > 0) {
               status = ARS_GOTDATA;
               dataPtr = data;
               for (wordsRead = 0; (wordsRead < *numberOfWords) && (wordsRead < unreadBufferSize); wordsRead++) {
                  *dataPtr = interruptQueue[board].intQueueEntries[tempTailPtr++];
                  if (tempTailPtr == INTERRUPT_QUEUE_SIZE) {
                     tempTailPtr = 0;
                  }
                  dataPtr++;
               }
               *numberOfWords = wordsRead;
               interruptQueue[board].intQueueTailPtr = tempTailPtr;
            }
         }
#ifndef TARGET_EMBEDDED
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   return (status);
}



EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_int_control (CDEV_BOARD_TYPE board, CDEV_CHAN_TYPE channel, CDEV_PARM_SSI_TYPE flag) 
{
   CDEV_API_RET_TYPE status = channel;
   if (flag == ARU_ENABLE_INT) {
      status = ar_set_device_config (board, 0, ARU_HW_INTERRUPT_ENABLE, AR_ON);
   }
   else {
      status = ar_set_device_config (board, 0, ARU_HW_INTERRUPT_ENABLE, AR_OFF);
   }
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                       A R _ I N T E R R U P T _ Q U E U E _ R E A D
 *===================================================================================================================*
 *
 * FUNCTION:    Provides application interrupt entry read access directly from the device interrupt queue.
 *
 * PARAMETERS:  board           -- (input)  device number to reference (0-127)
 *              numberOfWords   -- (input)  number of queue entries requested to read
 *                                 (output) number of entries actually read from the queue
 *              data            -- (output) the interrupt queue entry value(s) read
 *
 * RETURN VAL:  ARS_GOTDATA     -- Data was successfully retrieved
 *              ARS_NODATA      -- No data was retrieved
 *              ARS_INVBOARD    -- Invalid device ID parameter
 *              ARS_INVARG      -- Invalid data buffer parameter
 *              ARS_BOARD_MUTEX -- Access to the Board Lock timed-out/failed
 *              ARS_FAILURE     -- Interrupt Queue head pointer value read was invalid
 *
 * GLOBALS:     deviceConfig[].deviceAcquired
 *              device[].devicePtr
 *              lastApplicationReadIntQueueLocation[]
 *
 * DESCRIPTION: This routine provides application read access directly from the device interrupt queue. For each 
 *              invocation, the specified number of interrupt queue entries will be read from the buffer region 
 *              starting at the location last referenced by the API, limited to the last entry inserted by the 
 *              device, indicated via the interrupt queue pointer.
 *
 *===================================================================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_interrupt_queue_read 
(
   CEI_UINT32  board,                                             /* (input)  board to reference (0 to 127)          */
   pCEI_UINT32 numberOfWords,                                     /* (input)  number of entries to read              */
   pCEI_UINT32 data                                               /* (output) entries read from the interrupt buffer */
)
{
   CEI_UINT32  bufferHead = 0;                                                       /* temp register data parm      */
   CEI_UINT32  bufferTail = 0;                                                       /* temp API tail ptr tracker    */
   CEI_UINT32  unreadBufferSize = 0;                                                 /* unused space in interrupt q  */
   CEI_UINT32  wordsRead = 0;                                                        /* words to write index         */
   CEI_INT32   rxStatus = ARS_NODATA;                                                /* API routine status           */
   CEI_INT32   status = ARS_NORMAL;                                                  /* local API routine status     */
   pCEI_UINT32 appDataPtr;                                                           /* ptr to the app data array    */
   CEI_UINT32  memOffset;                                                    /* local offset to channel memory area  */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                                 /* ptr to board local structure */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   else if (data == NULL) {
      status = ARS_INVARG;
   }
   else {
#endif
      deviceLocalStructPtr = &device[board];
      
      /* Acquire the current interrupt queue head pointer */
      status = cei_read_32bit((CDEV_BOARD_TYPE)board, &deviceLocalStructPtr->devicePtr[CEIDEV_INT_QUEUE_HEAD_PTR], &bufferHead);
      bufferTail = interruptQueue[board].lastApplicationReadIntQueueLocation;
   
      /* Determine how many unread queue entries are available in the h/w interrupt queue */
      if (status == ARS_NORMAL) {
         if (bufferTail != bufferHead) {
            if (bufferTail < bufferHead) {
               unreadBufferSize = bufferHead - bufferTail;
            }
            else {
               unreadBufferSize = (CEIDEV_INT_QUEUE_SIZE - bufferTail) + bufferHead;
            }
            if (unreadBufferSize > 0) {
               rxStatus = ARS_GOTDATA;
            }
         }
      }
   
      if (rxStatus == ARS_GOTDATA) {
         appDataPtr = data;
         /* Read the lesser of the number of entries requested or the number of
          * unread entries available in the queue, starting from the tail ptr */
         memOffset = CEIDEV_INT_QUEUE_START + bufferTail;
   
#ifndef TARGET_EMBEDDED
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            status = cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (status == ARS_NORMAL) {
#endif
            for (wordsRead = 0; (wordsRead < *numberOfWords) && (wordsRead < unreadBufferSize); wordsRead++) {
               status = cei_read_32bit ((CDEV_BOARD_TYPE)board, &deviceLocalStructPtr->devicePtr[memOffset], appDataPtr++);
               /* If we reach the end of the h/w interrupt queue, wrap around to the start */
               if (bufferTail == (CEIDEV_INT_QUEUE_SIZE-1)) {
                  bufferTail = 0;
                  memOffset = CEIDEV_INT_QUEUE_START;
               }
               else {
                  bufferTail++;
                  memOffset++;
               }
            }
#ifndef TARGET_EMBEDDED
            if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
               status = cei_release_thread_lock(LOCK_IDENT);
            }
         }
#endif
      
         /* Update the API-tracked interrupt queue tail ptr to upon completion */
         interruptQueue[board].lastApplicationReadIntQueueLocation = bufferTail;
      }
   
      *numberOfWords = wordsRead;
   
#ifndef TARGET_EMBEDDED
   }
#endif
   if (status != ARS_NORMAL) {
      rxStatus = status;
   }

   return rxStatus;
}


/*===================================================================================================================*
 * ENTRY POINT:                        A R _ S E T _ I S R _ F U N C T I O N
 *===================================================================================================================*
 *
 * FUNCTION:    Define an application supplied ISR routine to override the API internal default ISR.
 *
 * PARAMETERS:  deviceIndex   -- (input) device to reference (0 to 127)
 *              function      -- (input) void function to execute from ISR
 *
 * RETURN VAL:  ARS_NORMAL    -- Success
 *              ARS_INVBOARD  -- Invalid device ID parameter
 *              ARS_INVARG    -- State parameter value is invalid
 *
 * GLOBALS:     x30InterruptFunctionPtr[]
 *
 * DESCRIPTION: This routine assigns the supplied function pointer to an array of pointers indexed by device id.  
 *              The referenced function will be invoked from cei_utl_interrupt_handler() instead of executing the 
 *              API default "flush the h/w interrupt queue" processing via cei_utl_interrupt_handler.
 *
 *===================================================================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_set_isr_function 
(
   CDEV_BOARD_TYPE deviceIndex,                                               /* the device identifier               */
   pCEI_VOID function                                                         /* multithread protect state selection */
)
{
   CEI_INT32 status = ARS_NORMAL;                                                        /* function return status   */

   if (deviceIndex < MAX_BOARDS) {
      if (function != NULL) {
         x30InterruptFunctionPtr[deviceIndex] = function;
      }
      else {
         status = ARS_INVARG;
      }
   }
   else {
      status = ARS_INVBOARD;
   }

   return status;
}

