/*====================================================================================================================*
 * FILE:                                     C D E V _ A P I _ I R I G. C                    
 *====================================================================================================================*
 *
 *   COPYRIGHT 2016 - 2019 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED
 *   ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH THE INCLUSION
 *   OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY OTHER COPIES THEREOF
 *   MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE
 *   AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS, INC. 
 *
 *====================================================================================================================*
 *
 * FUNCTION:    The IRIG related local API routines for a CEI-x30 ARINC device.
 *
 * DESCRIPTION: These routines are provided to demonstrate low-level IRIG-specific programming primitives for the 
 *              CEI-x30 API using the C programming language.
 *
 *              Most routines utilize a board number parameter, zero-based on the Device ID/Board Number assigned 
 *              during the installation of CEI-x30-SW distribution.  Channel parameter values are all zero-based.
 *
 * INTERNAL UTILITY ROUTINES:
 *
 *    a429_utl_adjust_irig_threshold
 *    a429_utl_find_dac_threshold
 *    a429_utl_is_irig_calibrated
 *    a429_utl_quick_adjust_irig_threshold
 *
 *--------------------------------------------------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date    By   Ver. Comments
 * -------- ----- ---- -----------------------------------------------------------------------------------------------
 * 10/18/16  BDW  4.90 Initial version.  For modification history related to the routines contained herein prior to
 *                     and including v4.90, see the history section in the header of the file CDEV_API.C.
 *
 *====================================================================================================================*/ 
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
   #include <unistd.h>
   #include <fcntl.h>
   #include <string.h>
   #include <pthread.h>
   #include <errno.h>
   #include <sys/stat.h> 
   #include <sys/shm.h>
   #include <sys/sem.h>
   #include <sys/syscall.h>
   #define VOLATILE volatile
#elif defined (VXW_PCI_X86) || defined (VXW_PCI_PPC)
   #include <stdio.h>
   #include <stdlib.h>
   #include <unistd.h>
   #define VOLATILE 
#elif defined (NIVISA_PXI)
   #ifdef LABVIEW_RT
      #include <ansi_c.h> 
   #endif
   #include <visa.h>  
   #define VOLATILE 
#elif defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
   #include <string.h>
   #define VOLATILE 
#else
   #define VOLATILE volatile
#endif

#include "cdev_glb.h"

extern VOLATILE ARINC_LOCAL_TYPE device[MAX_BOARDS];
extern ARINC_GLOBAL_TYPE * deviceGlobal[MAX_BOARDS];

/**********************************************************************************************************************
 **********************************************************************************************************************
 *                   O P E R A T I N G   S Y S T E M   C O M P I L A T I O N   D I R E C T I V E S                    *
 **********************************************************************************************************************
 **********************************************************************************************************************/

#ifndef LABVIEW_RT
   static pCEI_CHAR thresholdText[2] = {"lower", "upper"};
#endif
#if defined DEBUG_PROGRAMMING
   #include "irigdecl.h"
#endif

/*===================================================================================================================*
 * ENTRY POINT:                A 4 2 9 _ U T L _ A D J U S T _ I R I G _ T H R E S H O L D
 *===================================================================================================================*
 *
 * FUNCTION:    Performs a thorough adjustment of the IRIG DAC for an external input IRIG signal, using valid 
 *              IRIG-B 32-bit time reception for verification of signal presence.
 *
 * DESCRIPTION: This routine sets the DAC to Vmin + (0.825 * (Vmax-Vmin)).  
 *              Vmin is the lower peak value and Vmax is the upper peak value.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board        -- (input)  Board number (0 to 127) 
 *              CEI_INT32 displayDacSettings -- (input)  when set enables DAC adjustment print to console
 * 
 * RETURN VAL:  status   --  Any value from 0 to x7FFFFFFF is the DAC value
 *                           ARS_FAILURE (-1) Any issue encountered with incoming IRIG signal level
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_adjust_irig_threshold 
(
   CDEV_BOARD_TYPE board,                                             /* (input)  board to reference (0 to 127)      */
   CEI_INT32 displayDacSettings                         /* (input)  when set enables DAC adjustment print to console */
)
{
   CEI_UINT32 dacMaxVoltage = IRIG_DAC_MAX_VALUE;                              /* temp max DAC value for thold test  */
   CEI_UINT32 dacValue;                                                        /* dynamic DAC value during test      */
   CEI_UINT32 initialDacValue = IRIG_DAC_ADJUST_INITIAL_COUNT;                 /* initial DAC value                  */
   CEI_UINT32 initialDacAdjustment = INITIAL_SIGNAL_DETECT_ADJUST;             /* initial DAC increment              */
   CEI_UINT32 minimumVoltageValue = 0U;                                        /* lower bounds of final DAC value    */
   CEI_UINT32 maximumVoltageValue = 0U;                                        /* upper bounds of final DAC value    */
   CEI_UINT32 irigRegValue;                                                    /* temp contents of IRIG Receiver     */
   CEI_UINT32 dacBoudaryCushionValue = 0U;                                     /* minor adj to final DAC value       */
   CEI_INT32 signalFound = FALSE;                                              /* flag ind IRIG signal edge detected */
   CEI_INT32 accessStatus = ARS_NORMAL;                                        /* status of board read/write access  */
   CEI_INT32 status = ARS_NORMAL;                                              /* status of threshold detection      */
   CEI_UINT32 prevIrigRegValue = 0U;                                           /* IRIG Time sample value             */
   pCEI_UINT32 dacControlRegPtr = &device[board].devicePtr[CEIDEV_DAC_CONTROL];
   pCEI_UINT32 irigReceiverRegPtr = &device[board].devicePtr[CEIDEV_IRIG_INPUT_TIME];
   
   dacValue = IRIG_DAC_ADJUST_INITIAL_COUNT;

#ifndef LABVIEW_RT
   if (displayDacSettings) {
      printf ("Initial DAC Setting = %3d\n", (CEI_NATIVE_INT)dacValue);
   }
#endif
   /*=======================================================================
    * FIRST FIND THE INCOMING SIGNAL
    * Attempt to initially detect the incoming IRIG signal between the 
    * mid-upper peak by setting the DAC to 128 (0.0 Volts) and increasing
    * the threshold until the signal is found.
    *=======================================================================*/
   accessStatus = cei_write_32bit (board, dacControlRegPtr, dacValue);
   status = cei_delay_milliseconds(ADJUST_DELAY_FOR_EDGE_DETECT);         /* Wait 5 Msec for signal    */
   accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &prevIrigRegValue);
   status = cei_delay_milliseconds(ADJUST_DELAY_30_MILLISECONDS);
   
   /* Read IRIG value, clearing "edge received" bit, then  */
   /* Wait 2 seconds for a new incoming sample. */
   accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &prevIrigRegValue);
   status = cei_delay_milliseconds(ADJUST_DELAY_FOR_TIME_SAMPLE);
 
   /* Read IRIG value again, check the time sample changed */
   accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &irigRegValue);
   if (irigRegValue != prevIrigRegValue) {
      prevIrigRegValue = irigRegValue;
      signalFound = TRUE;
   }
   else {
      /* Signal wasn't immediately detected, raise the threshold until it's found */
      for (dacValue = IRIG_DAC_ADJUST_INITIAL_COUNT; (dacValue <= IRIG_DAC_MAX_VALUE) && (!signalFound); dacValue += INITIAL_SIGNAL_DETECT_ADJUST) {
         accessStatus = cei_write_32bit (board, dacControlRegPtr, dacValue);
         
#ifndef LABVIEW_RT
         if (displayDacSettings) {
            printf ("\rFinding the signal above the mid point -- DAC Setting = %3d", (CEI_NATIVE_INT)dacValue);
         }
#endif
         /* Read IRIG value, clearing "edge received" bit, then  */
         /* Wait 30 Msec for an incoming signal edge transition. */
         status = cei_delay_milliseconds(ADJUST_DELAY_FOR_EDGE_DETECT);
         accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &irigRegValue);

         status = cei_delay_milliseconds(ADJUST_DELAY_FOR_TIME_SAMPLE);
         accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &irigRegValue);
         if (irigRegValue != prevIrigRegValue) {
            prevIrigRegValue = irigRegValue;
            dacMaxVoltage = IRIG_DAC_MAX_VALUE;
            initialDacValue = dacValue;
            initialDacAdjustment = INITIAL_SIGNAL_DETECT_NO_SIG_ADJUST;
            signalFound = TRUE;
         }
      }

#ifndef LABVIEW_RT
      if (displayDacSettings) {
         printf ("\n");
      }
#endif
      /* Signal wasn't detected by raising the threshold? Lower it now until it's found */
      for (dacValue = IRIG_DAC_ADJUST_INITIAL_COUNT; (dacValue > 8U) && (!signalFound); dacValue -= INITIAL_SIGNAL_DETECT_ADJUST) {
         accessStatus = cei_write_32bit (board, dacControlRegPtr, dacValue);

#ifndef LABVIEW_RT
         if (displayDacSettings) {
            printf ("\rFinding the signal below the mid point -- DAC Setting = %3d", (CEI_NATIVE_INT)dacValue);
         }
#endif
         /* Read IRIG value, clearing "edge received" bit, then  */
         /* Wait 30 Msec for an incoming signal edge transition. */
         status = cei_delay_milliseconds(ADJUST_DELAY_FOR_EDGE_DETECT);
         accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &irigRegValue);
         
         status = cei_delay_milliseconds(ADJUST_DELAY_FOR_TIME_SAMPLE);
         accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &irigRegValue);
         if (irigRegValue != prevIrigRegValue) {
            prevIrigRegValue = irigRegValue;
            dacMaxVoltage = IRIG_DAC_ADJUST_INITIAL_COUNT;
            initialDacAdjustment = INITIAL_SIGNAL_DETECT_NO_SIG_ADJUST;
            initialDacValue = dacValue;
            signalFound = TRUE;
         }
      }

      if (!signalFound) {
#ifndef LABVIEW_RT
         if (displayDacSettings) {
            printf ("\nNever found a signal -- final DAC Setting = %3d\n", (CEI_NATIVE_INT)dacValue);
         }
#endif
         status = ARS_FAILURE;
      }
   }

   if (status == ARS_NORMAL) {
      /*=======================================================================
       * SECOND FIND THE LOW THRESHOLD AT WHICH SIGNAL BECOMES PRESENT
       * Starting with the theshold DAC at level the signal was detected above,
       * decrement the DAC voltage until the signal is no longer present;
       * then perform a fine adjustment until DAC setting is just above
       * the lower threshold where the signal becomes present.
       *=======================================================================*/
      dacValue = initialDacValue;
      status = a429_utl_find_dac_threshold (board, 
                                 TRUE, 
                                 displayDacSettings, 
                                 AR_LOWER, 
                                 dacMaxVoltage, 
                                 initialDacAdjustment, 
                                 &dacValue, 
                                 &dacBoudaryCushionValue); 
   }
   
   if (status == ARS_NORMAL) {
      minimumVoltageValue = dacValue + dacBoudaryCushionValue;
      
#ifndef LABVIEW_RT
      if (displayDacSettings) {
         printf ("\nLower transition final value = %3d\n", (CEI_NATIVE_INT)minimumVoltageValue);
      }
#endif
      /*=======================================================================
       * THIRD FIND THE AR_UPPER THRESHOLD AT WHICH SIGNAL BECOMES LOST
       * Starting with the theshold DAC at level the signal was detected first,
       * increment the DAC voltage until the signal is no longer present;
       * then perform a fine adjustment until DAC setting is just below
       * the upper threshold where the signal is lost.
       *=======================================================================*/
      dacValue = initialDacValue + initialDacAdjustment;
      status = a429_utl_find_dac_threshold (board, 
                                 TRUE, 
                                 displayDacSettings, 
                                 AR_UPPER, 
                                 minimumVoltageValue, 
                                 initialDacAdjustment, 
                                 &dacValue, 
                                 &dacBoudaryCushionValue); 
   }

   if (status == ARS_NORMAL) {
      maximumVoltageValue = dacValue - dacBoudaryCushionValue;
   
#ifndef LABVIEW_RT
      if (displayDacSettings) {
         printf ("\nUpper transition final value = %3d\n", (CEI_NATIVE_INT)maximumVoltageValue);
      }
#endif
   
      /* validate the span between the upper and lower thresholds */
      if ((maximumVoltageValue < minimumVoltageValue) || ((maximumVoltageValue-minimumVoltageValue) < MIN_DAC_SPAN)) {
#ifndef LABVIEW_RT
         if (displayDacSettings) {
            printf ("Error, Signal Detection Range is invalid, Vmax = %d  Vmin = %d\n", 
                    (CEI_NATIVE_INT)maximumVoltageValue, (CEI_NATIVE_INT)minimumVoltageValue);
         }
#endif
         accessStatus = cei_write_32bit (board, dacControlRegPtr, IRIG_DAC_ADJUST_INITIAL_COUNT);
         status = ARS_FAILURE;
      }
   }

   /* write the computed DAC counts value to the DAC control register, then convert to millivolts */
   if (status == ARS_NORMAL) {
      dacValue = minimumVoltageValue + (CEI_UINT32)((825*(maximumVoltageValue-minimumVoltageValue))/1000);
      accessStatus = cei_write_32bit (board, dacControlRegPtr, dacValue);
   
#ifndef LABVIEW_RT
      if (displayDacSettings) {
         printf ("Final DAC Setting = %3d\n", (CEI_NATIVE_INT)dacValue);
      }
#endif

      if (dacValue <= CEIDEV_IRIG_DAC_OFFSET) {
         device[board].irigDacThreshold = 0U;
         dacValue = 0U;
      }
      else {
         device[board].irigDacThreshold = (CEI_INT32)dacValue;
         dacValue = (CEI_UINT32)(CEIDEV_IRIG_DAC_MV_CONVERSION * (float)(dacValue - CEIDEV_IRIG_DAC_OFFSET));
      }
   }
   if (status == ARS_NORMAL) {
      if (accessStatus == ARS_NORMAL) {
         status = (CEI_INT32)(dacValue);
      }
      else {
         status = accessStatus;
      }
   }
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                    A 4 2 9 _ U T L _ F I N D _ D A C _ T H R E S H O L D
 *===================================================================================================================*
 *
 * FUNCTION:    Given a threshold type (lower or upper) and the bounds of the threshold search, find the DAC setting 
 *              at which the IRIG signal is lost, then return a DAC value just under that value.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board               -- (input)  Board number (0 to 127) 
 *              CEI_INT32 useActualTimeSample,      -- (input)  IRIG edge det. vs time cmpare      
 *              CEI_INT32 displayDacSettings,       -- (input)  conditional flag printf invoc      
 *              CEI_INT32 thresholdType,            -- (input)  find AR_LOWER or AR_UPPER threshold
 *              CEI_UINT32 voltageLimit,            -- (input)  outer bounds for the search        
 *              CEI_UINT32 initialDacAdjustment,    -- (input)  initial DAC increment              
 *              pCEI_UINT32 finalDacValue,          -- (output) dynamic DAC value during test      
 *              pCEI_UINT32 dacBoudaryCushionValue  -- (output) minor adj to final DAC value       
 *
 * GLOBALS:     device[]
 *
 * RETURN VAL:  CEI_INT32 -- ARS_NORMAL       threshold was detected
 *                           ARS_FAILURE (-1) Any issue encountered with incoming IRIG signal level
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_find_dac_threshold 
(
   CDEV_BOARD_TYPE board,                                            /* (input)  board to reference (0 to 127)       */
   CEI_INT32 useActualTimeSample,                                    /* (input)  IRIG edge det. vs time cmpare       */
   CEI_INT32 displayDacSettings,                                     /* (input)  conditional flag printf invoc       */
   CEI_INT32 thresholdType,                                          /* (input)  find AR_LOWER or AR_UPPER threshold */ 
   CEI_UINT32 voltageLimit,                                          /* (input)  outer bounds for the search         */
   CEI_UINT32 initialDacAdjustment,                                  /* (input)  initial DAC increment               */
   pCEI_UINT32 finalDacValue,                                        /* (output) dynamic DAC value during test       */
   pCEI_UINT32 dacBoudaryCushionValue                                /* (output) minor adj to final DAC value        */
)
{
   CEI_UINT32 adjustmentValue;                                                  /* dynamic increment to DAC voltage  */
   CEI_UINT32 dacValue;                                                         /* dynamic DAC value during test     */
   CEI_UINT32 irigRegValue;                                                     /* temp contents of IRIG Receiver    */
   CEI_INT32 sampleReceived = TRUE;                            /* flag indicating IRIG signal recvd or edge detected */
   CEI_INT32 adjustmentDirection = APPLY_ADJUSTMENT_AND_TEST;                   /* adjustment mode                   */
   CEI_INT32 thresholdFound = FALSE;                                            /* IRIG signal threshold determined  */
   CEI_INT32 status = ARS_NORMAL;                                               /* status of o/s delay calls         */
   CEI_INT32 accessStatus = ARS_NORMAL;                                         /* status of board read/write access */
   CEI_UINT32 prevIrigRegValue = 0U;                                            /* IRIG Time sample value            */
   pCEI_UINT32 irigReceiverRegPtr;                                              /* pointers to device registers      */
   pCEI_UINT32 dacControlRegPtr = &device[board].devicePtr[CEIDEV_DAC_CONTROL];
   
   if (useActualTimeSample) {
      irigReceiverRegPtr = &device[board].devicePtr[CEIDEV_IRIG_INPUT_TIME];
   }
   else {
      irigReceiverRegPtr = &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE];
   }
   
   thresholdFound = FALSE;
   adjustmentValue = initialDacAdjustment;
   dacValue = *finalDacValue;

   while (accessStatus == ARS_NORMAL) {
      /* start with the DAC "course adjustment until the signal is lost, then fine adjust until the threshold is found */
      
      /* For time "sample" usage, read the IRIG sample value.  Clear the "edge received" bit in the DAC Control Register.  
       * Delay 2 seconds if waiting for a full IRIG time sample, or 5msec for an just pulse edge detection.
       * Then wait an additional 30 msec between double reads (just to be sure).
       */
      if (useActualTimeSample) {
         accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &prevIrigRegValue);
      }
      accessStatus = cei_write_32bit (board, dacControlRegPtr, dacValue);
      
#ifndef LABVIEW_RT
      if (displayDacSettings) {
         printf ("\rFinding the %s transition -- DAC Setting = %3d", thresholdText[thresholdType], (CEI_NATIVE_INT)dacValue);
      }
#endif
      if (useActualTimeSample) {
         status = cei_delay_milliseconds(ADJUST_DELAY_FOR_TIME_SAMPLE);
      }
      else {
         status = cei_delay_milliseconds(ADJUST_DELAY_FOR_EDGE_DETECT);
      }

      /* double read will make sure any edge detect is not just some noise */
      accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &irigRegValue);
      status = cei_delay_milliseconds(ADJUST_DELAY_30_MILLISECONDS);
      accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &irigRegValue);
      
      /* determine if the "sample received" condition has been met */
      if (useActualTimeSample) {
         sampleReceived = (irigRegValue == prevIrigRegValue) ? FALSE : TRUE;
      }
      else {
         sampleReceived = irigRegValue & CEIDEV_IRIG_EDGE_DETECT_BIT;
      }

      /* if an IRIG signal is detected, continue with the DAC "course adjustment" until the DAC adjustment is reduced to 4 counts. */
      if (sampleReceived) {
         prevIrigRegValue = irigRegValue;

         /* adjust the DAC value in the appropriate direction for the next DAC assignment and signal check */
         if (thresholdType == AR_LOWER) {
            dacValue -= adjustmentValue;
         }
         else {
            dacValue += adjustmentValue;
         }

         /* When the signal is not detected the last time under the "course adjustment", (the last adjustment value was 4 counts),
          * enter the "fine adjustment", where DAC value is adjusted by only 1 count until the threshold is found.
          */
         if (adjustmentValue <= COURSE_ADJUST_MIN_DAC_COUNT) {
            adjustmentDirection = APPLY_ADJUSTMENT_AND_TEST;        /* start search in a neutral direction */
            adjustmentValue = FINE_ADJUST_DAC_COUNT;                /* with a fine granularity adjustment  */

#ifndef LABVIEW_RT
            if (displayDacSettings) {
               printf ("\n");
            }
#endif
            do {
               /* For time "sample" usage, read the IRIG sample value.  Clear the "edge received" bit  
                *  in the DAC Control Register.  Delay 2 seconds if waiting for a full IRIG time sample, 
                * Then wait an additional 30 msec between double reads (just to be sure).
                */
               accessStatus = cei_write_32bit (board, dacControlRegPtr, dacValue);

               if (useActualTimeSample) {
                  status = cei_delay_milliseconds(ADJUST_DELAY_FOR_TIME_SAMPLE);
               }
                                                
#ifndef LABVIEW_RT
               if (displayDacSettings) {
                  printf("\rFine tuning %s transition -- DAC Setting = %3d", thresholdText[thresholdType], (CEI_NATIVE_INT)dacValue);
               }
#endif
               accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &irigRegValue);
               status = cei_delay_milliseconds(ADJUST_DELAY_30_MILLISECONDS);
               accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &irigRegValue);
               
               /* determine if the "sample received" condition has been met */
               if (useActualTimeSample) {
                  sampleReceived = (irigRegValue == prevIrigRegValue) ? FALSE : TRUE;
               }
               else {
                  sampleReceived = irigRegValue & CEIDEV_IRIG_EDGE_DETECT_BIT;
               }
         
               /* If a signal is detected after the most recent adjustment but not detected with the previous adjustment,
                * you have found the threshold and the tolerance to apply to that threshold is one count.
                * If a signal is not detected after the most recent adjustment but was detected with the previous adjustment,
                * you have again found the threshold but the tolerance to apply to that threshold is two counts.
                */
               if (sampleReceived) {
                  prevIrigRegValue = irigRegValue;
                  if (adjustmentDirection == NO_PREV_SIGNAL_DET_FINE_ADJUST) {
                     /* threshold was found, exit */
                     thresholdFound = TRUE;
                     *dacBoudaryCushionValue = SIGNAL_DETECT_THRESHOLD_CUSHION;
                  } 
                  else {
                     /* continue adjusting the DAC value until the signal is lost */
                     adjustmentDirection = PREV_SIGNAL_DETECT_FINE_ADJUST; 
                     if (thresholdType == AR_LOWER) {
                        dacValue -= adjustmentValue;
                     }
                     else {
                        dacValue += adjustmentValue;
                     }
                  }
               } 
               else {
                  if (adjustmentDirection == PREV_SIGNAL_DETECT_FINE_ADJUST) {
                     /* threshold was found, exit */
                     thresholdFound = TRUE;
                     *dacBoudaryCushionValue = SIGNAL_LOST_THRESHOLD_CUSHION;
                  } 
                  else {
                     /* continue adjusting the DAC value until the signal is found */
                     adjustmentDirection = NO_PREV_SIGNAL_DET_FINE_ADJUST; 
                     if (thresholdType == AR_LOWER) {
                        dacValue += adjustmentValue;
                     }
                     else {
                        dacValue -= adjustmentValue;
                     }
                  }
               }
            } while (thresholdFound == FALSE);
            
            break;
         }
      }
      else {
         /* no signal was detected, determine if we're still within limits of a valid "course adjustment" of the DAC */
         if ( ((thresholdType == AR_LOWER) && (dacValue >= voltageLimit)) || ((thresholdType == AR_UPPER) && (dacValue <= voltageLimit)) ) {
#ifndef LABVIEW_RT
            if (displayDacSettings) {
               printf("\nCould not find a %s transition -- DAC Setting = %3d\n", thresholdText[thresholdType], (CEI_NATIVE_INT)dacValue);
            }
#endif
            accessStatus = ARS_FAILURE;
         }
   
         /* Lost the signal but still within valid limits, start backing out the adjusted DAC value until the signal is found again */
         if (thresholdType == AR_LOWER) {
            dacValue += adjustmentValue;
         }
         else {
            dacValue -= adjustmentValue;
         }
         
      }

      /* for the second pass through the course adjustment, alter the per-pass DAC adjustment value differently 
       * based on which threshold we're trying to find... if attempting to find the lower threshold, reduce the 
       * adjustment to the current DAC by half each pass; otherwise for the upper threshold reduce the adjustment
       * to the current DAC by 2 counts each pass (note - not the DAC value, just the adjustment to the DAC value).
       */
      if (thresholdType == AR_LOWER) {
         adjustmentValue /= COURSE_ADJUST_DAC_COUNT;
      }
      else {
         adjustmentValue -= COURSE_ADJUST_DAC_COUNT;
      }

      /* limit the last "course adjustment" adjusted value to the current DAC value to 4 counts */
      if (adjustmentValue < COURSE_ADJUST_MIN_DAC_COUNT) {
         adjustmentValue = COURSE_ADJUST_MIN_DAC_COUNT;
      }
   }
   *finalDacValue = dacValue;
   return accessStatus;
}


/*===================================================================================================================*
 * ENTRY POINT:                    A 4 2 9 _ U T L _ I S _ I R I G _ C A L I B R A T E D
 *===================================================================================================================*
 *
 * FUNCTION:    Determine if the given board is synched to an IRIG signal.
 *
 * DESCRIPTION: Check if the IRIG base time has been updated in the last one second.  If it has, return; otherwise, 
 *              wait a second for the reference IRIG time-tag to update.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board  -- (input)  Board number (0 to 127) 
 * 
 * GLOBALS:     deviceGlobal[]
 *
 * RETURN VAL:  TRUE or ARS_NORMAL indicates valid IRIG signal reception
 *              ARS_INVHARVAL or other non-zero value indicates bad API status
 *              FALSE or 0 indicates failure to detect valid IRIG input
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_is_irig_calibrated (CDEV_BOARD_TYPE board)
{
   AR_TIMETAG_TYPE updateStart;                                                 /* initial IRIG update timetag       */
   AR_TIMETAG_TYPE updateCur;                                                   /* current IRIG update timetag       */
   TIME_TAG_TYPE timeDiff;                                                      /* delta time between samples        */
   CEI_INT32 delay_count;                                                       /* tracks delay iterations           */
   CEI_INT32 status = ARS_NORMAL;                                               /* operation status value            */
   CEI_INT32 calibrationStatus = FALSE;                                         /* return IRIG valid signal indicat. */

   /* read the time of the last IRIG update */
   status = ar_get_time (board, AR_TIMETAG_EXT_IRIG_REF_64BIT, &updateStart);
   
   /* not calibrated if the last IRIG update time is more than 1 second ago */
   if (status == ARS_NORMAL) {
      if (updateStart.timeTag == (TIME_TAG_TYPE)0) {
         for (delay_count = 0; (delay_count < 5) && (status == ARS_NORMAL) && (updateStart.timeTag == (TIME_TAG_TYPE)0); delay_count++) {
            /* IRIG not up and running yet, delay and try reading it again */
            ar_sleep (100);
            status = ar_get_time (board, AR_TIMETAG_EXT_IRIG_REF_64BIT, &updateStart);
         }
      }
      if ((status == ARS_NORMAL) && (updateStart.timeTag == (TIME_TAG_TYPE)0)) {
         ar_sleep (200);
         status = FALSE;
      }
      else {
         if (updateStart.referenceTimeTag >= updateStart.timeTag) {
            timeDiff = updateStart.referenceTimeTag - updateStart.timeTag;
         }
         else {
            timeDiff = (AR_MAX_TIMETAG_VALUE - updateStart.timeTag) + updateStart.referenceTimeTag;
         }
         if (timeDiff > (TIME_TAG_TYPE)1000900) {
            status = FALSE;
         }
      }
   }
   
   /* wait for one second to obtain a valid IRIG update */
   for (delay_count = 100; (delay_count > 0) && (status == ARS_NORMAL); delay_count--)
   {
      /* read the time of the last IRIG update */
      status = ar_get_time (board, AR_TIMETAG_EXT_IRIG_REF_64BIT, &updateCur);
   
      /* we are calibrated if there is one second between updates (+/- 1ms) */
      if (status == ARS_NORMAL) {
         if (updateCur.timeTag != updateStart.timeTag) {
            if (updateCur.timeTag >= updateStart.timeTag) {
               timeDiff = updateCur.timeTag - updateStart.timeTag;
            }
            else {
               timeDiff = (AR_MAX_TIMETAG_VALUE - updateStart.timeTag) + updateCur.timeTag;
            }
            if ( (timeDiff < (TIME_TAG_TYPE)1001000) && (timeDiff > (TIME_TAG_TYPE)997000) ) {
               calibrationStatus = TRUE;
               break;
            }
         }
      
         /* check if we timed out waiting for the board time to increment */
         if (updateCur.referenceTimeTag >= updateStart.referenceTimeTag) {
            timeDiff = updateCur.referenceTimeTag - updateStart.referenceTimeTag;
         }
         else {
            timeDiff = (AR_MAX_TIMETAG_VALUE - updateStart.referenceTimeTag) + updateCur.referenceTimeTag;
         }
            
         if (timeDiff > (TIME_TAG_TYPE)1000900) {
            status = FALSE;
         }
         else {
            status = cei_delay_milliseconds(10U);
         }
      }
   }

   if ((!calibrationStatus) && (status != ARS_NORMAL)) {
      calibrationStatus = status;
   }

   return calibrationStatus;
}


/*===================================================================================================================*
 * ENTRY POINT:           A 4 2 9 _ U T L _ Q U I C K _ A D J U S T _ I R I G _ T H R E S H O L D
 *===================================================================================================================*
 *
 * FUNCTION:    Performs a quick adjustment of the IRIG DAC for an external input IRIG signal, using signal edge 
 *              detection for verification of signal presence.
 *
 * DESCRIPTION: This routine sets the DAC to Vmin + (0.825 * (Vmax-Vmin)).  
 *              Vmin is the lower peak value and Vmax is the upper peak value.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board        -- (input)  Board number (0 to 127) 
 *              CEI_INT32 displayDacSettings -- (input)  when set enables DAC adjustment print to console
 *
 * GLOBALS:     deviceGlobal[]
 *
 * RETURN VAL:  CEI_INT32    -- Any value from 0 to x7FFFFFFF is the DAC value
 *                              ARS_FAILURE (-1) Any issue encountered with incoming IRIG signal level
 *
 *===================================================================================================================*/
CEI_INT32 a429_utl_quick_adjust_irig_threshold 
(
   CDEV_BOARD_TYPE board,                                             /* (input)  board to reference (0 to 127)      */
   CEI_INT32 displayDacSettings                         /* (input)  when set enables DAC adjustment print to console */
)
{
   CEI_UINT32 dacMaxVoltage = IRIG_DAC_MAX_VALUE;                               /* temp max DAC value for thold test */
   CEI_UINT32 dacValue;                                                            /* dynamic DAC value during test  */
   CEI_UINT32 initialDacValue = IRIG_DAC_ADJUST_INITIAL_COUNT;                                /* initial DAC value   */
   CEI_UINT32 initialDacAdjustment = INITIAL_SIGNAL_DETECT_ADJUST;                               /* initial Adjust   */
   CEI_UINT32 minimumVoltageValue = 0U;                                         /* lower bounds of final DAC value   */
   CEI_UINT32 maximumVoltageValue = 0U;                                         /* upper bounds of final DAC value   */
   CEI_UINT32 irigRegValue;                                                     /* temp contents of IRIG Receiver    */
   CEI_UINT32 dacBoudaryCushionValue = 0U;                                         /* minor adj to final DAC value   */
   CEI_INT32 signalFound = FALSE;                                               /* flag indicating IRIG Recvd time   */
   CEI_INT32 accessStatus = ARS_NORMAL;                                         /* status of board read/write access */
   CEI_INT32 status = ARS_NORMAL;                                               /* status of threshold detection     */
   pCEI_UINT32 dacControlRegPtr = &device[board].devicePtr[CEIDEV_DAC_CONTROL];
   pCEI_UINT32 globalEnableRegPtr = &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE];
    
   dacValue = IRIG_DAC_ADJUST_INITIAL_COUNT;

#ifndef LABVIEW_RT
   if (displayDacSettings) {
      printf ("Initial DAC Setting = %3d\n", (CEI_NATIVE_INT)dacValue);
   }
#endif
   
   /*=======================================================================
    * FIRST FIND THE INCOMING SIGNAL
    * Attempt to initially detect the incoming IRIG signal between the 
    * mid-upper peak by setting the DAC to 128 (0.0 Volts) and increasing
    * the threshold until the signal is found.
    *=======================================================================*/
   accessStatus = cei_write_32bit (board, dacControlRegPtr, dacValue);
   /* Wait 5 Msec for signal    */
   status = cei_delay_milliseconds(ADJUST_DELAY_FOR_EDGE_DETECT);
   
   /* Read IRIG value, clearing "edge received" bit, then  */
   /* Wait 30 Msec for an incoming signal edge transition. */
   accessStatus = cei_read_32bit (board, globalEnableRegPtr, &irigRegValue);
   status = cei_delay_milliseconds(ADJUST_DELAY_30_MILLISECONDS);
   
   /* Read IRIG value again, check the "edge received" bit to be set */
   accessStatus = cei_read_32bit (board, globalEnableRegPtr, &irigRegValue);
   
   if (irigRegValue & CEIDEV_IRIG_EDGE_DETECT_BIT) {
      signalFound = TRUE;
   }
   else {
      /* Signal wasn't immediately detected, raise the threshold until it's found */
      for (dacValue = IRIG_DAC_ADJUST_INITIAL_COUNT; (dacValue <= IRIG_DAC_MAX_VALUE) && (!signalFound); dacValue += INITIAL_SIGNAL_DETECT_ADJUST) {
         accessStatus = cei_write_32bit (board, dacControlRegPtr, dacValue);
         
#ifndef LABVIEW_RT
         if (displayDacSettings) {
            printf ("\rFinding the signal above the mid point -- DAC Setting = %3d", (CEI_NATIVE_INT)dacValue);
         }
#endif
         /* Read IRIG value, clearing "edge received" bit, then  */
         /* Wait 30 Msec for an incoming signal edge transition. */
         status = cei_delay_milliseconds(ADJUST_DELAY_FOR_EDGE_DETECT);
         accessStatus = cei_read_32bit (board, globalEnableRegPtr, &irigRegValue);
         status = cei_delay_milliseconds(ADJUST_DELAY_30_MILLISECONDS);
         accessStatus = cei_read_32bit (board, globalEnableRegPtr, &irigRegValue);

         if (irigRegValue & CEIDEV_IRIG_EDGE_DETECT_BIT) {
            dacMaxVoltage = IRIG_DAC_MAX_VALUE;
            initialDacValue = dacValue;
            initialDacAdjustment = INITIAL_SIGNAL_DETECT_NO_SIG_ADJUST;
            signalFound = TRUE;
         }
      }

#ifndef LABVIEW_RT
      if (displayDacSettings) {
         printf ("\n");
      }
#endif
      /* Signal wasn't detected by raising the threshold? Lower it now until it's found */
      for (dacValue = IRIG_DAC_ADJUST_INITIAL_COUNT; (dacValue > 8U) && (!signalFound); dacValue -= INITIAL_SIGNAL_DETECT_ADJUST) {
         accessStatus = cei_write_32bit (board, dacControlRegPtr, dacValue);

#ifndef LABVIEW_RT
         if (displayDacSettings) {
            printf ("\rFinding the signal below the mid point -- DAC Setting = %3d", (CEI_NATIVE_INT)dacValue);
         }
#endif
         /* Read IRIG value, clearing "edge received" bit, then  */
         /* Wait 30 Msec for an incoming signal edge transition. */
         status = cei_delay_milliseconds(ADJUST_DELAY_FOR_EDGE_DETECT);
         accessStatus = cei_read_32bit (board, globalEnableRegPtr, &irigRegValue);
         status = cei_delay_milliseconds(ADJUST_DELAY_30_MILLISECONDS);
         accessStatus = cei_read_32bit (board, globalEnableRegPtr, &irigRegValue);
         
         if (irigRegValue & CEIDEV_IRIG_EDGE_DETECT_BIT) {
            dacMaxVoltage = IRIG_DAC_ADJUST_INITIAL_COUNT;
            initialDacAdjustment = INITIAL_SIGNAL_DETECT_NO_SIG_ADJUST;
            initialDacValue = dacValue;
            signalFound = TRUE;
         }
      }

      if (!signalFound) {
#ifndef LABVIEW_RT
         if (displayDacSettings) {
            printf ("\nNever found a signal -- final DAC Setting = %3d\n", (CEI_NATIVE_INT)dacValue);
         }
#endif
         status = ARS_FAILURE;
      }
   }

   if (status == ARS_NORMAL) {
      /*=======================================================================
       * SECOND FIND THE LOW THRESHOLD AT WHICH SIGNAL BECOMES PRESENT
       * Starting with the theshold DAC at level the signal was detected above,
       * decrement the DAC voltage until the signal is no longer present;
       * then perform a fine adjustment until DAC setting is just above
       * the lower threshold where the signal becomes present.
       *=======================================================================*/
      dacValue = initialDacValue;
      status = a429_utl_find_dac_threshold (board, 
                                            FALSE, 
                                            displayDacSettings, 
                                            AR_LOWER, 
                                            dacMaxVoltage, 
                                            initialDacAdjustment, 
                                            &dacValue, 
                                            &dacBoudaryCushionValue); 
   }
   
   if (status == ARS_NORMAL) {
   
      minimumVoltageValue = dacValue + dacBoudaryCushionValue;
   
#ifndef LABVIEW_RT
      if (displayDacSettings) {
         printf ("\nLower transition final value = %3d\n", (CEI_NATIVE_INT)minimumVoltageValue);
      }
#endif
      /*=======================================================================
       * THIRD FIND THE AR_UPPER THRESHOLD AT WHICH SIGNAL BECOMES LOST
       * Starting with the theshold DAC at level the signal was detected first,
       * increment the DAC voltage until the signal is no longer present;
       * then perform a fine adjustment until DAC setting is just below
       * the upper threshold where the signal is lost.
       *=======================================================================*/
      dacValue = initialDacValue;
      status = a429_utl_find_dac_threshold (board, 
                                            FALSE, 
                                            displayDacSettings, 
                                            AR_UPPER, 
                                            minimumVoltageValue, 
                                            initialDacAdjustment, 
                                            &dacValue, 
                                            &dacBoudaryCushionValue); 
   }
   
   if (status == ARS_NORMAL) {
      maximumVoltageValue = dacValue - dacBoudaryCushionValue;
   
#ifndef LABVIEW_RT
      if (displayDacSettings) {
         printf ("\nUpper transition final value = %3d\n", (CEI_NATIVE_INT)maximumVoltageValue);
      }
#endif

      /* validate the span between the upper and lower thresholds */
      if ((maximumVoltageValue < minimumVoltageValue) || ((maximumVoltageValue-minimumVoltageValue) < MIN_DAC_SPAN)) {
#ifndef LABVIEW_RT
         if (displayDacSettings) {
            printf ("Error, Signal Detection Range is invalid, Vmax = %d  Vmin = %d\n", (CEI_NATIVE_INT)maximumVoltageValue, (CEI_NATIVE_INT)minimumVoltageValue);
         }
#endif
         accessStatus = cei_write_32bit (board, dacControlRegPtr, IRIG_DAC_ADJUST_INITIAL_COUNT);
         status = ARS_FAILURE;
      }
   }

   /* write the computed DAC counts value to the DAC control register, then convert to millivolts */
   if (status == ARS_NORMAL) {
      dacValue = minimumVoltageValue + (CEI_UINT32)((86*(maximumVoltageValue-minimumVoltageValue))/100);
      accessStatus = cei_write_32bit (board, dacControlRegPtr, dacValue);

#ifndef LABVIEW_RT
      if (displayDacSettings) {
         printf ("Final DAC Setting = %3d\n", (CEI_NATIVE_INT)dacValue);
      }
#endif
      if (dacValue <= CEIDEV_IRIG_DAC_OFFSET) {
         device[board].irigDacThreshold = 0U;
         dacValue = 0U;
      }
      else {
         dacValue = (CEI_UINT32)(CEIDEV_IRIG_DAC_MV_CONVERSION * (float)(dacValue - CEIDEV_IRIG_DAC_OFFSET));
         device[board].irigDacThreshold = (CEI_INT32)dacValue;
      }
   }

   if (status == ARS_NORMAL) {
      if (accessStatus == ARS_NORMAL) {
         status = (CEI_INT32)(dacValue);
      }
      else {
         status = accessStatus;
      }
   }
   return status;
}



