/*====================================================================================================================*
 * FILE:                                     C D E V _ A P I _ S C H E D. C                    
 *====================================================================================================================*
 *
 *   COPYRIGHT 2016 - 2019 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED
 *   ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH THE INCLUSION
 *   OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY OTHER COPIES THEREOF
 *   MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE
 *   AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS, INC. 
 *
 *====================================================================================================================*
 *
 * FUNCTION:    The Message Scheduler Routines portion of the Application Programmer's Interface for a CEI-x30 ARINC device.
 *
 * DESCRIPTION: These routines are provided to demonstrate the message scheduler-specific programming primitives for 
 *              the CEI-x30 API using the C programming language.  Each routine implements a particular API function; 
 *              however, a user application program may need to combine functions in different ways to perform the 
 *              desired operation.
 *
 *              Most routines utilize a board number parameter, zero-based on the Device ID/Board Number assigned 
 *              during the installation of CEI-x30-SW distribution.  Channel parameter values are all zero-based.
 *
 * EXTERNALLY DEFINED ROUTINE ENTRY POINTS:         
 *
 *    CEI-x30 ARINC 429 Message Scheduler related API routines:
 *
 *    ar_assign_scheduler_start_offsets    - defines start offsets for a best estimate tx scenario without rate skew
 *    ar_define_msg                        - defines a message for scheduled transmission
 *    ar_define_msg_block                  - defines a block of messages for scheduled transmission
 *    ar_define_msg_block_lv               - defines a block of messages for scheduled transmission (for LabVIEW)
 *    ar_modify_msg                        - modifies a message value (scheduled transmission)
 *    ar_modify_msg_block                  - modifies a block of message values (scheduled transmission)
 *    ar_modify_msg_block_lv               - modifies a block of message values (scheduled transmission)  (for LabVIEW)
 *    ar_read_message_schedule_table_entry - reads a raw message scheduler table entry from the table
 *    ar_update_message_block              - updates only the message value for successive multiple scheduer entries
 *
 * INTERNAL UTILITY ROUTINES:
 *
 *    a429_utl_store_message_scheduler_entry
 *    a429_utl_validate_scheduled_message_parameters
 *
 *--------------------------------------------------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date    By   Ver. Comments
 * -------- ----- ---- -----------------------------------------------------------------------------------------------
 * 10/18/16  BDW  4.90 Initial version.  For modification history related to the routines contained herein prior to
 *                     and including v4.90, see the history section in the header of the file CDEV_API.C.
 * 01/22/18  BDW  5.10 Added verification of a valid open session in ar_assign_scheduler_start_offsets before accessing
 *                     the message scheduler table.
 *                     Added passed parameter NULL pointer reference check in ar_update_message_block.
 *====================================================================================================================*/ 
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
   #include <unistd.h>
   #include <fcntl.h>
   #include <string.h>
   #include <pthread.h>
   #include <errno.h>
   #include <sys/stat.h> 
   #include <sys/shm.h>
   #include <sys/sem.h>
   #include <sys/syscall.h>
   #define VOLATILE volatile
#elif defined (VXW_PCI_X86) || defined (VXW_PCI_PPC)
   #include <stdio.h>
   #include <stdlib.h>
   #include <unistd.h>
   #define VOLATILE 
#elif defined (NIVISA_PXI)
   #ifdef LABVIEW_RT
      #include <ansi_c.h> 
   #endif
   #ifdef CVI
      #include <ansi_c.h> 
   #endif
   #include <visa.h>  
   #define VOLATILE 
#elif defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
   #include <string.h>
   #define VOLATILE 
#else
   #define VOLATILE volatile
#endif

#include "cdev_glb.h"

extern VOLATILE ARINC_LOCAL_TYPE device[MAX_BOARDS];
extern ARINC_GLOBAL_TYPE * deviceGlobal[MAX_BOARDS];

/*===================================================================================================================*
 * ENTRY POINT:            A 4 2 9 _ U T L _ S T O R E _ M E S S A G E _ S C H E D U L E R _ E N T R Y
 *===================================================================================================================*
 *
 * FUNCTION:    Define or modify a single message scheduler table entry
 *
 * PARAMETERS:  CEI_INT32 modifyEntryFlag -- (input) indicates whether to create a new scheduled message 
 *                                                   or modify an existing message: FALSE - new : TRUE - modify
 *              CEI_INT32 numberOfEntries -- (input) the number of entries to define or modify, used as a
 *                                                   reference for the number of structure elements defined
 *                                                   via the messageEntry pointer
 *              pAR_SCHEDULED_MSG_ENTRY_TYPE
 *                        messageEntry    -- (input) structure of message content defined as follows:
 *
 *                                              struct {
 *                                                        CEI_UINT32 messageIndex;                  
 *                                                        CEI_UINT32 board;                         
 *                                                        CEI_UINT32 channel;                       
 *                                                        CEI_UINT32 rate;                          
 *                                                        CEI_UINT32 start;                         
 *                                                        CEI_UINT32 txCount;               
 *                                                        CEI_UINT32 data; 
 *                                                     }
 * RETURN VAL:  CEI_INT32  -- ARS_NORMAL      Success
 *                            ARS_FAILURE     message scheduler table full
 *
 * GLOBALS:     device[]
 * 
 * DESCRIPTION: This routine defines the supplied message(s) in the next available entires within the device 
 *              message scheduler table.
 * 
 *===================================================================================================================*/ 
CEI_INT32 a429_utl_store_message_scheduler_entry 
(
   CEI_INT32 modifyEntryFlag,                                                /* (input) modify existing entries flag */
   CEI_INT32 numberOfEntries,                                                /* (input) number of entries to process */
   pAR_SCHEDULED_MSG_ENTRY_TYPE messageEntry                                 /* (input) ptr to struct array of msgs  */
)
{
   CDEV_BOARD_TYPE board;                                                     /* temp board index                    */
   CEI_INT32 msgTableCounter;                                                 /* counter for multiple msg entry/mod  */
   CEI_UINT32 msgIndex = 0;                                                   /* temp index for table search         */
   CEI_UINT32 msgRate;                                                        /* temp table entry rate setting       */
   CEI_UINT32 localStart;                                                     /* temp table entry start setting      */
   CEI_UINT32 localRate;                                                      /* temp table entry rate setting       */
   CEI_UINT32 appProgTimerResolution;                                         /* Application specified time-tag res  */
   CEI_UINT32 currentRateValue;                                               /* modify entry current rate value     */
   CEI_UINT32 currentTxCountValue;                                            /* modify entry current tx cnt value   */
   CEI_UINT32 currentChannelValue;                                            /* modify entry current channel value  */
   CEI_INT32 status = ARS_NORMAL;                                             /* function return status              */
   CEI_INT32 accessStatus = ARS_NORMAL;                                       /* status of device read/write access  */
   CEI_INT32 disableMessage = FALSE;                                          /* flag indicatn need to disable msg   */
   CEI_UINT32 msgEntryOffset;                                                 /* local offset to Msg Sched table     */
   CEI_UINT32 memOffset;                                                      /* local offset to Msg Sched table     */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr = NULL;                   /* pointer to board local structure    */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr = NULL;                          /* pointer to board global structure   */
   pAR_SCHEDULED_MSG_ENTRY_TYPE messageEntryPtr = messageEntry;               /* pointer to applic msg stuct array   */
   
   for (msgTableCounter = 0; msgTableCounter < numberOfEntries; msgTableCounter++) {
      board = (CDEV_BOARD_TYPE)messageEntryPtr->board;
      deviceGlobalStructPtr = deviceGlobal[board];
      deviceLocalStructPtr = &device[board];

      /* if this invocation is to modify an existing entry, bypass search for first available */
      if (modifyEntryFlag) {
         msgIndex = messageEntryPtr->messageIndex;
         memOffset = CEIDEV_MSG_SCHED_TABLE_START + (msgIndex * CEIDEV_MSG_SCHED_TBL_ENTRY_SIZE);
         
         /* check for a change to this entry's rate value */
         accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset++], &currentRateValue);

         /* for non-zero rates, check for CEI-x20 timer compatibility */
         localRate = messageEntryPtr->rate;
         if (localRate && (deviceGlobalStructPtr->deviceTimeTagReference == AR_TIMER_X20_COMPAT_32BIT)) {
            appProgTimerResolution = deviceGlobalStructPtr->deviceX20CompatibleResolution;
            if (appProgTimerResolution >= AR_X20_QTR_MICROSEC_PER_MSEC) {
               localRate *= appProgTimerResolution/AR_X20_QTR_MICROSEC_PER_MSEC;
            }
         }
         /* when modifying an existing entry, disable the message to alter any
          * attribute (rate or channel) other than just the data value.
          */
         if (localRate != currentRateValue) {
            disableMessage = TRUE;
         }
         else {
            /* check for a change to this entry's channel value */
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &currentChannelValue);
            if (messageEntryPtr->channel != currentChannelValue) {
               disableMessage = TRUE;
            }
         }
      }
      else {
         /* Search for the first unused message scheduler table entry (rate = 0) */
         memOffset = CEIDEV_MSG_SCHED_TABLE_START;
         for (msgIndex = 0; msgIndex < deviceGlobalStructPtr->messageSchedulerTableSize; msgIndex++ ) {
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &msgRate);
            if ((accessStatus != ARS_NORMAL) || (msgRate == 0)) {
               break;
            }
            else {
               memOffset += CEIDEV_MSG_SCHED_TBL_ENTRY_SIZE;
            }
         }
      }
      
      if ((accessStatus == ARS_NORMAL) && (msgIndex < deviceGlobalStructPtr->messageSchedulerTableSize)) {
         msgEntryOffset = CEIDEV_MSG_SCHED_TABLE_START + (msgIndex * CEIDEV_MSG_SCHED_TBL_ENTRY_SIZE);
         memOffset = msgEntryOffset;
         if (disableMessage) {
            /* first disable the message by zeroing the entry rate */
            accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], (CEI_UINT32)0);
         }
         
         /* assign this entry's channel */
         memOffset++;
         accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], messageEntryPtr->channel);
      
         /* only allow modification of the transmit count if this message was created
          * with a limited transmit count and not in an active f/w countdown condition */
         /* assign this entry's transmit count */
         memOffset++;
         accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &currentTxCountValue);
         if (!modifyEntryFlag || (modifyEntryFlag && (currentTxCountValue == 0U)) ) {
            accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], messageEntryPtr->txCount);
         }
         
         /* only allow assignment of the start offset at message creation */
         if (!modifyEntryFlag) {
            /* assign this entry's start offset */
            localStart = messageEntryPtr->start;
            
            /* check for CEI-x20 timer compatibility */
            if (deviceGlobalStructPtr->deviceTimeTagReference == AR_TIMER_X20_COMPAT_32BIT) {
               appProgTimerResolution = deviceGlobalStructPtr->deviceX20CompatibleResolution;
               if (appProgTimerResolution < AR_X20_QTR_MICROSEC_PER_MSEC) {
                  localStart = 0U;
               }
               else {
                  appProgTimerResolution /= AR_X20_QTR_MICROSEC_PER_MSEC;
                  localStart *= appProgTimerResolution;
               }
            }
            memOffset++;
            accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], localStart);
         }

         /* assign this entry's message value */
         memOffset = msgEntryOffset + CEIDEV_MSG_SCHED_TBL_DATA_OFFSET;
         accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], messageEntryPtr->data);
            
         /* after all other entries are written, now assign this entry's periodic rate */
         localRate = messageEntryPtr->rate;
            
         /* check for CEI-x20 timer compatibility */
         if (deviceGlobalStructPtr->deviceTimeTagReference == AR_TIMER_X20_COMPAT_32BIT) {
            appProgTimerResolution = deviceGlobalStructPtr->deviceX20CompatibleResolution;
            if (appProgTimerResolution >= AR_X20_QTR_MICROSEC_PER_MSEC) {
               appProgTimerResolution /= AR_X20_QTR_MICROSEC_PER_MSEC;
               localRate *= appProgTimerResolution;
            }
         }
         accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[msgEntryOffset], localRate);
         
         /* on creation of a new message, assign the message index to the structure member supplied by the calling routine */
         if ((!modifyEntryFlag) && (accessStatus == ARS_NORMAL)) {
            messageEntryPtr->messageIndex = msgIndex;
         }
      }
      else {   /* either the message index is invalid or the scheduler table is full */
         status = ARS_FAILURE;
      }
      
      if ((accessStatus != ARS_NORMAL) || (status != ARS_NORMAL)) {
         break;
      }
      else {  /* when defining an array of entries, increment the pointer to the next structure */
         messageEntryPtr++;
      }
   }
   
   if (accessStatus != ARS_NORMAL) {
      status = accessStatus;
   }
   
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:     A 4 2 9 _ U T L _V A L I D A T E _ S C H E D U L E D _ M E S S A G E _ P A R A M E T E R S
 *===================================================================================================================*
 *
 * FUNCTION:    Validate the parameters for a scheduled transmission request.
 *
 * PARAMETERS:  CEI_UINT32 board     -- (input)  Board number (0 to 127) 
 *              CEI_UINT32 channel   -- (input)  ARINC 429 Tx channel selection, valid range is 0 to N, where N 
 *                                               refers to the ARINC 561 channel, and N-1 refers to the last 429 
 *                                               Tx channel installed.
 *              pCEI_INT32 hwChannel -- (output) resp. h/w-based channel number
 *
 * RETURN VAL:  CEI_INT32    -- ARS_NORMAL      Success
 *                              ARS_INVBOARD    Invalid or uninitialized board
 *                              ARS_INVARG      Invalid channel
 *                              ARS_INVHARVAL   Message scheduling is not supported on this device
 *
 * GLOBALS:     device[]
 * 
 * DESCRIPTION: This routine validates the board and channel parameter values for a host-requested message 
 *              scheduling function.  It then returns the 429-based channel-index-referenced hardware-based channel 
 *              number for use by the calling routine.
 * 
 *===================================================================================================================*/ 
CEI_INT32 a429_utl_validate_scheduled_message_parameters 
(
   CEI_UINT32 board,                                                  /* (input)  Board number (0 to 127)            */
   CEI_UINT32 channel,                                                /* (input)  API-based 429 Tx channel selection */
   pCEI_UINT32 hwChannel                                              /* (output) resp. h/w-based channel number     */
)
{
   CEI_INT32 status = ARS_NORMAL;                                                        /* function return status   */
   CEI_UINT32 localChannel = channel;                                                    /* local channel select     */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                                         /* ptr to brd global structure */

   /* verify the channel parameter validity - must be of type 5 or 6 (enhanced 429 or 561) */
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   else {
      deviceGlobalStructPtr = deviceGlobal[board];

      /* For current configurations, any message scheduler request is limited
       * to the number of enhanced ARINC 429 channels installed on the device
       */
      if (channel > deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX]) {
         status = ARS_INVARG;
      }
      /* If the channel requested in one larger than the last ARINC 429 channel,
       * assume the request is to schedule a message on the enhanced ARINC 561 Receive Channel.
       */
      else if (channel == deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX]) {
         if (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC561_TX] != CEIDEV_MAX_CHANNELS) {
            localChannel = deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC561_TX];
         }
         else {
            status = ARS_INVARG;
         }
      }
      else {
         localChannel = channel + deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
      }
   
      if (status == ARS_NORMAL) {
         status = (CEI_INT32)a429_utl_validate_channel_request ((CDEV_BOARD_TYPE)board, localChannel, ARU_ADD_SCHEDULED_MSG);
         if (status == ARS_NORMAL) {
            *hwChannel = localChannel;
         }
      }
   }
   
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                                A R _ D E F I N E _ M S G
 *===================================================================================================================*
 *
 * FUNCTION:    Define a message on the device for scheduled transmission.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board    -- (input) Board number (0 to 127)  
 *              CDEV_CHAN_TYPE channel   -- (input) ARINC 429 transmit channel selection.  Valid range is 0 to N, 
 *                                                  where N-1 refers to the last ARINC 429 transmit channel installed.
 *              CDEV_PARM_SSI_TYPE rate  -- (input) Message transmission rate (in milliseconds).
 *              CDEV_PARM_USI_TYPE start -- (input) Initial delay from the current time (or "scheduler enabled start
 *                                                  time") until the time this message will be transmitted for the 
 *                                                  first time (subsequent transmissions at "rate"), in milliseconds.
 *              CDEV_PARM_SI_TYPE data   -- (input) The ARINC message to transmit.
 *
 * RETURN VAL:  ARS_FAILURE    One of the following errors occurred:
 *                                board number out of range
 *                                board not loaded and initialized
 *                                an invalid channel was specified 
 *                                the maximum number of messages has already been defined
 *                                feature is not supported by the hardware
 *
 *              Message Index  a return value in the range of 0 to 2047 (or 1023 for the EOL'd CEI-830 only)
 *                             indicates success, to be used by the host application as the unique message number
 *                             index assigned to this message.
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 *
 * DESCRIPTION: This routine defines an ARINC 429 message for periodic retransmission, causing the message to be 
 *              transmitted at a specific rate.  Once a message has been defined, it can be modified by calling 
 *              AR_MODIFY_MSG.  Messages can be defined anytime after ar_open has been invoked.
 *
 *              Message rate skew is an ever present issue when defining a scheduled message scenario on an individual
 *              transmit channel.  The start parameter allows the application to prioritize high rate messages so they
 *              experience the least rate skew over low rate messages.  Utilize ar_assign_scheduler_start_offsets to
 *              assign the best start offset values to avoid rate skew on individual transmit channels.
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_define_msg 
(
   CDEV_BOARD_TYPE board,                                                  /* (input) Board number (0 to 127)        */
   CDEV_CHAN_TYPE channel,                                                 /* (input) 429 Tx channel selection       */
   CDEV_PARM_SSI_TYPE rate,                                                /* (input) transmission rate (in msec)    */
   CDEV_PARM_USI_TYPE start,                                               /* (input) initial delay from the current */
   CDEV_PARM_SI_TYPE data                                                  /* (input) the ARINC word to transmit     */
)
{
   CEI_UINT32 localChannel = 0U;                                                        /* transmit channel select   */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                               /* function return status    */
   AR_SCHEDULED_MSG_ENTRY_TYPE msgEntry;                                                /* message content structure */

#ifdef TARGET_EMBEDDED
   localChannel = (CEI_UINT32)channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
   msgEntry.messageIndex = (CEI_UINT32)0;
#else
   CEI_INT32 tempStatus = ARS_NORMAL;                                                   /* local temp return status  */
   msgEntry.messageIndex = (CEI_UINT32)0;

   /* verify calling parameter validity for h/w message scheduling */
   tempStatus = a429_utl_validate_scheduled_message_parameters ((CEI_UINT32)board, (CEI_UINT32)channel, &localChannel);

   if (tempStatus == ARS_NORMAL) {
      /* if multi-thread protection is active, acquire a lock on the host interface */
      if (device[board].multiThreadProtect != AR_CONC_NONE) {
         status = (CDEV_API_RET_TYPE)cei_acquire_thread_lock(LOCK_IDENT);
      }
         
      if (status == ARS_NORMAL) {
#endif
         msgEntry.board        = (CEI_UINT32)board;
         msgEntry.channel      = localChannel;
         msgEntry.rate         = (CEI_UINT32)rate;
         msgEntry.start        = (CEI_UINT32)start;
         msgEntry.txCount      = (CEI_UINT32)CEIDEV_MSG_SCHEDULER_TX_ALWAYS;
         msgEntry.data         = (CEI_UINT32)data;
         
         /* define a single scheduled message entry */
         status = (CDEV_API_RET_TYPE)a429_utl_store_message_scheduler_entry (FALSE, 1, &msgEntry);
   
#ifndef TARGET_EMBEDDED
         /* if multi-thread protection is active, free any acquired lock */
         if (device[board].multiThreadProtect != AR_CONC_NONE) {
            if (status == ARS_NORMAL) {
               status = (CDEV_API_RET_TYPE)cei_release_thread_lock(LOCK_IDENT);
            }
            else {
               tempStatus = cei_release_thread_lock(LOCK_IDENT);
            }
         }
      }
   }
   if ((status == ARS_NORMAL) && (tempStatus == ARS_NORMAL)) {
#else   
   if (status == ARS_NORMAL) {
#endif
      status = (CDEV_API_RET_TYPE)msgEntry.messageIndex;
   }
   else {
      status = ARS_FAILURE;
   }

   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                             A R _ D E F I N E _ M S G _ B L O C K
 *===================================================================================================================*
 *
 * FUNCTION:    Define a block of messages for scheduled transmission on the device.
 *
 * PARAMETERS:  CEI_INT32 numberOfEntries -- (input) Defines the number of scheduler entries to create on the device,
 *                                                   used as a reference for the number of structure array elements 
 *                                                   defined via the messageEntry pointer.
 *              pAR_SCHEDULED_MSG_ENTRY_TYPE
 *                        messageEntry    -- (input) array of structures of message properties, each element defined 
 *                                                   as follows:
 *                        struct 
 *                        {
 *                           CEI_UINT32 messageIndex;  Message scheduler table index for this msg assigned by the API
 *                           CEI_UINT32 board;         Board number (0 to 127)                
 *                           CEI_UINT32 channel;       ARINC 429 transmit channel selection                
 *                           CEI_UINT32 rate;          Message transmission rate (in milliseconds)                
 *                           CEI_UINT32 start;         Initial delay until message transmission (in milliseconds) 
 *                           CEI_UINT32 txCount;       Message transmission count (0xFFFFFFFF for infinite count)
 *                           CEI_UINT32 data;          ARINC 429 message
 *                        };
 *                                                    
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL      Success
 *                                ARS_INVBOARD    Invalid or uninitialized board
 *                                ARS_INVARG      Invalid number of entries argument
 *                                ARS_INVHARVAL   Invalid transmit channel entry
 *                                ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                                ARS_FAILURE     The number of entries requested exceeds the available number
 *                                                of remaining table entries
 * 
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: This routine defines a block of ARINC 429 messages for periodic retransmission, causing the messages 
 *              to be transmitted at specific rates.  Once the message block has been defined, message values can be 
 *              modified by calling AR_MODIFY_MSG_BLK. Messages can be defined anytime after ar_open has been invoked.
 *
 *              Message rate skew is an ever present issue when defining a scheduled message scenario on an individual
 *              transmit channel.  The start member allows the application to prioritize high rate messages so they
 *              experience the least rate skew over low rate messages.  Utilize ar_assign_scheduler_start_offsets to
 *              assign the best start offset values to avoid rate skew on individual transmit channels.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_define_msg_block
(
   CEI_INT32 numberOfEntries,                                                  /* (input) number of entries          */
   pAR_SCHEDULED_MSG_ENTRY_TYPE messageEntry                                   /* (input) ptr to struct of messages  */
)
{
   CDEV_BOARD_TYPE board;                                                       /* board id extracted from structure */
   CEI_INT32 arrayIndex;                                                             /* scheduler access array index */
   CEI_UINT32 localChannel = 0U;                                                       /* transmit channel select    */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                              /* function return status     */
#ifndef TARGET_EMBEDDED                                                            
   CEI_INT32 tempStatus = ARS_NORMAL;                                                  /* local temp return status   */
#endif
   AR_SCHEDULED_MSG_ENTRY_TYPE msgEntry;                                               /* message content structure  */
   pAR_SCHEDULED_MSG_ENTRY_TYPE msgPtr = messageEntry;                           /* local ptr to msg structure array */

   msgEntry.messageIndex = (CEI_UINT32)0;
   if ((numberOfEntries == 0) || (numberOfEntries >= MAX_SCHEDULED_MSGS)) {
      status = ARS_INVARG;
   }
   /* for a valid set of messages, insert the block of entries into the message scheduler table */
   for (arrayIndex = 0; (arrayIndex < numberOfEntries) && (status == ARS_NORMAL); arrayIndex++) {
      board = (CDEV_BOARD_TYPE)msgPtr->board;

#ifdef TARGET_EMBEDDED
      localChannel = msgPtr->channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
#else
      /* verify calling parameter validity for h/w message scheduling */
      tempStatus = a429_utl_validate_scheduled_message_parameters ((CEI_UINT32)board, msgPtr->channel, &localChannel);

      if (tempStatus == ARS_NORMAL) {
         /* if multi-thread protection is active, acquire a lock on the host interface */
         if (device[board].multiThreadProtect != AR_CONC_NONE) {
            status = (CDEV_API_RET_TYPE)cei_acquire_thread_lock(LOCK_IDENT);
         }
            
         if (status == ARS_NORMAL) {
#endif
            msgEntry.messageIndex = 0;
            msgEntry.board        = board;
            msgEntry.channel      = localChannel;
            msgEntry.rate         = msgPtr->rate;
            msgEntry.start        = msgPtr->start;
            msgEntry.txCount      = msgPtr->txCount;
            msgEntry.data         = msgPtr->data;
            
            /* define a single scheduled message entry */
            status = (CDEV_API_RET_TYPE)a429_utl_store_message_scheduler_entry (FALSE, 1, &msgEntry);
            
            /* assign the message index in the app-supplied structure array entry */
            msgPtr->messageIndex = msgEntry.messageIndex;
            msgPtr++;
      
#ifndef TARGET_EMBEDDED
            /* if multi-thread protection is active, free any acquired lock */
            if (device[board].multiThreadProtect != AR_CONC_NONE) {
               if (status == ARS_NORMAL) {
                  status = (CDEV_API_RET_TYPE)cei_release_thread_lock(LOCK_IDENT);
               }
               else {
                  tempStatus = cei_release_thread_lock(LOCK_IDENT);
               }
            }
         }
      }
      else {
         status = (CDEV_API_RET_TYPE)tempStatus;
      }
#endif
   }
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                     A R _ D E F I N E _ M S G _ B L O C K _ L V
 *===================================================================================================================*
 *
 * FUNCTION:    Define a block of messages for scheduled transmission, specifically included in support of
 *              LabVIEW structure array formatting
 *
 * PARAMETERS:  CEI_INT32 numberOfEntries -- (input) Defines the number of scheduler entries to create on the device,
 *                                                   used as a reference for the number of structure array elements 
 *                                                   defined via the messageEntry pointer.
 *
 *              pLV_AR_SCHEDULED_MSG_ENTRY_TYPE messageEntry
 *                                                    
 *                  CEI_UINT32 array_size -- (input) Size of the messageEntries array (in bytes), and should always
 *                                                   be assigned the value 57344 (2048 32-bit x 7 entries).  This 
 *                                                   parameter is unused outside of the respective LabVIEW wrapper.
 *                                                    
 *                  AR_SCHEDULED_MSG_ENTRY_TYPE msgEntries[2048]  
 *                                        -- (input) A preallocated array of structures of message properties,  
 *                                                   each element defined as follows:
 *                        struct 
 *                        {
 *                           CEI_UINT32 messageIndex;  Message scheduler table index for this msg assigned by the API
 *                           CEI_UINT32 board;         Board number (0 to 127)                
 *                           CEI_UINT32 channel;       ARINC 429 transmit channel selection                
 *                           CEI_UINT32 rate;          Message transmission rate (in milliseconds)                
 *                           CEI_UINT32 start;         Initial delay until message transmission (in milliseconds)
 *                           CEI_UINT32 txCount;       Message transmission count (0xFFFFFFFF for infinite count)
 *                           CEI_UINT32 data;          ARINC 429 message
 *                        };
 *                
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL      Success
 *                                ARS_INVBOARD    Invalid or uninitialized board
 *                                ARS_INVARG      Invalid number of entries argument
 *                                ARS_INVHARVAL   Invalid transmit channel entry
 *                                ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                                ARS_FAILURE     The number of entries requested exceeds the available number
 *                                                of remaining table entries
 * 
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: This routine defines a block of ARINC 429 messages for periodic retransmission, causing the messages 
 *              to be transmitted at specific rates.  Once the message block has been defined, message values can be 
 *              modified by calling AR_MODIFY_MSG_BLK. Messages can be defined anytime after ar_open has been invoked.
 *
 *              Message rate skew is an ever present issue when defining a scheduled message scenario on an individual
 *              transmit channel.  The start member allows the application to prioritize high rate messages so they
 *              experience the least rate skew over low rate messages.  Utilize ar_assign_scheduler_start_offsets to
 *              assign the best start offset values to avoid rate skew on individual transmit channels.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_define_msg_block_lv
(
   CEI_INT32 numberOfEntries,                                                           /* (input) number of entries */
   pLV_AR_SCHEDULED_MSG_ENTRY_TYPE messageEntry                                         /* (input) ptr to msg array  */
)
{
   CDEV_BOARD_TYPE board;                                                       /* board id extracted from structure */
   CEI_INT32 arrayIndex;                                                             /* scheduler access array index */
   CEI_UINT32 localChannel = 0U;                                                       /* transmit channel select    */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                              /* function return status     */
#ifndef TARGET_EMBEDDED                                                             
   CEI_INT32 tempStatus = ARS_NORMAL;                                                  /* local temp return status   */
#endif
   AR_SCHEDULED_MSG_ENTRY_TYPE msgEntry;                                               /* message content structure  */
   pAR_SCHEDULED_MSG_ENTRY_TYPE msgPtr = messageEntry->msgEntries;               /* local ptr to msg structure array */

   msgEntry.messageIndex = (CEI_UINT32)0;
   if ((numberOfEntries == 0) || (numberOfEntries >= MAX_SCHEDULED_MSGS)) {
      status = ARS_INVARG;
   }
   /* for a valid set of messages, insert the block of entries into the message scheduler table */
   for (arrayIndex = 0; (arrayIndex < numberOfEntries) && (status == ARS_NORMAL); arrayIndex++) {
      board = (CDEV_BOARD_TYPE)msgPtr->board;

#ifdef TARGET_EMBEDDED
      localChannel = msgPtr->channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
#else
      /* verify calling parameter validity for h/w message scheduling */
      tempStatus = a429_utl_validate_scheduled_message_parameters ((CEI_UINT32)board, msgPtr->channel, &localChannel);

      if (tempStatus == ARS_NORMAL) {
         /* if multi-thread protection is active, acquire a lock on the host interface */
         if (device[board].multiThreadProtect != AR_CONC_NONE) {
            status = (CDEV_API_RET_TYPE)cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (status == ARS_NORMAL) {
#endif
            msgEntry.messageIndex = 0;
            msgEntry.board        = board;
            msgEntry.channel      = localChannel;
            msgEntry.rate         = msgPtr->rate;
            msgEntry.start        = msgPtr->start;
            msgEntry.txCount      = msgPtr->txCount;
            msgEntry.data         = msgPtr->data;
            
            /* define a single scheduled message entry */
            status = (CDEV_API_RET_TYPE)a429_utl_store_message_scheduler_entry (FALSE, 1, &msgEntry);
            
            /* assign the message index in the app-supplied structure array entry */
            msgPtr->messageIndex = msgEntry.messageIndex;
            msgPtr++;
   
#ifndef TARGET_EMBEDDED
            /* if multi-thread protection is active, free any acquired lock */
            if (device[board].multiThreadProtect != AR_CONC_NONE) {
               if (status == ARS_NORMAL) {
                  status = (CDEV_API_RET_TYPE)cei_release_thread_lock(LOCK_IDENT);
               }
               else {
                  tempStatus = cei_release_thread_lock(LOCK_IDENT);
               }
            }
         }
      }
      else {
         status = (CDEV_API_RET_TYPE)tempStatus;
      }
#endif
   }

   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                               A R _ M O D I F Y _ M S G
 *===================================================================================================================*
 *
 * FUNCTION:    Modify a message already defined for scheduled transmission.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board        -- (input) Board number (0 to 127)  
 *              CDEV_CHAN_TYPE channel       -- (input) ARINC 429 transmit channel selection.  Valid range is 0 to N, 
 *                                                      where N-1 refers to the last ARINC 429 transmit channel.
 *              CDEV_PARM_SSI_TYPE msgNumber -- (input) The message scheduler table index of the message to modify
 *              CDEV_PARM_SSI_TYPE rate      -- (input) The message transmission rate (in milliseconds)
 *              CDEV_PARM_SI_TYPE data       -- (input) The ARINC message to transmit
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL      Success
 *                                ARS_INVBOARD    Invalid or uninitialized board
 *                                ARS_INVARG      Invalid channel
 *                                ARS_INVHARVAL   Message scheduling is not supported on the specified channel
 *                                ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                                ARS_FAILURE     The supplied msgNumber index value exceeds the available number
 *                                                of table entries
 *
 * GLOBALS:     device[board].multiThreadProtect
 *
 * DESCRIPTION: This routine modifies an existing ARINC 429 message already defined for periodic retransmission.  
 *              The message index parameter msgNumber should correspond to the respective message when it was created 
 *              (and the index returned from) in the invocation of ar_define_msg.
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_modify_msg 
(
   CDEV_BOARD_TYPE board,                                                     /* (input) Board number (0 to 127)     */
   CDEV_CHAN_TYPE  channel,                                                   /* (input) 429 Tx channel selection    */
   CDEV_PARM_SSI_TYPE msgNumber,                                              /* (input) index from ar_define_msg    */
   CDEV_PARM_SSI_TYPE rate,                                                   /* (input) transmission rate (in msec) */
   CDEV_PARM_SI_TYPE data                                                     /* (input) the ARINC msg to transmit   */
)
{
   CEI_UINT32 localChannel = 0U;                                                        /* transmit channel select   */
   CEI_INT32 status = ARS_NORMAL;                                                       /* function return status    */
#ifndef TARGET_EMBEDDED
   CEI_INT32 tempStatus = ARS_NORMAL;                                                   /* local temp return status  */
#endif
   AR_SCHEDULED_MSG_ENTRY_TYPE msgEntry;                                                /* message content structure */

#ifdef TARGET_EMBEDDED
   localChannel = (CEI_UINT32)channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
#else
   /* verify calling parameter validity for h/w message scheduling */
   status = a429_utl_validate_scheduled_message_parameters ((CEI_UINT32)board, (CEI_UINT32)channel, &localChannel);

   if (status == ARS_NORMAL) {
      if (msgNumber >= (CDEV_PARM_SSI_TYPE)deviceGlobal[board]->messageSchedulerTableSize) {
         status = ARS_INVARG;
      }
      else {
         /* if multi-thread protection is active, acquire a lock on the host interface */
         if (device[board].multiThreadProtect != AR_CONC_NONE) {
            status = cei_acquire_thread_lock(LOCK_IDENT);
         }
   
         if (status == ARS_NORMAL) {
#endif
            msgEntry.messageIndex = (CEI_UINT32)msgNumber;
            msgEntry.board        = (CEI_UINT32)board;
            msgEntry.channel      = localChannel;
            msgEntry.rate         = (CEI_UINT32)rate;
            msgEntry.start        = (CEI_UINT32)0;         /* start offset and transmit count mods */
            msgEntry.txCount      = (CEI_UINT32)0;         /* are not supported by this routine    */
            msgEntry.data         = (CEI_UINT32)data;
            
            /* modify a single scheduled message entry */
            status = a429_utl_store_message_scheduler_entry (TRUE, 1, &msgEntry);
         
#ifndef TARGET_EMBEDDED
            /* if multi-thread protection is active, free any acquired lock */
            if (device[board].multiThreadProtect != AR_CONC_NONE) {
               if (status == ARS_NORMAL) {
                  status = cei_release_thread_lock(LOCK_IDENT);
               }
               else {
                  tempStatus = cei_release_thread_lock(LOCK_IDENT);
               }
            }
         }
      }
   }
   if (tempStatus) ;
#endif   
   return (CDEV_API_RET_TYPE)status;
}


/*===================================================================================================================*
 * ENTRY POINT:                        A R _ M O D I F Y _ M S G _ B L O C K
 *===================================================================================================================*
 *
 * FUNCTION:    Modify multiple messages previously defined for scheduled transmission.
 *
 * PARAMETERS:  CEI_INT32 numberOfEntries -- (input) Defines the number of scheduler entries to modify on the device,
 *                                                   used as a reference for the number of structure array elements 
 *                                                   defined via the messageEntry pointer.
 *              pAR_SCHEDULED_MSG_ENTRY_TYPE
 *                        messageEntry    -- (input) array of structures of message properties, each element defined 
 *                                                   as follows:
 *                        struct 
 *                        {
 *                           CEI_UINT32 messageIndex;  Msg scheduler index previously assigned via ar_define_msg_blk
 *                           CEI_UINT32 board;         Board number (0 to 127)                
 *                           CEI_UINT32 channel;       ARINC 429 transmit channel selection                
 *                           CEI_UINT32 rate;          Message transmission rate (in milliseconds)                
 *                           CEI_UINT32 start;         Not applicable (start offset cannot be modified)
 *                           CEI_UINT32 txCount;       Message transmission count (0xFFFFFFFF for infinite count)
 *                           CEI_UINT32 data;          ARINC 429 message
 *                        };
 *                                                    
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL      Success
 *                                ARS_INVBOARD    Invalid or uninitialized board
 *                                ARS_INVHARVAL   Invalid transmit channel entry
 *                                ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                                ARS_INVARG      The supplied msg table index exceeds the available number
 *                                                of table entries
 *
 * GLOBALS:     device[board].multiThreadProtect
 *
 * DESCRIPTION: This routine modifies existing ARINC 429 messages already defined for periodic retransmission.
 *              The message index structure member messageIndex should correspond to the respective message index 
 *              for this structure member assigned in the structure during the invocation of ar_define_msg_blk.
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_modify_msg_block 
(
   CEI_INT32 numberOfEntries,                                                  /* (input) number of entries          */   
   pAR_SCHEDULED_MSG_ENTRY_TYPE messageEntry                                   /* (input) ptr to struct of messages  */   
)
{
   CEI_UINT32 board;                                                            /* board id extracted from structure */
   CEI_INT32 arrayIndex;                                                             /* scheduler access array index */
   CEI_UINT32 localChannel = 0U;                                                        /* transmit channel select   */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                              /* function return status     */
#ifndef TARGET_EMBEDDED
   CEI_INT32 tempStatus = ARS_NORMAL;                                                  /* local temp return status   */
#endif
   AR_SCHEDULED_MSG_ENTRY_TYPE msgEntry;                                               /* message content structure  */
   pAR_SCHEDULED_MSG_ENTRY_TYPE msgPtr = messageEntry;                                         /* local ptr to array */

   for (arrayIndex = 0; (arrayIndex < numberOfEntries) && (status == ARS_NORMAL); arrayIndex++) {
      board = msgPtr->board;

#ifdef TARGET_EMBEDDED
      localChannel = (CEI_UINT32)msgPtr->channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
#else
      /* verify calling parameter validity for h/w message scheduling */
      tempStatus = a429_utl_validate_scheduled_message_parameters (board, msgPtr->channel, &localChannel);

      if (tempStatus == ARS_NORMAL) {
         if (msgPtr->messageIndex >= deviceGlobal[board]->messageSchedulerTableSize) {
            status = ARS_INVARG;
         }

         else {
            /* if multi-thread protection is active, acquire a lock on the host interface */
            if (device[board].multiThreadProtect != AR_CONC_NONE) {
               status = (CDEV_API_RET_TYPE)cei_acquire_thread_lock(LOCK_IDENT);
            }
            if (status == ARS_NORMAL) {
#endif
               msgEntry.messageIndex = msgPtr->messageIndex;
               msgEntry.board        = board;
               msgEntry.channel      = localChannel;
               msgEntry.rate         = msgPtr->rate;
               msgEntry.start        = 0;  /* start offset cannot be modified from original def */ 
               msgEntry.txCount      = msgPtr->txCount;
               msgEntry.data         = msgPtr->data;
               
               /* modify a single scheduled message entry */
               status = (CDEV_API_RET_TYPE)a429_utl_store_message_scheduler_entry (TRUE, 1, &msgEntry);
            
              /* move to the next sructure array entry */
               msgPtr++;
   
#ifndef TARGET_EMBEDDED
               /* if multi-thread protection is active, free any acquired lock */
               if (device[board].multiThreadProtect != AR_CONC_NONE) {
                  if (status == ARS_NORMAL) {
                     status = (CDEV_API_RET_TYPE)cei_release_thread_lock(LOCK_IDENT);
                  }
                  else {
                     tempStatus = cei_release_thread_lock(LOCK_IDENT);
                  }
               }
            }
         }
      }
      else {
         status = (CDEV_API_RET_TYPE)tempStatus;
      }
#endif
   }
   
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                        A R _ M O D I F Y _ M S G _ B L O C K _ L V
 *===================================================================================================================*
 *
 * FUNCTION:    Modify a block of messages already defined for scheduled transmission, specifically included in 
 *              support of LabVIEW structure array formatting
 *
 * PARAMETERS:  CEI_INT32 numberOfEntries -- (input) Defines the number of scheduler entries to modify on the device,
 *                                                   used as a reference for the number of structure array elements 
 *                                                   defined via the messageEntry pointer.
 *
 *              pLV_AR_SCHEDULED_MSG_ENTRY_TYPE messageEntry
 *                                                    
 *                  CEI_UINT32 array_size -- (input) Size of the messageEntries array (in bytes), and should always
 *                                                   be assigned the value 57344 (2048 32-bit x 7 entries).  This 
 *                                                   parameter is unused outside of the respective LabVIEW wrapper.
 *                                                    
 *                  AR_SCHEDULED_MSG_ENTRY_TYPE msgEntries[2048]  
 *                                        -- (input) A preallocated array of structures of message properties,  
 *                                                   each element defined as follows:
 *                        struct 
 *                        {
 *                           CEI_UINT32 messageIndex;  Msg scheduler index previously assigned via ar_define_msg_blk
 *                           CEI_UINT32 board;         Board number (0 to 127)                
 *                           CEI_UINT32 channel;       ARINC 429 transmit channel selection                
 *                           CEI_UINT32 rate;          Message transmission rate (in milliseconds)                
 *                           CEI_UINT32 start;         Not applicable (start offset cannot be modified)
 *                           CEI_UINT32 txCount;       Message transmission count (0xFFFFFFFF for infinite count)
 *                           CEI_UINT32 data;          ARINC 429 message
 *                        };
 *                                                    
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL      Success
 *                                ARS_INVBOARD    Invalid or uninitialized board
 *                                ARS_INVHARVAL   Invalid transmit channel entry
 *                                ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                                ARS_INVARG      The supplied msg table index exceeds the available number
 *                                                of table entries
 *
 * GLOBALS:     device[board].multiThreadProtect
 *
 * DESCRIPTION: This routine modifies existing ARINC 429 messages already defined for periodic retransmission.
 *              The message index structure member messageIndex should correspond to the respective message index 
 *              for this structure member assigned in the structure during the invocation of ar_define_msg_blk_lv.
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_modify_msg_block_lv 
(
   CEI_INT32 numberOfEntries,                                                  /* (input) number of entries          */
   pLV_AR_SCHEDULED_MSG_ENTRY_TYPE messageEntry                                /* (input) ptr to struct of messages  */
)
{
   CEI_UINT32 board;                                                            /* board id extracted from structure */
   CEI_INT32 arrayIndex;                                                             /* scheduler access array index */
   CEI_UINT32 localChannel = 0U;                                                        /* transmit channel select   */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                              /* function return status     */
#ifndef TARGET_EMBEDDED
   CEI_INT32 tempStatus = ARS_NORMAL;                                                  /* local temp return status   */
#endif
   AR_SCHEDULED_MSG_ENTRY_TYPE msgEntry;                                               /* message content structure  */
   pAR_SCHEDULED_MSG_ENTRY_TYPE msgPtr = messageEntry->msgEntries; /* local ptr to array */

   for (arrayIndex = 0; (arrayIndex < numberOfEntries) && (status == ARS_NORMAL); arrayIndex++) {
      board = msgPtr->board;

#ifdef TARGET_EMBEDDED
      localChannel = (CEI_UINT32)msgPtr->channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
#else
      /* verify calling parameter validity for h/w message scheduling */
      tempStatus = a429_utl_validate_scheduled_message_parameters (board, msgPtr->channel, &localChannel);

      if (tempStatus == ARS_NORMAL) {
         if (msgPtr->messageIndex >= deviceGlobal[board]->messageSchedulerTableSize) {
            status = ARS_INVARG;
         }

         else {
            /* if multi-thread protection is active, acquire a lock on the host interface */
            if (device[board].multiThreadProtect != AR_CONC_NONE) {
               status = (CDEV_API_RET_TYPE)cei_acquire_thread_lock(LOCK_IDENT);
            }
            if (status == ARS_NORMAL) {
#endif
               msgEntry.messageIndex = msgPtr->messageIndex;
               msgEntry.board        = board;
               msgEntry.channel      = localChannel;
               msgEntry.rate         = msgPtr->rate;
               msgEntry.start        = 0;  /* start offset cannot be modified from original def */ 
               msgEntry.txCount      = msgPtr->txCount;
               msgEntry.data         = msgPtr->data;
               
               /* modify a single scheduled message entry */
               status = (CDEV_API_RET_TYPE)a429_utl_store_message_scheduler_entry (TRUE, 1, &msgEntry);
            
              /* move to the next sructure array entry */
               msgPtr++;
   
#ifndef TARGET_EMBEDDED
               /* if multi-thread protection is active, free any acquired lock */
               if (device[board].multiThreadProtect != AR_CONC_NONE) {
                  if (status == ARS_NORMAL) {
                     status = (CDEV_API_RET_TYPE)cei_release_thread_lock(LOCK_IDENT);
                  }
                  else {
                     tempStatus = cei_release_thread_lock(LOCK_IDENT);
                  }
               }
            }
         }
      }
      else {
         status = (CDEV_API_RET_TYPE)tempStatus;
      }
#endif
   }
   
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                   A R _ R E A D _ S C H E D U L E D _ M S G _ B L O C K
 *===================================================================================================================*
 *
 * FUNCTION:    Read a block of defined messages from the message scheduler table.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board                       -- (input) Board number (0 to 127)  
 *                                                    
 *              CEI_INT32 startingEntry                     -- (input) the first entry from which to begin reading
 *                                                    
 *              CEI_INT32 numberOfEntries                   -- (input) number of entries to read
 *                                                    
 *              pAR_SCHEDULED_MSG_ENTRY_TYPE  messageEntry  -- (input) array of structures of message content read, 
 *                                                                     each element defined as follows:
 *                        struct 
 *                        {
 *                           CEI_UINT32 messageIndex;  Msg scheduler index previously assigned via ar_define_msg_blk
 *                           CEI_UINT32 board;         Board number (0 to 127)                
 *                           CEI_UINT32 channel;       ARINC 429 transmit channel selection                
 *                           CEI_UINT32 rate;          Message transmission rate (in milliseconds)                
 *                           CEI_UINT32 start;         Start offset prior to board enable (once the scheduler is 
 *                                                     enabled, this value indicates the current rate countdown value)
 *                           CEI_UINT32 txCount;       Message transmission count (0xFFFFFFFF for infinite count)
 *                           CEI_UINT32 data;          ARINC 429 message
 *                        };
 *                                                    
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL      Success
 *                                ARS_INVBOARD    Invalid or uninitialized board
 *                                ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                                ARS_INVARG      The starting entry plus the number of entries requested to read 
 *                                                exceeds the size of the message scheduler table
 * 
 * GLOBALS:     device[]
 * 
 * DESCRIPTION: This routine returns the contents of the specified block of entries from the message scheduler table.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_read_scheduled_msg_block
(
   CDEV_BOARD_TYPE board,                                                      /* (input) board number (0-127)       */
   CEI_INT32 startingEntry,		                                              /* (input) index of the first entry   */
   CEI_INT32 numberOfEntries,                                                  /* (input) number of entries to read  */
   pAR_SCHEDULED_MSG_ENTRY_TYPE messageEntry                                   /* (input) ptr to struct of messages  */
)
{
   CEI_INT32 msgTableCounter;                                                 /* counter for message table read      */
   CEI_UINT32 appProgTimerResolution;                                         /* Application specified time-tag res  */
   CEI_UINT32 memOffset;                                                      /* local offset to channel memory area */
   CEI_UINT32 attributeRead;                                                  /* local value read from the board     */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                              /* function return status     */
#ifndef TARGET_EMBEDDED
   CEI_INT32 tempStatus = ARS_NORMAL;                                                  /* local temp return status   */
#endif
   CEI_INT32 accessStatus = ARS_NORMAL;                                                /* local temp return status   */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr = NULL;                   /* pointer to board local structure    */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr = NULL;                          /* pointer to board global structure   */
   pAR_SCHEDULED_MSG_ENTRY_TYPE messageEntryPtr = messageEntry;                  /* local ptr to msg structure array */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   else if ((numberOfEntries == 0) || ((numberOfEntries+startingEntry) >= MAX_SCHEDULED_MSGS)) {
      status = ARS_INVARG;
   }
   else {
#endif
      deviceLocalStructPtr = &device[board];
      deviceGlobalStructPtr = deviceGlobal[board];
      
#ifndef TARGET_EMBEDDED
      /* if multi-thread protection is active, acquire a lock on the host interface */
      if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
         status = (CDEV_API_RET_TYPE)cei_acquire_thread_lock(LOCK_IDENT);
      }
            
      if (status == ARS_NORMAL) {
#endif
      	memOffset = CEIDEV_MSG_SCHED_TABLE_START + (startingEntry * CEIDEV_MSG_SCHED_TBL_ENTRY_SIZE);
      	
      	for (msgTableCounter = 0; msgTableCounter < numberOfEntries; msgTableCounter++) {
      	
      	   /* read this entry's rate value, scale the rate if using x20 compatible resolution timing mode */
      	   accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &attributeRead);
            messageEntryPtr->rate = attributeRead;
      	   memOffset++;

           /* read this entry's channel value */
      	   accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &attributeRead);
            messageEntryPtr->channel = attributeRead - deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
      	   memOffset++;
      	
      	   /* read the entry transmit count */
      	   accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &attributeRead);
            messageEntryPtr->txCount = attributeRead;
      	   memOffset++;
      	
      	   /* read the entry start offset */
      	   accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &attributeRead);
            messageEntryPtr->start = attributeRead;
      	   memOffset++;
      	         
      	   /* scale the rate and start offset if using x20 compatible resolution timing mode */
      	   if (deviceGlobalStructPtr->deviceTimeTagReference == AR_TIMER_X20_COMPAT_32BIT) {
      	      appProgTimerResolution = deviceGlobalStructPtr->deviceX20CompatibleResolution;
      	      if (appProgTimerResolution >= AR_X20_QTR_MICROSEC_PER_MSEC) {
      	         messageEntryPtr->rate  *= appProgTimerResolution/AR_X20_QTR_MICROSEC_PER_MSEC;
      	         messageEntryPtr->start *= appProgTimerResolution/AR_X20_QTR_MICROSEC_PER_MSEC;
      	      }
      	   }
      	
      	   /* read the entry data */
      	   memOffset += 3;
      	   accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &attributeRead);
      	   memOffset++;
            messageEntryPtr->data = attributeRead;
      	         
      	   messageEntryPtr->messageIndex = msgTableCounter;
      	   messageEntryPtr++;
      	}
      	
      	if (accessStatus != ARS_NORMAL) {
      	   status = (CDEV_API_RET_TYPE)accessStatus;
         }

#ifndef TARGET_EMBEDDED
         /* if multi-thread protection is active, free any acquired lock */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            if (status == ARS_NORMAL) {
               status = (CDEV_API_RET_TYPE)cei_release_thread_lock(LOCK_IDENT);
            }
            else {
               tempStatus = cei_release_thread_lock(LOCK_IDENT);
            }
         }
      }
      else
         status = (CDEV_API_RET_TYPE)tempStatus;
   }
#endif
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:             A R _ R E A D _ M E S S A G E _ S C H E D U L E _ T A B L E _ E N T R Y
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve a single entry from the message scheduler table
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board          -- (input)  board to reference (0 to 127)
 *              CEI_UINT32  messageIndex,      -- (input)  index of the entry to read
 *              pCEI_UINT32 msgSchedTableEntry -- (output) array of 8 unsigned integer values reflecting the entry
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL      Success
 *                                ARS_INVBOARD    Invalid or uninitialized board
 *
 * GLOBALS:     device[]
 * 
 * DESCRIPTION: This routine returns the contents of the specified entry from the message scheduler table, in a raw
 *              format matching the message scheduler table entry format in the CEI-x30 Hardware User Manual.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_read_message_schedule_table_entry 
(
   CDEV_BOARD_TYPE board,                                                      /* (input) board number (0-127)       */
   CEI_UINT32  messageIndex,                                                   /* (input) index of the entry to read */
   pCEI_UINT32 msgSchedTableEntry                                              /* (input) ptr to array to store msg  */
)
{
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                      /* function return status             */
   CEI_INT32 accessStatus = ARS_NORMAL;                                        /* status of device read/write access */
   CEI_UINT32 msgIndex;                                                         /* temp index for table search       */
   CEI_UINT32 memOffset;                                                   /* local offset to scheduler table memory */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                                 /* ptr to board local structure */
   pCEI_UINT32 msgPtr = msgSchedTableEntry;                                         /* ptr to application data array */
   
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      deviceLocalStructPtr = &device[board];
      memOffset = CEIDEV_MSG_SCHED_TABLE_START + (messageIndex * CEIDEV_MSG_SCHED_TBL_ENTRY_SIZE);
      for (msgIndex = 0; msgIndex < CEIDEV_MSG_SCHED_TBL_ENTRY_SIZE; msgIndex++) {
         accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset++], msgPtr++);
      }
#ifndef TARGET_EMBEDDED
   }
   else
      status = ARS_INVBOARD;
#endif

   if (accessStatus) ;
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                          A R _ U P D A T E _ M E S S A G E _ B L O C K
 *===================================================================================================================*
 *
 * FUNCTION:    Modify the message value only for a block of message scheduler table entries
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board      -- (input) board number (0 to 127)  
 *
 *              CEI_UINT32 numberOfMsgs    -- (input) the number of message scheduler entries to modify
 *
 *              CEI_UINT32 msgIndex        -- (input) the message index of the first message to modify
 *
 *              pCEI_UINT32 msgValues      -- (input) an array of values for successive messages to be assigned   
 *
 * RETURN VAL:  CEI_INT32  -- ARS_NORMAL      Success
 *                            ARS_INVBOARD    Invalid or uninitialized board
 *                            ARS_INVARG      Invalid message scheduler table index reference
 *
 * GLOBALS:     device[]
 * 
 * DESCRIPTION: This routine modifies only the message value of a successive block of message scheduler table 
 *              entries previously defined via ar_define_msg or ar_define_msg_block.
 * 
 *===================================================================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_update_message_block 
(
   CDEV_BOARD_TYPE board,                                           /* (input) board number to reference (0 to 127)  */
   CEI_UINT32  numberOfMsgs,                                        /* (input) number of messages to modify          */
   CEI_UINT32  msgIndex,                                            /* (input) index of the first message to modify  */
   pCEI_UINT32 msgValues                                            /* (input) pointer to an array of message values */
)
{
   CEI_UINT32 msgTableCounter;                                                /* message scheduler table entry index */
   CEI_UINT32 memOffset;                                                /* local offset into the Msg Scheduler Table */
   CEI_INT32  status = ARS_NORMAL;                                                      /* function return status    */
#ifndef TARGET_EMBEDDED
   CEI_INT32  tempStatus = ARS_NORMAL;                                                  /* local temp return status  */
#endif
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                             /* pointer to board local structure */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   else if ( ((numberOfMsgs + msgIndex) >= deviceGlobal[board]->messageSchedulerTableSize) || (msgValues == NULL)) {
      status = ARS_INVARG;
   }
   else {
#endif
      deviceLocalStructPtr = &device[board];

#ifndef TARGET_EMBEDDED
      /* if multi-thread protection is active, acquire a lock on the host interface */
      if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
         status = cei_acquire_thread_lock(LOCK_IDENT);
      }
   
      if (status == ARS_NORMAL) {
#endif
         memOffset = CEIDEV_MSG_SCHED_TABLE_START + (msgIndex * CEIDEV_MSG_SCHED_TBL_ENTRY_SIZE) + CEIDEV_MSG_SCHED_TBL_DATA_OFFSET;
         
         for (msgTableCounter = 0; (msgTableCounter < numberOfMsgs) && (status == ARS_NORMAL); msgTableCounter++) {
            status = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], msgValues[msgTableCounter]);
            memOffset += CEIDEV_MSG_SCHED_TBL_ENTRY_SIZE;
         }   
         
#ifndef TARGET_EMBEDDED
         /* if multi-thread protection is active, free any acquired lock */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            if (status == ARS_NORMAL) {
               status = cei_release_thread_lock(LOCK_IDENT);
            }
            else {
               tempStatus = cei_release_thread_lock(LOCK_IDENT);
            }
         }
      }
   }
#endif
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                A R _ A S S I G N _ S C H E D U L E R _ S T A R T _ O F F S E T S
 *===================================================================================================================*
 *
 * FUNCTION:    This routine determines the applicable start offset values for each transmit channel message 
 *              scenario defined in the message scheduler, computed to avoid rate skew on the respective channel.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board -- (input) board to reference (0 to 127)
 *                                                    
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL      Success
 *                                ARS_INVBOARD    Invalid or uninitialized board
 *                                ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                                ARS_FAILURE     Message processing is enabled - can't touch the scheduler table
 * 
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: This routine reads all defined messages (rate > 0) from the message scheduler table, determines
 *              the appropriate start offset values for each transmit channel message scenario as a best estimate
 *              to avoid rate skew on the respective channel, then updates the start offset values in the table.
 *              This feature requires all message scheduler table entries for the board be defined prior to 
 *              invocation, as well as the board global enable to be set to 'disabled'.  The start offset values
 *              defined by this routine do not account for bus timing issues and rate skew due to bursts of 
 *              aperiodic messages invoked by the application.
 * 
 *===================================================================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_assign_scheduler_start_offsets (CDEV_BOARD_TYPE board) 
{
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (INTEGRITY_PCI_PPC) && !defined (INTEGRITY_PCI_X86)
   CHAN_CFG_ENTRY_TYPE txChanAttributes[CEIDEV_MAX_TX_CHAN];           /* transmit channel and bus speed assignments */
   CHAN_CFG_ENTRY_TYPE tempChanEntry;                                  /* temporary transmit channel and bus speed   */
   SCH_MSG_ENTRY_TYPE tempMsgEntry;                                           /* temporary local msg scheduler entry */
   pSCH_MSG_ENTRY_TYPE localMsgBlock = NULL;                               /* local copy of message scheduler table  */
   pSCH_MSG_ENTRY_TYPE channelMsgBlock = NULL;                             /* local copy of message scheduler table  */
   pCEI_UINT32 msgRateGroupRate = NULL;                                    /* array of rate groups for a single chnl */
   pCEI_UINT32 msgRateGroupOffsetStartOffset = NULL;                         /* cumm. start offset for each rate grp */
   pCEI_UINT32 msgRateGroupFirstStartOffset = NULL;                          /* forst start offset for each rate grp */
   pCEI_UINT32 msgRateGroupLastStartOffset = NULL;                        /* last start offset used for each rate gp */
   pCEI_UINT32 msgRateGroupCount = NULL;                            /* count of messages assigned to each rate group */
   CEI_UINT32 lastRateValue = 0;                                      /* rate of previous channel message rate group */
   CEI_UINT32 numberOfRateGoups = 0;                                    /* number of rate groups for current channel */
   CEI_UINT32 i, j, k, n;                                                                    /* temp looping indexes */
   CEI_UINT32 startOffset;                                              /* temp start offset value being computed    */
   CEI_UINT32 startOffsetAdjusted;               /* flag indicating this message introduces a new start offset value */
   CEI_UINT32 startOffsetIncrementValue;  /* the number of millisec to add to the previous start offset for the next */
   CEI_UINT32 msgCountThisMillisec;       /* the number of 100Kbps messages assigned to the current start offset     */
   CEI_UINT32 msgCountAllowed;            /* the number of 100Kbps messages allowed to use the current start offset  */
   CEI_UINT32 currentRateGroup;           /* the rate in milliseconds of the current rate group being processed      */
   CEI_UINT32 currentRateGroupCount;      /* the number of messages in the current rate group being processed        */
   CEI_UINT32 halfMsgRateGroupCount;      /* the number of messages to split into the next instance of this msg rate */
   CEI_UINT32 startOffsetBeforeHalfMsgRateGroup;/* the last start offset used in the first instance of this msg rate */
   CEI_UINT32 intoHalfMsgRateGroupCount;     /* running count of messages after split into next instance of msg rate */
   CEI_UINT32 increment100sStartOffset;      /* the number of 100ms multiples to increment with the current rate grp */
   CEI_UINT32 increment100sStartOffsetOffset = 100; /* the current 100ms multiple in use with the current rate group */
   CEI_UINT32 incrementStartOffset;      /* flag indicating the start offset was incremented during the current pass */
   CEI_UINT32 rateGroupIndex;                                      /* running index of the current rate group in use */
   CEI_UINT32 numberOfEntries = 0;              /* total number of defined messages overall and per transmit channel */
   CEI_UINT32 totalNumberOfEntries = 0;                                  /* total number of defined messages overall */
   CEI_UINT32 channelEntryTotal = 0;   /* total number of transmit channels having messages defined in the scheduler */
   CEI_UINT32 rate;                                                                        /* temp rate variable     */
   CEI_INT32 status = ARS_NORMAL;                                                  /* function return status */
   CDEV_CHAN_TYPE channelIndex;                                                            /* temp channel index     */
   CDEV_PARM_SSI_TYPE busSpeed;                                                            /* temp tx chan bus speed */
   CEI_INT32 accessStatus = ARS_NORMAL;                                        /* status of device read/write access */
   CEI_UINT32 msgIndex;                                                              /* temp index for table search  */
   CEI_UINT32 memOffset;                                                   /* local offset to scheduler table memory */
   CEI_UINT32 chanOffset;                                                  /* local offset to scheduler table memory */
   CEI_UINT32 globalEnableReg;                                             /* current value of the global enable reg */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                                 /* ptr to board local structure */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                                       /* ptr to board global structure */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      /* check "global enable" and assure this board is not actively processing messages before overwriting the start offset values */
      accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], &globalEnableReg);
      if (accessStatus == ARS_NORMAL) {
         if (globalEnableReg & CEIDEV_GLOBAL_ENABLE_ON) {
            status = ARS_FAILURE;
         }
      }
      else {
         status = ARS_FAILURE;
      }
#ifndef TARGET_EMBEDDED
   }
   else {
      status = ARS_FAILURE;
   }
#endif
   if (status == ARS_NORMAL) {
      /* allocate and initialize memory for the local working message scheduler table entry and rate sorting data structures */
      deviceGlobalStructPtr = deviceGlobal[board];
      localMsgBlock =   calloc (deviceGlobalStructPtr->messageSchedulerTableSize, sizeof(SCH_MSG_ENTRY_TYPE));
      channelMsgBlock = calloc (deviceGlobalStructPtr->messageSchedulerTableSize, sizeof(SCH_MSG_ENTRY_TYPE));
      msgRateGroupOffsetStartOffset = calloc (deviceGlobalStructPtr->messageSchedulerTableSize, sizeof(CEI_UINT32));
      msgRateGroupFirstStartOffset =  calloc (deviceGlobalStructPtr->messageSchedulerTableSize, sizeof(CEI_UINT32));
      msgRateGroupLastStartOffset =   calloc (deviceGlobalStructPtr->messageSchedulerTableSize, sizeof(CEI_UINT32));
      msgRateGroupCount =             calloc (deviceGlobalStructPtr->messageSchedulerTableSize, sizeof(CEI_UINT32));
      msgRateGroupRate =              calloc (deviceGlobalStructPtr->messageSchedulerTableSize, sizeof(CEI_UINT32));

      if ((localMsgBlock != NULL) && (channelMsgBlock != NULL) && (msgRateGroupRate != NULL) && (msgRateGroupOffsetStartOffset != NULL) && 
          (msgRateGroupFirstStartOffset != NULL) && (msgRateGroupLastStartOffset != NULL) && (msgRateGroupCount != NULL)) {
         /* initialize the local channel attribute array */
         for (i = 0; i < CEIDEV_MAX_TX_CHAN; i++) {
            txChanAttributes[i].channel = CEIDEV_MAX_TX_CHAN;
            txChanAttributes[i].busSpeed        = 0;
            txChanAttributes[i].numberOfEntries = 0;
         }
   
         /* First, retrieve all 'defined message entries' from the message scheduler table, bypassing entries having a rate value of 0 */
         deviceLocalStructPtr = &device[board];
         msgIndex = 0;
         memOffset = CEIDEV_MSG_SCHED_TABLE_START;
         chanOffset = CEIDEV_MSG_SCHED_TABLE_START + 1;
         for (i = 0; ((i < deviceGlobalStructPtr->messageSchedulerTableSize) && (accessStatus == ARS_NORMAL)); i++) {
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &rate);
            if (rate > 0U) {
               localMsgBlock[msgIndex].messageIndex = i;   /* copy this messages's channel, rate, and location */
               localMsgBlock[msgIndex].rate = rate;
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[chanOffset], &localMsgBlock[msgIndex].channel);
               msgIndex++;
            }
            memOffset += CEIDEV_MSG_SCHED_TBL_ENTRY_SIZE;
            chanOffset += CEIDEV_MSG_SCHED_TBL_ENTRY_SIZE;
         }
         
         if (accessStatus == ARS_NORMAL) {
            numberOfEntries = msgIndex;
         
            /* OK, we have copied the entire scheduled message set, search the local scheduler table entries and log all of the transmit channels used */
            channelEntryTotal = 1;
            txChanAttributes[0].channel = localMsgBlock[0].channel;
            txChanAttributes[0].numberOfEntries = 1;
         
            for (i = 1; i < numberOfEntries; i++) {
               for (j = 0; j < channelEntryTotal; j++) {
                  if (localMsgBlock[i].channel == txChanAttributes[j].channel) {
                     txChanAttributes[j].numberOfEntries++;
                     break;
                  }
               }
               if (j == channelEntryTotal) {
                  txChanAttributes[j].channel = localMsgBlock[i].channel;
                  txChanAttributes[j].numberOfEntries = 1;
                  channelEntryTotal++;
               }
            }
         
            /* Next log the bus speed for all of the transmit channels referenced in the message scheduler in the channel attribute array */
            for (channelIndex = 0; channelIndex < (CDEV_CHAN_TYPE)channelEntryTotal; channelIndex++) {
                status = ar_get_device_config (board, (CDEV_CHAN_TYPE)(txChanAttributes[channelIndex].channel - deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX]), ARU_TX_BITRATE, &busSpeed);
                txChanAttributes[channelIndex].busSpeed = (busSpeed == ARU_SPEED_LOW) ? ARU_SPEED_LOW : ARU_SPEED_HIGH;
            }
         
            /* Now sort the transmit channel attribute array based on the assigned transmit channel */
            for (j = 0; j < channelEntryTotal; j++) {
               for (i = j; i < channelEntryTotal; i++) {
                  if (txChanAttributes[i].channel < txChanAttributes[j].channel) {
                     memcpy (&tempChanEntry, &txChanAttributes[j], sizeof(CHAN_CFG_ENTRY_TYPE));
                     memcpy (&txChanAttributes[j], &txChanAttributes[i], sizeof(CHAN_CFG_ENTRY_TYPE));
                     memcpy (&txChanAttributes[i], &tempChanEntry, sizeof(CHAN_CFG_ENTRY_TYPE));
                  }
               }
            }
         
            /* Now sort the message block array based on the assigned transmit channel */
            for (j = 0; j < numberOfEntries; j++) {
               for (i = j; i < numberOfEntries; i++) {
                  if (localMsgBlock[i].channel < localMsgBlock[j].channel) {
                     memcpy (&tempMsgEntry, &localMsgBlock[j], sizeof(SCH_MSG_ENTRY_TYPE));
                     memcpy (&localMsgBlock[j], &localMsgBlock[i], sizeof(SCH_MSG_ENTRY_TYPE));
                     memcpy (&localMsgBlock[i], &tempMsgEntry, sizeof(SCH_MSG_ENTRY_TYPE));
                  }
               }
            }
         
            /* Now sort the individual channel message block array entries based on their rate value */
            j = 0;
            for (k = 0; k < channelEntryTotal; k++) {
               totalNumberOfEntries = j + txChanAttributes[k].numberOfEntries;
               while (j < totalNumberOfEntries) {
                  for (i = j; i < totalNumberOfEntries; i++) {
                     if (localMsgBlock[i].rate < localMsgBlock[j].rate) {
                        memcpy (&tempMsgEntry, &localMsgBlock[j], sizeof(SCH_MSG_ENTRY_TYPE));
                        memcpy (&localMsgBlock[j], &localMsgBlock[i], sizeof(SCH_MSG_ENTRY_TYPE));
                        memcpy (&localMsgBlock[i], &tempMsgEntry, sizeof(SCH_MSG_ENTRY_TYPE));
                     }
                  }
                  j++;
               }
            }
         
            /* Now loop on each channel's message group and determine the best start offset values to avoid rate skew on that channel */
            n = 0;
            totalNumberOfEntries = numberOfEntries;
            for (k = 0; k < channelEntryTotal; k++) {
               memcpy (channelMsgBlock, &localMsgBlock[n], (sizeof(SCH_MSG_ENTRY_TYPE) * txChanAttributes[k].numberOfEntries));
               numberOfEntries = txChanAttributes[k].numberOfEntries;
               numberOfRateGoups = 0;
         
               /* First determine the number of different rate groups defined in this channel's message scenario and log the group index for each entry */
               lastRateValue = channelMsgBlock[0].rate;   
               msgRateGroupRate[0] = channelMsgBlock[0].rate;
               msgRateGroupCount[0] = 1;
               for (j = 1; j < numberOfEntries; j++) {
                  if (channelMsgBlock[j].rate != lastRateValue) {
                     numberOfRateGoups++;
                     msgRateGroupRate[numberOfRateGoups] = channelMsgBlock[j].rate;
                     msgRateGroupCount[numberOfRateGoups]++;
                     lastRateValue = channelMsgBlock[j].rate;   
                  }
                  else
                     msgRateGroupCount[numberOfRateGoups]++;
               }
            
               /* now assign the start offset values to spread message transmissions out across the duration of the scenario */
               if (txChanAttributes[k].busSpeed == ARU_SPEED_HIGH) {
            
                  /* Process 100Kbps bus speed scheduled message scenario start offsets, never more than 3 messages per millisecond */
                  i = 1;
                  startOffset = 0;
                  channelMsgBlock[0].start = 0;   
                  lastRateValue = channelMsgBlock[0].rate;   
                  msgRateGroupFirstStartOffset[0] = 0;
                  msgCountThisMillisec = 1;
                  msgCountAllowed = 4;
                  currentRateGroup = 0;
                  currentRateGroupCount = 0;
                  incrementStartOffset = 0;
                  increment100sStartOffset = 0;
                  halfMsgRateGroupCount = 0;
                  intoHalfMsgRateGroupCount = 0;
                  startOffsetBeforeHalfMsgRateGroup = 0;
                  msgRateGroupOffsetStartOffset[0] = 0;
                  while (i < numberOfEntries) {
                     /* if the next entry is in a new rate group, flag it and limit the number of messages at this offset to 2 */
                     if (channelMsgBlock[i].rate != lastRateValue) {
                        if (intoHalfMsgRateGroupCount) {
                           msgRateGroupLastStartOffset[currentRateGroup] = startOffsetBeforeHalfMsgRateGroup;
                        }
                        else {
                           msgRateGroupLastStartOffset[currentRateGroup] = startOffset;
                        }
                        if ((msgRateGroupRate[currentRateGroup] > 100) && (intoHalfMsgRateGroupCount)) {
                           startOffset = startOffsetBeforeHalfMsgRateGroup;
                        }
                        currentRateGroup++;
                        currentRateGroupCount = 0;
                        intoHalfMsgRateGroupCount = 0;
                        msgRateGroupFirstStartOffset[currentRateGroup] = 0xFFFFFFFF;
                        msgRateGroupOffsetStartOffset[currentRateGroup] = 0;
                        if (++msgCountThisMillisec > 1) {
                           incrementStartOffset = 1;
                           msgCountThisMillisec = 0;
                        }
                        if (msgRateGroupRate[currentRateGroup] > 100) {
                           halfMsgRateGroupCount = msgRateGroupCount[currentRateGroup]/2;
                        }
                     }
                     else {
                        currentRateGroupCount++;
               
                        /* if the current offset selection is defined to allow for 3 messages but we're hitting a mulitple of message having a higher rate,
                         * don't allow 3 messages at this offset as it will delay the higher rate message transmission by 80 microseconds */
                        if (((msgCountThisMillisec+1) >= 2) && (msgCountAllowed == 3)) {
                           for (rateGroupIndex = 0; rateGroupIndex < currentRateGroup; rateGroupIndex++) {
                              if (((startOffset+1) % msgRateGroupRate[rateGroupIndex]) == 0) {
                                 msgCountAllowed = 2;
                              }
                           }
                        }
                        /* handle the standard toggle between allowing 2 and 3 messages per offset when there are no conflicts */
                        if (++msgCountThisMillisec >= msgCountAllowed) {
                           incrementStartOffset = 1;
                           msgCountThisMillisec = 0;
                           msgCountAllowed = msgCountAllowed > 2 ? 2 : 3;
                        }
                     }
                     /* if we're moving on to the next start offset value, assure it isn't a 'retransmission' multiple of a previous rate group */
                     if (incrementStartOffset) {
                        incrementStartOffset = 0;
                        startOffsetIncrementValue = startOffset+1;
                        startOffsetAdjusted = 0;
                        if (currentRateGroup > 0) {
               
                           for (rateGroupIndex = 0; rateGroupIndex < currentRateGroup; rateGroupIndex++) {
                              if ( (((startOffsetIncrementValue % msgRateGroupRate[rateGroupIndex]) >= msgRateGroupFirstStartOffset[rateGroupIndex]) && 
                                    ((startOffsetIncrementValue % msgRateGroupRate[rateGroupIndex]) <= msgRateGroupLastStartOffset[rateGroupIndex])) ||
                                   ((msgRateGroupRate[rateGroupIndex] > 100) &&
                                    (((startOffsetIncrementValue % 100) >= msgRateGroupFirstStartOffset[rateGroupIndex]) &&
                                     ((startOffsetIncrementValue % 100) <= msgRateGroupLastStartOffset[rateGroupIndex]))) ||
                                   (((startOffsetIncrementValue % msgRateGroupRate[rateGroupIndex]) == 0) || (((startOffsetIncrementValue+1) % msgRateGroupRate[rateGroupIndex]) == 0))
                                  ) {
                                 
                                 if ( (msgRateGroupRate[currentRateGroup] >= 100) && (increment100sStartOffset == 0) && (currentRateGroupCount > 1) && ((startOffsetIncrementValue % 100) == 99)) {
                                    msgRateGroupOffsetStartOffset[currentRateGroup] = increment100sStartOffsetOffset;
                                    increment100sStartOffset = 1;
                                 }
            
                                 startOffsetIncrementValue = (startOffsetIncrementValue+1) / msgRateGroupRate[rateGroupIndex];
                                 startOffsetIncrementValue *= msgRateGroupRate[rateGroupIndex];
                                 startOffsetIncrementValue += msgRateGroupLastStartOffset[rateGroupIndex] + 1;
                                 
                                 msgCountAllowed = 3;
                                 startOffsetAdjusted = 1;
                              }
                           }
                           startOffsetIncrementValue += msgRateGroupOffsetStartOffset[currentRateGroup];
                           msgRateGroupOffsetStartOffset[currentRateGroup] = 0;
               
                           if ((startOffsetAdjusted == 0) && (halfMsgRateGroupCount > 0)) {
                              if ((currentRateGroupCount >= halfMsgRateGroupCount) && (intoHalfMsgRateGroupCount == 0)) {
                                 startOffsetBeforeHalfMsgRateGroup = startOffsetIncrementValue;
                                 startOffsetIncrementValue = msgRateGroupFirstStartOffset[currentRateGroup] + 100 + (increment100sStartOffset * 100);
                                 increment100sStartOffsetOffset += 100;
                                 increment100sStartOffset = 0;
                                 intoHalfMsgRateGroupCount = 1;
                             }
                           }
                        }
                        /* apply the modified start offset value */
                        startOffset = startOffsetIncrementValue;
                     }
                     lastRateValue = channelMsgBlock[i].rate;
                     channelMsgBlock[i++].start = startOffset;   
                     if (msgRateGroupFirstStartOffset[currentRateGroup] == 0xFFFFFFFF) {
                        msgRateGroupFirstStartOffset[currentRateGroup] = startOffset;
                     }
                  }
               }
               else {
                  /* Process 12.5Kbps bus speed scheduled message scenario start offsets, never more than 1 message every 3 milliseconds */
                  i = 1;
                  startOffset = 0;
                  channelMsgBlock[0].start = 0;   
                  lastRateValue = channelMsgBlock[0].rate;   
                  msgRateGroupFirstStartOffset[0] = 0;
                  currentRateGroup = 0;
                  currentRateGroupCount = 0;
                  halfMsgRateGroupCount = 0;
                  intoHalfMsgRateGroupCount = 0;
                  startOffsetBeforeHalfMsgRateGroup = 0;
                  while (i < numberOfEntries) {
                     /* if the next entry is in a new rate group, flag it and limit to only one message during this 3 millisecond offset entry */
                     if (channelMsgBlock[i].rate != lastRateValue) {
                        if (intoHalfMsgRateGroupCount) {
                           msgRateGroupLastStartOffset[currentRateGroup] = startOffsetBeforeHalfMsgRateGroup-3;
                        }
                        else {
                           msgRateGroupLastStartOffset[currentRateGroup] = startOffset;
                        }
                        if ((msgRateGroupRate[currentRateGroup] > 100) && (intoHalfMsgRateGroupCount)) {
                           startOffset = startOffsetBeforeHalfMsgRateGroup-3;
                        }
                        currentRateGroup++;
                        msgRateGroupOffsetStartOffset[currentRateGroup] = 0;
                        currentRateGroupCount = 0;
                        intoHalfMsgRateGroupCount = 0;
                        msgRateGroupFirstStartOffset[currentRateGroup] = 0xFFFFFFFF;
                        if (msgRateGroupRate[currentRateGroup] > 100) {
                           halfMsgRateGroupCount = msgRateGroupCount[currentRateGroup]/2;
                           if ((msgRateGroupCount[currentRateGroup] % 2) == 1) {
                              halfMsgRateGroupCount++;
                           }
                        }
                     }
                     else {
                        currentRateGroupCount++;
                     }
            
                     /* when incrementing to the next available start offset value, assure it isn't a multiple of a previous rate group */
                     startOffsetIncrementValue = startOffset+3;
                     if (currentRateGroup > 0) {
               
                        for (rateGroupIndex = 0; rateGroupIndex < currentRateGroup; rateGroupIndex++) {
                           if ( ((startOffsetIncrementValue % msgRateGroupRate[rateGroupIndex]) == 0) || 
                                (((startOffsetIncrementValue+1) % msgRateGroupRate[rateGroupIndex]) == 0) || 
                                (((startOffsetIncrementValue+2) % msgRateGroupRate[rateGroupIndex]) == 0) 
                              ) {
            
                              startOffsetIncrementValue = (startOffsetIncrementValue+3) / msgRateGroupRate[rateGroupIndex];
                              startOffsetIncrementValue *= msgRateGroupRate[rateGroupIndex];
                              startOffsetIncrementValue += msgRateGroupLastStartOffset[rateGroupIndex] + 3;
                           }
                           if (((startOffsetIncrementValue % msgRateGroupRate[rateGroupIndex]) >= msgRateGroupFirstStartOffset[rateGroupIndex]) &&
                               ((startOffsetIncrementValue % msgRateGroupRate[rateGroupIndex]) <= msgRateGroupLastStartOffset[rateGroupIndex]) ) {
                              if (msgRateGroupRate[currentRateGroup] > 100) {
                                 startOffsetIncrementValue = msgRateGroupFirstStartOffset[currentRateGroup] + msgRateGroupOffsetStartOffset[currentRateGroup] + 100;
                                 msgRateGroupOffsetStartOffset[currentRateGroup] += 100;
                              }
                              else {
                                 startOffsetIncrementValue = msgRateGroupLastStartOffset[rateGroupIndex] + 3;
                              }
                           }
                        }
               
                        if (halfMsgRateGroupCount > 0) {
                           if ((currentRateGroupCount >= halfMsgRateGroupCount) && (intoHalfMsgRateGroupCount == 0)) {
                              startOffsetBeforeHalfMsgRateGroup = startOffsetIncrementValue;
                              startOffsetIncrementValue = msgRateGroupFirstStartOffset[currentRateGroup] + msgRateGroupOffsetStartOffset[currentRateGroup] + 100;
                              msgRateGroupOffsetStartOffset[currentRateGroup] += 100;
                              intoHalfMsgRateGroupCount = 1;
                           }
                        }
                     }
                     /* apply the modified start offset value */
                     startOffset = startOffsetIncrementValue;
            
                     lastRateValue = channelMsgBlock[i].rate;
                     channelMsgBlock[i++].start = startOffset;   
                     if (msgRateGroupFirstStartOffset[currentRateGroup] == 0xFFFFFFFF) {
                        msgRateGroupFirstStartOffset[currentRateGroup] = startOffset;
                     }
                  }
               }
         
               /* now replace this channel's message set in the pre-sorted full message scheduler table array for the board */
               memcpy (&localMsgBlock[n], channelMsgBlock, (sizeof(SCH_MSG_ENTRY_TYPE) * txChanAttributes[k].numberOfEntries));
               n += txChanAttributes[k].numberOfEntries;
            }
         
            /* Sorting and offset assignment is complete for all channel message scenarios...
             * For each scheduler table entry processed, write the computed start offset value to the respective message scheduler entry 
             */
   
#ifndef TARGET_EMBEDDED
            /* if multi-thread protection requested, acquire semaphore/mutex-locking */
            if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
               accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
            }
            if (accessStatus == ARS_NORMAL) {
#endif
               for (i = 0; ((i < totalNumberOfEntries) && (accessStatus == ARS_NORMAL)); i++) {
                  memOffset = CEIDEV_MSG_SCHED_TABLE_START + CEIDEV_MSG_SCHED_TBL_START_OFFSET + (localMsgBlock[i].messageIndex * CEIDEV_MSG_SCHED_TBL_ENTRY_SIZE);
                  accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], localMsgBlock[i].start);
               }
#ifndef TARGET_EMBEDDED
               /* if multi-thread protection is active, free any acquired lock */
               if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
                  accessStatus = cei_release_thread_lock(LOCK_IDENT);
               }
            }
#endif
         }
         free (localMsgBlock);
         free (channelMsgBlock);
         free (msgRateGroupRate);
         free (msgRateGroupOffsetStartOffset);
         free (msgRateGroupFirstStartOffset); 
         free (msgRateGroupLastStartOffset);
         free (msgRateGroupCount);
      }
   }

   if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL)) {
      status = (CEI_INT32)accessStatus;
   }
   
   return status;
#else
   return ARS_NORMAL;
#endif
}


