/*====================================================================================================================*
 * FILE:                                     C D E V _ A P I . C                    
 *====================================================================================================================*
 *
 *   COPYRIGHT 2002 - 2019 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED
 *   ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH THE INCLUSION
 *   OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY OTHER COPIES THEREOF
 *   MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE
 *   AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS, INC. 
 *
 *====================================================================================================================*
 *
 * FUNCTION:    The Basic Application Programmer's Interface for a CEI-x30 ARINC device.
 *
 * DESCRIPTION: These routines are provided to demonstrate the basic programming primitives for the CEI-x30 API 
 *              using the C programming language.  Each routine implements a particular API function; however, 
 *              a user application program may need to combine functions in different ways to perform the desired
 *              operation.
 *
 *              Most routines utilize a board number parameter, zero-based on the Device ID/Board Number assigned 
 *              during the installation of CEI-x30-SW distribution.  Channel parameter values are all zero-based.
 *
 * EXTERNALLY DEFINED ROUTINE ENTRY POINTS:         
 *
 *    CEI-x30 ARINC API routines:
 *
 *    ar_board_test               - init-time internal functionality test
 *    ar_bypass_wrap_test         - controls execution of init wrap test
 *    ar_close                    - relinquishes WinRT device resources
 *    ar_clr_rx_count             - resets the API received message counter for Rx channels
 *    ar_execute_bit              - performs BIT functionality testing of the device
 *    ar_get_device_config        - returns ARINC 429 & I/O chan attributes  
 *    ar_get_error                - returns error string for a status code
 *    ar_get_rx_channel_status    - retrieves the current receive channel buffer status 
 *    ar_get_rx_count             - retrieves the accululative count of messages received on a channel
 *    ar_get_time                 - returns the current 64-bit device timer value
 *    ar_getblock_t               - read a block of messages from a receiver FIFO
 *    ar_getword                  - read a single message data from a receiver FIFO
 *    ar_getword_xt               - read data and 64-bit timetag from a receiver FIFO
 *    ar_get_latest_t             - retrieves latest data and time-tag from the snapshot buffer
 *    ar_get_snap_data            - retrieves latest data from the snapshot buffer
 *    ar_go                       - enable data processing all channels on the board 
 *    ar_initialize_api           - initializes the device API, downloads firmware
 *    ar_initialize_device        - initializes the device
 *    ar_num_rchans               - returns the no. of recv channels
 *    ar_num_xchans               - returns the no. of xmit channels
 *    ar_open                     - opens a session, then initializes the API and the device
 *    ar_putblock                 - write a block of messages to a transmitter FIFO
 *    ar_putword                  - send an ARINC word on a channel   
 *    ar_reset                    - reset the device to an initial state
 *    ar_set_device_config        - set ARINC 429 and I/O chan attributes  
 *    ar_set_multithread_protect  - turn on specific hardware access protection
 *    ar_set_preload_config       - defines a pre-init (open) multi-thread/process setting
 *    ar_set_storage_mode         - select buffered/merged receive modes
 *    ar_set_time                 - sets the current 64-bit device timer value
 *    ar_sleep                    - suspends call thread by milliseconds
 *    ar_stop                     - disable the device via the global-enable
 *
 * INTERNAL UTILITY ROUTINES:
 *
 *    a429_utl_acquire_formatted_time_tag
 *    a429_utl_configure_programmable_channel
 *    a429_utl_determine_channel_support
 *    a429_utl_initialize_global_data
 *    a429_utl_initialize_device
 *    a429_utl_memory_test
 *    a429_utl_read_message
 *    a429_utl_validate_channel_request
 *    a429_utl_wrap_test
 *
 *--------------------------------------------------------------------------------------------------------------------*
 *          
 * HISTORY:   
 * 01/03/04 to 02/02/09 prior history for standard firmware changes found in v1.70 source for CDEV_API.C
 *
 *   Date    By   Ver. Comments
 * -------- ----- ---- -----------------------------------------------------------------------------------------------
 * 04/15/09  BDW  2.00 Incorporated support for the CEI-x30 Enhanced Firmware Design.
 *                     - Modified all methods to access Channel Registers and Buffers,
 *                          Message Schedule Table, Label Filter Table, and Snapshot Buffer.
 *                     - Added Interrupt Queue access routines and generic interrupt
 *                          handler (with provisions for designating a user-defined handler):
 *                             cei_utl_interrupt_queue_read
 *                             cei_utl_interrupt_handler
 *                             ar_hw_interrupt_buffer_read
 *                             ar_interrupt_queue_read
 *                             ar_set_isr_function
 *                     - Added the routine ar_get_latest_t to provide a method to 
 *                          retrieve the label-only snapshot mode message & time-stamp
 *                     - Implemented the functional support within the following API 
 *                          routines for backward compatibility to the CEI-x20 product line/API:
 *                             ar_get_boardname     ar_init_dual_port
 *                             ar_get_label_filter  ar_init_slave
 *                             ar_get_timercnt      ar_int_control
 *                             ar_sleep
 *                     - Added support for the ARS_BAD_MESSAGE return status from the following API routines, indicating
 *                          a "bad msg length" ARINC 429 message was encountered on the respective receiver:
 *                             ar_getword       ar_getwordt       ar_getword_xt
 *                             ar_getnext       ar_getnextt       ar_getnext_xt
 *                             ar_getblock      ar_getblock_t     ar_get_429_message
 *                             ar_get_data      ar_get_data_xt
 * 06/16/09  BDW  2.00 Replaced call to cei_open_shared_lock in ar_set_preload_config with the new function called
 *                        cei_open_preload_lock.
 *                     Removed support for compressed firmware download.
 * 06/23/09  BDW  2.00 Added a test for the Enhanced CEI-x30 Interface to ar_get_device_config and the invocation to update
 *                        any standard device to the enhanced i/f in ar_set_device_config.
 *                     Added the utility routines to support the EEPROM update.
 *                     Renamed all local subroutines to have an "a429_utl_" prefix.
 * 09/08/09  BDW  2.10 Fixed merged mode rx channel reporting issue in ar_getblock_t().
 * 09/30/09  BDW  2.10 Added support for ar_set_base_addresses routine for the host to assign device BAR0 and BAR2 resources
 *                        for embedded linux.
 * 05/19/10  BDW  2.10 Added support for the RAR-PCIE to the following items:
 *                        ar_initialize_api (bypass f/w download, copy BAR0 ptr to BAR2 ptr)
 *                        ar_query_device (identify RAR-PCIE board type)
 *                        boardNames array adds next board type name RAR-PCIE
 * 09/16/10  BDW  3.00 Renamed interruptFunctionPtr to x30InterruptFunctionPtr to eliminate duplicate name with the
 *                        PIO product line API
 * 11/01/10  BDW  3.00 Modified the return type for ar_get_base_addr to be a pointer
 * 01/05/11  BDW  3.10 Corrected slew rate assignment for custom baud rate selections in ar_set_device_config.
 * 02/01/11  BDW  3.10 Modified the operation of the routines ar_modify_msg_block and a429_utl_store_message_scheduler_entry
 *                        to allow a message scheduler table entry's transmit count to be modified if it had been created
 *                        with or previously modified to have a limited transmit count and the current transmit count is 0.
 * 02/15/11  BDW  3.10 Corrected the CEI-x20 rate and start offset "programmable timer" resolution conversion in
 *                        a429_utl_store_message_scheduler_entry.
 *                     Corrected the snapshot buffer table content returned from ar_get_latest when the entire table
 *                        is requested using the label parameter ARU_ALL_LABELS option.
 *                     Corrected the function of ar_init_slave so it only resets/flushes the transmit and receive buffers
 *                        and does not initialize the snapshot buffer, label filter table, or message scheduler table.
 * 04/05/11  BDW  3.10 Replaced LABVIEW_RT directive use for PLX accesses for LabVIEW Real-Time o/s with the
 *                        directive NIVISA_PXI.
 *                     Added support for creating a 1 second host time format, AR_TIMETAG_HOST_SEC_32BIT in
 *                        ar_convert_time_to_string.
 *                     Corrected an issue in ar_getblock_t, where in merged mode the actual receive channel isn't
 *                        returned unless time-tags are returned (time-tag array pointers are not NULL).
 * 04/29/11  BDW  3.10 Replaced use of cei_read/write_32bit function invocations in the utl_plx_eeprom_* routines with the
 *                        respective READ/WRITE_32_BIT macros to support PROM updates under LabVIEW Real-Time.
 * 08/15/11  BDW  3.10 Introduced support for CEI-x20 discrete I/O set/get options ARU_DISCRETE_VALUES, ARU_DISCRETE_INPUTS
 *                        and ARU_DISCRETE_OUTPUTS in ar_set_config and ar_get_config.
 * 09/16/11  BDW  3.10 Added CEI-430A support with additional boardNames element and modifications to the following functions:
 *                        a429_utl_enh & a429_utl_enh_check (formality, not necessary)
 *                        ar_initialize_api to support identification and f/w download
 *                        ar_query_device to report the device type
 *                   
 *   Date    By   Ver. Comments
 * -------- ----- ---- -----------------------------------------------------------------------------------------------
 * 01/10/12  BDW  4.00 Modified ar_setchparms to enable 717 transmitter/receiver fifo
 * 02/06/12  BDW  Beta Added support for ar_set_board_identity routine for the host to assign device PCI manufacturer and device
 *                        IDs for embedded linux.
 * 02/17/12  BDW   "   Modified the initialization of x30InterruptFunctionPtr[] to allow MAX_BOARDS of 128.
 * 02/21/12  BDW   "   Modified a429_utl_acquire_formatted_time_tag to support synchronization of ARINC 429 message time stamps
 *                        to RAR15-XMC 1553 channels via timeTagFormat parameter options AR_TIMETAG_SYNC_1553_CHn.
 *                     Switched default Snapshot Buffer storage to be label-only
 * 03/16/12  BDW   "   Added assignment of the ar_get_latest sequence_number output parameter to the time-stamp least
 *                        significant byte
 * 07/31/12  BDW  4.01 Updated the boardNames[] array to include the RAR15-XMC and RAR15-XMC-XT board ID strings
 *    to          Beta Added the routine ar_get_rx_channel_status to support the reporting of receive channel buffer status
 * 09/28/12            Added a channel indexing info structure type and routine AR_CHANNEL_INDEX_INFO_TYPE and
 *                        ar_get_channel_index_info for applications to deal with programmable and inactive channels
 * 11/28/12  BDW  4.10 Added the compiler directive MULTI_PROTOCOL_SUPPORT around external references to the BusTools/1553-API
 *                        time conversion references used in the RAR15-XMC-XT Linux, VxWorks, and Integrity compilations.
 * 02/19/13  BDW  4.10 Modified the routine ar_set_preload_config, when the invocation is for AR_CONC_MULTIPROC (multiprocess
 *                        support) and the API is compiled for a LINUX host (_LINUX_X86_ or _LINUX_PPC_ defined), and the 
 *                        directive MULTI_PROCESS_SUPPORT is not defined, return an error (avoiding semaphore usage conflicts).
 * 03/20/13  BDW  4.10 Added PCI Express I/O Synchronization macro to the SRAM Pattern Test in the 
 *                        routine a429_utl_memory_test(); this requires the host compiler to define
 *                        the directive PPC_SYNC for Integrity or VxWorks target operating systems.
 * 03/28/13  BDW  4.10 Added support for the RCEI-830X820 as follows:
 *                        Added the board name RCEI-830X820 to boardNames[]
 *                        Added return of the new board type CEI830X820 (34) to ar_get_boardtype
 *                        Added the board type check for PCI_DEVICE_ID_830X820 for enhancement checks
 *                        Added firmware download via a429_utl_program_xilinx_9056 in ar_initialize_api
 *                        Added return of the board type CEI830X820 to ar_query_device
 *                        Added additional check for CEIDEV_CONFIG_R830X820 for 9056 PCI Interrupt setup
 *                        in ar_set_device_config with item ARU_HW_INTERRUPT_ENABLE
 * 04/08/13  BDW  4.20 Added support for the RCEI-830X820 device detection and firmware download processing.
 * 06/13/13  BDW  4.30 Release full distribution for the (R)AR15-XMC-XT and RCEI-830X820 boards.  
 * 08/28/13  BDW  4.40 Modified the pipeline flush implementation for Integrity o/s; see use of compiler directive
 *                        INTEGRITY_PPC_SYNC; removed invocation of AppIntMemBarrier, replaced with inline assembly.
 * 10/01/13  BDW  4.40 Added support for PCI9056 byte-wide file-based f/w download to a429_utl_program_xilinx_9056,
 *                        via alternate build with the compiler directive FW_830_FILE_LOAD defined.
 * 10/03/13  BDW  4.40 Added support for the RCEI-830A device detection and firmware download processing.
 * 10/25/13  BDW  4.40 Change the address parameters for the routine ar_set_base_addresses from fixed 32-bit values
 *                        to compiler-directed 32-bit or 64-bit values.
 *                     Fixed a bug in ar_initialize_device where the Merged Receive Buffer flush didn't flush.
 *                     Restricted multi-thread protection functionality in a429_utl_acquire_formatted_time_tag()
 *                        to only execute for a board timer read and conversion, as receive buffer message & time-tag  
 *                        read routines have multi-thread protection already enforced within their execution.
 * 01/08/14  BDW  4.50 Modified data type for ARINC_LOCAL_TYPE and how it is used specifically as a pointer in Linux
 * 01/08/14  BDW  4.50 Modified data type for the Linux build definition of ARINC_LOCAL_TYPE member sharedMemoryHandle
 *                        and the corresponding cei_map_shared_memory parameter blockHandle used to acquire the handle,
 *                        so they are integer & pointer to and not void.  Added Linux specific compiler-directive  
 *                        controlled code to relinquish the handle in the ar_close routine.
 * 05/02/14  BDW  4.60 Corrected a bug in ar_get_data, where the time tag format structure member needed to be properly 
 *                        initialized to the value of the current board-wide receive message time-tag reference.
 * 05/27/14  BDW  4.60 Corrected a bug in any routine invoking a429_utl_read_message that returns a modified time-tag
 *                        (not directly returning the AR_TIMETAG_TYPE structure), to return the time-tag value when
 *                        an invalid ARINC 429 message had been detected in execution of a429_utl_read_message.
 * 06/03/14  BDW  4.60 Corrected the directive-specific invocation of cei_get_open_error for the Integrity compilation.
 * 06/11/14  BDW  4.60 Applied the 'U' suffix to unsigned integer constant values
 *                     Removed bypassWrapTest array initialization from the routine a429_utl_initialize_global_data and
 *                        made it a run-time initialized array.
 *                     Modified the third ar_get_time parameter namefrom timeTag to boardTime, usage remains the same
 *                     Removed the volatile declaration from the global array deviceGlobal[] and added it to device[]
 *                        (selective compilation) along with every local variable referencing device[] (see #define VOLATILE). 
 * 10/07/14  BDW  4.60 Modified ar_get_573_frame to bypass the built-in 1 second delay for reception of the "next 
 *                        sub-frame" if the receive buffer already contains a full 717 frame's worth of messages.
 *                     Assigned the default state of the Discrete Outputs to be tri-state within the ar_initialize_device
 *                        routine (invoked from ar_loadslv and ar_reset).
 * 01/07/15  BDW  4.60 Added utility routine a429_utl_configure_programmable_channel to support software programmable 
 *                     channel configuration as a transmitter or receiver, invoked from ar_set_device_config using the
 *                     item parameter value ARU_CONFIG_PROGRAMMABLE_CHAN.
 *                   
 *   Date    By   Ver. Comments
 * -------- ----- ---- -----------------------------------------------------------------------------------------------
 * 03/16/15  BDW  4.70 Added support for RAR15-XMC-VCP receiver BIT f/w to ar_execute_bit via AR_BIT_RX_LOOPBACK option.
 *                     Added API return status ARS_RX_BUFFER_OVERRUN indicating a receive buffer overrun was detected
 *                     in all ar_get**** routines (word/wordt/word_xt, next/nextt/next_xt, _data/_data_xt, _429_message, 
 *                     block, block_t).
 * 04/23/15  BDW  4.70 Corrected a bug in ar_put_573_frame when a transmit buffer overflow condition is encountered.
 * 06/23/15  BDW  4.71 Corrected a bug in a429_utl_wrap_test in which the wrap test will fail for unbalanced channel
 *                     configurations where transmitter channel count exceeds receiver channel count.
 * 07/01/15  BDW  4.71 Added the utility routine a429_utl_initialize_device so power-up initialization won't include
 *                     delays only desirable when a "reset" initialization sequence is called by the host app (same code).
 * 08/07/15  BDW  4.80 Increased MAX_BOARDS from 8 to 16 for VxWorks and Integrity targets
 *                     Added support for ARINC 585 24-bit message size feature via ar_set_device_config options
 *                        ARU_TX_MSG_SIZE_24BIT/ARU_RX_MSG_SIZE_24BIT
 *                     Reduced the amount of Label Filter Table initialization in the routine a429_utl_initialize_device
 *                        to only include memory allocation for the installed receive channel count; incorporated direct
 *                        access to the Label Filter Table SRAM mapped into the RAR15-XMC host interface.
 * 02/02/16  BDW  4.90 Added the function ar_modify_msg_block_lv in support of LabVIEW message scheduler usage.
 * 02/19/16  BDW  4.90 Added use of the ARINC_GLOBAL_TYPE structure member array receiveChannelBufferMode[] to store
 *                        the selected buffering mode for later reference, to remove the additional device read access
 *                        in the read receive message FIFO buffer routines (excluding the ar_get_573_frame routine).
 * 02/24/16  BDW  4.90 Added the API function ar_open as an alternative open session function excluding the unused
 *                        parameters required with the ar_loadslv function invocation.
 *                     Added the API function ar_assign_scheduler_start_offsets to perform a best estimate assignment
 *                        of start offset values to an existing message scheduler table definition to assist in
 *                        avoiding message rate skew on individual transmit channels; supported under Windows and
 *                        Linux only.
 * 04/12/16  BDW  4.90 For all receive message retrieval routines, modified the status checks to read the message or
 *                        specified number of messages when a "buffer overrun" is detected instead of leaving the
 *                        buffer in a 'full' state and reading nothing.
 *                     For ar_putblock and ar_put_573_frame, instead of reading the buffer status and checking for an
 *                        overrun with each message inserted in the buffer, extract the buffer fill level and insert
 *                        as much as the buffer will hold (if less than what was requested by the host application).
 * 05/05/16  BDW  4.90 Added support for the receive buffer clear option ARU_RX_FIFO_RESET in ar_set_device_config
 * 06/10/16  BDW  4.90 Modified a429_utl_is_irig_calibrated to query the IRIG Receiver 5 times over 500ms in support
 *                        of the scenario when the IRIG sample test is invoked immediately after a board reload.
 * 08/03/16  BDW  4.90 Added RAR-MPCIE support with additional boardNames element and modifications to the following functions:
 *                        ar_get_boardtype to report the CEI-x30 board type
 *                        ar_query_device to report the PCI device type
 * 10/18/16  BDW  4.90 Split the CEI-x30 API source into feature-specific and target build specific source files.
 *                        Introduced use of CEIx30 API-specific type defines via the file CEIX30_TYPES.H allowing
 *                        for easy build to common use of only signed integer return types and unsigned integer
 *                        parameter types (with the exception of 64-bit types).
 * 12/13/16  BDW  4.90 Corrected a bug in ar_open, the global structure deviceGlobal should be set to NULL on a 
 *                        failed session open attempt (ar_initialize_api).
 * 06/26/17  BDW  5.00 For RAR15-XMC, added an overall Certification firmware API-compliant directive control over source 
 *                        code build via use of the CERTIFICATION_BUILD compiler directive.  Features excluded include Merged 
 *                        Mode receive buffering, access to anything dealing with IRIG, and legacy time-tag conversions.
 *                        Inserted existing compiler directives around initialization of host interface registers and memory
 *                        that may not exist in the cert f/w: INCLUDE_IRIG, INCLUDE_ARINC_717, INCLUDE_DISCRETE_IO
 *                        referenced in the respective routines a429_utl_initialize_device and ar_set/get_device_config
 * 08/09/17  BDW  5.00 In ar_set_device_config, the routine previously forced an Rx channel FIFO reset prior to
 *                        changing any receiver attribute.  This requirement was reduced such that a FIFO reset
 *                        was only enforced when a message size, merged/ind rx buffer, or baud rate change was
 *                        requested from the caller.
 * 08/09/17  BDW  5.00 Modified several routines to return the status value ARS_INVHARVAL if an application attempts to 
 *                        request IRIG status or perform an IRIG operation with a board not configured with IRIG:
 *                        - ar_set_time modified to return an error if the application attempts to assign an IRIG time.
 *                        - ar_get_time modified to return an error if any IRIG format option is passed by the application.  
 *                        - ar_get_device_config modified to return an error if an application requests the API perform 
 *                          an IRIG calibration or request sample status.
 * 01/17/18  BDW  5.10 Modified a429_utl_validate_channel_request and ar_get_rx_channel_status to allow a request for 
 *                        receive channel status (ARU_RX_FIFO_COUNT) on the merged receive channel (63)
 * 12/06/19  BDW  5.10 Modified a429_utl_initialize_device, inserting a 10ms delay between conditionally successive
 *                        DAC register write accesses to allow DAC serial processing of threshold configuration.
 *====================================================================================================================*/ 

/*--------------------------------------------------------------------------------------------------------------------*
 *                                           GENERAL INCLUDE STATEMEMTS
 *--------------------------------------------------------------------------------------------------------------------*/

#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
   #include <unistd.h>
   #include <fcntl.h>
   #include <string.h>
   #include <pthread.h>
   #include <errno.h>
   #include <sys/stat.h> 
   #include <sys/shm.h>
   #include <sys/sem.h>
   #include <sys/syscall.h>
   #ifndef max
      #define max(a,b) (a>b?a:b)
   #endif
   #define VOLATILE volatile
#elif defined (VXW_PCI_X86) || defined (VXW_PCI_PPC)
   #include <stdio.h>
   #include <stdlib.h>
   #include <unistd.h>
	#include "target_defines.h"
   #define VOLATILE 
#elif defined (NIVISA_PXI)
   #ifdef LABVIEW_RT
      #include <ansi_c.h> 
   #endif
   #include <visa.h>  
   #define VOLATILE 
#elif defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
   #include <string.h>
   #define max(a,b) (a>b?a:b)
   #define VOLATILE 
#else
   #define VOLATILE volatile
#endif

/* include local header files, CDEV_API.H included in CDEV_GLB.H */
#include "cdev_glb.h"

#ifndef CERTIFICATION_BUILD
#include "Ar_error.h"
#include "cdev_fw.h"
#endif

#ifdef INCLUDE_ARINC_717
extern EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_set_573_config (CDEV_BOARD_TYPE board, CDEV_PARM_SSI_TYPE item, CDEV_PARM_SI_TYPE value);  
#endif

/**********************************************************************************************************************
 **********************************************************************************************************************
 *                                          A P I   P R I V A T E   D A T A                                           *
 **********************************************************************************************************************
 **********************************************************************************************************************/
   
/**********************************************************************************************************************
 *                             I N I T I A L I Z E D   P R I V A T E   D A T A                                        *
 **********************************************************************************************************************/
/* pointer to global (multi-process capable) shared board parameters */
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_) || defined (VXW_PCI_X86) || defined (VXW_PCI_PPC) || defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
ARINC_GLOBAL_TYPE * deviceGlobal[MAX_BOARDS] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};
#elif defined (NIVISA_PXI)
ARINC_GLOBAL_TYPE * deviceGlobal[MAX_BOARDS] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};
#else
ARINC_GLOBAL_TYPE * deviceGlobal[MAX_BOARDS] = {
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
};
#endif

/* multi-process execution is disabled by default, altered via ar_set_preload_config() invocation */
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_) || defined (VXW_PCI_X86) || defined (VXW_PCI_PPC) || defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
static CEI_UINT32 multiProcessProtect[MAX_BOARDS] = {AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, 
                                                     AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE};
#elif defined (NIVISA_PXI)
static CEI_UINT32 multiProcessProtect[MAX_BOARDS] = {AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, 
                                                     AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE};
#else
static CEI_UINT32 multiProcessProtect[MAX_BOARDS] = {
   AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, 
   AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE,
   AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, 
   AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE,
   AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, 
   AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE,
   AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, 
   AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE,
   AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, 
   AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE,
   AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, 
   AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE,
   AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, 
   AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE,
   AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, 
   AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE, AR_CONC_NONE
};
#endif

/* bypass wrap test by default, altered via ar_bypass_wrap_test() invocation */
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_) || defined (VXW_PCI_X86) || defined (VXW_PCI_PPC) || defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
static CEI_UINT32 bypassWrapTest[MAX_BOARDS] = {TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE};
#elif defined (NIVISA_PXI)
static CEI_UINT32 bypassWrapTest[MAX_BOARDS] = {TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE};
#else
#if defined DEBUG_PROGRAMMING
extern CEI_UINT32 baseConfig;
#endif
static CEI_UINT32 bypassWrapTest[MAX_BOARDS] = {
   TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE,
   TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE,
   TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE,
   TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE,
   TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE,
   TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE,
   TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE,
   TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE
};
#endif

/**********************************************************************************************************************
 *                P R I V A T E   D A T A   I N I T I A L I Z E D   A T   S E S S I O N   O P E N                     *
 **********************************************************************************************************************/

/* process-local board parameters and hardware interface pointers */
VOLATILE ARINC_LOCAL_TYPE device[MAX_BOARDS];

/* shared board parameters for single-process only implementation */
static ARINC_GLOBAL_TYPE deviceSingleProcessShared[MAX_BOARDS];


/**********************************************************************************************************************
 **********************************************************************************************************************
 *                                   A P I   P R I V A T E   F U N C T I O N S                                        *
 **********************************************************************************************************************
 **********************************************************************************************************************/


/*===================================================================================================================*
 * ENTRY POINT:         A 4 2 9 _ U T L _ A C Q U I R E _ F O R M A T T E D _ T I M E _ T A G
 *===================================================================================================================*
 *
 * FUNCTION:    Read and potentially convert the active receive buffer 
 *              or IRIG time-stamp to the specified format
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board       -- (input) Board number (0 to 127) 
 *
 *              CEI_UINT32 chanMemOffset    -- (input) Offset into channel memory area
 *                                                     to read the time-tag
 *
 *              CEI_UINT32 timeSource       -- (input) Source (timer or receiver FIFO)
 *                                                     AR_TIMER_REGISTER_SET
 *                                                     AR_RECEIVER_FIFO_TIMETAG
 *                                                     AR_BYPASS_TIMER_READ
 *                                              
 *              CEI_UINT32 timeTagFormat    -- (input) Selects the format of the time-tag value returned:
 *                                                     AR_TIMETAG_EXT_IRIG_64BIT
 *                                                     AR_TIMETAG_INT_USEC_64BIT
 *                                                     AR_TIMETAG_INT_20USEC_32BIT *
 *                                                     AR_TIMETAG_INT_MSEC_32BIT   *
 *                                                     AR_TIMER_X20_COMPAT_32BIT   *
 *                                                     AR_USE_PROGRAMMED_TIMETAG_MODE
 *                                                     AR_LOG_EXT_IRIG_SAMPLE
 *                                                     AR_TIMETAG_REF_IRIG_64BIT
 *                                                     AR_TIMETAG_SYNC_1553_CH1
 *                                                     AR_TIMETAG_SYNC_1553_CH2
 *                                                     AR_TIMETAG_SYNC_1553_CH3
 *                                                     AR_TIMETAG_SYNC_1553_CH4
 *                                              
 *              CEI_UINT32 mergedMode       -- (input) buffering mode (merged/dedicated)
 *                                                     When mergedMode < CEIDEV_MAX_CHANNELS, merged mode is active 
 *                                                     and then if a 32-bit time-stamp is requested, the channel 
 *                                                     number is inserted in the upper 5 bits of the time-stamp
 *                                              
 *              pAR_TIMETAG_TYPE timeTagRef -- (output) 64-bit time-tag structure
 *                                              
 * RETURN VAL:  CEI_INT32    -- ARS_NORMAL      Success
 *                              ARS_INVBOARD    Invalid or uninitialized board
 *                              ARS_INVARG      Invalid timeSource or timeTagFormat
 *
 * GLOBALS:     device
 *              deviceGlobal
 * 
 * DESCRIPTION: This routine assumes the desired receive channel has been selected by the calling routine.  It reads 
 *              the 64-bit microsecond resolution time-tag from the receive buffer, then reformats it based on either 
 *              the requested resolution and size or the actively selected resolution/size.
 * 
 *===================================================================================================================*/ 
CEI_INT32 a429_utl_acquire_formatted_time_tag
(
   CDEV_BOARD_TYPE board,                                                /* (input)  Board number (0 to 127)         */
   CEI_UINT32 chanMemOffset,                                             /* (input)  Offset into channel memory area */
   CEI_UINT32 timeSource,                                                /* (input)  Source (timer or receiver FIFO) */
   CEI_UINT32 timeTagFormat,                                             /* (input)  Format of the time-tag returned */
   CEI_UINT32 mergedMode,                                                /* (input)  Buffering mode merged/dedicated */
   pAR_TIMETAG_TYPE timeTagRef                                           /* (output) 64-bit time-tag structure       */
)
{
   CEI_UINT32 tempLowTimeTag = 0U;                                       /* 32-bit LSLW for 64-bit time-tag          */
   CEI_UINT32 tempHighTimeTag = 0U;                                      /* 32-bit MSLW for 64-bit time-tag          */
   CEI_UINT32 localChannel = 0U;                                         /* merged/individual receive channel        */
   CEI_UINT32 localTimeTagFormat;                                        /* local version of time tag format         */
   CEI_INT32 accessStatus = ARS_NORMAL;                                  /* status of device read/write access       */
   CEI_INT32 status = ARS_NORMAL;                                        /* operation status value                   */
   CEI_UINT32 ucaChanTimerLoOffset;                                      /* local index for 1553 sync timer          */
   CEI_UINT32 ucaChanTimerHiOffset;                                      /* local index for 1553 sync timer          */
   TIME_TAG_TYPE timeTag64LowValue;                                      /* temp 64-bit time-tag placeholder         */
   TIME_TAG_TYPE timeTag64HighValue;                                     /* temp 64-bit time-tag placeholder         */
   CEI_UINT32 localOffset;                                               /* local pointer to channel memory area     */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                     /* ptr to board local structure             */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                            /* ptr to brd global structure              */
   
#ifdef INCLUDE_IRIG
   CEI_UINT32 year;                                                      /* Current year (since 1900)                */
   CEI_UINT32 numberOfDays;                                              /* temp IRIG time conversion vars           */
   CEI_UINT32 numberOfMinutes;                                           /* temp IRIG time conversion vars           */
   CEI_UINT32 numberOfSeconds;                                           /* temp IRIG time conversion vars           */
   CEI_UINT32 remainderRoundUp = 0U;                                     /* round-up increment for low res.          */
   CEI_UINT32 appProgTimeTag = 0U;                                       /* Application specified time-tag res       */
   TIME_TAG_TYPE irigOffsetTime;                                         /* temp 64-bit time-tag placeholder         */
   time_t   now;                                                         /* Current time                             */
   struct   tm start;                                                    /* Formatted "now" current time             */
   TIME_TAG_TYPE maxMicroSecondsPerYear;                                 /* Max number of usec in this year          */
#endif

   /* assign references to this board's device and deviceGlobal structure-array elements */
   deviceLocalStructPtr = &device[board];
   deviceGlobalStructPtr = deviceGlobal[board];

   /* Process the source of this pass' time-tag, either the receive buffer or the timer */
   switch (timeSource) {
      case AR_RECEIVER_FIFO_TIMETAG:
         /* retrieve the merged channel identifier when in merged receive mode */
         if (mergedMode != CEIDEV_MAX_CHANNELS) {
            localChannel = (CEI_UINT32)mergedMode;
            localChannel = localChannel << MERGED_CHANNEL_SHIFT;
         }

         /* The low order 32-bits of time tag for the ARINC message is in the Receive 
          * Buffer FIFO Long Word 0 location; high order 32-bits in FIFO Long Word 1.
          * Location of Buffer FIFO Long Word 0 supplied by the calling routine. 
          */
         localOffset = chanMemOffset;
         accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[localOffset++], &tempLowTimeTag);
         if (accessStatus == ARS_NORMAL) {
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[localOffset], &tempHighTimeTag);
         }
         break;

      case AR_TIMER_REGISTER_SET:
#ifndef TARGET_EMBEDDED
         /* if multi-thread protection is active, acquire a lock on the host interface */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            status = cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (status == ARS_NORMAL) {
#endif
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_TIMER_LOW], &tempLowTimeTag);
            if (accessStatus == ARS_NORMAL) {
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_TIMER_HIGH], &tempHighTimeTag);
            }
#ifndef TARGET_EMBEDDED
         }

         /* if multi-thread protection is active, free any acquired lock */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
         }
#endif
         break;

      case AR_BYPASS_TIMER_READ:
         break;

      default:
         accessStatus = ARS_INVARG;
         break;
   }
   
   if (accessStatus == ARS_NORMAL) {
      timeTag64LowValue = (TIME_TAG_TYPE)tempLowTimeTag;
      timeTag64HighValue = (TIME_TAG_TYPE)tempHighTimeTag;
      timeTagRef->referenceTimeTag = timeTag64LowValue + (timeTag64HighValue << 32);
   
      localTimeTagFormat = (timeTagFormat == AR_USE_PROGRAMMED_TIMETAG_MODE) ? (CEI_INT32)deviceGlobalStructPtr->deviceTimeTagReference : timeTagFormat;
      
      switch (localTimeTagFormat) {
         case AR_TIMETAG_SYNC_1553_CH1:           /* Sync ARINC 429 Message Timestamp to UCA 1553 Channel 1 timer */
         case AR_TIMETAG_SYNC_1553_CH2:           /* Sync ARINC 429 Message Timestamp to UCA 1553 Channel 2 timer */
         case AR_TIMETAG_SYNC_1553_CH3:           /* Sync ARINC 429 Message Timestamp to UCA 1553 Channel 3 timer */
         case AR_TIMETAG_SYNC_1553_CH4:           /* Sync ARINC 429 Message Timestamp to UCA 1553 Channel 4 timer */

            ucaChanTimerLoOffset = CEIDEV_UCA_CH_1_TIME_LSW + (2 * (localTimeTagFormat - AR_TIMETAG_SYNC_1553_CH1));
            ucaChanTimerHiOffset = CEIDEV_UCA_CH_1_TIME_MSW + (2 * (localTimeTagFormat - AR_TIMETAG_SYNC_1553_CH1));

#ifndef TARGET_EMBEDDED
            /* if multi-thread protection is active, acquire a lock on the host interface */
            if ((timeSource == AR_TIMER_REGISTER_SET) && (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE)) {
               status = cei_acquire_thread_lock(LOCK_IDENT);
            }
   
            if (status == ARS_NORMAL) {
#endif
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_TIMER_LOW], &tempLowTimeTag);
               if (accessStatus == ARS_NORMAL) {
                  accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_TIMER_HIGH], &tempHighTimeTag);
                  if (accessStatus == ARS_NORMAL) {
                     timeTag64LowValue = (TIME_TAG_TYPE)tempLowTimeTag;
                     timeTag64HighValue = (TIME_TAG_TYPE)tempHighTimeTag;
                     timeTagRef->timeTag = timeTag64LowValue + (timeTag64HighValue << 32); /* x30 timer value in microseconds */
   
                     /* find the difference between the current x30 timer value and the logged message time-stamp reference */
                     timeTagRef->timeTag -= timeTagRef->referenceTimeTag;
                     /* convert the delta to nanoseconds */
                     timeTagRef->timeTag *= (TIME_TAG_TYPE)1000;
   
                     /* read the locked UCA 1553 Channel Sync time */
                     accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[ucaChanTimerLoOffset], &tempLowTimeTag);
                     if (accessStatus == ARS_NORMAL) {
                        accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[ucaChanTimerHiOffset], &tempHighTimeTag);
                     }
                     if (accessStatus == ARS_NORMAL) {
                        timeTag64LowValue = (TIME_TAG_TYPE)tempLowTimeTag;
                        timeTag64HighValue = (TIME_TAG_TYPE)tempHighTimeTag;
                        timeTag64LowValue += timeTag64HighValue << 32;  /* 64-bit UCA 1553 Channel Sync Time in nanoseconds  */
   
                        /* Calculate the Message Timestamp in 1553 Channel Sync Time terms w/1 nanaosecond resolution */
                        timeTag64LowValue -= timeTagRef->timeTag; 
                     }
                  }
               }
#ifndef TARGET_EMBEDDED
            }
   
            /* if multi-thread protection is active, free any acquired lock */
            if ((timeSource == AR_TIMER_REGISTER_SET) && (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE)) {
               accessStatus = cei_release_thread_lock(LOCK_IDENT);
            }
#endif
            break;

         case AR_TIMETAG_INT_USEC_64BIT:           /* time-tag already in usec 64-bit format */
            timeTag64LowValue += (timeTag64HighValue << 32);
            break;

#ifdef INCLUDE_IRIG
         case AR_TIMETAG_EXT_IRIG_64BIT:
         case AR_LOG_EXT_IRIG_SAMPLE:
#ifndef TARGET_EMBEDDED
            /* if multi-thread protection is active, acquire a lock on the host interface */
            if ((timeSource == AR_TIMER_REGISTER_SET) && (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE)) {
               status = cei_acquire_thread_lock(LOCK_IDENT);
            }
#endif
            if (status == ARS_NORMAL) {
               /* read the current IRIG time and the device time at which this IRIG time was received */
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_IRIG_INPUT_TIME], &appProgTimeTag);
   
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_IRIG_TIME_TAG_LOW], &tempLowTimeTag);
               if (accessStatus == ARS_NORMAL) {
                  accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_IRIG_TIME_TAG_HIGH], &tempHighTimeTag);
               }
   
#ifndef TARGET_EMBEDDED
               /* if multi-thread protection is active, free any acquired lock */
               if ((timeSource == AR_TIMER_REGISTER_SET) && (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE)) {
                  accessStatus = cei_release_thread_lock(LOCK_IDENT);
               }
#endif
               /* convert the IRIG time to number of seconds since 12:00AM of Jan 01 */
               numberOfDays     = ((appProgTimeTag >> IRIG_SHIFT_HUNDREDS_OF_DAYS) & IRIG_MASK_HUNDREDS_OF_DAYS) * 100;
               numberOfDays    += ((appProgTimeTag >> IRIG_SHIFT_TENS_OF_DAYS    ) & IRIG_MASK_TENS_OF_DAYS)     * 10;
               numberOfDays    +=  (appProgTimeTag >> IRIG_SHIFT_ONES_OF_DAYS    ) & IRIG_MASK_ONES_OF_DAYS;
               numberOfMinutes  = ((appProgTimeTag >> IRIG_SHIFT_TENS_OF_HOURS   ) & IRIG_MASK_TENS_OF_HOURS)    * 600;
               numberOfMinutes += ((appProgTimeTag >> IRIG_SHIFT_ONES_OF_HOURS   ) & IRIG_MASK_ONES_OF_HOURS)    * 60;
               numberOfMinutes += ((appProgTimeTag >> IRIG_SHIFT_TENS_OF_MINUTES ) & IRIG_MASK_TENS_OF_MINUTES)  * 10;
               numberOfMinutes +=  (appProgTimeTag >> IRIG_SHIFT_ONES_OF_MINUTES ) & IRIG_MASK_ONES_OF_MINUTES;
               numberOfSeconds  = ((appProgTimeTag >> IRIG_SHIFT_TENS_OF_SECONDS ) & IRIG_MASK_TENS_OF_SECONDS)  * 10;
               numberOfSeconds += appProgTimeTag & IRIG_MASK_ONES_OF_SECONDS;
               
               numberOfSeconds += numberOfMinutes * 60;
               numberOfSeconds += max(numberOfDays-1,0) * AR_SECONDS_PER_DAY; /* 0 days is invalid IRIG */
               deviceGlobalStructPtr->currentIrigTimeUsec = numberOfSeconds;
   
               timeTag64LowValue = (TIME_TAG_TYPE)tempLowTimeTag;
               timeTag64HighValue = (TIME_TAG_TYPE)tempHighTimeTag;
               deviceGlobalStructPtr->irigReferenceTimeTag = timeTag64LowValue + (timeTag64HighValue << 32);
            }
            
            /* if this invocation was only to read the IRIG time sample, exit here */
            if (localTimeTagFormat == AR_LOG_EXT_IRIG_SAMPLE) {
               break;
            }
            
         case AR_TIMETAG_REF_IRIG_64BIT:
            /* Offset the IRIG source time-tag by the difference of the time-tag 
             * reference and the time at which the IRIG update was received.
             *
             * Once we start applying offsets, the actual IRIG conversion could
             * cause the actual resulting time-of-year to exceed the number of
             * microseconds available in a year; therefore, we must force a
             * calendar rollover after applying the offsets.
             */
            irigOffsetTime = timeTagRef->referenceTimeTag - deviceGlobalStructPtr->irigReferenceTimeTag;

            /* now calculate respective IRIG time based on the IRIG reference offset */
            timeTag64LowValue = deviceGlobalStructPtr->currentIrigTimeUsec * AR_MICROSECONDS_PER_SECOND;
            timeTag64LowValue += irigOffsetTime;
            timeTag64LowValue += deviceGlobalStructPtr->userProgrammableIrigOffset + AR_IRIG_TIMETAG_BIAS;

            /* Acquire the current year */
            time(&now);               /* Time in seconds since 12AM Jan 1, 1970 GMT */
            start = *localtime(&now); /* Local time in sec, min, hr, day, mo, yr */
            year = (CEI_UINT32)start.tm_year + AR_LOCAL_TIME_YEAR_BASIS;
         
            /* Determine if this is a leap year for max # of usecs allowed this year */
            if ( ((year % AR_LEAP_YEAR_DIVISOR) == 0) && (year != AR_LEAP_YEAR_EXCEPTION) ) {
               maxMicroSecondsPerYear = AR_IRIG_ROLLOVER_LEAP_YEAR;
            }
            else {
               maxMicroSecondsPerYear = AR_IRIG_ROLLOVER_NON_LEAP_YEAR;
            }
            
            /* handle an IRIG rollover once the offsets are applied */
            if (timeTag64LowValue > maxMicroSecondsPerYear) {
               timeTag64LowValue -= maxMicroSecondsPerYear;
            }
            /*
            timeTagRef->referenceTimeTag = deviceGlobalStructPtr->irigReferenceTimeTag; */
            break;

         case AR_TIMETAG_EXT_IRIG_REF_64BIT:
#ifndef TARGET_EMBEDDED
            /* if multi-thread protection is active, acquire a lock on the host interface */
            if ((timeSource == AR_TIMER_REGISTER_SET) && (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE)) {
               status = cei_acquire_thread_lock(LOCK_IDENT);
            }
            if (status == ARS_NORMAL) {
#endif
               /* read the current IRIG time and the device time at which this IRIG time was received */
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_IRIG_INPUT_TIME], &appProgTimeTag);
   
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_IRIG_TIME_TAG_LOW], &tempLowTimeTag);
               if (accessStatus == ARS_NORMAL) {
                  accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_IRIG_TIME_TAG_HIGH], &tempHighTimeTag);
               }
   
               timeTag64LowValue = (TIME_TAG_TYPE)tempLowTimeTag + ((TIME_TAG_TYPE)tempHighTimeTag << 32);
#ifndef TARGET_EMBEDDED
            }

            /* if multi-thread protection is active, free any acquired lock */
            if ((timeSource == AR_TIMER_REGISTER_SET) && (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE)) {
               accessStatus = cei_release_thread_lock(LOCK_IDENT);
            }
#endif
            break;

         case AR_TIMETAG_EXT_IRIG_RAW:
#ifndef TARGET_EMBEDDED
            /* if multi-thread protection is active, acquire a lock on the host interface */
            if ((timeSource == AR_TIMER_REGISTER_SET) && (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE)) {
               status = cei_acquire_thread_lock(LOCK_IDENT);
            }
            if (status == ARS_NORMAL) {
#endif
               /* read the current IRIG time and the device time at which this IRIG time was received */
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_IRIG_INPUT_TIME], &appProgTimeTag);
   
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_IRIG_TIME_TAG_LOW], &tempLowTimeTag);
               if (accessStatus == ARS_NORMAL) {
                  accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_IRIG_TIME_TAG_HIGH], &tempHighTimeTag);
               }
   
               timeTag64LowValue = (TIME_TAG_TYPE)appProgTimeTag;
               timeTagRef->referenceTimeTag = (TIME_TAG_TYPE)tempLowTimeTag + ((TIME_TAG_TYPE)tempHighTimeTag << 32);
#ifndef TARGET_EMBEDDED
            }

            /* if multi-thread protection is active, free any acquired lock */
            if ((timeSource == AR_TIMER_REGISTER_SET) && (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE)) {
               accessStatus = cei_release_thread_lock(LOCK_IDENT);
            }
#endif
            break;
#endif

#ifndef CERTIFICATION_BUILD
         case AR_TIMETAG_INT_20USEC_32BIT:
            timeTag64LowValue = timeTag64LowValue + (timeTag64HighValue << 32);
            if (tempLowTimeTag%20 > 9) {
               remainderRoundUp = 1;
            }

            timeTag64LowValue /= (TIME_TAG_TYPE)20;
            timeTag64LowValue += (TIME_TAG_TYPE)remainderRoundUp;

            if (mergedMode != CEIDEV_MAX_CHANNELS) {
               timeTag64LowValue &= MERGED_CHANNEL_MASK;
               timeTag64LowValue |= (TIME_TAG_TYPE)localChannel;
            }
            break;

         case AR_TIMETAG_INT_MSEC_32BIT:
            timeTag64LowValue = timeTag64LowValue + (timeTag64HighValue << 32);
            if (tempLowTimeTag%1000 > 499U) {
               remainderRoundUp = 1U;
            }

            timeTag64LowValue /= AR_MICROSEC_PER_MILLISEC;
            timeTag64LowValue += (TIME_TAG_TYPE)remainderRoundUp;

            /*  If Merged Receive Mode is active, replace the upper 3 bits of 
             *  the time-tag with the channel number on which the data was received. */
            if (mergedMode != CEIDEV_MAX_CHANNELS) {
               timeTag64LowValue &= MERGED_CHANNEL_MASK;
               timeTag64LowValue |= (TIME_TAG_TYPE)localChannel;
            }
            break;

         case AR_TIMER_X20_COMPAT_32BIT:
            timeTag64LowValue = timeTag64LowValue + (timeTag64HighValue << 32);
            if (timeTag64LowValue < AR_X20_MAX_TIMETAG_VALUE) {
               timeTag64LowValue /= (TIME_TAG_TYPE)(deviceGlobalStructPtr->deviceX20CompatibleResolution/AR_X20_SCALE_TIMETAG_VALUE);
            }
            
            /*  If Merged Receive Mode is active, replace the upper 3 bits of 
             *  the time-tag with the channel number on which the data was received. */
            if (mergedMode != CEIDEV_MAX_CHANNELS) {
               timeTag64LowValue &= MERGED_CHANNEL_MASK;
               timeTag64LowValue |= (TIME_TAG_TYPE)localChannel;
            }
            break;
#endif

         default:
            break;
      }
      
      /* combine the low and high time-tag values into a single 64-bit time-tag */
      timeTagRef->timeTag = timeTag64LowValue;
      timeTagRef->timeTagFormat = (TIME_TAG_TYPE)localTimeTagFormat;
   }
   if (status != ARS_NORMAL) {
      accessStatus = status;
   }
   return accessStatus;
}


/*===================================================================================================================*
 * ENTRY POINT:         A 4 2 9 _ U T L _ C O N F I G U R E _ P R O G R A M M A B L E _ C H A N N E L
 *===================================================================================================================*
 *
 * FUNCTION:    Configure a programmable channel as either a transmitter or receiver
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board  -- (input) Board number (0 to 127) 
 *              CEI_UINT32 channel     -- (input) programmable channel index (0 to 127)
 *              CEI_UINT32 channelType -- (input) desired configuration for channel
 *                                                ARU_RECEIVER    0
 *                                                ARU_TRANSMITTER 1
 *                                                ARU_BOTH        2
 *                                                ARU_DISABLE     3
 * 
 * RETURN VAL:  CEI_INT32    -- ARS_NORMAL     Success updating or already done
 *                              ARS_INVHARVAL  The channel parameter was an invalid channel, or the channel was not 
 *                                             programmable, or the channelType parameter was neither ARU_RECEIVER 
 *                                             or ARU_TRANSMITTER.
 * GLOBALS:     device
 *              deviceGlobal
 *
 * DESCRIPTION: This function is invoked from ar_set_device_config when the item parameter 
 *              ARU_CONFIG_PROGRAMMABLE_CHAN is supplied.  It configures a programmable channel pair such that the 
 *              I/O pins and respective channel functions as a transmitter or receiver, both transmitter and receiver, 
 *              or both disabled.
 *
 *              Note the channel parameter is not indexed by receive channel... it is indexed by programmable 
 *              channel, typically the transmit channel index.
 *
 *===================================================================================================================*/ 
CEI_INT32 a429_utl_configure_programmable_channel 
(
   CDEV_BOARD_TYPE board,                                                  /* Board number (0 to 127)                */
   CEI_UINT32 channel,                                                     /* programmable channel index (0 to 127)  */
   CEI_UINT32 channelType                                                  /* desired configuration for channel      */
)
{
   CEI_UINT32  channelTypeIndex = 0U;                                      /* local Channel Type index               */
   CEI_UINT32  channelIndex;                                               /* local channel index                    */
   CEI_UINT32  rxMemOffset;                                                /* local offset to channel memory area    */
   CEI_UINT32  txMemOffset;                                                /* local offset to channel memory area    */
   CEI_UINT32  rxBitFieldValue = 0U;                                       /* temp for device register read content  */
   CEI_UINT32  txBitFieldValue = 0U;                                       /* temp for device register read content  */
   CEI_INT32   accessStatus = ARS_NORMAL;                                  /* status of device r/w access            */
   CEI_INT32   status = ARS_NORMAL;                                        /* function return status                 */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                       /* ptr to board local structure           */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                              /* ptr to brd global structure            */
   AR_CHANNEL_INDEX_INFO_TYPE * channelInfoTablePtr = (AR_CHANNEL_INDEX_INFO_TYPE *)&deviceGlobal[board]->boardChannelIndexInfoStruct;
   
   if ( ((channelType == ARU_RECEIVER) && (channel < deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX])) ||
        ((channelType == ARU_TRANSMITTER) && (channel < deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX])) ||
        (( (channelType == ARU_DISABLE) || (channelType == ARU_BOTH)) && 
         ((channel < deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) && (channel < deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX])) )
      ) {
   
      /* assign references to this board's device and deviceGlobal structure-array elements */
      deviceLocalStructPtr = &device[board];
      deviceGlobalStructPtr = deviceGlobal[board];
   
      /* find the first programmable channel on this board */
      while (channelTypeIndex < CEIDEV_MAX_CHANNELS) {
         if (channelInfoTablePtr->chan_type[channelTypeIndex] == AR_CHAN_TYPE_429_RX_OR_TX) {
            break;
         }
         channelTypeIndex++;
      }
   
      /* if this is a valid programmable channel, configure it as either a receiver or transmitter */
      if ((channelTypeIndex < CEIDEV_MAX_CHANNELS) && (channelInfoTablePtr->chan_rx_index[channelTypeIndex + channel] != AR_CHAN_TYPE_UNDEFINED)) {
         
         /* determine the hardware channel indices for the respective receive and transmit channel, calculate offset into host interface as such */
         channelIndex = channelInfoTablePtr->chan_rx_index[channelTypeIndex] + channel + deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
         rxMemOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (channelIndex * CEIDEV_CHANNEL_REGION_SIZE);

         channelIndex = channelInfoTablePtr->chan_tx_index[channelTypeIndex] + channel + deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
         txMemOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (channelIndex * CEIDEV_CHANNEL_REGION_SIZE);

         /* read the current respective channel configuration register values */
         accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[rxMemOffset], &rxBitFieldValue);
         if (accessStatus == ARS_NORMAL) {
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[txMemOffset], &txBitFieldValue);
         }

         /* assign the rx/tx channel enable and tx disable bits for the respective channel configuration register */
         if (accessStatus == ARS_NORMAL) {
            switch (channelType) {
               case ARU_RECEIVER:
                  rxBitFieldValue |= CEIDEV_CHANNEL_ENABLE;
                  txBitFieldValue &= CEIDEV_CHANNEL_ENABLE_MASK;
                  txBitFieldValue |= CEIDEV_TRANSMIT_DISABLE;
                  break;
               case ARU_TRANSMITTER:
                  rxBitFieldValue &= CEIDEV_CHANNEL_ENABLE_MASK;
                  txBitFieldValue &= CEIDEV_TRANSMIT_DISABLE_MASK;
                  txBitFieldValue |= CEIDEV_CHANNEL_ENABLE;
                  break;
               case ARU_BOTH:
                  rxBitFieldValue |= CEIDEV_CHANNEL_ENABLE;
                  txBitFieldValue &= CEIDEV_TRANSMIT_DISABLE_MASK;
                  txBitFieldValue |= CEIDEV_CHANNEL_ENABLE;
                  break;
               default:
               case ARU_DISABLE:
                  rxBitFieldValue &= CEIDEV_CHANNEL_ENABLE_MASK;
                  txBitFieldValue &= CEIDEV_CHANNEL_ENABLE_MASK;
                  txBitFieldValue |= CEIDEV_TRANSMIT_DISABLE;
                  break;
            }
            accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[rxMemOffset], rxBitFieldValue);
            if (accessStatus == ARS_NORMAL) {
               accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[txMemOffset], txBitFieldValue);
            }
         }
      }
      else {
         status = ARS_INVHARVAL;
      }
   }
   else {
      status = ARS_INVHARVAL;
   }
   if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL)) {
      status = (CEI_INT32)accessStatus;
   }
   
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:          A 4 2 9 _ U T L _ D E T E R M I N E _ C H A N N E L _ S U P P O R T
 *===================================================================================================================*
 *
 * FUNCTION:    Determine the channel type for each channel available on the specified device
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input) Board number (0 to 127) 
 * 
 * RETURN VAL:  N/A
 * 
 * GLOBALS:     device
 *              deviceGlobal
 *
 * DESCRIPTION: This function determines the number of channels installed on the board for each channel/protocol 
 *              type detected.
 *
 *===================================================================================================================*/ 
CEI_VOID  a429_utl_determine_channel_support
(
   CDEV_BOARD_TYPE board                                                          /* board to reference (0 to 127)   */
)
{
   CEI_INT32   accessStatus = ARS_NORMAL;                                         /* status of device r/w access     */
   CEI_INT32   channelTypeIndex = 0;                                              /* local Channel Type index        */
   CEI_UINT32  channelIndex;                                                      /* local channel index             */
   CEI_UINT32  genIndex = 0U;                                                     /* local Channel Type index        */
   CEI_UINT32  channelType;                                                       /* channel type for each channel   */
   CEI_UINT32  channelInactive;                                                   /* local inactive channel flag     */
   CEI_UINT32  channelOffset;                                                     /* channel region offset in HIFC   */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                              /* ptr to board local structure    */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                                     /* ptr to brd global structure     */
   AR_CHANNEL_INDEX_INFO_TYPE * channelInfoTablePtr = (AR_CHANNEL_INDEX_INFO_TYPE *)&deviceGlobal[board]->boardChannelIndexInfoStruct;
   
   /* assign references to this board's device and deviceGlobal structure-array elements */
   deviceLocalStructPtr = &device[board];
   deviceGlobalStructPtr = deviceGlobal[board];

   /* initialize all of the channel information for this board to be null */
   deviceGlobalStructPtr->alternateJconfiguration = 0U;
   channelInfoTablePtr->num_channels = 0U;
   for (channelIndex = 0; channelIndex < CEIDEV_MAX_CHANNELS; channelIndex++) {
      channelInfoTablePtr->chan_type[channelIndex] = AR_CHAN_TYPE_UNDEFINED;
      channelInfoTablePtr->chan_rx_index[channelIndex] = AR_CHAN_TYPE_UNDEFINED;
      channelInfoTablePtr->chan_tx_index[channelIndex] = AR_CHAN_TYPE_UNDEFINED;
   }

   /* initialize the channel counters and h/w channel mapping arrays to their "uninitialized" state */
   for (channelIndex = 0; channelIndex < NUMBER_OF_CHANNEL_TYPES; channelIndex++) {
      deviceGlobalStructPtr->numberOfChannels[channelIndex] = 0U;
      deviceGlobalStructPtr->firstChannel[channelIndex] = CEIDEV_MAX_CHANNELS;  
      deviceGlobalStructPtr->lastChannel[channelIndex] = CEIDEV_MAX_CHANNELS;
   }

   deviceGlobalStructPtr->numTx = 0U;
   deviceGlobalStructPtr->numRx = 0U;

   /* identify each receive channel and setup API local channel validation information */
   channelOffset = CEIDEV_CHANNEL_REGION_START;
   for (channelIndex = 0; channelIndex < CEIDEV_MAX_CHANNELS; channelIndex++) {
      
      /* retrieve the channel type (and protocol) for this channel index */
      accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &channelType);
      channelInactive = (channelType & CEIDEV_CSR_CHAN_INACTIVE_FLAG);
      channelType = (channelType & CEIDEV_CSR_CHAN_TYPE_MASK) >> CEIDEV_CSR_CHAN_TYPE_SHIFT;

      /* identify the receive channels and tally the protocol channel count for active channels only */
      if ((channelType != CEIDEV_CH_TYPE_NONE) && (!channelInactive)) {

         switch (channelType) {

            /* identify the serial receive channels and tally total device receive channel count */
            /* then fall through to next case to tally individual protocol channel counts        */
#ifndef CERTIFICATION_BUILD
            case CEIDEV_CH_TYPE_ARINC573_RX:
#endif
            case CEIDEV_CH_TYPE_ARINC429_RX:
               /* populate the channel configuration index table, added for programmable channels */
               if (channelType == CEIDEV_CH_TYPE_ARINC429_RX) {
                  channelInfoTablePtr->num_channels++;
                  channelInfoTablePtr->chan_rx_index[channelTypeIndex] = deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX];
                  channelInfoTablePtr->chan_type[channelTypeIndex++] = AR_CHAN_TYPE_429_RX;
               }
               else {
                  deviceGlobalStructPtr->alternateJconfiguration = CEIDEV_ALT_CONFIGURATION_717;
               }

#ifndef CERTIFICATION_BUILD
            case CEIDEV_CH_TYPE_ARINC561_RX:
            case CEIDEV_CH_TYPE_SERIAL_RX:
#endif
               deviceGlobalStructPtr->numRx++;

            /* identify state-type channels and tally only individual discrete/digital channel count */
#ifdef INCLUDE_DISCRETE_IO
            case CEIDEV_CH_TYPE_DISC_IN:
            case CEIDEV_CH_TYPE_DIGL_IN:
            case CEIDEV_CH_TYPE_DIFF_IN:
#endif
               deviceGlobalStructPtr->numberOfChannels[channelType]++; 
               if (deviceGlobalStructPtr->firstChannel[channelType] == CEIDEV_MAX_CHANNELS) {
                  deviceGlobalStructPtr->firstChannel[channelType] = channelIndex;  
                  deviceGlobalStructPtr->lastChannel[channelType] = channelIndex;  
               }
               else {
                  deviceGlobalStructPtr->lastChannel[channelType]++;  
               }
               break;
   
#ifndef CERTIFICATION_BUILD
            case CEIDEV_CH_TYPE_MERGED_RX:
               deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_MERGED_RX] = channelIndex;
               break;

            /* identify the serial transmit channels and tally total device transmit channel count */
            /* then fall through to next case to tally individual protocol channel counts      */
            case CEIDEV_CH_TYPE_ARINC573_TX:
#endif
            case CEIDEV_CH_TYPE_ARINC429_TX:
               /* populate the channel configuration index table, added for programmable channels */
               if (channelType == CEIDEV_CH_TYPE_ARINC429_TX) {
                  channelInfoTablePtr->num_channels++;
                  channelInfoTablePtr->chan_tx_index[channelTypeIndex] = deviceGlobalStructPtr->numberOfChannels[channelType];
                  channelInfoTablePtr->chan_type[channelTypeIndex++] = AR_CHAN_TYPE_429_TX;
               }
               else {
                  deviceGlobalStructPtr->alternateJconfiguration = CEIDEV_ALT_CONFIGURATION_717;
               }

#ifndef CERTIFICATION_BUILD
            case CEIDEV_CH_TYPE_ARINC561_TX:
            case CEIDEV_CH_TYPE_SERIAL_TX:
#endif
               deviceGlobalStructPtr->numTx++;
               
#ifdef INCLUDE_DISCRETE_IO
            /* identify state-type channels and tally only individual discrete/digital channel count */
            case CEIDEV_CH_TYPE_DISC_OUT:
            case CEIDEV_CH_TYPE_DIGL_OUT:
            case CEIDEV_CH_TYPE_DIFF_OUT:
#endif
               deviceGlobalStructPtr->numberOfChannels[channelType]++; 
               if (deviceGlobalStructPtr->firstChannel[channelType] == CEIDEV_MAX_CHANNELS) {
                  deviceGlobalStructPtr->firstChannel[channelType] = channelIndex;  
                  deviceGlobalStructPtr->lastChannel[channelType] = channelIndex;  
               }
               else {
                  deviceGlobalStructPtr->lastChannel[channelType]++;  
               }
               break;
      
            /* identify the ARINC 429 programmable receive channels and tally total device receive channel count */
            case CEIDEV_CH_TYPE_PROG_ARINC429_RX:
               for (genIndex = 0; genIndex < (channelInfoTablePtr->num_channels+1); genIndex++) {
                  if ( (channelInfoTablePtr->chan_type[genIndex] == AR_CHAN_TYPE_UNDEFINED) ||
                       ( (channelInfoTablePtr->chan_type[genIndex] == AR_CHAN_TYPE_429_RX_OR_TX) &&
                         (channelInfoTablePtr->chan_rx_index[genIndex] == AR_CHAN_TYPE_UNDEFINED) )
                     ) {
                     if (channelInfoTablePtr->chan_type[genIndex] == AR_CHAN_TYPE_UNDEFINED) {
                        channelInfoTablePtr->num_channels++;
                        channelInfoTablePtr->chan_type[genIndex] = AR_CHAN_TYPE_429_RX_OR_TX;
                        channelTypeIndex++;
                     }
                     channelInfoTablePtr->chan_rx_index[genIndex] = deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX];
                     break;
                  }
               }
                  
               deviceGlobalStructPtr->numRx++;
               deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]++; 
               if (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] == CEIDEV_MAX_CHANNELS) {
                  deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] = channelIndex;  
                  deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX] = channelIndex;  
               }
               else {
                  deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]++;  
               }
               break;

            /* identify the ARINC 429 programmable transmit channels and tally total device transmit channel count */
            case CEIDEV_CH_TYPE_PROG_ARINC429_TX:
               for (genIndex = 0; genIndex < (channelInfoTablePtr->num_channels+1); genIndex++) {
                  if ( (channelInfoTablePtr->chan_type[genIndex] == AR_CHAN_TYPE_UNDEFINED) ||
                       ( (channelInfoTablePtr->chan_type[genIndex] == AR_CHAN_TYPE_429_RX_OR_TX) &&
                         (channelInfoTablePtr->chan_tx_index[genIndex] == AR_CHAN_TYPE_UNDEFINED) )
                     ) {
                     if (channelInfoTablePtr->chan_type[genIndex] == AR_CHAN_TYPE_UNDEFINED) {
                        channelInfoTablePtr->num_channels++;
                        channelInfoTablePtr->chan_type[genIndex] = AR_CHAN_TYPE_429_RX_OR_TX;
                        channelTypeIndex++;
                     }
                     channelInfoTablePtr->chan_tx_index[genIndex] = deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX];
                     break;
                  }
               }

               deviceGlobalStructPtr->numTx++;
               deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX]++; 
               if (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX] == CEIDEV_MAX_CHANNELS) {
                  deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX] = channelIndex;  
                  deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_TX] = channelIndex;  
               }
               else {
                  deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_TX]++;  
               }
               break;

            default:
               break;
         }
      }
      /* increment to the next channel memory block */
      channelOffset += CEIDEV_CHANNEL_REGION_SIZE;
   }

   if (accessStatus) ;
#ifdef _DEBUG

   printf("\n\n");
#endif
}


/*===================================================================================================================*
 * ENTRY POINT:            A 4 2 9 _ U T L _ I N I T I A L I Z E _ G L O B A L _ D A T A
 *===================================================================================================================*
 *
 * FUNCTION:    Function that initializes most global data
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input) board to reference (0 to 127)
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     device
 *
 * DESCRIPTION: This routine is used to initialize global data structures and arrays shared between the API and the 
 *              operating system specifically-sized arrays based on MAX_BOARDS.
 *
 *===================================================================================================================*/ 
CEI_VOID  a429_utl_initialize_global_data 
(
   CDEV_BOARD_TYPE board                                                        /* board to reference (0 to 127)     */
)
{
#ifdef INCLUDE_ARINC_717
   a717_utl_initialize_global_data (board);
#endif
#ifdef INCLUDE_INTERRUPTS
   a429_utl_initialize_int_queue (board, TRUE);
#endif
#ifdef INCLUDE_IRIG
   device[board].irigDacThreshold = CEIDEV_IRIG_DAC_INIT_THRESHOLD;
#endif
}


/*===================================================================================================================*
 * ENTRY POINT:                 A 4 2 9 _ U T L _ I N I T I A L I Z E _ D E V I C E
 *===================================================================================================================*
 *
 * FUNCTION:    Initialize the device to a ready state
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input) board to reference (0 to 127)
 *              CEI_UINT32 powerOnState -- (input) state of calling function
 *                                                 TRUE  - power on initialization
 *                                                 FALSE - request to reset device
 * 
 * RETURN VAL:  CDEV_API_RET_TYPE -- ARS_NORMAL       Success
 *                                   ARS_BAD_STATIC   Register read/write/verify failure
 *                                   ARS_MEMWRERR     Memory test failure
 *                                   ARS_INVBOARD     Invalid board parameter
 *                                   ARS_BOARD_MUTEX  Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     device
 *              deviceGlobal
 *
 * DESCRIPTION: This routine initializes the device I/O configuration registers to a state ready for use by the host 
 *              application.
 *
 *===================================================================================================================*/ 
CDEV_API_RET_TYPE a429_utl_initialize_device
(
   CDEV_BOARD_TYPE  board,                                            /* board to reference (0 to 127)               */
   CEI_UINT32       powerOnState                                      /* state of calling function                   */
)
{
   CEI_UINT32  labelIndex;                                            /* local label filter table index              */
#ifdef INCLUDE_IRIG                                                           
   CEI_UINT32  initialIrigTime;                                       /* initial IRIG BCD time value                 */
#endif                                                                        
   CEI_UINT32  bitFieldValue;                                         /* temp var for device register data           */
   CEI_UINT32  configRegValue;                                        /* temp var for device register data           */
   CEI_UINT32  deviceChannel;                                         /* temp var for device register data           */
   CEI_UINT32  transmitChannelCount;                                  /* temp var for device register data           */
   CEI_UINT32  memOffset;                                             /* local offset to channel memory area         */
   CEI_UINT32  channelOffset;                                         /* chan reg region offset in host interface    */
   CEI_UINT32  bufferOffset;                                          /* chan buffer region offset in host interface */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                             /* function return status                      */
   CEI_INT32   accessStatus = ARS_NORMAL;                             /* status of device r/w access                 */
#ifndef TARGET_EMBEDDED
   CEI_INT32   tempStatus = ARS_NORMAL;                               /* local temp return status                    */
#endif
   CEI_UINT32  initProgress = INIT_DISABLE;                           /* init state of progression                   */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                  /* pointer to board local structure            */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                         /* pointer to board global structure           */
   pCEI_UINT32 sramPtr;                                               /* local ptr to board sram                     */

#ifndef TARGET_EMBEDDED
   if ( (board < MAX_BOARDS) && (deviceGlobal[board] != NULL) && (deviceGlobal[board]->attachedProcessCount > 0) ) {
#endif
      /* assign references to this board's device and deviceGlobal structure-array elements */
      deviceLocalStructPtr = &device[board];
      deviceGlobalStructPtr = deviceGlobal[board];

#ifndef TARGET_EMBEDDED
      /* if multi-thread protection requested, acquire semaphore/mutex-locking */
      if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
         accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
      }
#endif

      /* Execute the initialization sequence for the board */
      while ((accessStatus == ARS_NORMAL) && (status == ARS_NORMAL) && (initProgress != INIT_TERMINATE)) {
         switch (initProgress) {
            case INIT_DISABLE:
               /*--------------------------------------------------------------------*
                *  Disable all data processing
                *--------------------------------------------------------------------*/
               bitFieldValue = 0U; /* clear IRIG Output Enable, Snapshot Mode, Interrupt Enable, IRIG Wrap & Enable */
               accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
               /* for AR_RESET invocations, wait 20 milliseconds to assure application inactivity  */
               if (!powerOnState) {
#ifndef TARGET_EMBEDDED
                  accessStatus = cei_delay_milliseconds(20U);
#endif
               }
               if (deviceGlobalStructPtr->configuration == (CEI_UINT32)CEIDEV_CONFIG_RAR15XT) {
                  accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_RAR15_429_RX_BIT_CMD], 0U);
               }
               initProgress = INIT_TEST_SRAM;
               break;

            case INIT_TEST_SRAM:
               /*--------------------------------------------------------------------*
                *  Execute a quick memory test on a few unused SRAM locations
                *--------------------------------------------------------------------*/
               status = a429_utl_memory_test(board, UNUSED_SRAM, 0);
#ifdef _DEBUG
               if (status != ARS_NORMAL) {
                  printf("\nSRAM Memory Test Status: %d\n", status);
                  status = ARS_NORMAL;
               }
#endif
#ifdef INCLUDE_MSG_SCHEDULER
               initProgress = (deviceGlobalStructPtr->configuration == (CEI_UINT32)CEIDEV_CONFIG_R830RX) ? INIT_CLEAR_SNAPSHOT : INIT_CLEAR_SCHED_TABLE;
               break;

            case INIT_CLEAR_SCHED_TABLE:
               /*--------------------------------------------------------------------*
                * Initialize the message scheduler table to zero
                *--------------------------------------------------------------------*/
               /* check for availability of the expanded message scheduler table, if present read/write verify of entry 1024 rate works */
               accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_MSG_SCHED_TABLE_END], CEIDEV_MSG_SCHED_TBL_EXP_CHECK_PAT);
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_MSG_SCHED_TABLE_END], &configRegValue);
               if (configRegValue == CEIDEV_MSG_SCHED_TBL_EXP_CHECK_PAT) {
                  accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_MSG_SCHED_TABLE_END], 0U);
                  deviceGlobalStructPtr->messageSchedulerTableEnd = CEIDEV_MSG_SCHED_TABLE_EXT_END;
                  deviceGlobalStructPtr->messageSchedulerTableSize = CEIDEV_MSG_SCHED_TBL_ENTRY_EXP_MAX;
               }
               bufferOffset = deviceGlobalStructPtr->messageSchedulerTableEnd;
               
               for (labelIndex = CEIDEV_MSG_SCHED_TABLE_START; 
                    (labelIndex < bufferOffset) && (accessStatus == ARS_NORMAL) && (status == ARS_NORMAL); 
                    labelIndex++ ) {
                  accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[labelIndex], 0U);
                  if (accessStatus == ARS_NORMAL) {
#if defined (INTEGRITY_PPC_SYNC)
                     #pragma asm
                     eieio
                     sync
                     #pragma endasm
#elif defined (PPC_SYNC)
                     IO_SYNC;
#endif
                     /* Test for access conflict with the message scheduler table */
                     accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[labelIndex], &configRegValue);
                     if (configRegValue != 0U) {
                        status = ARS_BAD_STATIC;
                     }
                  }
               }
#endif               
               initProgress = INIT_CLEAR_SNAPSHOT;
               break;

            case INIT_CLEAR_SNAPSHOT:
               /*--------------------------------------------------------------------*
                * Initialize the snapshot buffer to zero
                *--------------------------------------------------------------------*/
               for (labelIndex = CEIDEV_SNAPSHOT_BUFFER_START; 
                    (labelIndex < CEIDEV_SNAPSHOT_BUFFER_END) && (accessStatus == ARS_NORMAL) && (status == ARS_NORMAL); 
                    labelIndex++ ) {
                  accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[labelIndex], (CEI_UINT32)0);
                  if (accessStatus == ARS_NORMAL) {
#if defined (INTEGRITY_PPC_SYNC)
                     #pragma asm
                     eieio
                     sync
                     #pragma endasm
#elif defined (PPC_SYNC)
                     IO_SYNC;
#endif
                     /* Assure the snapshot buffer entry was actually reset to zero */
                     accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[labelIndex], &configRegValue);
                     if (configRegValue != 0U) {
                        status = ARS_BAD_STATIC;
                     }
                  }
               }
#ifdef _DEBUG
               if (status != ARS_NORMAL) {
                  printf("\nInitialize the snapshot buffer Failed, Status: %d\n", status);
                  status = ARS_NORMAL;
               }
#endif
               initProgress = INIT_RESET_TIMER;
               break;

            case INIT_RESET_TIMER:
               /*----------------------------------------------------------------------*
                *  Initialize local timer reference to 64-bit integer
                *  x20 compatibility to a 5ms resolution
                *  host-programmable IRIG offset to zero
                *----------------------------------------------------------------------*/
               deviceGlobalStructPtr->deviceTimeTagReference = AR_TIMETAG_INT_USEC_64BIT;
               deviceGlobalStructPtr->deviceX20CompatibleResolution = AR_X20_INITIAL_RESOLUTION;
               deviceGlobalStructPtr->userProgrammableIrigOffset = (TIME_TAG_TYPE)0;
               deviceGlobalStructPtr->receiveMode = ARU_BUFFERED;
               deviceGlobalStructPtr->transmitMode = ARU_XMIT_573_BPRZ;
               initProgress = INIT_CONFIG_RX;
               break;

            case INIT_CONFIG_RX:
               if (deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX] > 0U) {
                  /*--------------------------------------------------------------------*
                   *  Initialize the ARINC 429 receiver configuration registers
                   *    Bus Speed of 100Kbps with Fast Slew Rate
                   *    Odd Parity
                   *    Buffered Rx mode
                   *    Internal Wrap OFF
                   *    Receiver FIFO Disabled
                   *--------------------------------------------------------------------*/
                  channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] * CEIDEV_CHANNEL_REGION_SIZE);
                  if ((deviceGlobalStructPtr->configuration == (CEI_UINT32)CEIDEV_CONFIG_RAR15) || (deviceGlobalStructPtr->configuration == (CEI_UINT32)CEIDEV_CONFIG_RAR15XT)) {
                     configRegValue = CEIDEV_20MHZ_RECEIVE_INIT;
                  }
                  else {
                     configRegValue = CEIDEV_RECEIVE_INIT;
                  }
                  for (deviceChannel = deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]; 
                       (deviceChannel <= deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]) && (accessStatus == ARS_NORMAL) && (status == ARS_NORMAL); 
                       deviceChannel++) {
                     deviceGlobalStructPtr->receiveChannelMessageCount[deviceChannel] = 0U;
                     deviceGlobalStructPtr->receiveChannelBufferMode[deviceChannel] = ARU_BUFFERED;

                     accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], configRegValue);
                     if (accessStatus == ARS_NORMAL) {
#if defined (INTEGRITY_PPC_SYNC)
                        #pragma asm
                        eieio
                        sync
                        #pragma endasm
#elif defined (PPC_SYNC)
                        IO_SYNC;
#endif
                        /* Test for access conflict with register set */
                        accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &bitFieldValue);
                        if (configRegValue != bitFieldValue) {
                           status = ARS_BAD_STATIC;
                        }
                        channelOffset += CEIDEV_CHANNEL_REGION_SIZE;
                     }
                  }
               }
               /* For ROM-based F/W PCIe ARINC boards, flush transmit buffers. */
               if ( (deviceGlobalStructPtr->configuration == (CEI_UINT32)CEIDEV_CONFIG_RAR_PCIE)  ||
                    (deviceGlobalStructPtr->configuration == (CEI_UINT32)CEIDEV_CONFIG_RAR_MPCIE) ||
                    (deviceGlobalStructPtr->configuration == (CEI_UINT32)CEIDEV_CONFIG_RAR_XMC)   ||
                    (deviceGlobalStructPtr->configuration == (CEI_UINT32)CEIDEV_CONFIG_RAR15)     ||
                    (deviceGlobalStructPtr->configuration == (CEI_UINT32)CEIDEV_CONFIG_RAR15XT) ) {
                  initProgress = INIT_FLUSH_TX;
               }
               else {
                  initProgress = INIT_CONFIG_TX;
               }
               break;

            case INIT_FLUSH_TX:
               /*--------------------------------------------------------------------*
                *  Since the ROM-based F/W PCIe ARINC boards retain transmitter
                *  channel buffer content, flush all transmit buffers here.
                *--------------------------------------------------------------------*/
               configRegValue = CEIDEV_FAST_SLEW_RATE + CEIDEV_TRANSMIT_DISABLE + CEIDEV_CHANNEL_ENABLE + ((CEI_UINT32)8U << CEIDEV_BAUD_RATE_SHIFT);
               channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX] * CEIDEV_CHANNEL_REGION_SIZE);
               for (deviceChannel = deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX]; 
                    deviceChannel <= deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_TX]; 
                    deviceChannel++) {
                  accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &bitFieldValue);
                  bitFieldValue = (bitFieldValue & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT;
                  if (bitFieldValue > (CEI_UINT32)0U) {
                     memOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (deviceChannel * CEIDEV_CHANNEL_REGION_SIZE);
                     accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], configRegValue);
                     if (accessStatus == ARS_NORMAL) {
#if defined (INTEGRITY_PPC_SYNC)
                        #pragma asm
                        eieio
                        sync
                        #pragma endasm
#elif defined (PPC_SYNC)
                        IO_SYNC;
#endif
                        accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], CEIDEV_GLOBAL_ENABLE_AUTON_SET + CEIDEV_GLOBAL_ENABLE_ON);
                        while ((bitFieldValue > (CEI_UINT32)0U) && (accessStatus == ARS_NORMAL)) {
                           accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &bitFieldValue);
                           bitFieldValue = (bitFieldValue & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT;
                        }
                        accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], CEIDEV_GLOBAL_ENABLE_AUTON_CLEAR + CEIDEV_GLOBAL_ENABLE_ON);
                     }
                  }
                  channelOffset += CEIDEV_CHANNEL_REGION_SIZE;
               }
#ifndef TARGET_EMBEDDED
               /* for AR_RESET invocations, wait 10 milliseconds for transmit bus inactivity  */
               if ((accessStatus == ARS_NORMAL) && (powerOnState == FALSE)) {
                  accessStatus = cei_delay_milliseconds(10U);
               }
#endif
               initProgress = INIT_CONFIG_TX;
               break;

            case INIT_CONFIG_TX:
               /*--------------------------------------------------------------------*
                *  Initialize the ARINC 429 transmitter configuration registers
                *    Transmitter FIFO Enabled, Bus Speed of 100Kbps, Odd Parity
                *--------------------------------------------------------------------*/
               if ((deviceGlobalStructPtr->configuration == (CEI_UINT32)CEIDEV_CONFIG_RAR15XT) ||
                   (deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX] > 0U)) {
                  channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX] * CEIDEV_CHANNEL_REGION_SIZE);
         
                  /* Assign the transmitter as enabled unless the board configuration supports s/w programmable channels,
                   * in which case all transmitters should be initialized to the tri-state setting
                   */
                  if (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15XT) {
                     bitFieldValue = CEIDEV_20MHZ_TRANSMIT_INIT + CEIDEV_CHANNEL_ENABLE + CEIDEV_TRANSMIT_DISABLE;
                  }
                  else {
                     bitFieldValue = CEIDEV_TRANSMIT_INIT + CEIDEV_CHANNEL_ENABLE;
                  }
                  /* in order to bail out f/w, disable all extraneous RAR15-XMC-XT transmitters present but not counted in channel config */
                  transmitChannelCount = (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15XT) ? 
                     deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX] + CEIDEV_RAR15_XT_MAX_TX_CHAN :   
                     deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_TX];
                  
                  for (deviceChannel = deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX]; 
                       (deviceChannel <= transmitChannelCount) && (status == ARS_NORMAL); 
                       deviceChannel++) {
                     accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], bitFieldValue);
                     if (accessStatus == ARS_NORMAL) {
#if defined (INTEGRITY_PPC_SYNC)
                        #pragma asm
                        eieio
                        sync
                        #pragma endasm
#elif defined (PPC_SYNC)
                        IO_SYNC;
#endif
                        /* Test for access conflict with register set */
                        accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &configRegValue);
                        if (configRegValue != bitFieldValue) {
                           status = ARS_BAD_STATIC;
                        }
                     }
                     channelOffset += CEIDEV_CHANNEL_REGION_SIZE;
                  }
               }
               initProgress = (deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX] > 0U) ? INIT_CLEAR_RX_FILTER : INIT_717;
               break;

            case INIT_CLEAR_RX_FILTER:
               /*---------------------------------------------------------------------------*
                * disable ARINC 429 enhanced label filtering, setting each table entry to 0 
                *---------------------------------------------------------------------------*/
               if (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15XT) {
                  sramPtr = &deviceLocalStructPtr->devicePtr[CEIDEV_RAR15_429_SRAM_OFFSET];
                  for (labelIndex = 0; 
                       (labelIndex < (deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX] * CEIDEV_ENH_LABEL_FILTER_CHAN_SIZE)) && (accessStatus == ARS_NORMAL); 
                       labelIndex++ ) {
                     accessStatus = cei_write_32bit (board, &sramPtr[labelIndex], 0U);
                  }
               }
               else {
                  channelOffset = CEIDEV_ENH_LABEL_FILTER_START + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] * CEIDEV_ENH_LABEL_FILTER_CHAN_SIZE);
                  memOffset = channelOffset + (deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX] * CEIDEV_ENH_LABEL_FILTER_CHAN_SIZE);
                  for (labelIndex = channelOffset; (labelIndex < memOffset) && (accessStatus == ARS_NORMAL); labelIndex++ ) {
                     /* load the External Memory Address Register with the Label Filter Table SRAM long-word address  */
                     accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_ACCESS_SRAM_ADDR], labelIndex);
                     if (accessStatus == ARS_NORMAL) {
#if defined (INTEGRITY_PPC_SYNC)
                        #pragma asm
                        eieio
                        sync
                        #pragma endasm
#elif defined (PPC_SYNC)
                     IO_SYNC;
#endif
                        /* write the filter table entry data pattern to this location */
                        accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_ACCESS_SRAM_DATA], 0U);
                     }
                  }
               }
#ifdef INCLUDE_ARINC_717
               initProgress = INIT_717;
               break;

            case INIT_717:
               /*--------------------------------------------------------------------*
                *  Initialize ARINC 573/717 receive and transmit channels
                *  Rx & Tx set to 768bps-64 wd subframe, FIFO Disabled, BPRZ encoding
                *  Receive API and h/w Sync Words set to 0
                *  Receive set to internal wrap OFF, individual buffered receive mode
                *     and auto-detect enabled after the FIFO is flushed
                *  Transmit set to External transmiter disabled, low slew rate
                *--------------------------------------------------------------------*/
               if ( deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC573_TX] > 0U ) {
                  bitFieldValue = CEIDEV_573_FRAME_768_64 + CEIDEV_573_TX_BPRZ_SELECT + CEIDEV_TRANSMIT_DISABLE;
                  channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX] * CEIDEV_CHANNEL_REGION_SIZE);
                  accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], bitFieldValue);
               }

               if ( deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC573_RX] > 0U ) {
                  deviceGlobalStructPtr->receiveChannelBufferMode[deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX]] = ARU_BUFFERED;
                  
                  a717_utl_initialize_global_data (board);
                  
                  bitFieldValue = CEIDEV_573_FRAME_768_64 + CEIDEV_573_RX_BPRZ_SELECT + CEIDEV_573_RX_MODE_RAW;
                  channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX] * CEIDEV_CHANNEL_REGION_SIZE);
                  accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], bitFieldValue);
                  if (accessStatus == ARS_NORMAL) {
#if defined (INTEGRITY_PPC_SYNC)
                     #pragma asm
                     eieio
                     sync
                     #pragma endasm
#elif defined (PPC_SYNC)
                     IO_SYNC;
#endif
                     channelOffset++;
                     accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], 0U);
                  }
                  if (accessStatus == ARS_NORMAL) {
#if defined (INTEGRITY_PPC_SYNC)
                     #pragma asm
                     eieio
                     sync
                     #pragma endasm
#elif defined (PPC_SYNC)
                     IO_SYNC;
#endif
                     channelOffset++;
                     accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], 0U);
                  }
                  /* flush the 717 receiver */
                  channelOffset = CEIDEV_CHANNEL_REGION_START + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX] * CEIDEV_CHANNEL_REGION_SIZE);
                  bufferOffset = channelOffset + CEIDEV_CHANNEL_BUFFER_3;
                  /* clear any outstanding error status bits */
                  bitFieldValue = CEIDEV_CSR_RECV_FIFO_NOT_EMPTY + CEIDEV_CSR_RECV_FIFO_INVALID_MSG;
                  accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], bitFieldValue);
                  if (accessStatus == ARS_NORMAL) {
#if defined (INTEGRITY_PPC_SYNC)
                     #pragma asm
                     eieio
                     sync
                     #pragma endasm
#elif defined (PPC_SYNC)
                     IO_SYNC;
#endif
                     while ((bitFieldValue & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) && (accessStatus == ARS_NORMAL)) {
                        accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[bufferOffset], &labelIndex);
                        if (accessStatus == ARS_NORMAL) {
#if defined (INTEGRITY_PPC_SYNC)
                           #pragma asm
                           eieio
                           sync
                           #pragma endasm
#elif defined (PPC_SYNC)
                           IO_SYNC;
#endif
                           accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &bitFieldValue);
                        }
                     }
                  }
                  /* now enable auto-sync */
                  bitFieldValue = CEIDEV_573_FRAME_768_64 + CEIDEV_573_RX_BPRZ_SELECT;
                  channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX] * CEIDEV_CHANNEL_REGION_SIZE);
                  accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], bitFieldValue);
               }
         
#endif
               initProgress = INIT_ENABLE_FLUSH_RX;
               break;

            case INIT_ENABLE_FLUSH_RX:
               /*------------------------------------------------------------------------------------------*
                *  Enable data processing and flush all of the 429 receive FIFOs
                *  Initialize ARINC 429/561 enhanced receiver snapshot storage to be on a label-only basis
                *------------------------------------------------------------------------------------------*/
               bitFieldValue = CEIDEV_GLOBAL_ENABLE_AUTON_SET + CEIDEV_GLOBAL_ENABLE_ON;
               accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
               
               /*--------------------------------------------------------------------*
                *  Enable and Flush the ARINC 429 receivers
                *--------------------------------------------------------------------*/
               channelOffset = CEIDEV_CHANNEL_REGION_START + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] * CEIDEV_CHANNEL_REGION_SIZE);
               memOffset = channelOffset + CEIDEV_CHANNEL_CONFIG_1;
               bufferOffset = channelOffset + CEIDEV_CHANNEL_BUFFER_3;
               if ((deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15) ||
                   (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15XT)) {
                  configRegValue = CEIDEV_CHANNEL_ENABLE + CEIDEV_20MHZ_RECEIVE_INIT;
               }
               else {
                  configRegValue = CEIDEV_CHANNEL_ENABLE + CEIDEV_RECEIVE_INIT;
               }
               if (deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX] > 0U) {
                  for (deviceChannel = deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]; 
                       (deviceChannel <= deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]) && (accessStatus == ARS_NORMAL); 
                       deviceChannel++) {
                     accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], configRegValue);  /* write the config register - enabling reception */
                     /* If the "receive buffer flushed and reset on enable" feature is implemented in this version firmware, 
                      * bypass the status register status bit reset operation and the manual buffer flushing sequence. 
                      */
                     if ((accessStatus == ARS_NORMAL) && (deviceGlobalStructPtr->firmwareVersion < CEI_RX_CLEAR_ON_ENABLE_IN_FW)) {
#if defined (INTEGRITY_PPC_SYNC)
                        #pragma asm
                        eieio
                        sync
                        #pragma endasm
#elif defined (PPC_SYNC)
                        IO_SYNC;
#endif
                        /* clear any outstanding error status bits */
                        bitFieldValue = CEIDEV_CSR_RECV_FIFO_NOT_EMPTY + CEIDEV_CSR_RECV_FIFO_INVALID_MSG + CEIDEV_CSR_RECV_FIFO_OVERFLOW;
                        accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], bitFieldValue);
                        
                        if (accessStatus == ARS_NORMAL) {
#if defined (INTEGRITY_PPC_SYNC)
                           #pragma asm
                           eieio
                           sync
                           #pragma endasm
#elif defined (PPC_SYNC)
                           IO_SYNC;
#endif
                           accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &bitFieldValue);    /* read the status register - check for presense of data */
                           while (((bitFieldValue & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) != (CEI_UINT32)0U) && (accessStatus == ARS_NORMAL)) {
                              accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[bufferOffset], &labelIndex);     /* read the data register to reset status for this message */
                              accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &bitFieldValue); /* read the status register - check for presense of more data */
                           }
                        }
                        channelOffset  += CEIDEV_CHANNEL_REGION_SIZE;                            /* setup for next channel */
                        bufferOffset += CEIDEV_CHANNEL_REGION_SIZE;
                     }
                     memOffset += CEIDEV_CHANNEL_REGION_SIZE;
                  }
               
#ifndef CERTIFICATION_BUILD
                  /*--------------------------------------------------------------------*
                   * Flush the merged mode receiver FIFO
                   *--------------------------------------------------------------------*/
                  channelOffset = CEIDEV_CHANNEL_REGION_START + (deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_MERGED_RX] * CEIDEV_CHANNEL_REGION_SIZE);
                  bufferOffset = channelOffset + CEIDEV_CHANNEL_BUFFER_3;
                  /* clear any outstanding error status bits */
                  bitFieldValue = CEIDEV_CSR_RECV_FIFO_NOT_EMPTY + CEIDEV_CSR_RECV_FIFO_INVALID_MSG;
                  accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], bitFieldValue);
#if defined (PPC_SYNC)
                  IO_SYNC;
#endif
                  accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &bitFieldValue);    /* read the status register - check for presense of data */
                  while ((bitFieldValue & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) && (accessStatus == ARS_NORMAL)) {
                     accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[bufferOffset], &labelIndex);     /* read the data register to reset status for this message */
                     accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &bitFieldValue); /* read the status register - check for presense of more data */
                  }
#endif
               
               }
               initProgress = INIT_DISABLE_AGAIN;
               break;

            case INIT_DISABLE_AGAIN:
               /*--------------------------------------------------------------------*
                *  Disable all data processing
                *--------------------------------------------------------------------*/
               bitFieldValue = CEIDEV_GLOBAL_ENABLE_AUTON_SET + CEIDEV_SNAPSHOT_STORAGE_ENHANCED;
               accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
               bitFieldValue = CEIDEV_GLOBAL_ENABLE_AUTON_CLEAR + CEIDEV_GLOBAL_ENABLE_ON;
               accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
#ifdef INCLUDE_IRIG
               initProgress = INIT_DAC_IRIG;
               break;

            case INIT_DAC_IRIG:
               /* For those board's that don't have a fixed Discrete Input Threshold, initialize the Discrete DAC register to a mid-level value */
               if (deviceGlobalStructPtr->configuration == (CEI_UINT32)CEIDEV_CONFIG_AMCA30) {
                  accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_DAC_CONTROL], 0x0180U);
                  if (deviceGlobalStructPtr->irigInstalled) {
                     ar_sleep (10);  /* since writes to the DAC registers are serial, pause before writing the IRIG register */
                  }
               }
               
               /* if present, initialize the IRIG DAC register to the last adjusted value, or initial if not yet adjusted */
               if (deviceGlobalStructPtr->irigInstalled) {
                  accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_DAC_CONTROL], deviceLocalStructPtr->irigDacThreshold);
               
                  /* initialize the IRIG timer to host current time-of-year */
                  if (accessStatus == ARS_NORMAL) {
                     cei_get_irig_initial_time (&initialIrigTime);
                     accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_IRIG_UPDATE], initialIrigTime);
                  }
               }
#endif
               
#ifdef INCLUDE_DISCRETE_IO
               /* Assign the default state of the Discrete Outputs to Tri-state */
               if (deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_DISC_OUT] > 0U) {
                  channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_DISC_OUT] * CEIDEV_CHANNEL_REGION_SIZE);
                  for (deviceChannel = deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_DISC_OUT]; 
                       (deviceChannel <= deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_DISC_OUT]) && (accessStatus == ARS_NORMAL); 
                       deviceChannel++) {
                     accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], 0);
                     channelOffset += CEIDEV_CHANNEL_REGION_SIZE;
                  }
               }
#endif

               /* Acquire the current interrupt queue head pointer */
#ifdef INCLUDE_INTERRUPTS
               a429_utl_initialize_int_queue(board, FALSE);
#endif
               initProgress = INIT_TERMINATE;
               break;

            default:
               break;
         }
      }
      
#ifndef TARGET_EMBEDDED
      /* if multi-thread protection is active, free any acquired lock */
      if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
         tempStatus = cei_release_thread_lock(LOCK_IDENT);
      }
#endif
      if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL)) {
         status = (CDEV_API_RET_TYPE)accessStatus;
      }
#ifndef TARGET_EMBEDDED
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                       A 4 2 9 _ U T L _ M E M O R Y _ T E S T
 *===================================================================================================================*
 *
 * FUNCTION:    Perform a memory test of the CEI-x30 SRAM device
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board -- (input) Board number (0 to 127) 
 * 
 *              CEI_UINT32 testType   -- (input) size of memory test to execute
 *                                               FULL_SRAM   0 test all SRAM
 *                                               UNUSED_SRAM 1 test only unused SRAM (Tx Chan 16-31)
 *                                               SELECT_SRAM 2 test a portion of SRAM
 * 
 *              CEI_UINT32 testOption -- (input) when the testType is SELECT_SRAM, the testOption parameter specifies 
 *                                               which block of SRAM will be verified. Block select is based on a  
 *                                               length of 512 locations starting at SRAM address = 512*testOption
 * 
 * RETURN VAL:  CDEV_API_RET_TYPE -- ARS_NORMAL     Normal succesful completion
 *                                   ARS_INVARG     Invalid memory test request
 *                                   ARS_MEMWRERR   Memory test failure
 * GLOBALS:     device
 *              deviceGlobal
 *
 * DESCRIPTION: This function performs SRAM memory tests, either on a small unused portion (active processing test) 
 *              a selected portion of, or on the entire SRAM (destructive test that should only be executed when the 
 *              board is disabled, duration is six seconds to test all 2Mb of SRAM).
 *
 *===================================================================================================================*/ 
CDEV_API_RET_TYPE a429_utl_memory_test 
(
   CDEV_BOARD_TYPE board,                                                     /* board to reference (0 to 127)       */
   CEI_UINT32 testType,                                                       /* stype of memory test to execute     */
   CEI_UINT32 testOption                                                      /* SELECT_SRAM block identifier        */
)
{
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                     /* status for the memory test          */
   CEI_INT32 accessStatus = ARS_NORMAL;                                       /* status of device read/write access  */
   CEI_UINT32 writeValue;                                                     /* the data pattern written to memory  */
   CEI_UINT32 readValue;                                                      /* the data pattern read from memory   */
   CEI_UINT32 memoryIndex = 0U;                                               /* local index into memory block       */
   CEI_UINT32 minMemoryIndex = 0U;                                            /* starting location of memory test    */
   CEI_UINT32 maxMemoryAddress = 0U;                                          /* ending location of memory test      */
   CEI_UINT32 testAddress;                                                    /* temp address to write/read          */
   CEI_UINT32 patternIndex;                                                   /* index into data patterns (6 total)  */
   pCEI_UINT32 boardPointer;                                                  /* pointer to FPGA local memory area   */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                          /* ptr to board local structure        */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                                 /* ptr to brd global structure         */
   /* pattern information used in the SRAM Memory Test */
   CEI_UINT32 patternTest[AR_NUM_PATTERNS] =    {       1, 0xFFFFFFFE, 0xAAAAAAAA, 0x55555555,      0,         0};
   CEI_UINT32 patternInc[AR_NUM_PATTERNS] =     {       0,          1,          0,          0,      1,         0};
   CEI_UINT32 patternProcess[AR_NUM_PATTERNS] = {AR_SHIFT,   AR_SHIFT,  AR_STATIC,  AR_STATIC, AR_ADD, AR_STATIC};
   CEI_UINT32 unusedSram[AR_NUM_UNUSED_SRAM_LOCATIONS] = {0x28000U, 0x29555U, 0x2A666U, 0x2B999U, 0x2CAAAU, 0x2DFFFU};
#ifdef _DEBUG
   int errorCount = 0;
#endif
   
   /* assign a reference to this board's device and deviceGlobal structure-array elements */
   deviceLocalStructPtr = &device[board];
   deviceGlobalStructPtr = deviceGlobal[board];

   /* A memory test of internal FPGA memory is supported for validation, not good for operational use. */
   if (testType == MSG_SCHEDULE_TABLE) {
      /* disable "global enable" */
      accessStatus = cei_write_32bit (board, 
                                      &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], 
                                      CEIDEV_GLOBAL_ENABLE_AUTON_CLEAR + CEIDEV_GLOBAL_ENABLE_ON);
      
      minMemoryIndex = CEIDEV_MSG_SCHED_TABLE_START;
      maxMemoryAddress = deviceGlobalStructPtr->messageSchedulerTableEnd;
      
      for (patternIndex = 0; (patternIndex < AR_NUM_PATTERNS) && (status == ARS_NORMAL); patternIndex++) {
         boardPointer = (pCEI_UINT32)(deviceLocalStructPtr->devicePtr + CEIDEV_MSG_SCHED_TABLE_START);
         writeValue = patternTest[patternIndex];
      
         /* Write the pattern out to SRAM in CEI_INT32-word accesses */
         for (memoryIndex = minMemoryIndex; (memoryIndex < maxMemoryAddress) && (accessStatus == ARS_NORMAL); memoryIndex++) {
      
            /* manipulate the pattern as requested */
            switch (patternProcess[patternIndex]) {
               case AR_SHIFT:
                  if ( ((memoryIndex+1U) % 32U) == 0U ) {
                     writeValue = patternTest[patternIndex];
                  }
                  else {
                     writeValue = writeValue << 1;
                     writeValue += patternInc[patternIndex];
                  }
                  break;
               case AR_ADD:
                  writeValue += patternInc[patternIndex];
                  break;
               default:
                  break;
            }
      
            /* write the data pattern to this location */
            accessStatus = cei_write_32bit (board, boardPointer, writeValue);
            boardPointer++;
         }
         
         /* Read and verify the written pattern */
         boardPointer = (pCEI_UINT32)(deviceLocalStructPtr->devicePtr + CEIDEV_MSG_SCHED_TABLE_START);
         writeValue = patternTest[patternIndex];
         for (memoryIndex = minMemoryIndex; (memoryIndex < maxMemoryAddress) && (accessStatus == ARS_NORMAL); memoryIndex++) {
      
            /* manipulate the pattern as requested */
            switch (patternProcess[patternIndex]) {
               case AR_SHIFT:
                  if ( ((memoryIndex+1U) % 32U) == 0U ) {
                     writeValue = patternTest[patternIndex];
                  }
                  else {
                     writeValue = writeValue << 1;
                     writeValue += patternInc[patternIndex];
                  }
                  break;
               case AR_ADD:
                  writeValue += patternInc[patternIndex];
                  break;
               default:
                  break;
            }
      
            /* read and verify the contents of this location */
            accessStatus = cei_read_32bit (board, boardPointer, &readValue);
            
            if (readValue != writeValue) {
               status = ARS_MEMWRERR;
               break;
            }
            else {
               boardPointer++;
            }
         }
      } 
   }
   else {
      /* Support the standard SRAM pattern testing */
      switch (testType) {
         /*  During active CEI-x30 processing, only test six locations in unused Tx Channel 16-31
          *  FIFO Buffer SRAM, verifying specific Address Bit access with varying data patterns.
          */
         case UNUSED_SRAM:
            minMemoryIndex = 0U;
            maxMemoryAddress = AR_NUM_UNUSED_SRAM_LOCATIONS;
            break;
   
         case FULL_SRAM:
            accessStatus = cei_write_32bit (board, 
                                            &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], 
                                            CEIDEV_GLOBAL_ENABLE_AUTON_CLEAR + CEIDEV_GLOBAL_ENABLE_ON);
            minMemoryIndex = 0U;
            maxMemoryAddress = CEIDEV_SRAM_MAX;
            break;
   
         case SELECT_SRAM:
            if ( (testOption >= AR_BIT_SELECT_SRAM_MIN) && (testOption <= AR_BIT_SELECT_SRAM_MAX) ) {
               testOption -= AR_BIT_SELECT_SRAM_MIN;
               minMemoryIndex = (CEI_UINT32)testOption * SELECT_SRAM_TEST_BLOCK_SIZE;
               maxMemoryAddress = minMemoryIndex + SELECT_SRAM_TEST_BLOCK_SIZE;
               accessStatus = cei_write_32bit (board, 
                                               &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], 
                                               CEIDEV_GLOBAL_ENABLE_AUTON_CLEAR + CEIDEV_GLOBAL_ENABLE_ON);
            }
            else {
               status = ARS_INVARG;
            }
            break;
   
         default:
            status = ARS_INVARG;
      }
   
      /* Loop through all of the various patterns to test SRAM */
      for (patternIndex = 0; (patternIndex < AR_NUM_PATTERNS) && (status == ARS_NORMAL); patternIndex++) {
         writeValue = patternTest[patternIndex];
   
         /* Write the pattern out to SRAM in CEI_INT32-word accesses */
         for (memoryIndex = minMemoryIndex; (memoryIndex < maxMemoryAddress) && (accessStatus == ARS_NORMAL); memoryIndex++) {
   
            /* manipulate the pattern as requested */
            if (patternProcess[patternIndex] == AR_SHIFT) {
               if ( ((memoryIndex+1U) % 32U) == 0U ) {
                  writeValue = patternTest[patternIndex];
               }
               else {
                  writeValue = writeValue << 1;
                  writeValue += patternInc[patternIndex];
               }
            }
            else if (patternProcess[patternIndex] == AR_ADD) {
               writeValue += patternInc[patternIndex];
            }
   
            testAddress = (testType == UNUSED_SRAM) ? unusedSram[memoryIndex] : memoryIndex;
         
            /* load the SRAM CEI_INT32-word address index */
            /* write the data pattern to this location */
            accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_ACCESS_SRAM_ADDR], testAddress);
   
            /* Sync and flush the PCI/PCIe bus */
#if defined (INTEGRITY_PPC_SYNC)
     #pragma asm
     eieio
     sync
     #pragma endasm
#elif defined (PPC_SYNC)
     IO_SYNC;
#endif
            accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_ACCESS_SRAM_DATA], writeValue);
         }
   
         /* Sync and flush the PCI/PCIe bus */
#if defined (INTEGRITY_PPC_SYNC)
     #pragma asm
     eieio
     sync
     #pragma endasm
#elif defined (PPC_SYNC)
     IO_SYNC;
#endif
         
         /* Read and verify the written pattern */
         writeValue = patternTest[patternIndex];
         for (memoryIndex = minMemoryIndex; (memoryIndex < maxMemoryAddress) && (accessStatus == ARS_NORMAL); memoryIndex++) {
   
            /* manipulate the pattern as requested */
            if (patternProcess[patternIndex] == AR_SHIFT) {
               if ( ((memoryIndex+1U) % 32U) == 0U ) {
                  writeValue = patternTest[patternIndex];
               }
               else {
                  writeValue = writeValue << 1;
                  writeValue += patternInc[patternIndex];
               }
            }
            else if (patternProcess[patternIndex] == AR_ADD) {
               writeValue += patternInc[patternIndex];
            }
   
            testAddress = (testType == UNUSED_SRAM) ? unusedSram[memoryIndex] : memoryIndex;
         
            /* load the SRAM CEI_INT32-word address index */
            /* read and verify the contents of this location */
            accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_ACCESS_SRAM_ADDR], testAddress);
            
            /* Sync and flush the PCI/PCIe bus */
#if defined (INTEGRITY_PPC_SYNC)
     #pragma asm
     eieio
     sync
     #pragma endasm
#elif defined (PPC_SYNC)
     IO_SYNC;
#endif
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_ACCESS_SRAM_DATA], &readValue);
            
            if (readValue != writeValue) {
#ifdef _DEBUG
               printf("\nSRAM Memory %d Write/Read Failed, Wrote: %08X  Read: %08X\n", testAddress, writeValue, readValue);
               if (++errorCount > 20) {
                  status = ARS_MEMWRERR;
                  printf("\n\n");
               }
#else
               status = ARS_MEMWRERR;
               break;
#endif
            }
         }
      }
   }

   if ( (status == ARS_NORMAL) && (accessStatus != ARS_NORMAL) ) {
      status = (CDEV_API_RET_TYPE)accessStatus;
   }

   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                      A 4 2 9 _ U T L _ R E A D _ M E S S A G E
 *===================================================================================================================*
 *
 * FUNCTION:    Read the next data and 64-bit time-tag from a receive FIFO,
 *              returns the extended time-tag structure supporting IRIG.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input)  Board number (0 to 127) 
 *              CDEV_CHAN_TYPE* channel -- (input)  recv channel to reference (0 to 255)
 *                                         (output) in merged mode, the channel this data was actually received on
 *              pCEI_UINT32 data        -- (output) location to store received data
 *
 *              pAR_TIMETAG_TYPE timeTagRef
 *                                      -- (output) location for time-tag struct
 *
 * RETURN VAL   CDEV_API_RET_TYPE ARS_GOTDATA           Success
 *                                ARS_INVARG            NULL data parameter value
 *                                ARS_BOARD_MUTEX       Access to the Board Lock timed-out/failed
 *                                ARS_NODATA            No data availalble
 *                                ARS_BAD_MESSAGE       An invalid length msg was received
 *                                ARS_RX_BUFFER_OVERRUN A receive buffer overrun was detected
 *
 * GLOBALS:     device
 *              deviceGlobal
 * 
 * DESCRIPTION: This routine reads the most recently received entry (data and timetag) from the specified receive 
 *              channel FIFO.  The channel number passed to this routine corresponds to the hardware channel 
 *              assignment on the device.
 *
 *===================================================================================================================*/ 
CDEV_API_RET_TYPE a429_utl_read_message
(
   CDEV_BOARD_TYPE board,                                               /* (input)  board to reference (0 to 127)    */
   CDEV_CHAN_TYPE * channel,                                            /* (input)  recv channel to ref (0 to 255)   */
   pCEI_UINT32 data,                                                    /* (output) location to store received data  */
   pAR_TIMETAG_TYPE timeTagRef                                          /* (output) location to store data time-tags */
)
{
   CEI_UINT32 mergedModeState = AR_OFF;                                 /* 429 Rx ch merged/individual state         */
   CDEV_API_RET_TYPE recvStatus = ARS_NORMAL;                           /* FIFO receive status                       */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                               /* function return status                    */
   CEI_INT32 accessStatus = ARS_NORMAL;                                 /* status of device read/write access        */
   CEI_INT32 lockStatus = ARS_NORMAL;                                   /* status of board access lock acquis.       */
   CEI_UINT32 memOffset;                                                /* local offset to channel memory area       */
   CEI_UINT32 channelOffset;                                            /* local offset to channel memory area       */
   CEI_UINT32 channelStatus;                                            /* Rx channel FIFO data avail status         */
   CEI_UINT32 channelToRead = 0U;                                       /* merged/individual channel select          */
   CEI_UINT32 localChannel = 0U;                                        /* merged/individual channel select          */
   CEI_UINT32 requestedTimeTagFormat;                                   /* Format of the time-tag returned           */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                    /* ptr to board local structure              */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                           /* ptr to brd global structure               */
   
   deviceGlobalStructPtr = deviceGlobal[board];
   deviceLocalStructPtr = &device[board];

#ifndef TARGET_EMBEDDED
   if (data == NULL) {
      status = ARS_INVARG;
   }
   else {
      /* if multi-thread protection is active, acquire a lock on the host interface */
      if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
         lockStatus = cei_acquire_thread_lock(LOCK_IDENT);
      }
      if (lockStatus == ARS_NORMAL) {
#endif
         /* If the supplied channel index is the Merged Receive Channel, setup to use the Merged Rx buffer. */
         localChannel = (CEI_UINT32)*channel;
         if (localChannel == deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_MERGED_RX]) {
            mergedModeState = AR_ON;
         }
         else {
            /* If the supplied channel index is an individual receive channel and Merged Receive Mode 
             * is active for this channel, read from the Merged receive channel buffer.
             */
            if (deviceGlobalStructPtr->receiveChannelBufferMode[localChannel] == ARU_MERGED) {
               localChannel = (CEI_UINT32)deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_MERGED_RX];
               mergedModeState = AR_ON;
            }
         }
         channelToRead = localChannel;
#ifndef TARGET_EMBEDDED
      }

      status = a429_utl_validate_channel_request (board, localChannel, ARU_GET_DATA);
   
      if ((status == ARS_NORMAL) && (accessStatus == ARS_NORMAL) && (lockStatus == ARS_NORMAL)) {
#endif
         /* Check the receive channel status and assign return status in the following priority order...
          * If it indicates "BUFFER OVERRUN", save and return a status indicating a buffer overrun was encountered, reset indicator.
          * Else If it indicates "BAD MESSAGE RECEIVED", save and return a status indicating an invalid 429 message was previously detected, reset indicator.
          * No matter what, if it indicates the buffer is "NOT EMPTY", then retrieve a message and optionally the time-stamp from the buffer. 
          * Optionally, if the active buffer is the Merged Receive Buffer, read the channel on which the message was actually received.
          */
         channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS + (channelToRead * CEIDEV_CHANNEL_REGION_SIZE);
         accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &channelStatus);
         if (channelStatus & CEIDEV_CSR_RECV_FIFO_OVERFLOW) {
            recvStatus = ARS_RX_BUFFER_OVERRUN;
            accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], CEIDEV_CSR_RECV_FIFO_OVERFLOW);
         }
         else if (channelStatus & CEIDEV_CSR_RECV_FIFO_INVALID_MSG) {
            recvStatus = ARS_BAD_MESSAGE;
            accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], CEIDEV_CSR_RECV_FIFO_INVALID_MSG);
         }
         if ((channelStatus & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) && (accessStatus == ARS_NORMAL)) {
            /* if merged mode is enabled for this channel or the merged channel was specified,
             * read the actual channel on which the message was received from the receive buffer */
            if (mergedModeState == AR_ON) {
               channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_BUFFER_2 + (channelToRead * CEIDEV_CHANNEL_REGION_SIZE);
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], (pCEI_UINT32)&localChannel);
               mergedModeState = (CEI_UINT32)localChannel;
            }
            else {
               mergedModeState = CEIDEV_MAX_CHANNELS;
            }
   
            /* if the application time-tag reference is valid, read the time-stamp */
            if (timeTagRef != NULL) {
               memOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_BUFFER_0 + (channelToRead * CEIDEV_CHANNEL_REGION_SIZE);
               /* if application is using the board-wide time-tag format, select it here */
               if (timeTagRef->timeTagFormat == AR_TIMETAG_USE_PROGRAMMED_MODE) {
                  requestedTimeTagFormat = AR_USE_PROGRAMMED_TIMETAG_MODE;
               }
               else {
                  /* for any application specified time-tag format, restrict to only a valid supplied format */
                  requestedTimeTagFormat = 
                     ( (((CEI_INT32)timeTagRef->timeTagFormat >= AR_TIMETAG_EXT_IRIG_64BIT) && (timeTagRef->timeTagFormat < AR_TIMETAG_EXT_IRIG_REF_64BIT)) ||
                       ((timeTagRef->timeTagFormat >= AR_TIMETAG_SYNC_1553_CH1) && (timeTagRef->timeTagFormat <= AR_TIMETAG_SYNC_1553_CH4))
                     ) ? (CEI_UINT32)timeTagRef->timeTagFormat : (CEI_UINT32)AR_USE_PROGRAMMED_TIMETAG_MODE;
               }
               accessStatus = a429_utl_acquire_formatted_time_tag ( board,
                                                                    memOffset,
                                                                    AR_RECEIVER_FIFO_TIMETAG,
                                                                    requestedTimeTagFormat, 
                                                                    mergedModeState, 
                                                                    timeTagRef );
            }
            
            /* read the actual message last, triggering the fifo head ptr to increment to the next message */
            if (accessStatus == ARS_NORMAL) {
               *channel = (CDEV_CHAN_TYPE)localChannel;
               channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_BUFFER_3 + (channelToRead * CEIDEV_CHANNEL_REGION_SIZE);
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], data);
               if (accessStatus == ARS_NORMAL) {
                  status = ARS_GOTDATA;
               }
            }
         }
         else { /* no data available, exit with status = ARS_NODATA */
            status = ARS_NODATA;
            *data = 0U;
            if (timeTagRef != NULL) {
               timeTagRef->timeTag = (TIME_TAG_TYPE)0U;
               timeTagRef->referenceTimeTag = (TIME_TAG_TYPE)0U;
            }
         }
#ifndef TARGET_EMBEDDED
      }
      /* if multi-thread protection is active, free any acquired lock */
      if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
         lockStatus = cei_release_thread_lock(LOCK_IDENT);
      }
   }
#endif
   
   /* handle the special return status cases for Buffer Overrun and Invalid Message Detected */
   if ( (recvStatus != ARS_NORMAL) && (accessStatus == ARS_NORMAL) && (lockStatus == ARS_NORMAL) ) {
      status = recvStatus;
   }

   return (status);
}

/*===================================================================================================================*
 * ENTRY POINT:           A 4 2 9 _ U T L _ V A L I D A T E _ C H A N N E L _ R E Q U E S T 
 *===================================================================================================================*
 *
 * FUNCTION:    Validate the channel / item combination specified by the
 *              host application.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input) board to reference (0 to 127)
 *              CEI_UINT32 channel      -- (input) h/w channel select value
 *              CDEV_PARM_SSI_TYPE item -- (input) config register attribute to check
 * 
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL     Normal succesful completion.
 *                                ARS_INVHARVAL  Item was invalid or channel not supported on this device
 *                                ARS_INVBOARD   Board parameter value invalid or board has not been initialized
 * 
 * GLOBALS:     deviceGlobal
 *
 * DESCRIPTION: This function assures the provided channel and configuration item are valid for the specified 
 *              board configuration.
 *
 *===================================================================================================================*/ 
CDEV_API_RET_TYPE a429_utl_validate_channel_request 
(
   CDEV_BOARD_TYPE board,                                                     /* board to reference (0 to 127)       */
   CEI_UINT32 channel,                                                        /* h/w channel select value            */
   CDEV_PARM_SSI_TYPE item                                                    /* config register attribute to check  */
)
{
   CDEV_API_RET_TYPE status;                                                    /* function return status            */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                                   /* ptr to brd global structure       */

   if ( (board >= MAX_BOARDS) || !(deviceGlobal[board]->boardInitialized) ) {
      status = ARS_INVBOARD;
   }
   else {
      status = ARS_NORMAL;
      deviceGlobalStructPtr = deviceGlobal[board];

      switch (item) {
         /* Validate ARINC 429 specific receive channel attribute requests */
         case ARU_RX_PARITY:
         case ARU_RX_MSG_SIZE_24BIT:
         case ARU_ENHANCED_FILTER_LABEL:
         case ARU_ACCESS_SNAPSHOT_BUFFER:
            if ( ( (channel > deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]) ||
                   (channel < deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]) ) 
               ) {
               status = ARS_INVHARVAL;
            }
            break;
   
         /* Validate ARINC 429 and 717 specific receive channel attribute requests */
         case ARU_RX_BITRATE:
         case ARU_RECV_MODE:
         case ARU_RX_STORAGE_MODE:
         case ARU_RX_MERGED_MODE:
         case ARU_RX_FIFO_ENABLE:
         case ARU_RX_DISABLE:
         case ARU_RX_RESET:
         case ARU_RX_FIFO_RESET:
            if ( ( (channel > deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]) ||
                   (channel < deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]) ) 
                 &&
                 ( (channel > deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC573_RX]) ||
                   (channel < deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX]) ) 
               ) {
               status = ARS_INVHARVAL;
            }
            break;

         /* Validate generic receive channel attribute requests */
         case ARU_RX_FIFO_COUNT:
         case ARU_GET_DATA:
            if ( ( (channel > deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]) ||
                   (channel < deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]) ) 
                 &&
                 ( (channel > deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC573_RX]) ||
                   (channel < deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX]) ) 
                 &&
                   (channel != deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_MERGED_RX])
               ) {
               status = ARS_INVHARVAL;
            }
            break;


         /* Validate ARINC 429 specific transmit channel attribute requests */
         case ARU_TX_BIT_ERROR:
         case ARU_TX_GAP_ERROR:
         case ARU_TX_PARITY:
         case ARU_TX_RESET:
         case ARU_TX_MSG_SIZE_24BIT:
         case ARU_ADD_SCHEDULED_MSG:
            if ( ( (channel > deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_TX]) ||
                   (channel < deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX]) ) 
               ) {
               status = ARS_INVHARVAL;
            }
            break;
   
         /* Validate ARINC 429 and 717 specific transmit channel attribute requests */
         case ARU_TX_BITRATE:
         case ARU_TX_DISABLE:
         case ARU_TX_FIFO_ENABLE:
         case ARU_TX_FIFO_COUNT:
         case ARU_DISCRETE_RAW:
         case ARU_FAST_SLEW_RATE:
         case ARU_PUT_DATA:
            if ( ( (channel > deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_TX]) ||
                   (channel < deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX]) ) 
                 &&
                 ( (channel > deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC573_TX]) ||
                   (channel < deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX]) ) 
               ) {
               status = ARS_INVHARVAL;
            }
            break;

         /* Validate ARINC 573 specific receive channel attribute requests */
         case ARU_573_RX_ENABLE:
         case ARU_573_RX_AUTO_DETECT:
            if ( (channel >= deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC573_RX]) &&
                 (channel < deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX]) ) {
               status = ARS_INVHARVAL;
            }
            break;

         case ARU_573_SYNC_WORD1:
         case ARU_573_SYNC_WORD2:
         case ARU_573_SYNC_WORD3:
         case ARU_573_SYNC_WORD4:
            if ( (channel > deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC573_RX]) &&
                 (channel < deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX]) ) {
               status = ARS_INVHARVAL;
            }
            break;

         /* Validate ARINC 573 generic channel attribute requests */
         case ARU_573_TX_BPRZ_SELECT:
         case ARU_573_RX_BPRZ_SELECT:
            if ( ( (channel > deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC573_TX]) ||
                   (channel < deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX]) ) &&
                 ( (channel > deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC573_RX]) ||
                   (channel < deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX]) ) ) {
               status = ARS_INVHARVAL;
            }
            break;
   
         case ARU_573_TX_HBP_SELECT:
         case ARU_573_TX_SLEW_RATE:
            if ( (channel > deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC573_TX]) &&
                 (channel < deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX]) ) {
               status = ARS_INVHARVAL;
            }
            break;
   
         /* Validate Discrete Output channel attribute requests */
         case ARU_DISCRETE_OUT:
            if ( (channel > deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_DISC_OUT]) &&
                 (channel < deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_DISC_OUT]) ) {
               status = ARS_INVHARVAL;
            }
            break;
   
         /* Validate Discrete Output channel attribute requests */
         case ARU_DISCRETE_IN:
            if ( (channel > deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_DISC_IN]) &&
                 (channel < deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_DISC_IN]) ) {
               status = ARS_INVHARVAL;
            }
            break;
   
         /* Validate Differential Input channel attribute requests */
         case ARU_DIFFERENTIAL_IN:
            if ( (channel > deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_DISC_IN]) &&
                 (channel < deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_DISC_IN]) ) {
               status = ARS_INVHARVAL;
            }
            break;
   
         /* Validate Differential Output channel attribute requests */
         case ARU_DIFFERENTIAL_OUT:
            if ( (channel > deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_DIFF_OUT]) &&
                 (channel < deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_DIFF_OUT]) ) {
               status = ARS_INVHARVAL;
            }
            break;
   
         case ARU_573_RX_SUBFRAME_SIZE:
         case ARU_DISCRETE_OUT1:
         case ARU_DISCRETE_OUT2:
            break;
      
         default : status = ARS_INVHARVAL; 
            break;
      }
   }
   
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                        A 4 2 9 _ U T L _ W R A P _ T E S T
 *===================================================================================================================*
 *
 * FUNCTION:    Perform an internal wrap test using the specified number of ARINC 429 messages on each transmit-
 *              receive channel pairs installed on the device.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board  -- (input) board to reference (0 to 127)
 * 
 *              CEI_UINT32 wrap        -- (input) type of wrap to enable: 
 *                                                LOCAL_INTERNAL_WRAP  for internal wrap 
 *                                                LOCAL_EXTERNAL_WRAP for external wrap
 * 
 *              CEI_UINT32 msgCount    -- (input) number of messages to transmit 
 *                                                limited to AR_WRAP_TEST_MSG_COUNT
 * 
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL          Normal succesful completion.
 *                                ARS_INVHARVAL       Item was invalid or channel not supported on this device
 *                                ARS_INVBOARD        Board parameter value invalid or board has not been initialized
 *                                ARS_BOARD_MUTEX     Access to the Board Lock timed-out/failed
 *                                ARS_WRAP_DROP_FAIL  Wrap Data Missing
 *                                ARS_WRAP_DATA_FAIL  Wrap Data Pattern Mismatch
 *                                ARS_WRAP_FLUSH_FAIL Unknown External data received
 *                                ARS_XMITOVRFLO      Transmit buffer overrun has occurred
 * GLOBALS:     device
 *              deviceGlobal
 *
 * DESCRIPTION: This function performs an internal wrap test based on a specified number of ARINC 429 messages on  
 *              each transmit/receive channel pair installed on the device.  Configuration parameters are set for  
 *              no parity and 100Kbps.
 *
 *===================================================================================================================*/ 
CDEV_API_RET_TYPE a429_utl_wrap_test 
(
   CDEV_BOARD_TYPE board,                                             /* (input)  board to reference (0 to 127)      */
   CEI_UINT32 wrap,                                                   /* (input)  type of wrap INTERNAL/EXTERNAL     */
   CEI_UINT32 msgCount                                                /* (input)  number of messages to transmit     */
)
{
   CEI_UINT32 numTxChannels;                                          /* number of transmitters/rcvrs                */
   CEI_UINT32 numRxChannels;                                          /* number of transmitters/rcvrs                */
   CEI_UINT32 channelIndex;                                           /* index of transmitters/rcvrs                 */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                             /* function return status                      */
   CEI_UINT32 currentTxConfig[MAX_CHANNEL];                           /* saved Chan Config on entry                  */
   CEI_UINT32 currentRxConfig[MAX_CHANNEL];                           /* saved Chan Config on entry                  */
   CEI_UINT32 bitFieldValue;                                          /* temp Chan Config variable                   */
   CEI_UINT32 channelBaudRate;                                        /* temp Chan baud rate setting                 */
   CEI_INT32  accessStatus = ARS_NORMAL;                              /* status of device read/write access          */
   CEI_UINT32 testIndex = 0U;                                         /* message counter/pattern gen                 */
   CEI_UINT32 deviceChannel;                                          /* general channel index                       */
   CEI_UINT32 channelStatus = 0;                                      /* Rx channel FIFO data avail status           */
   CEI_UINT32 globalEnableRegData;                                    /* contents of the Global Enable reg           */
   ARINC testOutput;                                                  /* Output test data                            */
   ARINC testInput[AR_WRAP_TEST_MSG_COUNT+2];                         /* Received data                               */
   CEI_UINT32  channelOffset;                                         /* chan reg region offset in host interface    */
   CEI_UINT32  bufferOffset;                                          /* chan buffer region offset in host interface */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                  /* ptr to board local structure                */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                         /* ptr to brd global structure                 */
   AR_CHANNEL_INDEX_INFO_TYPE * channelInfoTablePtr = (AR_CHANNEL_INDEX_INFO_TYPE *)&deviceGlobal[board]->boardChannelIndexInfoStruct;

   /* assign references to this board's device and deviceGlobal structure-array elements */
   deviceLocalStructPtr = &device[board];
   deviceGlobalStructPtr = deviceGlobal[board];

   /* limit the wrap test message count */
   if (msgCount > AR_WRAP_TEST_MSG_COUNT) {
      msgCount = AR_WRAP_TEST_MSG_COUNT;
   }

   /* for a combo board, use a different bus speed frequency value and limit external s/w programmable transmitter usage */
   if ( (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15) ||
        (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15XT) ) {
      channelBaudRate = CEIDEV_20MHZ_BAUD_RATE_100K;
      numRxChannels = deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX];
      if ((wrap == LOCAL_INTERNAL_WRAP) && (deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX] > 2U)) { 
         numTxChannels = deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX];
      }
      else {
         numTxChannels = 2U; 
      }
   }
   /* for a non-combo board with no fixed receivers, limit to a single s/w programmable transmitter usage on internal wrap */
   else if ( (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR_XMC) && (wrap == LOCAL_INTERNAL_WRAP) &&
             (channelInfoTablePtr->chan_type[0] == AR_CHAN_TYPE_429_RX_OR_TX) && (channelInfoTablePtr->num_channels == 16) &&
             (channelInfoTablePtr->chan_rx_index[0] == channelInfoTablePtr->chan_tx_index[0]) ) {
      numTxChannels = 1U; 
      numRxChannels = 1U;
      channelBaudRate = CEIDEV_BAUD_RATE_100K;
   }
   else {
      channelBaudRate = CEIDEV_BAUD_RATE_100K;
      numTxChannels = deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX]; 
      numRxChannels = deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX];
   }


#ifndef TARGET_EMBEDDED
   /* if multi-thread protection requested, acquire semaphore/mutex-locking */
   if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
      accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
   }
#endif

   /* save the existing channel global enable register state, disable data processing temporarily */
   accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], &globalEnableRegData);
   accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], CEIDEV_GLOBAL_ENABLE_AUTON_CLEAR + CEIDEV_GLOBAL_ENABLE_ON);

   /* save the existing channel configuration for all transmit and receive channels */
   channelIndex = 0U;
   channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX] * CEIDEV_CHANNEL_REGION_SIZE);
   for (deviceChannel = deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX]; 
        deviceChannel <= deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_TX]; 
        deviceChannel++) {
      accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &currentTxConfig[channelIndex]);
      channelOffset += CEIDEV_CHANNEL_REGION_SIZE;   /* increment to the next channel memory block */
      channelIndex++;
   }
   channelIndex = 0U;
   channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] * CEIDEV_CHANNEL_REGION_SIZE);
   for (deviceChannel = deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]; 
        deviceChannel <= deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]; 
        deviceChannel++) {
      accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &currentRxConfig[channelIndex]);
      channelOffset += CEIDEV_CHANNEL_REGION_SIZE;   /* increment to the next channel memory block */
      channelIndex++;
   }
   
   /* Configure each transmit/receive channel pair for 100Kbps with parity disabled 
    * Additionally configure the wrap, for internal wrap disable external transmission.
    */
   channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX] * CEIDEV_CHANNEL_REGION_SIZE);
   bitFieldValue = CEIDEV_FAST_SLEW_RATE + CEIDEV_CHANNEL_ENABLE + (channelBaudRate << CEIDEV_BAUD_RATE_SHIFT);
   if (wrap == LOCAL_INTERNAL_WRAP) {
      bitFieldValue |= CEIDEV_TRANSMIT_DISABLE;
   }
   for (deviceChannel = deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX]; 
        deviceChannel <= deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_TX]; 
        deviceChannel++) {
      if (deviceChannel >= (numTxChannels + deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX])) {
         accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], (bitFieldValue | CEIDEV_TRANSMIT_DISABLE));
      }
      else {
         accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], bitFieldValue);
      }
      channelOffset += CEIDEV_CHANNEL_REGION_SIZE;   /* increment to the next channel memory block */
   }
   
   channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] * CEIDEV_CHANNEL_REGION_SIZE);
   bitFieldValue = CEIDEV_FAST_SLEW_RATE + CEIDEV_CHANNEL_ENABLE + (channelBaudRate << CEIDEV_BAUD_RATE_SHIFT);
   if (wrap == LOCAL_INTERNAL_WRAP) {
      bitFieldValue += CEIDEV_INTERNAL_WRAP_ENABLE;
   }
   for (deviceChannel = deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]; 
        deviceChannel <= deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]; 
        deviceChannel++) {
      accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], bitFieldValue);
      channelOffset += CEIDEV_CHANNEL_REGION_SIZE;   /* increment to the next channel memory block */
   }
      
   /* enable ARINC data processing */
   accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], CEIDEV_GLOBAL_ENABLE_AUTON_SET + CEIDEV_GLOBAL_ENABLE_ON);
   accessStatus = cei_delay_milliseconds(1U);

   numTxChannels = numTxChannels > numRxChannels ? numRxChannels : numTxChannels;
   for (channelIndex = 0U; (channelIndex < numTxChannels) && (status == ARS_NORMAL); channelIndex++) {
      /* check for an actively transmitting connection attached to any receiver - bail if so */
      status = ARS_NORMAL;
      bufferOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_BUFFER_3 + ((deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] + channelIndex) * CEIDEV_CHANNEL_REGION_SIZE);
      channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS + ((deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] + channelIndex) * CEIDEV_CHANNEL_REGION_SIZE);
      for (testIndex = 0U; (testIndex < 10U) && (status == ARS_NORMAL); testIndex++) {
         accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &channelStatus);
         if ( (channelStatus & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) && (accessStatus == ARS_NORMAL) ) {
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[bufferOffset], &testInput[0]);     /* read the data register to reset status for this message */
         }
         else {
            break;
         }
      }
      if (channelStatus & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) {
         status = ARS_WRAP_FLUSH_FAIL;
         break;
      }
   
      /*---------------------------------------------------------------*
       *  Test ARINC 429 tranmsit & receive
       *---------------------------------------------------------------*/
      
      channelOffset = CEIDEV_CHANNEL_REGION_START + ((deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX] + channelIndex) * CEIDEV_CHANNEL_REGION_SIZE);
      bufferOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_BUFFER_3 + ((deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX] + channelIndex) * CEIDEV_CHANNEL_REGION_SIZE);
      for (testIndex = 0U; (testIndex < msgCount) && (status == ARS_NORMAL); testIndex++) { 
         testOutput = (CEI_UINT32)(testIndex + (CEI_UINT32)((testIndex+1U) << 16) + 1U);
         testInput[0] = 0U;
         /* check the transmit channel status, if it indicates "NOT FULL"
          * then stuff the data into the queue; otherwise, it's an overflow.
          */
         accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &channelStatus);
         if (accessStatus == ARS_NORMAL) {
            if (channelStatus & CEIDEV_CSR_XMIT_FIFO_FULL) {
               status = ARS_XMITOVRFLO;
            }
            else {
               /* stuff the data into the transmit FIFO */
               accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[bufferOffset], testOutput);
               status = ARS_NORMAL;
            }
         }
      }
      /* delay for the initial transmission to begin then more for each message transmitted */
      if (status == ARS_NORMAL) {
         accessStatus = cei_delay_milliseconds((CEI_UINT32)((CEI_UINT32)10U + msgCount));
      }
   
      /* attempt to receive the specified burst count */
      if (status == ARS_NORMAL) {
         status = ARS_NORMAL;
         testIndex = 0U;
         channelOffset = CEIDEV_CHANNEL_REGION_START + ((deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] + channelIndex) * CEIDEV_CHANNEL_REGION_SIZE);
         bufferOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_BUFFER_3 + ((deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] + channelIndex) * CEIDEV_CHANNEL_REGION_SIZE);
         
         while ( (status == ARS_NORMAL) && (testIndex < msgCount+1) ) {
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &channelStatus);
            if ( (channelStatus & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) && (accessStatus == ARS_NORMAL) ) {
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[bufferOffset], &testInput[testIndex]);
               testIndex++;
            }
            else {
               status = ARS_NODATA;
            }
         }
         
         /* check to see we received the correct number of messages */
         if (testIndex < msgCount) {
            status = ARS_WRAP_DROP_FAIL;
         }
         else if (testIndex > msgCount) {
            status = ARS_WRAP_FLUSH_FAIL;
         }
         else {
            status = ARS_NORMAL;
         }
      }

      /* if all messages were received, verify the data content for all received messages */
      if (status == ARS_NORMAL) {
         for ( testIndex = 0U; (testIndex < msgCount) && (status == ARS_NORMAL); testIndex++ ) { 
            if (testInput[testIndex] != (CEI_UINT32)(testIndex + (CEI_UINT32)((testIndex+1U) << 16) + 1U)) {
               status = ARS_WRAP_DATA_FAIL;
            }
         } 
      }
   }
   
   /* restore the previous global enable register setting */
   accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], globalEnableRegData);

   /* restore the previous channel configuration for all transmit and receive channels */
   channelIndex = 0U;
   channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX] * CEIDEV_CHANNEL_REGION_SIZE);
   for (deviceChannel = deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX]; 
        deviceChannel <= deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_TX]; 
        deviceChannel++) {
      accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], currentTxConfig[channelIndex]);
      channelOffset += CEIDEV_CHANNEL_REGION_SIZE;   /* increment to the next channel memory block */
      channelIndex++;
   }
   channelIndex = 0U;
   channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] * CEIDEV_CHANNEL_REGION_SIZE);
   for (deviceChannel = deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]; 
        deviceChannel <= deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]; 
        deviceChannel++) {
      accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], currentRxConfig[channelIndex]);
      channelOffset += CEIDEV_CHANNEL_REGION_SIZE;   /* increment to the next channel memory block */
      channelIndex++;
   }
#ifndef TARGET_EMBEDDED
   /* if multi-thread protection is active, free any acquired lock */
   if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
      accessStatus = cei_release_thread_lock(LOCK_IDENT);
   }
#endif
   
   return status;
}

/*****************************************************************************
 *****************************************************************************
 *                    A P I   P U B L I C   F U N C T I O N S                *
 *****************************************************************************
 *****************************************************************************/


/*===================================================================================================================*
 * ENTRY POINT:                      A R _ B O A R D _ T E S T
 *===================================================================================================================*
 *
 * FUNCTION:    This routine tests the functionality of the x30 device as part of the initialization process.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board       -- (input) board to reference (0 to 127)
 *              CDEV_PARM_SSI_TYPE testType -- (input) type of test to execute
 *                                                     INTERNAL_WRAP - perform an ARINC 429 internal wrap test 
 *                                                     EXTERNAL_WRAP - perform an ARINC 429 external wrap test 
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL          Success
 *                                ARS_INVBOARD        Invalid board parameter
 *                                ARS_BOARD_MUTEX     Access to the Board Lock timed-out/failed
 *                                ARS_MEMWRERR        Memory test failure
 *                                ARS_WRAP_DROP_FAIL  Wrap Data Missing
 *                                ARS_WRAP_DATA_FAIL  Wrap Data Pattern Mismatch
 *                                ARS_WRAP_FLUSH_FAIL Unknown External data received
 *                                ARS_XMITOVRFLO      Transmit buffer overrun has occurred
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine performs an internal or external wrap from each ARINC 429 transmit channel matched to 
 *              a respective receive channel, (unbalanced channels are not tested).  It is followed by a complete 
 *              reinitialization of the host interface to a semi-power-up state.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_board_test 
(
   CDEV_BOARD_TYPE board,                                                   /* (input) board to reference (0 to 127) */
   CDEV_PARM_SSI_TYPE testType                                              /* (input) type of test to execute       */
)
{
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                        /* function return status           */
   CEI_UINT32 wrapState;                                                         /* local internal wrap state select */
   
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      /*------------------------------------------------------------------*
       *  Perform a single message Wrap Test on all ARINC 429 channels
       *------------------------------------------------------------------*/
      wrapState = (testType == (CDEV_PARM_SSI_TYPE)INTERNAL_WRAP) ? LOCAL_INTERNAL_WRAP : LOCAL_INTERNAL_WRAP;
      status = a429_utl_wrap_test (board, wrapState, 1U);
   
      /* Initialize the device following the test */
      if (status == ARS_NORMAL) {
         status = ar_initialize_device (board);
      }
#ifndef TARGET_EMBEDDED
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   return (status);
}


/*===================================================================================================================*
 * ENTRY POINT:                      A R _ B Y P A S S _ W R A P _ T E S T
 *===================================================================================================================*
 *
 * FUNCTION:    This routine saves the application's request to bypass the internal wrap test during board 
 *              initialization.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board     -- (input) board to reference (0 to 127)
 *              CDEV_PARM_SSI_TYPE bypass -- (input) AR_ON  bypass internal wrap test
 *                                                   AR_OFF execute internal wrap test
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL       Success
 *                                ARS_INVBOARD     Invalid board parameter
 *
 * GLOBALS:     bypassWrapTest[]
 * 
 * DESCRIPTION: This routine saves the app's request for internal wrap execution during board initialization.  This 
 *              is supplied in case the board is initialized while ARINC data is active on any receiver channel, 
 *              which would cause the wrap test invoked via ar_open to fail.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_bypass_wrap_test 
(
   CDEV_BOARD_TYPE board,                                              /* (input) board to reference (0 to 127)      */
   CDEV_PARM_SSI_TYPE bypass                                           /* (input) bypass (or not) internal wrap test */
)
{
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                               /* function return status    */
   
#ifndef TARGET_EMBEDDED
   if (board < MAX_BOARDS) {
#endif
      bypassWrapTest[board] = (bypass == AR_ON) ? TRUE : FALSE;
#ifndef TARGET_EMBEDDED
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                             A R _ C L O S E 
 *===================================================================================================================*
 *
 * FUNCTION:    Release all resources for the specified device.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board -- (input) board to reference (0 to 127)
 *
 * RETURN:      CDEV_API_RET_TYPE ARS_NORMAL         - Succesful completion of routine
 *                                ARS_INVBOARD       - Invalid board number specified or board never initialized
 *                                ARS_INT_ISR        - Interrupt was not terminated successfully
 *                                ARS_BOARD_MUTEX    - failed to release the host interface lock
 *                                ARS_ERR_SH_MEM_OBJ - failed to release shared memory resource
 *                                ARS_DRIVERFAIL     - could not access the CEI_Install library or
 *                                                     the Driver Close Session invocation failed
 *                                ARS_FAILURE        - <Windows only> CEI_Install library failed to load in memory
 * 
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: Call into the respective low-level library to free the resources associated with this board,
 *              then terminate the session.
 *
 *===================================================================================================================*/                              
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_close
(
   CDEV_BOARD_TYPE board                                               /* (input) board to reference (0 to 127)      */
)
{    
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                     /* function return status              */
   CEI_INT32 accessStatus = ARS_NORMAL;                                       /* status of device read/write access  */
   pCEI_UINT32 intCtrlRegPtr = NULL;                                          /* ptr to the PLX Interrupt Ctrl Reg   */
   CEI_UINT32 interruptBit = 0U;                                              /* the respective PLX Int Enable bit   */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                          /* ptr to board local structure        */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                                 /* ptr to brd global structure         */
   CEI_UINT32 okToReleaseLock = 0U;                                           /* thread lock mechanism access status */

#ifndef TARGET_EMBEDDED
   if ((board < MAX_BOARDS) && (deviceGlobal[board] != NULL)) {
#endif
   
      /* assign references to this board's device and deviceGlobal structure-array elements */
      deviceLocalStructPtr = &device[board];
      deviceGlobalStructPtr = deviceGlobal[board];

#ifndef TARGET_EMBEDDED
      /* if multi-thread protection is active, acquire a lock on the host interface */
      if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
         accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
         okToReleaseLock = 1U;
      }
      if (accessStatus == ARS_NORMAL) {
#endif
         if ((deviceGlobalStructPtr->attachedProcessCount == 1U) || (deviceLocalStructPtr->multiThreadProtect == AR_CONC_NONE)) {
            if (deviceGlobalStructPtr->boardInitialized) {
               
               /* disable "global enable"  and "interrupt enable" */
               accessStatus = cei_write_32bit (board, 
                                               &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], 
                                               CEIDEV_GLOBAL_ENABLE_AUTON_CLEAR + CEIDEV_GLOBAL_ENABLE_ON + CEIDEV_INTERRUPTS_ENABLED);
               
               /* if the PCI interrupt via Label Filter feature was enabled, disable it here */
               if (deviceGlobalStructPtr->ceiPciInterruptEnabled) {
                  switch (deviceGlobalStructPtr->configuration) {
                     /* PLX PCI9030 Interrupt setup */
                     case CEIDEV_CONFIG_CEI430:
                     case CEIDEV_CONFIG_RAR_EC:
                     case CEIDEV_CONFIG_CEI430A:
                        intCtrlRegPtr = deviceLocalStructPtr->baseAddr + PLX9030_CFG_INT_CTRL_STAT_REG;
                        interruptBit = PLX9030_LOCAL_INT_INP_ENABLE;
                        break;

                     /* PLX PCI9056 Interrupt setup */
                     case CEIDEV_CONFIG_CEI830:
                     case CEIDEV_CONFIG_AMCA30:
                     case CEIDEV_CONFIG_CEI530:
                     case CEIDEV_CONFIG_R830RX:
                     case CEIDEV_CONFIG_RAR_CPCI:
                     case CEIDEV_CONFIG_R830X820:
                     case CEIDEV_CONFIG_RCEI830A:
                        intCtrlRegPtr = deviceLocalStructPtr->baseAddr + PLX_CONFIG_INT_CTRL_STAT_REG;
                        interruptBit = PLX9056_LOCAL_INT_INP_ENABLE;
                        break;

                     /* No PLX component for native PCI Express - RAR-PCIE, RAR15-XMC... */
                     case CEIDEV_CONFIG_RAR_PCIE: 
                     case CEIDEV_CONFIG_RAR_MPCIE: 
                     case CEIDEV_CONFIG_RAR15:    
                     case CEIDEV_CONFIG_RAR15XT:  
                     case CEIDEV_CONFIG_RAR_XMC:
                     default:  
                        break;
                  }
                  accessStatus = cei_terminate_interrupt(board, interruptBit, intCtrlRegPtr);
               }
               deviceGlobalStructPtr->boardInitialized = FALSE;
            }
            /* reset global structure members prior to possible object relinquish for multiprocess application usage */
            deviceGlobalStructPtr->ceiPciInterruptEnabled = FALSE;
            deviceGlobalStructPtr->attachedProcessCount = 0U;

            status = (CDEV_API_RET_TYPE)cei_release_device(board, (ARINC_LOCAL_TYPE *)&device[board]);
   
            deviceLocalStructPtr->devicePtr = 0U;
            deviceLocalStructPtr->baseAddr = 0U;
            okToReleaseLock = 0U;

#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
            /* for Linux multi-process application usage, cleanup of the System V shared memory allocation has to occur here */
            if (device->multiThreadProtect == AR_CONC_MULTIPROC) {
               shmdt(&deviceGlobalStructPtr); 
               shmctl((int)deviceLocalStructPtr->sharedMemoryHandle, IPC_RMID, NULL);
            }
#endif
            deviceGlobalStructPtr = NULL;
            deviceGlobal[board] = NULL;
            if (accessStatus != ARS_NORMAL) {
               status = (CDEV_API_RET_TYPE)accessStatus;
            }
         }
         else if (deviceGlobalStructPtr->attachedProcessCount > 0U) {
            deviceGlobalStructPtr->attachedProcessCount--;
         }
         
#ifndef TARGET_EMBEDDED
         if (okToReleaseLock) {
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
         }
         
      }
   }
   else {
      status = (CDEV_API_RET_TYPE)ARS_INVBOARD;
   }
#endif

   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                         A R _ C L R _ R X _ C O U N T
 *===================================================================================================================*
 *
 * FUNCTION:    Reset the counter for received messages on this channel
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input) Board to reference (0 to 127)
 *              CDEV_CHAN_TYPE  channel -- (input) Receive channel to use (0 to 31)
 * 
 * RETURN VAL   none
 * 
 * GLOBALS:     deviceGlobal[]
 *              device[]
 * 
 * DESCRIPTION: The x30 f/w maintains a count of the number of ARINC data messages received over the interface for 
 *              each channel since the device was last initialized (see ar_open).  The API maintains a separate
 *              local copy of that count that is assigned to the board value upon invocation of this routine.  The
 *              ar_get_rx_count returns the difference between the value read within this routine and the value
 *              read from the board upon invocation of ar_get_rx_count, thus providing a "running count" of the
 *              current receive message counter that can be reset (e.g. specifically for the purpose of providing
 *              legacy ARINC 429 API support).
 *
 *===================================================================================================================*/
EXPORT32 CEI_VOID DLL_EXPORTED ar_clr_rx_count 
( 
   CDEV_BOARD_TYPE board,                                                /* (input)  board to reference (0 to 127)   */
   CDEV_CHAN_TYPE channel                                                /* (input)  recv channel to ref (0 to 31)   */
)
{
   CEI_INT32  accessStatus = ARS_NORMAL;                                 /* status of dev read/write access          */
   CEI_UINT32 offsetChannelValue;                                        /* actual channel on the device             */
   CEI_UINT32 offsetIntoChannelWindow;                                   /* offset into chan mem region              */
   CEI_UINT32 bitFieldValue = 0;                                         /* temp for device reg read content         */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      if ((CEI_UINT32)channel < deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) {
#endif
         offsetChannelValue = (CEI_UINT32)channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
         offsetIntoChannelWindow = CEIDEV_CHANNEL_STATS_START + (CEI_UINT32)offsetChannelValue;
         accessStatus = cei_read_32bit (board, &device[board].devicePtr[offsetIntoChannelWindow], &bitFieldValue);
         deviceGlobal[board]->receiveChannelMessageCount[channel] = bitFieldValue;
#ifndef TARGET_EMBEDDED
      }
   }
#endif
   if (accessStatus) ;
}
 

/*===================================================================================================================*
 * ENTRY POINT:                          A R _ E X E C U T E _ B I T
 *===================================================================================================================*
 *
 * FUNCTION:    This routine performs initiated BIT functionality testing of the device as specified by the host 
 *              application.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board       -- (input) board to reference (0 to 127)
 *              CDEV_PARM_SSI_TYPE testType -- (input) type of BIT to execute
 *                                                     AR_BIT_BASIC_STARTUP
 *                                                     AR_BIT_FULL_STARTUP
 *                                                     AR_BIT_PERIODIC
 *                                                     AR_BIT_INT_LOOPBACK
 *                                                     AR_BIT_EXT_LOOPBACK
 *                                                     AR_BIT_PARTIAL_SRAM   
 *                                                     AR_BIT_FULL_SRAM      
 *                                                     AR_BIT_RX_LOOPBACK (RAR15-XMC only)     
 *                                                     AR_BIT_SELECT_SRAM_MIN
 *                                                     AR_BIT_SELECT_SRAM_MAX
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL                     Success
 *                                ARS_BOARD_MUTEX                Access to the Board Lock timed-out/failed
 *                                ARS_MEMWRERR                   Memory test failure
 *                                ARS_WRAP_DROP_FAIL             Wrap Data Missing
 *                                ARS_WRAP_DATA_FAIL             Wrap Data Pattern Mismatch
 *                                ARS_WRAP_FLUSH_FAIL            Unknown External data received
 *                                ARS_INVARG                     Invalid testType parameter
 *                                ARS_INVBOARD                   Invalid board parameter
 *                                ARS_FAILURE                    Timer-deviation test failed
 *                                ARS_RX_BIT_CMD_ERROR RAR15-XMC Receiver Bit Errors
 *                                ARS_RX_BIT_CH01_ERROR -
 *                                   ARS_RX_BIT_CH32_ERROR
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: This routine performs various tests of the hardware device.
 *              While none of these tests are "built-in" per se, they do provide basic health-status of the hardware.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_execute_bit
(
   CDEV_BOARD_TYPE board,                                                   /* (input) board to reference (0 to 127) */
   CDEV_PARM_SSI_TYPE testType                                              /* (input) type of test to execute       */
)
{
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                   /* function return status                */
   CEI_INT32  accessStatus = ARS_NORMAL;                                    /* status of device read/write access    */
   CEI_UINT32 firstTimerValue;                                              /* timer deviation test value            */
   CEI_UINT32 secondTimerValue;                                             /* timer deviation test value            */
   CEI_UINT32 differenceTimerValue;                                         /* timer deviation test value            */
   CEI_UINT32 hostFirstTimerValue;                                          /* timer deviation test value            */
   CEI_UINT32 hostSecondTimerValue;                                         /* timer deviation test value            */
   CEI_UINT32 hostDifferenceTimerValue;                                     /* timer deviation test value            */
   AR_TIMETAG_TYPE boardTimeStruct;                                         /* temporary time storage                */
   CEI_UINT32 bitIndexValue;                                                /* RAR15 Rx BIT Pattern index            */
   CEI_UINT32 bitStatusValue;                                               /* RAR15 Rx BIT Status values            */
   CEI_UINT32 rxChanBitIndex;                                               /* RAR15 Rx BIT Rx Chan index            */
   CEI_UINT32 expBitStatusValue;                                            /* RAR15 Rx BIT Status values            */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                               /* ptr to brd global structure           */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                        /* ptr to board local structure          */
   CEI_UINT32 channelOffset;                                             /* chan reg region offset in the host intfc */
 
   
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      if ( (device[board].multiThreadProtect == AR_CONC_MULTIPROC) && (deviceGlobal[board]->attachedProcessCount > 1) ) {
         status = ARS_FAILURE;
      }
      else {
#endif
         switch (testType) {
            case AR_BIT_BASIC_STARTUP:
               /* Initialize the device */
               status = ar_initialize_device (board);
               break;
   
            case AR_BIT_FULL_STARTUP:
               /* Initialize the device */
               status = ar_initialize_device (board);
               /* Perform a 1 msg Internal Wrap Test on each ARINC 429 channel */
               if (status == ARS_NORMAL) {
                  status = a429_utl_wrap_test (board, LOCAL_INTERNAL_WRAP, 1U);
               }
               /* Perform a Full SRAM Memory Test (6 seconds) */
               if (status == ARS_NORMAL) {
                  status = a429_utl_memory_test (board, FULL_SRAM, 0U);
               }
               /* Initialize the device following the test */
               if (status == ARS_NORMAL) {
                  status = ar_initialize_device (board);
               }
               break;
   
            case AR_BIT_PERIODIC:
               status = a429_utl_memory_test (board, UNUSED_SRAM, 0U);
               if (status == ARS_NORMAL) {
                  /* check consecutive reads of the h/w timer to check if "alive" */
                  status = ar_get_time (board, AR_TIMETAG_INT_USEC_64BIT, &boardTimeStruct);
                  firstTimerValue = (CEI_UINT32)boardTimeStruct.timeTag;
                  cei_get_host_system_time(&hostFirstTimerValue);
      
                  ar_sleep (2);
                  
                  status = ar_get_time (board, AR_TIMETAG_INT_USEC_64BIT, &boardTimeStruct);
                  secondTimerValue = (CEI_UINT32)boardTimeStruct.timeTag;
                  
                  cei_get_host_system_time(&hostSecondTimerValue);
                  if (hostSecondTimerValue > hostFirstTimerValue) {
                     hostDifferenceTimerValue = hostSecondTimerValue - hostFirstTimerValue;
                  }
                  else {
                     hostDifferenceTimerValue = hostSecondTimerValue + 1U + (0xFFFFFFFFU - hostFirstTimerValue);
                  }
                  /* compare to host, host is in milliseconds so convert to microseconds */
                  if (hostDifferenceTimerValue > 0U) {
                     hostDifferenceTimerValue *= 1000U;
                  }
                  else {
                     hostDifferenceTimerValue = 1001U;
                  }
      
                  if (secondTimerValue > firstTimerValue) {
                     differenceTimerValue = secondTimerValue - firstTimerValue;
                  }
                  else {
                     differenceTimerValue = secondTimerValue + 1U + (0xFFFFFFFFU - firstTimerValue);
                  }
                  /* timer deviation should be between 1000 and 3000 microseconds */
                  if ( (differenceTimerValue < (hostDifferenceTimerValue - AR_BIT_TIMER_DEVIATION)) || (differenceTimerValue > (hostDifferenceTimerValue + AR_BIT_TIMER_DEVIATION)) ) {
                     status = ARS_FAILURE;
                  }
   
               }
               break;
   
            case AR_BIT_INT_LOOPBACK:
               /* Perform an Internal Wrap Test on the first ARINC 429 channel */
               status = a429_utl_wrap_test (board, LOCAL_INTERNAL_WRAP, AR_WRAP_TEST_MSG_COUNT);
               break;
   
            case AR_BIT_EXT_LOOPBACK:
               /* Perform an External Wrap Test on the first ARINC 429 channel */
               status = a429_utl_wrap_test (board, LOCAL_EXTERNAL_WRAP, AR_WRAP_TEST_MSG_COUNT);
               break;
   
            case AR_BIT_RX_LOOPBACK:
               /* Perform the ARINC 429 Receiver BIT Test on all installed ARINC 429 channels */
               deviceGlobalStructPtr = deviceGlobal[board];
               deviceLocalStructPtr = &device[board];

               if (deviceGlobalStructPtr->configuration == (CEI_UINT32)CEIDEV_CONFIG_RAR15XT ) {
   
                  for (bitIndexValue = 3U; (bitIndexValue > 0U) && (accessStatus == ARS_NORMAL) && (status == ARS_NORMAL); bitIndexValue--) {
                     /* Write the respective value to the Rx BIT command register */
                     accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_RAR15_429_RX_BIT_CMD], bitIndexValue);
                     accessStatus = cei_read_32bit  (board, &device[board].devicePtr[CEIDEV_RAR15_429_RX_BIT_CMD], &bitStatusValue);
                     /* verify Rx BIT command register readback value */
                     if (bitStatusValue != bitIndexValue) {
                        status = ARS_RX_BIT_CMD_ERROR;
                     }
                     else {
                        /* verify Rx BIT "2-bit field" self-test reporting value for each command:
                         *   CMD A/B     RX A/B
                         *     11          00
                         *     01          01
                         *     10          10
                         * channels 0-9 are encoded in the first status register, while channels 10-18 are in a second status register.
                         */
                        expBitStatusValue = (bitIndexValue < 3) ? bitIndexValue : 0U;  /* expected BIT status */

                        /* read the first status register and verify the Receiver 1->10 BIT field values */
                        accessStatus = cei_read_32bit  (board, &device[board].devicePtr[CEIDEV_RAR15_429_RX_BIT_STATE_1], &bitStatusValue);
                        for (rxChanBitIndex = 0U; 
                             (rxChanBitIndex < 10U) && (rxChanBitIndex < deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) && (accessStatus == ARS_NORMAL) && (status == ARS_NORMAL); 
                             rxChanBitIndex++) {
                           if ( (bitStatusValue & 3) != expBitStatusValue) {
                              status = (CDEV_API_RET_TYPE)(ARS_RX_BIT_CH01_ERROR + rxChanBitIndex);
                           }
                           bitStatusValue = bitStatusValue >> 2;
                        }
                        /* read the second status register and verify the Receiver 11->18 BIT field values, if installed */
                        if (accessStatus == ARS_NORMAL) {
                           accessStatus = cei_read_32bit  (board, &device[board].devicePtr[CEIDEV_RAR15_429_RX_BIT_STATE_2], &bitStatusValue);
                        }
                        for (rxChanBitIndex = 0U; 
                             (rxChanBitIndex < 8U) && ((rxChanBitIndex+10U) < deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) && (accessStatus == ARS_NORMAL) && (status == ARS_NORMAL); 
                             rxChanBitIndex++) {
                           if ( (bitStatusValue & 3) != expBitStatusValue) {
                              status = (CDEV_API_RET_TYPE)(ARS_RX_BIT_CH01_ERROR + rxChanBitIndex + 10U);
                           }
                           bitStatusValue = bitStatusValue >> 2;
                        }
                     }
                  }
                  /* reset/clear the Receiver BIT command register, and clear all receiver "Invalid ARINC 429 Message Detected" status as well */
                  accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_RAR15_429_RX_BIT_CMD], 0U);
                  ar_sleep(5);
                  channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS + (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] * CEIDEV_CHANNEL_REGION_SIZE);
                  for (rxChanBitIndex = 0; rxChanBitIndex < deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]; rxChanBitIndex++) {
                     accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], CEIDEV_CSR_RECV_FIFO_INVALID_MSG);
                     channelOffset += CEIDEV_CHANNEL_REGION_SIZE;
                  }
               }
               else {
                  status = ARS_INVARG;
               }
               break;
   
            case AR_BIT_PARTIAL_SRAM:
               /* Perform an Unused SRAM Memory Test */
               status = a429_utl_memory_test (board, UNUSED_SRAM, 0U);
               break;
   
            case AR_BIT_FULL_SRAM:
               /* Perform a Full SRAM Memory Test (6 seconds) */
               status = a429_utl_memory_test (board, FULL_SRAM, 0U);
               if (status == ARS_NORMAL) {
                  status = ar_initialize_device (board);
               }
               break;
   
            default:
               if ( (testType >= AR_BIT_SELECT_SRAM_MIN) && (testType <= AR_BIT_SELECT_SRAM_MAX) ) {
                  status = a429_utl_memory_test (board, SELECT_SRAM, (CEI_UINT32)testType);
               }
               else if (testType == (CDEV_PARM_SSI_TYPE)0x5A5A) {
                  status = a429_utl_memory_test (board, MSG_SCHEDULE_TABLE, 0U);
               }
               else {
                  status = ARS_INVARG;
               }
               break;
         }
#ifndef TARGET_EMBEDDED
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   if (accessStatus) ;
   return (status);
}


/*===================================================================================================================*
 * ENTRY POINT:                    A R _ G E T _ D E V I C E _ C O N F I G          
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the current state of protocol-generic device configuration register and board-level/API 
 *              attributes.
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE        board   -- (input)  board to reference (0 to 127)
 *              CDEV_CHAN_TYPE         channel -- (input)  channel number (0 to 31)
 *              CDEV_PARM_SSI_TYPE     item    -- (input)  selected channel config item
 *              CDEV_PARM_SSI_PTR_TYPE value   -- (output) current bit or bit field state
 * 
 *              Valid combinations of item / value parameters are:
 * 
 *              for this "item"        valid options for "value" returned are:
 * 
 *              ARU_RX_FIFO_ENABLE        AR_ON           AR_OFF
 *              ARU_TX_FIFO_ENABLE        AR_ON           AR_OFF
 *              ARU_TX_DISABLE            AR_ON           AR_OFF
 *
 *              ARU_RX_BITRATE            ARU_SPEED_LOW   ARU_SPEED_HIGH  Custom Divisor
 *              ARU_RX_PARITY             AR_ON           AR_OFF
 *              ARU_RX_MERGED_MODE        AR_ON           AR_OFF
 *              ARU_RECV_MODE             AR_WRAP_OFF     AR_WRAP_ON
 * 
 *              ARU_TX_BITRATE            ARU_SPEED_LOW   ARU_SPEED_HIGH  Custom Divisor
 *              ARU_TX_PARITY             ARU_PARITY_NONE ARU_PARITY_ODD  ARU_PARITY_EVEN
 *              ARU_TX_GAP_ERROR          AR_ON           AR_OFF
 *              ARU_TX_BIT_ERROR          AR_HI   AR_LO   AR_OFF
 *              ARU_FAST_SLEW_RATE        AR_ON           AR_OFF
 *
 *              ARU_DISCRETE_VALUES       $0000           $FFFF
 *              ARU_DISCRETE_INPUTS       $0000           $FFFF
 *              ARU_DISCRETE_IN           AR_HI           AR_LO
 *              ARU_DIGITAL_IN            AR_HI           AR_LO
 *              ARU_DISCRETE_OUT          AR_HI           AR_LO
 *              ARU_DIGITAL_OUT           AR_HI           AR_LO
 *              ARU_DIFFERENTIAL_IN       AR_HI           AR_LO
 *              ARU_DIFFERENTIAL_OUT      AR_ON           AR_OFF
 *              ARU_HW_INTERRUPT_ENABLE   AR_ON           AR_OFF       
 *
 *              ARU_IRIG_AVAILABLE        TRUE            FALSE
 *              ARU_IRIG_WRAP_ENABLE      AR_ON           AR_OFF
 *              ARU_IRIG_OUTPUT_ENABLE    AR_ON           AR_OFF
 *              ARU_IRIG_CALIBRATED
 *              ARU_IRIG_INPUT_TIME
 *              ARU_ACCESS_SNAPSHOT_BUFFER     ARU_LABEL_ONLY  ARU_LABEL_WITH_SDI
 *              ARU_RX_FIFO_COUNT         0 to 8191
 *              ARU_TX_FIFO_COUNT         0 to 8191
 *              ARU_RX_MSG_COUNT          0 to 65535
 *              ARU_TX_MSG_COUNT          0 to 65535
 *              ARU_RX_TIMETAG_MODE w/values of:                          
 *
 *                 AR_TIMETAG_EXT_IRIG_64BIT                         
 *                 AR_TIMETAG_INT_USEC_64BIT                         
 *                 AR_TIMETAG_INT_20USEC_32BIT                       
 *                 AR_TIMETAG_INT_MSEC_32BIT                         
 *                 AR_TIMER_X20_COMPAT_32BIT
 *                 AR_TIMETAG_SYNC_1553_CH1 (for combo cards only)               
 *                 AR_TIMETAG_SYNC_1553_CH2                
 *                 AR_TIMETAG_SYNC_1553_CH3                
 *                 AR_TIMETAG_SYNC_1553_CH4                
 *
 *              ARU_CHAN_COUNT_429
 *              ARU_CHAN_COUNT_573
 *              ARU_CHAN_COUNT_DISC
 *              ARU_CHAN_COUNT_DIFF
 *              ARU_CONFIGURATION
 *              ARU_DEVICE_DISABLE
 *              ARU_FW_VERSION
 *              ARU_HW_ENHANCE_CHECK
 *              ARU_HW_FPGA_TEMPERATURE
 *              ARU_HW_1PT0V_PWR_SUPPLY
 *              ARU_HW_2PT5V_PWR_SUPPLY
 *              ARU_READ_RAR_XMC_FW_BUILD_NUMBER
 *
 *              ARU_1553_CHANNEL_INITIALIZED
 *              ARU_1553_CHANNEL_STATUS     
 *              ARU_1553_DISCRETE_USE       
 *              ARU_1553_TRIGGER_USE        
 *              ARU_1553_RT_ADDRESS         
 *              ARU_1553_OPERATIONAL_MODE   
 *              ARU_1553_TIME_TAG_MODE      
 *              ARU_1553_TIME_TAG_INIT      
 *              ARU_1553_TIME_TAG_DISPLAY   
 *              ARU_1553_TIME_PERIOD        
 *              ARU_1553_INITIAL_TIME_LOW   
 *              ARU_1553_INITIAL_TIME_HIGH  
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL      Normal succesful completion.
 *                                ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                                ARS_INVARG      "item" parameter value is not supported by this function
 *                                ARS_INVHARVAL   Item was invalid or channel not supported on this device
 *                                ARS_INVBOARD    Board parameter value invalid or board has not been initialized
 * 
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: This function returns the state of the device configuration register attribute based on the combined 
 *              item/value parameter selections.  It is designed to support the ARINC 429 and all discrete and 
 *              digital I/O configuration register attributes available to the device.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_get_device_config 
( 
   CDEV_BOARD_TYPE board,                                            /* (input)  board to reference (0 to 127)       */
   CDEV_CHAN_TYPE  channel,                                          /* (input)  channel number (0 to 31)            */
   CDEV_PARM_SSI_TYPE item,                                          /* (input)  selected channel configuration item */
   CDEV_PARM_SSI_PTR_TYPE value                                      /* (output) current bit field state             */
)
{
   CEI_INT32 accessStatus;                                                  /* status of device read/write access    */
   CEI_UINT32 itemValue = 0U;                                               /* temp placeholder for return value     */
   CEI_UINT32 deviceChannel;                                                /* current channel to select             */
   CEI_UINT32 offsetIntoChannelWindow = 0U;                                 /* offset into chan mem region           */
   CEI_UINT32 bitFieldValue;                                                /* temp for device register read content */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                   /* function return status                */
   CEI_UINT32 memOffset;                                                    /* local offset to channel memory area   */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                        /* ptr to board local structure          */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                               /* ptr to brd global structure           */
#ifndef  CERTIFICATION_BUILD
   CEI_FLOAT parametricValue;                                               /* FPGA parametric working variable      */
#endif      
   AR_CHANNEL_INDEX_INFO_TYPE * channelInfoTablePtr = (AR_CHANNEL_INDEX_INFO_TYPE *)&deviceGlobal[board]->boardChannelIndexInfoStruct;

#ifdef INCLUDE_PLX_SUPPORT
   if ((board < MAX_BOARDS) && (item == ARU_HW_ENHANCE_CHECK)) {
      status = a429_utl_enh (board, PLX_ENHANCED_CHECK);
   }
   else if (VALID_BOARD_AND_INIT) {
#else   
   if (VALID_BOARD_AND_INIT) {
#endif      
      deviceChannel = (CEI_UINT32)channel;
   
      /* assign references to this board's device and deviceGlobal structure-array elements */
      deviceLocalStructPtr = &device[board];
      deviceGlobalStructPtr = deviceGlobal[board];
   
      /* Validate the channel parameter for the specific item parameter and setup the offset for the next device read operation */
      /* For most board level items, determine the return value and set the status to exit */
      switch (item) {
         case ARU_RECV_MODE:
         case ARU_RX_BITRATE:
         case ARU_RX_PARITY:
         case ARU_RX_FIFO_ENABLE:
         case ARU_RX_MERGED_MODE:
            offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
            deviceChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
            break;
      
         case ARU_TX_BIT_ERROR:
         case ARU_TX_GAP_ERROR:
         case ARU_TX_BITRATE:
         case ARU_TX_PARITY:
         case ARU_TX_FIFO_ENABLE:
         case ARU_TX_DISABLE:
         case ARU_FAST_SLEW_RATE:
            offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
            deviceChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
            break;
   
         case ARU_TX_FIFO_COUNT:
            offsetIntoChannelWindow = CEIDEV_CHANNEL_STATUS;
            deviceChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
            break;

         case ARU_RX_FIFO_COUNT:
            offsetIntoChannelWindow = CEIDEV_CHANNEL_STATUS;
            deviceChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
            break;

#ifdef INCLUDE_DISCRETE_IO
         case ARU_DISCRETE_OUT:
            deviceChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_DISC_OUT];
            offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
            break;
      
         case ARU_DISCRETE_VALUES:
         case ARU_DISCRETE_INPUTS:
            /* read and return the raw discrete input register value */
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_DISCRETE_INPUT_0], &bitFieldValue);
            *value = (CDEV_PARM_SSI_TYPE)(bitFieldValue & 0xFFFF);
            status = EXIT_ROUTINE;
            break;
      
         case ARU_DISCRETE_RAW:
            status = EXIT_ROUTINE;
            break;

         case ARU_DISCRETE_IN:
            /* read and return an individual discrete bit value from the raw input register value */
#ifndef TARGET_EMBEDDED
            status = a429_utl_validate_channel_request (board, deviceChannel, (CDEV_PARM_SSI_TYPE)item);
            if ( status == ARS_NORMAL ) {
#endif
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_DISCRETE_INPUT_0], &bitFieldValue);
               *value = (CDEV_PARM_SSI_TYPE)(((bitFieldValue >> deviceChannel) & 1) ? AR_HI : AR_LO);
               status = EXIT_ROUTINE;
#ifndef TARGET_EMBEDDED
            }
#endif
            break;
      
         case ARU_DIFFERENTIAL_IN:
            /* read and return an individual differential I/O bit value from the raw input register value */
#ifndef TARGET_EMBEDDED
            status = a429_utl_validate_channel_request (board, deviceChannel, (CDEV_PARM_SSI_TYPE)item);
            if ( status == ARS_NORMAL ) {
#endif
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_DISCRETE_INPUT_1], &bitFieldValue);
               *value = (CDEV_PARM_SSI_TYPE)(((bitFieldValue >> deviceChannel) & 1) ? AR_HI : AR_LO);
               status = EXIT_ROUTINE;
#ifndef TARGET_EMBEDDED
            }
#endif
            break;

         case ARU_DIFFERENTIAL_OUT:
            deviceChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_DIFF_OUT];
            offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
            break;
#endif

         case ARU_ACCESS_SNAPSHOT_BUFFER:
            /* read and return the snapshot storage mode bit value from the global enable register */
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], &bitFieldValue);
            *value = (CDEV_PARM_SSI_TYPE)((bitFieldValue & CEIDEV_SNAPSHOT_STORAGE_ENHANCED) ? ARU_LABEL_ONLY : ARU_LABEL_WITH_SDI);
            status = EXIT_ROUTINE;
            break;

#ifdef INCLUDE_INTERRUPTS
         case ARU_HW_INTERRUPT_ENABLE:
            /* read and return the hardware interrupt enable bit value from the global enable register */
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], &bitFieldValue);
            *value = (CDEV_PARM_SSI_TYPE)((bitFieldValue & CEIDEV_INTERRUPTS_ENABLED) ? AR_ON : AR_OFF);
            status = EXIT_ROUTINE;
            break;
#endif

         case ARU_CHAN_COUNT_429:
            *value = (CDEV_PARM_SSI_TYPE)deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX]; 
            status = EXIT_ROUTINE;
            break;

         case ARU_CHAN_COUNT_573:
            *value = (CDEV_PARM_SSI_TYPE)deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC573_TX]; 
            status = EXIT_ROUTINE;
            break;

         case ARU_CHAN_COUNT_DISC:
            *value = (CDEV_PARM_SSI_TYPE)deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_DISC_OUT]; 
            status = EXIT_ROUTINE;
            break;

         case ARU_CHAN_COUNT_DIFF:
            *value = (CDEV_PARM_SSI_TYPE)deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_DIFF_OUT]; 
            status = EXIT_ROUTINE;
            break;

         case ARU_DEVICE_DISABLE:
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], &bitFieldValue);
            *value = (CDEV_PARM_SSI_TYPE)( (bitFieldValue & CEIDEV_DEVICE_DISABLED) >> 31 );
            status = EXIT_ROUTINE;
            break;

#ifdef INCLUDE_IRIG
         case ARU_IRIG_OUTPUT_ENABLE:
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], (pCEI_UINT32)&bitFieldValue);
            *value = (CDEV_PARM_SSI_TYPE)( (bitFieldValue & CEIDEV_IRIG_830RX_IRIG_ENABLE) ? AR_ON : AR_OFF);
            status = EXIT_ROUTINE;
            break;
#endif
   
         case ARU_FW_VERSION:
            *value = (CDEV_PARM_SSI_TYPE)deviceGlobalStructPtr->firmwareVersion;
            status = EXIT_ROUTINE;
            break;
      
         case ARU_CONFIGURATION:
            *value = (CDEV_PARM_SSI_TYPE)(deviceGlobalStructPtr->configuration + deviceGlobalStructPtr->alternateJconfiguration);
            status = EXIT_ROUTINE;
            break;
      
         case ARU_SPECIAL_PROGRAMMABLE_CONFIG:
            if (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR_XMC) {
               if ( (channelInfoTablePtr->chan_type[0] == AR_CHAN_TYPE_429_RX_OR_TX) && (channelInfoTablePtr->num_channels == 16) &&
                    (channelInfoTablePtr->chan_rx_index[0] == channelInfoTablePtr->chan_tx_index[0]) ) {
                  *value = (CDEV_PARM_SSI_TYPE)(RAR_XMC_P16_SPECIAL_CFG);
               }
               else if ((channelInfoTablePtr->chan_type[16] == AR_CHAN_TYPE_429_RX_OR_TX) && (channelInfoTablePtr->num_channels == 32)) {
                  *value = (CDEV_PARM_SSI_TYPE)(RAR_XMC_16P16_SPECIAL_CFG);
               }
               else if ((channelInfoTablePtr->chan_type[16] == AR_CHAN_TYPE_429_RX_OR_TX) && (channelInfoTablePtr->num_channels == 31)) {
                  *value = (CDEV_PARM_SSI_TYPE)(RAR_XMC_15P15_SPECIAL_CFG);
               }
               else {
                  *value = (CDEV_PARM_SSI_TYPE)0;
               }
            }
            else if (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15XT) {
               if (channelInfoTablePtr->chan_type[channelInfoTablePtr->num_channels-1] == AR_CHAN_TYPE_429_RX_OR_TX) {
                  *value = (CDEV_PARM_SSI_TYPE)(RAR15_XMC_SPECIAL_CFG);
               }
               else {
                  *value = (CDEV_PARM_SSI_TYPE)0;
               }
            }
            else {
               *value = (CDEV_PARM_SSI_TYPE)0;
            }
            status = EXIT_ROUTINE;
            break;
      
         case ARU_RX_TIMETAG_MODE:
            *value = (CDEV_PARM_SSI_TYPE)deviceGlobalStructPtr->deviceTimeTagReference;
            status = EXIT_ROUTINE;
            break;

         /* return IRIG Internal Wrap State */
         #if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC) && !defined (INTEGRITY_PCI_X86) && !defined (NIVISA_PXI)
         #if defined DEBUG_PROGRAMMING
         case 0: *value=(CDEV_PARM_SSI_TYPE)baseConfig; status = EXIT_ROUTINE; break;
         #endif
         #endif
#ifdef INCLUDE_IRIG
         case ARU_IRIG_WRAP_ENABLE:
            if (deviceGlobalStructPtr->irigInstalled) {
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], (pCEI_UINT32)&bitFieldValue);
      
               if (bitFieldValue & CEIDEV_IRIG_INTERNAL_ENABLE) {
                  *value = AR_ON;
               }
               else {
                  *value = AR_OFF;
               }
               status = EXIT_ROUTINE;
            }
            else {
               status = ARS_INVHARVAL;
            }
            break;
      
         case ARU_IRIG_AVAILABLE:
            #if defined DEBUG_PROGRAMMING
               #include "irigdefs.h"
            #else
               *value = (CDEV_PARM_SSI_TYPE)deviceGlobalStructPtr->irigInstalled;
            #endif
               status = EXIT_ROUTINE;
            break;
     
         case ARU_IRIG_CALIBRATED:
            /* make sure IRIG is available on this board */
            if (deviceGlobalStructPtr->irigInstalled) {
               *value = (CDEV_PARM_SSI_TYPE)a429_utl_is_irig_calibrated(board);
               status = EXIT_ROUTINE;
            }
            else {
               status = ARS_INVHARVAL;
            }
            break;

         /* return IRIG Receiver Raw Time-code Value */
         case ARU_IRIG_INPUT_TIME:
            /* make sure IRIG is available on this board */
            if (deviceGlobalStructPtr->irigInstalled) {
#ifndef TARGET_EMBEDDED
               /* if multi-thread protection requested, semaphore/mutex-lock IRIG sample time register access */
               if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
                  status = (CDEV_API_RET_TYPE)cei_acquire_thread_lock(LOCK_IDENT);
                  if (status == ARS_NORMAL) {
#endif
                     accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_IRIG_INPUT_TIME], (pCEI_UINT32)&bitFieldValue);  
                     bitFieldValue &= 0x3FFFFFFF;
                     *value = (CDEV_PARM_SSI_TYPE)bitFieldValue;
                     status = EXIT_ROUTINE;
            
#ifndef TARGET_EMBEDDED
                     /* if multi-thread protection is active, free any acquired lock */
                     if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
                        accessStatus = cei_release_thread_lock(LOCK_IDENT);
                     }
                  }
               }
#endif
            }
            else {
               status = ARS_INVHARVAL;
            }
            break;
#endif

         case ARU_RX_MSG_COUNT:
            if (deviceChannel < deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) {
               offsetIntoChannelWindow = CEIDEV_CHANNEL_STATS_START + deviceChannel + (CEI_INT32)deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[offsetIntoChannelWindow], &bitFieldValue);
               *value = (CDEV_PARM_SSI_TYPE)bitFieldValue;
               status = EXIT_ROUTINE;
            }
            else {
               status = ARS_INVHARVAL;
            }
            break;

         case ARU_TX_MSG_COUNT:
            if (deviceChannel < deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX]) {
               offsetIntoChannelWindow = CEIDEV_CHANNEL_STATS_START + deviceChannel + (CEI_INT32)deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[offsetIntoChannelWindow], &bitFieldValue);
               *value = (CDEV_PARM_SSI_TYPE)bitFieldValue;
               status = EXIT_ROUTINE;
            }
            else {
               status = ARS_INVHARVAL;
            }
            break;

#ifndef  CERTIFICATION_BUILD
         case ARU_HW_FPGA_TEMPERATURE:
            if (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_CEI430A) {
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_FPGA_FPGA_TEMPERATURE], &bitFieldValue);
               ar_sleep (100);
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_FPGA_FPGA_TEMPERATURE], &bitFieldValue);
               ar_sleep (100);
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_FPGA_FPGA_TEMPERATURE], &bitFieldValue);
               if (bitFieldValue & 0x80) {
                  bitFieldValue &= 0x7F;
                  bitFieldValue = 127 - bitFieldValue;
                  bitFieldValue *= (CEI_UINT32)-1;
               }
               *value = (CDEV_PARM_SSI_TYPE)bitFieldValue;
            }
            else if ((deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR_XMC) || (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR_MPCIE)) {
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_FPGA_FPGA_TEMPERATURE], &bitFieldValue);
               bitFieldValue &= 0xFFF;
               parametricValue = ((CEI_FLOAT)bitFieldValue * ((CEI_FLOAT)503.975 / (CEI_FLOAT)4096.0)) - (CEI_FLOAT)273.15;
               *value = (CDEV_PARM_SSI_TYPE)(parametricValue); /* return temperature in degrees C */
            }
            else {
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_FPGA_FPGA_TEMPERATURE], &bitFieldValue);
               parametricValue = ((CEI_FLOAT)(bitFieldValue & 0xFFF) * ((CEI_FLOAT)503.975 / (CEI_FLOAT)1024.0)) - (CEI_FLOAT)273.15;
               *value = (CDEV_PARM_SSI_TYPE)((CEI_FLOAT)1000 * parametricValue); /* return temperature in millidegrees C */
            }
            status = EXIT_ROUTINE;
            break;

         case ARU_HW_1PT0V_PWR_SUPPLY:
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_FPGA_1PT0V_PWR_SUPPLY], &bitFieldValue);
            parametricValue = (CEI_FLOAT)(bitFieldValue & 0xFFF) * (CEI_FLOAT)0.00293;
            *value = (CDEV_PARM_SSI_TYPE)((CEI_FLOAT)1000 * parametricValue);  /* return 1.0V supply level in millivolts */
            status = EXIT_ROUTINE;
            break;

         case ARU_HW_2PT5V_PWR_SUPPLY:
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_FPGA_2PT5V_PWR_SUPPLY], &bitFieldValue);
            parametricValue = (CEI_FLOAT)(bitFieldValue & 0xFFF) * (CEI_FLOAT)0.00293;
            *value = (CDEV_PARM_SSI_TYPE)((CEI_FLOAT)1000 * parametricValue);  /* return 2.5V supply level in millivolts */
            status = EXIT_ROUTINE;
            break;

         case ARU_READ_RAR_XMC_FW_BUILD_NUMBER:
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_RAR_XMC_BUILD_NUMBER], &bitFieldValue);
            *value = (CDEV_PARM_SSI_TYPE)(bitFieldValue & 0x0000FFFFU);
            status = EXIT_ROUTINE;
            break;
#endif

         case ARU_MULTITHREAD_PROTECT:
            *value = (deviceLocalStructPtr->multiThreadProtect == AR_CONC_MULTITHRD) ? (CDEV_PARM_SSI_TYPE)AR_ON : (CDEV_PARM_SSI_TYPE)AR_OFF;
            status = EXIT_ROUTINE;
            break;

         case ARU_1553_CHANNEL_INITIALIZED:
         case ARU_1553_CHANNEL_STATUS:
         case ARU_1553_DISCRETE_USE:
         case ARU_1553_TRIGGER_USE:
         case ARU_1553_RT_ADDRESS:
         case ARU_1553_OPERATIONAL_MODE:
         case ARU_1553_TIME_TAG_MODE:
         case ARU_1553_TIME_TAG_INIT:
         case ARU_1553_TIME_TAG_DISPLAY:
         case ARU_1553_TIME_PERIOD:
         case ARU_1553_INITIAL_TIME_LOW:
         case ARU_1553_INITIAL_TIME_HIGH:
            if ( (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15) ||
                 (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15XT) ) {
               memOffset = CEIDEV_RAR15_SM_1553_CH_OFFSET + (channel * CEIDEV_RAR15_SM_1553_CH_SIZE) + (item - ARU_1553_CHANNEL_INITIALIZED);
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->baseAddr[memOffset], &bitFieldValue);
               *value = (CDEV_PARM_SSI_TYPE)bitFieldValue;
               status = EXIT_ROUTINE;
            }
            else {
               status = ARS_INVHARVAL;
            }
            break;

         default : status = ARS_INVARG; 
            break;
      }
      
      /* validate the specified channel */
#ifndef TARGET_EMBEDDED
      if ( status == ARS_NORMAL ) {
         status = a429_utl_validate_channel_request (board, deviceChannel, (CDEV_PARM_SSI_TYPE)item);
      }
#endif

      if ((status == ARS_NORMAL) && (deviceGlobalStructPtr != NULL) && (deviceLocalStructPtr != NULL)) {
         /* read the device register with respect to the attribute specified */
         memOffset = CEIDEV_CHANNEL_REGION_START + offsetIntoChannelWindow + (deviceChannel * CEIDEV_CHANNEL_REGION_SIZE);
         accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &bitFieldValue);
         if (accessStatus == ARS_NORMAL) {
   
            /* test the specified bit field and return the respective state */
            switch (item) {
               case ARU_RECV_MODE:
                  itemValue = (bitFieldValue & CEIDEV_INTERNAL_WRAP_ENABLE) ? AR_WRAP_ON : AR_WRAP_OFF;
                  break;
      
               case ARU_RX_BITRATE:
                  bitFieldValue &= CEIDEV_BAUD_RATE_FIELD;
                  bitFieldValue = bitFieldValue >> CEIDEV_BAUD_RATE_SHIFT;
      
                  /* flag default ARINC 429 bus speeds with API constant values, otherwise it's custom */
                  if ((deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15) ||
                      (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15XT)) {
                     if (bitFieldValue == CEIDEV_20MHZ_BAUD_RATE_100K) {
                        itemValue = ARU_SPEED_HIGH;
                     }
                     else if (bitFieldValue == CEIDEV_20MHZ_BAUD_RATE_12pt5K) {
                        itemValue = ARU_SPEED_LOW;
                     }
                     else  {
                        itemValue = (CEI_UINT32)bitFieldValue;
                     }
                  }
                  else {
                     if (bitFieldValue == CEIDEV_BAUD_RATE_100K) {
                        itemValue = ARU_SPEED_HIGH;
                     }
                     else if (bitFieldValue == CEIDEV_BAUD_RATE_12pt5K) {
                        itemValue = ARU_SPEED_LOW;
                     }
                     else  {
                        itemValue = (CEI_UINT32)bitFieldValue;
                     }
                  }
                  break;
      
               case ARU_RX_FIFO_ENABLE:
                  itemValue = (bitFieldValue & CEIDEV_CHANNEL_ENABLE) ? AR_ON : AR_OFF;
                  break;
         
#ifndef  CERTIFICATION_BUILD
               case ARU_RX_MERGED_MODE:
                  itemValue = (bitFieldValue & CEIDEV_MERGE_MODE_ENABLE) ? AR_ON : AR_OFF;
                  break;
#endif
      
               case ARU_RX_PARITY:
                  itemValue = (bitFieldValue & CEIDEV_PARITY_ENABLE) ? AR_ON : AR_OFF;
                  break;
      
               case ARU_TX_BITRATE:
                  bitFieldValue &= CEIDEV_BAUD_RATE_FIELD;
                  bitFieldValue = bitFieldValue >> CEIDEV_BAUD_RATE_SHIFT;
      
                  /* flag default ARINC 429 bus speeds with API constant values, otherwise it's custom */
                  if ((deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15) ||
                      (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15XT)) {
                     if (bitFieldValue == CEIDEV_20MHZ_BAUD_RATE_100K) {
                        itemValue = ARU_SPEED_HIGH;
                     }
                     else if (bitFieldValue == CEIDEV_20MHZ_BAUD_RATE_12pt5K) {
                        itemValue = ARU_SPEED_LOW;
                     }
                     else {
                        itemValue = (CEI_UINT32)bitFieldValue;
                     }
                  }
                  else {
                     if (bitFieldValue == CEIDEV_BAUD_RATE_100K) {
                        itemValue = ARU_SPEED_HIGH;
                     }
                     else if (bitFieldValue == CEIDEV_BAUD_RATE_12pt5K) {
                        itemValue = ARU_SPEED_LOW;
                     }
                     else {
                        itemValue = (CEI_UINT32)bitFieldValue;
                     }
                  }
                  break;
                  
#ifndef  CERTIFICATION_BUILD
               case ARU_TX_BIT_ERROR:
                  if (bitFieldValue & CEIDEV_429_TXCFG_BIT_COUNT_LO) {
                     itemValue = AR_LO;
                  }
                  else if (bitFieldValue & CEIDEV_429_TXCFG_BIT_COUNT_HI) {
                     itemValue = AR_HI;
                  }
                  else {
                     itemValue = AR_OFF;
                  }
                  break;
      
               case ARU_TX_GAP_ERROR:
                  itemValue = (bitFieldValue & CEIDEV_429_TXCFG_GAP_ERROR) ? AR_ON : AR_OFF;
                  break;
#endif
      
               case ARU_TX_PARITY:
                  if (bitFieldValue & CEIDEV_PARITY_ENABLE) {
                     itemValue = (bitFieldValue & CEIDEV_PARITY_SELECT_EVEN) ? ARU_PARITY_EVEN : ARU_PARITY_ODD;
                  }
                  else {
                     itemValue = ARU_PARITY_NONE;
                  }
                  break;
      
               case ARU_TX_FIFO_ENABLE:
                  itemValue = (bitFieldValue & CEIDEV_CHANNEL_ENABLE) ? AR_ON : AR_OFF;
                  break;
            
               case ARU_TX_DISABLE:
                  itemValue = (bitFieldValue & CEIDEV_TRANSMIT_DISABLE) ? AR_ON : AR_OFF;
                  break;
            
               case ARU_FAST_SLEW_RATE:
                  itemValue = (bitFieldValue & CEIDEV_FAST_SLEW_RATE) ? AR_ON : AR_OFF;
                  break;
            
#ifdef INCLUDE_DISCRETE_IO
               case ARU_DISCRETE_OUT:
                  itemValue = (bitFieldValue & 1) ? AR_HI : AR_LO;
                  break;
            
               case ARU_DIFFERENTIAL_OUT:
                  if (bitFieldValue & CEIDEV_CHANNEL_ENABLE) {
                     itemValue = AR_ON;
                  }
                  else {
                     itemValue = AR_OFF;
                  }
                  break;
#endif
            
               case ARU_TX_FIFO_COUNT:
                  itemValue = (bitFieldValue & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT;
                  break;

               case ARU_RX_FIFO_COUNT:
                  itemValue = (bitFieldValue & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT;
                  break;

               default : status = ARS_INVARG; 
            }

            *value = (CDEV_PARM_SSI_TYPE)itemValue;
         }
         else {
            status = (CDEV_API_RET_TYPE)accessStatus;
         }
      }
   }
   else {
      status = ARS_INVBOARD;
   }

   if (status == EXIT_ROUTINE) {
      status = ARS_NORMAL;
   }

   return (status);
}


/*===================================================================================================================*
 * ENTRY POINT:                     A R _ G E T _ E R R O R
 *===================================================================================================================*
 *
 * FUNCTION:    Function to return a message string associated with an error.
 * 
 * PARAMETERS:  CDEV_API_RET_TYPE error_num -- (input) A status value returned by any API routine
 * 
 *              pCEI_CHAR errorStringOut (LV-x30 LabVIEW usage only as an output) error description string
 * 
 * RETURN VAL:  pCEI_CHAR -- (std API) string containing a description of the associated error 
 *
 * GLOBALS:     ar_x30_errors_1[]
 *              ar_x30_errors_2[]
 *              ar_x30_bit_errors[]
 *              ar_x30_intfc_errors[]
 *              arw_x30_errors[]
 * 
 * DESCRIPTION: Many of the ARINC functions described here return status values which may indicate an error 
 *              condition.  This function is passed such an error value and returns a pointer to a message string 
 *              describing the error. 
 *
 *===================================================================================================================*/  
#ifndef CERTIFICATION_BUILD
#ifdef LABVIEW
EXPORT32 CEI_VOID  DLL_EXPORTED ar_get_error
(
   CDEV_API_RET_TYPE error_num,                                      /* (input)  error code returned by an API funct */
   pCEI_CHAR errorStringOut                                          /* (output) string description of the error     */
)
#else
EXPORT32 pCEI_CHAR DLL_EXPORTED ar_get_error
(
   CDEV_API_RET_TYPE error_num                                       /* (input)  error code returned by an API funct */
)
#endif
{   
   pCEI_CHAR error_string = NULL;

   if ( error_num == ARS_FAILURE) {
      error_string = "API Call Failure";
   }
   else if ((error_num == ARS_WINRTFAIL) || (error_num == ARS_DRIVERFAIL) || (error_num == ARS_ERR_SH_MEM_OBJ)) {
#if defined INTEGRITY_PCI_PPC || defined (INTEGRITY_PCI_X86)
      cei_get_open_error(error_string);
#else
      error_string = cei_get_open_error();
#endif
   }
   else if ( (error_num >= 0) && (error_num < MAX_X30_1ST_ERRORS) ) {
      error_string = ar_x30_errors_1[error_num];
   }
   else if ( (error_num >= ARS_INVHARCMD) && (error_num < ARS_RX_BIT_CMD_ERROR) && ((error_num - 1000) < MAX_X30_2ND_ERRORS) ) {
      error_string = ar_x30_errors_2[error_num - 1000];
   }
   else if ( (error_num >= ARS_RX_BIT_CMD_ERROR) && (error_num <= ARS_RX_BIT_CH32_ERROR) && ((error_num - ARS_RX_BIT_CMD_ERROR) < MAX_X30_BIT_ERRORS) ) {
      error_string = ar_x30_bit_errors[error_num - 1000];
   }
   else if ( (error_num > 1899) && (error_num < 1999) && ((error_num - 1900) < MAX_X30_INTFC_ERRORS) ) {
      error_string = ar_830_intfc_errors[error_num - 1900];
   }
   else if ( (error_num > 2000) && (error_num < 3000) && ((error_num - 2000) < MAX_X30_WMS_ERRORS) ) {
      error_string = arw_x30_errors[error_num - 2000];
   }
   else {
      error_string = "Invalid error passed to AR_GET_ERROR";
   }
#ifdef LABVIEW
   strcpy (errorStringOut, error_string);
#else
   return error_string;
#endif
}   
#endif


/*===================================================================================================================*
 * ENTRY POINT:                  A R _ G E T _ R X _ C H A N N E L _ S T A T U S
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the current receive channel buffer status and 
 *              fill level.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board               -- (input)  Board to reference (0 to 127)
 *              CDEV_CHAN_TYPE  channel             -- (input)  Receive channel to use (0 to 31, or 63-merged)
 *              CDEV_PARM_SI_PTR_TYPE channelStatus -- (output) Bitwise reflection of the Rx channel status register 
 *                                                              bits 0-2:
 *                                                                 AR_BUFFER_MSG_AVAILABLE     1
 *                                                                 AR_INVALID_MSG_DETECTED     2
 *                                                                 AR_BUFFER_OVERFLOW_DETECTED 4
 *              CDEV_PARM_SI_PTR_TYPE messageCount  -- (output) Buffer fill level acquired from the MSW of the status
 *                                                              register; valid if b0 is set, range is 1 - 2047.
 * 
 * RETURN VAL   CEI_UINT32 -- ARS_NORMAL     Success
 *                            ARS_INVHARVAL  Unsupported channel parameter
 *                            ARS_INVARG     NULL channelStatus or messageCount parameter value
 *                            ARS_INVBOARD   Invalid board parameter
 * 
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: This routine returns the current status of the specified receive channel buffer.  If either an 
 *              ARINC 429 protocol error or buffer overflow bit was set in the receive channel buffer status 
 *              register, it will be cleared upon return from this routine.
 *
 *===================================================================================================================*/
EXPORT32 CEI_UINT32 DLL_EXPORTED ar_get_rx_channel_status 
( 
   CDEV_BOARD_TYPE board,                                           /* (input)  board to reference (0 to 127)        */
   CDEV_CHAN_TYPE  channel,                                         /* (input)  channel to reference (0 to 32)       */
   CDEV_PARM_SI_PTR_TYPE channelStatus,                             /* (output) status bits from each Rx status reg. */
   CDEV_PARM_SI_PTR_TYPE messageCount                               /* (output) buffer fill level from ea rx st reg. */
)
{
   CEI_UINT32 bufferStatus;                                                  /* temp FIFO status by channel          */
   CEI_UINT32 memOffset;                                                     /* local offset to channel memory area  */
   CEI_INT32 status = ARS_NORMAL;                                            /* function return status               */
   CEI_INT32 accessStatus = ARS_NORMAL;                                      /* status of device read/write access   */
   
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   else if ((channelStatus == NULL) || (messageCount == NULL)) {
      status = ARS_INVARG;
   }
   else {
#endif
      if ( ((CEI_UINT32)channel < deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) ||
           ((CEI_UINT32)channel == deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_MERGED_RX]) )
         {
         memOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS + ((CEI_UINT32)channel * CEIDEV_CHANNEL_REGION_SIZE);
         /* read and report the channel status register content for the specified receiver */
         accessStatus = cei_read_32bit (board, &device[board].devicePtr[memOffset], &bufferStatus);
         if (accessStatus == ARS_NORMAL) {
            *channelStatus = (CDEV_PARM_SI_TYPE)(bufferStatus & (CEIDEV_CSR_RECV_FIFO_NOT_EMPTY+CEIDEV_CSR_RECV_FIFO_INVALID_MSG+CEIDEV_CSR_RECV_FIFO_OVERFLOW));

            /* if either an ARINC 429 protocol error or buffer overflow bit is set, clear it here */
            if (bufferStatus & (CEIDEV_CSR_RECV_FIFO_INVALID_MSG+CEIDEV_CSR_RECV_FIFO_OVERFLOW)) {
               accessStatus = cei_write_32bit (board, &device[board].devicePtr[memOffset], bufferStatus);
            }
            if (bufferStatus & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) {
               *messageCount = (CDEV_PARM_SI_TYPE)((bufferStatus & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT);
            }
            else {
               *messageCount = (CDEV_PARM_SI_TYPE)0U;
            }
         }
      }
      else {
         status = ARS_INVHARVAL;
      }
#ifndef TARGET_EMBEDDED
      if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL)) {
         status = (CEI_UINT32)accessStatus;
      }
   }
#endif
   
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                         A R _ G E T _ R X _ C O U N T
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the count of received messages on this channel
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input) Board to reference (0 to 127)
 *              CDEV_CHAN_TYPE channel  -- (input) Receive channel to use (0 to 31)
 * 
 * RETURN VAL   CEI_UINT32                   Count of ARINC messages received
 * 
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: The x30 f/w maintains a count of the number of ARINC data messages received over the interface for 
 *              each channel since the device was last initialized (see ar_open).  The API maintains a separate
 *              local copy of that count that is assigned to the board value upon invocation of this ar_clr_rx_count.
 *              This routine returns the difference between the value maintained by the API and the value
 *              read from the board upon invocation of this routine, thus providing a "running count" of the
 *              current receive message counter that can be reset (e.g. specifically for the purpose of providing
 *              legacy ARINC 429 API support).
 *
 *===================================================================================================================*/
EXPORT32 CEI_UINT32 DLL_EXPORTED ar_get_rx_count
( 
   CDEV_BOARD_TYPE board,                                                /* (input)  board to reference (0 to 127)   */
   CDEV_CHAN_TYPE channel                                                /* (input)  recv channel to ref (0 to 31)   */
)
{
   CEI_INT32  accessStatus = ARS_NORMAL;                                         /* status of dev read/write access  */
   CEI_UINT32 offsetIntoChannelWindow;                                           /* offset into chan mem region      */
   CEI_UINT32 bitFieldValue = 0;                                                 /* temp for device reg read content */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                                    /* ptr to brd global structure      */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      deviceGlobalStructPtr = deviceGlobal[board];
      if ((CEI_UINT32)channel < deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) {
         offsetIntoChannelWindow = CEIDEV_CHANNEL_STATS_START + (CEI_INT32)channel + (CEI_INT32)deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
         accessStatus = cei_read_32bit (board, &device[board].devicePtr[offsetIntoChannelWindow], &bitFieldValue);
         if (bitFieldValue >= deviceGlobalStructPtr->receiveChannelMessageCount[channel]) {
            bitFieldValue -= deviceGlobalStructPtr->receiveChannelMessageCount[channel];
         }
         else {
            bitFieldValue += (~deviceGlobalStructPtr->receiveChannelMessageCount[channel]) + 1;
         }
      }
#ifndef TARGET_EMBEDDED
   }
#endif
   
   if (accessStatus) ;
   return bitFieldValue;
}


/*===================================================================================================================*
 * ENTRY POINT:                         A R _ G E T _ T I M E
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the current board time
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board      -- (input)  Board number (0 to 127)  
 * 
 *              CDEV_PARM_SSI_TYPE format  -- (input)  Source and format of time value to be returned:
 *                                                        AR_TIMETAG_EXT_IRIG_64BIT       
 *                                                        AR_TIMETAG_INT_USEC_64BIT       
 *                                                        AR_TIMETAG_HOST_USEC_64BIT      
 *                                                        AR_TIMETAG_INT_20USEC_32BIT     
 *                                                        AR_TIMETAG_INT_MSEC_32BIT       
 *                                                        AR_TIMER_X20_COMPAT_32BIT       
 *                                                        AR_TIMETAG_SYNC_1553_CH1
 *                                                        AR_TIMETAG_SYNC_1553_CH2
 *                                                        AR_TIMETAG_SYNC_1553_CH3
 *                                                        AR_TIMETAG_SYNC_1553_CH4
 * 
 *              pAR_TIMETAG_TYPE boardTime -- (output) ->timeTagFormat -- format of timeTag member returned     
 *                                                                                                      
 *                                                     ->timeTag -- timer value in the resolution and size selected    
 *                                                                  via the "format" parameter
 *                                                                                                      
 *                                                     ->referenceTimeTag -- a 64-bit 1 microsecond indicating one of      
 *                                                                           the following:    
 *                                                                                                      
 *                                                                           1) The current raw device timer value, or       
 *                                                                           2) For IRIG, the device timer value         
 *                                                                              when the last IRIG update was received.  
 *                             
 * 
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL      Success
 *                                ARS_INVBOARD    Invalid or uninitialized board
 *                                ARS_INVARG      Invalid timeTagFormat structure member
 *                                ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: This routine returns the current time reference value as either a scaled value based on the CEI-x30 
 *              internal 64-bit timer or the most recently received IRIG timer reference, as specified via the format  
 *              parameter.  If the requested timer reference is  IRIG, the syncTime parameter will contain the 
 *              timer-referenced time-tag assigned when the IRIG time reference was received by the CEI-x30 board.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_get_time 
(
   CDEV_BOARD_TYPE board,                                                  /* (input)  board to reference (0 to 127) */
   CDEV_PARM_SSI_TYPE format,                                              /* (input)  requested format of time      */
   pAR_TIMETAG_TYPE boardTime                                              /* (output) formatted board timer struct  */
)
{
   CDEV_API_RET_TYPE status;
   TIME_TAG_TYPE hostTime;

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   else if ( format == AR_TIMETAG_HOST_USEC_64BIT ) {
#else
   if ( format == AR_TIMETAG_HOST_USEC_64BIT ) {
#endif
      status = (CDEV_API_RET_TYPE)a429_utl_acquire_formatted_time_tag ( board,
                                                                        0U,
                                                                        AR_TIMER_REGISTER_SET, 
                                                                        AR_TIMETAG_INT_USEC_64BIT, 
                                                                        AR_OFF, 
                                                                        boardTime );
      boardTime->timeTagFormat = (TIME_TAG_TYPE)AR_TIMETAG_HOST_USEC_64BIT;
      cei_get_host_system_time_x (&hostTime);
      boardTime->timeTag = hostTime;
   }
   else {
      if ( ((format == AR_TIMETAG_EXT_IRIG_64BIT) || (format == AR_TIMETAG_EXT_IRIG_REF_64BIT) || (format == AR_TIMETAG_REF_IRIG_64BIT) ||
            (format == AR_LOG_EXT_IRIG_SAMPLE)    || (format == AR_TIMETAG_EXT_IRIG_RAW)) && (deviceGlobal[board]->irigInstalled == FALSE) ) {
         status = ARS_INVHARVAL;
      }
      else {
         status = (CDEV_API_RET_TYPE)a429_utl_acquire_formatted_time_tag ( board,
                                                                           0U,
                                                                           AR_TIMER_REGISTER_SET, 
                                                                           (CEI_UINT32)format, 
                                                                           AR_OFF, 
                                                                           boardTime );
      }
   }
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                          A R _ G E T W O R D
 *===================================================================================================================*
 *
 * FUNCTION:    Read the ARINC message from a receiver FIFO buffer.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board               -- (input)  board to reference (0 to 127)
 *              CDEV_CHAN_TYPE channel              -- (input)  receive channel to use (0 to 1 less than Rx count)
 *              CDEV_PARM_VOID_PTR_TYPE destination -- (output) location to receive data
 *
 * RETURN VAL   CDEV_API_RET_TYPE ARS_GOTDATA           Success
 *                                ARS_NODATA            No data available
 *                                ARS_INVHARVAL         Protocol not available on device
 *                                ARS_INVARG            NULL data parameter value was detected
 *                                ARS_INVBOARD          Invalid board parameter or board not initialized
 *                                ARS_BOARD_MUTEX       Access to the Board Lock timed-out/failed
 *                                ARS_BAD_MESSAGE       An invalid length msg was received
 *                                ARS_RX_BUFFER_OVERRUN A receive buffer overrun was detected
 *
 * GLOBALS:     deviceGlobal[]
 * 
 * DESCRIPTION: This routine reads the most recently received ARINC message.  The channel number passed to this 
 *              routine corresponds to the ARINC 429 receive channel index, starting with zero. If the channel number 
 *              exceeds the 429 receive channel count and an ARINC 573/717 receiver exists in the board configuration
 *              the ARINC 573/717 receive channel will be used as the designated receiver.
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_getword 
( 
   CDEV_BOARD_TYPE board,                                                /* (input)  board to reference (0 to 127)   */
   CDEV_CHAN_TYPE channel,                                               /* (input)  recv channel to ref (0 to 31)   */
   CDEV_PARM_VOID_PTR_TYPE destination                                   /* (output) location to store received data */
)
{
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                                  /* function return status */
   CDEV_CHAN_TYPE localChannel;                                       /* channel reference for a429_utl_read_message */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      /* legacy ar_getword channel is 0-referenced ARINC 429 receive channel index;
       * a429_utl_read_message uses the board-referenced channel number 0-255, where
       * ARINC 429 receive channels can be located anywhere in that range (but they're not, they're indexed 0 - 31).
       */
      if (channel >= (CDEV_CHAN_TYPE)deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) {
         if (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX] != CEIDEV_MAX_CHANNELS) {
            localChannel = (CDEV_CHAN_TYPE)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX];
         }
         else {
            status = ARS_INVHARVAL;
         }
      }
      else {
         localChannel = (CDEV_PARM_SSI_TYPE)(channel + (CDEV_CHAN_TYPE)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]);
      }
      if (status == ARS_NORMAL) {
         status = a429_utl_read_message ( board, &localChannel, (CDEV_PARM_VOID_PTR_TYPE)destination, NULL );
      }

#ifndef TARGET_EMBEDDED
   }
   else {
      status = ARS_INVBOARD;
   }
#endif
   return (status);
}
           

/*===================================================================================================================*
 * ENTRY POINT:                                 A R _ G E T W O R D _ X T
 *===================================================================================================================*
 *
 * FUNCTION:    Read the ARINC word and time-tag from a receiver FIFO buffer,
 *              returning the extended time-tag structure supporting IRIG
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board    -- (input)  board to reference (0 to 127)
 *              CDEV_CHAN_TYPE channel   -- (input)  recvr channel to ref (0 to 31)
 *              pCEI_UINT32 data         -- (output) location for receive data
 *              pAR_TIMETAG_TYPE timeTagRef
 *                                       -- (input)  ->timeTagFormat structure member only.  This format defines the
 *                                                   desired resolution and format for the returned time-tag, with 
 *                                                   valid selections of:
 *                                                     AR_TIMETAG_EXT_IRIG_64BIT      0
 *                                                     AR_TIMETAG_INT_USEC_64BIT      1
 *                                                     AR_TIMETAG_HOST_USEC_64BIT     2
 *                                                     AR_TIMETAG_INT_20USEC_32BIT    3
 *                                                     AR_TIMETAG_INT_MSEC_32BIT      4
 *                                                     AR_TIMETAG_SYNC_1553_CH1       11
 *                                                     AR_TIMETAG_SYNC_1553_CH2       12
 *                                                     AR_TIMETAG_SYNC_1553_CH3       13
 *                                                     AR_TIMETAG_SYNC_1553_CH4       14
 *                                                     AR_TIMETAG_USE_PROGRAMMED_MODE 99
 *
 *                                       -- (output) location to store the acquired time-tag information
 *
 * RETURN VAL   CDEV_API_RET_TYPE ARS_GOTDATA           Success
 *                                ARS_NODATA            No data availalble
 *                                ARS_INVHARVAL         Protocol not available on device
 *                                ARS_INVARG            NULL data parameter value
 *                                ARS_INVBOARD          Invalid board parameter or board not initialized
 *                                ARS_BOARD_MUTEX       Access to the Board Lock timed-out/failed
 *                                ARS_BAD_MESSAGE       An invalid length msg was received
 *                                ARS_RX_BUFFER_OVERRUN A receive buffer overrun was detected
 *
 * GLOBALS:     deviceGlobal[]
 * 
 * DESCRIPTION: This routine reads the most recently received ARINC message and time-stamp in the requested time-tag 
 *              format from the specified receive channel buffer.  The channel number passed to this routine 
 *              corresponds to the ARINC 429 channel number index, starting with zero.
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_getword_xt
(
   CDEV_BOARD_TYPE board,                                               /* (input)  board to reference (0 to 127)    */
   CDEV_CHAN_TYPE channel,                                              /* (input)  recv channel to ref (0 to 31)    */
   pCEI_UINT32 data,                                                    /* (output) location to store received msg   */
   pAR_TIMETAG_TYPE timeTagRef                                          /* (output) location to store data time-tags */
)
{
   CDEV_CHAN_TYPE localChannel;                                                 /* merged/individual channel select  */
   CDEV_API_RET_TYPE status;                                                    /* function return status            */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      /* legacy ar_getword_xt channel is 0-referenced ARINC 429 receive channel index;
       * a429_utl_read_message uses the board-referenced channel number 0-255, where
       * ARINC 429 receive channels can be located anywhere in that range.
       */
      if (channel < (CDEV_CHAN_TYPE)(deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX])) {
#endif
         localChannel = (CDEV_PARM_SSI_TYPE)(channel + (CDEV_CHAN_TYPE)(deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]));
         status = a429_utl_read_message (board, &localChannel, data, timeTagRef);

#ifndef TARGET_EMBEDDED
      }
      else {
         status = ARS_INVHARVAL;
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   return (status);
}


/*===================================================================================================================*
 * ENTRY POINT:                               A R _ G E T B L O C K _ T
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve multiple messages from a specified receive buffer,
 *              supporting 64-bit time-tag retrieval.
 *
 * PARAMETERS:  CEI_UINT32  board       -- (input)  board to reference (0 to 127)
 *              CEI_UINT32  channel     -- (input)  recvr channel to use (0 to 31)
 *              CEI_INT32   maxMessages -- (input)  number of messages to retrieve
 *              pCEI_INT32  actualCount -- (output) number of messages retrieved
 *              pCEI_UINT32 msgChan     -- (output) location to store receive channel
 *              pCEI_INT32  data        -- (output) location to store data
 *              pCEI_INT32  timeTagMsw  -- (output) location to store time-tag MSW
 *              pCEI_INT32  timeTagLsw  -- (output) location to store time-tag LSW
 *
 * RETURN VAL:  ARS_GOTDATA           one or more ARINC messages were retrieved
 *              ARS_NODATA            no ARINC messages were available in the buffer
 *              ARS_INVBOARD          invalid board number
 *              ARS_INVHARVAL         Unsupported channel parameter
 *              ARS_INVARG            invalid parameter value
 *              ARS_BOARD_MUTEX       Access to the Board Lock timed-out/failed
 *              ARS_BAD_MESSAGE       An invalid length msg was received
 *              ARS_RX_BUFFER_OVERRUN A receive buffer overrun was detected
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: This routine retrieves all of the available ARINC messages from the requested receive channel buffer 
 *              and copies them to the desired destination.
 *
 *              If the TimeTag pointers are not NULL, the 64-bit time-tag associated with each retrieved message are 
 *              also copied, in separate 32-bit MSW and LSW arrays.
 *
 *              If the buffer is empty this routine returns ARS_NODATA.
 *
 *              If the return value for actualCount is equal to maxMessages, additional unread data may be available 
 *              in the receive buffer. 
 *
 *===================================================================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_getblock_t
(
   CEI_UINT32  board,                                                  /* (input)  board to reference (0 to 127)     */
   CEI_UINT32  channel,                                                /* (input)  recvr channel to use (0 to 31)    */
   CEI_INT32   maxMessages,                                            /* (input)  number of messages to retrieve    */
   pCEI_INT32  actualCount,                                            /* (output) number of messages retrieved      */
   pCEI_UINT32 msgChan,                                                /* (output) location to store rx chan (mrg md)*/
   pCEI_INT32  data,                                                   /* (output) location to store data            */
   pCEI_INT32  timeTagMsw,                                             /* (output) location to store time-tag MSW    */
   pCEI_INT32  timeTagLsw                                              /* (output) location to store time-tag LSW    */
)
{
   CEI_UINT32 channelStatus = 0U;                                             /* Rx channel FIFO data avail status   */
   CEI_UINT32 localChannel = 0U;                                              /* merged/individual channel select    */
   pCEI_UINT32 rxChanPtr = NULL;                                              /* local pointer to store Rx channel   */
   pCEI_INT32 dataPtr;                                                        /* local pointer to store Rx data      */
   pCEI_INT32 timeTagPtr1 = NULL;                                             /* local pointer to store Rx timetag   */
   pCEI_INT32 timeTagPtr2 = NULL;                                             /* local pointer to store Rx timetag   */
   CEI_INT32 messageCount;                                                    /* receiver buffer message counter     */
   CEI_UINT32 mergedModeState = AR_OFF;                                       /* 429 Rx ch merged/individual state   */
   CEI_UINT32 mergedModeChan = 0U;                                            /* 429 Rx ch merged/individual state   */
   AR_TIMETAG_TYPE timeTagRef;                                                /* local time-tag reference structure  */
   CEI_INT32 recvStatus = ARS_NORMAL;                                         /* FIFO receive status                 */
#ifndef TARGET_EMBEDDED
   CEI_INT32 lockStatus = ARS_NORMAL;                                         /* status of board access lock acquis. */
#endif
   CEI_INT32 status = ARS_NORMAL;                                             /* function return status              */
   CEI_INT32 retrieveTimeTag = FALSE;                                         /* indicates to pass time-tags back    */
   CEI_INT32 retrieveChannel = FALSE;                                         /* indicates to pass channel back      */
   CEI_INT32 accessStatus = ARS_NORMAL;                                       /* status of device read/write access  */
   CEI_UINT32 timeTagReference = 0;                                           /* local current time-tag reference    */
   CEI_UINT32 chanOffset;                                                     /* local offset to channel memory area */
   CEI_UINT32 dataOffset;                                                     /* local offset to channel memory area */
   CEI_UINT32 memOffset;                                                      /* local offset to channel memory area */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr = NULL;                   /* pointer to board local structure    */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr = NULL;                          /* pointer to board global structure   */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   else if ((data == NULL) || (actualCount == NULL) || (maxMessages < 0)) {
      status = ARS_INVARG;
   }
   else if (channel < deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) {
#endif
      /* initialize local return message count and board pointers */
      *actualCount = 0;
      timeTagPtr1 = timeTagLsw;
      timeTagPtr2 = timeTagMsw;
      rxChanPtr = msgChan;
      
      deviceGlobalStructPtr = deviceGlobal[board];
      deviceLocalStructPtr = &device[board];

      /* if External IRIG time-tagging is selected, convert IRIG once per block */
      timeTagReference = deviceGlobalStructPtr->deviceTimeTagReference;
      if (timeTagReference == AR_TIMETAG_EXT_IRIG_64BIT) {
         timeTagReference = AR_TIMETAG_REF_IRIG_64BIT;
      }
      
      localChannel = channel + deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
      
      /* Check to see which Receive Mode is active for this channel, Individual or Merged */
      if (deviceGlobalStructPtr->receiveChannelBufferMode[localChannel] == ARU_MERGED) {
         localChannel = (CEI_UINT32)deviceGlobalStructPtr->lastChannel[CEIDEV_CH_TYPE_MERGED_RX];
         mergedModeState = AR_ON;
      }
      else {
         mergedModeChan = CEIDEV_MAX_CHANNELS;
      }
      
#ifndef TARGET_EMBEDDED
      status = a429_utl_validate_channel_request ((CDEV_BOARD_TYPE)board, localChannel, ARU_GET_DATA);
   }
   else {
      status = ARS_INVHARVAL;
   }

   if ((accessStatus == ARS_NORMAL) && (lockStatus == ARS_NORMAL) && (status == ARS_NORMAL) && (deviceGlobalStructPtr != NULL) && (deviceLocalStructPtr != NULL)) {
      /* if multi-thread protection requested, acquire semaphore/mutex-locking */
      if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
         lockStatus = cei_acquire_thread_lock(LOCK_IDENT);
      }
      if (lockStatus == ARS_NORMAL) {
#endif
         /* check the receive channel status and assign return status in the following priority order...
          * If it indicates "BUFFER OVERRUN", save and return a status indicating a buffer overrun was encountered
          * If it indicates "BAD MESSAGE RECEIVED", save and return a status indicating an invalid 429 message was previously detected
          * If it indicates "NOT EMPTY", then retrieve a message and optionally it's time-stamp from the buffer. 
          */
         memOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS + (localChannel * CEIDEV_CHANNEL_REGION_SIZE);
         accessStatus = cei_read_32bit ((CDEV_BOARD_TYPE)board, &deviceLocalStructPtr->devicePtr[memOffset], &channelStatus);
         if (accessStatus == ARS_NORMAL) {
            if (channelStatus & CEIDEV_CSR_RECV_FIFO_OVERFLOW) {
               recvStatus = ARS_RX_BUFFER_OVERRUN;
               accessStatus = cei_write_32bit ((CDEV_BOARD_TYPE)board, &deviceLocalStructPtr->devicePtr[memOffset], CEIDEV_CSR_RECV_FIFO_OVERFLOW);
            }
            else if (channelStatus & CEIDEV_CSR_RECV_FIFO_INVALID_MSG) {
               recvStatus = ARS_BAD_MESSAGE;
               accessStatus = cei_write_32bit ((CDEV_BOARD_TYPE)board, &deviceLocalStructPtr->devicePtr[memOffset], CEIDEV_CSR_RECV_FIFO_INVALID_MSG);
            }
         }
         if ((accessStatus == ARS_NORMAL) && (channelStatus & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY)) {
            /* To avoid having to poll the channel status register for each message read, read
             * how many unread entries exist in the respective buffer from the channel status register.
             */
            messageCount = (CEI_INT32)((channelStatus & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT);
            if (maxMessages > messageCount) {
               maxMessages = messageCount;
            }
   
            /* reference supplied pointers locally */
            dataPtr = data;
            if ( (timeTagLsw != NULL) && (timeTagMsw != NULL) ) {
               retrieveTimeTag = TRUE;
            }
            if (msgChan != NULL) {
               retrieveChannel = TRUE;
            }
   
            /* If the reference is valid and the active Time-Tag Mode is IRIG, 
             * read the time-stamp in a two-step process. First, read the current
             * IRIG sample time value, then extrapolate the individual received 
             * message time-stamps to IRIG-referenced time as they are read.
             * This saves time by not having to read and convert the IRIG sample
             * for each received message time-stamp extrapolation.
             */
            memOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_BUFFER_0 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE);
   
            if ((retrieveTimeTag) && (timeTagReference == AR_TIMETAG_REF_IRIG_64BIT)) {
               accessStatus = a429_utl_acquire_formatted_time_tag ((CDEV_BOARD_TYPE)board, 
                                                                    memOffset,
                                                                    AR_BYPASS_TIMER_READ, 
                                                                    AR_LOG_EXT_IRIG_SAMPLE, 
                                                                    mergedModeChan, 
                                                                    &timeTagRef );
            }
            
            /* setup all offsets that might be used in the message/time-stamp retreival loop */
            chanOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_BUFFER_2 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE);
            dataOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_BUFFER_3 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE);
   
            /* attempt to retrieve as much data as requested */
            messageCount = 0;
            while ((messageCount < maxMessages) && (accessStatus == ARS_NORMAL)) {
               /* if merged mode is enabled for this channel read the actual 
                * channel on which the message was received from the Rx buffer */
               if (mergedModeState == AR_ON) {
                  accessStatus = cei_read_32bit ((CDEV_BOARD_TYPE)board, &deviceLocalStructPtr->devicePtr[chanOffset], (pCEI_UINT32)&mergedModeChan);
               }
   
               if (retrieveTimeTag) {
                  accessStatus = a429_utl_acquire_formatted_time_tag ((CDEV_BOARD_TYPE)board, 
                                                                      memOffset,
                                                                      AR_RECEIVER_FIFO_TIMETAG,
                                                                      timeTagReference, 
                                                                      mergedModeChan, 
                                                                      &timeTagRef );
                  if (accessStatus == ARS_NORMAL) {
                     /* return the 64-bit time-tag as two 32-bit values */
                     pCEI_UINT32 timeStampPtr = (pCEI_UINT32)&timeTagRef.timeTag;
                     *timeTagPtr1 = *timeStampPtr;
                     timeStampPtr++;
                     *timeTagPtr2 = *timeStampPtr;
                     timeTagPtr1++;
                     timeTagPtr2++;
                  }
               }
               if (retrieveChannel) {
                  /* return the channel identifier from the FIFO */
                  *rxChanPtr = (mergedModeState == AR_ON) ? mergedModeChan: channel;  
                  rxChanPtr++;
               }
   
               /* return the 32-bit data word */
               accessStatus = cei_read_32bit ((CDEV_BOARD_TYPE)board, &deviceLocalStructPtr->devicePtr[dataOffset], (pCEI_UINT32)dataPtr);
               dataPtr++;
               messageCount++;
            }
            
            if (messageCount) {
               status = ARS_GOTDATA;
               *actualCount = messageCount;
            }
         }
         else { /* no data available, exit with status = ARS_NODATA */
            status = ARS_NODATA;
         }
#ifndef TARGET_EMBEDDED
         /* if multi-thread protection is active, free any acquired lock */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            lockStatus = cei_release_thread_lock(LOCK_IDENT);
            if (lockStatus != ARS_NORMAL) {
               status = lockStatus;
            }
         }
      }
      else {
         status = lockStatus;
      }
   }
   if ( ((status == ARS_NORMAL) || (status == ARS_NODATA)) && (accessStatus != ARS_NORMAL)) {
      status = accessStatus;
   }
   if ((recvStatus != ARS_NORMAL) && (accessStatus == ARS_NORMAL) && (lockStatus == ARS_NORMAL)) {
#else
   if ((recvStatus != ARS_NORMAL) && (accessStatus == ARS_NORMAL)) {
#endif
      status = recvStatus;
   }
   
   return (status);
}


/*===================================================================================================================*
 * ENTRY POINT:                                A R _ G E T _ L A T E S T _ T
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the latest message & time-stamp for the specified channel/label combination from 
 *              the snapshot buffer
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board    -- (input)  Board number (0 to 127)  
 *              CDEV_CHAN_TYPE channel   -- (input)  429 Rx channel selection, valid range is 0 to one less than the 
 *                                                   number of ARINC 429 channels installed
 *              CDEV_PARM_USI_TYPE label -- (input)  Label of interest
 *              pCEI_UINT32 data         -- (output) pointer to the host data storage
 *              TIME_TAG_TYPE timeTag    -- (output) pointer to time-stamp storage
 * 
 * 
 * RETURN VAL:  ARS_NORMAL        Operation completed successfully
 *              ARS_INVBOARD      Invalid board number or board not initialized
 *              ARS_INVHARVAL     Invalid receive channel argument
 *              ARS_INVARG        Invalid label argument, null data ptr
 *              ARS_BOARD_MUTEX   Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 *
 * DESCRIPTION: This routine copies the latest ARINC 492 message received for the specified channel/label comb. 
 *              from the snapshot buffer.  When using this routine, the host application should set the
 *              snapshot storage mode to "label field only".  This will setup the device to store snapshot data 
 *              based on the label field value only, ignoring the SDI bit field value.
 *
 *              The snapshot buffer for a single channel is defined as follows:
 *
 *                 CEI_UCHAR snapshot[MAX_SDI][MAX_LABEL]
 *
 *              where the bits of the ARINC word are split up as follows:
 *
 *                     field   -SDI-  --------label---------
 *                      bits    9 8   7  6  5  4  3  2  1  0
 *                    
 *===================================================================================================================*/ 
EXPORT32 CEI_INT32 DLL_EXPORTED ar_get_latest_t 
(
   CDEV_BOARD_TYPE board,                                              /* (input)  Board number (0 to 127)           */
   CDEV_CHAN_TYPE channel,                                             /* (input)  429 Rx channel selection          */
   CDEV_PARM_USI_TYPE label,                                           /* (input)  Label of interest                 */
   pCEI_UINT32 data,                                                   /* (output) pointer to the host data storage  */
   TIME_TAG_TYPE * timeTag                                             /* (output) pointer to time-stamp storage     */
)
{
   CEI_UINT32 snapStorageMode;                                                 /* local snapshot storage selection   */
   CEI_UINT32 localChannel = 0U;                                               /* merged/individual channel select   */
   CEI_UINT32 localLabel;                                                      /* local label of interest            */
   CEI_INT32  status = ARS_NORMAL;                                             /* function return status             */
   CEI_INT32  accessStatus = ARS_NORMAL;                                       /* status of device read/write access */
   AR_TIMETAG_TYPE timeTagRef;                                                 /* structure for the 64-bit time-tag  */
   CEI_UINT32 memOffset;                                                     /* local offset to channel memory area  */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                                 /* ptr to board local structure */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                                       /* ptr to board global structure */

   /* assign references to this board's device and deviceGlobal structure-array elements */
   deviceLocalStructPtr = &device[board];
   deviceGlobalStructPtr = deviceGlobal[board];

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   /* Snapshot Buffer access requests are limited to the number of ARINC 429 receive channels installed on the device */
   else if ( ((CEI_UINT32)channel > deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) || (data == NULL) ) {
      status = ARS_INVARG;
   }
   else {
#endif
      localChannel = (CEI_UINT32)channel + deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
      
#ifndef TARGET_EMBEDDED
      status = a429_utl_validate_channel_request (board, localChannel, ARU_ACCESS_SNAPSHOT_BUFFER);
      if (status == ARS_NORMAL) {
         /* if multi-thread protection requested, acquire semaphore/mutex-locking */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (accessStatus == ARS_NORMAL) {
#endif
            /* read the current snapshot storage mode selection: label only or label/SDI combined */
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], &snapStorageMode);
   
            /* this routine must only be invoked when the snapshot storage mode selection is label only */
            if ((snapStorageMode & CEIDEV_SNAPSHOT_STORAGE_ENHANCED) == 0U) {
               status = ARS_INVARG;
            }
            else {
               localLabel = (CEI_UINT32)(label & 0x00FFU);
               memOffset = CEIDEV_SNAPSHOT_BUFFER_START + (CEI_UINT32)(localChannel << CEIDEV_SNAPSHOT_BUFFER_CH_SHIFT) + (localLabel << CEIDEV_SNAPSHOT_BUFFER_LB_SHIFT);
               if (timeTag != NULL) {
                  accessStatus = a429_utl_acquire_formatted_time_tag ( board,
                                                                       memOffset,
                                                                       AR_RECEIVER_FIFO_TIMETAG,
                                                                       AR_USE_PROGRAMMED_TIMETAG_MODE, 
                                                                       CEIDEV_MAX_CHANNELS, 
                                                                       &timeTagRef );
                  *timeTag = timeTagRef.timeTag;
               }
               memOffset += CEIDEV_SNAPSHOT_BUFFER_DATA_OFF;
               accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], (pCEI_UINT32)data);
            }
#ifndef TARGET_EMBEDDED
            /* if multi-thread protection is active, free any acquired lock */
            if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
               accessStatus = cei_release_thread_lock(LOCK_IDENT);
            }
         }
         
         if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL)) {
            status = accessStatus;
         }
      }
   }
#endif
   
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                                  A R _ G E T _ S N A P _ D A T A
 *===================================================================================================================*
 *
 * FUNCTION:    Retrieve the latest message received for the specified channel/label/SDI combination from the  
 *              snapshot buffer.
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board    -- (input)  Board number (0 to 127)  
 *              CDEV_CHAN_TYPE channel   -- (input)  429 Rx channel selection, valid range is 0 to one less
 *                                                   than the number of  ARINC 429 channels installed
 *              CDEV_PARM_USI_TYPE label -- (input)  Label of interest
 *              CDEV_PARM_USI_TYPE sdi   -- (input)  SDI Field Value of interest
 *              pCEI_INT32 data          -- (output) pointer to the host data storage
 * 
 * RETURN VAL:  ARS_NORMAL        Operation completed successfully
 *              ARS_INVBOARD      Invalid board number or board not initialized
 *              ARS_INVHARVAL     Invalid receive channel argument
 *              ARS_INVARG        Invalid label or SDI argument, null data ptr
 *              ARS_BOARD_MUTEX   Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 *
 * DESCRIPTION: This routine copies the latest ARINC 492 message received for the specified channel/label/SDI 
 *              combination from the snapshot buffer.  When using this routine, the host application should set the
 *              snapshot storage mode to "label/sdi storage".  This will setup the device to store snapshot data 
 *              based on the label field value in combination with the SDI bit field value.
 *
 *              The snapshot buffer for a single channel is defined as follows:
 *
 *                 CEI_UCHAR snapshot[MAX_SDI][MAX_LABEL]
 *
 *              where the bits of the ARINC word are split up as follows:
 *
 *                     field   -SDI-  --------label---------
 *                      bits    9 8   7  6  5  4  3  2  1  0
 *                    
 *===================================================================================================================*/ 
EXPORT32 CEI_INT32 DLL_EXPORTED ar_get_snap_data 
(
 CDEV_BOARD_TYPE board,                                                /* (input)  Board number (0 to 127)           */
 CDEV_CHAN_TYPE channel,                                               /* (input)  429 Rx channel selection          */
 CDEV_PARM_USI_TYPE label,                                             /* (input)  Label of interest                 */
 CDEV_PARM_USI_TYPE sdi,                                               /* (input)  SDI Field Value of interest       */
 pCEI_UINT32 data                                                      /* (output) pointer to the host data storage  */
)
{
   CEI_UINT32 snapshotBufferIndex;                                        /* snapshot buffer offset label/sdi        */
   CEI_UINT32 snapStorageMode;                                            /* local snapshot storage selection        */
   CEI_UINT32 localChannel = 0U;                                          /* merged/individual channel select        */
   CEI_UINT32 localLabel;                                                 /* local label value of interest           */
   CEI_UINT32 localSdi;                                                   /* local SDI field value of interest       */
   CEI_UINT32 memOffset;                                                  /* local offset to channel memory area     */
   CEI_INT32 status = ARS_NORMAL;                                         /* function return status                  */
   CEI_INT32 accessStatus = ARS_NORMAL;                                   /* status of device read/write access      */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr = NULL;               /* board local structure pointer           */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   /* Snapshot Buffer access requests are limited to the number of ARINC 429 receive channels installed on the device */
   else if ( ((CEI_UINT32)channel > deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) || (data == NULL) ) {
      status = ARS_INVARG;
   }
   else {
#endif
      deviceLocalStructPtr = &device[board];
      localChannel = (CEI_UINT32)channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
      
#ifndef TARGET_EMBEDDED
      status = a429_utl_validate_channel_request (board, localChannel, ARU_ACCESS_SNAPSHOT_BUFFER);
      if (status == ARS_NORMAL) {
         /* if multi-thread protection requested, acquire semaphore/mutex-locking */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (accessStatus == ARS_NORMAL) {
#endif
            localLabel = (CEI_UINT32)(label & 0x00FFU);
            localSdi = (CEI_UINT32)(sdi & 3U);
            
            /* read the current snapshot storage mode selection: label only or label/SDI combined */
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], &snapStorageMode);
   
            if (snapStorageMode & CEIDEV_SNAPSHOT_STORAGE_ENHANCED) { 
               snapshotBufferIndex = (localLabel << CEIDEV_SNAPSHOT_BUFFER_LB_SHIFT) + CEIDEV_SNAPSHOT_BUFFER_DATA_OFF;
            }
            else {
               snapshotBufferIndex = (localLabel << CEIDEV_SNAPSHOT_BUFFER_LB_SHIFT) + localSdi;
            }

            snapshotBufferIndex += localChannel << CEIDEV_SNAPSHOT_BUFFER_CH_SHIFT;
            memOffset = CEIDEV_SNAPSHOT_BUFFER_START + snapshotBufferIndex;
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], data);
            
#ifndef TARGET_EMBEDDED
            /* if multi-thread protection is active, free any acquired lock */
            if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
               accessStatus = cei_release_thread_lock(LOCK_IDENT);
            }
         }
         
         if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL)) {
            status = accessStatus;
         }
      }
   }
#endif
   
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                                       A R _ G O
 *===================================================================================================================*
 *
 * FUNCTION:    Enable the transmitter and receiver FIFOs on the device
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board  -- (input) board to reference (0 to 127)
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL      Success
 *                                ARS_INVBOARD    Invalid board parameter or board not initialized
 *                                ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: Sets the global enable register global enable bit to ENABLE, enabling the Message Scheduler and all
 *              circular buffer to external bus operations.
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_go
(
   CDEV_BOARD_TYPE board                                                   /* (input)  board to reference (0 to 127) */
)
{
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                      /* function return status             */
   CEI_INT32 accessStatus = ARS_NORMAL;                                        /* status of device read/write access */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      /* enable "global enable" */
      accessStatus = cei_write_32bit (board, 
                                      &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], 
                                      CEIDEV_GLOBAL_ENABLE_AUTON_SET + CEIDEV_GLOBAL_ENABLE_ON);
#ifndef TARGET_EMBEDDED
      status = (CDEV_API_RET_TYPE)accessStatus;
   }
   else
      status = ARS_INVBOARD;
#endif

   return (status);
}


/*===================================================================================================================*
 * ENTRY POINT:                              A R _ I N I T I A L I Z E _ A P I
 *===================================================================================================================*
 *
 * FUNCTION:    Initialize the API local variables
 * 
 * PARAMETERS:  CEI_INT32 board   -- (input) device number to open (0-127)
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL            Success
 *                                ARS_DRIVERFAIL        Board session open failed
 *                                ARS_BOARD_MUTEX       <Windows only> Creation of the Board Lock mutex failed
 *                                ARS_INVBOARD          Invalid board parameter
 *                                ARS_BADLOAD           FPGA load process failed
 *                                ARS_HW_DETECT         Supported device not detected
 *                                ARS_FW_NOT_SUPPORTED  Device firmware revision not supported
 *                                ARS_FAILURE           <Windows only> CEI_Install library failed to load in memory
 *                                                      <Linux only> Creation of the Board Lock semaphore failed
 *                                                      <VxWorks only> Legacy PCI driver open session failure
 *                                API_VXB_METHOD_ERR    <VxWorks only> VxBus open session failure
 * GLOBALS:     device[]
 *              deviceGlobal[]
 *                                    
 * DESCRIPTION: This routine opens a session to the device, loads firmware if a PLX interface is detected, and 
 *              defines the API I/O register pointers.  This routine should not be directly invoked without first 
 *              invoking ar_open.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_initialize_api
(
   CDEV_BOARD_TYPE board                                                   /* (input)  board to reference (0 to 127) */
)
{
   CEI_UINT32 globalEnableRegValue;                                            /* Global Enable register contents    */
   CEI_UINT32 boardConfiguration;                                              /* device channel configuration       */
   CEI_UINT32 fwVersion;                                                       /* firmware ver. programmed on device */
   CEI_INT32  vendorIdent;                                                     /* PCI Vendor ID for opened device    */
   CEI_INT32  deviceIdent;                                                     /* PCI Device ID for opened device    */
   CEI_INT32  status = ARS_NORMAL;                                             /* OS-specific function return status */
   CEI_INT32  altStatus = ARS_NORMAL;                                          /* OS-specific function return status */
#if !defined (NIVISA_PXI) && defined (INCLUDE_PLX_SUPPORT)
   CEI_UINT32 interruptBit;                                                     /* the respective PLX Int Enable bit */
#endif   
   CEI_UINT32 comboBoardChanCount;                                              /* combo board 1553 channel count    */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                                /* ptr to board local structure  */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                                       /* ptr to board global structure */

#ifndef TARGET_EMBEDDED
   if ( (board < MAX_BOARDS) && (deviceGlobal[board] != NULL) ) {
#endif
      /*-------------------------------------------------------------------*
       * Attempt to open a session with the device.
       *-------------------------------------------------------------------*/
      status = cei_acquire_device (board, (ARINC_LOCAL_TYPE *)&device[board]);
      
      /* if successful, validate the device configuration is that of an x30 supported board */
      if (status == ARS_NORMAL) { 
         status = cei_acquire_device_identity ((CEI_INT32)board, &vendorIdent, &deviceIdent);
         if ((status == ARS_FAILURE) || (status == ARS_INVBOARD))  {
            /* could not identify the board at all... close and exit */
            altStatus = cei_release_device(board, (ARINC_LOCAL_TYPE *)&device[board]);
         }
         else {  
            /* assign a reference to this board's device structure element */
            deviceLocalStructPtr = &device[board];
            deviceGlobalStructPtr = deviceGlobal[board];

            /*-----------------------------------------------------------------------------------------*
             * Assure this session is either the master in a multi-process app or a single process app 
             *-----------------------------------------------------------------------------------------*/
            if ((deviceGlobalStructPtr->attachedProcessCount == 0) || (deviceLocalStructPtr->multiThreadProtect == AR_CONC_NONE)) {

               /*-------------------------------------------------------------------------------------------------*
                * For legacy PCI devices, assure the device is programmed with enhanced firmware (not old std f/w)
                *-------------------------------------------------------------------------------------------------*/
               if (vendorIdent == PCI_X30_MANUF_ID) {
#ifdef INCLUDE_PLX_SUPPORT
#if defined (_WIN32) || defined (_LINUX_X86_) || defined (_LINUX_PPC_)
                  if ((deviceIdent != PCI_DEVICE_ID_RAR_PCIE) && 
                      (deviceIdent != PCI_DEVICE_ID_RAR_XMC) && 
                      (deviceIdent != PCI_DEVICE_ID_RAR_MPCIE) && 
                      (deviceIdent != PCI_DEVICE_ID_RAR15) && 
                      (deviceIdent != PCI_DEVICE_ID_RAR15XT)) {
                     status = a429_utl_enh_check (board, deviceIdent);
                  }
#endif
#endif
                  /*-------------------------------------------------------------------------------------*
                   * If this is an uninitialized PCI device, load the FPGA and check board configuration.
                   *-------------------------------------------------------------------------------------*/
                  if (status == ARS_NORMAL) { 
                     switch (deviceIdent) { /* load routine is dependent on the PLX PCI interface */
#ifdef INCLUDE_PLX_SUPPORT
                        case PCI_DEVICE_ID_830:
                           status = a429_utl_program_xilinx_9056 ( board,
                                                                   deviceIdent, 
                                                                   (CEI_UINT32)sizeof(fpga_830), 
                                                                   (pCEI_UCHAR)fpga_830 );
                           break;
         
                        case PCI_DEVICE_ID_830RX:
                           status = a429_utl_program_xilinx_9056 ( board,
                                                                   deviceIdent, 
                                                                   (CEI_UINT32)sizeof(fpga_830_rx), 
                                                                   (pCEI_UCHAR)fpga_830_rx );
                           break;
         
                        case PCI_DEVICE_ID_830X820:
                           status = a429_utl_program_xilinx_9056 ( board,
                                                                   deviceIdent, 
                                                                   (CEI_UINT32)sizeof(fpga830x820), 
                                                                   (pCEI_UCHAR)fpga830x820 );
                           break;
         
                        case PCI_DEVICE_ID_RCEI830A:
                           status = a429_utl_program_xilinx_9056 ( board,
                                                                   deviceIdent, 
                                                                   (CEI_UINT32)sizeof(fpga_830a), 
                                                                   (pCEI_UCHAR)fpga_830a );
                           break;
         
                        case PCI_DEVICE_ID_630:
                           status = a429_utl_program_xilinx_9056_32bit ( board,
                                                                         deviceIdent, 
                                                                         (CEI_UINT32)((sizeof(fpga_630))/4), 
                                                                         (pCEI_UINT32)fpga_630 );
                           break;
         
                        case PCI_DEVICE_ID_530:
                           status = a429_utl_program_xilinx_9056_32bit ( board,
                                                                         deviceIdent, 
                                                                         (CEI_UINT32)((sizeof(fpga_530))/4), 
                                                                         (pCEI_UINT32)fpga_530 );
                           break;
#ifndef NIVISA_PXI
                        case PCI_DEVICE_ID_A30:
                           status = a429_utl_program_xilinx_9056_32bit ( board,
                                                                         deviceIdent, 
                                                                         (CEI_UINT32)((sizeof(fpga_A30))/4), 
                                                                         (pCEI_UINT32)fpga_A30 );
                           break;
         
                        case PCI_DEVICE_ID_RAREC:
                           status = cei_read_32bit (board, &deviceLocalStructPtr->baseAddr[PLX9030_CFG_INT_CTRL_STAT_REG], &interruptBit);
                           interruptBit &= ~PLX9030_LOCAL_INT_INP_ENABLE;
                           status = cei_write_32bit (board, &deviceLocalStructPtr->baseAddr[PLX9030_CFG_INT_CTRL_STAT_REG], interruptBit);
                           status = a429_utl_program_xilinx_9030_32bit ( board,
                                                                         deviceIdent, 
                                                                         (CEI_UINT32)(sizeof(fpga_ec)/4), 
                                                                         (pCEI_UINT32)fpga_ec );
                           break;
         
                        case PCI_DEVICE_ID_430:
                           status = cei_read_32bit (board, &deviceLocalStructPtr->baseAddr[PLX9030_CFG_INT_CTRL_STAT_REG], &interruptBit);
                           interruptBit &= ~PLX9030_LOCAL_INT_INP_ENABLE;
                           status = cei_write_32bit (board, &deviceLocalStructPtr->baseAddr[PLX9030_CFG_INT_CTRL_STAT_REG], interruptBit);
                           status = a429_utl_program_xilinx_9030 ( board,
                                                                   deviceIdent, 
                                                                   (CEI_UINT32)sizeof(fpga_430), 
                                                                   (pCEI_UCHAR)fpga_430 );
                           break;
         
                        case PCI_DEVICE_ID_430A:
                           status = cei_read_32bit (board, &deviceLocalStructPtr->baseAddr[PLX9030_CFG_INT_CTRL_STAT_REG], &interruptBit);
                           interruptBit &= ~PLX9030_LOCAL_INT_INP_ENABLE;
                           status = cei_write_32bit (board, &deviceLocalStructPtr->baseAddr[PLX9030_CFG_INT_CTRL_STAT_REG], interruptBit);
                           status = a429_utl_program_xilinx_9030 ( board,
                                                                   deviceIdent, 
                                                                   (CEI_UINT32)sizeof(fpga_430A), 
                                                                   (pCEI_UCHAR)fpga_430A );
                           break;
#endif
#endif
                        /* PCI Express device firmware is loaded from flash on power-up, host interface is in BAR0, so fake it */
                        case PCI_DEVICE_ID_RAR_PCIE:
                        case PCI_DEVICE_ID_RAR_XMC:
                        case PCI_DEVICE_ID_RAR_MPCIE:
                           deviceLocalStructPtr->devicePtr = deviceLocalStructPtr->baseAddr;
                           break;

                        case PCI_DEVICE_ID_RAR15:
                        case PCI_DEVICE_ID_RAR15XT:
                           deviceLocalStructPtr->devicePtr = &deviceLocalStructPtr->baseAddr[CEIDEV_RAR15_429_OFFSET];
                           break;
         
                        default:
                           status = ARS_HW_DETECT;
                           break;
                     }
                  }
               }
               else {  /* if (vendorIdent == PCI_X30_MANUF_ID) is false, this is not a Abaco Systems Avionics / Condor board */
                  status = ARS_HW_DETECT;
               }
            }
            else if (vendorIdent == PCI_X30_MANUF_ID) { /* for secondary multi-process situations, bypassing f/w load */
               /* PCI Express device firmware is loaded from flash on power-up, host interface is in BAR0, so fake it */
               switch (deviceIdent) {
                  case PCI_DEVICE_ID_RAR_PCIE:
                  case PCI_DEVICE_ID_RAR_XMC:
                  case PCI_DEVICE_ID_RAR_MPCIE:
                     deviceLocalStructPtr->devicePtr = deviceLocalStructPtr->baseAddr;
                     break;

                  case PCI_DEVICE_ID_RAR15:
                  case PCI_DEVICE_ID_RAR15XT:
                     deviceLocalStructPtr->devicePtr = &deviceLocalStructPtr->baseAddr[CEIDEV_RAR15_429_OFFSET];
                     break;
                 
                  default:
                     break;
               }
            }

            if (status == ARS_NORMAL) {
               /* read the board boardConfiguration field */
               status = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], &globalEnableRegValue);
               boardConfiguration = (globalEnableRegValue >> 8) & 0x000000FFU;
               if ( (boardConfiguration >= (CEI_UINT32)CEIDEV_FIRST_CEIX30_CONFIG) && (boardConfiguration <= (CEI_UINT32)CEIDEV_LAST_CEIX30_CONFIG) ) {  
                  fwVersion = (globalEnableRegValue >> CEIDEV_FIRMWARE_VERSION_SHIFT) & CEIDEV_FIRMWARE_VERSION_MASK;

                  /* this version of the CEI-x30 API only works with firmware version 4.13 and later */
                  if (fwVersion < CEI_EARLIEST_COMPATIBLE_FW) {
                     status = ARS_FW_NOT_SUPPORTED;
                  }
                  else {
                     deviceGlobalStructPtr->configuration = boardConfiguration;
                     deviceGlobalStructPtr->firmwareVersion = fwVersion;
      
#ifdef INCLUDE_IRIG
                     if (globalEnableRegValue & CEIDEV_IRIG_AVAILABLE) {
                        deviceGlobalStructPtr->irigInstalled = TRUE;
                     }
#endif
                     /* query the board for the channel boardConfiguration */
                     a429_utl_determine_channel_support (board);
                     
                     /* multi-protocol boards have an additional "combo channel count" attribute */
                     if ( (boardConfiguration == (CEI_UINT32)CEIDEV_CONFIG_RAR15) || (boardConfiguration == (CEI_UINT32)CEIDEV_CONFIG_RAR15XT) ) {
                        status = cei_read_32bit (board, &deviceLocalStructPtr->baseAddr[0], &comboBoardChanCount);
                        deviceGlobalStructPtr->comboBoard1553chanCount = 
                           (comboBoardChanCount & CEIDEV_RAR15_1553_CH_COUNT_MASK) >> CEIDEV_RAR15_1553_CH_COUNT_SHIFT;
                     }
                     else {
                        deviceGlobalStructPtr->comboBoard1553chanCount = 0U;
                     }
                  }
               }
            }
            else {
               status = ARS_HW_DETECT;
            }
         }
      }
#ifndef TARGET_EMBEDDED
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   if (altStatus) ;
   return (CDEV_API_RET_TYPE)status;
}


/*===================================================================================================================*
 * ENTRY POINT:                          A R _ I N I T I A L I Z E _ D E V I C E
 *===================================================================================================================*
 *
 * FUNCTION:    Initialize the device to a ready state
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board  -- (input) device number to reference (0-127)
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL       Success
 *                                ARS_BAD_STATIC   Register read/write/verify failure
 *                                ARS_MEMWRERR     Memory test failure
 *                                ARS_INVBOARD     Invalid board parameter
 *                                ARS_BOARD_MUTEX  Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: This routine initializes the device I/O configuration registers to a state ready for use by the 
 *              host application.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_initialize_device 
(
   CDEV_BOARD_TYPE board                                                   /* (input)  board to reference (0 to 127) */
)
{
   /* a power-up initialization of the device will do */
   return a429_utl_initialize_device (board, TRUE);
}


/*===================================================================================================================*
 * ENTRY POINT:                               A R _ N U M _ R C H A N S
 *===================================================================================================================*
 *
 * FUNCTION:    Return the number of receive channels on the board.
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board -- (input) board to reference (0 to 127)
 *
 * RETURN VAL:  CDEV_CHAN_TYPE  number of receive channels on this board.
 *                              ARS_INVBOARD  board parm value not supported
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: Report the number of receiver channels detected on this device
 *              NOTE: The board must have been initialized via ar_open.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_CHAN_TYPE DLL_EXPORTED ar_num_rchans
(
   CDEV_BOARD_TYPE board                                                   /* (input)  board to reference (0 to 127) */
)
{
   CDEV_API_RET_TYPE nChannels = ARS_INVBOARD;
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      nChannels = (CDEV_CHAN_TYPE)(deviceGlobal[board]->numRx);
#ifndef TARGET_EMBEDDED
   }
#endif
   return nChannels;
}               


/*===================================================================================================================*
 * ENTRY POINT:                               A R _ N U M _ X C H A N S
 *===================================================================================================================*
 *
 * FUNCTION:    Return the number of transmit channels on the board.
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board -- (input) board to reference (0 to 127)
 *
 * RETURN VAL:  CDEV_CHAN_TYPE  number of transmit channels on this board.
 *                              ARS_INVBOARD  board parm value not supported
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: Report the number of receiver channels detected on this device
 *              NOTE: The board must have been initialized via ar_open.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_CHAN_TYPE DLL_EXPORTED ar_num_xchans
(
   CDEV_BOARD_TYPE board                                                   /* (input)  board to reference (0 to 127) */
)
{
   CDEV_API_RET_TYPE nChannels = ARS_INVBOARD;
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      nChannels = (CDEV_CHAN_TYPE)(deviceGlobal[board]->numTx);
#ifndef TARGET_EMBEDDED
   }
#endif
   return nChannels;
}               


/*===================================================================================================================*
 * ENTRY POINT:                                     A R _ O P E N
 *===================================================================================================================*
 *
 * FUNCTION:    This routine opens a session with a board, then initializes the API and board.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board     -- (input) device number (0-127)
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL           Success
 *                                ARS_INVBOARD         board parm value not supported
 *                                ARS_BADLOAD          FPGA load process failed
 *                                ARS_BOARD_MUTEX      Creation of the Board Lock timed-out/failed
 *                                ARS_DRIVERFAIL       Board session open failed
 *                                ARS_HW_DETECT        Supported device not detected
 *                                ARS_ERR_SH_MEM_MAP   failed to map shared memory block (multi-process
 *                                ARS_ERR_SH_MEM_OBJ   failed to allocate shared memory block      only)
 *                                ARS_FW_NOT_SUPPORTED Device firmware revision not supported
 *                                ARS_BADLOAD          FPGA load process failed
 *                                ARS_MEMWRERR         Device memory test failure
 *                                ARS_FAILURE          <Windows only> CEI_Install library failed to load in memory
 *                                                     <Linux multi-process> failed to create a shared memory block
 * GLOBALS:     deviceGlobal[]
 * 
 * DESCRIPTION: This routine fulfills the standard ARINC product API initialization functionality.  It invokes the 
 *              following functions herein:
 *
 *                 ar_initialize_api()    - acquires the device resources & loads PLX-based board firmware
 *                 ar_initialize_device() - initializes the device to the default settings
 *                 ar_board_test()        - performs a simple internal wrap test
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_open 
(
   CDEV_BOARD_TYPE board                                                   /* (input)  board to reference (0 to 127) */
)
{
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                  /* function return status                 */
   CEI_INT32  accessStatus = ARS_NORMAL;                                   /* status of device lock access           */
   CEI_INT32  mapStatus = ARS_NORMAL;                                      /* function return status                 */
   CEI_INT32  index;                                                       /* utility index                          */
   CEI_UINT32 alreadyExists = FALSE;                                  /* indicates if shared board mem block exists  */
   CEI_UINT32 isExclusive = FALSE;                                    /* indicates if this is only mapping process   */
   CEI_UINT32 forceInit = FALSE;                                      /* indicates if shared mem init should be done */
   ARINC_GLOBAL_TYPE * localDevGlobalPtr = NULL;                           /* pointer to the board global structure  */

#ifndef TARGET_EMBEDDED
   if (board < MAX_BOARDS) {
#endif
	   if (deviceGlobal[board] == NULL) {
         /* for a multi-process application, either create the shared block or just map it*/
         if (multiProcessProtect[board] == AR_CONC_MULTIPROC) {

            /* assign the global thread/process protection mode to the lib-initialized value */
            device[board].multiThreadProtect = multiProcessProtect[board];

#if defined (MULTI_PROCESS_SUPPORT) && !defined (TARGET_EMBEDDED)
            /* assure we are alone in creating the shared memory block */
            accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
#endif

            if (accessStatus == ARS_NORMAL) {

               /* attempt to map global shared board memory block */
               mapStatus = cei_map_shared_memory( (CEI_UCHAR)board,
                                                  sizeof(ARINC_GLOBAL_TYPE),
                                                  (CEI_VOID **)&deviceGlobal[board],
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
                                                  (CEI_INT32 *)&device[board].sharedMemoryHandle,
#else                                                  
                                                  (pCEI_VOID)&device[board].sharedMemoryHandle,
#endif
                                                  &alreadyExists,
                                                  &isExclusive );
                               
               if (mapStatus != ARS_NORMAL) {
                  /* flag shared memory as invalid and bail out */
                  deviceGlobal[board] = NULL;
#ifndef TARGET_EMBEDDED
                  accessStatus = cei_release_thread_lock(LOCK_IDENT);
#endif
                  status = (CDEV_API_RET_TYPE)mapStatus;
               }
               else {
                  localDevGlobalPtr = deviceGlobal[board];

                  /* force initialization of shared memory if all attached processes previously terminated but the board is still open */
                  forceInit = ((alreadyExists) && (isExclusive) && (deviceGlobal[board]->attachedProcessCount == 0U));
      
                  /* check if this is the first process to map the block or init required */
                  if ( (alreadyExists == FALSE) || (forceInit) ) {
                     localDevGlobalPtr->attachedProcessCount = 0U;
                     localDevGlobalPtr->configuration = 0U;
                     localDevGlobalPtr->messageSchedulerTableEnd = CEIDEV_MSG_SCHED_TABLE_END;
                     localDevGlobalPtr->messageSchedulerTableSize = CEIDEV_MSG_SCHED_TBL_ENTRY_MAX;
                     localDevGlobalPtr->ceiPciInterruptEnabled = FALSE;
                     for (index = 0; index < CEIDEV_MAX_RX_CHAN; index++) {
                        localDevGlobalPtr->receiveChannelMessageCount[index] = 0U;
                        localDevGlobalPtr->receiveChannelBufferMode[index] = ARU_BUFFERED;
                     }
                  }
               }
#if defined (MULTI_PROCESS_SUPPORT) && !defined (TARGET_EMBEDDED)
               /* first process in a multiprocess initialization is done, safe to release the lock */
               accessStatus = cei_release_thread_lock(LOCK_IDENT);
#endif
            }
            else { /* failed to acquire the lock, this should never happen, but if it does fail initialization */
               status = ARS_BOARD_MUTEX;
            }
         }
         else {
            /* assign the global thread/process protection mode to the lib-initialized value */
            device[board].multiThreadProtect = multiProcessProtect[board];
            device[board].sharedMemoryHandle = NULL;

            /* For a single-process application, just point to local pseudo-shared block for single-process implementation */
            localDevGlobalPtr = &deviceSingleProcessShared[board];
            localDevGlobalPtr->boardInitialized = FALSE;
            localDevGlobalPtr->attachedProcessCount = 0U;
            localDevGlobalPtr->configuration = 0U;
            localDevGlobalPtr->messageSchedulerTableEnd = CEIDEV_MSG_SCHED_TABLE_END;
            localDevGlobalPtr->messageSchedulerTableSize = CEIDEV_MSG_SCHED_TBL_ENTRY_MAX;
            localDevGlobalPtr->ceiPciInterruptEnabled = FALSE;
            for (index = 0; index < CEIDEV_MAX_RX_CHAN; index++) {
               localDevGlobalPtr->receiveChannelMessageCount[index] = 0U;
            }
            deviceGlobal[board] = (ARINC_GLOBAL_TYPE *)localDevGlobalPtr;
         }
      }

      if ((localDevGlobalPtr != NULL) && (status == ARS_NORMAL)) {
         /* acquire the device resources and initialize the API */
         status = ar_initialize_api (board);
      
         if (status == ARS_NORMAL) {
            localDevGlobalPtr->attachedProcessCount++;
            a429_utl_initialize_global_data (board);
   
            if ( (localDevGlobalPtr->attachedProcessCount == 1) || (device[board].multiThreadProtect == AR_CONC_NONE) ) {
               
               /* initialize the device to a known state, then perform an internal wrap test (unless bypassed). */
               status = ar_initialize_device (board);
   
               if (status == ARS_NORMAL) {
                  localDevGlobalPtr->boardInitialized = TRUE;
               }
            }
            if ((status == ARS_NORMAL) && (bypassWrapTest[board] == FALSE)) {
               status = ar_board_test (board, (CDEV_PARM_SSI_TYPE)INTERNAL_WRAP);
            }
         }
         else {
            /* device session open failed... clean up here */
            if (multiProcessProtect[board] == AR_CONC_MULTIPROC) {
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
               /* for Linux multi-process application usage, cleanup of the System V shared memory allocation has to occur here */
               shmdt(&deviceGlobal[board]); 
               shmctl((int)device[board].sharedMemoryHandle, IPC_RMID, NULL);
#endif
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (INTEGRITY_PCI_PPC) && !defined (INTEGRITY_PCI_X86)
               {
                  CDEV_API_RET_TYPE localStatus;
                  localStatus = (CDEV_API_RET_TYPE)cei_release_device(board, (ARINC_LOCAL_TYPE *)&device[board]);
               }
#endif
            }
            deviceGlobal[board] = NULL;
         }
      }

#ifndef TARGET_EMBEDDED
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                                    A R _ P U T W O R D
 *===================================================================================================================*
 *
 * FUNCTION:    Insert a single message into the specified channel's sequential transmit buffer.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board       -- (input) board to reference (0 to 127)
 *              CDEV_CHAN_TYPE channel      -- (input) Transmit channel (0 to *)
 *              CDEV_PARM_SI_TYPE arincdata -- (input) message data to transmit
 * 
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL      Success
 *                                ARS_INVBOARD    Invalid board or board not initialized
 *                                ARS_INVHARVAL   Invalid transmit channel argument
 *                                ARS_XMITOVRFLO  Transmit buffer overrun has occurred
 *                                ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: This routine will place the specified message in the respective channel FIFO.  The channel number 
 *              passed to this routine corresponds to the ARINC 429 transmit channel index, starting with zero.  
 *              If the channel parameter value exceeds the 429 transmit channel count and an ARINC 573/717 
 *              transmitter exists in the board configuration, it will be used as the designated transmitter.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_putword 
( 
   CDEV_BOARD_TYPE board,                                                   /* (input) board to reference (0 to 127) */
   CDEV_CHAN_TYPE channel,                                                  /* (input) transmit channel to use       */
   CDEV_PARM_SI_TYPE arincdata                                              /* (input) 32/16-bit data to transmit    */
) 
{
   CEI_UINT32 channelStatus;                                                   /* Rx channel FIFO data avail status  */
   CEI_UINT32 channelOffset;                                          /* chan reg region offset in host interface    */
   CEI_UINT32 bufferOffset;                                           /* chan buffer region offset in host interface */
   CEI_INT32 accessStatus = ARS_NORMAL;                                        /* status of device read/write access */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                      /* function return status             */
   CEI_UINT32 localChannel = (CEI_UINT32)channel;                              /* hardware referenced channel number */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                               /* ptr to board local structure   */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                                      /* ptr to board global structure  */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      /* assign references to this board's device and deviceGlobal structure-array elements */
      deviceLocalStructPtr = &device[board];
      deviceGlobalStructPtr = deviceGlobal[board];
   
      /* First verify the specified channel is either a valid 429 transmitter or if not, 
       * defaulting to the optional 717 transmitter is available 
       */
      if (localChannel < deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX]) {
         localChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
      }
      else if (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX] != CEIDEV_MAX_CHANNELS) {
         localChannel = deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX];
      }
#ifndef TARGET_EMBEDDED
      else {
         status = ARS_INVHARVAL;
      }
      if (status == ARS_NORMAL) {
#endif
         channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS + (localChannel * CEIDEV_CHANNEL_REGION_SIZE);
         bufferOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_BUFFER_3 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE);
   
#ifndef TARGET_EMBEDDED
         /* if multi-thread protection requested, acquire semaphore/mutex-locking */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (accessStatus == ARS_NORMAL) {
#endif
            /* check the transmit channel status, if it indicates "NOT FULL"
             * then stuff the data into the queue; otherwise, it's an overflow.
             */
            accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[channelOffset], &channelStatus);
            if (accessStatus == ARS_NORMAL) {
               if ((channelStatus & CEIDEV_CSR_XMIT_FIFO_FULL) != (CEI_UINT32)0U) {
                  status = ARS_XMITOVRFLO;
               }
               else {
                  /* stuff the data into the transmit FIFO */
                  accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[bufferOffset], (CEI_UINT32)arincdata);
                  status =  ARS_NORMAL;
               }
            }
#ifndef TARGET_EMBEDDED
            /* if multi-thread protection is active, free any acquired lock */
            if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
               accessStatus = cei_release_thread_lock(LOCK_IDENT);
            }
         }
#endif
         if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL)) {
            status = (CDEV_API_RET_TYPE)accessStatus;
         }
#ifndef TARGET_EMBEDDED
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                                  A R _ P U T B L O C K
 *===================================================================================================================*
 *
 * FUNCTION:    Insert a block of messages into the specified channel's sequential transmit buffer.
 *
 * PARAMETERS:  CEI_UINT32 board       -- (input)  board to reference (0 to 127)
 *              CEI_UINT32 channel     -- (input)  transmit channel (0 to 31)
 *              CEI_INT32  maxMessages -- (input)  number of messages to transmit
 *              CEI_INT32  offset      -- (input)  unused parameter (legacy support)
 *              pCEI_INT32 data        -- (input)  source of data
 *              pCEI_INT32 actualCount -- (output) number of messages transmitted
 *
 * RETURN VAL:  ARS_NORMAL        Success
 *              ARS_INVBOARD      Invalid board number
 *              ARS_INVHARVAL     Invalid receive channel argument
 *              ARS_INVARG        Invalid argument value
 *              ARS_BOARD_MUTEX   Access to the Board Lock timed-out/failed
 *              ARS_XMITOVRFLO    Transmit buffer overflow - indicates that 
 *                                there was not enough room in the transmit 
 *                                buffer for all of the ARINC words specified
 *                                ('ActualCount' receives the number of words
 *                                actually transfered)
 *
 * GLOBALS:     device[]
 *              deviceGlobal[]
 *
 * DESCRIPTION: This routine puts multiple ARINC messages in the transmit queue for a particular channel. When this 
 *              routine returns, the data has not necessarily been sent, it has only been placed in the transmit 
 *              buffer. If other data is in the transmit buffer ahead of it, it will be transmitted in turn.
 *
 *===================================================================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_putblock
(
   CEI_UINT32 board,                                                   /* (input)  board to reference (0 to 127)     */
   CEI_UINT32 channel,                                                 /* (input)  transmit channel (0 to 31)        */
   CEI_INT32  maxMessages,                                             /* (input)  number of messages to transmit    */
   CEI_INT32  offset,                                                  /* (input)  unused parameter (legacy support) */
   pCEI_INT32 data,                                                    /* (input)  source of data                    */
   pCEI_INT32 actualCount                                              /* (output) number of messages transmitted    */
)
{
   CEI_UINT32 channelStatus;                                                   /* Transmit channel buffer status     */
   CEI_UINT32 channelOffset;                                          /* chan reg region offset in host interface    */
   CEI_UINT32 bufferOffset;                                           /* chan buffer region offset in host interface */
   CEI_INT32  arincdata;                                                       /* temp ARINC message container       */
   pCEI_INT32 localDataPtr;                                                    /* temp pointer to application data   */
   CEI_UINT32 messageCount = offset;                                           /* transmitted message counter        */
   CEI_UINT32 messageCtMax = 0U;                                               /* number of messages to transmit     */
   CEI_INT32 accessStatus = ARS_NORMAL;                                        /* status of device read/write access */
   CEI_INT32 status = ARS_NORMAL;                                              /* function return status             */
   CEI_UINT32 localChannel = (CEI_UINT32)channel;                              /* hardware referenced channel number */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                           /* pointer to board local structure   */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                                  /* pointer to board global structure  */


#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      /* assign references to this board's device and deviceGlobal structure-array elements */
      deviceLocalStructPtr = &device[board];
      deviceGlobalStructPtr = deviceGlobal[board];
      
      /* verify and then offset the logical channel supplied by the hardware transmit channel index */
      if (channel < deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX]) {
         localChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
      }
      else if (deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX] != CEIDEV_MAX_CHANNELS) {
         localChannel = deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX];
      }
      else {
         status = ARS_INVHARVAL;
      }
#ifdef TARGET_EMBEDDED
      if (status == ARS_NORMAL) {
#else      
      if (status == ARS_NORMAL) {
         status = a429_utl_validate_channel_request ((CDEV_BOARD_TYPE)board, localChannel, ARU_PUT_DATA);
      
         /* verify required parameters */
         if ( (status == ARS_NORMAL) && ((data == NULL) || (actualCount == NULL)) ) {
            status = ARS_INVARG;
         }
         else {
#endif
            *actualCount = 0;

#ifndef TARGET_EMBEDDED
            /* if multi-thread protection requested, acquire semaphore/mutex-locking */
            if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
               accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
            }
            if (accessStatus == ARS_NORMAL) {
#endif
               channelOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS + (localChannel * CEIDEV_CHANNEL_REGION_SIZE);
               
               /* check the transmit channel status, if it indicates the buffer isn't already full,
                * attempt to transmit as many messages as possible, hopefully as much as requested 
                */
               accessStatus = cei_read_32bit ((CDEV_BOARD_TYPE)board,  &deviceLocalStructPtr->devicePtr[channelOffset], &channelStatus);
               if (accessStatus == ARS_NORMAL) {
                  if (channelStatus & CEIDEV_CSR_XMIT_FIFO_FULL) {
                     status = ARS_XMITOVRFLO;
                  }
                  else {
                     /* limit the number of messages allowed to transmit based on available space in the buffer */
                     channelStatus = (channelStatus & CEIDEV_CSR_TX_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT;
                     messageCtMax = (CEI_UINT32)maxMessages <= (CEIDEV_CIRCULAR_BUFFER_SIZE - channelStatus) ? (CEI_UINT32)maxMessages : CEIDEV_CIRCULAR_BUFFER_SIZE - channelStatus; 

                     messageCount = 0;
                     localDataPtr = data;
                     bufferOffset = CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_BUFFER_3 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE);
                     
                     /* insert transmit messages into the circular buffer */
                     while ((messageCount < messageCtMax) && (accessStatus == ARS_NORMAL)) {
                        arincdata = *localDataPtr;
                     
                        /* stuff the data into the transmit FIFO */
                        accessStatus = cei_write_32bit ((CDEV_BOARD_TYPE)board, &deviceLocalStructPtr->devicePtr[bufferOffset], arincdata);
                        if (accessStatus == ARS_NORMAL) {
                           messageCount++;
                           localDataPtr++;
                        }
                     }
                  }
#ifndef TARGET_EMBEDDED
               }
               /* if multi-thread protection is active, free any acquired lock */
               if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
                  accessStatus = cei_release_thread_lock(LOCK_IDENT);
               }
#endif
            }
      
            *actualCount = messageCount;
         }
   
#ifndef TARGET_EMBEDDED
      }
      if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL)) {
         status = accessStatus;
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif
   
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                                       A R _ R E S E T
 *===================================================================================================================*
 *
 * FUNCTION:    Reset the device to the initial state
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board  -- (input) board to reference (0 to 127)
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL    success
 *                                ARS_INVBOARD  board parm value not supported
 *
 * GLOBALS:     none
 * 
 * DESCRIPTION: see description for ar_initialize_device()
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_reset
(
   CDEV_BOARD_TYPE board                                                   /* (input)  board to reference (0 to 127) */
)
{
   /* a simple non-power-up re-initialization of the device will do */
   return a429_utl_initialize_device (board, FALSE);
}    
     

/*===================================================================================================================*
 * ENTRY POINT:                             A R _ S E T _ D E V I C E _ C O N F I G          
 *===================================================================================================================*
 *
 * FUNCTION:    Assign protocol-specific device configuration register attributes 
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE board    -- (input) board to reference (0 to 127)
 *              CDEV_CHAN_TYPE channel   -- (input) channel number (0 to # 429 channels)
 *              CDEV_PARM_SSI_TYPE item  -- (input) channel config item to (re)set
 *              CDEV_PARM_SSI_TYPE value -- (input) value to set item to
 * 
 *              Valid combinations of item / value parameters are:
 * 
 *              for this "item"        valid options for "value" are:
 * 
 *              ARU_RX_BITRATE         ARU_SPEED_LOW   ARU_SPEED_HIGH  Custom Speed
 *              ARU_RX_PARITY          AR_OFF          AR_ON    or ARU_PARITY_NONE and ARU_PARITY_ODD/ARU_PARITY_EVEN
 *                                                              representing AR_OFF               AR_ON  
 *              ARU_RECV_MODE          AR_WRAP_OFF     AR_WRAP_ON
 * 
 *              ARU_RX_FIFO_RESET      N/A
 * 
 *              ARU_TX_BITRATE         ARU_SPEED_LOW   ARU_SPEED_HIGH  Custom
 *              ARU_TX_PARITY          ARU_PARITY_NONE ARU_PARITY_ODD  ARU_PARITY_EVEN
 *
 *              ARU_RX_DISABLE         AR_ON           AR_OFF
 *              ARU_RX_FIFO_ENABLE     AR_ON           AR_OFF
 *              ARU_TX_DISABLE         AR_ON           AR_OFF
 *              ARU_TX_FIFO_ENABLE     AR_ON           AR_OFF
 *
 *              ARU_RX_MSG_SIZE_24BIT  AR_ON           AR_OFF
 *              ARU_TX_MSG_SIZE_24BIT  AR_ON           AR_OFF
 *
 *              ARU_TX_GAP_ERROR       AR_ON           AR_OFF
 *              ARU_TX_BIT_ERROR       AR_HI   AR_LO   AR_OFF
 *
 *              ARU_DISCRETE_OUT       AR_HI           AR_LO
 *              ARU_DIGITAL_OUT        AR_HI           AR_LO
 *              ARU_DIFFERENTIAL_OUT   AR_ON   AR_OFF  AR_HI    AR_LO
 *
 *              ARU_RX_MERGED_MODE     AR_ON           AR_OFF
 *              ARU_FAST_SLEW_RATE     AR_ON           AR_OFF
 * 
 *              ARU_CONFIG_PROGRAMMABLE_CHAN  ARU_RECEIVER  ARU_TRANSMITTER
 *
 *              ARU_ACCESS_SNAPSHOT_BUFFER ARU_LABEL_ONLY   ARU_LABEL_WITH_SDI
 *              ARU_BYPASS_INIT_WRAP_TEST  AR_ON       AR_OFF
 *              ARU_MULTITHREAD_PROTECT    AR_ON       AR_OFF
 *              ARU_HW_INTERRUPT_ENABLE    AR_ON       AR_OFF       
 * 
 *              ARU_IRIG_WRAP_ENABLE   AR_ON           AR_OFF
 *              ARU_IRIG_OUTPUT_ENABLE AR_ON           AR_OFF
 *              ARU_IRIG_INPUT_THRESHOLD      threashold voltage in millivolts
 *              ARU_IRIG_ADJUST_THRESHOLD     display DAC settings AR_ON/AR_OFF
 *              ARU_IRIG_QUICK_ADJUSTMENT
 *              ARU_IRIG_SET_BIAS             user IRIG offset 0 to +/-32768 milliseconds
 *              ARU_HW_ENHANCE_UPDATE  AR_ON           AR_OFF
 *              ARU_RX_TIMETAG_MODE w/values of:                          
 *                                     AR_TIMETAG_EXT_IRIG_64BIT                         
 *                                     AR_TIMETAG_INT_USEC_64BIT                         
 *                                     AR_TIMETAG_INT_20USEC_32BIT                       
 *                                     AR_TIMETAG_INT_MSEC_32BIT
 *                                     AR_TIMETAG_SYNC_1553_CH1 (for combo cards only)               
 *                                     AR_TIMETAG_SYNC_1553_CH2                
 *                                     AR_TIMETAG_SYNC_1553_CH3                
 *                                     AR_TIMETAG_SYNC_1553_CH4                
 *                                                     
 * RETURN VAL:  CDEV_API_RET_TYPE status  ARS_NORMAL         Normal succesful completion.
 *                                        ARS_INVARG         "item" parameter value is not supported by this function
 *                                        ARS_INVHARVAL      "value" was invalid or channel not supported on this device
 *                                        ARS_INVBOARD       Board parameter value invalid or board has not been initialized
 *                                        ARS_INT_ISR  &     Interrupt was either not initialized
 *                                        ARS_DRIVERFAIL     or not terminated successfully
 *                                        ARS_NO_INT_SUPPORT Driver interrupt handling not supported for this board
 *                                        ARS_NO_INT_ENABLED Driver Interrupt handling not enabled for this board
 *                                        ARS_BOARD_MUTEX    Access to the Board Lock timed-out/failed
 * 
 * GLOBALS:     device[]
 *              deviceGlobal[]
 * 
 * DESCRIPTION: This function assigns the respective device configuration register attributes based on the combined 
 *              item/value parameter selections.  It is designed to support the ARINC 429 and all discrete and 
 *              digital I/O configuration register attributes available to the device.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_set_device_config 
( 
   CDEV_BOARD_TYPE board,                                            /* (input) board to reference (0 to 127)        */
   CDEV_CHAN_TYPE channel,                                           /* (input) channel number (0 to # 429 channels) */
   CDEV_PARM_SSI_TYPE item,                                          /* (input) channel config item to (re)set       */
   CDEV_PARM_SSI_TYPE value                                          /* (input) value to set item to                 */
)
{
   CEI_UINT32 deviceChannel = 0U;                                           /* current channel to select             */
   CEI_UINT32 bitFieldValue = 0U;                                           /* temp for device register read content */
   CEI_UINT32 tempValue;                                                    /* temp for baud rate setting            */
   CEI_INT32  accessStatus = ARS_NORMAL;                                    /* status of read/write access           */
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                   /* function return status                */
   CEI_INT32  tempStatus = ARS_NORMAL;                                      /* local temp return status              */
   CEI_UINT32 offsetIntoChannelWindow = 0U;                                 /* offset into chan mem region           */
#ifndef TARGET_EMBEDDED                                                                                              
   CEI_UINT32 lockAcquired = FALSE;                                         /* multi-thread lock aquired flag        */
#endif                                                                                                               
   CEI_UINT32 memOffset;                                                    /* local offset to channel memory area   */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                        /* ptr to board local structure          */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                               /* ptr to board global structure         */

#ifndef TARGET_EMBEDDED
#ifdef INCLUDE_PLX_SUPPORT
   if ((item == ARU_HW_ENHANCE_UPDATE) && (board < MAX_BOARDS)) {
      if (value == AR_OFF) {
         status = a429_utl_enh (board, PLX_ORIGINAL_RESTORE);
      }
      else if (value == AR_ON) {
         status = a429_utl_enh (board, PLX_ENHANCED_UPDATE);
      }
      else {
         status = ARS_INVARG;
      }
      if (status == ARS_NORMAL) {
         status = EXIT_ROUTINE;
      }
   }
   else if ((item == ARU_WAIT_ON_EVENT) && (board < MAX_BOARDS)) {
#else   
   if ((item == ARU_WAIT_ON_EVENT) && (board < MAX_BOARDS)) {
#endif
      accessStatus = (CEI_INT32)value;
      tempStatus = cei_open_shared_lock (board, &accessStatus);
      if (tempStatus == ARS_NORMAL) {
         status = EXIT_ROUTINE;
      }
      else {
         status = (CDEV_API_RET_TYPE)tempStatus;
      }
   }
   else if (VALID_BOARD_AND_INIT) {
#endif
      /* assign references to this board's device and deviceGlobal structure-array elements */
      deviceLocalStructPtr = &device[board];
      deviceGlobalStructPtr = deviceGlobal[board];
   
#ifndef TARGET_EMBEDDED
      if (status == ARS_NORMAL) {
         /* if multi-thread protection is active, acquire a lock on the host interface */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            status = (CDEV_API_RET_TYPE)cei_acquire_thread_lock(LOCK_IDENT);
            if (status == ARS_NORMAL) {
               lockAcquired = TRUE;
            }
         }
      }
#endif
            
      if (status == ARS_NORMAL) {
         deviceChannel = (CEI_UINT32)channel;
   
         /* Validate the channel parameter for the specific item parameter,
          * along with predefining the Channel Type, Channel Select, and Channel
          * Offset, for options requiring configuration register operations.
          */
         switch (item) {
            case ARU_RECV_MODE:
            case ARU_RX_BITRATE:
            case ARU_RX_PARITY:
            case ARU_RX_DISABLE:
            case ARU_RX_FIFO_ENABLE:
            case ARU_RX_MERGED_MODE:
            case ARU_RX_RESET:
            case ARU_RX_MSG_SIZE_24BIT:
            case ARU_RX_FIFO_RESET:
               offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
               deviceChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
               break;
         
            case ARU_TX_BIT_ERROR:
            case ARU_TX_GAP_ERROR:
            case ARU_TX_BITRATE:
            case ARU_TX_PARITY:
            case ARU_TX_FIFO_ENABLE:
            case ARU_TX_DISABLE:
            case ARU_FAST_SLEW_RATE:
            case ARU_TX_RESET:
            case ARU_TX_MSG_SIZE_24BIT:
               offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
               deviceChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
               break;
      
#ifdef INCLUDE_DISCRETE_IO
            /* support legacy CEI-710 Discrete item parameters */
            case ARU_DISCRETE_OUT1:
            case ARU_DISCRETE_OUT2:
               deviceChannel = (item == ARU_DISCRETE_OUT1) ? 0U : 1U;
               /* no break here, now process below as a generic ARU_DISCRETE_OUT selection */
            case ARU_DISCRETE_OUT:
               deviceChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_DISC_OUT];
               offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
               break;
         
            case ARU_DIFFERENTIAL_OUT:
               deviceChannel += deviceGlobalStructPtr->firstChannel[CEIDEV_CH_TYPE_DIFF_OUT];
               offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
               break;
#endif
      
            case ARU_ACCESS_SNAPSHOT_BUFFER:
               if (value == ARU_LABEL_WITH_SDI) {
                  bitFieldValue = CEIDEV_GLOBAL_ENABLE_AUTON_CLEAR + CEIDEV_SNAPSHOT_STORAGE_ENHANCED;
               }
               else if (value == ARU_LABEL_ONLY) {
                  bitFieldValue = CEIDEV_GLOBAL_ENABLE_AUTON_SET + CEIDEV_SNAPSHOT_STORAGE_ENHANCED;
               }
               else {
                  status = ARS_INVHARVAL;
               }
               if (status == ARS_NORMAL) {
                  accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
                  status = EXIT_ROUTINE;
               }
               break;
       
#ifdef INCLUDE_INTERRUPTS
            case ARU_HW_INTERRUPT_ENABLE:
               status = a429_utl_process_hw_int (board, value);
               break;
      
            case ARU_INSERT_INT_Q_ENTRY:
               accessStatus = cei_write_32bit(board, &deviceLocalStructPtr->devicePtr[CEIDEV_INT_QUEUE_HEAD_PTR], value);
               status = EXIT_ROUTINE;
               break;
      
            case ARU_PCI_INTERRUPT_ENABLE:
               if (value == AR_ON) {
                  /* enable interrupts */
                  bitFieldValue = CEIDEV_GLOBAL_ENABLE_AUTON_SET + CEIDEV_INTERRUPTS_ENABLED;
                  accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
               }
               else if (value == AR_OFF) {
                  /* disable interrupts */
                  bitFieldValue = CEIDEV_GLOBAL_ENABLE_AUTON_CLEAR + CEIDEV_INTERRUPTS_ENABLED;
                  accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
               }
               else {
                  status = ARS_INVHARVAL;
               }
               if (status == ARS_NORMAL) {
                  status = EXIT_ROUTINE;
               }
               break;
#endif
      
#ifdef INCLUDE_IRIG
            /* The remaining items are handled either in utility routines or they 
             * don't actually require interaction with common h/w channel register
             * selection.  For these options, perform the operation then set the
             * status so as to skip the subsequent processing.
             */
            case ARU_IRIG_WRAP_ENABLE:
            case ARU_IRIG_INPUT_THRESHOLD:
            case ARU_IRIG_ADJUST_THRESHOLD:
            case ARU_IRIG_QUICK_ADJUSTMENT:
            case ARU_IRIG_SET_BIAS:
            case ARU_IRIG_OUTPUT_ENABLE:
               if (deviceGlobalStructPtr->irigInstalled) {
                  switch (item) {
                     case ARU_IRIG_WRAP_ENABLE:
                        if ( (value == AR_ON) || (value == AR_WRAP_ON) ) {
                           bitFieldValue = CEIDEV_GLOBAL_ENABLE_AUTON_SET + CEIDEV_IRIG_INTERNAL_ENABLE;
                        }
                        else {
                           bitFieldValue = CEIDEV_GLOBAL_ENABLE_AUTON_CLEAR + CEIDEV_IRIG_INTERNAL_ENABLE;
                        }
                        accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
                        break;
               
                     case ARU_IRIG_OUTPUT_ENABLE:
                        if ( (value == AR_ON) || (value == AR_WRAP_ON) ) {
                           bitFieldValue = CEIDEV_GLOBAL_ENABLE_AUTON_SET + CEIDEV_IRIG_830RX_IRIG_ENABLE;
                        }
                        else {
                           bitFieldValue = CEIDEV_GLOBAL_ENABLE_AUTON_CLEAR + CEIDEV_IRIG_830RX_IRIG_ENABLE;
                        }
                        accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
                        break;
               
                     case ARU_IRIG_INPUT_THRESHOLD:
#ifdef _DEBUG
                        bitFieldValue = (CEI_UINT32)value;
#else
                        bitFieldValue = (CEI_UINT32)((128L + (CEI_INT32)(((float)256/(float)3.3)*((float)4.99/(float)22.1) * ((float)value/(float)1000) )) & 0x000000FF);
#endif                  
                        accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_DAC_CONTROL], bitFieldValue);
                        deviceLocalStructPtr->irigDacThreshold = (CEI_INT32)bitFieldValue;
                        break;
               
                     case ARU_IRIG_ADJUST_THRESHOLD:
                        accessStatus = a429_utl_adjust_irig_threshold (board, (CEI_INT32)value);
                        break;
               
                     case ARU_IRIG_QUICK_ADJUSTMENT:
                        accessStatus = a429_utl_quick_adjust_irig_threshold (board, (CEI_INT32)value);
                        break;
               
                     case ARU_IRIG_SET_BIAS:
                        deviceGlobalStructPtr->userProgrammableIrigOffset = (TIME_TAG_TYPE)value * (TIME_TAG_TYPE)1000;
                        break;
                  }
                  if (accessStatus == ARS_NORMAL) {
                     status = EXIT_ROUTINE;
                  }
                  else {
                     status = (CDEV_API_RET_TYPE)accessStatus;
                  }
               }
               else {
                  status = ARS_INVHARVAL;
               }
               break;
#endif
      
            case ARU_BYPASS_INIT_WRAP_TEST:
               status = ar_bypass_wrap_test (board, value);
               if (status == ARS_NORMAL) {
                  status = EXIT_ROUTINE;
               }
               break;
            
            case ARU_MULTITHREAD_PROTECT:
               status = ar_set_multithread_protect (board, value);
               if (status == ARS_NORMAL) {
                  status = EXIT_ROUTINE;
               }
               break;
      
            case ARU_RX_TIMETAG_MODE:
               if ( (value == AR_TIMETAG_EXT_IRIG_64BIT) && (deviceGlobalStructPtr->irigInstalled == FALSE) ) {
                  status = ARS_INVHARVAL;
               }
               /* for a Combo Card, allow 1553 Synchronized time-tag operations if the requested 1553 channel is initialized */
               else if ( (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15) || (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15XT) ) {
                  if ( (value == AR_TIMETAG_INT_USEC_64BIT)   ||
                       (value == AR_TIMETAG_HOST_USEC_64BIT)  ||
                       (value == AR_TIMETAG_INT_20USEC_32BIT) ||
                       (value == AR_TIMETAG_INT_MSEC_32BIT)   ||
                       (value == AR_TIMETAG_HOST_SEC_32BIT)   ||
                       (value == AR_TIMETAG_SYNC_1553_CH1)    ||
                       (value == AR_TIMETAG_SYNC_1553_CH2)    ||
                       (value == AR_TIMETAG_SYNC_1553_CH3)    ||
                       (value == AR_TIMETAG_SYNC_1553_CH4) ) {
                     /* first check if this is a 1553 channel time synchronization request, and assure the requested channel is installed */
                     if ((value >= AR_TIMETAG_SYNC_1553_CH1) && (value <= AR_TIMETAG_SYNC_1553_CH4)) {
                        if ((value - AR_TIMETAG_SYNC_1553_CH1) < (CDEV_PARM_SSI_TYPE)deviceGlobalStructPtr->comboBoard1553chanCount) {
                           /* if so, assure the 1553 Channel is initialized via the on-board 1553 channel shared memory region */
                           memOffset = CEIDEV_RAR15_SM_1553_CH_OFFSET + (((CEI_UINT32)value - AR_TIMETAG_SYNC_1553_CH1) * CEIDEV_RAR15_SM_1553_CH_SIZE);
                           accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->baseAddr[memOffset], &bitFieldValue);
                           status = (bitFieldValue > 0) ? (CDEV_API_RET_TYPE)EXIT_ROUTINE : (CDEV_API_RET_TYPE)ARS_INVHARVAL;
                        }
                        else {
                           status = ARS_INVHARVAL;
                        }
                     }
                     else {
                        status = EXIT_ROUTINE;
                     }
                  }
                  else {
                     status = ARS_INVHARVAL;
                  }
               }
               else if (value < AR_TIMER_X20_COMPAT_32BIT) {
                  status = EXIT_ROUTINE;
               }
               else {
                  status = ARS_INVHARVAL;
               }
      
               /* after all that time-tag mode validation, status is either ok-EXIT or INVALID, if ok, allow the selection */
               if (status == EXIT_ROUTINE) {
                  deviceGlobalStructPtr->deviceTimeTagReference = (CEI_UINT32)value;
               }
               break;
      
#ifdef INCLUDE_ARINC_717
            case ARU_573_RX_ENABLE:
            case ARU_573_RX_AUTO_DETECT:
            case ARU_573_RX_BPRZ_SELECT:
            case ARU_573_RX_SUBFRAME_SIZE:
            case ARU_573_TX_SLEW_RATE:
            case ARU_573_SYNC_WORD1:
            case ARU_573_SYNC_WORD2:
            case ARU_573_SYNC_WORD3:
            case ARU_573_SYNC_WORD4:
               /* if a 573 item was referenced, let the 573 Set Config routine handle it */
               status = ar_set_573_config (board, item, value);
               if (status == ARS_NORMAL) {
                  status = EXIT_ROUTINE;
               }
               break;
#endif
      
            /* configure a programmable channel as either a transmitter or receiver */
            case ARU_CONFIG_PROGRAMMABLE_CHAN:
               deviceChannel = (CEI_UINT32)(channel & 0xFF);
               accessStatus = a429_utl_configure_programmable_channel (board, deviceChannel, (CEI_UINT32)(value & 0xFF));
               if (accessStatus == ARS_NORMAL) {
                  status = EXIT_ROUTINE;
               }
               else {
                  status = (CDEV_API_RET_TYPE)accessStatus;
               }
               break;

            default : status = ARS_INVARG; 
               break;
         }

      }
      
#ifndef TARGET_EMBEDDED
      /* validate the specified channel */
      if (status == ARS_NORMAL) {
         status = a429_utl_validate_channel_request (board, deviceChannel, (CDEV_PARM_SSI_TYPE)item);
      }
#endif
   
      if (status == ARS_NORMAL) {
         /* read the device register with respect to the attribute specified */
         memOffset = CEIDEV_CHANNEL_REGION_START + offsetIntoChannelWindow + (deviceChannel * CEIDEV_CHANNEL_REGION_SIZE);
         accessStatus = cei_read_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], &bitFieldValue);
   
         if (accessStatus == ARS_NORMAL) {
   
            /* now adjust the respective bit field and output the new configuration */
            switch (item) {
               case ARU_RECV_MODE:
               case ARU_RX_PARITY:
               case ARU_RX_BITRATE:
#ifndef CERTIFICATION_BUILD
               case ARU_RX_MERGED_MODE:
               case ARU_RX_MSG_SIZE_24BIT:
#endif
                  if (item == ARU_RECV_MODE) {
                     /* When assigning certain Receive Configuration Register attributes, force an Rx channel FIFO reset prior 
                      * to modification of attribute, then restore the FIFO enable state afterwards */
                     accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], (bitFieldValue & CEIDEV_CHANNEL_ENABLE_MASK));
                     if (accessStatus != ARS_NORMAL) {
                        break;
                     }
                     
                     if (value == AR_WRAP_ON) {
                        bitFieldValue |= CEIDEV_INTERNAL_WRAP_ENABLE;
                     }
                     else if (value == AR_WRAP_OFF) {
                        bitFieldValue &= CEIDEV_INTERNAL_WRAP_MASK;
                     }
                     else {
                        status = ARS_INVHARVAL;
                     }
                  }
                  else if (item == ARU_RX_BITRATE) {
                     /* flag default ARINC 429 bus speeds with API constant values, otherwise it's custom, diff clock value for RAR15 */
                     if ((deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15) || (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15XT)) {
                        if (value == ARU_SPEED_HIGH) {
                           tempValue = CEIDEV_20MHZ_BAUD_RATE_100K;
                        }
                        else if (value == ARU_SPEED_LOW) {
                           tempValue = CEIDEV_20MHZ_BAUD_RATE_12pt5K;
                        }
                        else /* non-standard baud rates are allowed, where rate = 20,000,000 / value+2 */ {
                           tempValue = value;
                        }
                     }
                     else {
                        if (value == ARU_SPEED_HIGH) {
                           tempValue = CEIDEV_BAUD_RATE_100K;
                        }
                        else if (value == ARU_SPEED_LOW) {
                           tempValue = CEIDEV_BAUD_RATE_12pt5K;
                        }
                        else /* non-standard baud rates are allowed, where rate = 16,000,000 / value+2 */ {
                           tempValue = value;
                        }
                     }
                     if ((bitFieldValue & CEIDEV_BAUD_RATE_FIELD) != (tempValue << CEIDEV_BAUD_RATE_SHIFT)) {
                        /* When assigning certain Receive Configuration Register attributes, force an Rx channel FIFO reset prior 
                         * to modification of attribute, then restore the FIFO enable state afterwards */
                        accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], (bitFieldValue & CEIDEV_CHANNEL_ENABLE_MASK));
                        if (accessStatus == ARS_NORMAL) {
                           bitFieldValue &= CEIDEV_BAUD_RATE_MASK;
                           bitFieldValue |= (tempValue << CEIDEV_BAUD_RATE_SHIFT) & CEIDEV_BAUD_RATE_FIELD;
                        }
                     }
                  }
#ifndef CERTIFICATION_BUILD
                  else if (item == ARU_RX_MERGED_MODE) {
                     /* When assigning certain Receive Configuration Register attributes, force an Rx channel FIFO reset prior 
                      * to modification of attribute, then restore the FIFO enable state afterwards */
                     accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], (bitFieldValue & CEIDEV_CHANNEL_ENABLE_MASK));
                     if (accessStatus != ARS_NORMAL) {
                        break;
                     }
                     
                     if (value == AR_ON) {
                        deviceGlobalStructPtr->receiveChannelBufferMode[deviceChannel] = ARU_MERGED;
                        bitFieldValue |= CEIDEV_MERGE_MODE_ENABLE;
                     }
                     else if (value == AR_OFF) {
                        deviceGlobalStructPtr->receiveChannelBufferMode[deviceChannel] = ARU_BUFFERED;
                        bitFieldValue &= CEIDEV_MERGE_MODE_ENABLE_MASK;
                     }
                     else {
                        status = ARS_INVHARVAL;
                     }
                  }
                  else if (item == ARU_RX_MSG_SIZE_24BIT) {
                     /* When assigning certain Receive Configuration Register attributes, force an Rx channel FIFO reset prior 
                      * to modification of attribute, then restore the FIFO enable state afterwards */
                     accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], (bitFieldValue & CEIDEV_CHANNEL_ENABLE_MASK));
                     if (accessStatus != ARS_NORMAL) {
                        break;
                     }
                     
                     if (value == AR_ON) {
                        bitFieldValue |= CEIDEV_MSG_LENGTH_24_BIT;
                     }
                     else if (value == AR_OFF) {
                        bitFieldValue &= CEIDEV_MSG_LENGTH_24_BIT_MASK;
                     }
                     else {
                        status = ARS_INVHARVAL;
                     }
                  }
#endif
                  else {   /* item == ARU_RX_PARITY */
                     if ( (value == ARU_PARITY_ODD) || (value == ARU_PARITY_EVEN) || (value == AR_ON) ) {
                        bitFieldValue |= CEIDEV_PARITY_ENABLE;
                     }
                     else if ( (value == ARU_PARITY_NONE) || (value == AR_OFF) || (value == AR_RAW) ) {
                        bitFieldValue &= CEIDEV_PARITY_MASK;
                     }
                     else {
                        status = ARS_INVHARVAL;
                     }
                  }
                  break;
      
               case ARU_RX_DISABLE:
                  if (value == AR_ON) {
                     bitFieldValue &= CEIDEV_CHANNEL_ENABLE_MASK;
                  }
                  else if (value == AR_OFF) {
                     bitFieldValue |= CEIDEV_CHANNEL_ENABLE;
                  }
                  else {
                     status = ARS_INVHARVAL;
                  }
                  break;
            
               case ARU_TX_BITRATE:
               case ARU_TX_PARITY:
               case ARU_FAST_SLEW_RATE:
#ifndef CERTIFICATION_BUILD
               case ARU_TX_BIT_ERROR:
               case ARU_TX_GAP_ERROR:
               case ARU_TX_MSG_SIZE_24BIT:
#endif
                  /* When assigning Transmit Configuration Register attributes, force a Tx channel FIFO disable
                   * prior to modification of attribute, then restore the FIFO enable state afterwards */
                  accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], (bitFieldValue & CEIDEV_CHANNEL_ENABLE_MASK));
                  if (accessStatus != ARS_NORMAL) {
                     break;
                  }
                  
                  if (item == ARU_TX_BITRATE) {
                     /* flag default ARINC 429 bus speeds with API constant values, otherwise it's custom, diff clock value for RAR15 */
                     if ((deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15) || (deviceGlobalStructPtr->configuration == CEIDEV_CONFIG_RAR15XT)) {
                        if (value == ARU_SPEED_HIGH) {
                           bitFieldValue |= CEIDEV_FAST_SLEW_RATE;
                           tempValue = CEIDEV_20MHZ_BAUD_RATE_100K;
                        }
                        else if (value == ARU_SPEED_LOW) {
                           bitFieldValue &= CEIDEV_SLEW_RATE_MASK;
                           tempValue = CEIDEV_20MHZ_BAUD_RATE_12pt5K;
                        }
                        else { /* non-standard baud rates are allowed, where rate = 20,000,000 / value+2 */
                           tempValue = value;
                           if (value < CEIDEV_20MHZ_BAUD_RATE_15K) {
                              bitFieldValue |= CEIDEV_FAST_SLEW_RATE;
                           }
                           else {
                              bitFieldValue &= CEIDEV_SLEW_RATE_MASK;
                           }
                        }
                     }
                     else {
                        if (value == ARU_SPEED_HIGH) {
                           bitFieldValue |= CEIDEV_FAST_SLEW_RATE;
                           tempValue = CEIDEV_BAUD_RATE_100K;
                        }
                        else if (value == ARU_SPEED_LOW) {
                           bitFieldValue &= CEIDEV_SLEW_RATE_MASK;
                           tempValue = CEIDEV_BAUD_RATE_12pt5K;
                        }
                        else { /* non-standard baud rates are allowed, where rate = 16,000,000 / value+2 */
                           tempValue = value;
                           if (value < CEIDEV_BAUD_RATE_15K) {
                              bitFieldValue |= CEIDEV_FAST_SLEW_RATE;
                           }
                           else {
                              bitFieldValue &= CEIDEV_SLEW_RATE_MASK;
                           }
                        }
                     }
                     bitFieldValue &= CEIDEV_BAUD_RATE_MASK;
                     bitFieldValue |= (tempValue << CEIDEV_BAUD_RATE_SHIFT) & CEIDEV_BAUD_RATE_FIELD;
                  }
                  else if (item == ARU_FAST_SLEW_RATE) {
                     if (value == AR_ON) {
                        bitFieldValue |= CEIDEV_FAST_SLEW_RATE;
                     }
                     else if (value == AR_OFF) {
                        bitFieldValue &= CEIDEV_SLEW_RATE_MASK;
                     }
                     else {
                        status = ARS_INVHARVAL;
                     }
                  }
#ifndef CERTIFICATION_BUILD
                  else if (item == ARU_TX_BIT_ERROR) {
                     if (bitFieldValue & CEIDEV_429_TXCFG_GAP_ERROR) {
                        bitFieldValue &= ~(CEIDEV_429_TXCFG_BIT_COUNT_LO | CEIDEV_429_TXCFG_BIT_COUNT_HI);
                     }
                     else {
                        bitFieldValue &= ~(CEIDEV_429_TXCFG_BIT_COUNT_LO | CEIDEV_429_TXCFG_BIT_COUNT_HI | CEIDEV_429_TXCFG_PARAM_ENABLE);
                     }
   
                     if (value == AR_LO) {
                        bitFieldValue |= CEIDEV_429_TXCFG_BIT_COUNT_LO + CEIDEV_429_TXCFG_PARAM_ENABLE;
                     }
                     else if (value == AR_HI) {
                        bitFieldValue |= CEIDEV_429_TXCFG_BIT_COUNT_HI + CEIDEV_429_TXCFG_PARAM_ENABLE;
                     }
                     else if (value != AR_OFF) {
                        status = ARS_INVHARVAL;
                     }
                  }
                  else if (item == ARU_TX_GAP_ERROR) {
                     if (bitFieldValue & (CEIDEV_429_TXCFG_BIT_COUNT_LO | CEIDEV_429_TXCFG_BIT_COUNT_HI)) {
                        bitFieldValue &= ~(CEIDEV_429_TXCFG_GAP_ERROR);
                     }
                     else {
                        bitFieldValue &= ~(CEIDEV_429_TXCFG_GAP_ERROR | CEIDEV_429_TXCFG_PARAM_ENABLE);
                     }
   
                     if (value == AR_ON) {
                        bitFieldValue |= CEIDEV_429_TXCFG_GAP_ERROR + CEIDEV_429_TXCFG_PARAM_ENABLE;
                     }
                     else if (value != AR_OFF) {
                        status = ARS_INVHARVAL;
                     }
                  }
                  else if (item == ARU_TX_MSG_SIZE_24BIT) {
                     if (value == AR_ON) {
                        bitFieldValue |= CEIDEV_MSG_LENGTH_24_BIT;
                     }
                     else if (value == AR_OFF) {
                        bitFieldValue &= CEIDEV_MSG_LENGTH_24_BIT_MASK;
                     }
                     else {
                        status = ARS_INVHARVAL;
                     }
                  }
#endif
                  else {   /* item == ARU_TX_PARITY */
                     bitFieldValue &= CEIDEV_PARITY_SELECT_MASK;
      
                     if ( (value == ARU_PARITY_NONE) || (value == AR_OFF) || (value == AR_RAW) ) {
                        bitFieldValue &= CEIDEV_PARITY_MASK;
                     }
                     else if (value == ARU_PARITY_ODD) {
                        bitFieldValue |= CEIDEV_PARITY_ENABLE;
                     }
                     else if (value == ARU_PARITY_EVEN) {
                        bitFieldValue |= CEIDEV_PARITY_SELECT_EVEN + CEIDEV_PARITY_ENABLE;
                     }
                     else {
                        status = ARS_INVHARVAL;
                     }
                  }
                  break;
      
               case ARU_TX_DISABLE:
                  if (value == AR_OFF) {
                     bitFieldValue &= CEIDEV_TRANSMIT_DISABLE_MASK;
                     if ( (deviceGlobalStructPtr->configuration != CEIDEV_CONFIG_RAR15) &&
                          (deviceGlobalStructPtr->configuration != CEIDEV_CONFIG_RAR_XMC) &&
                          (deviceGlobalStructPtr->configuration != CEIDEV_CONFIG_RAR15XT) ) {
                        bitFieldValue |= CEIDEV_CHANNEL_ENABLE;
                     }
                  }
                  else if (value == AR_ON) {
                     bitFieldValue |= CEIDEV_TRANSMIT_DISABLE;
                  }
                  else {
                     status = ARS_INVHARVAL;
                  }
                  break;
            
               case ARU_RX_FIFO_ENABLE:
               case ARU_TX_FIFO_ENABLE:
                  if (value == AR_OFF) {
                     bitFieldValue &= CEIDEV_CHANNEL_ENABLE_MASK;
                  }
                  else if (value == AR_ON) {
                     bitFieldValue |= CEIDEV_CHANNEL_ENABLE;
                  }
                  else {
                     status = ARS_INVHARVAL;
                  }
                  break;
            
               case ARU_RX_FIFO_RESET:
#if defined (INTEGRITY_PPC_SYNC)
                  #pragma asm
                  eieio
                  sync
                  #pragma endasm
#elif defined (PPC_SYNC)
                  IO_SYNC;
#endif
                  /* first clear the FIFO enable bit and write that to the receive config register, then setup for FIFO enable */
                  bitFieldValue &= CEIDEV_CHANNEL_ENABLE_MASK;
                  accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], bitFieldValue);
                  bitFieldValue |= CEIDEV_CHANNEL_ENABLE;
#if defined (INTEGRITY_PPC_SYNC)
                  #pragma asm
                  eieio
                  sync
                  #pragma endasm
#elif defined (PPC_SYNC)
                  IO_SYNC;
#endif
                  break;
#ifdef INCLUDE_DISCRETE_IO      
               case ARU_DISCRETE_OUT1:
               case ARU_DISCRETE_OUT2:
               case ARU_DISCRETE_OUT:
                     if ( (value == AR_ON) || (value == AR_HI) ) {
                        bitFieldValue = 0;
                     }
                     else if ( (value == AR_OFF) || (value == AR_LO) ) {
                        bitFieldValue = 1;
                     }
                     else {
                        status = ARS_INVHARVAL;
                     }
                  break;
      
               case ARU_DIFFERENTIAL_OUT:
                     if (value == AR_ON) {
                        bitFieldValue = CEIDEV_DIFF_OUT_ENABLE;
                     }
                     else if (value == AR_OFF) {
                        bitFieldValue = 0;
                     }
                     else if (value == AR_LO) {
                        bitFieldValue = 1;
                     }
                     else if (value == AR_HI) {
                        bitFieldValue = 0;
                     }
                     else {
                        status = ARS_INVHARVAL;
                     }
                  break;
#endif
      
               case ARU_RX_RESET:
               case ARU_TX_RESET:
                  bitFieldValue = (CEI_UINT32)value;
                  break;
            }
         }
   
         /* update the configuration register with the specified attribute setting */
         if (accessStatus == ARS_NORMAL) {
            accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[memOffset], bitFieldValue);
         }
         
         if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL)) {
            status = (CDEV_API_RET_TYPE)accessStatus;
         }
      }
   
#ifndef TARGET_EMBEDDED
      /* if multi-thread protection is active, free any acquired lock */
      if ((deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) && lockAcquired) {
         if (status == ARS_NORMAL) {
            status = (CDEV_API_RET_TYPE)cei_release_thread_lock(LOCK_IDENT);
         }
         else {
            tempStatus = cei_release_thread_lock(LOCK_IDENT);
         }
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif   
   if (status == EXIT_ROUTINE) {
      status = ARS_NORMAL;
   }

   if (tempStatus) ;
   return (status);
}


/*===================================================================================================================*
 * ENTRY POINT:                       A R _ S E T _ M U L T I T H R E A D _ P R O T E C T
 *===================================================================================================================*
 *
 * FUNCTION:    Enable/Disable Multi-thread Protection around device accesses
 * 
 * PARAMETERS:  CDEV_BOARD_TYPE    board  -- (input) board to reference (0 to 127)
 *              CDEV_PARM_SSI_TYPE state  -- (input) AR_ON / AR_OFF 
 *
 * GLOBALS:     device[]
 *              multiProcessProtect[]
 * 
 * RETURN VAL:  ARS_NORMAL       - Success
 *              ARS_INVARG       - state value not supported
 * 
 * DESCRIPTION: This function controls the execution of multi-thread protection around every global host interface
 *              firmware access in API routines.  AR_ON enables multi-thread protection, AR_OFF disables it.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_set_multithread_protect 
( 
   CDEV_BOARD_TYPE board,                                                   /* (input) board to reference (0 to 127) */ 
   CDEV_PARM_SSI_TYPE state                                                 /* (input) AR_ON / AR_OFF                */
)
{
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                                 /* function return status  */

   if (board < MAX_BOARDS) {
      if (state == AR_OFF)  {
         multiProcessProtect[board] = (CEI_UINT32)AR_CONC_NONE;
         device[board].multiThreadProtect = (CEI_UINT32)AR_CONC_NONE;
      }
      else if (state == AR_ON) {
         multiProcessProtect[board] = (CEI_UINT32)AR_CONC_MULTITHRD;
         device[board].multiThreadProtect = (CEI_UINT32)AR_CONC_MULTITHRD;
      }
      else {
         status = ARS_INVARG;
      }
   }
   else {
      status = ARS_INVBOARD;
   }
   
   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                            A R _ S E T _ P R E L O A D _ C O N F I G
 *===================================================================================================================*
 *
 * FUNCTION:    Update a pre-load configuration setting.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input) board to reference (0 to 127)
 *           
 *              CDEV_PARM_SSI_TYPE item -- (input) the pre-load config item to define, the only valid selection being:
 *                                                 ARU_CONCURRENCY_MODE           
 *           
 *              CEI_UINT32 value        -- (input) the value to define the item valid selections from:
 *                                                 AR_CONC_NONE
 *                                                 AR_CONC_MULTITHRD
 *                                                 AR_CONC_MULTIPROC
 *           
 * GLOBALS:     device[]
 *              multiProcessProtect[]
 *           
 * RETURN VAL:  ARS_NORMAL   -- semaphore was opened successfully
 *              ARS_INVBOARD -- invalid board number 
 *              ARS_INVARG   -- invalid item or value parameter
 *              ARS_NO_OS_SUPPORT -- item not supported on host o/s
 *              ARS_BOARD_MUTEX -- failed to create/open the semaphore
 *
 * DESCRIPTION: Call this routine before calling AR_OPEN to update the value of a particular load configuration 
 *              setting.  This routine should not be called after the calling process has executed AR_OPEN.
 *
 *              If 'item' is ARU_CONCURRENCY_MODE, the 'value' parameter specifies the board concurrency mode.  One 
 *              of three modes may be selected: AR_CONC_NONE, AR_CONC_MULTITHRD, or AR_CONC_MULTIPROC.  Note that 
 *              some modes are only supported on certain boards and operating systems.  Refer to the routine
 *              AR_SET_PRELOAD_CONFIG documentation in the CEI-x30-SW Software User's Manual to determine if a 
 *              particular concurrency mode is supported on a certain board and operating system.
 *
 *              The default concurrency mode, AR_CONC_NONE, provides no multi-thread protection and no multi-process 
 *              support.  The user application must ensure that at most one thread is calling into the API at any  
 *              given time, and only a single process may interface with a particular board.
 *
 *              If AR_CONC_MULTITHRD concurrency mode is selected, thread protection is provided internally within the 
 *              API.  The user application may call into the API from multiple threads, but all threads must belong to 
 *              a single process.  The main user application thread should initialize the board with a call to
 *              AR_OPEN before other threads attempt to call into the API.
 *
 *              If AR_CONC_MULTIPROC concurrency mode is selected, thread protection is provided internally within the 
 *              API and multiple processes may interface with a single board.  If any process requests multi-process 
 *              mode, all other processes must also request multi-process mode.
 *
 *              Some coordination is required when using multiple processes in conjunction with a single device.  A
 *              single "board" process should be launched to configure board-specific and application-wide attributes 
 *              and invoke data processing using AR_GO.  After the "board" process launches the board it is then 
 *              permissible to launch one or more "channel" processes.  A "channel" process should first attach to the 
 *              board via AR_OPEN, then execute operations strictly confined to the particular channel(s) that process 
 *              is associated with (e.g. AR_GETWORD, AR_PUTWORD, AR_SET_CONFIG, etc.).  When finished, the "channel" 
 *              process should invoke AR_CLOSE and terminate. The "board" process should remain running until all 
 *              "channel" processes have terminated; afterwhich, it should invoke AR_CLOSE and terminate.
 *
 *              All processes must exercise caution when modifying board settings that may impact other concurrently
 *              running processes.
 *
 *              Note that board setup/initialization is only executed in AR_OPEN if no other processes have the board 
 *              open.  If another process has the board open (that is, if another process has opened the board using 
 *              AR_OPEN but hasn't yet closed the board using AR_CLOSE), AR_OPEN attaches to the device without 
 *              re-initializing the board or modifying board settings.  Similarly, AR_CLOSE only shuts down the board 
 *              if no other processes have the board open.  If another process has the board open, AR_CLOSE detaches 
 *              from the board without shutting it down.  Thus, board settings are preserved across "channel" 
 *              process invocations of AR_OPEN and AR_CLOSE.  Note the importance of matching every call to AR_OPEN 
 *              with a call to AR_CLOSE in multi-process mode.
 *              
 *              Multi-process mode is only required when accessing a single board from multiple processes.  If 
 *              multiple boards are installed, AR_CONC_NONE concurrency mode can be used as long as only one 
 *              process interfaces with each board.
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_set_preload_config
(
   CDEV_BOARD_TYPE board,                                                   /* (input) board to reference (0 to 127) */ 
   CDEV_PARM_SSI_TYPE item,                                                 /* (input) the item to define            */
   CEI_UINT32 value                                                         /* (input) the value to define the item  */
)
{
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                                 /* function return status  */

   /* make sure the board number is valid */
#ifndef TARGET_EMBEDDED
   if (board < MAX_BOARDS) {
#endif
      /* process the configuration command */
      if (item == ARU_CONCURRENCY_MODE) {
         switch (value) {
            case AR_CONC_MULTIPROC:
#if defined (VXW_PCI_PPC) || defined (VXW_PCI_X86) || defined (INTEGRITY_PCI_PPC) || defined (INTEGRITY_PCI_X86)
               status = ARS_NO_OS_SUPPORT;
               break;
#else
  #if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
    #ifdef MULTI_PROCESS_SUPPORT
               status = (CDEV_API_RET_TYPE)cei_open_preload_lock (board);
    #else
               status = ARS_NO_OS_SUPPORT;
               break;
    #endif
  #else
    #ifndef TARGET_EMBEDDED
               status = (CDEV_API_RET_TYPE)cei_open_preload_lock (board);
    #endif
  #endif
#endif
            case AR_CONC_MULTITHRD:
            case AR_CONC_NONE:
               device[board].multiThreadProtect = value;
               multiProcessProtect[board] = value;
               break;
            default:
               status = ARS_INVARG;
               break;
         }
      }
      else {
         status = ARS_INVARG;
      }
#ifndef TARGET_EMBEDDED
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                           A R _ S E T _ S T O R A G E _ M O D E
 *===================================================================================================================*
 *
 * FUNCTION:    Set board-level receive data storage mode (legacy API support)
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board   -- (input) board index (0 to 127)
 *              CDEV_PARM_SSI_TYPE mode -- (input) valid values are:
 *                                                 ARU_BUFFERED  individual buffer per channel
 *                                                 ARU_DEDICATED individual buffer per channel
 *                                                 ARU_MERGED    one merged circular buffer
 *
 * RETURN VAL:  ARS_NORMAL      Success
 *              ARS_INVARG      "mode" parameter value is not supported
 *              ARS_INVBOARD    Invalid board parameter or board not initialized
 *              ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine assigns the respective storage mode to all receive channels, all Merged or all Buffered.
 *
 *              This routine is provided for backward compatibility with the CEI-710 API.  The proper method to 
 *              assign the Merged vs Individual Buffer Receive Mode is provided for individual channels via the 
 *              respective channel set-config routine using the item parameter value ARU_RX_MERGED_MODE and its 
 *              ON/OFF selection.
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_set_storage_mode 
(  
   CDEV_BOARD_TYPE board,                                       /* (input) board index (0 to 127)                    */
   CDEV_PARM_SSI_TYPE mode                                      /* (input) storage mode (ARU_BUFFERED or ARU_MERGED) */
) 
{
   CDEV_API_RET_TYPE status = ARS_NORMAL;                                           /* function return status        */
   CEI_UINT32 channelIndex;                                                         /* local channel index           */
   CEI_INT32 mergedSelect = AR_ON;                                                  /* channel merged mode selection */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                                       /* ptr to board global structure */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      /* convert the specified buffer mode to ar_set_device_config options */
      if ((mode == ARU_BUFFERED) || (mode == ARU_DEDICATED)) {
         mergedSelect = AR_OFF;
      }
      else if (mode == ARU_MERGED) {
         mergedSelect = AR_ON;
      }
#ifndef TARGET_EMBEDDED
      else {
         status = ARS_INVARG;
      }

      if (status == ARS_NORMAL) {
#endif
         /* log the assigned receive mode in the API for reporting via ar_get_storage_mode */
         deviceGlobalStructPtr = deviceGlobal[board];
         deviceGlobalStructPtr->receiveMode = (CEI_UINT32)mode;

         /* now apply the requested storage mode to all receive channels, include any 573/717 channel if present */
         for (channelIndex = 0; channelIndex < deviceGlobalStructPtr->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]; channelIndex++) {
            ar_set_device_config (board, (CDEV_CHAN_TYPE)channelIndex, ARU_RX_MERGED_MODE, (CDEV_PARM_SSI_TYPE)mergedSelect);
         }
#ifdef INCLUDE_ARINC_717
         if ( deviceGlobalStructPtr->alternateJconfiguration == CEIDEV_ALT_CONFIGURATION_717 ) {
            ar_set_573_config (board, ARU_RX_MERGED_MODE, (CDEV_PARM_SI_TYPE)mergedSelect);
         }
#endif
#ifndef TARGET_EMBEDDED
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   return status;
}


/*===================================================================================================================*
 * ENTRY POINT:                                  A R _ S E T _ T I M E
 *===================================================================================================================*
 *
 * FUNCTION:    Assign an application-provided time value to the CEI-x30 internal 64-bit timer or the IRIG generator  
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board -- (input)  Board number (0 to 127)  
 * 
 *              pAR_TIMETAG_TYPE setTimeTag -- (input)
 *                               ->timeTagFormat -- selected timer to assign the value, this also defines the format
 *                                                  of the "timeTag" structure member value:
 *                                                     AR_TIMETAG_EXT_IRIG_64BIT
 *                                                     AR_TIMETAG_INT_USEC_64BIT
 *                             
 *                               ->timeTag       -- timer value to assign, with the format specified via the 
 *                                                  "timeTagFormat" structure member above, as follows:
 *                             
 *                                                     For a timeTagFormat of ...IRIG... the format of timeTag is a
 *                                                     BCD encoded value as defined in the CEI-x30 Hardware User 
 *                                                     Manual for the IRIG time register.
 *                             
 *                                                     For a timeTagFormat of ...USEC... the format of timeTag is a 
 *                                                     64-bit integer value having a 1 microsecond resolution.
 *                             
 *                               ->referenceTimeTag -- not used
 * 
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL      Success
 *                                ARS_INVBOARD    Invalid or uninitialized board
 *                                ARS_INVARG      Invalid timeTagFormat selection
 *                                ARS_INVHARVAL   Attempt to assign an IRIG time on a board not configured with IRIG
 *                                ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     device[]
 * 
 * DESCRIPTION: This routine assigns an application-provided time value to either the CEI-x30 internal 64-bit timer   
 *              or the IRIG generator, as specified by the timeTagFormat structure member.
 * 
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_set_time 
(
   CDEV_BOARD_TYPE board,                                                    /* (input) board index (0 to 127)       */
   pAR_TIMETAG_TYPE setTimeTag                                               /* (input) selected timer and value     */
)
{
   CEI_UINT32 timeTagHigh;
   CEI_UINT32 timeTagLow;
   CEI_INT32 accessStatus = ARS_NORMAL;                                      /* status of device read/write access   */
   VOLATILE ARINC_LOCAL_TYPE * deviceLocalStructPtr;                         /* pointer to board local structure     */
   ARINC_GLOBAL_TYPE * deviceGlobalStructPtr;                                /* pointer to board global structure    */
   
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      accessStatus = ARS_INVBOARD;
   }
   else {
#endif
      deviceLocalStructPtr = &device[board];
      deviceGlobalStructPtr = deviceGlobal[board];

      if (setTimeTag->timeTagFormat == (CEI_UINT32)AR_TIMETAG_INT_USEC_64BIT) {
         /* when assigning an internal 64-bit timer value, it must be parsed into 2 32-bit values */
         timeTagLow = (CEI_UINT32)(setTimeTag->timeTag);
         timeTagHigh = (CEI_UINT32)(setTimeTag->timeTag >> 32);

#ifndef TARGET_EMBEDDED
         /* if multi-thread protection is active, acquire a lock on the host interface */
         if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
            accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
         }
#endif
         /* when assigning an internal 64-bit timer value, it must be written to the timer set register in an LSW->MSW order */
         if (accessStatus == ARS_NORMAL) {
            accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_TIMER_LOW], timeTagLow);
            if (accessStatus == ARS_NORMAL) {
               accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_TIMER_HIGH], timeTagHigh);
            }
   
#ifndef TARGET_EMBEDDED
            /* if multi-thread protection is active, free any acquired lock */
            if (deviceLocalStructPtr->multiThreadProtect != AR_CONC_NONE) {
               accessStatus = cei_release_thread_lock(LOCK_IDENT);
            }
#endif
         }
      }
      else if (setTimeTag->timeTagFormat == AR_TIMETAG_EXT_IRIG_64BIT) {
         if (deviceGlobalStructPtr->irigInstalled) {
            /* when assigning an IRIG time value, the value is written to the IRIG set register as is, without validation */
            accessStatus = cei_write_32bit (board, &deviceLocalStructPtr->devicePtr[CEIDEV_IRIG_UPDATE], (CEI_UINT32)(setTimeTag->timeTag));
         }
         else {  /* board not configured with IRIG... flag attempt as error */
            accessStatus = ARS_INVHARVAL;
         }
      }
      else {
         accessStatus = ARS_INVARG;
      }

#ifndef TARGET_EMBEDDED
   }
#endif
   
   return (CDEV_API_RET_TYPE)accessStatus;
}


/*===================================================================================================================*
 * ENTRY POINT:                                       A R _ S L E E P
 *===================================================================================================================*
 *
 * FUNCTION:    Execute the o/s specific C library Sleep function
 *
 * PARAMETERS:  CEI_UINT32 sleep_ms    (i) The number of milliseconds to sleep
 *
 * DESCRIPTION: This function is used to invoke the current o/s supported C Sleep routine with the specified number 
 *              of milliseconds to sleep.  If the current o/s uses a task delay invocation instead of a sleep 
 *              routine, the sleep_ms will be converted based on that task delay duration resolution.
 *
 *===================================================================================================================*/
EXPORT32 CEI_VOID  DLL_EXPORTED ar_sleep (CEI_UINT32 sleep_ms)
{
   CEI_INT32 status;
   status = cei_delay_milliseconds(sleep_ms);
}


/*===================================================================================================================*
 * ENTRY POINT:                                       A R _ S T O P
 *===================================================================================================================*
 *
 * FUNCTION:    Disable the transmitter and receiver FIFOs on the device
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board  -- (input) board to reference (0 to 127)
 *
 * RETURN VAL:  CDEV_API_RET_TYPE ARS_NORMAL       Success
 *                                ARS_INVBOARD     Invalid board parameter or board not initialized
 *                                ARS_BOARD_MUTEX  Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     device[]
 * 
 * DESCRIPTION: Sets the global enable register global enable bit to DISABLE, disabling the Message Scheduler and all
 *              circular buffer to external bus operations.
 *
 *===================================================================================================================*/
EXPORT32 CDEV_API_RET_TYPE DLL_EXPORTED ar_stop
(
   CDEV_BOARD_TYPE board                                                   /* (input)  board to reference (0 to 127) */
)
{
   CEI_INT32 status = ARS_NORMAL;                                                      /* function return status     */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      /* disable "global enable" */
      status = cei_write_32bit (board, 
                                &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], 
                                CEIDEV_GLOBAL_ENABLE_AUTON_CLEAR + CEIDEV_GLOBAL_ENABLE_ON);
#ifndef TARGET_EMBEDDED
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   return (CDEV_API_RET_TYPE)status;
}


#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
/*===================================================================================================================*
 * ENTRY POINT:                         A R _ S E T _ B A S E _ A D D R E S S E S
 *===================================================================================================================*
 *
 * FUNCTION:    This routine provides a method to supply the BAR0 and BAR2 virtual address pointers to the base 
 *              address assigned to the device acquired from a custom embedded host device driver.
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board  -- (input) board to reference (0 to 127)
 *              CEI_ULONG  bar0Address -- (input) virtual address for BAR0
 *              CEI_ULONG  bar2Address -- (input) virtual address for BAR2
 *
 * RETURN VAL:  ARS_INVBOARD  board parm invalid
 * 
 * GLOBALS:     none
 * 
 * DESCRIPTION: This routine provides a method for any embedded platform application to bypass the standard device d
 *              river/host interface to acquire the device memory resources and supply the device BAR0 and BAR2 
 *              virtual base addresses to the API.
 * 
 *===================================================================================================================*/
extern CEI_VOID cei_assign_device_resources (CEI_INT16 deviceIndex, CEI_ULONG bar0Addr, CEI_ULONG bar2Addr);

EXPORT32 CEI_INT32 DLL_EXPORTED ar_set_base_addresses 
(
   CDEV_BOARD_TYPE board,                                                   /* (input) board to reference (0 to 127) */
   CEI_ULONG bar0Address,                                                   /* (input) virtual address for BAR0      */
   CEI_ULONG bar2Address                                                    /* (input) virtual address for BAR2      */
)
{
   CEI_INT32 status = ARS_NORMAL;
   if (board < MAX_BOARDS) {
      cei_assign_device_resources (board, bar0Address, bar2Address);
   }
   else {
      status = ARS_INVBOARD;
   }
   return status;
}

/*===================================================================================================================*
 * ENTRY POINT:                           A R _ S E T _ B O A R D _ I D E N T I T Y
 *===================================================================================================================*
 *
 * FUNCTION:    This routine provides a method to supply the PCI Configuration space attributes for Manufacturer and 
 *              Device ID when the API is used in conjunction with a user-supplied Linux driver
 *
 * PARAMETERS:  CDEV_BOARD_TYPE board     -- (input) board to reference (0 to 127)
 *              CEI_UINT32 manufacturerId -- (input) virtual address for BAR0
 *              CEI_UINT32 deviceId       -- (input) virtual address for BAR2
 *
 * RETURN VAL:  ARS_INVBOARD  board parm invalid
 * 
 * GLOBALS:     none
 * 
 * DESCRIPTION: This routine provides a method for any embedded platform application to bypass the standard device 
 *              driver/host interface to acquire the board's PCI Configuration space attributes for Manufacturer and 
 *              Device ID and assign these values from the application.
 * 
 *===================================================================================================================*/
extern CEI_VOID cei_assign_board_identity (CEI_INT16 deviceIndex, CEI_UINT32 manufacturerId, CEI_UINT32 deviceId);

EXPORT32 CEI_INT32 DLL_EXPORTED ar_set_board_identity
(
   CDEV_BOARD_TYPE board,                                                 /* (input) board to reference (0 to 127)   */
   CEI_UINT32 manufacturerId,                                             /* (input) PCI Manufacturer Identification */
   CEI_UINT32 deviceId                                                    /* (input) PCI Device Identification       */
)
{
   CEI_INT32 status = ARS_NORMAL;
   if (board < MAX_BOARDS) {
      cei_assign_board_identity (board, manufacturerId, deviceId);
   }
   else {
      status = ARS_INVBOARD;
   }
   return status;
}
#endif


